/* Generated by CCAN configurator */
#ifndef CCAN_CONFIG_H
#define CCAN_CONFIG_H
#ifndef _GNU_SOURCE
#define _GNU_SOURCE /* Always use GNU extensions. */
#endif
#define CCAN_COMPILER "gcc"
#define CCAN_CFLAGS "-std=gnu99"
#define CCAN_OUTPUT_EXE_CFLAG "-o"

#define HAVE_CCAN 1
#define HAVE_32BIT_OFF_T 1
#define HAVE_ALIGNOF 1
#define HAVE_ASPRINTF 1
#define HAVE_ATTRIBUTE_COLD 1
#define HAVE_ATTRIBUTE_CONST 1
#define HAVE_ATTRIBUTE_DEPRECATED 0
#define HAVE_ATTRIBUTE_NONNULL 1
#define HAVE_ATTRIBUTE_SENTINEL 1
#define HAVE_ATTRIBUTE_PURE 1
#define HAVE_ATTRIBUTE_MAY_ALIAS 1
#define HAVE_ATTRIBUTE_NORETURN 1
#define HAVE_ATTRIBUTE_PRINTF 1
#define HAVE_ATTRIBUTE_UNUSED 1
#define HAVE_ATTRIBUTE_USED 1
#define HAVE_BACKTRACE 1
#define HAVE_BIG_ENDIAN 0
#define HAVE_BSWAP_64 1
#define HAVE_BUILTIN_CHOOSE_EXPR 1
#define HAVE_BUILTIN_CLZ 1
#define HAVE_BUILTIN_CLZL 1
#define HAVE_BUILTIN_CLZLL 1
#define HAVE_BUILTIN_CTZ 1
#define HAVE_BUILTIN_CTZL 1
#define HAVE_BUILTIN_CTZLL 1
#define HAVE_BUILTIN_CONSTANT_P 1
#define HAVE_BUILTIN_EXPECT 1
#define HAVE_BUILTIN_FFS 1
#define HAVE_BUILTIN_FFSL 1
#define HAVE_BUILTIN_FFSLL 1
#define HAVE_BUILTIN_POPCOUNT 1
#define HAVE_BUILTIN_POPCOUNTL 1
#define HAVE_BUILTIN_POPCOUNTLL 1
#define HAVE_BUILTIN_TYPES_COMPATIBLE_P 1
#define HAVE_ICCARM_INTRINSICS 0
#define HAVE_BYTESWAP_H 1
#define HAVE_CLOCK_GETTIME 0
#define HAVE_CLOCK_GETTIME_IN_LIBRT 0
#define HAVE_COMPOUND_LITERALS 1
#define HAVE_FCHDIR 1
#define HAVE_ERR_H 1
#define HAVE_FILE_OFFSET_BITS 1
#define HAVE_FOR_LOOP_DECLARATION 1
#define HAVE_FLEXIBLE_ARRAY_MEMBER 1
#define HAVE_GETPAGESIZE 1
#define HAVE_ISBLANK 1
#define HAVE_LITTLE_ENDIAN 1
#define HAVE_MEMMEM 1
#define HAVE_MEMRCHR 1
#define HAVE_MMAP 1
#define HAVE_PROC_SELF_MAPS 1
#define HAVE_QSORT_R_PRIVATE_LAST 1
#define HAVE_STRUCT_TIMESPEC 1
#define HAVE_SECTION_START_STOP 1
#define HAVE_STACK_GROWS_UPWARDS 0
#define HAVE_STATEMENT_EXPR 1
#define HAVE_SYS_FILIO_H 0
#define HAVE_SYS_TERMIOS_H 1
#define HAVE_SYS_UNISTD_H 1
#define HAVE_TYPEOF 1
#define HAVE_UNALIGNED_ACCESS 1
#define HAVE_UTIME 1
#define HAVE_WARN_UNUSED_RESULT 1
#define HAVE_OPENMP 1
#define HAVE_VALGRIND_MEMCHECK_H 0
#define HAVE_UCONTEXT 1
#define HAVE_POINTER_SAFE_MAKECONTEXT 1
#endif /* CCAN_CONFIG_H */
