#ifndef ML99_NAT_DEC_H
#define ML99_NAT_DEC_H

#define ML99_PRIV_DEC(x)     ML99_PRIV_DEC_AUX(x)
#define ML99_PRIV_DEC_AUX(x) ML99_PRIV_DEC_##x

#define ML99_PRIV_DEC_0   255
#define ML99_PRIV_DEC_1   0
#define ML99_PRIV_DEC_2   1
#define ML99_PRIV_DEC_3   2
#define ML99_PRIV_DEC_4   3
#define ML99_PRIV_DEC_5   4
#define ML99_PRIV_DEC_6   5
#define ML99_PRIV_DEC_7   6
#define ML99_PRIV_DEC_8   7
#define ML99_PRIV_DEC_9   8
#define ML99_PRIV_DEC_10  9
#define ML99_PRIV_DEC_11  10
#define ML99_PRIV_DEC_12  11
#define ML99_PRIV_DEC_13  12
#define ML99_PRIV_DEC_14  13
#define ML99_PRIV_DEC_15  14
#define ML99_PRIV_DEC_16  15
#define ML99_PRIV_DEC_17  16
#define ML99_PRIV_DEC_18  17
#define ML99_PRIV_DEC_19  18
#define ML99_PRIV_DEC_20  19
#define ML99_PRIV_DEC_21  20
#define ML99_PRIV_DEC_22  21
#define ML99_PRIV_DEC_23  22
#define ML99_PRIV_DEC_24  23
#define ML99_PRIV_DEC_25  24
#define ML99_PRIV_DEC_26  25
#define ML99_PRIV_DEC_27  26
#define ML99_PRIV_DEC_28  27
#define ML99_PRIV_DEC_29  28
#define ML99_PRIV_DEC_30  29
#define ML99_PRIV_DEC_31  30
#define ML99_PRIV_DEC_32  31
#define ML99_PRIV_DEC_33  32
#define ML99_PRIV_DEC_34  33
#define ML99_PRIV_DEC_35  34
#define ML99_PRIV_DEC_36  35
#define ML99_PRIV_DEC_37  36
#define ML99_PRIV_DEC_38  37
#define ML99_PRIV_DEC_39  38
#define ML99_PRIV_DEC_40  39
#define ML99_PRIV_DEC_41  40
#define ML99_PRIV_DEC_42  41
#define ML99_PRIV_DEC_43  42
#define ML99_PRIV_DEC_44  43
#define ML99_PRIV_DEC_45  44
#define ML99_PRIV_DEC_46  45
#define ML99_PRIV_DEC_47  46
#define ML99_PRIV_DEC_48  47
#define ML99_PRIV_DEC_49  48
#define ML99_PRIV_DEC_50  49
#define ML99_PRIV_DEC_51  50
#define ML99_PRIV_DEC_52  51
#define ML99_PRIV_DEC_53  52
#define ML99_PRIV_DEC_54  53
#define ML99_PRIV_DEC_55  54
#define ML99_PRIV_DEC_56  55
#define ML99_PRIV_DEC_57  56
#define ML99_PRIV_DEC_58  57
#define ML99_PRIV_DEC_59  58
#define ML99_PRIV_DEC_60  59
#define ML99_PRIV_DEC_61  60
#define ML99_PRIV_DEC_62  61
#define ML99_PRIV_DEC_63  62
#define ML99_PRIV_DEC_64  63
#define ML99_PRIV_DEC_65  64
#define ML99_PRIV_DEC_66  65
#define ML99_PRIV_DEC_67  66
#define ML99_PRIV_DEC_68  67
#define ML99_PRIV_DEC_69  68
#define ML99_PRIV_DEC_70  69
#define ML99_PRIV_DEC_71  70
#define ML99_PRIV_DEC_72  71
#define ML99_PRIV_DEC_73  72
#define ML99_PRIV_DEC_74  73
#define ML99_PRIV_DEC_75  74
#define ML99_PRIV_DEC_76  75
#define ML99_PRIV_DEC_77  76
#define ML99_PRIV_DEC_78  77
#define ML99_PRIV_DEC_79  78
#define ML99_PRIV_DEC_80  79
#define ML99_PRIV_DEC_81  80
#define ML99_PRIV_DEC_82  81
#define ML99_PRIV_DEC_83  82
#define ML99_PRIV_DEC_84  83
#define ML99_PRIV_DEC_85  84
#define ML99_PRIV_DEC_86  85
#define ML99_PRIV_DEC_87  86
#define ML99_PRIV_DEC_88  87
#define ML99_PRIV_DEC_89  88
#define ML99_PRIV_DEC_90  89
#define ML99_PRIV_DEC_91  90
#define ML99_PRIV_DEC_92  91
#define ML99_PRIV_DEC_93  92
#define ML99_PRIV_DEC_94  93
#define ML99_PRIV_DEC_95  94
#define ML99_PRIV_DEC_96  95
#define ML99_PRIV_DEC_97  96
#define ML99_PRIV_DEC_98  97
#define ML99_PRIV_DEC_99  98
#define ML99_PRIV_DEC_100 99
#define ML99_PRIV_DEC_101 100
#define ML99_PRIV_DEC_102 101
#define ML99_PRIV_DEC_103 102
#define ML99_PRIV_DEC_104 103
#define ML99_PRIV_DEC_105 104
#define ML99_PRIV_DEC_106 105
#define ML99_PRIV_DEC_107 106
#define ML99_PRIV_DEC_108 107
#define ML99_PRIV_DEC_109 108
#define ML99_PRIV_DEC_110 109
#define ML99_PRIV_DEC_111 110
#define ML99_PRIV_DEC_112 111
#define ML99_PRIV_DEC_113 112
#define ML99_PRIV_DEC_114 113
#define ML99_PRIV_DEC_115 114
#define ML99_PRIV_DEC_116 115
#define ML99_PRIV_DEC_117 116
#define ML99_PRIV_DEC_118 117
#define ML99_PRIV_DEC_119 118
#define ML99_PRIV_DEC_120 119
#define ML99_PRIV_DEC_121 120
#define ML99_PRIV_DEC_122 121
#define ML99_PRIV_DEC_123 122
#define ML99_PRIV_DEC_124 123
#define ML99_PRIV_DEC_125 124
#define ML99_PRIV_DEC_126 125
#define ML99_PRIV_DEC_127 126
#define ML99_PRIV_DEC_128 127
#define ML99_PRIV_DEC_129 128
#define ML99_PRIV_DEC_130 129
#define ML99_PRIV_DEC_131 130
#define ML99_PRIV_DEC_132 131
#define ML99_PRIV_DEC_133 132
#define ML99_PRIV_DEC_134 133
#define ML99_PRIV_DEC_135 134
#define ML99_PRIV_DEC_136 135
#define ML99_PRIV_DEC_137 136
#define ML99_PRIV_DEC_138 137
#define ML99_PRIV_DEC_139 138
#define ML99_PRIV_DEC_140 139
#define ML99_PRIV_DEC_141 140
#define ML99_PRIV_DEC_142 141
#define ML99_PRIV_DEC_143 142
#define ML99_PRIV_DEC_144 143
#define ML99_PRIV_DEC_145 144
#define ML99_PRIV_DEC_146 145
#define ML99_PRIV_DEC_147 146
#define ML99_PRIV_DEC_148 147
#define ML99_PRIV_DEC_149 148
#define ML99_PRIV_DEC_150 149
#define ML99_PRIV_DEC_151 150
#define ML99_PRIV_DEC_152 151
#define ML99_PRIV_DEC_153 152
#define ML99_PRIV_DEC_154 153
#define ML99_PRIV_DEC_155 154
#define ML99_PRIV_DEC_156 155
#define ML99_PRIV_DEC_157 156
#define ML99_PRIV_DEC_158 157
#define ML99_PRIV_DEC_159 158
#define ML99_PRIV_DEC_160 159
#define ML99_PRIV_DEC_161 160
#define ML99_PRIV_DEC_162 161
#define ML99_PRIV_DEC_163 162
#define ML99_PRIV_DEC_164 163
#define ML99_PRIV_DEC_165 164
#define ML99_PRIV_DEC_166 165
#define ML99_PRIV_DEC_167 166
#define ML99_PRIV_DEC_168 167
#define ML99_PRIV_DEC_169 168
#define ML99_PRIV_DEC_170 169
#define ML99_PRIV_DEC_171 170
#define ML99_PRIV_DEC_172 171
#define ML99_PRIV_DEC_173 172
#define ML99_PRIV_DEC_174 173
#define ML99_PRIV_DEC_175 174
#define ML99_PRIV_DEC_176 175
#define ML99_PRIV_DEC_177 176
#define ML99_PRIV_DEC_178 177
#define ML99_PRIV_DEC_179 178
#define ML99_PRIV_DEC_180 179
#define ML99_PRIV_DEC_181 180
#define ML99_PRIV_DEC_182 181
#define ML99_PRIV_DEC_183 182
#define ML99_PRIV_DEC_184 183
#define ML99_PRIV_DEC_185 184
#define ML99_PRIV_DEC_186 185
#define ML99_PRIV_DEC_187 186
#define ML99_PRIV_DEC_188 187
#define ML99_PRIV_DEC_189 188
#define ML99_PRIV_DEC_190 189
#define ML99_PRIV_DEC_191 190
#define ML99_PRIV_DEC_192 191
#define ML99_PRIV_DEC_193 192
#define ML99_PRIV_DEC_194 193
#define ML99_PRIV_DEC_195 194
#define ML99_PRIV_DEC_196 195
#define ML99_PRIV_DEC_197 196
#define ML99_PRIV_DEC_198 197
#define ML99_PRIV_DEC_199 198
#define ML99_PRIV_DEC_200 199
#define ML99_PRIV_DEC_201 200
#define ML99_PRIV_DEC_202 201
#define ML99_PRIV_DEC_203 202
#define ML99_PRIV_DEC_204 203
#define ML99_PRIV_DEC_205 204
#define ML99_PRIV_DEC_206 205
#define ML99_PRIV_DEC_207 206
#define ML99_PRIV_DEC_208 207
#define ML99_PRIV_DEC_209 208
#define ML99_PRIV_DEC_210 209
#define ML99_PRIV_DEC_211 210
#define ML99_PRIV_DEC_212 211
#define ML99_PRIV_DEC_213 212
#define ML99_PRIV_DEC_214 213
#define ML99_PRIV_DEC_215 214
#define ML99_PRIV_DEC_216 215
#define ML99_PRIV_DEC_217 216
#define ML99_PRIV_DEC_218 217
#define ML99_PRIV_DEC_219 218
#define ML99_PRIV_DEC_220 219
#define ML99_PRIV_DEC_221 220
#define ML99_PRIV_DEC_222 221
#define ML99_PRIV_DEC_223 222
#define ML99_PRIV_DEC_224 223
#define ML99_PRIV_DEC_225 224
#define ML99_PRIV_DEC_226 225
#define ML99_PRIV_DEC_227 226
#define ML99_PRIV_DEC_228 227
#define ML99_PRIV_DEC_229 228
#define ML99_PRIV_DEC_230 229
#define ML99_PRIV_DEC_231 230
#define ML99_PRIV_DEC_232 231
#define ML99_PRIV_DEC_233 232
#define ML99_PRIV_DEC_234 233
#define ML99_PRIV_DEC_235 234
#define ML99_PRIV_DEC_236 235
#define ML99_PRIV_DEC_237 236
#define ML99_PRIV_DEC_238 237
#define ML99_PRIV_DEC_239 238
#define ML99_PRIV_DEC_240 239
#define ML99_PRIV_DEC_241 240
#define ML99_PRIV_DEC_242 241
#define ML99_PRIV_DEC_243 242
#define ML99_PRIV_DEC_244 243
#define ML99_PRIV_DEC_245 244
#define ML99_PRIV_DEC_246 245
#define ML99_PRIV_DEC_247 246
#define ML99_PRIV_DEC_248 247
#define ML99_PRIV_DEC_249 248
#define ML99_PRIV_DEC_250 249
#define ML99_PRIV_DEC_251 250
#define ML99_PRIV_DEC_252 251
#define ML99_PRIV_DEC_253 252
#define ML99_PRIV_DEC_254 253
#define ML99_PRIV_DEC_255 254

#endif // ML99_NAT_DEC_H
