#ifndef ML99_NAT_INC_H
#define ML99_NAT_INC_H

#define ML99_PRIV_INC(x)     ML99_PRIV_INC_AUX(x)
#define ML99_PRIV_INC_AUX(x) ML99_PRIV_INC_##x

#define ML99_PRIV_INC_0   1
#define ML99_PRIV_INC_1   2
#define ML99_PRIV_INC_2   3
#define ML99_PRIV_INC_3   4
#define ML99_PRIV_INC_4   5
#define ML99_PRIV_INC_5   6
#define ML99_PRIV_INC_6   7
#define ML99_PRIV_INC_7   8
#define ML99_PRIV_INC_8   9
#define ML99_PRIV_INC_9   10
#define ML99_PRIV_INC_10  11
#define ML99_PRIV_INC_11  12
#define ML99_PRIV_INC_12  13
#define ML99_PRIV_INC_13  14
#define ML99_PRIV_INC_14  15
#define ML99_PRIV_INC_15  16
#define ML99_PRIV_INC_16  17
#define ML99_PRIV_INC_17  18
#define ML99_PRIV_INC_18  19
#define ML99_PRIV_INC_19  20
#define ML99_PRIV_INC_20  21
#define ML99_PRIV_INC_21  22
#define ML99_PRIV_INC_22  23
#define ML99_PRIV_INC_23  24
#define ML99_PRIV_INC_24  25
#define ML99_PRIV_INC_25  26
#define ML99_PRIV_INC_26  27
#define ML99_PRIV_INC_27  28
#define ML99_PRIV_INC_28  29
#define ML99_PRIV_INC_29  30
#define ML99_PRIV_INC_30  31
#define ML99_PRIV_INC_31  32
#define ML99_PRIV_INC_32  33
#define ML99_PRIV_INC_33  34
#define ML99_PRIV_INC_34  35
#define ML99_PRIV_INC_35  36
#define ML99_PRIV_INC_36  37
#define ML99_PRIV_INC_37  38
#define ML99_PRIV_INC_38  39
#define ML99_PRIV_INC_39  40
#define ML99_PRIV_INC_40  41
#define ML99_PRIV_INC_41  42
#define ML99_PRIV_INC_42  43
#define ML99_PRIV_INC_43  44
#define ML99_PRIV_INC_44  45
#define ML99_PRIV_INC_45  46
#define ML99_PRIV_INC_46  47
#define ML99_PRIV_INC_47  48
#define ML99_PRIV_INC_48  49
#define ML99_PRIV_INC_49  50
#define ML99_PRIV_INC_50  51
#define ML99_PRIV_INC_51  52
#define ML99_PRIV_INC_52  53
#define ML99_PRIV_INC_53  54
#define ML99_PRIV_INC_54  55
#define ML99_PRIV_INC_55  56
#define ML99_PRIV_INC_56  57
#define ML99_PRIV_INC_57  58
#define ML99_PRIV_INC_58  59
#define ML99_PRIV_INC_59  60
#define ML99_PRIV_INC_60  61
#define ML99_PRIV_INC_61  62
#define ML99_PRIV_INC_62  63
#define ML99_PRIV_INC_63  64
#define ML99_PRIV_INC_64  65
#define ML99_PRIV_INC_65  66
#define ML99_PRIV_INC_66  67
#define ML99_PRIV_INC_67  68
#define ML99_PRIV_INC_68  69
#define ML99_PRIV_INC_69  70
#define ML99_PRIV_INC_70  71
#define ML99_PRIV_INC_71  72
#define ML99_PRIV_INC_72  73
#define ML99_PRIV_INC_73  74
#define ML99_PRIV_INC_74  75
#define ML99_PRIV_INC_75  76
#define ML99_PRIV_INC_76  77
#define ML99_PRIV_INC_77  78
#define ML99_PRIV_INC_78  79
#define ML99_PRIV_INC_79  80
#define ML99_PRIV_INC_80  81
#define ML99_PRIV_INC_81  82
#define ML99_PRIV_INC_82  83
#define ML99_PRIV_INC_83  84
#define ML99_PRIV_INC_84  85
#define ML99_PRIV_INC_85  86
#define ML99_PRIV_INC_86  87
#define ML99_PRIV_INC_87  88
#define ML99_PRIV_INC_88  89
#define ML99_PRIV_INC_89  90
#define ML99_PRIV_INC_90  91
#define ML99_PRIV_INC_91  92
#define ML99_PRIV_INC_92  93
#define ML99_PRIV_INC_93  94
#define ML99_PRIV_INC_94  95
#define ML99_PRIV_INC_95  96
#define ML99_PRIV_INC_96  97
#define ML99_PRIV_INC_97  98
#define ML99_PRIV_INC_98  99
#define ML99_PRIV_INC_99  100
#define ML99_PRIV_INC_100 101
#define ML99_PRIV_INC_101 102
#define ML99_PRIV_INC_102 103
#define ML99_PRIV_INC_103 104
#define ML99_PRIV_INC_104 105
#define ML99_PRIV_INC_105 106
#define ML99_PRIV_INC_106 107
#define ML99_PRIV_INC_107 108
#define ML99_PRIV_INC_108 109
#define ML99_PRIV_INC_109 110
#define ML99_PRIV_INC_110 111
#define ML99_PRIV_INC_111 112
#define ML99_PRIV_INC_112 113
#define ML99_PRIV_INC_113 114
#define ML99_PRIV_INC_114 115
#define ML99_PRIV_INC_115 116
#define ML99_PRIV_INC_116 117
#define ML99_PRIV_INC_117 118
#define ML99_PRIV_INC_118 119
#define ML99_PRIV_INC_119 120
#define ML99_PRIV_INC_120 121
#define ML99_PRIV_INC_121 122
#define ML99_PRIV_INC_122 123
#define ML99_PRIV_INC_123 124
#define ML99_PRIV_INC_124 125
#define ML99_PRIV_INC_125 126
#define ML99_PRIV_INC_126 127
#define ML99_PRIV_INC_127 128
#define ML99_PRIV_INC_128 129
#define ML99_PRIV_INC_129 130
#define ML99_PRIV_INC_130 131
#define ML99_PRIV_INC_131 132
#define ML99_PRIV_INC_132 133
#define ML99_PRIV_INC_133 134
#define ML99_PRIV_INC_134 135
#define ML99_PRIV_INC_135 136
#define ML99_PRIV_INC_136 137
#define ML99_PRIV_INC_137 138
#define ML99_PRIV_INC_138 139
#define ML99_PRIV_INC_139 140
#define ML99_PRIV_INC_140 141
#define ML99_PRIV_INC_141 142
#define ML99_PRIV_INC_142 143
#define ML99_PRIV_INC_143 144
#define ML99_PRIV_INC_144 145
#define ML99_PRIV_INC_145 146
#define ML99_PRIV_INC_146 147
#define ML99_PRIV_INC_147 148
#define ML99_PRIV_INC_148 149
#define ML99_PRIV_INC_149 150
#define ML99_PRIV_INC_150 151
#define ML99_PRIV_INC_151 152
#define ML99_PRIV_INC_152 153
#define ML99_PRIV_INC_153 154
#define ML99_PRIV_INC_154 155
#define ML99_PRIV_INC_155 156
#define ML99_PRIV_INC_156 157
#define ML99_PRIV_INC_157 158
#define ML99_PRIV_INC_158 159
#define ML99_PRIV_INC_159 160
#define ML99_PRIV_INC_160 161
#define ML99_PRIV_INC_161 162
#define ML99_PRIV_INC_162 163
#define ML99_PRIV_INC_163 164
#define ML99_PRIV_INC_164 165
#define ML99_PRIV_INC_165 166
#define ML99_PRIV_INC_166 167
#define ML99_PRIV_INC_167 168
#define ML99_PRIV_INC_168 169
#define ML99_PRIV_INC_169 170
#define ML99_PRIV_INC_170 171
#define ML99_PRIV_INC_171 172
#define ML99_PRIV_INC_172 173
#define ML99_PRIV_INC_173 174
#define ML99_PRIV_INC_174 175
#define ML99_PRIV_INC_175 176
#define ML99_PRIV_INC_176 177
#define ML99_PRIV_INC_177 178
#define ML99_PRIV_INC_178 179
#define ML99_PRIV_INC_179 180
#define ML99_PRIV_INC_180 181
#define ML99_PRIV_INC_181 182
#define ML99_PRIV_INC_182 183
#define ML99_PRIV_INC_183 184
#define ML99_PRIV_INC_184 185
#define ML99_PRIV_INC_185 186
#define ML99_PRIV_INC_186 187
#define ML99_PRIV_INC_187 188
#define ML99_PRIV_INC_188 189
#define ML99_PRIV_INC_189 190
#define ML99_PRIV_INC_190 191
#define ML99_PRIV_INC_191 192
#define ML99_PRIV_INC_192 193
#define ML99_PRIV_INC_193 194
#define ML99_PRIV_INC_194 195
#define ML99_PRIV_INC_195 196
#define ML99_PRIV_INC_196 197
#define ML99_PRIV_INC_197 198
#define ML99_PRIV_INC_198 199
#define ML99_PRIV_INC_199 200
#define ML99_PRIV_INC_200 201
#define ML99_PRIV_INC_201 202
#define ML99_PRIV_INC_202 203
#define ML99_PRIV_INC_203 204
#define ML99_PRIV_INC_204 205
#define ML99_PRIV_INC_205 206
#define ML99_PRIV_INC_206 207
#define ML99_PRIV_INC_207 208
#define ML99_PRIV_INC_208 209
#define ML99_PRIV_INC_209 210
#define ML99_PRIV_INC_210 211
#define ML99_PRIV_INC_211 212
#define ML99_PRIV_INC_212 213
#define ML99_PRIV_INC_213 214
#define ML99_PRIV_INC_214 215
#define ML99_PRIV_INC_215 216
#define ML99_PRIV_INC_216 217
#define ML99_PRIV_INC_217 218
#define ML99_PRIV_INC_218 219
#define ML99_PRIV_INC_219 220
#define ML99_PRIV_INC_220 221
#define ML99_PRIV_INC_221 222
#define ML99_PRIV_INC_222 223
#define ML99_PRIV_INC_223 224
#define ML99_PRIV_INC_224 225
#define ML99_PRIV_INC_225 226
#define ML99_PRIV_INC_226 227
#define ML99_PRIV_INC_227 228
#define ML99_PRIV_INC_228 229
#define ML99_PRIV_INC_229 230
#define ML99_PRIV_INC_230 231
#define ML99_PRIV_INC_231 232
#define ML99_PRIV_INC_232 233
#define ML99_PRIV_INC_233 234
#define ML99_PRIV_INC_234 235
#define ML99_PRIV_INC_235 236
#define ML99_PRIV_INC_236 237
#define ML99_PRIV_INC_237 238
#define ML99_PRIV_INC_238 239
#define ML99_PRIV_INC_239 240
#define ML99_PRIV_INC_240 241
#define ML99_PRIV_INC_241 242
#define ML99_PRIV_INC_242 243
#define ML99_PRIV_INC_243 244
#define ML99_PRIV_INC_244 245
#define ML99_PRIV_INC_245 246
#define ML99_PRIV_INC_246 247
#define ML99_PRIV_INC_247 248
#define ML99_PRIV_INC_248 249
#define ML99_PRIV_INC_249 250
#define ML99_PRIV_INC_250 251
#define ML99_PRIV_INC_251 252
#define ML99_PRIV_INC_252 253
#define ML99_PRIV_INC_253 254
#define ML99_PRIV_INC_254 255
#define ML99_PRIV_INC_255 0

#endif // ML99_NAT_INC_H
