#ifndef METALANG99_UINT_INC_H
#define METALANG99_UINT_INC_H

#include <metalang99/priv/util.h>

#define METALANG99_PRIV_inc(x) METALANG99_PRIV_PRIMITIVE_CAT(METALANG99_PRIV_inc_, x)()

#define METALANG99_PRIV_inc_0()   1
#define METALANG99_PRIV_inc_1()   2
#define METALANG99_PRIV_inc_2()   3
#define METALANG99_PRIV_inc_3()   4
#define METALANG99_PRIV_inc_4()   5
#define METALANG99_PRIV_inc_5()   6
#define METALANG99_PRIV_inc_6()   7
#define METALANG99_PRIV_inc_7()   8
#define METALANG99_PRIV_inc_8()   9
#define METALANG99_PRIV_inc_9()   10
#define METALANG99_PRIV_inc_10()  11
#define METALANG99_PRIV_inc_11()  12
#define METALANG99_PRIV_inc_12()  13
#define METALANG99_PRIV_inc_13()  14
#define METALANG99_PRIV_inc_14()  15
#define METALANG99_PRIV_inc_15()  16
#define METALANG99_PRIV_inc_16()  17
#define METALANG99_PRIV_inc_17()  18
#define METALANG99_PRIV_inc_18()  19
#define METALANG99_PRIV_inc_19()  20
#define METALANG99_PRIV_inc_20()  21
#define METALANG99_PRIV_inc_21()  22
#define METALANG99_PRIV_inc_22()  23
#define METALANG99_PRIV_inc_23()  24
#define METALANG99_PRIV_inc_24()  25
#define METALANG99_PRIV_inc_25()  26
#define METALANG99_PRIV_inc_26()  27
#define METALANG99_PRIV_inc_27()  28
#define METALANG99_PRIV_inc_28()  29
#define METALANG99_PRIV_inc_29()  30
#define METALANG99_PRIV_inc_30()  31
#define METALANG99_PRIV_inc_31()  32
#define METALANG99_PRIV_inc_32()  33
#define METALANG99_PRIV_inc_33()  34
#define METALANG99_PRIV_inc_34()  35
#define METALANG99_PRIV_inc_35()  36
#define METALANG99_PRIV_inc_36()  37
#define METALANG99_PRIV_inc_37()  38
#define METALANG99_PRIV_inc_38()  39
#define METALANG99_PRIV_inc_39()  40
#define METALANG99_PRIV_inc_40()  41
#define METALANG99_PRIV_inc_41()  42
#define METALANG99_PRIV_inc_42()  43
#define METALANG99_PRIV_inc_43()  44
#define METALANG99_PRIV_inc_44()  45
#define METALANG99_PRIV_inc_45()  46
#define METALANG99_PRIV_inc_46()  47
#define METALANG99_PRIV_inc_47()  48
#define METALANG99_PRIV_inc_48()  49
#define METALANG99_PRIV_inc_49()  50
#define METALANG99_PRIV_inc_50()  51
#define METALANG99_PRIV_inc_51()  52
#define METALANG99_PRIV_inc_52()  53
#define METALANG99_PRIV_inc_53()  54
#define METALANG99_PRIV_inc_54()  55
#define METALANG99_PRIV_inc_55()  56
#define METALANG99_PRIV_inc_56()  57
#define METALANG99_PRIV_inc_57()  58
#define METALANG99_PRIV_inc_58()  59
#define METALANG99_PRIV_inc_59()  60
#define METALANG99_PRIV_inc_60()  61
#define METALANG99_PRIV_inc_61()  62
#define METALANG99_PRIV_inc_62()  63
#define METALANG99_PRIV_inc_63()  64
#define METALANG99_PRIV_inc_64()  65
#define METALANG99_PRIV_inc_65()  66
#define METALANG99_PRIV_inc_66()  67
#define METALANG99_PRIV_inc_67()  68
#define METALANG99_PRIV_inc_68()  69
#define METALANG99_PRIV_inc_69()  70
#define METALANG99_PRIV_inc_70()  71
#define METALANG99_PRIV_inc_71()  72
#define METALANG99_PRIV_inc_72()  73
#define METALANG99_PRIV_inc_73()  74
#define METALANG99_PRIV_inc_74()  75
#define METALANG99_PRIV_inc_75()  76
#define METALANG99_PRIV_inc_76()  77
#define METALANG99_PRIV_inc_77()  78
#define METALANG99_PRIV_inc_78()  79
#define METALANG99_PRIV_inc_79()  80
#define METALANG99_PRIV_inc_80()  81
#define METALANG99_PRIV_inc_81()  82
#define METALANG99_PRIV_inc_82()  83
#define METALANG99_PRIV_inc_83()  84
#define METALANG99_PRIV_inc_84()  85
#define METALANG99_PRIV_inc_85()  86
#define METALANG99_PRIV_inc_86()  87
#define METALANG99_PRIV_inc_87()  88
#define METALANG99_PRIV_inc_88()  89
#define METALANG99_PRIV_inc_89()  90
#define METALANG99_PRIV_inc_90()  91
#define METALANG99_PRIV_inc_91()  92
#define METALANG99_PRIV_inc_92()  93
#define METALANG99_PRIV_inc_93()  94
#define METALANG99_PRIV_inc_94()  95
#define METALANG99_PRIV_inc_95()  96
#define METALANG99_PRIV_inc_96()  97
#define METALANG99_PRIV_inc_97()  98
#define METALANG99_PRIV_inc_98()  99
#define METALANG99_PRIV_inc_99()  100
#define METALANG99_PRIV_inc_100() 101
#define METALANG99_PRIV_inc_101() 102
#define METALANG99_PRIV_inc_102() 103
#define METALANG99_PRIV_inc_103() 104
#define METALANG99_PRIV_inc_104() 105
#define METALANG99_PRIV_inc_105() 106
#define METALANG99_PRIV_inc_106() 107
#define METALANG99_PRIV_inc_107() 108
#define METALANG99_PRIV_inc_108() 109
#define METALANG99_PRIV_inc_109() 110
#define METALANG99_PRIV_inc_110() 111
#define METALANG99_PRIV_inc_111() 112
#define METALANG99_PRIV_inc_112() 113
#define METALANG99_PRIV_inc_113() 114
#define METALANG99_PRIV_inc_114() 115
#define METALANG99_PRIV_inc_115() 116
#define METALANG99_PRIV_inc_116() 117
#define METALANG99_PRIV_inc_117() 118
#define METALANG99_PRIV_inc_118() 119
#define METALANG99_PRIV_inc_119() 120
#define METALANG99_PRIV_inc_120() 121
#define METALANG99_PRIV_inc_121() 122
#define METALANG99_PRIV_inc_122() 123
#define METALANG99_PRIV_inc_123() 124
#define METALANG99_PRIV_inc_124() 125
#define METALANG99_PRIV_inc_125() 126
#define METALANG99_PRIV_inc_126() 127
#define METALANG99_PRIV_inc_127() 128
#define METALANG99_PRIV_inc_128() 129
#define METALANG99_PRIV_inc_129() 130
#define METALANG99_PRIV_inc_130() 131
#define METALANG99_PRIV_inc_131() 132
#define METALANG99_PRIV_inc_132() 133
#define METALANG99_PRIV_inc_133() 134
#define METALANG99_PRIV_inc_134() 135
#define METALANG99_PRIV_inc_135() 136
#define METALANG99_PRIV_inc_136() 137
#define METALANG99_PRIV_inc_137() 138
#define METALANG99_PRIV_inc_138() 139
#define METALANG99_PRIV_inc_139() 140
#define METALANG99_PRIV_inc_140() 141
#define METALANG99_PRIV_inc_141() 142
#define METALANG99_PRIV_inc_142() 143
#define METALANG99_PRIV_inc_143() 144
#define METALANG99_PRIV_inc_144() 145
#define METALANG99_PRIV_inc_145() 146
#define METALANG99_PRIV_inc_146() 147
#define METALANG99_PRIV_inc_147() 148
#define METALANG99_PRIV_inc_148() 149
#define METALANG99_PRIV_inc_149() 150
#define METALANG99_PRIV_inc_150() 151
#define METALANG99_PRIV_inc_151() 152
#define METALANG99_PRIV_inc_152() 153
#define METALANG99_PRIV_inc_153() 154
#define METALANG99_PRIV_inc_154() 155
#define METALANG99_PRIV_inc_155() 156
#define METALANG99_PRIV_inc_156() 157
#define METALANG99_PRIV_inc_157() 158
#define METALANG99_PRIV_inc_158() 159
#define METALANG99_PRIV_inc_159() 160
#define METALANG99_PRIV_inc_160() 161
#define METALANG99_PRIV_inc_161() 162
#define METALANG99_PRIV_inc_162() 163
#define METALANG99_PRIV_inc_163() 164
#define METALANG99_PRIV_inc_164() 165
#define METALANG99_PRIV_inc_165() 166
#define METALANG99_PRIV_inc_166() 167
#define METALANG99_PRIV_inc_167() 168
#define METALANG99_PRIV_inc_168() 169
#define METALANG99_PRIV_inc_169() 170
#define METALANG99_PRIV_inc_170() 171
#define METALANG99_PRIV_inc_171() 172
#define METALANG99_PRIV_inc_172() 173
#define METALANG99_PRIV_inc_173() 174
#define METALANG99_PRIV_inc_174() 175
#define METALANG99_PRIV_inc_175() 176
#define METALANG99_PRIV_inc_176() 177
#define METALANG99_PRIV_inc_177() 178
#define METALANG99_PRIV_inc_178() 179
#define METALANG99_PRIV_inc_179() 180
#define METALANG99_PRIV_inc_180() 181
#define METALANG99_PRIV_inc_181() 182
#define METALANG99_PRIV_inc_182() 183
#define METALANG99_PRIV_inc_183() 184
#define METALANG99_PRIV_inc_184() 185
#define METALANG99_PRIV_inc_185() 186
#define METALANG99_PRIV_inc_186() 187
#define METALANG99_PRIV_inc_187() 188
#define METALANG99_PRIV_inc_188() 189
#define METALANG99_PRIV_inc_189() 190
#define METALANG99_PRIV_inc_190() 191
#define METALANG99_PRIV_inc_191() 192
#define METALANG99_PRIV_inc_192() 193
#define METALANG99_PRIV_inc_193() 194
#define METALANG99_PRIV_inc_194() 195
#define METALANG99_PRIV_inc_195() 196
#define METALANG99_PRIV_inc_196() 197
#define METALANG99_PRIV_inc_197() 198
#define METALANG99_PRIV_inc_198() 199
#define METALANG99_PRIV_inc_199() 200
#define METALANG99_PRIV_inc_200() 201
#define METALANG99_PRIV_inc_201() 202
#define METALANG99_PRIV_inc_202() 203
#define METALANG99_PRIV_inc_203() 204
#define METALANG99_PRIV_inc_204() 205
#define METALANG99_PRIV_inc_205() 206
#define METALANG99_PRIV_inc_206() 207
#define METALANG99_PRIV_inc_207() 208
#define METALANG99_PRIV_inc_208() 209
#define METALANG99_PRIV_inc_209() 210
#define METALANG99_PRIV_inc_210() 211
#define METALANG99_PRIV_inc_211() 212
#define METALANG99_PRIV_inc_212() 213
#define METALANG99_PRIV_inc_213() 214
#define METALANG99_PRIV_inc_214() 215
#define METALANG99_PRIV_inc_215() 216
#define METALANG99_PRIV_inc_216() 217
#define METALANG99_PRIV_inc_217() 218
#define METALANG99_PRIV_inc_218() 219
#define METALANG99_PRIV_inc_219() 220
#define METALANG99_PRIV_inc_220() 221
#define METALANG99_PRIV_inc_221() 222
#define METALANG99_PRIV_inc_222() 223
#define METALANG99_PRIV_inc_223() 224
#define METALANG99_PRIV_inc_224() 225
#define METALANG99_PRIV_inc_225() 226
#define METALANG99_PRIV_inc_226() 227
#define METALANG99_PRIV_inc_227() 228
#define METALANG99_PRIV_inc_228() 229
#define METALANG99_PRIV_inc_229() 230
#define METALANG99_PRIV_inc_230() 231
#define METALANG99_PRIV_inc_231() 232
#define METALANG99_PRIV_inc_232() 233
#define METALANG99_PRIV_inc_233() 234
#define METALANG99_PRIV_inc_234() 235
#define METALANG99_PRIV_inc_235() 236
#define METALANG99_PRIV_inc_236() 237
#define METALANG99_PRIV_inc_237() 238
#define METALANG99_PRIV_inc_238() 239
#define METALANG99_PRIV_inc_239() 240
#define METALANG99_PRIV_inc_240() 241
#define METALANG99_PRIV_inc_241() 242
#define METALANG99_PRIV_inc_242() 243
#define METALANG99_PRIV_inc_243() 244
#define METALANG99_PRIV_inc_244() 245
#define METALANG99_PRIV_inc_245() 246
#define METALANG99_PRIV_inc_246() 247
#define METALANG99_PRIV_inc_247() 248
#define METALANG99_PRIV_inc_248() 249
#define METALANG99_PRIV_inc_249() 250
#define METALANG99_PRIV_inc_250() 251
#define METALANG99_PRIV_inc_251() 252
#define METALANG99_PRIV_inc_252() 253
#define METALANG99_PRIV_inc_253() 254
#define METALANG99_PRIV_inc_254() 255
#define METALANG99_PRIV_inc_255() 0

#endif // METALANG99_UINT_INC_H
