#ifndef METALANG99_UINT_EQ_H
#define METALANG99_UINT_EQ_H

#include <metalang99/priv/variadics/get.h>

#define METALANG99_PRIV_uintEq(x, y) METALANG99_PRIV_uintEq_AUX(x, y)
#define METALANG99_PRIV_uintEq_AUX(x, y)                                                           \
    METALANG99_PRIV_VARIADICS_SND(METALANG99_PRIV_uintEq_##x##_##y(), 0)

#define METALANG99_PRIV_uintEq_0_0()     ~, 1
#define METALANG99_PRIV_uintEq_1_1()     ~, 1
#define METALANG99_PRIV_uintEq_2_2()     ~, 1
#define METALANG99_PRIV_uintEq_3_3()     ~, 1
#define METALANG99_PRIV_uintEq_4_4()     ~, 1
#define METALANG99_PRIV_uintEq_5_5()     ~, 1
#define METALANG99_PRIV_uintEq_6_6()     ~, 1
#define METALANG99_PRIV_uintEq_7_7()     ~, 1
#define METALANG99_PRIV_uintEq_8_8()     ~, 1
#define METALANG99_PRIV_uintEq_9_9()     ~, 1
#define METALANG99_PRIV_uintEq_10_10()   ~, 1
#define METALANG99_PRIV_uintEq_11_11()   ~, 1
#define METALANG99_PRIV_uintEq_12_12()   ~, 1
#define METALANG99_PRIV_uintEq_13_13()   ~, 1
#define METALANG99_PRIV_uintEq_14_14()   ~, 1
#define METALANG99_PRIV_uintEq_15_15()   ~, 1
#define METALANG99_PRIV_uintEq_16_16()   ~, 1
#define METALANG99_PRIV_uintEq_17_17()   ~, 1
#define METALANG99_PRIV_uintEq_18_18()   ~, 1
#define METALANG99_PRIV_uintEq_19_19()   ~, 1
#define METALANG99_PRIV_uintEq_20_20()   ~, 1
#define METALANG99_PRIV_uintEq_21_21()   ~, 1
#define METALANG99_PRIV_uintEq_22_22()   ~, 1
#define METALANG99_PRIV_uintEq_23_23()   ~, 1
#define METALANG99_PRIV_uintEq_24_24()   ~, 1
#define METALANG99_PRIV_uintEq_25_25()   ~, 1
#define METALANG99_PRIV_uintEq_26_26()   ~, 1
#define METALANG99_PRIV_uintEq_27_27()   ~, 1
#define METALANG99_PRIV_uintEq_28_28()   ~, 1
#define METALANG99_PRIV_uintEq_29_29()   ~, 1
#define METALANG99_PRIV_uintEq_30_30()   ~, 1
#define METALANG99_PRIV_uintEq_31_31()   ~, 1
#define METALANG99_PRIV_uintEq_32_32()   ~, 1
#define METALANG99_PRIV_uintEq_33_33()   ~, 1
#define METALANG99_PRIV_uintEq_34_34()   ~, 1
#define METALANG99_PRIV_uintEq_35_35()   ~, 1
#define METALANG99_PRIV_uintEq_36_36()   ~, 1
#define METALANG99_PRIV_uintEq_37_37()   ~, 1
#define METALANG99_PRIV_uintEq_38_38()   ~, 1
#define METALANG99_PRIV_uintEq_39_39()   ~, 1
#define METALANG99_PRIV_uintEq_40_40()   ~, 1
#define METALANG99_PRIV_uintEq_41_41()   ~, 1
#define METALANG99_PRIV_uintEq_42_42()   ~, 1
#define METALANG99_PRIV_uintEq_43_43()   ~, 1
#define METALANG99_PRIV_uintEq_44_44()   ~, 1
#define METALANG99_PRIV_uintEq_45_45()   ~, 1
#define METALANG99_PRIV_uintEq_46_46()   ~, 1
#define METALANG99_PRIV_uintEq_47_47()   ~, 1
#define METALANG99_PRIV_uintEq_48_48()   ~, 1
#define METALANG99_PRIV_uintEq_49_49()   ~, 1
#define METALANG99_PRIV_uintEq_50_50()   ~, 1
#define METALANG99_PRIV_uintEq_51_51()   ~, 1
#define METALANG99_PRIV_uintEq_52_52()   ~, 1
#define METALANG99_PRIV_uintEq_53_53()   ~, 1
#define METALANG99_PRIV_uintEq_54_54()   ~, 1
#define METALANG99_PRIV_uintEq_55_55()   ~, 1
#define METALANG99_PRIV_uintEq_56_56()   ~, 1
#define METALANG99_PRIV_uintEq_57_57()   ~, 1
#define METALANG99_PRIV_uintEq_58_58()   ~, 1
#define METALANG99_PRIV_uintEq_59_59()   ~, 1
#define METALANG99_PRIV_uintEq_60_60()   ~, 1
#define METALANG99_PRIV_uintEq_61_61()   ~, 1
#define METALANG99_PRIV_uintEq_62_62()   ~, 1
#define METALANG99_PRIV_uintEq_63_63()   ~, 1
#define METALANG99_PRIV_uintEq_64_64()   ~, 1
#define METALANG99_PRIV_uintEq_65_65()   ~, 1
#define METALANG99_PRIV_uintEq_66_66()   ~, 1
#define METALANG99_PRIV_uintEq_67_67()   ~, 1
#define METALANG99_PRIV_uintEq_68_68()   ~, 1
#define METALANG99_PRIV_uintEq_69_69()   ~, 1
#define METALANG99_PRIV_uintEq_70_70()   ~, 1
#define METALANG99_PRIV_uintEq_71_71()   ~, 1
#define METALANG99_PRIV_uintEq_72_72()   ~, 1
#define METALANG99_PRIV_uintEq_73_73()   ~, 1
#define METALANG99_PRIV_uintEq_74_74()   ~, 1
#define METALANG99_PRIV_uintEq_75_75()   ~, 1
#define METALANG99_PRIV_uintEq_76_76()   ~, 1
#define METALANG99_PRIV_uintEq_77_77()   ~, 1
#define METALANG99_PRIV_uintEq_78_78()   ~, 1
#define METALANG99_PRIV_uintEq_79_79()   ~, 1
#define METALANG99_PRIV_uintEq_80_80()   ~, 1
#define METALANG99_PRIV_uintEq_81_81()   ~, 1
#define METALANG99_PRIV_uintEq_82_82()   ~, 1
#define METALANG99_PRIV_uintEq_83_83()   ~, 1
#define METALANG99_PRIV_uintEq_84_84()   ~, 1
#define METALANG99_PRIV_uintEq_85_85()   ~, 1
#define METALANG99_PRIV_uintEq_86_86()   ~, 1
#define METALANG99_PRIV_uintEq_87_87()   ~, 1
#define METALANG99_PRIV_uintEq_88_88()   ~, 1
#define METALANG99_PRIV_uintEq_89_89()   ~, 1
#define METALANG99_PRIV_uintEq_90_90()   ~, 1
#define METALANG99_PRIV_uintEq_91_91()   ~, 1
#define METALANG99_PRIV_uintEq_92_92()   ~, 1
#define METALANG99_PRIV_uintEq_93_93()   ~, 1
#define METALANG99_PRIV_uintEq_94_94()   ~, 1
#define METALANG99_PRIV_uintEq_95_95()   ~, 1
#define METALANG99_PRIV_uintEq_96_96()   ~, 1
#define METALANG99_PRIV_uintEq_97_97()   ~, 1
#define METALANG99_PRIV_uintEq_98_98()   ~, 1
#define METALANG99_PRIV_uintEq_99_99()   ~, 1
#define METALANG99_PRIV_uintEq_100_100() ~, 1
#define METALANG99_PRIV_uintEq_101_101() ~, 1
#define METALANG99_PRIV_uintEq_102_102() ~, 1
#define METALANG99_PRIV_uintEq_103_103() ~, 1
#define METALANG99_PRIV_uintEq_104_104() ~, 1
#define METALANG99_PRIV_uintEq_105_105() ~, 1
#define METALANG99_PRIV_uintEq_106_106() ~, 1
#define METALANG99_PRIV_uintEq_107_107() ~, 1
#define METALANG99_PRIV_uintEq_108_108() ~, 1
#define METALANG99_PRIV_uintEq_109_109() ~, 1
#define METALANG99_PRIV_uintEq_110_110() ~, 1
#define METALANG99_PRIV_uintEq_111_111() ~, 1
#define METALANG99_PRIV_uintEq_112_112() ~, 1
#define METALANG99_PRIV_uintEq_113_113() ~, 1
#define METALANG99_PRIV_uintEq_114_114() ~, 1
#define METALANG99_PRIV_uintEq_115_115() ~, 1
#define METALANG99_PRIV_uintEq_116_116() ~, 1
#define METALANG99_PRIV_uintEq_117_117() ~, 1
#define METALANG99_PRIV_uintEq_118_118() ~, 1
#define METALANG99_PRIV_uintEq_119_119() ~, 1
#define METALANG99_PRIV_uintEq_120_120() ~, 1
#define METALANG99_PRIV_uintEq_121_121() ~, 1
#define METALANG99_PRIV_uintEq_122_122() ~, 1
#define METALANG99_PRIV_uintEq_123_123() ~, 1
#define METALANG99_PRIV_uintEq_124_124() ~, 1
#define METALANG99_PRIV_uintEq_125_125() ~, 1
#define METALANG99_PRIV_uintEq_126_126() ~, 1
#define METALANG99_PRIV_uintEq_127_127() ~, 1
#define METALANG99_PRIV_uintEq_128_128() ~, 1
#define METALANG99_PRIV_uintEq_129_129() ~, 1
#define METALANG99_PRIV_uintEq_130_130() ~, 1
#define METALANG99_PRIV_uintEq_131_131() ~, 1
#define METALANG99_PRIV_uintEq_132_132() ~, 1
#define METALANG99_PRIV_uintEq_133_133() ~, 1
#define METALANG99_PRIV_uintEq_134_134() ~, 1
#define METALANG99_PRIV_uintEq_135_135() ~, 1
#define METALANG99_PRIV_uintEq_136_136() ~, 1
#define METALANG99_PRIV_uintEq_137_137() ~, 1
#define METALANG99_PRIV_uintEq_138_138() ~, 1
#define METALANG99_PRIV_uintEq_139_139() ~, 1
#define METALANG99_PRIV_uintEq_140_140() ~, 1
#define METALANG99_PRIV_uintEq_141_141() ~, 1
#define METALANG99_PRIV_uintEq_142_142() ~, 1
#define METALANG99_PRIV_uintEq_143_143() ~, 1
#define METALANG99_PRIV_uintEq_144_144() ~, 1
#define METALANG99_PRIV_uintEq_145_145() ~, 1
#define METALANG99_PRIV_uintEq_146_146() ~, 1
#define METALANG99_PRIV_uintEq_147_147() ~, 1
#define METALANG99_PRIV_uintEq_148_148() ~, 1
#define METALANG99_PRIV_uintEq_149_149() ~, 1
#define METALANG99_PRIV_uintEq_150_150() ~, 1
#define METALANG99_PRIV_uintEq_151_151() ~, 1
#define METALANG99_PRIV_uintEq_152_152() ~, 1
#define METALANG99_PRIV_uintEq_153_153() ~, 1
#define METALANG99_PRIV_uintEq_154_154() ~, 1
#define METALANG99_PRIV_uintEq_155_155() ~, 1
#define METALANG99_PRIV_uintEq_156_156() ~, 1
#define METALANG99_PRIV_uintEq_157_157() ~, 1
#define METALANG99_PRIV_uintEq_158_158() ~, 1
#define METALANG99_PRIV_uintEq_159_159() ~, 1
#define METALANG99_PRIV_uintEq_160_160() ~, 1
#define METALANG99_PRIV_uintEq_161_161() ~, 1
#define METALANG99_PRIV_uintEq_162_162() ~, 1
#define METALANG99_PRIV_uintEq_163_163() ~, 1
#define METALANG99_PRIV_uintEq_164_164() ~, 1
#define METALANG99_PRIV_uintEq_165_165() ~, 1
#define METALANG99_PRIV_uintEq_166_166() ~, 1
#define METALANG99_PRIV_uintEq_167_167() ~, 1
#define METALANG99_PRIV_uintEq_168_168() ~, 1
#define METALANG99_PRIV_uintEq_169_169() ~, 1
#define METALANG99_PRIV_uintEq_170_170() ~, 1
#define METALANG99_PRIV_uintEq_171_171() ~, 1
#define METALANG99_PRIV_uintEq_172_172() ~, 1
#define METALANG99_PRIV_uintEq_173_173() ~, 1
#define METALANG99_PRIV_uintEq_174_174() ~, 1
#define METALANG99_PRIV_uintEq_175_175() ~, 1
#define METALANG99_PRIV_uintEq_176_176() ~, 1
#define METALANG99_PRIV_uintEq_177_177() ~, 1
#define METALANG99_PRIV_uintEq_178_178() ~, 1
#define METALANG99_PRIV_uintEq_179_179() ~, 1
#define METALANG99_PRIV_uintEq_180_180() ~, 1
#define METALANG99_PRIV_uintEq_181_181() ~, 1
#define METALANG99_PRIV_uintEq_182_182() ~, 1
#define METALANG99_PRIV_uintEq_183_183() ~, 1
#define METALANG99_PRIV_uintEq_184_184() ~, 1
#define METALANG99_PRIV_uintEq_185_185() ~, 1
#define METALANG99_PRIV_uintEq_186_186() ~, 1
#define METALANG99_PRIV_uintEq_187_187() ~, 1
#define METALANG99_PRIV_uintEq_188_188() ~, 1
#define METALANG99_PRIV_uintEq_189_189() ~, 1
#define METALANG99_PRIV_uintEq_190_190() ~, 1
#define METALANG99_PRIV_uintEq_191_191() ~, 1
#define METALANG99_PRIV_uintEq_192_192() ~, 1
#define METALANG99_PRIV_uintEq_193_193() ~, 1
#define METALANG99_PRIV_uintEq_194_194() ~, 1
#define METALANG99_PRIV_uintEq_195_195() ~, 1
#define METALANG99_PRIV_uintEq_196_196() ~, 1
#define METALANG99_PRIV_uintEq_197_197() ~, 1
#define METALANG99_PRIV_uintEq_198_198() ~, 1
#define METALANG99_PRIV_uintEq_199_199() ~, 1
#define METALANG99_PRIV_uintEq_200_200() ~, 1
#define METALANG99_PRIV_uintEq_201_201() ~, 1
#define METALANG99_PRIV_uintEq_202_202() ~, 1
#define METALANG99_PRIV_uintEq_203_203() ~, 1
#define METALANG99_PRIV_uintEq_204_204() ~, 1
#define METALANG99_PRIV_uintEq_205_205() ~, 1
#define METALANG99_PRIV_uintEq_206_206() ~, 1
#define METALANG99_PRIV_uintEq_207_207() ~, 1
#define METALANG99_PRIV_uintEq_208_208() ~, 1
#define METALANG99_PRIV_uintEq_209_209() ~, 1
#define METALANG99_PRIV_uintEq_210_210() ~, 1
#define METALANG99_PRIV_uintEq_211_211() ~, 1
#define METALANG99_PRIV_uintEq_212_212() ~, 1
#define METALANG99_PRIV_uintEq_213_213() ~, 1
#define METALANG99_PRIV_uintEq_214_214() ~, 1
#define METALANG99_PRIV_uintEq_215_215() ~, 1
#define METALANG99_PRIV_uintEq_216_216() ~, 1
#define METALANG99_PRIV_uintEq_217_217() ~, 1
#define METALANG99_PRIV_uintEq_218_218() ~, 1
#define METALANG99_PRIV_uintEq_219_219() ~, 1
#define METALANG99_PRIV_uintEq_220_220() ~, 1
#define METALANG99_PRIV_uintEq_221_221() ~, 1
#define METALANG99_PRIV_uintEq_222_222() ~, 1
#define METALANG99_PRIV_uintEq_223_223() ~, 1
#define METALANG99_PRIV_uintEq_224_224() ~, 1
#define METALANG99_PRIV_uintEq_225_225() ~, 1
#define METALANG99_PRIV_uintEq_226_226() ~, 1
#define METALANG99_PRIV_uintEq_227_227() ~, 1
#define METALANG99_PRIV_uintEq_228_228() ~, 1
#define METALANG99_PRIV_uintEq_229_229() ~, 1
#define METALANG99_PRIV_uintEq_230_230() ~, 1
#define METALANG99_PRIV_uintEq_231_231() ~, 1
#define METALANG99_PRIV_uintEq_232_232() ~, 1
#define METALANG99_PRIV_uintEq_233_233() ~, 1
#define METALANG99_PRIV_uintEq_234_234() ~, 1
#define METALANG99_PRIV_uintEq_235_235() ~, 1
#define METALANG99_PRIV_uintEq_236_236() ~, 1
#define METALANG99_PRIV_uintEq_237_237() ~, 1
#define METALANG99_PRIV_uintEq_238_238() ~, 1
#define METALANG99_PRIV_uintEq_239_239() ~, 1
#define METALANG99_PRIV_uintEq_240_240() ~, 1
#define METALANG99_PRIV_uintEq_241_241() ~, 1
#define METALANG99_PRIV_uintEq_242_242() ~, 1
#define METALANG99_PRIV_uintEq_243_243() ~, 1
#define METALANG99_PRIV_uintEq_244_244() ~, 1
#define METALANG99_PRIV_uintEq_245_245() ~, 1
#define METALANG99_PRIV_uintEq_246_246() ~, 1
#define METALANG99_PRIV_uintEq_247_247() ~, 1
#define METALANG99_PRIV_uintEq_248_248() ~, 1
#define METALANG99_PRIV_uintEq_249_249() ~, 1
#define METALANG99_PRIV_uintEq_250_250() ~, 1
#define METALANG99_PRIV_uintEq_251_251() ~, 1
#define METALANG99_PRIV_uintEq_252_252() ~, 1
#define METALANG99_PRIV_uintEq_253_253() ~, 1
#define METALANG99_PRIV_uintEq_254_254() ~, 1
#define METALANG99_PRIV_uintEq_255_255() ~, 1

#endif // METALANG99_UINT_EQ_H
