/**
 * @file
 * This module is deprecated and exists only for backwards compatibility.
 */

#ifndef ML99_CONTROL_H
#define ML99_CONTROL_H

#include <metalang99/bool.h>      // ML99_if, ML99_IF
#include <metalang99/gen.h>       // ML99_times, ML99_repeat
#include <metalang99/variadics.h> // ML99_OVERLOAD

#endif // ML99_CONTROL_H
