
#ifndef METALANG99_EVAL_REC_UNROLL_H
#define METALANG99_EVAL_REC_UNROLL_H

#include <metalang99/eval/rec/progress.h>

#define METALANG99_PRIV_REC_0(choice, ...)     METALANG99_PRIV_REC_NEXT(1, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1(choice, ...)     METALANG99_PRIV_REC_NEXT(2, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2(choice, ...)     METALANG99_PRIV_REC_NEXT(3, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3(choice, ...)     METALANG99_PRIV_REC_NEXT(4, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4(choice, ...)     METALANG99_PRIV_REC_NEXT(5, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5(choice, ...)     METALANG99_PRIV_REC_NEXT(6, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6(choice, ...)     METALANG99_PRIV_REC_NEXT(7, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7(choice, ...)     METALANG99_PRIV_REC_NEXT(8, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8(choice, ...)     METALANG99_PRIV_REC_NEXT(9, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9(choice, ...)     METALANG99_PRIV_REC_NEXT(10, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10(choice, ...)    METALANG99_PRIV_REC_NEXT(11, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11(choice, ...)    METALANG99_PRIV_REC_NEXT(12, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12(choice, ...)    METALANG99_PRIV_REC_NEXT(13, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13(choice, ...)    METALANG99_PRIV_REC_NEXT(14, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14(choice, ...)    METALANG99_PRIV_REC_NEXT(15, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15(choice, ...)    METALANG99_PRIV_REC_NEXT(16, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16(choice, ...)    METALANG99_PRIV_REC_NEXT(17, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_17(choice, ...)    METALANG99_PRIV_REC_NEXT(18, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_18(choice, ...)    METALANG99_PRIV_REC_NEXT(19, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_19(choice, ...)    METALANG99_PRIV_REC_NEXT(20, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_20(choice, ...)    METALANG99_PRIV_REC_NEXT(21, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_21(choice, ...)    METALANG99_PRIV_REC_NEXT(22, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_22(choice, ...)    METALANG99_PRIV_REC_NEXT(23, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_23(choice, ...)    METALANG99_PRIV_REC_NEXT(24, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_24(choice, ...)    METALANG99_PRIV_REC_NEXT(25, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_25(choice, ...)    METALANG99_PRIV_REC_NEXT(26, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_26(choice, ...)    METALANG99_PRIV_REC_NEXT(27, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_27(choice, ...)    METALANG99_PRIV_REC_NEXT(28, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_28(choice, ...)    METALANG99_PRIV_REC_NEXT(29, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_29(choice, ...)    METALANG99_PRIV_REC_NEXT(30, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_30(choice, ...)    METALANG99_PRIV_REC_NEXT(31, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_31(choice, ...)    METALANG99_PRIV_REC_NEXT(32, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_32(choice, ...)    METALANG99_PRIV_REC_NEXT(33, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_33(choice, ...)    METALANG99_PRIV_REC_NEXT(34, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_34(choice, ...)    METALANG99_PRIV_REC_NEXT(35, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_35(choice, ...)    METALANG99_PRIV_REC_NEXT(36, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_36(choice, ...)    METALANG99_PRIV_REC_NEXT(37, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_37(choice, ...)    METALANG99_PRIV_REC_NEXT(38, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_38(choice, ...)    METALANG99_PRIV_REC_NEXT(39, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_39(choice, ...)    METALANG99_PRIV_REC_NEXT(40, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_40(choice, ...)    METALANG99_PRIV_REC_NEXT(41, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_41(choice, ...)    METALANG99_PRIV_REC_NEXT(42, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_42(choice, ...)    METALANG99_PRIV_REC_NEXT(43, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_43(choice, ...)    METALANG99_PRIV_REC_NEXT(44, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_44(choice, ...)    METALANG99_PRIV_REC_NEXT(45, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_45(choice, ...)    METALANG99_PRIV_REC_NEXT(46, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_46(choice, ...)    METALANG99_PRIV_REC_NEXT(47, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_47(choice, ...)    METALANG99_PRIV_REC_NEXT(48, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_48(choice, ...)    METALANG99_PRIV_REC_NEXT(49, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_49(choice, ...)    METALANG99_PRIV_REC_NEXT(50, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_50(choice, ...)    METALANG99_PRIV_REC_NEXT(51, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_51(choice, ...)    METALANG99_PRIV_REC_NEXT(52, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_52(choice, ...)    METALANG99_PRIV_REC_NEXT(53, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_53(choice, ...)    METALANG99_PRIV_REC_NEXT(54, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_54(choice, ...)    METALANG99_PRIV_REC_NEXT(55, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_55(choice, ...)    METALANG99_PRIV_REC_NEXT(56, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_56(choice, ...)    METALANG99_PRIV_REC_NEXT(57, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_57(choice, ...)    METALANG99_PRIV_REC_NEXT(58, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_58(choice, ...)    METALANG99_PRIV_REC_NEXT(59, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_59(choice, ...)    METALANG99_PRIV_REC_NEXT(60, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_60(choice, ...)    METALANG99_PRIV_REC_NEXT(61, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_61(choice, ...)    METALANG99_PRIV_REC_NEXT(62, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_62(choice, ...)    METALANG99_PRIV_REC_NEXT(63, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_63(choice, ...)    METALANG99_PRIV_REC_NEXT(64, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_64(choice, ...)    METALANG99_PRIV_REC_NEXT(65, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_65(choice, ...)    METALANG99_PRIV_REC_NEXT(66, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_66(choice, ...)    METALANG99_PRIV_REC_NEXT(67, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_67(choice, ...)    METALANG99_PRIV_REC_NEXT(68, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_68(choice, ...)    METALANG99_PRIV_REC_NEXT(69, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_69(choice, ...)    METALANG99_PRIV_REC_NEXT(70, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_70(choice, ...)    METALANG99_PRIV_REC_NEXT(71, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_71(choice, ...)    METALANG99_PRIV_REC_NEXT(72, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_72(choice, ...)    METALANG99_PRIV_REC_NEXT(73, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_73(choice, ...)    METALANG99_PRIV_REC_NEXT(74, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_74(choice, ...)    METALANG99_PRIV_REC_NEXT(75, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_75(choice, ...)    METALANG99_PRIV_REC_NEXT(76, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_76(choice, ...)    METALANG99_PRIV_REC_NEXT(77, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_77(choice, ...)    METALANG99_PRIV_REC_NEXT(78, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_78(choice, ...)    METALANG99_PRIV_REC_NEXT(79, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_79(choice, ...)    METALANG99_PRIV_REC_NEXT(80, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_80(choice, ...)    METALANG99_PRIV_REC_NEXT(81, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_81(choice, ...)    METALANG99_PRIV_REC_NEXT(82, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_82(choice, ...)    METALANG99_PRIV_REC_NEXT(83, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_83(choice, ...)    METALANG99_PRIV_REC_NEXT(84, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_84(choice, ...)    METALANG99_PRIV_REC_NEXT(85, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_85(choice, ...)    METALANG99_PRIV_REC_NEXT(86, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_86(choice, ...)    METALANG99_PRIV_REC_NEXT(87, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_87(choice, ...)    METALANG99_PRIV_REC_NEXT(88, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_88(choice, ...)    METALANG99_PRIV_REC_NEXT(89, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_89(choice, ...)    METALANG99_PRIV_REC_NEXT(90, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_90(choice, ...)    METALANG99_PRIV_REC_NEXT(91, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_91(choice, ...)    METALANG99_PRIV_REC_NEXT(92, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_92(choice, ...)    METALANG99_PRIV_REC_NEXT(93, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_93(choice, ...)    METALANG99_PRIV_REC_NEXT(94, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_94(choice, ...)    METALANG99_PRIV_REC_NEXT(95, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_95(choice, ...)    METALANG99_PRIV_REC_NEXT(96, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_96(choice, ...)    METALANG99_PRIV_REC_NEXT(97, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_97(choice, ...)    METALANG99_PRIV_REC_NEXT(98, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_98(choice, ...)    METALANG99_PRIV_REC_NEXT(99, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_99(choice, ...)    METALANG99_PRIV_REC_NEXT(100, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_100(choice, ...)   METALANG99_PRIV_REC_NEXT(101, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_101(choice, ...)   METALANG99_PRIV_REC_NEXT(102, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_102(choice, ...)   METALANG99_PRIV_REC_NEXT(103, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_103(choice, ...)   METALANG99_PRIV_REC_NEXT(104, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_104(choice, ...)   METALANG99_PRIV_REC_NEXT(105, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_105(choice, ...)   METALANG99_PRIV_REC_NEXT(106, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_106(choice, ...)   METALANG99_PRIV_REC_NEXT(107, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_107(choice, ...)   METALANG99_PRIV_REC_NEXT(108, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_108(choice, ...)   METALANG99_PRIV_REC_NEXT(109, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_109(choice, ...)   METALANG99_PRIV_REC_NEXT(110, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_110(choice, ...)   METALANG99_PRIV_REC_NEXT(111, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_111(choice, ...)   METALANG99_PRIV_REC_NEXT(112, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_112(choice, ...)   METALANG99_PRIV_REC_NEXT(113, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_113(choice, ...)   METALANG99_PRIV_REC_NEXT(114, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_114(choice, ...)   METALANG99_PRIV_REC_NEXT(115, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_115(choice, ...)   METALANG99_PRIV_REC_NEXT(116, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_116(choice, ...)   METALANG99_PRIV_REC_NEXT(117, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_117(choice, ...)   METALANG99_PRIV_REC_NEXT(118, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_118(choice, ...)   METALANG99_PRIV_REC_NEXT(119, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_119(choice, ...)   METALANG99_PRIV_REC_NEXT(120, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_120(choice, ...)   METALANG99_PRIV_REC_NEXT(121, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_121(choice, ...)   METALANG99_PRIV_REC_NEXT(122, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_122(choice, ...)   METALANG99_PRIV_REC_NEXT(123, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_123(choice, ...)   METALANG99_PRIV_REC_NEXT(124, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_124(choice, ...)   METALANG99_PRIV_REC_NEXT(125, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_125(choice, ...)   METALANG99_PRIV_REC_NEXT(126, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_126(choice, ...)   METALANG99_PRIV_REC_NEXT(127, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_127(choice, ...)   METALANG99_PRIV_REC_NEXT(128, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_128(choice, ...)   METALANG99_PRIV_REC_NEXT(129, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_129(choice, ...)   METALANG99_PRIV_REC_NEXT(130, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_130(choice, ...)   METALANG99_PRIV_REC_NEXT(131, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_131(choice, ...)   METALANG99_PRIV_REC_NEXT(132, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_132(choice, ...)   METALANG99_PRIV_REC_NEXT(133, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_133(choice, ...)   METALANG99_PRIV_REC_NEXT(134, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_134(choice, ...)   METALANG99_PRIV_REC_NEXT(135, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_135(choice, ...)   METALANG99_PRIV_REC_NEXT(136, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_136(choice, ...)   METALANG99_PRIV_REC_NEXT(137, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_137(choice, ...)   METALANG99_PRIV_REC_NEXT(138, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_138(choice, ...)   METALANG99_PRIV_REC_NEXT(139, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_139(choice, ...)   METALANG99_PRIV_REC_NEXT(140, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_140(choice, ...)   METALANG99_PRIV_REC_NEXT(141, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_141(choice, ...)   METALANG99_PRIV_REC_NEXT(142, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_142(choice, ...)   METALANG99_PRIV_REC_NEXT(143, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_143(choice, ...)   METALANG99_PRIV_REC_NEXT(144, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_144(choice, ...)   METALANG99_PRIV_REC_NEXT(145, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_145(choice, ...)   METALANG99_PRIV_REC_NEXT(146, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_146(choice, ...)   METALANG99_PRIV_REC_NEXT(147, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_147(choice, ...)   METALANG99_PRIV_REC_NEXT(148, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_148(choice, ...)   METALANG99_PRIV_REC_NEXT(149, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_149(choice, ...)   METALANG99_PRIV_REC_NEXT(150, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_150(choice, ...)   METALANG99_PRIV_REC_NEXT(151, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_151(choice, ...)   METALANG99_PRIV_REC_NEXT(152, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_152(choice, ...)   METALANG99_PRIV_REC_NEXT(153, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_153(choice, ...)   METALANG99_PRIV_REC_NEXT(154, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_154(choice, ...)   METALANG99_PRIV_REC_NEXT(155, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_155(choice, ...)   METALANG99_PRIV_REC_NEXT(156, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_156(choice, ...)   METALANG99_PRIV_REC_NEXT(157, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_157(choice, ...)   METALANG99_PRIV_REC_NEXT(158, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_158(choice, ...)   METALANG99_PRIV_REC_NEXT(159, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_159(choice, ...)   METALANG99_PRIV_REC_NEXT(160, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_160(choice, ...)   METALANG99_PRIV_REC_NEXT(161, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_161(choice, ...)   METALANG99_PRIV_REC_NEXT(162, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_162(choice, ...)   METALANG99_PRIV_REC_NEXT(163, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_163(choice, ...)   METALANG99_PRIV_REC_NEXT(164, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_164(choice, ...)   METALANG99_PRIV_REC_NEXT(165, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_165(choice, ...)   METALANG99_PRIV_REC_NEXT(166, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_166(choice, ...)   METALANG99_PRIV_REC_NEXT(167, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_167(choice, ...)   METALANG99_PRIV_REC_NEXT(168, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_168(choice, ...)   METALANG99_PRIV_REC_NEXT(169, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_169(choice, ...)   METALANG99_PRIV_REC_NEXT(170, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_170(choice, ...)   METALANG99_PRIV_REC_NEXT(171, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_171(choice, ...)   METALANG99_PRIV_REC_NEXT(172, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_172(choice, ...)   METALANG99_PRIV_REC_NEXT(173, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_173(choice, ...)   METALANG99_PRIV_REC_NEXT(174, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_174(choice, ...)   METALANG99_PRIV_REC_NEXT(175, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_175(choice, ...)   METALANG99_PRIV_REC_NEXT(176, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_176(choice, ...)   METALANG99_PRIV_REC_NEXT(177, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_177(choice, ...)   METALANG99_PRIV_REC_NEXT(178, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_178(choice, ...)   METALANG99_PRIV_REC_NEXT(179, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_179(choice, ...)   METALANG99_PRIV_REC_NEXT(180, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_180(choice, ...)   METALANG99_PRIV_REC_NEXT(181, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_181(choice, ...)   METALANG99_PRIV_REC_NEXT(182, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_182(choice, ...)   METALANG99_PRIV_REC_NEXT(183, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_183(choice, ...)   METALANG99_PRIV_REC_NEXT(184, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_184(choice, ...)   METALANG99_PRIV_REC_NEXT(185, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_185(choice, ...)   METALANG99_PRIV_REC_NEXT(186, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_186(choice, ...)   METALANG99_PRIV_REC_NEXT(187, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_187(choice, ...)   METALANG99_PRIV_REC_NEXT(188, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_188(choice, ...)   METALANG99_PRIV_REC_NEXT(189, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_189(choice, ...)   METALANG99_PRIV_REC_NEXT(190, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_190(choice, ...)   METALANG99_PRIV_REC_NEXT(191, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_191(choice, ...)   METALANG99_PRIV_REC_NEXT(192, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_192(choice, ...)   METALANG99_PRIV_REC_NEXT(193, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_193(choice, ...)   METALANG99_PRIV_REC_NEXT(194, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_194(choice, ...)   METALANG99_PRIV_REC_NEXT(195, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_195(choice, ...)   METALANG99_PRIV_REC_NEXT(196, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_196(choice, ...)   METALANG99_PRIV_REC_NEXT(197, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_197(choice, ...)   METALANG99_PRIV_REC_NEXT(198, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_198(choice, ...)   METALANG99_PRIV_REC_NEXT(199, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_199(choice, ...)   METALANG99_PRIV_REC_NEXT(200, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_200(choice, ...)   METALANG99_PRIV_REC_NEXT(201, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_201(choice, ...)   METALANG99_PRIV_REC_NEXT(202, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_202(choice, ...)   METALANG99_PRIV_REC_NEXT(203, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_203(choice, ...)   METALANG99_PRIV_REC_NEXT(204, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_204(choice, ...)   METALANG99_PRIV_REC_NEXT(205, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_205(choice, ...)   METALANG99_PRIV_REC_NEXT(206, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_206(choice, ...)   METALANG99_PRIV_REC_NEXT(207, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_207(choice, ...)   METALANG99_PRIV_REC_NEXT(208, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_208(choice, ...)   METALANG99_PRIV_REC_NEXT(209, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_209(choice, ...)   METALANG99_PRIV_REC_NEXT(210, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_210(choice, ...)   METALANG99_PRIV_REC_NEXT(211, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_211(choice, ...)   METALANG99_PRIV_REC_NEXT(212, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_212(choice, ...)   METALANG99_PRIV_REC_NEXT(213, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_213(choice, ...)   METALANG99_PRIV_REC_NEXT(214, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_214(choice, ...)   METALANG99_PRIV_REC_NEXT(215, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_215(choice, ...)   METALANG99_PRIV_REC_NEXT(216, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_216(choice, ...)   METALANG99_PRIV_REC_NEXT(217, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_217(choice, ...)   METALANG99_PRIV_REC_NEXT(218, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_218(choice, ...)   METALANG99_PRIV_REC_NEXT(219, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_219(choice, ...)   METALANG99_PRIV_REC_NEXT(220, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_220(choice, ...)   METALANG99_PRIV_REC_NEXT(221, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_221(choice, ...)   METALANG99_PRIV_REC_NEXT(222, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_222(choice, ...)   METALANG99_PRIV_REC_NEXT(223, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_223(choice, ...)   METALANG99_PRIV_REC_NEXT(224, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_224(choice, ...)   METALANG99_PRIV_REC_NEXT(225, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_225(choice, ...)   METALANG99_PRIV_REC_NEXT(226, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_226(choice, ...)   METALANG99_PRIV_REC_NEXT(227, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_227(choice, ...)   METALANG99_PRIV_REC_NEXT(228, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_228(choice, ...)   METALANG99_PRIV_REC_NEXT(229, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_229(choice, ...)   METALANG99_PRIV_REC_NEXT(230, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_230(choice, ...)   METALANG99_PRIV_REC_NEXT(231, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_231(choice, ...)   METALANG99_PRIV_REC_NEXT(232, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_232(choice, ...)   METALANG99_PRIV_REC_NEXT(233, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_233(choice, ...)   METALANG99_PRIV_REC_NEXT(234, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_234(choice, ...)   METALANG99_PRIV_REC_NEXT(235, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_235(choice, ...)   METALANG99_PRIV_REC_NEXT(236, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_236(choice, ...)   METALANG99_PRIV_REC_NEXT(237, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_237(choice, ...)   METALANG99_PRIV_REC_NEXT(238, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_238(choice, ...)   METALANG99_PRIV_REC_NEXT(239, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_239(choice, ...)   METALANG99_PRIV_REC_NEXT(240, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_240(choice, ...)   METALANG99_PRIV_REC_NEXT(241, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_241(choice, ...)   METALANG99_PRIV_REC_NEXT(242, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_242(choice, ...)   METALANG99_PRIV_REC_NEXT(243, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_243(choice, ...)   METALANG99_PRIV_REC_NEXT(244, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_244(choice, ...)   METALANG99_PRIV_REC_NEXT(245, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_245(choice, ...)   METALANG99_PRIV_REC_NEXT(246, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_246(choice, ...)   METALANG99_PRIV_REC_NEXT(247, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_247(choice, ...)   METALANG99_PRIV_REC_NEXT(248, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_248(choice, ...)   METALANG99_PRIV_REC_NEXT(249, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_249(choice, ...)   METALANG99_PRIV_REC_NEXT(250, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_250(choice, ...)   METALANG99_PRIV_REC_NEXT(251, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_251(choice, ...)   METALANG99_PRIV_REC_NEXT(252, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_252(choice, ...)   METALANG99_PRIV_REC_NEXT(253, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_253(choice, ...)   METALANG99_PRIV_REC_NEXT(254, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_254(choice, ...)   METALANG99_PRIV_REC_NEXT(255, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_255(choice, ...)   METALANG99_PRIV_REC_NEXT(256, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_256(choice, ...)   METALANG99_PRIV_REC_NEXT(257, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_257(choice, ...)   METALANG99_PRIV_REC_NEXT(258, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_258(choice, ...)   METALANG99_PRIV_REC_NEXT(259, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_259(choice, ...)   METALANG99_PRIV_REC_NEXT(260, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_260(choice, ...)   METALANG99_PRIV_REC_NEXT(261, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_261(choice, ...)   METALANG99_PRIV_REC_NEXT(262, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_262(choice, ...)   METALANG99_PRIV_REC_NEXT(263, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_263(choice, ...)   METALANG99_PRIV_REC_NEXT(264, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_264(choice, ...)   METALANG99_PRIV_REC_NEXT(265, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_265(choice, ...)   METALANG99_PRIV_REC_NEXT(266, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_266(choice, ...)   METALANG99_PRIV_REC_NEXT(267, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_267(choice, ...)   METALANG99_PRIV_REC_NEXT(268, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_268(choice, ...)   METALANG99_PRIV_REC_NEXT(269, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_269(choice, ...)   METALANG99_PRIV_REC_NEXT(270, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_270(choice, ...)   METALANG99_PRIV_REC_NEXT(271, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_271(choice, ...)   METALANG99_PRIV_REC_NEXT(272, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_272(choice, ...)   METALANG99_PRIV_REC_NEXT(273, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_273(choice, ...)   METALANG99_PRIV_REC_NEXT(274, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_274(choice, ...)   METALANG99_PRIV_REC_NEXT(275, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_275(choice, ...)   METALANG99_PRIV_REC_NEXT(276, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_276(choice, ...)   METALANG99_PRIV_REC_NEXT(277, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_277(choice, ...)   METALANG99_PRIV_REC_NEXT(278, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_278(choice, ...)   METALANG99_PRIV_REC_NEXT(279, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_279(choice, ...)   METALANG99_PRIV_REC_NEXT(280, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_280(choice, ...)   METALANG99_PRIV_REC_NEXT(281, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_281(choice, ...)   METALANG99_PRIV_REC_NEXT(282, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_282(choice, ...)   METALANG99_PRIV_REC_NEXT(283, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_283(choice, ...)   METALANG99_PRIV_REC_NEXT(284, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_284(choice, ...)   METALANG99_PRIV_REC_NEXT(285, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_285(choice, ...)   METALANG99_PRIV_REC_NEXT(286, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_286(choice, ...)   METALANG99_PRIV_REC_NEXT(287, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_287(choice, ...)   METALANG99_PRIV_REC_NEXT(288, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_288(choice, ...)   METALANG99_PRIV_REC_NEXT(289, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_289(choice, ...)   METALANG99_PRIV_REC_NEXT(290, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_290(choice, ...)   METALANG99_PRIV_REC_NEXT(291, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_291(choice, ...)   METALANG99_PRIV_REC_NEXT(292, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_292(choice, ...)   METALANG99_PRIV_REC_NEXT(293, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_293(choice, ...)   METALANG99_PRIV_REC_NEXT(294, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_294(choice, ...)   METALANG99_PRIV_REC_NEXT(295, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_295(choice, ...)   METALANG99_PRIV_REC_NEXT(296, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_296(choice, ...)   METALANG99_PRIV_REC_NEXT(297, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_297(choice, ...)   METALANG99_PRIV_REC_NEXT(298, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_298(choice, ...)   METALANG99_PRIV_REC_NEXT(299, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_299(choice, ...)   METALANG99_PRIV_REC_NEXT(300, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_300(choice, ...)   METALANG99_PRIV_REC_NEXT(301, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_301(choice, ...)   METALANG99_PRIV_REC_NEXT(302, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_302(choice, ...)   METALANG99_PRIV_REC_NEXT(303, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_303(choice, ...)   METALANG99_PRIV_REC_NEXT(304, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_304(choice, ...)   METALANG99_PRIV_REC_NEXT(305, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_305(choice, ...)   METALANG99_PRIV_REC_NEXT(306, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_306(choice, ...)   METALANG99_PRIV_REC_NEXT(307, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_307(choice, ...)   METALANG99_PRIV_REC_NEXT(308, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_308(choice, ...)   METALANG99_PRIV_REC_NEXT(309, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_309(choice, ...)   METALANG99_PRIV_REC_NEXT(310, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_310(choice, ...)   METALANG99_PRIV_REC_NEXT(311, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_311(choice, ...)   METALANG99_PRIV_REC_NEXT(312, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_312(choice, ...)   METALANG99_PRIV_REC_NEXT(313, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_313(choice, ...)   METALANG99_PRIV_REC_NEXT(314, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_314(choice, ...)   METALANG99_PRIV_REC_NEXT(315, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_315(choice, ...)   METALANG99_PRIV_REC_NEXT(316, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_316(choice, ...)   METALANG99_PRIV_REC_NEXT(317, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_317(choice, ...)   METALANG99_PRIV_REC_NEXT(318, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_318(choice, ...)   METALANG99_PRIV_REC_NEXT(319, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_319(choice, ...)   METALANG99_PRIV_REC_NEXT(320, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_320(choice, ...)   METALANG99_PRIV_REC_NEXT(321, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_321(choice, ...)   METALANG99_PRIV_REC_NEXT(322, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_322(choice, ...)   METALANG99_PRIV_REC_NEXT(323, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_323(choice, ...)   METALANG99_PRIV_REC_NEXT(324, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_324(choice, ...)   METALANG99_PRIV_REC_NEXT(325, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_325(choice, ...)   METALANG99_PRIV_REC_NEXT(326, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_326(choice, ...)   METALANG99_PRIV_REC_NEXT(327, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_327(choice, ...)   METALANG99_PRIV_REC_NEXT(328, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_328(choice, ...)   METALANG99_PRIV_REC_NEXT(329, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_329(choice, ...)   METALANG99_PRIV_REC_NEXT(330, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_330(choice, ...)   METALANG99_PRIV_REC_NEXT(331, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_331(choice, ...)   METALANG99_PRIV_REC_NEXT(332, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_332(choice, ...)   METALANG99_PRIV_REC_NEXT(333, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_333(choice, ...)   METALANG99_PRIV_REC_NEXT(334, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_334(choice, ...)   METALANG99_PRIV_REC_NEXT(335, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_335(choice, ...)   METALANG99_PRIV_REC_NEXT(336, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_336(choice, ...)   METALANG99_PRIV_REC_NEXT(337, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_337(choice, ...)   METALANG99_PRIV_REC_NEXT(338, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_338(choice, ...)   METALANG99_PRIV_REC_NEXT(339, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_339(choice, ...)   METALANG99_PRIV_REC_NEXT(340, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_340(choice, ...)   METALANG99_PRIV_REC_NEXT(341, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_341(choice, ...)   METALANG99_PRIV_REC_NEXT(342, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_342(choice, ...)   METALANG99_PRIV_REC_NEXT(343, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_343(choice, ...)   METALANG99_PRIV_REC_NEXT(344, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_344(choice, ...)   METALANG99_PRIV_REC_NEXT(345, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_345(choice, ...)   METALANG99_PRIV_REC_NEXT(346, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_346(choice, ...)   METALANG99_PRIV_REC_NEXT(347, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_347(choice, ...)   METALANG99_PRIV_REC_NEXT(348, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_348(choice, ...)   METALANG99_PRIV_REC_NEXT(349, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_349(choice, ...)   METALANG99_PRIV_REC_NEXT(350, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_350(choice, ...)   METALANG99_PRIV_REC_NEXT(351, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_351(choice, ...)   METALANG99_PRIV_REC_NEXT(352, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_352(choice, ...)   METALANG99_PRIV_REC_NEXT(353, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_353(choice, ...)   METALANG99_PRIV_REC_NEXT(354, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_354(choice, ...)   METALANG99_PRIV_REC_NEXT(355, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_355(choice, ...)   METALANG99_PRIV_REC_NEXT(356, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_356(choice, ...)   METALANG99_PRIV_REC_NEXT(357, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_357(choice, ...)   METALANG99_PRIV_REC_NEXT(358, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_358(choice, ...)   METALANG99_PRIV_REC_NEXT(359, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_359(choice, ...)   METALANG99_PRIV_REC_NEXT(360, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_360(choice, ...)   METALANG99_PRIV_REC_NEXT(361, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_361(choice, ...)   METALANG99_PRIV_REC_NEXT(362, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_362(choice, ...)   METALANG99_PRIV_REC_NEXT(363, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_363(choice, ...)   METALANG99_PRIV_REC_NEXT(364, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_364(choice, ...)   METALANG99_PRIV_REC_NEXT(365, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_365(choice, ...)   METALANG99_PRIV_REC_NEXT(366, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_366(choice, ...)   METALANG99_PRIV_REC_NEXT(367, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_367(choice, ...)   METALANG99_PRIV_REC_NEXT(368, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_368(choice, ...)   METALANG99_PRIV_REC_NEXT(369, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_369(choice, ...)   METALANG99_PRIV_REC_NEXT(370, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_370(choice, ...)   METALANG99_PRIV_REC_NEXT(371, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_371(choice, ...)   METALANG99_PRIV_REC_NEXT(372, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_372(choice, ...)   METALANG99_PRIV_REC_NEXT(373, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_373(choice, ...)   METALANG99_PRIV_REC_NEXT(374, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_374(choice, ...)   METALANG99_PRIV_REC_NEXT(375, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_375(choice, ...)   METALANG99_PRIV_REC_NEXT(376, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_376(choice, ...)   METALANG99_PRIV_REC_NEXT(377, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_377(choice, ...)   METALANG99_PRIV_REC_NEXT(378, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_378(choice, ...)   METALANG99_PRIV_REC_NEXT(379, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_379(choice, ...)   METALANG99_PRIV_REC_NEXT(380, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_380(choice, ...)   METALANG99_PRIV_REC_NEXT(381, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_381(choice, ...)   METALANG99_PRIV_REC_NEXT(382, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_382(choice, ...)   METALANG99_PRIV_REC_NEXT(383, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_383(choice, ...)   METALANG99_PRIV_REC_NEXT(384, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_384(choice, ...)   METALANG99_PRIV_REC_NEXT(385, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_385(choice, ...)   METALANG99_PRIV_REC_NEXT(386, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_386(choice, ...)   METALANG99_PRIV_REC_NEXT(387, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_387(choice, ...)   METALANG99_PRIV_REC_NEXT(388, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_388(choice, ...)   METALANG99_PRIV_REC_NEXT(389, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_389(choice, ...)   METALANG99_PRIV_REC_NEXT(390, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_390(choice, ...)   METALANG99_PRIV_REC_NEXT(391, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_391(choice, ...)   METALANG99_PRIV_REC_NEXT(392, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_392(choice, ...)   METALANG99_PRIV_REC_NEXT(393, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_393(choice, ...)   METALANG99_PRIV_REC_NEXT(394, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_394(choice, ...)   METALANG99_PRIV_REC_NEXT(395, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_395(choice, ...)   METALANG99_PRIV_REC_NEXT(396, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_396(choice, ...)   METALANG99_PRIV_REC_NEXT(397, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_397(choice, ...)   METALANG99_PRIV_REC_NEXT(398, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_398(choice, ...)   METALANG99_PRIV_REC_NEXT(399, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_399(choice, ...)   METALANG99_PRIV_REC_NEXT(400, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_400(choice, ...)   METALANG99_PRIV_REC_NEXT(401, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_401(choice, ...)   METALANG99_PRIV_REC_NEXT(402, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_402(choice, ...)   METALANG99_PRIV_REC_NEXT(403, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_403(choice, ...)   METALANG99_PRIV_REC_NEXT(404, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_404(choice, ...)   METALANG99_PRIV_REC_NEXT(405, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_405(choice, ...)   METALANG99_PRIV_REC_NEXT(406, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_406(choice, ...)   METALANG99_PRIV_REC_NEXT(407, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_407(choice, ...)   METALANG99_PRIV_REC_NEXT(408, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_408(choice, ...)   METALANG99_PRIV_REC_NEXT(409, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_409(choice, ...)   METALANG99_PRIV_REC_NEXT(410, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_410(choice, ...)   METALANG99_PRIV_REC_NEXT(411, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_411(choice, ...)   METALANG99_PRIV_REC_NEXT(412, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_412(choice, ...)   METALANG99_PRIV_REC_NEXT(413, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_413(choice, ...)   METALANG99_PRIV_REC_NEXT(414, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_414(choice, ...)   METALANG99_PRIV_REC_NEXT(415, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_415(choice, ...)   METALANG99_PRIV_REC_NEXT(416, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_416(choice, ...)   METALANG99_PRIV_REC_NEXT(417, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_417(choice, ...)   METALANG99_PRIV_REC_NEXT(418, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_418(choice, ...)   METALANG99_PRIV_REC_NEXT(419, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_419(choice, ...)   METALANG99_PRIV_REC_NEXT(420, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_420(choice, ...)   METALANG99_PRIV_REC_NEXT(421, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_421(choice, ...)   METALANG99_PRIV_REC_NEXT(422, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_422(choice, ...)   METALANG99_PRIV_REC_NEXT(423, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_423(choice, ...)   METALANG99_PRIV_REC_NEXT(424, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_424(choice, ...)   METALANG99_PRIV_REC_NEXT(425, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_425(choice, ...)   METALANG99_PRIV_REC_NEXT(426, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_426(choice, ...)   METALANG99_PRIV_REC_NEXT(427, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_427(choice, ...)   METALANG99_PRIV_REC_NEXT(428, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_428(choice, ...)   METALANG99_PRIV_REC_NEXT(429, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_429(choice, ...)   METALANG99_PRIV_REC_NEXT(430, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_430(choice, ...)   METALANG99_PRIV_REC_NEXT(431, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_431(choice, ...)   METALANG99_PRIV_REC_NEXT(432, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_432(choice, ...)   METALANG99_PRIV_REC_NEXT(433, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_433(choice, ...)   METALANG99_PRIV_REC_NEXT(434, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_434(choice, ...)   METALANG99_PRIV_REC_NEXT(435, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_435(choice, ...)   METALANG99_PRIV_REC_NEXT(436, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_436(choice, ...)   METALANG99_PRIV_REC_NEXT(437, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_437(choice, ...)   METALANG99_PRIV_REC_NEXT(438, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_438(choice, ...)   METALANG99_PRIV_REC_NEXT(439, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_439(choice, ...)   METALANG99_PRIV_REC_NEXT(440, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_440(choice, ...)   METALANG99_PRIV_REC_NEXT(441, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_441(choice, ...)   METALANG99_PRIV_REC_NEXT(442, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_442(choice, ...)   METALANG99_PRIV_REC_NEXT(443, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_443(choice, ...)   METALANG99_PRIV_REC_NEXT(444, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_444(choice, ...)   METALANG99_PRIV_REC_NEXT(445, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_445(choice, ...)   METALANG99_PRIV_REC_NEXT(446, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_446(choice, ...)   METALANG99_PRIV_REC_NEXT(447, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_447(choice, ...)   METALANG99_PRIV_REC_NEXT(448, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_448(choice, ...)   METALANG99_PRIV_REC_NEXT(449, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_449(choice, ...)   METALANG99_PRIV_REC_NEXT(450, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_450(choice, ...)   METALANG99_PRIV_REC_NEXT(451, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_451(choice, ...)   METALANG99_PRIV_REC_NEXT(452, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_452(choice, ...)   METALANG99_PRIV_REC_NEXT(453, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_453(choice, ...)   METALANG99_PRIV_REC_NEXT(454, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_454(choice, ...)   METALANG99_PRIV_REC_NEXT(455, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_455(choice, ...)   METALANG99_PRIV_REC_NEXT(456, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_456(choice, ...)   METALANG99_PRIV_REC_NEXT(457, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_457(choice, ...)   METALANG99_PRIV_REC_NEXT(458, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_458(choice, ...)   METALANG99_PRIV_REC_NEXT(459, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_459(choice, ...)   METALANG99_PRIV_REC_NEXT(460, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_460(choice, ...)   METALANG99_PRIV_REC_NEXT(461, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_461(choice, ...)   METALANG99_PRIV_REC_NEXT(462, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_462(choice, ...)   METALANG99_PRIV_REC_NEXT(463, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_463(choice, ...)   METALANG99_PRIV_REC_NEXT(464, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_464(choice, ...)   METALANG99_PRIV_REC_NEXT(465, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_465(choice, ...)   METALANG99_PRIV_REC_NEXT(466, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_466(choice, ...)   METALANG99_PRIV_REC_NEXT(467, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_467(choice, ...)   METALANG99_PRIV_REC_NEXT(468, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_468(choice, ...)   METALANG99_PRIV_REC_NEXT(469, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_469(choice, ...)   METALANG99_PRIV_REC_NEXT(470, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_470(choice, ...)   METALANG99_PRIV_REC_NEXT(471, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_471(choice, ...)   METALANG99_PRIV_REC_NEXT(472, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_472(choice, ...)   METALANG99_PRIV_REC_NEXT(473, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_473(choice, ...)   METALANG99_PRIV_REC_NEXT(474, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_474(choice, ...)   METALANG99_PRIV_REC_NEXT(475, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_475(choice, ...)   METALANG99_PRIV_REC_NEXT(476, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_476(choice, ...)   METALANG99_PRIV_REC_NEXT(477, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_477(choice, ...)   METALANG99_PRIV_REC_NEXT(478, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_478(choice, ...)   METALANG99_PRIV_REC_NEXT(479, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_479(choice, ...)   METALANG99_PRIV_REC_NEXT(480, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_480(choice, ...)   METALANG99_PRIV_REC_NEXT(481, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_481(choice, ...)   METALANG99_PRIV_REC_NEXT(482, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_482(choice, ...)   METALANG99_PRIV_REC_NEXT(483, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_483(choice, ...)   METALANG99_PRIV_REC_NEXT(484, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_484(choice, ...)   METALANG99_PRIV_REC_NEXT(485, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_485(choice, ...)   METALANG99_PRIV_REC_NEXT(486, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_486(choice, ...)   METALANG99_PRIV_REC_NEXT(487, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_487(choice, ...)   METALANG99_PRIV_REC_NEXT(488, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_488(choice, ...)   METALANG99_PRIV_REC_NEXT(489, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_489(choice, ...)   METALANG99_PRIV_REC_NEXT(490, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_490(choice, ...)   METALANG99_PRIV_REC_NEXT(491, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_491(choice, ...)   METALANG99_PRIV_REC_NEXT(492, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_492(choice, ...)   METALANG99_PRIV_REC_NEXT(493, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_493(choice, ...)   METALANG99_PRIV_REC_NEXT(494, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_494(choice, ...)   METALANG99_PRIV_REC_NEXT(495, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_495(choice, ...)   METALANG99_PRIV_REC_NEXT(496, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_496(choice, ...)   METALANG99_PRIV_REC_NEXT(497, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_497(choice, ...)   METALANG99_PRIV_REC_NEXT(498, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_498(choice, ...)   METALANG99_PRIV_REC_NEXT(499, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_499(choice, ...)   METALANG99_PRIV_REC_NEXT(500, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_500(choice, ...)   METALANG99_PRIV_REC_NEXT(501, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_501(choice, ...)   METALANG99_PRIV_REC_NEXT(502, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_502(choice, ...)   METALANG99_PRIV_REC_NEXT(503, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_503(choice, ...)   METALANG99_PRIV_REC_NEXT(504, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_504(choice, ...)   METALANG99_PRIV_REC_NEXT(505, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_505(choice, ...)   METALANG99_PRIV_REC_NEXT(506, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_506(choice, ...)   METALANG99_PRIV_REC_NEXT(507, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_507(choice, ...)   METALANG99_PRIV_REC_NEXT(508, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_508(choice, ...)   METALANG99_PRIV_REC_NEXT(509, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_509(choice, ...)   METALANG99_PRIV_REC_NEXT(510, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_510(choice, ...)   METALANG99_PRIV_REC_NEXT(511, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_511(choice, ...)   METALANG99_PRIV_REC_NEXT(512, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_512(choice, ...)   METALANG99_PRIV_REC_NEXT(513, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_513(choice, ...)   METALANG99_PRIV_REC_NEXT(514, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_514(choice, ...)   METALANG99_PRIV_REC_NEXT(515, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_515(choice, ...)   METALANG99_PRIV_REC_NEXT(516, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_516(choice, ...)   METALANG99_PRIV_REC_NEXT(517, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_517(choice, ...)   METALANG99_PRIV_REC_NEXT(518, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_518(choice, ...)   METALANG99_PRIV_REC_NEXT(519, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_519(choice, ...)   METALANG99_PRIV_REC_NEXT(520, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_520(choice, ...)   METALANG99_PRIV_REC_NEXT(521, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_521(choice, ...)   METALANG99_PRIV_REC_NEXT(522, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_522(choice, ...)   METALANG99_PRIV_REC_NEXT(523, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_523(choice, ...)   METALANG99_PRIV_REC_NEXT(524, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_524(choice, ...)   METALANG99_PRIV_REC_NEXT(525, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_525(choice, ...)   METALANG99_PRIV_REC_NEXT(526, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_526(choice, ...)   METALANG99_PRIV_REC_NEXT(527, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_527(choice, ...)   METALANG99_PRIV_REC_NEXT(528, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_528(choice, ...)   METALANG99_PRIV_REC_NEXT(529, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_529(choice, ...)   METALANG99_PRIV_REC_NEXT(530, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_530(choice, ...)   METALANG99_PRIV_REC_NEXT(531, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_531(choice, ...)   METALANG99_PRIV_REC_NEXT(532, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_532(choice, ...)   METALANG99_PRIV_REC_NEXT(533, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_533(choice, ...)   METALANG99_PRIV_REC_NEXT(534, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_534(choice, ...)   METALANG99_PRIV_REC_NEXT(535, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_535(choice, ...)   METALANG99_PRIV_REC_NEXT(536, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_536(choice, ...)   METALANG99_PRIV_REC_NEXT(537, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_537(choice, ...)   METALANG99_PRIV_REC_NEXT(538, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_538(choice, ...)   METALANG99_PRIV_REC_NEXT(539, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_539(choice, ...)   METALANG99_PRIV_REC_NEXT(540, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_540(choice, ...)   METALANG99_PRIV_REC_NEXT(541, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_541(choice, ...)   METALANG99_PRIV_REC_NEXT(542, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_542(choice, ...)   METALANG99_PRIV_REC_NEXT(543, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_543(choice, ...)   METALANG99_PRIV_REC_NEXT(544, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_544(choice, ...)   METALANG99_PRIV_REC_NEXT(545, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_545(choice, ...)   METALANG99_PRIV_REC_NEXT(546, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_546(choice, ...)   METALANG99_PRIV_REC_NEXT(547, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_547(choice, ...)   METALANG99_PRIV_REC_NEXT(548, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_548(choice, ...)   METALANG99_PRIV_REC_NEXT(549, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_549(choice, ...)   METALANG99_PRIV_REC_NEXT(550, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_550(choice, ...)   METALANG99_PRIV_REC_NEXT(551, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_551(choice, ...)   METALANG99_PRIV_REC_NEXT(552, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_552(choice, ...)   METALANG99_PRIV_REC_NEXT(553, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_553(choice, ...)   METALANG99_PRIV_REC_NEXT(554, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_554(choice, ...)   METALANG99_PRIV_REC_NEXT(555, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_555(choice, ...)   METALANG99_PRIV_REC_NEXT(556, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_556(choice, ...)   METALANG99_PRIV_REC_NEXT(557, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_557(choice, ...)   METALANG99_PRIV_REC_NEXT(558, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_558(choice, ...)   METALANG99_PRIV_REC_NEXT(559, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_559(choice, ...)   METALANG99_PRIV_REC_NEXT(560, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_560(choice, ...)   METALANG99_PRIV_REC_NEXT(561, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_561(choice, ...)   METALANG99_PRIV_REC_NEXT(562, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_562(choice, ...)   METALANG99_PRIV_REC_NEXT(563, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_563(choice, ...)   METALANG99_PRIV_REC_NEXT(564, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_564(choice, ...)   METALANG99_PRIV_REC_NEXT(565, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_565(choice, ...)   METALANG99_PRIV_REC_NEXT(566, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_566(choice, ...)   METALANG99_PRIV_REC_NEXT(567, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_567(choice, ...)   METALANG99_PRIV_REC_NEXT(568, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_568(choice, ...)   METALANG99_PRIV_REC_NEXT(569, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_569(choice, ...)   METALANG99_PRIV_REC_NEXT(570, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_570(choice, ...)   METALANG99_PRIV_REC_NEXT(571, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_571(choice, ...)   METALANG99_PRIV_REC_NEXT(572, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_572(choice, ...)   METALANG99_PRIV_REC_NEXT(573, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_573(choice, ...)   METALANG99_PRIV_REC_NEXT(574, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_574(choice, ...)   METALANG99_PRIV_REC_NEXT(575, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_575(choice, ...)   METALANG99_PRIV_REC_NEXT(576, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_576(choice, ...)   METALANG99_PRIV_REC_NEXT(577, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_577(choice, ...)   METALANG99_PRIV_REC_NEXT(578, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_578(choice, ...)   METALANG99_PRIV_REC_NEXT(579, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_579(choice, ...)   METALANG99_PRIV_REC_NEXT(580, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_580(choice, ...)   METALANG99_PRIV_REC_NEXT(581, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_581(choice, ...)   METALANG99_PRIV_REC_NEXT(582, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_582(choice, ...)   METALANG99_PRIV_REC_NEXT(583, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_583(choice, ...)   METALANG99_PRIV_REC_NEXT(584, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_584(choice, ...)   METALANG99_PRIV_REC_NEXT(585, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_585(choice, ...)   METALANG99_PRIV_REC_NEXT(586, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_586(choice, ...)   METALANG99_PRIV_REC_NEXT(587, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_587(choice, ...)   METALANG99_PRIV_REC_NEXT(588, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_588(choice, ...)   METALANG99_PRIV_REC_NEXT(589, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_589(choice, ...)   METALANG99_PRIV_REC_NEXT(590, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_590(choice, ...)   METALANG99_PRIV_REC_NEXT(591, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_591(choice, ...)   METALANG99_PRIV_REC_NEXT(592, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_592(choice, ...)   METALANG99_PRIV_REC_NEXT(593, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_593(choice, ...)   METALANG99_PRIV_REC_NEXT(594, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_594(choice, ...)   METALANG99_PRIV_REC_NEXT(595, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_595(choice, ...)   METALANG99_PRIV_REC_NEXT(596, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_596(choice, ...)   METALANG99_PRIV_REC_NEXT(597, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_597(choice, ...)   METALANG99_PRIV_REC_NEXT(598, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_598(choice, ...)   METALANG99_PRIV_REC_NEXT(599, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_599(choice, ...)   METALANG99_PRIV_REC_NEXT(600, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_600(choice, ...)   METALANG99_PRIV_REC_NEXT(601, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_601(choice, ...)   METALANG99_PRIV_REC_NEXT(602, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_602(choice, ...)   METALANG99_PRIV_REC_NEXT(603, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_603(choice, ...)   METALANG99_PRIV_REC_NEXT(604, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_604(choice, ...)   METALANG99_PRIV_REC_NEXT(605, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_605(choice, ...)   METALANG99_PRIV_REC_NEXT(606, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_606(choice, ...)   METALANG99_PRIV_REC_NEXT(607, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_607(choice, ...)   METALANG99_PRIV_REC_NEXT(608, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_608(choice, ...)   METALANG99_PRIV_REC_NEXT(609, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_609(choice, ...)   METALANG99_PRIV_REC_NEXT(610, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_610(choice, ...)   METALANG99_PRIV_REC_NEXT(611, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_611(choice, ...)   METALANG99_PRIV_REC_NEXT(612, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_612(choice, ...)   METALANG99_PRIV_REC_NEXT(613, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_613(choice, ...)   METALANG99_PRIV_REC_NEXT(614, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_614(choice, ...)   METALANG99_PRIV_REC_NEXT(615, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_615(choice, ...)   METALANG99_PRIV_REC_NEXT(616, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_616(choice, ...)   METALANG99_PRIV_REC_NEXT(617, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_617(choice, ...)   METALANG99_PRIV_REC_NEXT(618, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_618(choice, ...)   METALANG99_PRIV_REC_NEXT(619, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_619(choice, ...)   METALANG99_PRIV_REC_NEXT(620, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_620(choice, ...)   METALANG99_PRIV_REC_NEXT(621, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_621(choice, ...)   METALANG99_PRIV_REC_NEXT(622, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_622(choice, ...)   METALANG99_PRIV_REC_NEXT(623, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_623(choice, ...)   METALANG99_PRIV_REC_NEXT(624, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_624(choice, ...)   METALANG99_PRIV_REC_NEXT(625, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_625(choice, ...)   METALANG99_PRIV_REC_NEXT(626, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_626(choice, ...)   METALANG99_PRIV_REC_NEXT(627, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_627(choice, ...)   METALANG99_PRIV_REC_NEXT(628, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_628(choice, ...)   METALANG99_PRIV_REC_NEXT(629, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_629(choice, ...)   METALANG99_PRIV_REC_NEXT(630, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_630(choice, ...)   METALANG99_PRIV_REC_NEXT(631, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_631(choice, ...)   METALANG99_PRIV_REC_NEXT(632, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_632(choice, ...)   METALANG99_PRIV_REC_NEXT(633, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_633(choice, ...)   METALANG99_PRIV_REC_NEXT(634, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_634(choice, ...)   METALANG99_PRIV_REC_NEXT(635, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_635(choice, ...)   METALANG99_PRIV_REC_NEXT(636, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_636(choice, ...)   METALANG99_PRIV_REC_NEXT(637, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_637(choice, ...)   METALANG99_PRIV_REC_NEXT(638, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_638(choice, ...)   METALANG99_PRIV_REC_NEXT(639, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_639(choice, ...)   METALANG99_PRIV_REC_NEXT(640, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_640(choice, ...)   METALANG99_PRIV_REC_NEXT(641, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_641(choice, ...)   METALANG99_PRIV_REC_NEXT(642, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_642(choice, ...)   METALANG99_PRIV_REC_NEXT(643, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_643(choice, ...)   METALANG99_PRIV_REC_NEXT(644, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_644(choice, ...)   METALANG99_PRIV_REC_NEXT(645, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_645(choice, ...)   METALANG99_PRIV_REC_NEXT(646, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_646(choice, ...)   METALANG99_PRIV_REC_NEXT(647, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_647(choice, ...)   METALANG99_PRIV_REC_NEXT(648, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_648(choice, ...)   METALANG99_PRIV_REC_NEXT(649, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_649(choice, ...)   METALANG99_PRIV_REC_NEXT(650, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_650(choice, ...)   METALANG99_PRIV_REC_NEXT(651, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_651(choice, ...)   METALANG99_PRIV_REC_NEXT(652, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_652(choice, ...)   METALANG99_PRIV_REC_NEXT(653, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_653(choice, ...)   METALANG99_PRIV_REC_NEXT(654, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_654(choice, ...)   METALANG99_PRIV_REC_NEXT(655, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_655(choice, ...)   METALANG99_PRIV_REC_NEXT(656, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_656(choice, ...)   METALANG99_PRIV_REC_NEXT(657, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_657(choice, ...)   METALANG99_PRIV_REC_NEXT(658, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_658(choice, ...)   METALANG99_PRIV_REC_NEXT(659, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_659(choice, ...)   METALANG99_PRIV_REC_NEXT(660, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_660(choice, ...)   METALANG99_PRIV_REC_NEXT(661, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_661(choice, ...)   METALANG99_PRIV_REC_NEXT(662, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_662(choice, ...)   METALANG99_PRIV_REC_NEXT(663, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_663(choice, ...)   METALANG99_PRIV_REC_NEXT(664, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_664(choice, ...)   METALANG99_PRIV_REC_NEXT(665, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_665(choice, ...)   METALANG99_PRIV_REC_NEXT(666, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_666(choice, ...)   METALANG99_PRIV_REC_NEXT(667, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_667(choice, ...)   METALANG99_PRIV_REC_NEXT(668, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_668(choice, ...)   METALANG99_PRIV_REC_NEXT(669, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_669(choice, ...)   METALANG99_PRIV_REC_NEXT(670, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_670(choice, ...)   METALANG99_PRIV_REC_NEXT(671, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_671(choice, ...)   METALANG99_PRIV_REC_NEXT(672, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_672(choice, ...)   METALANG99_PRIV_REC_NEXT(673, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_673(choice, ...)   METALANG99_PRIV_REC_NEXT(674, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_674(choice, ...)   METALANG99_PRIV_REC_NEXT(675, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_675(choice, ...)   METALANG99_PRIV_REC_NEXT(676, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_676(choice, ...)   METALANG99_PRIV_REC_NEXT(677, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_677(choice, ...)   METALANG99_PRIV_REC_NEXT(678, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_678(choice, ...)   METALANG99_PRIV_REC_NEXT(679, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_679(choice, ...)   METALANG99_PRIV_REC_NEXT(680, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_680(choice, ...)   METALANG99_PRIV_REC_NEXT(681, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_681(choice, ...)   METALANG99_PRIV_REC_NEXT(682, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_682(choice, ...)   METALANG99_PRIV_REC_NEXT(683, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_683(choice, ...)   METALANG99_PRIV_REC_NEXT(684, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_684(choice, ...)   METALANG99_PRIV_REC_NEXT(685, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_685(choice, ...)   METALANG99_PRIV_REC_NEXT(686, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_686(choice, ...)   METALANG99_PRIV_REC_NEXT(687, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_687(choice, ...)   METALANG99_PRIV_REC_NEXT(688, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_688(choice, ...)   METALANG99_PRIV_REC_NEXT(689, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_689(choice, ...)   METALANG99_PRIV_REC_NEXT(690, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_690(choice, ...)   METALANG99_PRIV_REC_NEXT(691, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_691(choice, ...)   METALANG99_PRIV_REC_NEXT(692, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_692(choice, ...)   METALANG99_PRIV_REC_NEXT(693, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_693(choice, ...)   METALANG99_PRIV_REC_NEXT(694, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_694(choice, ...)   METALANG99_PRIV_REC_NEXT(695, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_695(choice, ...)   METALANG99_PRIV_REC_NEXT(696, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_696(choice, ...)   METALANG99_PRIV_REC_NEXT(697, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_697(choice, ...)   METALANG99_PRIV_REC_NEXT(698, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_698(choice, ...)   METALANG99_PRIV_REC_NEXT(699, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_699(choice, ...)   METALANG99_PRIV_REC_NEXT(700, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_700(choice, ...)   METALANG99_PRIV_REC_NEXT(701, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_701(choice, ...)   METALANG99_PRIV_REC_NEXT(702, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_702(choice, ...)   METALANG99_PRIV_REC_NEXT(703, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_703(choice, ...)   METALANG99_PRIV_REC_NEXT(704, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_704(choice, ...)   METALANG99_PRIV_REC_NEXT(705, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_705(choice, ...)   METALANG99_PRIV_REC_NEXT(706, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_706(choice, ...)   METALANG99_PRIV_REC_NEXT(707, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_707(choice, ...)   METALANG99_PRIV_REC_NEXT(708, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_708(choice, ...)   METALANG99_PRIV_REC_NEXT(709, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_709(choice, ...)   METALANG99_PRIV_REC_NEXT(710, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_710(choice, ...)   METALANG99_PRIV_REC_NEXT(711, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_711(choice, ...)   METALANG99_PRIV_REC_NEXT(712, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_712(choice, ...)   METALANG99_PRIV_REC_NEXT(713, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_713(choice, ...)   METALANG99_PRIV_REC_NEXT(714, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_714(choice, ...)   METALANG99_PRIV_REC_NEXT(715, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_715(choice, ...)   METALANG99_PRIV_REC_NEXT(716, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_716(choice, ...)   METALANG99_PRIV_REC_NEXT(717, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_717(choice, ...)   METALANG99_PRIV_REC_NEXT(718, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_718(choice, ...)   METALANG99_PRIV_REC_NEXT(719, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_719(choice, ...)   METALANG99_PRIV_REC_NEXT(720, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_720(choice, ...)   METALANG99_PRIV_REC_NEXT(721, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_721(choice, ...)   METALANG99_PRIV_REC_NEXT(722, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_722(choice, ...)   METALANG99_PRIV_REC_NEXT(723, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_723(choice, ...)   METALANG99_PRIV_REC_NEXT(724, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_724(choice, ...)   METALANG99_PRIV_REC_NEXT(725, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_725(choice, ...)   METALANG99_PRIV_REC_NEXT(726, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_726(choice, ...)   METALANG99_PRIV_REC_NEXT(727, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_727(choice, ...)   METALANG99_PRIV_REC_NEXT(728, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_728(choice, ...)   METALANG99_PRIV_REC_NEXT(729, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_729(choice, ...)   METALANG99_PRIV_REC_NEXT(730, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_730(choice, ...)   METALANG99_PRIV_REC_NEXT(731, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_731(choice, ...)   METALANG99_PRIV_REC_NEXT(732, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_732(choice, ...)   METALANG99_PRIV_REC_NEXT(733, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_733(choice, ...)   METALANG99_PRIV_REC_NEXT(734, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_734(choice, ...)   METALANG99_PRIV_REC_NEXT(735, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_735(choice, ...)   METALANG99_PRIV_REC_NEXT(736, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_736(choice, ...)   METALANG99_PRIV_REC_NEXT(737, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_737(choice, ...)   METALANG99_PRIV_REC_NEXT(738, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_738(choice, ...)   METALANG99_PRIV_REC_NEXT(739, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_739(choice, ...)   METALANG99_PRIV_REC_NEXT(740, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_740(choice, ...)   METALANG99_PRIV_REC_NEXT(741, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_741(choice, ...)   METALANG99_PRIV_REC_NEXT(742, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_742(choice, ...)   METALANG99_PRIV_REC_NEXT(743, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_743(choice, ...)   METALANG99_PRIV_REC_NEXT(744, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_744(choice, ...)   METALANG99_PRIV_REC_NEXT(745, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_745(choice, ...)   METALANG99_PRIV_REC_NEXT(746, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_746(choice, ...)   METALANG99_PRIV_REC_NEXT(747, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_747(choice, ...)   METALANG99_PRIV_REC_NEXT(748, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_748(choice, ...)   METALANG99_PRIV_REC_NEXT(749, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_749(choice, ...)   METALANG99_PRIV_REC_NEXT(750, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_750(choice, ...)   METALANG99_PRIV_REC_NEXT(751, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_751(choice, ...)   METALANG99_PRIV_REC_NEXT(752, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_752(choice, ...)   METALANG99_PRIV_REC_NEXT(753, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_753(choice, ...)   METALANG99_PRIV_REC_NEXT(754, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_754(choice, ...)   METALANG99_PRIV_REC_NEXT(755, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_755(choice, ...)   METALANG99_PRIV_REC_NEXT(756, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_756(choice, ...)   METALANG99_PRIV_REC_NEXT(757, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_757(choice, ...)   METALANG99_PRIV_REC_NEXT(758, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_758(choice, ...)   METALANG99_PRIV_REC_NEXT(759, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_759(choice, ...)   METALANG99_PRIV_REC_NEXT(760, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_760(choice, ...)   METALANG99_PRIV_REC_NEXT(761, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_761(choice, ...)   METALANG99_PRIV_REC_NEXT(762, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_762(choice, ...)   METALANG99_PRIV_REC_NEXT(763, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_763(choice, ...)   METALANG99_PRIV_REC_NEXT(764, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_764(choice, ...)   METALANG99_PRIV_REC_NEXT(765, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_765(choice, ...)   METALANG99_PRIV_REC_NEXT(766, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_766(choice, ...)   METALANG99_PRIV_REC_NEXT(767, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_767(choice, ...)   METALANG99_PRIV_REC_NEXT(768, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_768(choice, ...)   METALANG99_PRIV_REC_NEXT(769, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_769(choice, ...)   METALANG99_PRIV_REC_NEXT(770, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_770(choice, ...)   METALANG99_PRIV_REC_NEXT(771, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_771(choice, ...)   METALANG99_PRIV_REC_NEXT(772, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_772(choice, ...)   METALANG99_PRIV_REC_NEXT(773, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_773(choice, ...)   METALANG99_PRIV_REC_NEXT(774, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_774(choice, ...)   METALANG99_PRIV_REC_NEXT(775, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_775(choice, ...)   METALANG99_PRIV_REC_NEXT(776, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_776(choice, ...)   METALANG99_PRIV_REC_NEXT(777, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_777(choice, ...)   METALANG99_PRIV_REC_NEXT(778, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_778(choice, ...)   METALANG99_PRIV_REC_NEXT(779, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_779(choice, ...)   METALANG99_PRIV_REC_NEXT(780, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_780(choice, ...)   METALANG99_PRIV_REC_NEXT(781, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_781(choice, ...)   METALANG99_PRIV_REC_NEXT(782, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_782(choice, ...)   METALANG99_PRIV_REC_NEXT(783, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_783(choice, ...)   METALANG99_PRIV_REC_NEXT(784, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_784(choice, ...)   METALANG99_PRIV_REC_NEXT(785, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_785(choice, ...)   METALANG99_PRIV_REC_NEXT(786, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_786(choice, ...)   METALANG99_PRIV_REC_NEXT(787, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_787(choice, ...)   METALANG99_PRIV_REC_NEXT(788, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_788(choice, ...)   METALANG99_PRIV_REC_NEXT(789, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_789(choice, ...)   METALANG99_PRIV_REC_NEXT(790, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_790(choice, ...)   METALANG99_PRIV_REC_NEXT(791, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_791(choice, ...)   METALANG99_PRIV_REC_NEXT(792, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_792(choice, ...)   METALANG99_PRIV_REC_NEXT(793, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_793(choice, ...)   METALANG99_PRIV_REC_NEXT(794, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_794(choice, ...)   METALANG99_PRIV_REC_NEXT(795, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_795(choice, ...)   METALANG99_PRIV_REC_NEXT(796, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_796(choice, ...)   METALANG99_PRIV_REC_NEXT(797, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_797(choice, ...)   METALANG99_PRIV_REC_NEXT(798, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_798(choice, ...)   METALANG99_PRIV_REC_NEXT(799, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_799(choice, ...)   METALANG99_PRIV_REC_NEXT(800, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_800(choice, ...)   METALANG99_PRIV_REC_NEXT(801, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_801(choice, ...)   METALANG99_PRIV_REC_NEXT(802, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_802(choice, ...)   METALANG99_PRIV_REC_NEXT(803, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_803(choice, ...)   METALANG99_PRIV_REC_NEXT(804, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_804(choice, ...)   METALANG99_PRIV_REC_NEXT(805, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_805(choice, ...)   METALANG99_PRIV_REC_NEXT(806, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_806(choice, ...)   METALANG99_PRIV_REC_NEXT(807, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_807(choice, ...)   METALANG99_PRIV_REC_NEXT(808, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_808(choice, ...)   METALANG99_PRIV_REC_NEXT(809, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_809(choice, ...)   METALANG99_PRIV_REC_NEXT(810, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_810(choice, ...)   METALANG99_PRIV_REC_NEXT(811, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_811(choice, ...)   METALANG99_PRIV_REC_NEXT(812, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_812(choice, ...)   METALANG99_PRIV_REC_NEXT(813, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_813(choice, ...)   METALANG99_PRIV_REC_NEXT(814, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_814(choice, ...)   METALANG99_PRIV_REC_NEXT(815, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_815(choice, ...)   METALANG99_PRIV_REC_NEXT(816, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_816(choice, ...)   METALANG99_PRIV_REC_NEXT(817, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_817(choice, ...)   METALANG99_PRIV_REC_NEXT(818, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_818(choice, ...)   METALANG99_PRIV_REC_NEXT(819, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_819(choice, ...)   METALANG99_PRIV_REC_NEXT(820, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_820(choice, ...)   METALANG99_PRIV_REC_NEXT(821, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_821(choice, ...)   METALANG99_PRIV_REC_NEXT(822, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_822(choice, ...)   METALANG99_PRIV_REC_NEXT(823, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_823(choice, ...)   METALANG99_PRIV_REC_NEXT(824, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_824(choice, ...)   METALANG99_PRIV_REC_NEXT(825, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_825(choice, ...)   METALANG99_PRIV_REC_NEXT(826, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_826(choice, ...)   METALANG99_PRIV_REC_NEXT(827, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_827(choice, ...)   METALANG99_PRIV_REC_NEXT(828, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_828(choice, ...)   METALANG99_PRIV_REC_NEXT(829, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_829(choice, ...)   METALANG99_PRIV_REC_NEXT(830, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_830(choice, ...)   METALANG99_PRIV_REC_NEXT(831, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_831(choice, ...)   METALANG99_PRIV_REC_NEXT(832, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_832(choice, ...)   METALANG99_PRIV_REC_NEXT(833, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_833(choice, ...)   METALANG99_PRIV_REC_NEXT(834, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_834(choice, ...)   METALANG99_PRIV_REC_NEXT(835, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_835(choice, ...)   METALANG99_PRIV_REC_NEXT(836, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_836(choice, ...)   METALANG99_PRIV_REC_NEXT(837, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_837(choice, ...)   METALANG99_PRIV_REC_NEXT(838, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_838(choice, ...)   METALANG99_PRIV_REC_NEXT(839, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_839(choice, ...)   METALANG99_PRIV_REC_NEXT(840, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_840(choice, ...)   METALANG99_PRIV_REC_NEXT(841, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_841(choice, ...)   METALANG99_PRIV_REC_NEXT(842, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_842(choice, ...)   METALANG99_PRIV_REC_NEXT(843, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_843(choice, ...)   METALANG99_PRIV_REC_NEXT(844, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_844(choice, ...)   METALANG99_PRIV_REC_NEXT(845, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_845(choice, ...)   METALANG99_PRIV_REC_NEXT(846, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_846(choice, ...)   METALANG99_PRIV_REC_NEXT(847, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_847(choice, ...)   METALANG99_PRIV_REC_NEXT(848, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_848(choice, ...)   METALANG99_PRIV_REC_NEXT(849, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_849(choice, ...)   METALANG99_PRIV_REC_NEXT(850, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_850(choice, ...)   METALANG99_PRIV_REC_NEXT(851, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_851(choice, ...)   METALANG99_PRIV_REC_NEXT(852, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_852(choice, ...)   METALANG99_PRIV_REC_NEXT(853, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_853(choice, ...)   METALANG99_PRIV_REC_NEXT(854, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_854(choice, ...)   METALANG99_PRIV_REC_NEXT(855, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_855(choice, ...)   METALANG99_PRIV_REC_NEXT(856, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_856(choice, ...)   METALANG99_PRIV_REC_NEXT(857, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_857(choice, ...)   METALANG99_PRIV_REC_NEXT(858, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_858(choice, ...)   METALANG99_PRIV_REC_NEXT(859, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_859(choice, ...)   METALANG99_PRIV_REC_NEXT(860, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_860(choice, ...)   METALANG99_PRIV_REC_NEXT(861, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_861(choice, ...)   METALANG99_PRIV_REC_NEXT(862, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_862(choice, ...)   METALANG99_PRIV_REC_NEXT(863, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_863(choice, ...)   METALANG99_PRIV_REC_NEXT(864, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_864(choice, ...)   METALANG99_PRIV_REC_NEXT(865, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_865(choice, ...)   METALANG99_PRIV_REC_NEXT(866, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_866(choice, ...)   METALANG99_PRIV_REC_NEXT(867, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_867(choice, ...)   METALANG99_PRIV_REC_NEXT(868, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_868(choice, ...)   METALANG99_PRIV_REC_NEXT(869, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_869(choice, ...)   METALANG99_PRIV_REC_NEXT(870, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_870(choice, ...)   METALANG99_PRIV_REC_NEXT(871, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_871(choice, ...)   METALANG99_PRIV_REC_NEXT(872, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_872(choice, ...)   METALANG99_PRIV_REC_NEXT(873, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_873(choice, ...)   METALANG99_PRIV_REC_NEXT(874, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_874(choice, ...)   METALANG99_PRIV_REC_NEXT(875, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_875(choice, ...)   METALANG99_PRIV_REC_NEXT(876, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_876(choice, ...)   METALANG99_PRIV_REC_NEXT(877, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_877(choice, ...)   METALANG99_PRIV_REC_NEXT(878, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_878(choice, ...)   METALANG99_PRIV_REC_NEXT(879, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_879(choice, ...)   METALANG99_PRIV_REC_NEXT(880, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_880(choice, ...)   METALANG99_PRIV_REC_NEXT(881, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_881(choice, ...)   METALANG99_PRIV_REC_NEXT(882, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_882(choice, ...)   METALANG99_PRIV_REC_NEXT(883, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_883(choice, ...)   METALANG99_PRIV_REC_NEXT(884, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_884(choice, ...)   METALANG99_PRIV_REC_NEXT(885, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_885(choice, ...)   METALANG99_PRIV_REC_NEXT(886, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_886(choice, ...)   METALANG99_PRIV_REC_NEXT(887, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_887(choice, ...)   METALANG99_PRIV_REC_NEXT(888, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_888(choice, ...)   METALANG99_PRIV_REC_NEXT(889, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_889(choice, ...)   METALANG99_PRIV_REC_NEXT(890, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_890(choice, ...)   METALANG99_PRIV_REC_NEXT(891, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_891(choice, ...)   METALANG99_PRIV_REC_NEXT(892, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_892(choice, ...)   METALANG99_PRIV_REC_NEXT(893, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_893(choice, ...)   METALANG99_PRIV_REC_NEXT(894, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_894(choice, ...)   METALANG99_PRIV_REC_NEXT(895, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_895(choice, ...)   METALANG99_PRIV_REC_NEXT(896, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_896(choice, ...)   METALANG99_PRIV_REC_NEXT(897, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_897(choice, ...)   METALANG99_PRIV_REC_NEXT(898, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_898(choice, ...)   METALANG99_PRIV_REC_NEXT(899, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_899(choice, ...)   METALANG99_PRIV_REC_NEXT(900, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_900(choice, ...)   METALANG99_PRIV_REC_NEXT(901, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_901(choice, ...)   METALANG99_PRIV_REC_NEXT(902, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_902(choice, ...)   METALANG99_PRIV_REC_NEXT(903, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_903(choice, ...)   METALANG99_PRIV_REC_NEXT(904, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_904(choice, ...)   METALANG99_PRIV_REC_NEXT(905, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_905(choice, ...)   METALANG99_PRIV_REC_NEXT(906, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_906(choice, ...)   METALANG99_PRIV_REC_NEXT(907, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_907(choice, ...)   METALANG99_PRIV_REC_NEXT(908, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_908(choice, ...)   METALANG99_PRIV_REC_NEXT(909, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_909(choice, ...)   METALANG99_PRIV_REC_NEXT(910, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_910(choice, ...)   METALANG99_PRIV_REC_NEXT(911, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_911(choice, ...)   METALANG99_PRIV_REC_NEXT(912, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_912(choice, ...)   METALANG99_PRIV_REC_NEXT(913, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_913(choice, ...)   METALANG99_PRIV_REC_NEXT(914, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_914(choice, ...)   METALANG99_PRIV_REC_NEXT(915, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_915(choice, ...)   METALANG99_PRIV_REC_NEXT(916, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_916(choice, ...)   METALANG99_PRIV_REC_NEXT(917, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_917(choice, ...)   METALANG99_PRIV_REC_NEXT(918, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_918(choice, ...)   METALANG99_PRIV_REC_NEXT(919, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_919(choice, ...)   METALANG99_PRIV_REC_NEXT(920, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_920(choice, ...)   METALANG99_PRIV_REC_NEXT(921, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_921(choice, ...)   METALANG99_PRIV_REC_NEXT(922, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_922(choice, ...)   METALANG99_PRIV_REC_NEXT(923, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_923(choice, ...)   METALANG99_PRIV_REC_NEXT(924, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_924(choice, ...)   METALANG99_PRIV_REC_NEXT(925, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_925(choice, ...)   METALANG99_PRIV_REC_NEXT(926, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_926(choice, ...)   METALANG99_PRIV_REC_NEXT(927, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_927(choice, ...)   METALANG99_PRIV_REC_NEXT(928, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_928(choice, ...)   METALANG99_PRIV_REC_NEXT(929, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_929(choice, ...)   METALANG99_PRIV_REC_NEXT(930, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_930(choice, ...)   METALANG99_PRIV_REC_NEXT(931, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_931(choice, ...)   METALANG99_PRIV_REC_NEXT(932, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_932(choice, ...)   METALANG99_PRIV_REC_NEXT(933, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_933(choice, ...)   METALANG99_PRIV_REC_NEXT(934, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_934(choice, ...)   METALANG99_PRIV_REC_NEXT(935, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_935(choice, ...)   METALANG99_PRIV_REC_NEXT(936, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_936(choice, ...)   METALANG99_PRIV_REC_NEXT(937, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_937(choice, ...)   METALANG99_PRIV_REC_NEXT(938, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_938(choice, ...)   METALANG99_PRIV_REC_NEXT(939, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_939(choice, ...)   METALANG99_PRIV_REC_NEXT(940, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_940(choice, ...)   METALANG99_PRIV_REC_NEXT(941, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_941(choice, ...)   METALANG99_PRIV_REC_NEXT(942, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_942(choice, ...)   METALANG99_PRIV_REC_NEXT(943, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_943(choice, ...)   METALANG99_PRIV_REC_NEXT(944, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_944(choice, ...)   METALANG99_PRIV_REC_NEXT(945, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_945(choice, ...)   METALANG99_PRIV_REC_NEXT(946, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_946(choice, ...)   METALANG99_PRIV_REC_NEXT(947, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_947(choice, ...)   METALANG99_PRIV_REC_NEXT(948, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_948(choice, ...)   METALANG99_PRIV_REC_NEXT(949, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_949(choice, ...)   METALANG99_PRIV_REC_NEXT(950, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_950(choice, ...)   METALANG99_PRIV_REC_NEXT(951, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_951(choice, ...)   METALANG99_PRIV_REC_NEXT(952, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_952(choice, ...)   METALANG99_PRIV_REC_NEXT(953, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_953(choice, ...)   METALANG99_PRIV_REC_NEXT(954, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_954(choice, ...)   METALANG99_PRIV_REC_NEXT(955, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_955(choice, ...)   METALANG99_PRIV_REC_NEXT(956, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_956(choice, ...)   METALANG99_PRIV_REC_NEXT(957, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_957(choice, ...)   METALANG99_PRIV_REC_NEXT(958, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_958(choice, ...)   METALANG99_PRIV_REC_NEXT(959, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_959(choice, ...)   METALANG99_PRIV_REC_NEXT(960, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_960(choice, ...)   METALANG99_PRIV_REC_NEXT(961, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_961(choice, ...)   METALANG99_PRIV_REC_NEXT(962, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_962(choice, ...)   METALANG99_PRIV_REC_NEXT(963, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_963(choice, ...)   METALANG99_PRIV_REC_NEXT(964, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_964(choice, ...)   METALANG99_PRIV_REC_NEXT(965, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_965(choice, ...)   METALANG99_PRIV_REC_NEXT(966, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_966(choice, ...)   METALANG99_PRIV_REC_NEXT(967, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_967(choice, ...)   METALANG99_PRIV_REC_NEXT(968, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_968(choice, ...)   METALANG99_PRIV_REC_NEXT(969, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_969(choice, ...)   METALANG99_PRIV_REC_NEXT(970, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_970(choice, ...)   METALANG99_PRIV_REC_NEXT(971, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_971(choice, ...)   METALANG99_PRIV_REC_NEXT(972, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_972(choice, ...)   METALANG99_PRIV_REC_NEXT(973, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_973(choice, ...)   METALANG99_PRIV_REC_NEXT(974, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_974(choice, ...)   METALANG99_PRIV_REC_NEXT(975, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_975(choice, ...)   METALANG99_PRIV_REC_NEXT(976, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_976(choice, ...)   METALANG99_PRIV_REC_NEXT(977, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_977(choice, ...)   METALANG99_PRIV_REC_NEXT(978, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_978(choice, ...)   METALANG99_PRIV_REC_NEXT(979, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_979(choice, ...)   METALANG99_PRIV_REC_NEXT(980, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_980(choice, ...)   METALANG99_PRIV_REC_NEXT(981, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_981(choice, ...)   METALANG99_PRIV_REC_NEXT(982, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_982(choice, ...)   METALANG99_PRIV_REC_NEXT(983, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_983(choice, ...)   METALANG99_PRIV_REC_NEXT(984, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_984(choice, ...)   METALANG99_PRIV_REC_NEXT(985, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_985(choice, ...)   METALANG99_PRIV_REC_NEXT(986, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_986(choice, ...)   METALANG99_PRIV_REC_NEXT(987, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_987(choice, ...)   METALANG99_PRIV_REC_NEXT(988, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_988(choice, ...)   METALANG99_PRIV_REC_NEXT(989, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_989(choice, ...)   METALANG99_PRIV_REC_NEXT(990, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_990(choice, ...)   METALANG99_PRIV_REC_NEXT(991, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_991(choice, ...)   METALANG99_PRIV_REC_NEXT(992, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_992(choice, ...)   METALANG99_PRIV_REC_NEXT(993, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_993(choice, ...)   METALANG99_PRIV_REC_NEXT(994, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_994(choice, ...)   METALANG99_PRIV_REC_NEXT(995, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_995(choice, ...)   METALANG99_PRIV_REC_NEXT(996, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_996(choice, ...)   METALANG99_PRIV_REC_NEXT(997, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_997(choice, ...)   METALANG99_PRIV_REC_NEXT(998, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_998(choice, ...)   METALANG99_PRIV_REC_NEXT(999, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_999(choice, ...)   METALANG99_PRIV_REC_NEXT(1000, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1000(choice, ...)  METALANG99_PRIV_REC_NEXT(1001, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1001(choice, ...)  METALANG99_PRIV_REC_NEXT(1002, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1002(choice, ...)  METALANG99_PRIV_REC_NEXT(1003, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1003(choice, ...)  METALANG99_PRIV_REC_NEXT(1004, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1004(choice, ...)  METALANG99_PRIV_REC_NEXT(1005, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1005(choice, ...)  METALANG99_PRIV_REC_NEXT(1006, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1006(choice, ...)  METALANG99_PRIV_REC_NEXT(1007, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1007(choice, ...)  METALANG99_PRIV_REC_NEXT(1008, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1008(choice, ...)  METALANG99_PRIV_REC_NEXT(1009, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1009(choice, ...)  METALANG99_PRIV_REC_NEXT(1010, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1010(choice, ...)  METALANG99_PRIV_REC_NEXT(1011, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1011(choice, ...)  METALANG99_PRIV_REC_NEXT(1012, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1012(choice, ...)  METALANG99_PRIV_REC_NEXT(1013, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1013(choice, ...)  METALANG99_PRIV_REC_NEXT(1014, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1014(choice, ...)  METALANG99_PRIV_REC_NEXT(1015, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1015(choice, ...)  METALANG99_PRIV_REC_NEXT(1016, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1016(choice, ...)  METALANG99_PRIV_REC_NEXT(1017, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1017(choice, ...)  METALANG99_PRIV_REC_NEXT(1018, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1018(choice, ...)  METALANG99_PRIV_REC_NEXT(1019, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1019(choice, ...)  METALANG99_PRIV_REC_NEXT(1020, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1020(choice, ...)  METALANG99_PRIV_REC_NEXT(1021, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1021(choice, ...)  METALANG99_PRIV_REC_NEXT(1022, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1022(choice, ...)  METALANG99_PRIV_REC_NEXT(1023, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1023(choice, ...)  METALANG99_PRIV_REC_NEXT(1024, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1024(choice, ...)  METALANG99_PRIV_REC_NEXT(1025, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1025(choice, ...)  METALANG99_PRIV_REC_NEXT(1026, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1026(choice, ...)  METALANG99_PRIV_REC_NEXT(1027, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1027(choice, ...)  METALANG99_PRIV_REC_NEXT(1028, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1028(choice, ...)  METALANG99_PRIV_REC_NEXT(1029, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1029(choice, ...)  METALANG99_PRIV_REC_NEXT(1030, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1030(choice, ...)  METALANG99_PRIV_REC_NEXT(1031, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1031(choice, ...)  METALANG99_PRIV_REC_NEXT(1032, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1032(choice, ...)  METALANG99_PRIV_REC_NEXT(1033, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1033(choice, ...)  METALANG99_PRIV_REC_NEXT(1034, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1034(choice, ...)  METALANG99_PRIV_REC_NEXT(1035, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1035(choice, ...)  METALANG99_PRIV_REC_NEXT(1036, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1036(choice, ...)  METALANG99_PRIV_REC_NEXT(1037, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1037(choice, ...)  METALANG99_PRIV_REC_NEXT(1038, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1038(choice, ...)  METALANG99_PRIV_REC_NEXT(1039, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1039(choice, ...)  METALANG99_PRIV_REC_NEXT(1040, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1040(choice, ...)  METALANG99_PRIV_REC_NEXT(1041, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1041(choice, ...)  METALANG99_PRIV_REC_NEXT(1042, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1042(choice, ...)  METALANG99_PRIV_REC_NEXT(1043, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1043(choice, ...)  METALANG99_PRIV_REC_NEXT(1044, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1044(choice, ...)  METALANG99_PRIV_REC_NEXT(1045, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1045(choice, ...)  METALANG99_PRIV_REC_NEXT(1046, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1046(choice, ...)  METALANG99_PRIV_REC_NEXT(1047, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1047(choice, ...)  METALANG99_PRIV_REC_NEXT(1048, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1048(choice, ...)  METALANG99_PRIV_REC_NEXT(1049, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1049(choice, ...)  METALANG99_PRIV_REC_NEXT(1050, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1050(choice, ...)  METALANG99_PRIV_REC_NEXT(1051, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1051(choice, ...)  METALANG99_PRIV_REC_NEXT(1052, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1052(choice, ...)  METALANG99_PRIV_REC_NEXT(1053, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1053(choice, ...)  METALANG99_PRIV_REC_NEXT(1054, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1054(choice, ...)  METALANG99_PRIV_REC_NEXT(1055, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1055(choice, ...)  METALANG99_PRIV_REC_NEXT(1056, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1056(choice, ...)  METALANG99_PRIV_REC_NEXT(1057, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1057(choice, ...)  METALANG99_PRIV_REC_NEXT(1058, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1058(choice, ...)  METALANG99_PRIV_REC_NEXT(1059, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1059(choice, ...)  METALANG99_PRIV_REC_NEXT(1060, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1060(choice, ...)  METALANG99_PRIV_REC_NEXT(1061, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1061(choice, ...)  METALANG99_PRIV_REC_NEXT(1062, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1062(choice, ...)  METALANG99_PRIV_REC_NEXT(1063, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1063(choice, ...)  METALANG99_PRIV_REC_NEXT(1064, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1064(choice, ...)  METALANG99_PRIV_REC_NEXT(1065, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1065(choice, ...)  METALANG99_PRIV_REC_NEXT(1066, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1066(choice, ...)  METALANG99_PRIV_REC_NEXT(1067, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1067(choice, ...)  METALANG99_PRIV_REC_NEXT(1068, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1068(choice, ...)  METALANG99_PRIV_REC_NEXT(1069, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1069(choice, ...)  METALANG99_PRIV_REC_NEXT(1070, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1070(choice, ...)  METALANG99_PRIV_REC_NEXT(1071, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1071(choice, ...)  METALANG99_PRIV_REC_NEXT(1072, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1072(choice, ...)  METALANG99_PRIV_REC_NEXT(1073, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1073(choice, ...)  METALANG99_PRIV_REC_NEXT(1074, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1074(choice, ...)  METALANG99_PRIV_REC_NEXT(1075, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1075(choice, ...)  METALANG99_PRIV_REC_NEXT(1076, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1076(choice, ...)  METALANG99_PRIV_REC_NEXT(1077, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1077(choice, ...)  METALANG99_PRIV_REC_NEXT(1078, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1078(choice, ...)  METALANG99_PRIV_REC_NEXT(1079, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1079(choice, ...)  METALANG99_PRIV_REC_NEXT(1080, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1080(choice, ...)  METALANG99_PRIV_REC_NEXT(1081, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1081(choice, ...)  METALANG99_PRIV_REC_NEXT(1082, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1082(choice, ...)  METALANG99_PRIV_REC_NEXT(1083, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1083(choice, ...)  METALANG99_PRIV_REC_NEXT(1084, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1084(choice, ...)  METALANG99_PRIV_REC_NEXT(1085, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1085(choice, ...)  METALANG99_PRIV_REC_NEXT(1086, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1086(choice, ...)  METALANG99_PRIV_REC_NEXT(1087, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1087(choice, ...)  METALANG99_PRIV_REC_NEXT(1088, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1088(choice, ...)  METALANG99_PRIV_REC_NEXT(1089, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1089(choice, ...)  METALANG99_PRIV_REC_NEXT(1090, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1090(choice, ...)  METALANG99_PRIV_REC_NEXT(1091, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1091(choice, ...)  METALANG99_PRIV_REC_NEXT(1092, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1092(choice, ...)  METALANG99_PRIV_REC_NEXT(1093, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1093(choice, ...)  METALANG99_PRIV_REC_NEXT(1094, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1094(choice, ...)  METALANG99_PRIV_REC_NEXT(1095, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1095(choice, ...)  METALANG99_PRIV_REC_NEXT(1096, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1096(choice, ...)  METALANG99_PRIV_REC_NEXT(1097, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1097(choice, ...)  METALANG99_PRIV_REC_NEXT(1098, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1098(choice, ...)  METALANG99_PRIV_REC_NEXT(1099, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1099(choice, ...)  METALANG99_PRIV_REC_NEXT(1100, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1100(choice, ...)  METALANG99_PRIV_REC_NEXT(1101, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1101(choice, ...)  METALANG99_PRIV_REC_NEXT(1102, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1102(choice, ...)  METALANG99_PRIV_REC_NEXT(1103, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1103(choice, ...)  METALANG99_PRIV_REC_NEXT(1104, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1104(choice, ...)  METALANG99_PRIV_REC_NEXT(1105, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1105(choice, ...)  METALANG99_PRIV_REC_NEXT(1106, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1106(choice, ...)  METALANG99_PRIV_REC_NEXT(1107, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1107(choice, ...)  METALANG99_PRIV_REC_NEXT(1108, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1108(choice, ...)  METALANG99_PRIV_REC_NEXT(1109, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1109(choice, ...)  METALANG99_PRIV_REC_NEXT(1110, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1110(choice, ...)  METALANG99_PRIV_REC_NEXT(1111, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1111(choice, ...)  METALANG99_PRIV_REC_NEXT(1112, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1112(choice, ...)  METALANG99_PRIV_REC_NEXT(1113, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1113(choice, ...)  METALANG99_PRIV_REC_NEXT(1114, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1114(choice, ...)  METALANG99_PRIV_REC_NEXT(1115, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1115(choice, ...)  METALANG99_PRIV_REC_NEXT(1116, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1116(choice, ...)  METALANG99_PRIV_REC_NEXT(1117, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1117(choice, ...)  METALANG99_PRIV_REC_NEXT(1118, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1118(choice, ...)  METALANG99_PRIV_REC_NEXT(1119, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1119(choice, ...)  METALANG99_PRIV_REC_NEXT(1120, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1120(choice, ...)  METALANG99_PRIV_REC_NEXT(1121, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1121(choice, ...)  METALANG99_PRIV_REC_NEXT(1122, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1122(choice, ...)  METALANG99_PRIV_REC_NEXT(1123, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1123(choice, ...)  METALANG99_PRIV_REC_NEXT(1124, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1124(choice, ...)  METALANG99_PRIV_REC_NEXT(1125, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1125(choice, ...)  METALANG99_PRIV_REC_NEXT(1126, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1126(choice, ...)  METALANG99_PRIV_REC_NEXT(1127, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1127(choice, ...)  METALANG99_PRIV_REC_NEXT(1128, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1128(choice, ...)  METALANG99_PRIV_REC_NEXT(1129, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1129(choice, ...)  METALANG99_PRIV_REC_NEXT(1130, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1130(choice, ...)  METALANG99_PRIV_REC_NEXT(1131, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1131(choice, ...)  METALANG99_PRIV_REC_NEXT(1132, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1132(choice, ...)  METALANG99_PRIV_REC_NEXT(1133, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1133(choice, ...)  METALANG99_PRIV_REC_NEXT(1134, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1134(choice, ...)  METALANG99_PRIV_REC_NEXT(1135, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1135(choice, ...)  METALANG99_PRIV_REC_NEXT(1136, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1136(choice, ...)  METALANG99_PRIV_REC_NEXT(1137, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1137(choice, ...)  METALANG99_PRIV_REC_NEXT(1138, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1138(choice, ...)  METALANG99_PRIV_REC_NEXT(1139, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1139(choice, ...)  METALANG99_PRIV_REC_NEXT(1140, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1140(choice, ...)  METALANG99_PRIV_REC_NEXT(1141, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1141(choice, ...)  METALANG99_PRIV_REC_NEXT(1142, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1142(choice, ...)  METALANG99_PRIV_REC_NEXT(1143, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1143(choice, ...)  METALANG99_PRIV_REC_NEXT(1144, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1144(choice, ...)  METALANG99_PRIV_REC_NEXT(1145, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1145(choice, ...)  METALANG99_PRIV_REC_NEXT(1146, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1146(choice, ...)  METALANG99_PRIV_REC_NEXT(1147, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1147(choice, ...)  METALANG99_PRIV_REC_NEXT(1148, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1148(choice, ...)  METALANG99_PRIV_REC_NEXT(1149, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1149(choice, ...)  METALANG99_PRIV_REC_NEXT(1150, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1150(choice, ...)  METALANG99_PRIV_REC_NEXT(1151, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1151(choice, ...)  METALANG99_PRIV_REC_NEXT(1152, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1152(choice, ...)  METALANG99_PRIV_REC_NEXT(1153, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1153(choice, ...)  METALANG99_PRIV_REC_NEXT(1154, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1154(choice, ...)  METALANG99_PRIV_REC_NEXT(1155, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1155(choice, ...)  METALANG99_PRIV_REC_NEXT(1156, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1156(choice, ...)  METALANG99_PRIV_REC_NEXT(1157, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1157(choice, ...)  METALANG99_PRIV_REC_NEXT(1158, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1158(choice, ...)  METALANG99_PRIV_REC_NEXT(1159, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1159(choice, ...)  METALANG99_PRIV_REC_NEXT(1160, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1160(choice, ...)  METALANG99_PRIV_REC_NEXT(1161, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1161(choice, ...)  METALANG99_PRIV_REC_NEXT(1162, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1162(choice, ...)  METALANG99_PRIV_REC_NEXT(1163, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1163(choice, ...)  METALANG99_PRIV_REC_NEXT(1164, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1164(choice, ...)  METALANG99_PRIV_REC_NEXT(1165, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1165(choice, ...)  METALANG99_PRIV_REC_NEXT(1166, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1166(choice, ...)  METALANG99_PRIV_REC_NEXT(1167, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1167(choice, ...)  METALANG99_PRIV_REC_NEXT(1168, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1168(choice, ...)  METALANG99_PRIV_REC_NEXT(1169, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1169(choice, ...)  METALANG99_PRIV_REC_NEXT(1170, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1170(choice, ...)  METALANG99_PRIV_REC_NEXT(1171, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1171(choice, ...)  METALANG99_PRIV_REC_NEXT(1172, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1172(choice, ...)  METALANG99_PRIV_REC_NEXT(1173, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1173(choice, ...)  METALANG99_PRIV_REC_NEXT(1174, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1174(choice, ...)  METALANG99_PRIV_REC_NEXT(1175, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1175(choice, ...)  METALANG99_PRIV_REC_NEXT(1176, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1176(choice, ...)  METALANG99_PRIV_REC_NEXT(1177, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1177(choice, ...)  METALANG99_PRIV_REC_NEXT(1178, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1178(choice, ...)  METALANG99_PRIV_REC_NEXT(1179, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1179(choice, ...)  METALANG99_PRIV_REC_NEXT(1180, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1180(choice, ...)  METALANG99_PRIV_REC_NEXT(1181, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1181(choice, ...)  METALANG99_PRIV_REC_NEXT(1182, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1182(choice, ...)  METALANG99_PRIV_REC_NEXT(1183, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1183(choice, ...)  METALANG99_PRIV_REC_NEXT(1184, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1184(choice, ...)  METALANG99_PRIV_REC_NEXT(1185, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1185(choice, ...)  METALANG99_PRIV_REC_NEXT(1186, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1186(choice, ...)  METALANG99_PRIV_REC_NEXT(1187, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1187(choice, ...)  METALANG99_PRIV_REC_NEXT(1188, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1188(choice, ...)  METALANG99_PRIV_REC_NEXT(1189, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1189(choice, ...)  METALANG99_PRIV_REC_NEXT(1190, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1190(choice, ...)  METALANG99_PRIV_REC_NEXT(1191, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1191(choice, ...)  METALANG99_PRIV_REC_NEXT(1192, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1192(choice, ...)  METALANG99_PRIV_REC_NEXT(1193, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1193(choice, ...)  METALANG99_PRIV_REC_NEXT(1194, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1194(choice, ...)  METALANG99_PRIV_REC_NEXT(1195, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1195(choice, ...)  METALANG99_PRIV_REC_NEXT(1196, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1196(choice, ...)  METALANG99_PRIV_REC_NEXT(1197, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1197(choice, ...)  METALANG99_PRIV_REC_NEXT(1198, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1198(choice, ...)  METALANG99_PRIV_REC_NEXT(1199, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1199(choice, ...)  METALANG99_PRIV_REC_NEXT(1200, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1200(choice, ...)  METALANG99_PRIV_REC_NEXT(1201, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1201(choice, ...)  METALANG99_PRIV_REC_NEXT(1202, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1202(choice, ...)  METALANG99_PRIV_REC_NEXT(1203, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1203(choice, ...)  METALANG99_PRIV_REC_NEXT(1204, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1204(choice, ...)  METALANG99_PRIV_REC_NEXT(1205, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1205(choice, ...)  METALANG99_PRIV_REC_NEXT(1206, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1206(choice, ...)  METALANG99_PRIV_REC_NEXT(1207, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1207(choice, ...)  METALANG99_PRIV_REC_NEXT(1208, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1208(choice, ...)  METALANG99_PRIV_REC_NEXT(1209, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1209(choice, ...)  METALANG99_PRIV_REC_NEXT(1210, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1210(choice, ...)  METALANG99_PRIV_REC_NEXT(1211, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1211(choice, ...)  METALANG99_PRIV_REC_NEXT(1212, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1212(choice, ...)  METALANG99_PRIV_REC_NEXT(1213, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1213(choice, ...)  METALANG99_PRIV_REC_NEXT(1214, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1214(choice, ...)  METALANG99_PRIV_REC_NEXT(1215, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1215(choice, ...)  METALANG99_PRIV_REC_NEXT(1216, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1216(choice, ...)  METALANG99_PRIV_REC_NEXT(1217, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1217(choice, ...)  METALANG99_PRIV_REC_NEXT(1218, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1218(choice, ...)  METALANG99_PRIV_REC_NEXT(1219, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1219(choice, ...)  METALANG99_PRIV_REC_NEXT(1220, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1220(choice, ...)  METALANG99_PRIV_REC_NEXT(1221, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1221(choice, ...)  METALANG99_PRIV_REC_NEXT(1222, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1222(choice, ...)  METALANG99_PRIV_REC_NEXT(1223, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1223(choice, ...)  METALANG99_PRIV_REC_NEXT(1224, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1224(choice, ...)  METALANG99_PRIV_REC_NEXT(1225, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1225(choice, ...)  METALANG99_PRIV_REC_NEXT(1226, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1226(choice, ...)  METALANG99_PRIV_REC_NEXT(1227, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1227(choice, ...)  METALANG99_PRIV_REC_NEXT(1228, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1228(choice, ...)  METALANG99_PRIV_REC_NEXT(1229, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1229(choice, ...)  METALANG99_PRIV_REC_NEXT(1230, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1230(choice, ...)  METALANG99_PRIV_REC_NEXT(1231, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1231(choice, ...)  METALANG99_PRIV_REC_NEXT(1232, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1232(choice, ...)  METALANG99_PRIV_REC_NEXT(1233, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1233(choice, ...)  METALANG99_PRIV_REC_NEXT(1234, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1234(choice, ...)  METALANG99_PRIV_REC_NEXT(1235, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1235(choice, ...)  METALANG99_PRIV_REC_NEXT(1236, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1236(choice, ...)  METALANG99_PRIV_REC_NEXT(1237, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1237(choice, ...)  METALANG99_PRIV_REC_NEXT(1238, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1238(choice, ...)  METALANG99_PRIV_REC_NEXT(1239, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1239(choice, ...)  METALANG99_PRIV_REC_NEXT(1240, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1240(choice, ...)  METALANG99_PRIV_REC_NEXT(1241, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1241(choice, ...)  METALANG99_PRIV_REC_NEXT(1242, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1242(choice, ...)  METALANG99_PRIV_REC_NEXT(1243, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1243(choice, ...)  METALANG99_PRIV_REC_NEXT(1244, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1244(choice, ...)  METALANG99_PRIV_REC_NEXT(1245, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1245(choice, ...)  METALANG99_PRIV_REC_NEXT(1246, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1246(choice, ...)  METALANG99_PRIV_REC_NEXT(1247, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1247(choice, ...)  METALANG99_PRIV_REC_NEXT(1248, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1248(choice, ...)  METALANG99_PRIV_REC_NEXT(1249, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1249(choice, ...)  METALANG99_PRIV_REC_NEXT(1250, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1250(choice, ...)  METALANG99_PRIV_REC_NEXT(1251, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1251(choice, ...)  METALANG99_PRIV_REC_NEXT(1252, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1252(choice, ...)  METALANG99_PRIV_REC_NEXT(1253, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1253(choice, ...)  METALANG99_PRIV_REC_NEXT(1254, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1254(choice, ...)  METALANG99_PRIV_REC_NEXT(1255, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1255(choice, ...)  METALANG99_PRIV_REC_NEXT(1256, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1256(choice, ...)  METALANG99_PRIV_REC_NEXT(1257, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1257(choice, ...)  METALANG99_PRIV_REC_NEXT(1258, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1258(choice, ...)  METALANG99_PRIV_REC_NEXT(1259, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1259(choice, ...)  METALANG99_PRIV_REC_NEXT(1260, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1260(choice, ...)  METALANG99_PRIV_REC_NEXT(1261, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1261(choice, ...)  METALANG99_PRIV_REC_NEXT(1262, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1262(choice, ...)  METALANG99_PRIV_REC_NEXT(1263, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1263(choice, ...)  METALANG99_PRIV_REC_NEXT(1264, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1264(choice, ...)  METALANG99_PRIV_REC_NEXT(1265, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1265(choice, ...)  METALANG99_PRIV_REC_NEXT(1266, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1266(choice, ...)  METALANG99_PRIV_REC_NEXT(1267, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1267(choice, ...)  METALANG99_PRIV_REC_NEXT(1268, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1268(choice, ...)  METALANG99_PRIV_REC_NEXT(1269, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1269(choice, ...)  METALANG99_PRIV_REC_NEXT(1270, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1270(choice, ...)  METALANG99_PRIV_REC_NEXT(1271, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1271(choice, ...)  METALANG99_PRIV_REC_NEXT(1272, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1272(choice, ...)  METALANG99_PRIV_REC_NEXT(1273, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1273(choice, ...)  METALANG99_PRIV_REC_NEXT(1274, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1274(choice, ...)  METALANG99_PRIV_REC_NEXT(1275, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1275(choice, ...)  METALANG99_PRIV_REC_NEXT(1276, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1276(choice, ...)  METALANG99_PRIV_REC_NEXT(1277, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1277(choice, ...)  METALANG99_PRIV_REC_NEXT(1278, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1278(choice, ...)  METALANG99_PRIV_REC_NEXT(1279, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1279(choice, ...)  METALANG99_PRIV_REC_NEXT(1280, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1280(choice, ...)  METALANG99_PRIV_REC_NEXT(1281, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1281(choice, ...)  METALANG99_PRIV_REC_NEXT(1282, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1282(choice, ...)  METALANG99_PRIV_REC_NEXT(1283, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1283(choice, ...)  METALANG99_PRIV_REC_NEXT(1284, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1284(choice, ...)  METALANG99_PRIV_REC_NEXT(1285, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1285(choice, ...)  METALANG99_PRIV_REC_NEXT(1286, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1286(choice, ...)  METALANG99_PRIV_REC_NEXT(1287, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1287(choice, ...)  METALANG99_PRIV_REC_NEXT(1288, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1288(choice, ...)  METALANG99_PRIV_REC_NEXT(1289, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1289(choice, ...)  METALANG99_PRIV_REC_NEXT(1290, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1290(choice, ...)  METALANG99_PRIV_REC_NEXT(1291, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1291(choice, ...)  METALANG99_PRIV_REC_NEXT(1292, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1292(choice, ...)  METALANG99_PRIV_REC_NEXT(1293, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1293(choice, ...)  METALANG99_PRIV_REC_NEXT(1294, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1294(choice, ...)  METALANG99_PRIV_REC_NEXT(1295, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1295(choice, ...)  METALANG99_PRIV_REC_NEXT(1296, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1296(choice, ...)  METALANG99_PRIV_REC_NEXT(1297, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1297(choice, ...)  METALANG99_PRIV_REC_NEXT(1298, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1298(choice, ...)  METALANG99_PRIV_REC_NEXT(1299, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1299(choice, ...)  METALANG99_PRIV_REC_NEXT(1300, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1300(choice, ...)  METALANG99_PRIV_REC_NEXT(1301, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1301(choice, ...)  METALANG99_PRIV_REC_NEXT(1302, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1302(choice, ...)  METALANG99_PRIV_REC_NEXT(1303, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1303(choice, ...)  METALANG99_PRIV_REC_NEXT(1304, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1304(choice, ...)  METALANG99_PRIV_REC_NEXT(1305, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1305(choice, ...)  METALANG99_PRIV_REC_NEXT(1306, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1306(choice, ...)  METALANG99_PRIV_REC_NEXT(1307, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1307(choice, ...)  METALANG99_PRIV_REC_NEXT(1308, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1308(choice, ...)  METALANG99_PRIV_REC_NEXT(1309, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1309(choice, ...)  METALANG99_PRIV_REC_NEXT(1310, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1310(choice, ...)  METALANG99_PRIV_REC_NEXT(1311, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1311(choice, ...)  METALANG99_PRIV_REC_NEXT(1312, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1312(choice, ...)  METALANG99_PRIV_REC_NEXT(1313, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1313(choice, ...)  METALANG99_PRIV_REC_NEXT(1314, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1314(choice, ...)  METALANG99_PRIV_REC_NEXT(1315, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1315(choice, ...)  METALANG99_PRIV_REC_NEXT(1316, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1316(choice, ...)  METALANG99_PRIV_REC_NEXT(1317, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1317(choice, ...)  METALANG99_PRIV_REC_NEXT(1318, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1318(choice, ...)  METALANG99_PRIV_REC_NEXT(1319, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1319(choice, ...)  METALANG99_PRIV_REC_NEXT(1320, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1320(choice, ...)  METALANG99_PRIV_REC_NEXT(1321, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1321(choice, ...)  METALANG99_PRIV_REC_NEXT(1322, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1322(choice, ...)  METALANG99_PRIV_REC_NEXT(1323, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1323(choice, ...)  METALANG99_PRIV_REC_NEXT(1324, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1324(choice, ...)  METALANG99_PRIV_REC_NEXT(1325, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1325(choice, ...)  METALANG99_PRIV_REC_NEXT(1326, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1326(choice, ...)  METALANG99_PRIV_REC_NEXT(1327, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1327(choice, ...)  METALANG99_PRIV_REC_NEXT(1328, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1328(choice, ...)  METALANG99_PRIV_REC_NEXT(1329, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1329(choice, ...)  METALANG99_PRIV_REC_NEXT(1330, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1330(choice, ...)  METALANG99_PRIV_REC_NEXT(1331, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1331(choice, ...)  METALANG99_PRIV_REC_NEXT(1332, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1332(choice, ...)  METALANG99_PRIV_REC_NEXT(1333, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1333(choice, ...)  METALANG99_PRIV_REC_NEXT(1334, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1334(choice, ...)  METALANG99_PRIV_REC_NEXT(1335, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1335(choice, ...)  METALANG99_PRIV_REC_NEXT(1336, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1336(choice, ...)  METALANG99_PRIV_REC_NEXT(1337, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1337(choice, ...)  METALANG99_PRIV_REC_NEXT(1338, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1338(choice, ...)  METALANG99_PRIV_REC_NEXT(1339, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1339(choice, ...)  METALANG99_PRIV_REC_NEXT(1340, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1340(choice, ...)  METALANG99_PRIV_REC_NEXT(1341, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1341(choice, ...)  METALANG99_PRIV_REC_NEXT(1342, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1342(choice, ...)  METALANG99_PRIV_REC_NEXT(1343, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1343(choice, ...)  METALANG99_PRIV_REC_NEXT(1344, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1344(choice, ...)  METALANG99_PRIV_REC_NEXT(1345, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1345(choice, ...)  METALANG99_PRIV_REC_NEXT(1346, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1346(choice, ...)  METALANG99_PRIV_REC_NEXT(1347, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1347(choice, ...)  METALANG99_PRIV_REC_NEXT(1348, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1348(choice, ...)  METALANG99_PRIV_REC_NEXT(1349, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1349(choice, ...)  METALANG99_PRIV_REC_NEXT(1350, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1350(choice, ...)  METALANG99_PRIV_REC_NEXT(1351, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1351(choice, ...)  METALANG99_PRIV_REC_NEXT(1352, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1352(choice, ...)  METALANG99_PRIV_REC_NEXT(1353, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1353(choice, ...)  METALANG99_PRIV_REC_NEXT(1354, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1354(choice, ...)  METALANG99_PRIV_REC_NEXT(1355, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1355(choice, ...)  METALANG99_PRIV_REC_NEXT(1356, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1356(choice, ...)  METALANG99_PRIV_REC_NEXT(1357, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1357(choice, ...)  METALANG99_PRIV_REC_NEXT(1358, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1358(choice, ...)  METALANG99_PRIV_REC_NEXT(1359, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1359(choice, ...)  METALANG99_PRIV_REC_NEXT(1360, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1360(choice, ...)  METALANG99_PRIV_REC_NEXT(1361, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1361(choice, ...)  METALANG99_PRIV_REC_NEXT(1362, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1362(choice, ...)  METALANG99_PRIV_REC_NEXT(1363, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1363(choice, ...)  METALANG99_PRIV_REC_NEXT(1364, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1364(choice, ...)  METALANG99_PRIV_REC_NEXT(1365, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1365(choice, ...)  METALANG99_PRIV_REC_NEXT(1366, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1366(choice, ...)  METALANG99_PRIV_REC_NEXT(1367, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1367(choice, ...)  METALANG99_PRIV_REC_NEXT(1368, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1368(choice, ...)  METALANG99_PRIV_REC_NEXT(1369, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1369(choice, ...)  METALANG99_PRIV_REC_NEXT(1370, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1370(choice, ...)  METALANG99_PRIV_REC_NEXT(1371, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1371(choice, ...)  METALANG99_PRIV_REC_NEXT(1372, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1372(choice, ...)  METALANG99_PRIV_REC_NEXT(1373, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1373(choice, ...)  METALANG99_PRIV_REC_NEXT(1374, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1374(choice, ...)  METALANG99_PRIV_REC_NEXT(1375, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1375(choice, ...)  METALANG99_PRIV_REC_NEXT(1376, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1376(choice, ...)  METALANG99_PRIV_REC_NEXT(1377, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1377(choice, ...)  METALANG99_PRIV_REC_NEXT(1378, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1378(choice, ...)  METALANG99_PRIV_REC_NEXT(1379, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1379(choice, ...)  METALANG99_PRIV_REC_NEXT(1380, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1380(choice, ...)  METALANG99_PRIV_REC_NEXT(1381, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1381(choice, ...)  METALANG99_PRIV_REC_NEXT(1382, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1382(choice, ...)  METALANG99_PRIV_REC_NEXT(1383, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1383(choice, ...)  METALANG99_PRIV_REC_NEXT(1384, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1384(choice, ...)  METALANG99_PRIV_REC_NEXT(1385, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1385(choice, ...)  METALANG99_PRIV_REC_NEXT(1386, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1386(choice, ...)  METALANG99_PRIV_REC_NEXT(1387, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1387(choice, ...)  METALANG99_PRIV_REC_NEXT(1388, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1388(choice, ...)  METALANG99_PRIV_REC_NEXT(1389, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1389(choice, ...)  METALANG99_PRIV_REC_NEXT(1390, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1390(choice, ...)  METALANG99_PRIV_REC_NEXT(1391, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1391(choice, ...)  METALANG99_PRIV_REC_NEXT(1392, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1392(choice, ...)  METALANG99_PRIV_REC_NEXT(1393, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1393(choice, ...)  METALANG99_PRIV_REC_NEXT(1394, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1394(choice, ...)  METALANG99_PRIV_REC_NEXT(1395, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1395(choice, ...)  METALANG99_PRIV_REC_NEXT(1396, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1396(choice, ...)  METALANG99_PRIV_REC_NEXT(1397, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1397(choice, ...)  METALANG99_PRIV_REC_NEXT(1398, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1398(choice, ...)  METALANG99_PRIV_REC_NEXT(1399, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1399(choice, ...)  METALANG99_PRIV_REC_NEXT(1400, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1400(choice, ...)  METALANG99_PRIV_REC_NEXT(1401, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1401(choice, ...)  METALANG99_PRIV_REC_NEXT(1402, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1402(choice, ...)  METALANG99_PRIV_REC_NEXT(1403, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1403(choice, ...)  METALANG99_PRIV_REC_NEXT(1404, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1404(choice, ...)  METALANG99_PRIV_REC_NEXT(1405, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1405(choice, ...)  METALANG99_PRIV_REC_NEXT(1406, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1406(choice, ...)  METALANG99_PRIV_REC_NEXT(1407, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1407(choice, ...)  METALANG99_PRIV_REC_NEXT(1408, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1408(choice, ...)  METALANG99_PRIV_REC_NEXT(1409, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1409(choice, ...)  METALANG99_PRIV_REC_NEXT(1410, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1410(choice, ...)  METALANG99_PRIV_REC_NEXT(1411, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1411(choice, ...)  METALANG99_PRIV_REC_NEXT(1412, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1412(choice, ...)  METALANG99_PRIV_REC_NEXT(1413, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1413(choice, ...)  METALANG99_PRIV_REC_NEXT(1414, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1414(choice, ...)  METALANG99_PRIV_REC_NEXT(1415, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1415(choice, ...)  METALANG99_PRIV_REC_NEXT(1416, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1416(choice, ...)  METALANG99_PRIV_REC_NEXT(1417, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1417(choice, ...)  METALANG99_PRIV_REC_NEXT(1418, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1418(choice, ...)  METALANG99_PRIV_REC_NEXT(1419, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1419(choice, ...)  METALANG99_PRIV_REC_NEXT(1420, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1420(choice, ...)  METALANG99_PRIV_REC_NEXT(1421, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1421(choice, ...)  METALANG99_PRIV_REC_NEXT(1422, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1422(choice, ...)  METALANG99_PRIV_REC_NEXT(1423, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1423(choice, ...)  METALANG99_PRIV_REC_NEXT(1424, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1424(choice, ...)  METALANG99_PRIV_REC_NEXT(1425, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1425(choice, ...)  METALANG99_PRIV_REC_NEXT(1426, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1426(choice, ...)  METALANG99_PRIV_REC_NEXT(1427, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1427(choice, ...)  METALANG99_PRIV_REC_NEXT(1428, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1428(choice, ...)  METALANG99_PRIV_REC_NEXT(1429, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1429(choice, ...)  METALANG99_PRIV_REC_NEXT(1430, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1430(choice, ...)  METALANG99_PRIV_REC_NEXT(1431, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1431(choice, ...)  METALANG99_PRIV_REC_NEXT(1432, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1432(choice, ...)  METALANG99_PRIV_REC_NEXT(1433, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1433(choice, ...)  METALANG99_PRIV_REC_NEXT(1434, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1434(choice, ...)  METALANG99_PRIV_REC_NEXT(1435, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1435(choice, ...)  METALANG99_PRIV_REC_NEXT(1436, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1436(choice, ...)  METALANG99_PRIV_REC_NEXT(1437, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1437(choice, ...)  METALANG99_PRIV_REC_NEXT(1438, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1438(choice, ...)  METALANG99_PRIV_REC_NEXT(1439, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1439(choice, ...)  METALANG99_PRIV_REC_NEXT(1440, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1440(choice, ...)  METALANG99_PRIV_REC_NEXT(1441, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1441(choice, ...)  METALANG99_PRIV_REC_NEXT(1442, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1442(choice, ...)  METALANG99_PRIV_REC_NEXT(1443, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1443(choice, ...)  METALANG99_PRIV_REC_NEXT(1444, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1444(choice, ...)  METALANG99_PRIV_REC_NEXT(1445, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1445(choice, ...)  METALANG99_PRIV_REC_NEXT(1446, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1446(choice, ...)  METALANG99_PRIV_REC_NEXT(1447, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1447(choice, ...)  METALANG99_PRIV_REC_NEXT(1448, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1448(choice, ...)  METALANG99_PRIV_REC_NEXT(1449, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1449(choice, ...)  METALANG99_PRIV_REC_NEXT(1450, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1450(choice, ...)  METALANG99_PRIV_REC_NEXT(1451, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1451(choice, ...)  METALANG99_PRIV_REC_NEXT(1452, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1452(choice, ...)  METALANG99_PRIV_REC_NEXT(1453, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1453(choice, ...)  METALANG99_PRIV_REC_NEXT(1454, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1454(choice, ...)  METALANG99_PRIV_REC_NEXT(1455, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1455(choice, ...)  METALANG99_PRIV_REC_NEXT(1456, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1456(choice, ...)  METALANG99_PRIV_REC_NEXT(1457, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1457(choice, ...)  METALANG99_PRIV_REC_NEXT(1458, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1458(choice, ...)  METALANG99_PRIV_REC_NEXT(1459, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1459(choice, ...)  METALANG99_PRIV_REC_NEXT(1460, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1460(choice, ...)  METALANG99_PRIV_REC_NEXT(1461, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1461(choice, ...)  METALANG99_PRIV_REC_NEXT(1462, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1462(choice, ...)  METALANG99_PRIV_REC_NEXT(1463, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1463(choice, ...)  METALANG99_PRIV_REC_NEXT(1464, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1464(choice, ...)  METALANG99_PRIV_REC_NEXT(1465, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1465(choice, ...)  METALANG99_PRIV_REC_NEXT(1466, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1466(choice, ...)  METALANG99_PRIV_REC_NEXT(1467, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1467(choice, ...)  METALANG99_PRIV_REC_NEXT(1468, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1468(choice, ...)  METALANG99_PRIV_REC_NEXT(1469, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1469(choice, ...)  METALANG99_PRIV_REC_NEXT(1470, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1470(choice, ...)  METALANG99_PRIV_REC_NEXT(1471, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1471(choice, ...)  METALANG99_PRIV_REC_NEXT(1472, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1472(choice, ...)  METALANG99_PRIV_REC_NEXT(1473, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1473(choice, ...)  METALANG99_PRIV_REC_NEXT(1474, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1474(choice, ...)  METALANG99_PRIV_REC_NEXT(1475, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1475(choice, ...)  METALANG99_PRIV_REC_NEXT(1476, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1476(choice, ...)  METALANG99_PRIV_REC_NEXT(1477, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1477(choice, ...)  METALANG99_PRIV_REC_NEXT(1478, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1478(choice, ...)  METALANG99_PRIV_REC_NEXT(1479, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1479(choice, ...)  METALANG99_PRIV_REC_NEXT(1480, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1480(choice, ...)  METALANG99_PRIV_REC_NEXT(1481, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1481(choice, ...)  METALANG99_PRIV_REC_NEXT(1482, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1482(choice, ...)  METALANG99_PRIV_REC_NEXT(1483, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1483(choice, ...)  METALANG99_PRIV_REC_NEXT(1484, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1484(choice, ...)  METALANG99_PRIV_REC_NEXT(1485, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1485(choice, ...)  METALANG99_PRIV_REC_NEXT(1486, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1486(choice, ...)  METALANG99_PRIV_REC_NEXT(1487, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1487(choice, ...)  METALANG99_PRIV_REC_NEXT(1488, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1488(choice, ...)  METALANG99_PRIV_REC_NEXT(1489, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1489(choice, ...)  METALANG99_PRIV_REC_NEXT(1490, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1490(choice, ...)  METALANG99_PRIV_REC_NEXT(1491, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1491(choice, ...)  METALANG99_PRIV_REC_NEXT(1492, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1492(choice, ...)  METALANG99_PRIV_REC_NEXT(1493, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1493(choice, ...)  METALANG99_PRIV_REC_NEXT(1494, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1494(choice, ...)  METALANG99_PRIV_REC_NEXT(1495, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1495(choice, ...)  METALANG99_PRIV_REC_NEXT(1496, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1496(choice, ...)  METALANG99_PRIV_REC_NEXT(1497, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1497(choice, ...)  METALANG99_PRIV_REC_NEXT(1498, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1498(choice, ...)  METALANG99_PRIV_REC_NEXT(1499, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1499(choice, ...)  METALANG99_PRIV_REC_NEXT(1500, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1500(choice, ...)  METALANG99_PRIV_REC_NEXT(1501, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1501(choice, ...)  METALANG99_PRIV_REC_NEXT(1502, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1502(choice, ...)  METALANG99_PRIV_REC_NEXT(1503, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1503(choice, ...)  METALANG99_PRIV_REC_NEXT(1504, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1504(choice, ...)  METALANG99_PRIV_REC_NEXT(1505, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1505(choice, ...)  METALANG99_PRIV_REC_NEXT(1506, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1506(choice, ...)  METALANG99_PRIV_REC_NEXT(1507, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1507(choice, ...)  METALANG99_PRIV_REC_NEXT(1508, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1508(choice, ...)  METALANG99_PRIV_REC_NEXT(1509, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1509(choice, ...)  METALANG99_PRIV_REC_NEXT(1510, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1510(choice, ...)  METALANG99_PRIV_REC_NEXT(1511, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1511(choice, ...)  METALANG99_PRIV_REC_NEXT(1512, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1512(choice, ...)  METALANG99_PRIV_REC_NEXT(1513, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1513(choice, ...)  METALANG99_PRIV_REC_NEXT(1514, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1514(choice, ...)  METALANG99_PRIV_REC_NEXT(1515, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1515(choice, ...)  METALANG99_PRIV_REC_NEXT(1516, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1516(choice, ...)  METALANG99_PRIV_REC_NEXT(1517, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1517(choice, ...)  METALANG99_PRIV_REC_NEXT(1518, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1518(choice, ...)  METALANG99_PRIV_REC_NEXT(1519, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1519(choice, ...)  METALANG99_PRIV_REC_NEXT(1520, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1520(choice, ...)  METALANG99_PRIV_REC_NEXT(1521, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1521(choice, ...)  METALANG99_PRIV_REC_NEXT(1522, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1522(choice, ...)  METALANG99_PRIV_REC_NEXT(1523, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1523(choice, ...)  METALANG99_PRIV_REC_NEXT(1524, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1524(choice, ...)  METALANG99_PRIV_REC_NEXT(1525, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1525(choice, ...)  METALANG99_PRIV_REC_NEXT(1526, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1526(choice, ...)  METALANG99_PRIV_REC_NEXT(1527, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1527(choice, ...)  METALANG99_PRIV_REC_NEXT(1528, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1528(choice, ...)  METALANG99_PRIV_REC_NEXT(1529, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1529(choice, ...)  METALANG99_PRIV_REC_NEXT(1530, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1530(choice, ...)  METALANG99_PRIV_REC_NEXT(1531, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1531(choice, ...)  METALANG99_PRIV_REC_NEXT(1532, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1532(choice, ...)  METALANG99_PRIV_REC_NEXT(1533, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1533(choice, ...)  METALANG99_PRIV_REC_NEXT(1534, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1534(choice, ...)  METALANG99_PRIV_REC_NEXT(1535, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1535(choice, ...)  METALANG99_PRIV_REC_NEXT(1536, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1536(choice, ...)  METALANG99_PRIV_REC_NEXT(1537, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1537(choice, ...)  METALANG99_PRIV_REC_NEXT(1538, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1538(choice, ...)  METALANG99_PRIV_REC_NEXT(1539, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1539(choice, ...)  METALANG99_PRIV_REC_NEXT(1540, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1540(choice, ...)  METALANG99_PRIV_REC_NEXT(1541, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1541(choice, ...)  METALANG99_PRIV_REC_NEXT(1542, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1542(choice, ...)  METALANG99_PRIV_REC_NEXT(1543, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1543(choice, ...)  METALANG99_PRIV_REC_NEXT(1544, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1544(choice, ...)  METALANG99_PRIV_REC_NEXT(1545, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1545(choice, ...)  METALANG99_PRIV_REC_NEXT(1546, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1546(choice, ...)  METALANG99_PRIV_REC_NEXT(1547, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1547(choice, ...)  METALANG99_PRIV_REC_NEXT(1548, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1548(choice, ...)  METALANG99_PRIV_REC_NEXT(1549, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1549(choice, ...)  METALANG99_PRIV_REC_NEXT(1550, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1550(choice, ...)  METALANG99_PRIV_REC_NEXT(1551, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1551(choice, ...)  METALANG99_PRIV_REC_NEXT(1552, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1552(choice, ...)  METALANG99_PRIV_REC_NEXT(1553, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1553(choice, ...)  METALANG99_PRIV_REC_NEXT(1554, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1554(choice, ...)  METALANG99_PRIV_REC_NEXT(1555, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1555(choice, ...)  METALANG99_PRIV_REC_NEXT(1556, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1556(choice, ...)  METALANG99_PRIV_REC_NEXT(1557, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1557(choice, ...)  METALANG99_PRIV_REC_NEXT(1558, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1558(choice, ...)  METALANG99_PRIV_REC_NEXT(1559, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1559(choice, ...)  METALANG99_PRIV_REC_NEXT(1560, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1560(choice, ...)  METALANG99_PRIV_REC_NEXT(1561, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1561(choice, ...)  METALANG99_PRIV_REC_NEXT(1562, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1562(choice, ...)  METALANG99_PRIV_REC_NEXT(1563, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1563(choice, ...)  METALANG99_PRIV_REC_NEXT(1564, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1564(choice, ...)  METALANG99_PRIV_REC_NEXT(1565, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1565(choice, ...)  METALANG99_PRIV_REC_NEXT(1566, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1566(choice, ...)  METALANG99_PRIV_REC_NEXT(1567, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1567(choice, ...)  METALANG99_PRIV_REC_NEXT(1568, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1568(choice, ...)  METALANG99_PRIV_REC_NEXT(1569, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1569(choice, ...)  METALANG99_PRIV_REC_NEXT(1570, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1570(choice, ...)  METALANG99_PRIV_REC_NEXT(1571, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1571(choice, ...)  METALANG99_PRIV_REC_NEXT(1572, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1572(choice, ...)  METALANG99_PRIV_REC_NEXT(1573, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1573(choice, ...)  METALANG99_PRIV_REC_NEXT(1574, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1574(choice, ...)  METALANG99_PRIV_REC_NEXT(1575, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1575(choice, ...)  METALANG99_PRIV_REC_NEXT(1576, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1576(choice, ...)  METALANG99_PRIV_REC_NEXT(1577, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1577(choice, ...)  METALANG99_PRIV_REC_NEXT(1578, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1578(choice, ...)  METALANG99_PRIV_REC_NEXT(1579, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1579(choice, ...)  METALANG99_PRIV_REC_NEXT(1580, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1580(choice, ...)  METALANG99_PRIV_REC_NEXT(1581, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1581(choice, ...)  METALANG99_PRIV_REC_NEXT(1582, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1582(choice, ...)  METALANG99_PRIV_REC_NEXT(1583, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1583(choice, ...)  METALANG99_PRIV_REC_NEXT(1584, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1584(choice, ...)  METALANG99_PRIV_REC_NEXT(1585, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1585(choice, ...)  METALANG99_PRIV_REC_NEXT(1586, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1586(choice, ...)  METALANG99_PRIV_REC_NEXT(1587, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1587(choice, ...)  METALANG99_PRIV_REC_NEXT(1588, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1588(choice, ...)  METALANG99_PRIV_REC_NEXT(1589, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1589(choice, ...)  METALANG99_PRIV_REC_NEXT(1590, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1590(choice, ...)  METALANG99_PRIV_REC_NEXT(1591, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1591(choice, ...)  METALANG99_PRIV_REC_NEXT(1592, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1592(choice, ...)  METALANG99_PRIV_REC_NEXT(1593, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1593(choice, ...)  METALANG99_PRIV_REC_NEXT(1594, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1594(choice, ...)  METALANG99_PRIV_REC_NEXT(1595, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1595(choice, ...)  METALANG99_PRIV_REC_NEXT(1596, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1596(choice, ...)  METALANG99_PRIV_REC_NEXT(1597, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1597(choice, ...)  METALANG99_PRIV_REC_NEXT(1598, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1598(choice, ...)  METALANG99_PRIV_REC_NEXT(1599, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1599(choice, ...)  METALANG99_PRIV_REC_NEXT(1600, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1600(choice, ...)  METALANG99_PRIV_REC_NEXT(1601, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1601(choice, ...)  METALANG99_PRIV_REC_NEXT(1602, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1602(choice, ...)  METALANG99_PRIV_REC_NEXT(1603, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1603(choice, ...)  METALANG99_PRIV_REC_NEXT(1604, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1604(choice, ...)  METALANG99_PRIV_REC_NEXT(1605, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1605(choice, ...)  METALANG99_PRIV_REC_NEXT(1606, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1606(choice, ...)  METALANG99_PRIV_REC_NEXT(1607, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1607(choice, ...)  METALANG99_PRIV_REC_NEXT(1608, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1608(choice, ...)  METALANG99_PRIV_REC_NEXT(1609, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1609(choice, ...)  METALANG99_PRIV_REC_NEXT(1610, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1610(choice, ...)  METALANG99_PRIV_REC_NEXT(1611, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1611(choice, ...)  METALANG99_PRIV_REC_NEXT(1612, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1612(choice, ...)  METALANG99_PRIV_REC_NEXT(1613, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1613(choice, ...)  METALANG99_PRIV_REC_NEXT(1614, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1614(choice, ...)  METALANG99_PRIV_REC_NEXT(1615, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1615(choice, ...)  METALANG99_PRIV_REC_NEXT(1616, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1616(choice, ...)  METALANG99_PRIV_REC_NEXT(1617, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1617(choice, ...)  METALANG99_PRIV_REC_NEXT(1618, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1618(choice, ...)  METALANG99_PRIV_REC_NEXT(1619, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1619(choice, ...)  METALANG99_PRIV_REC_NEXT(1620, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1620(choice, ...)  METALANG99_PRIV_REC_NEXT(1621, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1621(choice, ...)  METALANG99_PRIV_REC_NEXT(1622, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1622(choice, ...)  METALANG99_PRIV_REC_NEXT(1623, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1623(choice, ...)  METALANG99_PRIV_REC_NEXT(1624, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1624(choice, ...)  METALANG99_PRIV_REC_NEXT(1625, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1625(choice, ...)  METALANG99_PRIV_REC_NEXT(1626, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1626(choice, ...)  METALANG99_PRIV_REC_NEXT(1627, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1627(choice, ...)  METALANG99_PRIV_REC_NEXT(1628, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1628(choice, ...)  METALANG99_PRIV_REC_NEXT(1629, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1629(choice, ...)  METALANG99_PRIV_REC_NEXT(1630, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1630(choice, ...)  METALANG99_PRIV_REC_NEXT(1631, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1631(choice, ...)  METALANG99_PRIV_REC_NEXT(1632, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1632(choice, ...)  METALANG99_PRIV_REC_NEXT(1633, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1633(choice, ...)  METALANG99_PRIV_REC_NEXT(1634, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1634(choice, ...)  METALANG99_PRIV_REC_NEXT(1635, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1635(choice, ...)  METALANG99_PRIV_REC_NEXT(1636, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1636(choice, ...)  METALANG99_PRIV_REC_NEXT(1637, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1637(choice, ...)  METALANG99_PRIV_REC_NEXT(1638, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1638(choice, ...)  METALANG99_PRIV_REC_NEXT(1639, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1639(choice, ...)  METALANG99_PRIV_REC_NEXT(1640, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1640(choice, ...)  METALANG99_PRIV_REC_NEXT(1641, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1641(choice, ...)  METALANG99_PRIV_REC_NEXT(1642, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1642(choice, ...)  METALANG99_PRIV_REC_NEXT(1643, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1643(choice, ...)  METALANG99_PRIV_REC_NEXT(1644, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1644(choice, ...)  METALANG99_PRIV_REC_NEXT(1645, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1645(choice, ...)  METALANG99_PRIV_REC_NEXT(1646, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1646(choice, ...)  METALANG99_PRIV_REC_NEXT(1647, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1647(choice, ...)  METALANG99_PRIV_REC_NEXT(1648, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1648(choice, ...)  METALANG99_PRIV_REC_NEXT(1649, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1649(choice, ...)  METALANG99_PRIV_REC_NEXT(1650, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1650(choice, ...)  METALANG99_PRIV_REC_NEXT(1651, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1651(choice, ...)  METALANG99_PRIV_REC_NEXT(1652, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1652(choice, ...)  METALANG99_PRIV_REC_NEXT(1653, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1653(choice, ...)  METALANG99_PRIV_REC_NEXT(1654, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1654(choice, ...)  METALANG99_PRIV_REC_NEXT(1655, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1655(choice, ...)  METALANG99_PRIV_REC_NEXT(1656, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1656(choice, ...)  METALANG99_PRIV_REC_NEXT(1657, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1657(choice, ...)  METALANG99_PRIV_REC_NEXT(1658, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1658(choice, ...)  METALANG99_PRIV_REC_NEXT(1659, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1659(choice, ...)  METALANG99_PRIV_REC_NEXT(1660, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1660(choice, ...)  METALANG99_PRIV_REC_NEXT(1661, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1661(choice, ...)  METALANG99_PRIV_REC_NEXT(1662, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1662(choice, ...)  METALANG99_PRIV_REC_NEXT(1663, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1663(choice, ...)  METALANG99_PRIV_REC_NEXT(1664, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1664(choice, ...)  METALANG99_PRIV_REC_NEXT(1665, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1665(choice, ...)  METALANG99_PRIV_REC_NEXT(1666, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1666(choice, ...)  METALANG99_PRIV_REC_NEXT(1667, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1667(choice, ...)  METALANG99_PRIV_REC_NEXT(1668, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1668(choice, ...)  METALANG99_PRIV_REC_NEXT(1669, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1669(choice, ...)  METALANG99_PRIV_REC_NEXT(1670, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1670(choice, ...)  METALANG99_PRIV_REC_NEXT(1671, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1671(choice, ...)  METALANG99_PRIV_REC_NEXT(1672, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1672(choice, ...)  METALANG99_PRIV_REC_NEXT(1673, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1673(choice, ...)  METALANG99_PRIV_REC_NEXT(1674, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1674(choice, ...)  METALANG99_PRIV_REC_NEXT(1675, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1675(choice, ...)  METALANG99_PRIV_REC_NEXT(1676, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1676(choice, ...)  METALANG99_PRIV_REC_NEXT(1677, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1677(choice, ...)  METALANG99_PRIV_REC_NEXT(1678, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1678(choice, ...)  METALANG99_PRIV_REC_NEXT(1679, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1679(choice, ...)  METALANG99_PRIV_REC_NEXT(1680, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1680(choice, ...)  METALANG99_PRIV_REC_NEXT(1681, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1681(choice, ...)  METALANG99_PRIV_REC_NEXT(1682, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1682(choice, ...)  METALANG99_PRIV_REC_NEXT(1683, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1683(choice, ...)  METALANG99_PRIV_REC_NEXT(1684, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1684(choice, ...)  METALANG99_PRIV_REC_NEXT(1685, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1685(choice, ...)  METALANG99_PRIV_REC_NEXT(1686, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1686(choice, ...)  METALANG99_PRIV_REC_NEXT(1687, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1687(choice, ...)  METALANG99_PRIV_REC_NEXT(1688, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1688(choice, ...)  METALANG99_PRIV_REC_NEXT(1689, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1689(choice, ...)  METALANG99_PRIV_REC_NEXT(1690, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1690(choice, ...)  METALANG99_PRIV_REC_NEXT(1691, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1691(choice, ...)  METALANG99_PRIV_REC_NEXT(1692, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1692(choice, ...)  METALANG99_PRIV_REC_NEXT(1693, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1693(choice, ...)  METALANG99_PRIV_REC_NEXT(1694, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1694(choice, ...)  METALANG99_PRIV_REC_NEXT(1695, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1695(choice, ...)  METALANG99_PRIV_REC_NEXT(1696, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1696(choice, ...)  METALANG99_PRIV_REC_NEXT(1697, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1697(choice, ...)  METALANG99_PRIV_REC_NEXT(1698, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1698(choice, ...)  METALANG99_PRIV_REC_NEXT(1699, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1699(choice, ...)  METALANG99_PRIV_REC_NEXT(1700, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1700(choice, ...)  METALANG99_PRIV_REC_NEXT(1701, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1701(choice, ...)  METALANG99_PRIV_REC_NEXT(1702, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1702(choice, ...)  METALANG99_PRIV_REC_NEXT(1703, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1703(choice, ...)  METALANG99_PRIV_REC_NEXT(1704, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1704(choice, ...)  METALANG99_PRIV_REC_NEXT(1705, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1705(choice, ...)  METALANG99_PRIV_REC_NEXT(1706, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1706(choice, ...)  METALANG99_PRIV_REC_NEXT(1707, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1707(choice, ...)  METALANG99_PRIV_REC_NEXT(1708, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1708(choice, ...)  METALANG99_PRIV_REC_NEXT(1709, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1709(choice, ...)  METALANG99_PRIV_REC_NEXT(1710, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1710(choice, ...)  METALANG99_PRIV_REC_NEXT(1711, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1711(choice, ...)  METALANG99_PRIV_REC_NEXT(1712, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1712(choice, ...)  METALANG99_PRIV_REC_NEXT(1713, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1713(choice, ...)  METALANG99_PRIV_REC_NEXT(1714, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1714(choice, ...)  METALANG99_PRIV_REC_NEXT(1715, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1715(choice, ...)  METALANG99_PRIV_REC_NEXT(1716, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1716(choice, ...)  METALANG99_PRIV_REC_NEXT(1717, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1717(choice, ...)  METALANG99_PRIV_REC_NEXT(1718, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1718(choice, ...)  METALANG99_PRIV_REC_NEXT(1719, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1719(choice, ...)  METALANG99_PRIV_REC_NEXT(1720, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1720(choice, ...)  METALANG99_PRIV_REC_NEXT(1721, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1721(choice, ...)  METALANG99_PRIV_REC_NEXT(1722, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1722(choice, ...)  METALANG99_PRIV_REC_NEXT(1723, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1723(choice, ...)  METALANG99_PRIV_REC_NEXT(1724, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1724(choice, ...)  METALANG99_PRIV_REC_NEXT(1725, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1725(choice, ...)  METALANG99_PRIV_REC_NEXT(1726, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1726(choice, ...)  METALANG99_PRIV_REC_NEXT(1727, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1727(choice, ...)  METALANG99_PRIV_REC_NEXT(1728, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1728(choice, ...)  METALANG99_PRIV_REC_NEXT(1729, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1729(choice, ...)  METALANG99_PRIV_REC_NEXT(1730, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1730(choice, ...)  METALANG99_PRIV_REC_NEXT(1731, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1731(choice, ...)  METALANG99_PRIV_REC_NEXT(1732, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1732(choice, ...)  METALANG99_PRIV_REC_NEXT(1733, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1733(choice, ...)  METALANG99_PRIV_REC_NEXT(1734, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1734(choice, ...)  METALANG99_PRIV_REC_NEXT(1735, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1735(choice, ...)  METALANG99_PRIV_REC_NEXT(1736, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1736(choice, ...)  METALANG99_PRIV_REC_NEXT(1737, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1737(choice, ...)  METALANG99_PRIV_REC_NEXT(1738, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1738(choice, ...)  METALANG99_PRIV_REC_NEXT(1739, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1739(choice, ...)  METALANG99_PRIV_REC_NEXT(1740, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1740(choice, ...)  METALANG99_PRIV_REC_NEXT(1741, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1741(choice, ...)  METALANG99_PRIV_REC_NEXT(1742, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1742(choice, ...)  METALANG99_PRIV_REC_NEXT(1743, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1743(choice, ...)  METALANG99_PRIV_REC_NEXT(1744, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1744(choice, ...)  METALANG99_PRIV_REC_NEXT(1745, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1745(choice, ...)  METALANG99_PRIV_REC_NEXT(1746, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1746(choice, ...)  METALANG99_PRIV_REC_NEXT(1747, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1747(choice, ...)  METALANG99_PRIV_REC_NEXT(1748, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1748(choice, ...)  METALANG99_PRIV_REC_NEXT(1749, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1749(choice, ...)  METALANG99_PRIV_REC_NEXT(1750, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1750(choice, ...)  METALANG99_PRIV_REC_NEXT(1751, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1751(choice, ...)  METALANG99_PRIV_REC_NEXT(1752, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1752(choice, ...)  METALANG99_PRIV_REC_NEXT(1753, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1753(choice, ...)  METALANG99_PRIV_REC_NEXT(1754, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1754(choice, ...)  METALANG99_PRIV_REC_NEXT(1755, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1755(choice, ...)  METALANG99_PRIV_REC_NEXT(1756, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1756(choice, ...)  METALANG99_PRIV_REC_NEXT(1757, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1757(choice, ...)  METALANG99_PRIV_REC_NEXT(1758, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1758(choice, ...)  METALANG99_PRIV_REC_NEXT(1759, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1759(choice, ...)  METALANG99_PRIV_REC_NEXT(1760, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1760(choice, ...)  METALANG99_PRIV_REC_NEXT(1761, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1761(choice, ...)  METALANG99_PRIV_REC_NEXT(1762, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1762(choice, ...)  METALANG99_PRIV_REC_NEXT(1763, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1763(choice, ...)  METALANG99_PRIV_REC_NEXT(1764, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1764(choice, ...)  METALANG99_PRIV_REC_NEXT(1765, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1765(choice, ...)  METALANG99_PRIV_REC_NEXT(1766, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1766(choice, ...)  METALANG99_PRIV_REC_NEXT(1767, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1767(choice, ...)  METALANG99_PRIV_REC_NEXT(1768, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1768(choice, ...)  METALANG99_PRIV_REC_NEXT(1769, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1769(choice, ...)  METALANG99_PRIV_REC_NEXT(1770, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1770(choice, ...)  METALANG99_PRIV_REC_NEXT(1771, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1771(choice, ...)  METALANG99_PRIV_REC_NEXT(1772, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1772(choice, ...)  METALANG99_PRIV_REC_NEXT(1773, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1773(choice, ...)  METALANG99_PRIV_REC_NEXT(1774, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1774(choice, ...)  METALANG99_PRIV_REC_NEXT(1775, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1775(choice, ...)  METALANG99_PRIV_REC_NEXT(1776, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1776(choice, ...)  METALANG99_PRIV_REC_NEXT(1777, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1777(choice, ...)  METALANG99_PRIV_REC_NEXT(1778, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1778(choice, ...)  METALANG99_PRIV_REC_NEXT(1779, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1779(choice, ...)  METALANG99_PRIV_REC_NEXT(1780, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1780(choice, ...)  METALANG99_PRIV_REC_NEXT(1781, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1781(choice, ...)  METALANG99_PRIV_REC_NEXT(1782, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1782(choice, ...)  METALANG99_PRIV_REC_NEXT(1783, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1783(choice, ...)  METALANG99_PRIV_REC_NEXT(1784, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1784(choice, ...)  METALANG99_PRIV_REC_NEXT(1785, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1785(choice, ...)  METALANG99_PRIV_REC_NEXT(1786, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1786(choice, ...)  METALANG99_PRIV_REC_NEXT(1787, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1787(choice, ...)  METALANG99_PRIV_REC_NEXT(1788, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1788(choice, ...)  METALANG99_PRIV_REC_NEXT(1789, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1789(choice, ...)  METALANG99_PRIV_REC_NEXT(1790, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1790(choice, ...)  METALANG99_PRIV_REC_NEXT(1791, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1791(choice, ...)  METALANG99_PRIV_REC_NEXT(1792, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1792(choice, ...)  METALANG99_PRIV_REC_NEXT(1793, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1793(choice, ...)  METALANG99_PRIV_REC_NEXT(1794, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1794(choice, ...)  METALANG99_PRIV_REC_NEXT(1795, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1795(choice, ...)  METALANG99_PRIV_REC_NEXT(1796, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1796(choice, ...)  METALANG99_PRIV_REC_NEXT(1797, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1797(choice, ...)  METALANG99_PRIV_REC_NEXT(1798, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1798(choice, ...)  METALANG99_PRIV_REC_NEXT(1799, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1799(choice, ...)  METALANG99_PRIV_REC_NEXT(1800, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1800(choice, ...)  METALANG99_PRIV_REC_NEXT(1801, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1801(choice, ...)  METALANG99_PRIV_REC_NEXT(1802, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1802(choice, ...)  METALANG99_PRIV_REC_NEXT(1803, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1803(choice, ...)  METALANG99_PRIV_REC_NEXT(1804, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1804(choice, ...)  METALANG99_PRIV_REC_NEXT(1805, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1805(choice, ...)  METALANG99_PRIV_REC_NEXT(1806, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1806(choice, ...)  METALANG99_PRIV_REC_NEXT(1807, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1807(choice, ...)  METALANG99_PRIV_REC_NEXT(1808, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1808(choice, ...)  METALANG99_PRIV_REC_NEXT(1809, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1809(choice, ...)  METALANG99_PRIV_REC_NEXT(1810, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1810(choice, ...)  METALANG99_PRIV_REC_NEXT(1811, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1811(choice, ...)  METALANG99_PRIV_REC_NEXT(1812, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1812(choice, ...)  METALANG99_PRIV_REC_NEXT(1813, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1813(choice, ...)  METALANG99_PRIV_REC_NEXT(1814, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1814(choice, ...)  METALANG99_PRIV_REC_NEXT(1815, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1815(choice, ...)  METALANG99_PRIV_REC_NEXT(1816, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1816(choice, ...)  METALANG99_PRIV_REC_NEXT(1817, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1817(choice, ...)  METALANG99_PRIV_REC_NEXT(1818, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1818(choice, ...)  METALANG99_PRIV_REC_NEXT(1819, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1819(choice, ...)  METALANG99_PRIV_REC_NEXT(1820, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1820(choice, ...)  METALANG99_PRIV_REC_NEXT(1821, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1821(choice, ...)  METALANG99_PRIV_REC_NEXT(1822, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1822(choice, ...)  METALANG99_PRIV_REC_NEXT(1823, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1823(choice, ...)  METALANG99_PRIV_REC_NEXT(1824, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1824(choice, ...)  METALANG99_PRIV_REC_NEXT(1825, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1825(choice, ...)  METALANG99_PRIV_REC_NEXT(1826, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1826(choice, ...)  METALANG99_PRIV_REC_NEXT(1827, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1827(choice, ...)  METALANG99_PRIV_REC_NEXT(1828, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1828(choice, ...)  METALANG99_PRIV_REC_NEXT(1829, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1829(choice, ...)  METALANG99_PRIV_REC_NEXT(1830, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1830(choice, ...)  METALANG99_PRIV_REC_NEXT(1831, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1831(choice, ...)  METALANG99_PRIV_REC_NEXT(1832, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1832(choice, ...)  METALANG99_PRIV_REC_NEXT(1833, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1833(choice, ...)  METALANG99_PRIV_REC_NEXT(1834, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1834(choice, ...)  METALANG99_PRIV_REC_NEXT(1835, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1835(choice, ...)  METALANG99_PRIV_REC_NEXT(1836, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1836(choice, ...)  METALANG99_PRIV_REC_NEXT(1837, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1837(choice, ...)  METALANG99_PRIV_REC_NEXT(1838, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1838(choice, ...)  METALANG99_PRIV_REC_NEXT(1839, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1839(choice, ...)  METALANG99_PRIV_REC_NEXT(1840, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1840(choice, ...)  METALANG99_PRIV_REC_NEXT(1841, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1841(choice, ...)  METALANG99_PRIV_REC_NEXT(1842, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1842(choice, ...)  METALANG99_PRIV_REC_NEXT(1843, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1843(choice, ...)  METALANG99_PRIV_REC_NEXT(1844, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1844(choice, ...)  METALANG99_PRIV_REC_NEXT(1845, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1845(choice, ...)  METALANG99_PRIV_REC_NEXT(1846, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1846(choice, ...)  METALANG99_PRIV_REC_NEXT(1847, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1847(choice, ...)  METALANG99_PRIV_REC_NEXT(1848, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1848(choice, ...)  METALANG99_PRIV_REC_NEXT(1849, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1849(choice, ...)  METALANG99_PRIV_REC_NEXT(1850, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1850(choice, ...)  METALANG99_PRIV_REC_NEXT(1851, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1851(choice, ...)  METALANG99_PRIV_REC_NEXT(1852, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1852(choice, ...)  METALANG99_PRIV_REC_NEXT(1853, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1853(choice, ...)  METALANG99_PRIV_REC_NEXT(1854, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1854(choice, ...)  METALANG99_PRIV_REC_NEXT(1855, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1855(choice, ...)  METALANG99_PRIV_REC_NEXT(1856, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1856(choice, ...)  METALANG99_PRIV_REC_NEXT(1857, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1857(choice, ...)  METALANG99_PRIV_REC_NEXT(1858, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1858(choice, ...)  METALANG99_PRIV_REC_NEXT(1859, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1859(choice, ...)  METALANG99_PRIV_REC_NEXT(1860, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1860(choice, ...)  METALANG99_PRIV_REC_NEXT(1861, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1861(choice, ...)  METALANG99_PRIV_REC_NEXT(1862, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1862(choice, ...)  METALANG99_PRIV_REC_NEXT(1863, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1863(choice, ...)  METALANG99_PRIV_REC_NEXT(1864, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1864(choice, ...)  METALANG99_PRIV_REC_NEXT(1865, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1865(choice, ...)  METALANG99_PRIV_REC_NEXT(1866, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1866(choice, ...)  METALANG99_PRIV_REC_NEXT(1867, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1867(choice, ...)  METALANG99_PRIV_REC_NEXT(1868, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1868(choice, ...)  METALANG99_PRIV_REC_NEXT(1869, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1869(choice, ...)  METALANG99_PRIV_REC_NEXT(1870, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1870(choice, ...)  METALANG99_PRIV_REC_NEXT(1871, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1871(choice, ...)  METALANG99_PRIV_REC_NEXT(1872, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1872(choice, ...)  METALANG99_PRIV_REC_NEXT(1873, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1873(choice, ...)  METALANG99_PRIV_REC_NEXT(1874, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1874(choice, ...)  METALANG99_PRIV_REC_NEXT(1875, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1875(choice, ...)  METALANG99_PRIV_REC_NEXT(1876, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1876(choice, ...)  METALANG99_PRIV_REC_NEXT(1877, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1877(choice, ...)  METALANG99_PRIV_REC_NEXT(1878, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1878(choice, ...)  METALANG99_PRIV_REC_NEXT(1879, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1879(choice, ...)  METALANG99_PRIV_REC_NEXT(1880, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1880(choice, ...)  METALANG99_PRIV_REC_NEXT(1881, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1881(choice, ...)  METALANG99_PRIV_REC_NEXT(1882, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1882(choice, ...)  METALANG99_PRIV_REC_NEXT(1883, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1883(choice, ...)  METALANG99_PRIV_REC_NEXT(1884, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1884(choice, ...)  METALANG99_PRIV_REC_NEXT(1885, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1885(choice, ...)  METALANG99_PRIV_REC_NEXT(1886, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1886(choice, ...)  METALANG99_PRIV_REC_NEXT(1887, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1887(choice, ...)  METALANG99_PRIV_REC_NEXT(1888, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1888(choice, ...)  METALANG99_PRIV_REC_NEXT(1889, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1889(choice, ...)  METALANG99_PRIV_REC_NEXT(1890, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1890(choice, ...)  METALANG99_PRIV_REC_NEXT(1891, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1891(choice, ...)  METALANG99_PRIV_REC_NEXT(1892, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1892(choice, ...)  METALANG99_PRIV_REC_NEXT(1893, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1893(choice, ...)  METALANG99_PRIV_REC_NEXT(1894, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1894(choice, ...)  METALANG99_PRIV_REC_NEXT(1895, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1895(choice, ...)  METALANG99_PRIV_REC_NEXT(1896, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1896(choice, ...)  METALANG99_PRIV_REC_NEXT(1897, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1897(choice, ...)  METALANG99_PRIV_REC_NEXT(1898, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1898(choice, ...)  METALANG99_PRIV_REC_NEXT(1899, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1899(choice, ...)  METALANG99_PRIV_REC_NEXT(1900, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1900(choice, ...)  METALANG99_PRIV_REC_NEXT(1901, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1901(choice, ...)  METALANG99_PRIV_REC_NEXT(1902, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1902(choice, ...)  METALANG99_PRIV_REC_NEXT(1903, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1903(choice, ...)  METALANG99_PRIV_REC_NEXT(1904, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1904(choice, ...)  METALANG99_PRIV_REC_NEXT(1905, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1905(choice, ...)  METALANG99_PRIV_REC_NEXT(1906, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1906(choice, ...)  METALANG99_PRIV_REC_NEXT(1907, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1907(choice, ...)  METALANG99_PRIV_REC_NEXT(1908, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1908(choice, ...)  METALANG99_PRIV_REC_NEXT(1909, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1909(choice, ...)  METALANG99_PRIV_REC_NEXT(1910, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1910(choice, ...)  METALANG99_PRIV_REC_NEXT(1911, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1911(choice, ...)  METALANG99_PRIV_REC_NEXT(1912, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1912(choice, ...)  METALANG99_PRIV_REC_NEXT(1913, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1913(choice, ...)  METALANG99_PRIV_REC_NEXT(1914, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1914(choice, ...)  METALANG99_PRIV_REC_NEXT(1915, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1915(choice, ...)  METALANG99_PRIV_REC_NEXT(1916, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1916(choice, ...)  METALANG99_PRIV_REC_NEXT(1917, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1917(choice, ...)  METALANG99_PRIV_REC_NEXT(1918, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1918(choice, ...)  METALANG99_PRIV_REC_NEXT(1919, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1919(choice, ...)  METALANG99_PRIV_REC_NEXT(1920, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1920(choice, ...)  METALANG99_PRIV_REC_NEXT(1921, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1921(choice, ...)  METALANG99_PRIV_REC_NEXT(1922, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1922(choice, ...)  METALANG99_PRIV_REC_NEXT(1923, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1923(choice, ...)  METALANG99_PRIV_REC_NEXT(1924, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1924(choice, ...)  METALANG99_PRIV_REC_NEXT(1925, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1925(choice, ...)  METALANG99_PRIV_REC_NEXT(1926, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1926(choice, ...)  METALANG99_PRIV_REC_NEXT(1927, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1927(choice, ...)  METALANG99_PRIV_REC_NEXT(1928, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1928(choice, ...)  METALANG99_PRIV_REC_NEXT(1929, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1929(choice, ...)  METALANG99_PRIV_REC_NEXT(1930, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1930(choice, ...)  METALANG99_PRIV_REC_NEXT(1931, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1931(choice, ...)  METALANG99_PRIV_REC_NEXT(1932, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1932(choice, ...)  METALANG99_PRIV_REC_NEXT(1933, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1933(choice, ...)  METALANG99_PRIV_REC_NEXT(1934, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1934(choice, ...)  METALANG99_PRIV_REC_NEXT(1935, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1935(choice, ...)  METALANG99_PRIV_REC_NEXT(1936, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1936(choice, ...)  METALANG99_PRIV_REC_NEXT(1937, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1937(choice, ...)  METALANG99_PRIV_REC_NEXT(1938, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1938(choice, ...)  METALANG99_PRIV_REC_NEXT(1939, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1939(choice, ...)  METALANG99_PRIV_REC_NEXT(1940, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1940(choice, ...)  METALANG99_PRIV_REC_NEXT(1941, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1941(choice, ...)  METALANG99_PRIV_REC_NEXT(1942, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1942(choice, ...)  METALANG99_PRIV_REC_NEXT(1943, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1943(choice, ...)  METALANG99_PRIV_REC_NEXT(1944, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1944(choice, ...)  METALANG99_PRIV_REC_NEXT(1945, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1945(choice, ...)  METALANG99_PRIV_REC_NEXT(1946, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1946(choice, ...)  METALANG99_PRIV_REC_NEXT(1947, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1947(choice, ...)  METALANG99_PRIV_REC_NEXT(1948, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1948(choice, ...)  METALANG99_PRIV_REC_NEXT(1949, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1949(choice, ...)  METALANG99_PRIV_REC_NEXT(1950, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1950(choice, ...)  METALANG99_PRIV_REC_NEXT(1951, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1951(choice, ...)  METALANG99_PRIV_REC_NEXT(1952, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1952(choice, ...)  METALANG99_PRIV_REC_NEXT(1953, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1953(choice, ...)  METALANG99_PRIV_REC_NEXT(1954, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1954(choice, ...)  METALANG99_PRIV_REC_NEXT(1955, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1955(choice, ...)  METALANG99_PRIV_REC_NEXT(1956, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1956(choice, ...)  METALANG99_PRIV_REC_NEXT(1957, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1957(choice, ...)  METALANG99_PRIV_REC_NEXT(1958, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1958(choice, ...)  METALANG99_PRIV_REC_NEXT(1959, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1959(choice, ...)  METALANG99_PRIV_REC_NEXT(1960, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1960(choice, ...)  METALANG99_PRIV_REC_NEXT(1961, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1961(choice, ...)  METALANG99_PRIV_REC_NEXT(1962, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1962(choice, ...)  METALANG99_PRIV_REC_NEXT(1963, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1963(choice, ...)  METALANG99_PRIV_REC_NEXT(1964, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1964(choice, ...)  METALANG99_PRIV_REC_NEXT(1965, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1965(choice, ...)  METALANG99_PRIV_REC_NEXT(1966, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1966(choice, ...)  METALANG99_PRIV_REC_NEXT(1967, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1967(choice, ...)  METALANG99_PRIV_REC_NEXT(1968, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1968(choice, ...)  METALANG99_PRIV_REC_NEXT(1969, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1969(choice, ...)  METALANG99_PRIV_REC_NEXT(1970, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1970(choice, ...)  METALANG99_PRIV_REC_NEXT(1971, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1971(choice, ...)  METALANG99_PRIV_REC_NEXT(1972, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1972(choice, ...)  METALANG99_PRIV_REC_NEXT(1973, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1973(choice, ...)  METALANG99_PRIV_REC_NEXT(1974, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1974(choice, ...)  METALANG99_PRIV_REC_NEXT(1975, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1975(choice, ...)  METALANG99_PRIV_REC_NEXT(1976, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1976(choice, ...)  METALANG99_PRIV_REC_NEXT(1977, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1977(choice, ...)  METALANG99_PRIV_REC_NEXT(1978, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1978(choice, ...)  METALANG99_PRIV_REC_NEXT(1979, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1979(choice, ...)  METALANG99_PRIV_REC_NEXT(1980, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1980(choice, ...)  METALANG99_PRIV_REC_NEXT(1981, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1981(choice, ...)  METALANG99_PRIV_REC_NEXT(1982, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1982(choice, ...)  METALANG99_PRIV_REC_NEXT(1983, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1983(choice, ...)  METALANG99_PRIV_REC_NEXT(1984, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1984(choice, ...)  METALANG99_PRIV_REC_NEXT(1985, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1985(choice, ...)  METALANG99_PRIV_REC_NEXT(1986, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1986(choice, ...)  METALANG99_PRIV_REC_NEXT(1987, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1987(choice, ...)  METALANG99_PRIV_REC_NEXT(1988, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1988(choice, ...)  METALANG99_PRIV_REC_NEXT(1989, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1989(choice, ...)  METALANG99_PRIV_REC_NEXT(1990, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1990(choice, ...)  METALANG99_PRIV_REC_NEXT(1991, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1991(choice, ...)  METALANG99_PRIV_REC_NEXT(1992, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1992(choice, ...)  METALANG99_PRIV_REC_NEXT(1993, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1993(choice, ...)  METALANG99_PRIV_REC_NEXT(1994, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1994(choice, ...)  METALANG99_PRIV_REC_NEXT(1995, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1995(choice, ...)  METALANG99_PRIV_REC_NEXT(1996, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1996(choice, ...)  METALANG99_PRIV_REC_NEXT(1997, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1997(choice, ...)  METALANG99_PRIV_REC_NEXT(1998, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1998(choice, ...)  METALANG99_PRIV_REC_NEXT(1999, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_1999(choice, ...)  METALANG99_PRIV_REC_NEXT(2000, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2000(choice, ...)  METALANG99_PRIV_REC_NEXT(2001, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2001(choice, ...)  METALANG99_PRIV_REC_NEXT(2002, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2002(choice, ...)  METALANG99_PRIV_REC_NEXT(2003, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2003(choice, ...)  METALANG99_PRIV_REC_NEXT(2004, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2004(choice, ...)  METALANG99_PRIV_REC_NEXT(2005, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2005(choice, ...)  METALANG99_PRIV_REC_NEXT(2006, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2006(choice, ...)  METALANG99_PRIV_REC_NEXT(2007, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2007(choice, ...)  METALANG99_PRIV_REC_NEXT(2008, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2008(choice, ...)  METALANG99_PRIV_REC_NEXT(2009, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2009(choice, ...)  METALANG99_PRIV_REC_NEXT(2010, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2010(choice, ...)  METALANG99_PRIV_REC_NEXT(2011, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2011(choice, ...)  METALANG99_PRIV_REC_NEXT(2012, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2012(choice, ...)  METALANG99_PRIV_REC_NEXT(2013, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2013(choice, ...)  METALANG99_PRIV_REC_NEXT(2014, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2014(choice, ...)  METALANG99_PRIV_REC_NEXT(2015, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2015(choice, ...)  METALANG99_PRIV_REC_NEXT(2016, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2016(choice, ...)  METALANG99_PRIV_REC_NEXT(2017, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2017(choice, ...)  METALANG99_PRIV_REC_NEXT(2018, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2018(choice, ...)  METALANG99_PRIV_REC_NEXT(2019, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2019(choice, ...)  METALANG99_PRIV_REC_NEXT(2020, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2020(choice, ...)  METALANG99_PRIV_REC_NEXT(2021, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2021(choice, ...)  METALANG99_PRIV_REC_NEXT(2022, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2022(choice, ...)  METALANG99_PRIV_REC_NEXT(2023, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2023(choice, ...)  METALANG99_PRIV_REC_NEXT(2024, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2024(choice, ...)  METALANG99_PRIV_REC_NEXT(2025, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2025(choice, ...)  METALANG99_PRIV_REC_NEXT(2026, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2026(choice, ...)  METALANG99_PRIV_REC_NEXT(2027, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2027(choice, ...)  METALANG99_PRIV_REC_NEXT(2028, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2028(choice, ...)  METALANG99_PRIV_REC_NEXT(2029, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2029(choice, ...)  METALANG99_PRIV_REC_NEXT(2030, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2030(choice, ...)  METALANG99_PRIV_REC_NEXT(2031, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2031(choice, ...)  METALANG99_PRIV_REC_NEXT(2032, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2032(choice, ...)  METALANG99_PRIV_REC_NEXT(2033, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2033(choice, ...)  METALANG99_PRIV_REC_NEXT(2034, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2034(choice, ...)  METALANG99_PRIV_REC_NEXT(2035, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2035(choice, ...)  METALANG99_PRIV_REC_NEXT(2036, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2036(choice, ...)  METALANG99_PRIV_REC_NEXT(2037, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2037(choice, ...)  METALANG99_PRIV_REC_NEXT(2038, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2038(choice, ...)  METALANG99_PRIV_REC_NEXT(2039, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2039(choice, ...)  METALANG99_PRIV_REC_NEXT(2040, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2040(choice, ...)  METALANG99_PRIV_REC_NEXT(2041, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2041(choice, ...)  METALANG99_PRIV_REC_NEXT(2042, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2042(choice, ...)  METALANG99_PRIV_REC_NEXT(2043, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2043(choice, ...)  METALANG99_PRIV_REC_NEXT(2044, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2044(choice, ...)  METALANG99_PRIV_REC_NEXT(2045, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2045(choice, ...)  METALANG99_PRIV_REC_NEXT(2046, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2046(choice, ...)  METALANG99_PRIV_REC_NEXT(2047, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2047(choice, ...)  METALANG99_PRIV_REC_NEXT(2048, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2048(choice, ...)  METALANG99_PRIV_REC_NEXT(2049, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2049(choice, ...)  METALANG99_PRIV_REC_NEXT(2050, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2050(choice, ...)  METALANG99_PRIV_REC_NEXT(2051, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2051(choice, ...)  METALANG99_PRIV_REC_NEXT(2052, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2052(choice, ...)  METALANG99_PRIV_REC_NEXT(2053, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2053(choice, ...)  METALANG99_PRIV_REC_NEXT(2054, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2054(choice, ...)  METALANG99_PRIV_REC_NEXT(2055, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2055(choice, ...)  METALANG99_PRIV_REC_NEXT(2056, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2056(choice, ...)  METALANG99_PRIV_REC_NEXT(2057, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2057(choice, ...)  METALANG99_PRIV_REC_NEXT(2058, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2058(choice, ...)  METALANG99_PRIV_REC_NEXT(2059, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2059(choice, ...)  METALANG99_PRIV_REC_NEXT(2060, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2060(choice, ...)  METALANG99_PRIV_REC_NEXT(2061, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2061(choice, ...)  METALANG99_PRIV_REC_NEXT(2062, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2062(choice, ...)  METALANG99_PRIV_REC_NEXT(2063, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2063(choice, ...)  METALANG99_PRIV_REC_NEXT(2064, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2064(choice, ...)  METALANG99_PRIV_REC_NEXT(2065, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2065(choice, ...)  METALANG99_PRIV_REC_NEXT(2066, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2066(choice, ...)  METALANG99_PRIV_REC_NEXT(2067, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2067(choice, ...)  METALANG99_PRIV_REC_NEXT(2068, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2068(choice, ...)  METALANG99_PRIV_REC_NEXT(2069, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2069(choice, ...)  METALANG99_PRIV_REC_NEXT(2070, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2070(choice, ...)  METALANG99_PRIV_REC_NEXT(2071, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2071(choice, ...)  METALANG99_PRIV_REC_NEXT(2072, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2072(choice, ...)  METALANG99_PRIV_REC_NEXT(2073, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2073(choice, ...)  METALANG99_PRIV_REC_NEXT(2074, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2074(choice, ...)  METALANG99_PRIV_REC_NEXT(2075, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2075(choice, ...)  METALANG99_PRIV_REC_NEXT(2076, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2076(choice, ...)  METALANG99_PRIV_REC_NEXT(2077, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2077(choice, ...)  METALANG99_PRIV_REC_NEXT(2078, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2078(choice, ...)  METALANG99_PRIV_REC_NEXT(2079, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2079(choice, ...)  METALANG99_PRIV_REC_NEXT(2080, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2080(choice, ...)  METALANG99_PRIV_REC_NEXT(2081, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2081(choice, ...)  METALANG99_PRIV_REC_NEXT(2082, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2082(choice, ...)  METALANG99_PRIV_REC_NEXT(2083, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2083(choice, ...)  METALANG99_PRIV_REC_NEXT(2084, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2084(choice, ...)  METALANG99_PRIV_REC_NEXT(2085, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2085(choice, ...)  METALANG99_PRIV_REC_NEXT(2086, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2086(choice, ...)  METALANG99_PRIV_REC_NEXT(2087, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2087(choice, ...)  METALANG99_PRIV_REC_NEXT(2088, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2088(choice, ...)  METALANG99_PRIV_REC_NEXT(2089, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2089(choice, ...)  METALANG99_PRIV_REC_NEXT(2090, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2090(choice, ...)  METALANG99_PRIV_REC_NEXT(2091, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2091(choice, ...)  METALANG99_PRIV_REC_NEXT(2092, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2092(choice, ...)  METALANG99_PRIV_REC_NEXT(2093, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2093(choice, ...)  METALANG99_PRIV_REC_NEXT(2094, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2094(choice, ...)  METALANG99_PRIV_REC_NEXT(2095, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2095(choice, ...)  METALANG99_PRIV_REC_NEXT(2096, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2096(choice, ...)  METALANG99_PRIV_REC_NEXT(2097, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2097(choice, ...)  METALANG99_PRIV_REC_NEXT(2098, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2098(choice, ...)  METALANG99_PRIV_REC_NEXT(2099, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2099(choice, ...)  METALANG99_PRIV_REC_NEXT(2100, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2100(choice, ...)  METALANG99_PRIV_REC_NEXT(2101, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2101(choice, ...)  METALANG99_PRIV_REC_NEXT(2102, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2102(choice, ...)  METALANG99_PRIV_REC_NEXT(2103, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2103(choice, ...)  METALANG99_PRIV_REC_NEXT(2104, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2104(choice, ...)  METALANG99_PRIV_REC_NEXT(2105, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2105(choice, ...)  METALANG99_PRIV_REC_NEXT(2106, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2106(choice, ...)  METALANG99_PRIV_REC_NEXT(2107, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2107(choice, ...)  METALANG99_PRIV_REC_NEXT(2108, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2108(choice, ...)  METALANG99_PRIV_REC_NEXT(2109, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2109(choice, ...)  METALANG99_PRIV_REC_NEXT(2110, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2110(choice, ...)  METALANG99_PRIV_REC_NEXT(2111, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2111(choice, ...)  METALANG99_PRIV_REC_NEXT(2112, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2112(choice, ...)  METALANG99_PRIV_REC_NEXT(2113, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2113(choice, ...)  METALANG99_PRIV_REC_NEXT(2114, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2114(choice, ...)  METALANG99_PRIV_REC_NEXT(2115, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2115(choice, ...)  METALANG99_PRIV_REC_NEXT(2116, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2116(choice, ...)  METALANG99_PRIV_REC_NEXT(2117, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2117(choice, ...)  METALANG99_PRIV_REC_NEXT(2118, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2118(choice, ...)  METALANG99_PRIV_REC_NEXT(2119, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2119(choice, ...)  METALANG99_PRIV_REC_NEXT(2120, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2120(choice, ...)  METALANG99_PRIV_REC_NEXT(2121, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2121(choice, ...)  METALANG99_PRIV_REC_NEXT(2122, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2122(choice, ...)  METALANG99_PRIV_REC_NEXT(2123, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2123(choice, ...)  METALANG99_PRIV_REC_NEXT(2124, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2124(choice, ...)  METALANG99_PRIV_REC_NEXT(2125, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2125(choice, ...)  METALANG99_PRIV_REC_NEXT(2126, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2126(choice, ...)  METALANG99_PRIV_REC_NEXT(2127, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2127(choice, ...)  METALANG99_PRIV_REC_NEXT(2128, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2128(choice, ...)  METALANG99_PRIV_REC_NEXT(2129, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2129(choice, ...)  METALANG99_PRIV_REC_NEXT(2130, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2130(choice, ...)  METALANG99_PRIV_REC_NEXT(2131, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2131(choice, ...)  METALANG99_PRIV_REC_NEXT(2132, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2132(choice, ...)  METALANG99_PRIV_REC_NEXT(2133, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2133(choice, ...)  METALANG99_PRIV_REC_NEXT(2134, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2134(choice, ...)  METALANG99_PRIV_REC_NEXT(2135, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2135(choice, ...)  METALANG99_PRIV_REC_NEXT(2136, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2136(choice, ...)  METALANG99_PRIV_REC_NEXT(2137, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2137(choice, ...)  METALANG99_PRIV_REC_NEXT(2138, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2138(choice, ...)  METALANG99_PRIV_REC_NEXT(2139, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2139(choice, ...)  METALANG99_PRIV_REC_NEXT(2140, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2140(choice, ...)  METALANG99_PRIV_REC_NEXT(2141, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2141(choice, ...)  METALANG99_PRIV_REC_NEXT(2142, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2142(choice, ...)  METALANG99_PRIV_REC_NEXT(2143, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2143(choice, ...)  METALANG99_PRIV_REC_NEXT(2144, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2144(choice, ...)  METALANG99_PRIV_REC_NEXT(2145, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2145(choice, ...)  METALANG99_PRIV_REC_NEXT(2146, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2146(choice, ...)  METALANG99_PRIV_REC_NEXT(2147, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2147(choice, ...)  METALANG99_PRIV_REC_NEXT(2148, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2148(choice, ...)  METALANG99_PRIV_REC_NEXT(2149, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2149(choice, ...)  METALANG99_PRIV_REC_NEXT(2150, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2150(choice, ...)  METALANG99_PRIV_REC_NEXT(2151, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2151(choice, ...)  METALANG99_PRIV_REC_NEXT(2152, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2152(choice, ...)  METALANG99_PRIV_REC_NEXT(2153, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2153(choice, ...)  METALANG99_PRIV_REC_NEXT(2154, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2154(choice, ...)  METALANG99_PRIV_REC_NEXT(2155, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2155(choice, ...)  METALANG99_PRIV_REC_NEXT(2156, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2156(choice, ...)  METALANG99_PRIV_REC_NEXT(2157, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2157(choice, ...)  METALANG99_PRIV_REC_NEXT(2158, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2158(choice, ...)  METALANG99_PRIV_REC_NEXT(2159, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2159(choice, ...)  METALANG99_PRIV_REC_NEXT(2160, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2160(choice, ...)  METALANG99_PRIV_REC_NEXT(2161, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2161(choice, ...)  METALANG99_PRIV_REC_NEXT(2162, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2162(choice, ...)  METALANG99_PRIV_REC_NEXT(2163, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2163(choice, ...)  METALANG99_PRIV_REC_NEXT(2164, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2164(choice, ...)  METALANG99_PRIV_REC_NEXT(2165, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2165(choice, ...)  METALANG99_PRIV_REC_NEXT(2166, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2166(choice, ...)  METALANG99_PRIV_REC_NEXT(2167, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2167(choice, ...)  METALANG99_PRIV_REC_NEXT(2168, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2168(choice, ...)  METALANG99_PRIV_REC_NEXT(2169, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2169(choice, ...)  METALANG99_PRIV_REC_NEXT(2170, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2170(choice, ...)  METALANG99_PRIV_REC_NEXT(2171, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2171(choice, ...)  METALANG99_PRIV_REC_NEXT(2172, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2172(choice, ...)  METALANG99_PRIV_REC_NEXT(2173, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2173(choice, ...)  METALANG99_PRIV_REC_NEXT(2174, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2174(choice, ...)  METALANG99_PRIV_REC_NEXT(2175, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2175(choice, ...)  METALANG99_PRIV_REC_NEXT(2176, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2176(choice, ...)  METALANG99_PRIV_REC_NEXT(2177, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2177(choice, ...)  METALANG99_PRIV_REC_NEXT(2178, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2178(choice, ...)  METALANG99_PRIV_REC_NEXT(2179, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2179(choice, ...)  METALANG99_PRIV_REC_NEXT(2180, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2180(choice, ...)  METALANG99_PRIV_REC_NEXT(2181, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2181(choice, ...)  METALANG99_PRIV_REC_NEXT(2182, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2182(choice, ...)  METALANG99_PRIV_REC_NEXT(2183, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2183(choice, ...)  METALANG99_PRIV_REC_NEXT(2184, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2184(choice, ...)  METALANG99_PRIV_REC_NEXT(2185, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2185(choice, ...)  METALANG99_PRIV_REC_NEXT(2186, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2186(choice, ...)  METALANG99_PRIV_REC_NEXT(2187, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2187(choice, ...)  METALANG99_PRIV_REC_NEXT(2188, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2188(choice, ...)  METALANG99_PRIV_REC_NEXT(2189, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2189(choice, ...)  METALANG99_PRIV_REC_NEXT(2190, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2190(choice, ...)  METALANG99_PRIV_REC_NEXT(2191, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2191(choice, ...)  METALANG99_PRIV_REC_NEXT(2192, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2192(choice, ...)  METALANG99_PRIV_REC_NEXT(2193, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2193(choice, ...)  METALANG99_PRIV_REC_NEXT(2194, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2194(choice, ...)  METALANG99_PRIV_REC_NEXT(2195, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2195(choice, ...)  METALANG99_PRIV_REC_NEXT(2196, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2196(choice, ...)  METALANG99_PRIV_REC_NEXT(2197, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2197(choice, ...)  METALANG99_PRIV_REC_NEXT(2198, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2198(choice, ...)  METALANG99_PRIV_REC_NEXT(2199, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2199(choice, ...)  METALANG99_PRIV_REC_NEXT(2200, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2200(choice, ...)  METALANG99_PRIV_REC_NEXT(2201, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2201(choice, ...)  METALANG99_PRIV_REC_NEXT(2202, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2202(choice, ...)  METALANG99_PRIV_REC_NEXT(2203, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2203(choice, ...)  METALANG99_PRIV_REC_NEXT(2204, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2204(choice, ...)  METALANG99_PRIV_REC_NEXT(2205, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2205(choice, ...)  METALANG99_PRIV_REC_NEXT(2206, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2206(choice, ...)  METALANG99_PRIV_REC_NEXT(2207, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2207(choice, ...)  METALANG99_PRIV_REC_NEXT(2208, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2208(choice, ...)  METALANG99_PRIV_REC_NEXT(2209, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2209(choice, ...)  METALANG99_PRIV_REC_NEXT(2210, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2210(choice, ...)  METALANG99_PRIV_REC_NEXT(2211, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2211(choice, ...)  METALANG99_PRIV_REC_NEXT(2212, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2212(choice, ...)  METALANG99_PRIV_REC_NEXT(2213, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2213(choice, ...)  METALANG99_PRIV_REC_NEXT(2214, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2214(choice, ...)  METALANG99_PRIV_REC_NEXT(2215, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2215(choice, ...)  METALANG99_PRIV_REC_NEXT(2216, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2216(choice, ...)  METALANG99_PRIV_REC_NEXT(2217, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2217(choice, ...)  METALANG99_PRIV_REC_NEXT(2218, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2218(choice, ...)  METALANG99_PRIV_REC_NEXT(2219, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2219(choice, ...)  METALANG99_PRIV_REC_NEXT(2220, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2220(choice, ...)  METALANG99_PRIV_REC_NEXT(2221, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2221(choice, ...)  METALANG99_PRIV_REC_NEXT(2222, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2222(choice, ...)  METALANG99_PRIV_REC_NEXT(2223, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2223(choice, ...)  METALANG99_PRIV_REC_NEXT(2224, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2224(choice, ...)  METALANG99_PRIV_REC_NEXT(2225, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2225(choice, ...)  METALANG99_PRIV_REC_NEXT(2226, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2226(choice, ...)  METALANG99_PRIV_REC_NEXT(2227, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2227(choice, ...)  METALANG99_PRIV_REC_NEXT(2228, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2228(choice, ...)  METALANG99_PRIV_REC_NEXT(2229, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2229(choice, ...)  METALANG99_PRIV_REC_NEXT(2230, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2230(choice, ...)  METALANG99_PRIV_REC_NEXT(2231, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2231(choice, ...)  METALANG99_PRIV_REC_NEXT(2232, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2232(choice, ...)  METALANG99_PRIV_REC_NEXT(2233, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2233(choice, ...)  METALANG99_PRIV_REC_NEXT(2234, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2234(choice, ...)  METALANG99_PRIV_REC_NEXT(2235, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2235(choice, ...)  METALANG99_PRIV_REC_NEXT(2236, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2236(choice, ...)  METALANG99_PRIV_REC_NEXT(2237, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2237(choice, ...)  METALANG99_PRIV_REC_NEXT(2238, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2238(choice, ...)  METALANG99_PRIV_REC_NEXT(2239, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2239(choice, ...)  METALANG99_PRIV_REC_NEXT(2240, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2240(choice, ...)  METALANG99_PRIV_REC_NEXT(2241, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2241(choice, ...)  METALANG99_PRIV_REC_NEXT(2242, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2242(choice, ...)  METALANG99_PRIV_REC_NEXT(2243, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2243(choice, ...)  METALANG99_PRIV_REC_NEXT(2244, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2244(choice, ...)  METALANG99_PRIV_REC_NEXT(2245, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2245(choice, ...)  METALANG99_PRIV_REC_NEXT(2246, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2246(choice, ...)  METALANG99_PRIV_REC_NEXT(2247, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2247(choice, ...)  METALANG99_PRIV_REC_NEXT(2248, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2248(choice, ...)  METALANG99_PRIV_REC_NEXT(2249, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2249(choice, ...)  METALANG99_PRIV_REC_NEXT(2250, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2250(choice, ...)  METALANG99_PRIV_REC_NEXT(2251, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2251(choice, ...)  METALANG99_PRIV_REC_NEXT(2252, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2252(choice, ...)  METALANG99_PRIV_REC_NEXT(2253, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2253(choice, ...)  METALANG99_PRIV_REC_NEXT(2254, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2254(choice, ...)  METALANG99_PRIV_REC_NEXT(2255, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2255(choice, ...)  METALANG99_PRIV_REC_NEXT(2256, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2256(choice, ...)  METALANG99_PRIV_REC_NEXT(2257, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2257(choice, ...)  METALANG99_PRIV_REC_NEXT(2258, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2258(choice, ...)  METALANG99_PRIV_REC_NEXT(2259, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2259(choice, ...)  METALANG99_PRIV_REC_NEXT(2260, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2260(choice, ...)  METALANG99_PRIV_REC_NEXT(2261, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2261(choice, ...)  METALANG99_PRIV_REC_NEXT(2262, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2262(choice, ...)  METALANG99_PRIV_REC_NEXT(2263, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2263(choice, ...)  METALANG99_PRIV_REC_NEXT(2264, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2264(choice, ...)  METALANG99_PRIV_REC_NEXT(2265, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2265(choice, ...)  METALANG99_PRIV_REC_NEXT(2266, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2266(choice, ...)  METALANG99_PRIV_REC_NEXT(2267, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2267(choice, ...)  METALANG99_PRIV_REC_NEXT(2268, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2268(choice, ...)  METALANG99_PRIV_REC_NEXT(2269, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2269(choice, ...)  METALANG99_PRIV_REC_NEXT(2270, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2270(choice, ...)  METALANG99_PRIV_REC_NEXT(2271, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2271(choice, ...)  METALANG99_PRIV_REC_NEXT(2272, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2272(choice, ...)  METALANG99_PRIV_REC_NEXT(2273, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2273(choice, ...)  METALANG99_PRIV_REC_NEXT(2274, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2274(choice, ...)  METALANG99_PRIV_REC_NEXT(2275, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2275(choice, ...)  METALANG99_PRIV_REC_NEXT(2276, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2276(choice, ...)  METALANG99_PRIV_REC_NEXT(2277, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2277(choice, ...)  METALANG99_PRIV_REC_NEXT(2278, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2278(choice, ...)  METALANG99_PRIV_REC_NEXT(2279, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2279(choice, ...)  METALANG99_PRIV_REC_NEXT(2280, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2280(choice, ...)  METALANG99_PRIV_REC_NEXT(2281, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2281(choice, ...)  METALANG99_PRIV_REC_NEXT(2282, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2282(choice, ...)  METALANG99_PRIV_REC_NEXT(2283, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2283(choice, ...)  METALANG99_PRIV_REC_NEXT(2284, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2284(choice, ...)  METALANG99_PRIV_REC_NEXT(2285, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2285(choice, ...)  METALANG99_PRIV_REC_NEXT(2286, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2286(choice, ...)  METALANG99_PRIV_REC_NEXT(2287, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2287(choice, ...)  METALANG99_PRIV_REC_NEXT(2288, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2288(choice, ...)  METALANG99_PRIV_REC_NEXT(2289, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2289(choice, ...)  METALANG99_PRIV_REC_NEXT(2290, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2290(choice, ...)  METALANG99_PRIV_REC_NEXT(2291, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2291(choice, ...)  METALANG99_PRIV_REC_NEXT(2292, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2292(choice, ...)  METALANG99_PRIV_REC_NEXT(2293, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2293(choice, ...)  METALANG99_PRIV_REC_NEXT(2294, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2294(choice, ...)  METALANG99_PRIV_REC_NEXT(2295, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2295(choice, ...)  METALANG99_PRIV_REC_NEXT(2296, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2296(choice, ...)  METALANG99_PRIV_REC_NEXT(2297, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2297(choice, ...)  METALANG99_PRIV_REC_NEXT(2298, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2298(choice, ...)  METALANG99_PRIV_REC_NEXT(2299, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2299(choice, ...)  METALANG99_PRIV_REC_NEXT(2300, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2300(choice, ...)  METALANG99_PRIV_REC_NEXT(2301, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2301(choice, ...)  METALANG99_PRIV_REC_NEXT(2302, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2302(choice, ...)  METALANG99_PRIV_REC_NEXT(2303, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2303(choice, ...)  METALANG99_PRIV_REC_NEXT(2304, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2304(choice, ...)  METALANG99_PRIV_REC_NEXT(2305, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2305(choice, ...)  METALANG99_PRIV_REC_NEXT(2306, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2306(choice, ...)  METALANG99_PRIV_REC_NEXT(2307, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2307(choice, ...)  METALANG99_PRIV_REC_NEXT(2308, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2308(choice, ...)  METALANG99_PRIV_REC_NEXT(2309, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2309(choice, ...)  METALANG99_PRIV_REC_NEXT(2310, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2310(choice, ...)  METALANG99_PRIV_REC_NEXT(2311, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2311(choice, ...)  METALANG99_PRIV_REC_NEXT(2312, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2312(choice, ...)  METALANG99_PRIV_REC_NEXT(2313, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2313(choice, ...)  METALANG99_PRIV_REC_NEXT(2314, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2314(choice, ...)  METALANG99_PRIV_REC_NEXT(2315, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2315(choice, ...)  METALANG99_PRIV_REC_NEXT(2316, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2316(choice, ...)  METALANG99_PRIV_REC_NEXT(2317, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2317(choice, ...)  METALANG99_PRIV_REC_NEXT(2318, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2318(choice, ...)  METALANG99_PRIV_REC_NEXT(2319, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2319(choice, ...)  METALANG99_PRIV_REC_NEXT(2320, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2320(choice, ...)  METALANG99_PRIV_REC_NEXT(2321, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2321(choice, ...)  METALANG99_PRIV_REC_NEXT(2322, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2322(choice, ...)  METALANG99_PRIV_REC_NEXT(2323, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2323(choice, ...)  METALANG99_PRIV_REC_NEXT(2324, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2324(choice, ...)  METALANG99_PRIV_REC_NEXT(2325, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2325(choice, ...)  METALANG99_PRIV_REC_NEXT(2326, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2326(choice, ...)  METALANG99_PRIV_REC_NEXT(2327, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2327(choice, ...)  METALANG99_PRIV_REC_NEXT(2328, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2328(choice, ...)  METALANG99_PRIV_REC_NEXT(2329, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2329(choice, ...)  METALANG99_PRIV_REC_NEXT(2330, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2330(choice, ...)  METALANG99_PRIV_REC_NEXT(2331, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2331(choice, ...)  METALANG99_PRIV_REC_NEXT(2332, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2332(choice, ...)  METALANG99_PRIV_REC_NEXT(2333, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2333(choice, ...)  METALANG99_PRIV_REC_NEXT(2334, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2334(choice, ...)  METALANG99_PRIV_REC_NEXT(2335, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2335(choice, ...)  METALANG99_PRIV_REC_NEXT(2336, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2336(choice, ...)  METALANG99_PRIV_REC_NEXT(2337, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2337(choice, ...)  METALANG99_PRIV_REC_NEXT(2338, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2338(choice, ...)  METALANG99_PRIV_REC_NEXT(2339, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2339(choice, ...)  METALANG99_PRIV_REC_NEXT(2340, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2340(choice, ...)  METALANG99_PRIV_REC_NEXT(2341, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2341(choice, ...)  METALANG99_PRIV_REC_NEXT(2342, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2342(choice, ...)  METALANG99_PRIV_REC_NEXT(2343, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2343(choice, ...)  METALANG99_PRIV_REC_NEXT(2344, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2344(choice, ...)  METALANG99_PRIV_REC_NEXT(2345, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2345(choice, ...)  METALANG99_PRIV_REC_NEXT(2346, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2346(choice, ...)  METALANG99_PRIV_REC_NEXT(2347, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2347(choice, ...)  METALANG99_PRIV_REC_NEXT(2348, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2348(choice, ...)  METALANG99_PRIV_REC_NEXT(2349, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2349(choice, ...)  METALANG99_PRIV_REC_NEXT(2350, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2350(choice, ...)  METALANG99_PRIV_REC_NEXT(2351, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2351(choice, ...)  METALANG99_PRIV_REC_NEXT(2352, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2352(choice, ...)  METALANG99_PRIV_REC_NEXT(2353, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2353(choice, ...)  METALANG99_PRIV_REC_NEXT(2354, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2354(choice, ...)  METALANG99_PRIV_REC_NEXT(2355, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2355(choice, ...)  METALANG99_PRIV_REC_NEXT(2356, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2356(choice, ...)  METALANG99_PRIV_REC_NEXT(2357, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2357(choice, ...)  METALANG99_PRIV_REC_NEXT(2358, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2358(choice, ...)  METALANG99_PRIV_REC_NEXT(2359, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2359(choice, ...)  METALANG99_PRIV_REC_NEXT(2360, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2360(choice, ...)  METALANG99_PRIV_REC_NEXT(2361, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2361(choice, ...)  METALANG99_PRIV_REC_NEXT(2362, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2362(choice, ...)  METALANG99_PRIV_REC_NEXT(2363, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2363(choice, ...)  METALANG99_PRIV_REC_NEXT(2364, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2364(choice, ...)  METALANG99_PRIV_REC_NEXT(2365, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2365(choice, ...)  METALANG99_PRIV_REC_NEXT(2366, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2366(choice, ...)  METALANG99_PRIV_REC_NEXT(2367, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2367(choice, ...)  METALANG99_PRIV_REC_NEXT(2368, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2368(choice, ...)  METALANG99_PRIV_REC_NEXT(2369, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2369(choice, ...)  METALANG99_PRIV_REC_NEXT(2370, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2370(choice, ...)  METALANG99_PRIV_REC_NEXT(2371, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2371(choice, ...)  METALANG99_PRIV_REC_NEXT(2372, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2372(choice, ...)  METALANG99_PRIV_REC_NEXT(2373, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2373(choice, ...)  METALANG99_PRIV_REC_NEXT(2374, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2374(choice, ...)  METALANG99_PRIV_REC_NEXT(2375, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2375(choice, ...)  METALANG99_PRIV_REC_NEXT(2376, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2376(choice, ...)  METALANG99_PRIV_REC_NEXT(2377, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2377(choice, ...)  METALANG99_PRIV_REC_NEXT(2378, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2378(choice, ...)  METALANG99_PRIV_REC_NEXT(2379, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2379(choice, ...)  METALANG99_PRIV_REC_NEXT(2380, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2380(choice, ...)  METALANG99_PRIV_REC_NEXT(2381, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2381(choice, ...)  METALANG99_PRIV_REC_NEXT(2382, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2382(choice, ...)  METALANG99_PRIV_REC_NEXT(2383, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2383(choice, ...)  METALANG99_PRIV_REC_NEXT(2384, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2384(choice, ...)  METALANG99_PRIV_REC_NEXT(2385, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2385(choice, ...)  METALANG99_PRIV_REC_NEXT(2386, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2386(choice, ...)  METALANG99_PRIV_REC_NEXT(2387, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2387(choice, ...)  METALANG99_PRIV_REC_NEXT(2388, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2388(choice, ...)  METALANG99_PRIV_REC_NEXT(2389, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2389(choice, ...)  METALANG99_PRIV_REC_NEXT(2390, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2390(choice, ...)  METALANG99_PRIV_REC_NEXT(2391, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2391(choice, ...)  METALANG99_PRIV_REC_NEXT(2392, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2392(choice, ...)  METALANG99_PRIV_REC_NEXT(2393, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2393(choice, ...)  METALANG99_PRIV_REC_NEXT(2394, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2394(choice, ...)  METALANG99_PRIV_REC_NEXT(2395, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2395(choice, ...)  METALANG99_PRIV_REC_NEXT(2396, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2396(choice, ...)  METALANG99_PRIV_REC_NEXT(2397, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2397(choice, ...)  METALANG99_PRIV_REC_NEXT(2398, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2398(choice, ...)  METALANG99_PRIV_REC_NEXT(2399, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2399(choice, ...)  METALANG99_PRIV_REC_NEXT(2400, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2400(choice, ...)  METALANG99_PRIV_REC_NEXT(2401, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2401(choice, ...)  METALANG99_PRIV_REC_NEXT(2402, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2402(choice, ...)  METALANG99_PRIV_REC_NEXT(2403, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2403(choice, ...)  METALANG99_PRIV_REC_NEXT(2404, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2404(choice, ...)  METALANG99_PRIV_REC_NEXT(2405, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2405(choice, ...)  METALANG99_PRIV_REC_NEXT(2406, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2406(choice, ...)  METALANG99_PRIV_REC_NEXT(2407, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2407(choice, ...)  METALANG99_PRIV_REC_NEXT(2408, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2408(choice, ...)  METALANG99_PRIV_REC_NEXT(2409, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2409(choice, ...)  METALANG99_PRIV_REC_NEXT(2410, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2410(choice, ...)  METALANG99_PRIV_REC_NEXT(2411, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2411(choice, ...)  METALANG99_PRIV_REC_NEXT(2412, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2412(choice, ...)  METALANG99_PRIV_REC_NEXT(2413, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2413(choice, ...)  METALANG99_PRIV_REC_NEXT(2414, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2414(choice, ...)  METALANG99_PRIV_REC_NEXT(2415, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2415(choice, ...)  METALANG99_PRIV_REC_NEXT(2416, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2416(choice, ...)  METALANG99_PRIV_REC_NEXT(2417, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2417(choice, ...)  METALANG99_PRIV_REC_NEXT(2418, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2418(choice, ...)  METALANG99_PRIV_REC_NEXT(2419, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2419(choice, ...)  METALANG99_PRIV_REC_NEXT(2420, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2420(choice, ...)  METALANG99_PRIV_REC_NEXT(2421, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2421(choice, ...)  METALANG99_PRIV_REC_NEXT(2422, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2422(choice, ...)  METALANG99_PRIV_REC_NEXT(2423, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2423(choice, ...)  METALANG99_PRIV_REC_NEXT(2424, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2424(choice, ...)  METALANG99_PRIV_REC_NEXT(2425, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2425(choice, ...)  METALANG99_PRIV_REC_NEXT(2426, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2426(choice, ...)  METALANG99_PRIV_REC_NEXT(2427, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2427(choice, ...)  METALANG99_PRIV_REC_NEXT(2428, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2428(choice, ...)  METALANG99_PRIV_REC_NEXT(2429, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2429(choice, ...)  METALANG99_PRIV_REC_NEXT(2430, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2430(choice, ...)  METALANG99_PRIV_REC_NEXT(2431, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2431(choice, ...)  METALANG99_PRIV_REC_NEXT(2432, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2432(choice, ...)  METALANG99_PRIV_REC_NEXT(2433, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2433(choice, ...)  METALANG99_PRIV_REC_NEXT(2434, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2434(choice, ...)  METALANG99_PRIV_REC_NEXT(2435, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2435(choice, ...)  METALANG99_PRIV_REC_NEXT(2436, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2436(choice, ...)  METALANG99_PRIV_REC_NEXT(2437, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2437(choice, ...)  METALANG99_PRIV_REC_NEXT(2438, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2438(choice, ...)  METALANG99_PRIV_REC_NEXT(2439, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2439(choice, ...)  METALANG99_PRIV_REC_NEXT(2440, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2440(choice, ...)  METALANG99_PRIV_REC_NEXT(2441, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2441(choice, ...)  METALANG99_PRIV_REC_NEXT(2442, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2442(choice, ...)  METALANG99_PRIV_REC_NEXT(2443, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2443(choice, ...)  METALANG99_PRIV_REC_NEXT(2444, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2444(choice, ...)  METALANG99_PRIV_REC_NEXT(2445, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2445(choice, ...)  METALANG99_PRIV_REC_NEXT(2446, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2446(choice, ...)  METALANG99_PRIV_REC_NEXT(2447, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2447(choice, ...)  METALANG99_PRIV_REC_NEXT(2448, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2448(choice, ...)  METALANG99_PRIV_REC_NEXT(2449, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2449(choice, ...)  METALANG99_PRIV_REC_NEXT(2450, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2450(choice, ...)  METALANG99_PRIV_REC_NEXT(2451, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2451(choice, ...)  METALANG99_PRIV_REC_NEXT(2452, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2452(choice, ...)  METALANG99_PRIV_REC_NEXT(2453, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2453(choice, ...)  METALANG99_PRIV_REC_NEXT(2454, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2454(choice, ...)  METALANG99_PRIV_REC_NEXT(2455, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2455(choice, ...)  METALANG99_PRIV_REC_NEXT(2456, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2456(choice, ...)  METALANG99_PRIV_REC_NEXT(2457, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2457(choice, ...)  METALANG99_PRIV_REC_NEXT(2458, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2458(choice, ...)  METALANG99_PRIV_REC_NEXT(2459, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2459(choice, ...)  METALANG99_PRIV_REC_NEXT(2460, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2460(choice, ...)  METALANG99_PRIV_REC_NEXT(2461, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2461(choice, ...)  METALANG99_PRIV_REC_NEXT(2462, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2462(choice, ...)  METALANG99_PRIV_REC_NEXT(2463, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2463(choice, ...)  METALANG99_PRIV_REC_NEXT(2464, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2464(choice, ...)  METALANG99_PRIV_REC_NEXT(2465, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2465(choice, ...)  METALANG99_PRIV_REC_NEXT(2466, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2466(choice, ...)  METALANG99_PRIV_REC_NEXT(2467, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2467(choice, ...)  METALANG99_PRIV_REC_NEXT(2468, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2468(choice, ...)  METALANG99_PRIV_REC_NEXT(2469, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2469(choice, ...)  METALANG99_PRIV_REC_NEXT(2470, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2470(choice, ...)  METALANG99_PRIV_REC_NEXT(2471, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2471(choice, ...)  METALANG99_PRIV_REC_NEXT(2472, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2472(choice, ...)  METALANG99_PRIV_REC_NEXT(2473, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2473(choice, ...)  METALANG99_PRIV_REC_NEXT(2474, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2474(choice, ...)  METALANG99_PRIV_REC_NEXT(2475, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2475(choice, ...)  METALANG99_PRIV_REC_NEXT(2476, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2476(choice, ...)  METALANG99_PRIV_REC_NEXT(2477, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2477(choice, ...)  METALANG99_PRIV_REC_NEXT(2478, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2478(choice, ...)  METALANG99_PRIV_REC_NEXT(2479, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2479(choice, ...)  METALANG99_PRIV_REC_NEXT(2480, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2480(choice, ...)  METALANG99_PRIV_REC_NEXT(2481, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2481(choice, ...)  METALANG99_PRIV_REC_NEXT(2482, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2482(choice, ...)  METALANG99_PRIV_REC_NEXT(2483, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2483(choice, ...)  METALANG99_PRIV_REC_NEXT(2484, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2484(choice, ...)  METALANG99_PRIV_REC_NEXT(2485, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2485(choice, ...)  METALANG99_PRIV_REC_NEXT(2486, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2486(choice, ...)  METALANG99_PRIV_REC_NEXT(2487, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2487(choice, ...)  METALANG99_PRIV_REC_NEXT(2488, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2488(choice, ...)  METALANG99_PRIV_REC_NEXT(2489, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2489(choice, ...)  METALANG99_PRIV_REC_NEXT(2490, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2490(choice, ...)  METALANG99_PRIV_REC_NEXT(2491, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2491(choice, ...)  METALANG99_PRIV_REC_NEXT(2492, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2492(choice, ...)  METALANG99_PRIV_REC_NEXT(2493, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2493(choice, ...)  METALANG99_PRIV_REC_NEXT(2494, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2494(choice, ...)  METALANG99_PRIV_REC_NEXT(2495, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2495(choice, ...)  METALANG99_PRIV_REC_NEXT(2496, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2496(choice, ...)  METALANG99_PRIV_REC_NEXT(2497, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2497(choice, ...)  METALANG99_PRIV_REC_NEXT(2498, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2498(choice, ...)  METALANG99_PRIV_REC_NEXT(2499, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2499(choice, ...)  METALANG99_PRIV_REC_NEXT(2500, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2500(choice, ...)  METALANG99_PRIV_REC_NEXT(2501, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2501(choice, ...)  METALANG99_PRIV_REC_NEXT(2502, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2502(choice, ...)  METALANG99_PRIV_REC_NEXT(2503, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2503(choice, ...)  METALANG99_PRIV_REC_NEXT(2504, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2504(choice, ...)  METALANG99_PRIV_REC_NEXT(2505, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2505(choice, ...)  METALANG99_PRIV_REC_NEXT(2506, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2506(choice, ...)  METALANG99_PRIV_REC_NEXT(2507, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2507(choice, ...)  METALANG99_PRIV_REC_NEXT(2508, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2508(choice, ...)  METALANG99_PRIV_REC_NEXT(2509, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2509(choice, ...)  METALANG99_PRIV_REC_NEXT(2510, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2510(choice, ...)  METALANG99_PRIV_REC_NEXT(2511, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2511(choice, ...)  METALANG99_PRIV_REC_NEXT(2512, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2512(choice, ...)  METALANG99_PRIV_REC_NEXT(2513, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2513(choice, ...)  METALANG99_PRIV_REC_NEXT(2514, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2514(choice, ...)  METALANG99_PRIV_REC_NEXT(2515, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2515(choice, ...)  METALANG99_PRIV_REC_NEXT(2516, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2516(choice, ...)  METALANG99_PRIV_REC_NEXT(2517, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2517(choice, ...)  METALANG99_PRIV_REC_NEXT(2518, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2518(choice, ...)  METALANG99_PRIV_REC_NEXT(2519, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2519(choice, ...)  METALANG99_PRIV_REC_NEXT(2520, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2520(choice, ...)  METALANG99_PRIV_REC_NEXT(2521, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2521(choice, ...)  METALANG99_PRIV_REC_NEXT(2522, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2522(choice, ...)  METALANG99_PRIV_REC_NEXT(2523, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2523(choice, ...)  METALANG99_PRIV_REC_NEXT(2524, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2524(choice, ...)  METALANG99_PRIV_REC_NEXT(2525, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2525(choice, ...)  METALANG99_PRIV_REC_NEXT(2526, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2526(choice, ...)  METALANG99_PRIV_REC_NEXT(2527, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2527(choice, ...)  METALANG99_PRIV_REC_NEXT(2528, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2528(choice, ...)  METALANG99_PRIV_REC_NEXT(2529, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2529(choice, ...)  METALANG99_PRIV_REC_NEXT(2530, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2530(choice, ...)  METALANG99_PRIV_REC_NEXT(2531, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2531(choice, ...)  METALANG99_PRIV_REC_NEXT(2532, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2532(choice, ...)  METALANG99_PRIV_REC_NEXT(2533, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2533(choice, ...)  METALANG99_PRIV_REC_NEXT(2534, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2534(choice, ...)  METALANG99_PRIV_REC_NEXT(2535, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2535(choice, ...)  METALANG99_PRIV_REC_NEXT(2536, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2536(choice, ...)  METALANG99_PRIV_REC_NEXT(2537, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2537(choice, ...)  METALANG99_PRIV_REC_NEXT(2538, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2538(choice, ...)  METALANG99_PRIV_REC_NEXT(2539, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2539(choice, ...)  METALANG99_PRIV_REC_NEXT(2540, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2540(choice, ...)  METALANG99_PRIV_REC_NEXT(2541, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2541(choice, ...)  METALANG99_PRIV_REC_NEXT(2542, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2542(choice, ...)  METALANG99_PRIV_REC_NEXT(2543, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2543(choice, ...)  METALANG99_PRIV_REC_NEXT(2544, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2544(choice, ...)  METALANG99_PRIV_REC_NEXT(2545, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2545(choice, ...)  METALANG99_PRIV_REC_NEXT(2546, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2546(choice, ...)  METALANG99_PRIV_REC_NEXT(2547, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2547(choice, ...)  METALANG99_PRIV_REC_NEXT(2548, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2548(choice, ...)  METALANG99_PRIV_REC_NEXT(2549, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2549(choice, ...)  METALANG99_PRIV_REC_NEXT(2550, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2550(choice, ...)  METALANG99_PRIV_REC_NEXT(2551, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2551(choice, ...)  METALANG99_PRIV_REC_NEXT(2552, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2552(choice, ...)  METALANG99_PRIV_REC_NEXT(2553, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2553(choice, ...)  METALANG99_PRIV_REC_NEXT(2554, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2554(choice, ...)  METALANG99_PRIV_REC_NEXT(2555, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2555(choice, ...)  METALANG99_PRIV_REC_NEXT(2556, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2556(choice, ...)  METALANG99_PRIV_REC_NEXT(2557, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2557(choice, ...)  METALANG99_PRIV_REC_NEXT(2558, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2558(choice, ...)  METALANG99_PRIV_REC_NEXT(2559, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2559(choice, ...)  METALANG99_PRIV_REC_NEXT(2560, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2560(choice, ...)  METALANG99_PRIV_REC_NEXT(2561, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2561(choice, ...)  METALANG99_PRIV_REC_NEXT(2562, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2562(choice, ...)  METALANG99_PRIV_REC_NEXT(2563, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2563(choice, ...)  METALANG99_PRIV_REC_NEXT(2564, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2564(choice, ...)  METALANG99_PRIV_REC_NEXT(2565, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2565(choice, ...)  METALANG99_PRIV_REC_NEXT(2566, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2566(choice, ...)  METALANG99_PRIV_REC_NEXT(2567, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2567(choice, ...)  METALANG99_PRIV_REC_NEXT(2568, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2568(choice, ...)  METALANG99_PRIV_REC_NEXT(2569, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2569(choice, ...)  METALANG99_PRIV_REC_NEXT(2570, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2570(choice, ...)  METALANG99_PRIV_REC_NEXT(2571, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2571(choice, ...)  METALANG99_PRIV_REC_NEXT(2572, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2572(choice, ...)  METALANG99_PRIV_REC_NEXT(2573, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2573(choice, ...)  METALANG99_PRIV_REC_NEXT(2574, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2574(choice, ...)  METALANG99_PRIV_REC_NEXT(2575, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2575(choice, ...)  METALANG99_PRIV_REC_NEXT(2576, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2576(choice, ...)  METALANG99_PRIV_REC_NEXT(2577, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2577(choice, ...)  METALANG99_PRIV_REC_NEXT(2578, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2578(choice, ...)  METALANG99_PRIV_REC_NEXT(2579, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2579(choice, ...)  METALANG99_PRIV_REC_NEXT(2580, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2580(choice, ...)  METALANG99_PRIV_REC_NEXT(2581, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2581(choice, ...)  METALANG99_PRIV_REC_NEXT(2582, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2582(choice, ...)  METALANG99_PRIV_REC_NEXT(2583, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2583(choice, ...)  METALANG99_PRIV_REC_NEXT(2584, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2584(choice, ...)  METALANG99_PRIV_REC_NEXT(2585, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2585(choice, ...)  METALANG99_PRIV_REC_NEXT(2586, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2586(choice, ...)  METALANG99_PRIV_REC_NEXT(2587, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2587(choice, ...)  METALANG99_PRIV_REC_NEXT(2588, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2588(choice, ...)  METALANG99_PRIV_REC_NEXT(2589, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2589(choice, ...)  METALANG99_PRIV_REC_NEXT(2590, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2590(choice, ...)  METALANG99_PRIV_REC_NEXT(2591, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2591(choice, ...)  METALANG99_PRIV_REC_NEXT(2592, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2592(choice, ...)  METALANG99_PRIV_REC_NEXT(2593, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2593(choice, ...)  METALANG99_PRIV_REC_NEXT(2594, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2594(choice, ...)  METALANG99_PRIV_REC_NEXT(2595, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2595(choice, ...)  METALANG99_PRIV_REC_NEXT(2596, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2596(choice, ...)  METALANG99_PRIV_REC_NEXT(2597, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2597(choice, ...)  METALANG99_PRIV_REC_NEXT(2598, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2598(choice, ...)  METALANG99_PRIV_REC_NEXT(2599, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2599(choice, ...)  METALANG99_PRIV_REC_NEXT(2600, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2600(choice, ...)  METALANG99_PRIV_REC_NEXT(2601, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2601(choice, ...)  METALANG99_PRIV_REC_NEXT(2602, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2602(choice, ...)  METALANG99_PRIV_REC_NEXT(2603, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2603(choice, ...)  METALANG99_PRIV_REC_NEXT(2604, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2604(choice, ...)  METALANG99_PRIV_REC_NEXT(2605, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2605(choice, ...)  METALANG99_PRIV_REC_NEXT(2606, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2606(choice, ...)  METALANG99_PRIV_REC_NEXT(2607, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2607(choice, ...)  METALANG99_PRIV_REC_NEXT(2608, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2608(choice, ...)  METALANG99_PRIV_REC_NEXT(2609, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2609(choice, ...)  METALANG99_PRIV_REC_NEXT(2610, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2610(choice, ...)  METALANG99_PRIV_REC_NEXT(2611, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2611(choice, ...)  METALANG99_PRIV_REC_NEXT(2612, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2612(choice, ...)  METALANG99_PRIV_REC_NEXT(2613, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2613(choice, ...)  METALANG99_PRIV_REC_NEXT(2614, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2614(choice, ...)  METALANG99_PRIV_REC_NEXT(2615, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2615(choice, ...)  METALANG99_PRIV_REC_NEXT(2616, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2616(choice, ...)  METALANG99_PRIV_REC_NEXT(2617, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2617(choice, ...)  METALANG99_PRIV_REC_NEXT(2618, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2618(choice, ...)  METALANG99_PRIV_REC_NEXT(2619, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2619(choice, ...)  METALANG99_PRIV_REC_NEXT(2620, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2620(choice, ...)  METALANG99_PRIV_REC_NEXT(2621, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2621(choice, ...)  METALANG99_PRIV_REC_NEXT(2622, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2622(choice, ...)  METALANG99_PRIV_REC_NEXT(2623, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2623(choice, ...)  METALANG99_PRIV_REC_NEXT(2624, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2624(choice, ...)  METALANG99_PRIV_REC_NEXT(2625, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2625(choice, ...)  METALANG99_PRIV_REC_NEXT(2626, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2626(choice, ...)  METALANG99_PRIV_REC_NEXT(2627, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2627(choice, ...)  METALANG99_PRIV_REC_NEXT(2628, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2628(choice, ...)  METALANG99_PRIV_REC_NEXT(2629, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2629(choice, ...)  METALANG99_PRIV_REC_NEXT(2630, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2630(choice, ...)  METALANG99_PRIV_REC_NEXT(2631, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2631(choice, ...)  METALANG99_PRIV_REC_NEXT(2632, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2632(choice, ...)  METALANG99_PRIV_REC_NEXT(2633, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2633(choice, ...)  METALANG99_PRIV_REC_NEXT(2634, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2634(choice, ...)  METALANG99_PRIV_REC_NEXT(2635, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2635(choice, ...)  METALANG99_PRIV_REC_NEXT(2636, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2636(choice, ...)  METALANG99_PRIV_REC_NEXT(2637, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2637(choice, ...)  METALANG99_PRIV_REC_NEXT(2638, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2638(choice, ...)  METALANG99_PRIV_REC_NEXT(2639, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2639(choice, ...)  METALANG99_PRIV_REC_NEXT(2640, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2640(choice, ...)  METALANG99_PRIV_REC_NEXT(2641, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2641(choice, ...)  METALANG99_PRIV_REC_NEXT(2642, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2642(choice, ...)  METALANG99_PRIV_REC_NEXT(2643, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2643(choice, ...)  METALANG99_PRIV_REC_NEXT(2644, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2644(choice, ...)  METALANG99_PRIV_REC_NEXT(2645, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2645(choice, ...)  METALANG99_PRIV_REC_NEXT(2646, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2646(choice, ...)  METALANG99_PRIV_REC_NEXT(2647, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2647(choice, ...)  METALANG99_PRIV_REC_NEXT(2648, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2648(choice, ...)  METALANG99_PRIV_REC_NEXT(2649, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2649(choice, ...)  METALANG99_PRIV_REC_NEXT(2650, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2650(choice, ...)  METALANG99_PRIV_REC_NEXT(2651, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2651(choice, ...)  METALANG99_PRIV_REC_NEXT(2652, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2652(choice, ...)  METALANG99_PRIV_REC_NEXT(2653, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2653(choice, ...)  METALANG99_PRIV_REC_NEXT(2654, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2654(choice, ...)  METALANG99_PRIV_REC_NEXT(2655, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2655(choice, ...)  METALANG99_PRIV_REC_NEXT(2656, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2656(choice, ...)  METALANG99_PRIV_REC_NEXT(2657, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2657(choice, ...)  METALANG99_PRIV_REC_NEXT(2658, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2658(choice, ...)  METALANG99_PRIV_REC_NEXT(2659, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2659(choice, ...)  METALANG99_PRIV_REC_NEXT(2660, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2660(choice, ...)  METALANG99_PRIV_REC_NEXT(2661, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2661(choice, ...)  METALANG99_PRIV_REC_NEXT(2662, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2662(choice, ...)  METALANG99_PRIV_REC_NEXT(2663, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2663(choice, ...)  METALANG99_PRIV_REC_NEXT(2664, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2664(choice, ...)  METALANG99_PRIV_REC_NEXT(2665, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2665(choice, ...)  METALANG99_PRIV_REC_NEXT(2666, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2666(choice, ...)  METALANG99_PRIV_REC_NEXT(2667, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2667(choice, ...)  METALANG99_PRIV_REC_NEXT(2668, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2668(choice, ...)  METALANG99_PRIV_REC_NEXT(2669, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2669(choice, ...)  METALANG99_PRIV_REC_NEXT(2670, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2670(choice, ...)  METALANG99_PRIV_REC_NEXT(2671, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2671(choice, ...)  METALANG99_PRIV_REC_NEXT(2672, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2672(choice, ...)  METALANG99_PRIV_REC_NEXT(2673, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2673(choice, ...)  METALANG99_PRIV_REC_NEXT(2674, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2674(choice, ...)  METALANG99_PRIV_REC_NEXT(2675, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2675(choice, ...)  METALANG99_PRIV_REC_NEXT(2676, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2676(choice, ...)  METALANG99_PRIV_REC_NEXT(2677, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2677(choice, ...)  METALANG99_PRIV_REC_NEXT(2678, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2678(choice, ...)  METALANG99_PRIV_REC_NEXT(2679, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2679(choice, ...)  METALANG99_PRIV_REC_NEXT(2680, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2680(choice, ...)  METALANG99_PRIV_REC_NEXT(2681, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2681(choice, ...)  METALANG99_PRIV_REC_NEXT(2682, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2682(choice, ...)  METALANG99_PRIV_REC_NEXT(2683, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2683(choice, ...)  METALANG99_PRIV_REC_NEXT(2684, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2684(choice, ...)  METALANG99_PRIV_REC_NEXT(2685, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2685(choice, ...)  METALANG99_PRIV_REC_NEXT(2686, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2686(choice, ...)  METALANG99_PRIV_REC_NEXT(2687, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2687(choice, ...)  METALANG99_PRIV_REC_NEXT(2688, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2688(choice, ...)  METALANG99_PRIV_REC_NEXT(2689, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2689(choice, ...)  METALANG99_PRIV_REC_NEXT(2690, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2690(choice, ...)  METALANG99_PRIV_REC_NEXT(2691, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2691(choice, ...)  METALANG99_PRIV_REC_NEXT(2692, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2692(choice, ...)  METALANG99_PRIV_REC_NEXT(2693, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2693(choice, ...)  METALANG99_PRIV_REC_NEXT(2694, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2694(choice, ...)  METALANG99_PRIV_REC_NEXT(2695, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2695(choice, ...)  METALANG99_PRIV_REC_NEXT(2696, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2696(choice, ...)  METALANG99_PRIV_REC_NEXT(2697, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2697(choice, ...)  METALANG99_PRIV_REC_NEXT(2698, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2698(choice, ...)  METALANG99_PRIV_REC_NEXT(2699, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2699(choice, ...)  METALANG99_PRIV_REC_NEXT(2700, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2700(choice, ...)  METALANG99_PRIV_REC_NEXT(2701, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2701(choice, ...)  METALANG99_PRIV_REC_NEXT(2702, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2702(choice, ...)  METALANG99_PRIV_REC_NEXT(2703, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2703(choice, ...)  METALANG99_PRIV_REC_NEXT(2704, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2704(choice, ...)  METALANG99_PRIV_REC_NEXT(2705, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2705(choice, ...)  METALANG99_PRIV_REC_NEXT(2706, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2706(choice, ...)  METALANG99_PRIV_REC_NEXT(2707, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2707(choice, ...)  METALANG99_PRIV_REC_NEXT(2708, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2708(choice, ...)  METALANG99_PRIV_REC_NEXT(2709, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2709(choice, ...)  METALANG99_PRIV_REC_NEXT(2710, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2710(choice, ...)  METALANG99_PRIV_REC_NEXT(2711, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2711(choice, ...)  METALANG99_PRIV_REC_NEXT(2712, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2712(choice, ...)  METALANG99_PRIV_REC_NEXT(2713, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2713(choice, ...)  METALANG99_PRIV_REC_NEXT(2714, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2714(choice, ...)  METALANG99_PRIV_REC_NEXT(2715, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2715(choice, ...)  METALANG99_PRIV_REC_NEXT(2716, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2716(choice, ...)  METALANG99_PRIV_REC_NEXT(2717, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2717(choice, ...)  METALANG99_PRIV_REC_NEXT(2718, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2718(choice, ...)  METALANG99_PRIV_REC_NEXT(2719, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2719(choice, ...)  METALANG99_PRIV_REC_NEXT(2720, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2720(choice, ...)  METALANG99_PRIV_REC_NEXT(2721, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2721(choice, ...)  METALANG99_PRIV_REC_NEXT(2722, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2722(choice, ...)  METALANG99_PRIV_REC_NEXT(2723, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2723(choice, ...)  METALANG99_PRIV_REC_NEXT(2724, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2724(choice, ...)  METALANG99_PRIV_REC_NEXT(2725, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2725(choice, ...)  METALANG99_PRIV_REC_NEXT(2726, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2726(choice, ...)  METALANG99_PRIV_REC_NEXT(2727, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2727(choice, ...)  METALANG99_PRIV_REC_NEXT(2728, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2728(choice, ...)  METALANG99_PRIV_REC_NEXT(2729, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2729(choice, ...)  METALANG99_PRIV_REC_NEXT(2730, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2730(choice, ...)  METALANG99_PRIV_REC_NEXT(2731, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2731(choice, ...)  METALANG99_PRIV_REC_NEXT(2732, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2732(choice, ...)  METALANG99_PRIV_REC_NEXT(2733, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2733(choice, ...)  METALANG99_PRIV_REC_NEXT(2734, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2734(choice, ...)  METALANG99_PRIV_REC_NEXT(2735, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2735(choice, ...)  METALANG99_PRIV_REC_NEXT(2736, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2736(choice, ...)  METALANG99_PRIV_REC_NEXT(2737, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2737(choice, ...)  METALANG99_PRIV_REC_NEXT(2738, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2738(choice, ...)  METALANG99_PRIV_REC_NEXT(2739, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2739(choice, ...)  METALANG99_PRIV_REC_NEXT(2740, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2740(choice, ...)  METALANG99_PRIV_REC_NEXT(2741, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2741(choice, ...)  METALANG99_PRIV_REC_NEXT(2742, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2742(choice, ...)  METALANG99_PRIV_REC_NEXT(2743, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2743(choice, ...)  METALANG99_PRIV_REC_NEXT(2744, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2744(choice, ...)  METALANG99_PRIV_REC_NEXT(2745, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2745(choice, ...)  METALANG99_PRIV_REC_NEXT(2746, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2746(choice, ...)  METALANG99_PRIV_REC_NEXT(2747, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2747(choice, ...)  METALANG99_PRIV_REC_NEXT(2748, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2748(choice, ...)  METALANG99_PRIV_REC_NEXT(2749, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2749(choice, ...)  METALANG99_PRIV_REC_NEXT(2750, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2750(choice, ...)  METALANG99_PRIV_REC_NEXT(2751, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2751(choice, ...)  METALANG99_PRIV_REC_NEXT(2752, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2752(choice, ...)  METALANG99_PRIV_REC_NEXT(2753, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2753(choice, ...)  METALANG99_PRIV_REC_NEXT(2754, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2754(choice, ...)  METALANG99_PRIV_REC_NEXT(2755, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2755(choice, ...)  METALANG99_PRIV_REC_NEXT(2756, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2756(choice, ...)  METALANG99_PRIV_REC_NEXT(2757, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2757(choice, ...)  METALANG99_PRIV_REC_NEXT(2758, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2758(choice, ...)  METALANG99_PRIV_REC_NEXT(2759, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2759(choice, ...)  METALANG99_PRIV_REC_NEXT(2760, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2760(choice, ...)  METALANG99_PRIV_REC_NEXT(2761, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2761(choice, ...)  METALANG99_PRIV_REC_NEXT(2762, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2762(choice, ...)  METALANG99_PRIV_REC_NEXT(2763, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2763(choice, ...)  METALANG99_PRIV_REC_NEXT(2764, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2764(choice, ...)  METALANG99_PRIV_REC_NEXT(2765, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2765(choice, ...)  METALANG99_PRIV_REC_NEXT(2766, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2766(choice, ...)  METALANG99_PRIV_REC_NEXT(2767, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2767(choice, ...)  METALANG99_PRIV_REC_NEXT(2768, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2768(choice, ...)  METALANG99_PRIV_REC_NEXT(2769, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2769(choice, ...)  METALANG99_PRIV_REC_NEXT(2770, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2770(choice, ...)  METALANG99_PRIV_REC_NEXT(2771, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2771(choice, ...)  METALANG99_PRIV_REC_NEXT(2772, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2772(choice, ...)  METALANG99_PRIV_REC_NEXT(2773, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2773(choice, ...)  METALANG99_PRIV_REC_NEXT(2774, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2774(choice, ...)  METALANG99_PRIV_REC_NEXT(2775, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2775(choice, ...)  METALANG99_PRIV_REC_NEXT(2776, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2776(choice, ...)  METALANG99_PRIV_REC_NEXT(2777, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2777(choice, ...)  METALANG99_PRIV_REC_NEXT(2778, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2778(choice, ...)  METALANG99_PRIV_REC_NEXT(2779, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2779(choice, ...)  METALANG99_PRIV_REC_NEXT(2780, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2780(choice, ...)  METALANG99_PRIV_REC_NEXT(2781, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2781(choice, ...)  METALANG99_PRIV_REC_NEXT(2782, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2782(choice, ...)  METALANG99_PRIV_REC_NEXT(2783, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2783(choice, ...)  METALANG99_PRIV_REC_NEXT(2784, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2784(choice, ...)  METALANG99_PRIV_REC_NEXT(2785, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2785(choice, ...)  METALANG99_PRIV_REC_NEXT(2786, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2786(choice, ...)  METALANG99_PRIV_REC_NEXT(2787, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2787(choice, ...)  METALANG99_PRIV_REC_NEXT(2788, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2788(choice, ...)  METALANG99_PRIV_REC_NEXT(2789, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2789(choice, ...)  METALANG99_PRIV_REC_NEXT(2790, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2790(choice, ...)  METALANG99_PRIV_REC_NEXT(2791, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2791(choice, ...)  METALANG99_PRIV_REC_NEXT(2792, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2792(choice, ...)  METALANG99_PRIV_REC_NEXT(2793, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2793(choice, ...)  METALANG99_PRIV_REC_NEXT(2794, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2794(choice, ...)  METALANG99_PRIV_REC_NEXT(2795, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2795(choice, ...)  METALANG99_PRIV_REC_NEXT(2796, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2796(choice, ...)  METALANG99_PRIV_REC_NEXT(2797, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2797(choice, ...)  METALANG99_PRIV_REC_NEXT(2798, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2798(choice, ...)  METALANG99_PRIV_REC_NEXT(2799, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2799(choice, ...)  METALANG99_PRIV_REC_NEXT(2800, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2800(choice, ...)  METALANG99_PRIV_REC_NEXT(2801, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2801(choice, ...)  METALANG99_PRIV_REC_NEXT(2802, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2802(choice, ...)  METALANG99_PRIV_REC_NEXT(2803, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2803(choice, ...)  METALANG99_PRIV_REC_NEXT(2804, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2804(choice, ...)  METALANG99_PRIV_REC_NEXT(2805, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2805(choice, ...)  METALANG99_PRIV_REC_NEXT(2806, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2806(choice, ...)  METALANG99_PRIV_REC_NEXT(2807, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2807(choice, ...)  METALANG99_PRIV_REC_NEXT(2808, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2808(choice, ...)  METALANG99_PRIV_REC_NEXT(2809, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2809(choice, ...)  METALANG99_PRIV_REC_NEXT(2810, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2810(choice, ...)  METALANG99_PRIV_REC_NEXT(2811, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2811(choice, ...)  METALANG99_PRIV_REC_NEXT(2812, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2812(choice, ...)  METALANG99_PRIV_REC_NEXT(2813, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2813(choice, ...)  METALANG99_PRIV_REC_NEXT(2814, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2814(choice, ...)  METALANG99_PRIV_REC_NEXT(2815, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2815(choice, ...)  METALANG99_PRIV_REC_NEXT(2816, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2816(choice, ...)  METALANG99_PRIV_REC_NEXT(2817, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2817(choice, ...)  METALANG99_PRIV_REC_NEXT(2818, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2818(choice, ...)  METALANG99_PRIV_REC_NEXT(2819, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2819(choice, ...)  METALANG99_PRIV_REC_NEXT(2820, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2820(choice, ...)  METALANG99_PRIV_REC_NEXT(2821, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2821(choice, ...)  METALANG99_PRIV_REC_NEXT(2822, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2822(choice, ...)  METALANG99_PRIV_REC_NEXT(2823, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2823(choice, ...)  METALANG99_PRIV_REC_NEXT(2824, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2824(choice, ...)  METALANG99_PRIV_REC_NEXT(2825, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2825(choice, ...)  METALANG99_PRIV_REC_NEXT(2826, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2826(choice, ...)  METALANG99_PRIV_REC_NEXT(2827, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2827(choice, ...)  METALANG99_PRIV_REC_NEXT(2828, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2828(choice, ...)  METALANG99_PRIV_REC_NEXT(2829, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2829(choice, ...)  METALANG99_PRIV_REC_NEXT(2830, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2830(choice, ...)  METALANG99_PRIV_REC_NEXT(2831, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2831(choice, ...)  METALANG99_PRIV_REC_NEXT(2832, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2832(choice, ...)  METALANG99_PRIV_REC_NEXT(2833, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2833(choice, ...)  METALANG99_PRIV_REC_NEXT(2834, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2834(choice, ...)  METALANG99_PRIV_REC_NEXT(2835, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2835(choice, ...)  METALANG99_PRIV_REC_NEXT(2836, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2836(choice, ...)  METALANG99_PRIV_REC_NEXT(2837, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2837(choice, ...)  METALANG99_PRIV_REC_NEXT(2838, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2838(choice, ...)  METALANG99_PRIV_REC_NEXT(2839, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2839(choice, ...)  METALANG99_PRIV_REC_NEXT(2840, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2840(choice, ...)  METALANG99_PRIV_REC_NEXT(2841, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2841(choice, ...)  METALANG99_PRIV_REC_NEXT(2842, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2842(choice, ...)  METALANG99_PRIV_REC_NEXT(2843, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2843(choice, ...)  METALANG99_PRIV_REC_NEXT(2844, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2844(choice, ...)  METALANG99_PRIV_REC_NEXT(2845, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2845(choice, ...)  METALANG99_PRIV_REC_NEXT(2846, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2846(choice, ...)  METALANG99_PRIV_REC_NEXT(2847, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2847(choice, ...)  METALANG99_PRIV_REC_NEXT(2848, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2848(choice, ...)  METALANG99_PRIV_REC_NEXT(2849, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2849(choice, ...)  METALANG99_PRIV_REC_NEXT(2850, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2850(choice, ...)  METALANG99_PRIV_REC_NEXT(2851, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2851(choice, ...)  METALANG99_PRIV_REC_NEXT(2852, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2852(choice, ...)  METALANG99_PRIV_REC_NEXT(2853, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2853(choice, ...)  METALANG99_PRIV_REC_NEXT(2854, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2854(choice, ...)  METALANG99_PRIV_REC_NEXT(2855, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2855(choice, ...)  METALANG99_PRIV_REC_NEXT(2856, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2856(choice, ...)  METALANG99_PRIV_REC_NEXT(2857, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2857(choice, ...)  METALANG99_PRIV_REC_NEXT(2858, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2858(choice, ...)  METALANG99_PRIV_REC_NEXT(2859, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2859(choice, ...)  METALANG99_PRIV_REC_NEXT(2860, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2860(choice, ...)  METALANG99_PRIV_REC_NEXT(2861, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2861(choice, ...)  METALANG99_PRIV_REC_NEXT(2862, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2862(choice, ...)  METALANG99_PRIV_REC_NEXT(2863, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2863(choice, ...)  METALANG99_PRIV_REC_NEXT(2864, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2864(choice, ...)  METALANG99_PRIV_REC_NEXT(2865, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2865(choice, ...)  METALANG99_PRIV_REC_NEXT(2866, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2866(choice, ...)  METALANG99_PRIV_REC_NEXT(2867, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2867(choice, ...)  METALANG99_PRIV_REC_NEXT(2868, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2868(choice, ...)  METALANG99_PRIV_REC_NEXT(2869, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2869(choice, ...)  METALANG99_PRIV_REC_NEXT(2870, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2870(choice, ...)  METALANG99_PRIV_REC_NEXT(2871, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2871(choice, ...)  METALANG99_PRIV_REC_NEXT(2872, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2872(choice, ...)  METALANG99_PRIV_REC_NEXT(2873, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2873(choice, ...)  METALANG99_PRIV_REC_NEXT(2874, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2874(choice, ...)  METALANG99_PRIV_REC_NEXT(2875, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2875(choice, ...)  METALANG99_PRIV_REC_NEXT(2876, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2876(choice, ...)  METALANG99_PRIV_REC_NEXT(2877, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2877(choice, ...)  METALANG99_PRIV_REC_NEXT(2878, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2878(choice, ...)  METALANG99_PRIV_REC_NEXT(2879, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2879(choice, ...)  METALANG99_PRIV_REC_NEXT(2880, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2880(choice, ...)  METALANG99_PRIV_REC_NEXT(2881, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2881(choice, ...)  METALANG99_PRIV_REC_NEXT(2882, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2882(choice, ...)  METALANG99_PRIV_REC_NEXT(2883, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2883(choice, ...)  METALANG99_PRIV_REC_NEXT(2884, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2884(choice, ...)  METALANG99_PRIV_REC_NEXT(2885, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2885(choice, ...)  METALANG99_PRIV_REC_NEXT(2886, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2886(choice, ...)  METALANG99_PRIV_REC_NEXT(2887, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2887(choice, ...)  METALANG99_PRIV_REC_NEXT(2888, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2888(choice, ...)  METALANG99_PRIV_REC_NEXT(2889, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2889(choice, ...)  METALANG99_PRIV_REC_NEXT(2890, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2890(choice, ...)  METALANG99_PRIV_REC_NEXT(2891, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2891(choice, ...)  METALANG99_PRIV_REC_NEXT(2892, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2892(choice, ...)  METALANG99_PRIV_REC_NEXT(2893, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2893(choice, ...)  METALANG99_PRIV_REC_NEXT(2894, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2894(choice, ...)  METALANG99_PRIV_REC_NEXT(2895, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2895(choice, ...)  METALANG99_PRIV_REC_NEXT(2896, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2896(choice, ...)  METALANG99_PRIV_REC_NEXT(2897, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2897(choice, ...)  METALANG99_PRIV_REC_NEXT(2898, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2898(choice, ...)  METALANG99_PRIV_REC_NEXT(2899, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2899(choice, ...)  METALANG99_PRIV_REC_NEXT(2900, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2900(choice, ...)  METALANG99_PRIV_REC_NEXT(2901, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2901(choice, ...)  METALANG99_PRIV_REC_NEXT(2902, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2902(choice, ...)  METALANG99_PRIV_REC_NEXT(2903, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2903(choice, ...)  METALANG99_PRIV_REC_NEXT(2904, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2904(choice, ...)  METALANG99_PRIV_REC_NEXT(2905, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2905(choice, ...)  METALANG99_PRIV_REC_NEXT(2906, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2906(choice, ...)  METALANG99_PRIV_REC_NEXT(2907, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2907(choice, ...)  METALANG99_PRIV_REC_NEXT(2908, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2908(choice, ...)  METALANG99_PRIV_REC_NEXT(2909, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2909(choice, ...)  METALANG99_PRIV_REC_NEXT(2910, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2910(choice, ...)  METALANG99_PRIV_REC_NEXT(2911, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2911(choice, ...)  METALANG99_PRIV_REC_NEXT(2912, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2912(choice, ...)  METALANG99_PRIV_REC_NEXT(2913, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2913(choice, ...)  METALANG99_PRIV_REC_NEXT(2914, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2914(choice, ...)  METALANG99_PRIV_REC_NEXT(2915, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2915(choice, ...)  METALANG99_PRIV_REC_NEXT(2916, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2916(choice, ...)  METALANG99_PRIV_REC_NEXT(2917, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2917(choice, ...)  METALANG99_PRIV_REC_NEXT(2918, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2918(choice, ...)  METALANG99_PRIV_REC_NEXT(2919, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2919(choice, ...)  METALANG99_PRIV_REC_NEXT(2920, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2920(choice, ...)  METALANG99_PRIV_REC_NEXT(2921, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2921(choice, ...)  METALANG99_PRIV_REC_NEXT(2922, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2922(choice, ...)  METALANG99_PRIV_REC_NEXT(2923, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2923(choice, ...)  METALANG99_PRIV_REC_NEXT(2924, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2924(choice, ...)  METALANG99_PRIV_REC_NEXT(2925, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2925(choice, ...)  METALANG99_PRIV_REC_NEXT(2926, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2926(choice, ...)  METALANG99_PRIV_REC_NEXT(2927, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2927(choice, ...)  METALANG99_PRIV_REC_NEXT(2928, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2928(choice, ...)  METALANG99_PRIV_REC_NEXT(2929, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2929(choice, ...)  METALANG99_PRIV_REC_NEXT(2930, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2930(choice, ...)  METALANG99_PRIV_REC_NEXT(2931, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2931(choice, ...)  METALANG99_PRIV_REC_NEXT(2932, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2932(choice, ...)  METALANG99_PRIV_REC_NEXT(2933, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2933(choice, ...)  METALANG99_PRIV_REC_NEXT(2934, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2934(choice, ...)  METALANG99_PRIV_REC_NEXT(2935, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2935(choice, ...)  METALANG99_PRIV_REC_NEXT(2936, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2936(choice, ...)  METALANG99_PRIV_REC_NEXT(2937, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2937(choice, ...)  METALANG99_PRIV_REC_NEXT(2938, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2938(choice, ...)  METALANG99_PRIV_REC_NEXT(2939, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2939(choice, ...)  METALANG99_PRIV_REC_NEXT(2940, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2940(choice, ...)  METALANG99_PRIV_REC_NEXT(2941, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2941(choice, ...)  METALANG99_PRIV_REC_NEXT(2942, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2942(choice, ...)  METALANG99_PRIV_REC_NEXT(2943, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2943(choice, ...)  METALANG99_PRIV_REC_NEXT(2944, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2944(choice, ...)  METALANG99_PRIV_REC_NEXT(2945, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2945(choice, ...)  METALANG99_PRIV_REC_NEXT(2946, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2946(choice, ...)  METALANG99_PRIV_REC_NEXT(2947, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2947(choice, ...)  METALANG99_PRIV_REC_NEXT(2948, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2948(choice, ...)  METALANG99_PRIV_REC_NEXT(2949, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2949(choice, ...)  METALANG99_PRIV_REC_NEXT(2950, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2950(choice, ...)  METALANG99_PRIV_REC_NEXT(2951, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2951(choice, ...)  METALANG99_PRIV_REC_NEXT(2952, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2952(choice, ...)  METALANG99_PRIV_REC_NEXT(2953, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2953(choice, ...)  METALANG99_PRIV_REC_NEXT(2954, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2954(choice, ...)  METALANG99_PRIV_REC_NEXT(2955, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2955(choice, ...)  METALANG99_PRIV_REC_NEXT(2956, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2956(choice, ...)  METALANG99_PRIV_REC_NEXT(2957, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2957(choice, ...)  METALANG99_PRIV_REC_NEXT(2958, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2958(choice, ...)  METALANG99_PRIV_REC_NEXT(2959, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2959(choice, ...)  METALANG99_PRIV_REC_NEXT(2960, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2960(choice, ...)  METALANG99_PRIV_REC_NEXT(2961, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2961(choice, ...)  METALANG99_PRIV_REC_NEXT(2962, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2962(choice, ...)  METALANG99_PRIV_REC_NEXT(2963, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2963(choice, ...)  METALANG99_PRIV_REC_NEXT(2964, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2964(choice, ...)  METALANG99_PRIV_REC_NEXT(2965, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2965(choice, ...)  METALANG99_PRIV_REC_NEXT(2966, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2966(choice, ...)  METALANG99_PRIV_REC_NEXT(2967, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2967(choice, ...)  METALANG99_PRIV_REC_NEXT(2968, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2968(choice, ...)  METALANG99_PRIV_REC_NEXT(2969, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2969(choice, ...)  METALANG99_PRIV_REC_NEXT(2970, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2970(choice, ...)  METALANG99_PRIV_REC_NEXT(2971, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2971(choice, ...)  METALANG99_PRIV_REC_NEXT(2972, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2972(choice, ...)  METALANG99_PRIV_REC_NEXT(2973, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2973(choice, ...)  METALANG99_PRIV_REC_NEXT(2974, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2974(choice, ...)  METALANG99_PRIV_REC_NEXT(2975, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2975(choice, ...)  METALANG99_PRIV_REC_NEXT(2976, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2976(choice, ...)  METALANG99_PRIV_REC_NEXT(2977, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2977(choice, ...)  METALANG99_PRIV_REC_NEXT(2978, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2978(choice, ...)  METALANG99_PRIV_REC_NEXT(2979, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2979(choice, ...)  METALANG99_PRIV_REC_NEXT(2980, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2980(choice, ...)  METALANG99_PRIV_REC_NEXT(2981, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2981(choice, ...)  METALANG99_PRIV_REC_NEXT(2982, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2982(choice, ...)  METALANG99_PRIV_REC_NEXT(2983, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2983(choice, ...)  METALANG99_PRIV_REC_NEXT(2984, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2984(choice, ...)  METALANG99_PRIV_REC_NEXT(2985, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2985(choice, ...)  METALANG99_PRIV_REC_NEXT(2986, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2986(choice, ...)  METALANG99_PRIV_REC_NEXT(2987, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2987(choice, ...)  METALANG99_PRIV_REC_NEXT(2988, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2988(choice, ...)  METALANG99_PRIV_REC_NEXT(2989, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2989(choice, ...)  METALANG99_PRIV_REC_NEXT(2990, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2990(choice, ...)  METALANG99_PRIV_REC_NEXT(2991, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2991(choice, ...)  METALANG99_PRIV_REC_NEXT(2992, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2992(choice, ...)  METALANG99_PRIV_REC_NEXT(2993, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2993(choice, ...)  METALANG99_PRIV_REC_NEXT(2994, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2994(choice, ...)  METALANG99_PRIV_REC_NEXT(2995, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2995(choice, ...)  METALANG99_PRIV_REC_NEXT(2996, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2996(choice, ...)  METALANG99_PRIV_REC_NEXT(2997, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2997(choice, ...)  METALANG99_PRIV_REC_NEXT(2998, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2998(choice, ...)  METALANG99_PRIV_REC_NEXT(2999, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_2999(choice, ...)  METALANG99_PRIV_REC_NEXT(3000, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3000(choice, ...)  METALANG99_PRIV_REC_NEXT(3001, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3001(choice, ...)  METALANG99_PRIV_REC_NEXT(3002, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3002(choice, ...)  METALANG99_PRIV_REC_NEXT(3003, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3003(choice, ...)  METALANG99_PRIV_REC_NEXT(3004, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3004(choice, ...)  METALANG99_PRIV_REC_NEXT(3005, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3005(choice, ...)  METALANG99_PRIV_REC_NEXT(3006, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3006(choice, ...)  METALANG99_PRIV_REC_NEXT(3007, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3007(choice, ...)  METALANG99_PRIV_REC_NEXT(3008, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3008(choice, ...)  METALANG99_PRIV_REC_NEXT(3009, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3009(choice, ...)  METALANG99_PRIV_REC_NEXT(3010, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3010(choice, ...)  METALANG99_PRIV_REC_NEXT(3011, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3011(choice, ...)  METALANG99_PRIV_REC_NEXT(3012, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3012(choice, ...)  METALANG99_PRIV_REC_NEXT(3013, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3013(choice, ...)  METALANG99_PRIV_REC_NEXT(3014, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3014(choice, ...)  METALANG99_PRIV_REC_NEXT(3015, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3015(choice, ...)  METALANG99_PRIV_REC_NEXT(3016, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3016(choice, ...)  METALANG99_PRIV_REC_NEXT(3017, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3017(choice, ...)  METALANG99_PRIV_REC_NEXT(3018, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3018(choice, ...)  METALANG99_PRIV_REC_NEXT(3019, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3019(choice, ...)  METALANG99_PRIV_REC_NEXT(3020, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3020(choice, ...)  METALANG99_PRIV_REC_NEXT(3021, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3021(choice, ...)  METALANG99_PRIV_REC_NEXT(3022, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3022(choice, ...)  METALANG99_PRIV_REC_NEXT(3023, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3023(choice, ...)  METALANG99_PRIV_REC_NEXT(3024, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3024(choice, ...)  METALANG99_PRIV_REC_NEXT(3025, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3025(choice, ...)  METALANG99_PRIV_REC_NEXT(3026, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3026(choice, ...)  METALANG99_PRIV_REC_NEXT(3027, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3027(choice, ...)  METALANG99_PRIV_REC_NEXT(3028, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3028(choice, ...)  METALANG99_PRIV_REC_NEXT(3029, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3029(choice, ...)  METALANG99_PRIV_REC_NEXT(3030, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3030(choice, ...)  METALANG99_PRIV_REC_NEXT(3031, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3031(choice, ...)  METALANG99_PRIV_REC_NEXT(3032, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3032(choice, ...)  METALANG99_PRIV_REC_NEXT(3033, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3033(choice, ...)  METALANG99_PRIV_REC_NEXT(3034, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3034(choice, ...)  METALANG99_PRIV_REC_NEXT(3035, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3035(choice, ...)  METALANG99_PRIV_REC_NEXT(3036, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3036(choice, ...)  METALANG99_PRIV_REC_NEXT(3037, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3037(choice, ...)  METALANG99_PRIV_REC_NEXT(3038, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3038(choice, ...)  METALANG99_PRIV_REC_NEXT(3039, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3039(choice, ...)  METALANG99_PRIV_REC_NEXT(3040, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3040(choice, ...)  METALANG99_PRIV_REC_NEXT(3041, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3041(choice, ...)  METALANG99_PRIV_REC_NEXT(3042, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3042(choice, ...)  METALANG99_PRIV_REC_NEXT(3043, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3043(choice, ...)  METALANG99_PRIV_REC_NEXT(3044, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3044(choice, ...)  METALANG99_PRIV_REC_NEXT(3045, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3045(choice, ...)  METALANG99_PRIV_REC_NEXT(3046, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3046(choice, ...)  METALANG99_PRIV_REC_NEXT(3047, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3047(choice, ...)  METALANG99_PRIV_REC_NEXT(3048, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3048(choice, ...)  METALANG99_PRIV_REC_NEXT(3049, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3049(choice, ...)  METALANG99_PRIV_REC_NEXT(3050, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3050(choice, ...)  METALANG99_PRIV_REC_NEXT(3051, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3051(choice, ...)  METALANG99_PRIV_REC_NEXT(3052, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3052(choice, ...)  METALANG99_PRIV_REC_NEXT(3053, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3053(choice, ...)  METALANG99_PRIV_REC_NEXT(3054, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3054(choice, ...)  METALANG99_PRIV_REC_NEXT(3055, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3055(choice, ...)  METALANG99_PRIV_REC_NEXT(3056, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3056(choice, ...)  METALANG99_PRIV_REC_NEXT(3057, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3057(choice, ...)  METALANG99_PRIV_REC_NEXT(3058, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3058(choice, ...)  METALANG99_PRIV_REC_NEXT(3059, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3059(choice, ...)  METALANG99_PRIV_REC_NEXT(3060, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3060(choice, ...)  METALANG99_PRIV_REC_NEXT(3061, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3061(choice, ...)  METALANG99_PRIV_REC_NEXT(3062, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3062(choice, ...)  METALANG99_PRIV_REC_NEXT(3063, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3063(choice, ...)  METALANG99_PRIV_REC_NEXT(3064, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3064(choice, ...)  METALANG99_PRIV_REC_NEXT(3065, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3065(choice, ...)  METALANG99_PRIV_REC_NEXT(3066, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3066(choice, ...)  METALANG99_PRIV_REC_NEXT(3067, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3067(choice, ...)  METALANG99_PRIV_REC_NEXT(3068, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3068(choice, ...)  METALANG99_PRIV_REC_NEXT(3069, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3069(choice, ...)  METALANG99_PRIV_REC_NEXT(3070, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3070(choice, ...)  METALANG99_PRIV_REC_NEXT(3071, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3071(choice, ...)  METALANG99_PRIV_REC_NEXT(3072, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3072(choice, ...)  METALANG99_PRIV_REC_NEXT(3073, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3073(choice, ...)  METALANG99_PRIV_REC_NEXT(3074, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3074(choice, ...)  METALANG99_PRIV_REC_NEXT(3075, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3075(choice, ...)  METALANG99_PRIV_REC_NEXT(3076, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3076(choice, ...)  METALANG99_PRIV_REC_NEXT(3077, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3077(choice, ...)  METALANG99_PRIV_REC_NEXT(3078, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3078(choice, ...)  METALANG99_PRIV_REC_NEXT(3079, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3079(choice, ...)  METALANG99_PRIV_REC_NEXT(3080, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3080(choice, ...)  METALANG99_PRIV_REC_NEXT(3081, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3081(choice, ...)  METALANG99_PRIV_REC_NEXT(3082, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3082(choice, ...)  METALANG99_PRIV_REC_NEXT(3083, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3083(choice, ...)  METALANG99_PRIV_REC_NEXT(3084, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3084(choice, ...)  METALANG99_PRIV_REC_NEXT(3085, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3085(choice, ...)  METALANG99_PRIV_REC_NEXT(3086, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3086(choice, ...)  METALANG99_PRIV_REC_NEXT(3087, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3087(choice, ...)  METALANG99_PRIV_REC_NEXT(3088, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3088(choice, ...)  METALANG99_PRIV_REC_NEXT(3089, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3089(choice, ...)  METALANG99_PRIV_REC_NEXT(3090, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3090(choice, ...)  METALANG99_PRIV_REC_NEXT(3091, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3091(choice, ...)  METALANG99_PRIV_REC_NEXT(3092, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3092(choice, ...)  METALANG99_PRIV_REC_NEXT(3093, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3093(choice, ...)  METALANG99_PRIV_REC_NEXT(3094, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3094(choice, ...)  METALANG99_PRIV_REC_NEXT(3095, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3095(choice, ...)  METALANG99_PRIV_REC_NEXT(3096, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3096(choice, ...)  METALANG99_PRIV_REC_NEXT(3097, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3097(choice, ...)  METALANG99_PRIV_REC_NEXT(3098, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3098(choice, ...)  METALANG99_PRIV_REC_NEXT(3099, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3099(choice, ...)  METALANG99_PRIV_REC_NEXT(3100, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3100(choice, ...)  METALANG99_PRIV_REC_NEXT(3101, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3101(choice, ...)  METALANG99_PRIV_REC_NEXT(3102, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3102(choice, ...)  METALANG99_PRIV_REC_NEXT(3103, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3103(choice, ...)  METALANG99_PRIV_REC_NEXT(3104, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3104(choice, ...)  METALANG99_PRIV_REC_NEXT(3105, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3105(choice, ...)  METALANG99_PRIV_REC_NEXT(3106, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3106(choice, ...)  METALANG99_PRIV_REC_NEXT(3107, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3107(choice, ...)  METALANG99_PRIV_REC_NEXT(3108, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3108(choice, ...)  METALANG99_PRIV_REC_NEXT(3109, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3109(choice, ...)  METALANG99_PRIV_REC_NEXT(3110, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3110(choice, ...)  METALANG99_PRIV_REC_NEXT(3111, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3111(choice, ...)  METALANG99_PRIV_REC_NEXT(3112, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3112(choice, ...)  METALANG99_PRIV_REC_NEXT(3113, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3113(choice, ...)  METALANG99_PRIV_REC_NEXT(3114, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3114(choice, ...)  METALANG99_PRIV_REC_NEXT(3115, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3115(choice, ...)  METALANG99_PRIV_REC_NEXT(3116, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3116(choice, ...)  METALANG99_PRIV_REC_NEXT(3117, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3117(choice, ...)  METALANG99_PRIV_REC_NEXT(3118, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3118(choice, ...)  METALANG99_PRIV_REC_NEXT(3119, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3119(choice, ...)  METALANG99_PRIV_REC_NEXT(3120, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3120(choice, ...)  METALANG99_PRIV_REC_NEXT(3121, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3121(choice, ...)  METALANG99_PRIV_REC_NEXT(3122, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3122(choice, ...)  METALANG99_PRIV_REC_NEXT(3123, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3123(choice, ...)  METALANG99_PRIV_REC_NEXT(3124, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3124(choice, ...)  METALANG99_PRIV_REC_NEXT(3125, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3125(choice, ...)  METALANG99_PRIV_REC_NEXT(3126, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3126(choice, ...)  METALANG99_PRIV_REC_NEXT(3127, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3127(choice, ...)  METALANG99_PRIV_REC_NEXT(3128, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3128(choice, ...)  METALANG99_PRIV_REC_NEXT(3129, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3129(choice, ...)  METALANG99_PRIV_REC_NEXT(3130, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3130(choice, ...)  METALANG99_PRIV_REC_NEXT(3131, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3131(choice, ...)  METALANG99_PRIV_REC_NEXT(3132, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3132(choice, ...)  METALANG99_PRIV_REC_NEXT(3133, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3133(choice, ...)  METALANG99_PRIV_REC_NEXT(3134, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3134(choice, ...)  METALANG99_PRIV_REC_NEXT(3135, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3135(choice, ...)  METALANG99_PRIV_REC_NEXT(3136, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3136(choice, ...)  METALANG99_PRIV_REC_NEXT(3137, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3137(choice, ...)  METALANG99_PRIV_REC_NEXT(3138, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3138(choice, ...)  METALANG99_PRIV_REC_NEXT(3139, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3139(choice, ...)  METALANG99_PRIV_REC_NEXT(3140, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3140(choice, ...)  METALANG99_PRIV_REC_NEXT(3141, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3141(choice, ...)  METALANG99_PRIV_REC_NEXT(3142, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3142(choice, ...)  METALANG99_PRIV_REC_NEXT(3143, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3143(choice, ...)  METALANG99_PRIV_REC_NEXT(3144, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3144(choice, ...)  METALANG99_PRIV_REC_NEXT(3145, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3145(choice, ...)  METALANG99_PRIV_REC_NEXT(3146, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3146(choice, ...)  METALANG99_PRIV_REC_NEXT(3147, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3147(choice, ...)  METALANG99_PRIV_REC_NEXT(3148, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3148(choice, ...)  METALANG99_PRIV_REC_NEXT(3149, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3149(choice, ...)  METALANG99_PRIV_REC_NEXT(3150, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3150(choice, ...)  METALANG99_PRIV_REC_NEXT(3151, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3151(choice, ...)  METALANG99_PRIV_REC_NEXT(3152, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3152(choice, ...)  METALANG99_PRIV_REC_NEXT(3153, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3153(choice, ...)  METALANG99_PRIV_REC_NEXT(3154, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3154(choice, ...)  METALANG99_PRIV_REC_NEXT(3155, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3155(choice, ...)  METALANG99_PRIV_REC_NEXT(3156, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3156(choice, ...)  METALANG99_PRIV_REC_NEXT(3157, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3157(choice, ...)  METALANG99_PRIV_REC_NEXT(3158, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3158(choice, ...)  METALANG99_PRIV_REC_NEXT(3159, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3159(choice, ...)  METALANG99_PRIV_REC_NEXT(3160, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3160(choice, ...)  METALANG99_PRIV_REC_NEXT(3161, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3161(choice, ...)  METALANG99_PRIV_REC_NEXT(3162, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3162(choice, ...)  METALANG99_PRIV_REC_NEXT(3163, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3163(choice, ...)  METALANG99_PRIV_REC_NEXT(3164, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3164(choice, ...)  METALANG99_PRIV_REC_NEXT(3165, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3165(choice, ...)  METALANG99_PRIV_REC_NEXT(3166, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3166(choice, ...)  METALANG99_PRIV_REC_NEXT(3167, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3167(choice, ...)  METALANG99_PRIV_REC_NEXT(3168, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3168(choice, ...)  METALANG99_PRIV_REC_NEXT(3169, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3169(choice, ...)  METALANG99_PRIV_REC_NEXT(3170, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3170(choice, ...)  METALANG99_PRIV_REC_NEXT(3171, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3171(choice, ...)  METALANG99_PRIV_REC_NEXT(3172, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3172(choice, ...)  METALANG99_PRIV_REC_NEXT(3173, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3173(choice, ...)  METALANG99_PRIV_REC_NEXT(3174, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3174(choice, ...)  METALANG99_PRIV_REC_NEXT(3175, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3175(choice, ...)  METALANG99_PRIV_REC_NEXT(3176, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3176(choice, ...)  METALANG99_PRIV_REC_NEXT(3177, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3177(choice, ...)  METALANG99_PRIV_REC_NEXT(3178, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3178(choice, ...)  METALANG99_PRIV_REC_NEXT(3179, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3179(choice, ...)  METALANG99_PRIV_REC_NEXT(3180, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3180(choice, ...)  METALANG99_PRIV_REC_NEXT(3181, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3181(choice, ...)  METALANG99_PRIV_REC_NEXT(3182, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3182(choice, ...)  METALANG99_PRIV_REC_NEXT(3183, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3183(choice, ...)  METALANG99_PRIV_REC_NEXT(3184, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3184(choice, ...)  METALANG99_PRIV_REC_NEXT(3185, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3185(choice, ...)  METALANG99_PRIV_REC_NEXT(3186, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3186(choice, ...)  METALANG99_PRIV_REC_NEXT(3187, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3187(choice, ...)  METALANG99_PRIV_REC_NEXT(3188, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3188(choice, ...)  METALANG99_PRIV_REC_NEXT(3189, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3189(choice, ...)  METALANG99_PRIV_REC_NEXT(3190, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3190(choice, ...)  METALANG99_PRIV_REC_NEXT(3191, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3191(choice, ...)  METALANG99_PRIV_REC_NEXT(3192, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3192(choice, ...)  METALANG99_PRIV_REC_NEXT(3193, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3193(choice, ...)  METALANG99_PRIV_REC_NEXT(3194, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3194(choice, ...)  METALANG99_PRIV_REC_NEXT(3195, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3195(choice, ...)  METALANG99_PRIV_REC_NEXT(3196, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3196(choice, ...)  METALANG99_PRIV_REC_NEXT(3197, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3197(choice, ...)  METALANG99_PRIV_REC_NEXT(3198, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3198(choice, ...)  METALANG99_PRIV_REC_NEXT(3199, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3199(choice, ...)  METALANG99_PRIV_REC_NEXT(3200, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3200(choice, ...)  METALANG99_PRIV_REC_NEXT(3201, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3201(choice, ...)  METALANG99_PRIV_REC_NEXT(3202, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3202(choice, ...)  METALANG99_PRIV_REC_NEXT(3203, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3203(choice, ...)  METALANG99_PRIV_REC_NEXT(3204, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3204(choice, ...)  METALANG99_PRIV_REC_NEXT(3205, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3205(choice, ...)  METALANG99_PRIV_REC_NEXT(3206, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3206(choice, ...)  METALANG99_PRIV_REC_NEXT(3207, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3207(choice, ...)  METALANG99_PRIV_REC_NEXT(3208, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3208(choice, ...)  METALANG99_PRIV_REC_NEXT(3209, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3209(choice, ...)  METALANG99_PRIV_REC_NEXT(3210, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3210(choice, ...)  METALANG99_PRIV_REC_NEXT(3211, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3211(choice, ...)  METALANG99_PRIV_REC_NEXT(3212, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3212(choice, ...)  METALANG99_PRIV_REC_NEXT(3213, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3213(choice, ...)  METALANG99_PRIV_REC_NEXT(3214, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3214(choice, ...)  METALANG99_PRIV_REC_NEXT(3215, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3215(choice, ...)  METALANG99_PRIV_REC_NEXT(3216, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3216(choice, ...)  METALANG99_PRIV_REC_NEXT(3217, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3217(choice, ...)  METALANG99_PRIV_REC_NEXT(3218, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3218(choice, ...)  METALANG99_PRIV_REC_NEXT(3219, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3219(choice, ...)  METALANG99_PRIV_REC_NEXT(3220, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3220(choice, ...)  METALANG99_PRIV_REC_NEXT(3221, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3221(choice, ...)  METALANG99_PRIV_REC_NEXT(3222, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3222(choice, ...)  METALANG99_PRIV_REC_NEXT(3223, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3223(choice, ...)  METALANG99_PRIV_REC_NEXT(3224, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3224(choice, ...)  METALANG99_PRIV_REC_NEXT(3225, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3225(choice, ...)  METALANG99_PRIV_REC_NEXT(3226, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3226(choice, ...)  METALANG99_PRIV_REC_NEXT(3227, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3227(choice, ...)  METALANG99_PRIV_REC_NEXT(3228, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3228(choice, ...)  METALANG99_PRIV_REC_NEXT(3229, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3229(choice, ...)  METALANG99_PRIV_REC_NEXT(3230, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3230(choice, ...)  METALANG99_PRIV_REC_NEXT(3231, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3231(choice, ...)  METALANG99_PRIV_REC_NEXT(3232, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3232(choice, ...)  METALANG99_PRIV_REC_NEXT(3233, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3233(choice, ...)  METALANG99_PRIV_REC_NEXT(3234, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3234(choice, ...)  METALANG99_PRIV_REC_NEXT(3235, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3235(choice, ...)  METALANG99_PRIV_REC_NEXT(3236, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3236(choice, ...)  METALANG99_PRIV_REC_NEXT(3237, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3237(choice, ...)  METALANG99_PRIV_REC_NEXT(3238, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3238(choice, ...)  METALANG99_PRIV_REC_NEXT(3239, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3239(choice, ...)  METALANG99_PRIV_REC_NEXT(3240, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3240(choice, ...)  METALANG99_PRIV_REC_NEXT(3241, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3241(choice, ...)  METALANG99_PRIV_REC_NEXT(3242, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3242(choice, ...)  METALANG99_PRIV_REC_NEXT(3243, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3243(choice, ...)  METALANG99_PRIV_REC_NEXT(3244, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3244(choice, ...)  METALANG99_PRIV_REC_NEXT(3245, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3245(choice, ...)  METALANG99_PRIV_REC_NEXT(3246, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3246(choice, ...)  METALANG99_PRIV_REC_NEXT(3247, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3247(choice, ...)  METALANG99_PRIV_REC_NEXT(3248, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3248(choice, ...)  METALANG99_PRIV_REC_NEXT(3249, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3249(choice, ...)  METALANG99_PRIV_REC_NEXT(3250, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3250(choice, ...)  METALANG99_PRIV_REC_NEXT(3251, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3251(choice, ...)  METALANG99_PRIV_REC_NEXT(3252, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3252(choice, ...)  METALANG99_PRIV_REC_NEXT(3253, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3253(choice, ...)  METALANG99_PRIV_REC_NEXT(3254, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3254(choice, ...)  METALANG99_PRIV_REC_NEXT(3255, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3255(choice, ...)  METALANG99_PRIV_REC_NEXT(3256, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3256(choice, ...)  METALANG99_PRIV_REC_NEXT(3257, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3257(choice, ...)  METALANG99_PRIV_REC_NEXT(3258, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3258(choice, ...)  METALANG99_PRIV_REC_NEXT(3259, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3259(choice, ...)  METALANG99_PRIV_REC_NEXT(3260, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3260(choice, ...)  METALANG99_PRIV_REC_NEXT(3261, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3261(choice, ...)  METALANG99_PRIV_REC_NEXT(3262, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3262(choice, ...)  METALANG99_PRIV_REC_NEXT(3263, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3263(choice, ...)  METALANG99_PRIV_REC_NEXT(3264, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3264(choice, ...)  METALANG99_PRIV_REC_NEXT(3265, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3265(choice, ...)  METALANG99_PRIV_REC_NEXT(3266, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3266(choice, ...)  METALANG99_PRIV_REC_NEXT(3267, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3267(choice, ...)  METALANG99_PRIV_REC_NEXT(3268, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3268(choice, ...)  METALANG99_PRIV_REC_NEXT(3269, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3269(choice, ...)  METALANG99_PRIV_REC_NEXT(3270, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3270(choice, ...)  METALANG99_PRIV_REC_NEXT(3271, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3271(choice, ...)  METALANG99_PRIV_REC_NEXT(3272, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3272(choice, ...)  METALANG99_PRIV_REC_NEXT(3273, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3273(choice, ...)  METALANG99_PRIV_REC_NEXT(3274, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3274(choice, ...)  METALANG99_PRIV_REC_NEXT(3275, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3275(choice, ...)  METALANG99_PRIV_REC_NEXT(3276, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3276(choice, ...)  METALANG99_PRIV_REC_NEXT(3277, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3277(choice, ...)  METALANG99_PRIV_REC_NEXT(3278, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3278(choice, ...)  METALANG99_PRIV_REC_NEXT(3279, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3279(choice, ...)  METALANG99_PRIV_REC_NEXT(3280, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3280(choice, ...)  METALANG99_PRIV_REC_NEXT(3281, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3281(choice, ...)  METALANG99_PRIV_REC_NEXT(3282, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3282(choice, ...)  METALANG99_PRIV_REC_NEXT(3283, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3283(choice, ...)  METALANG99_PRIV_REC_NEXT(3284, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3284(choice, ...)  METALANG99_PRIV_REC_NEXT(3285, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3285(choice, ...)  METALANG99_PRIV_REC_NEXT(3286, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3286(choice, ...)  METALANG99_PRIV_REC_NEXT(3287, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3287(choice, ...)  METALANG99_PRIV_REC_NEXT(3288, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3288(choice, ...)  METALANG99_PRIV_REC_NEXT(3289, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3289(choice, ...)  METALANG99_PRIV_REC_NEXT(3290, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3290(choice, ...)  METALANG99_PRIV_REC_NEXT(3291, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3291(choice, ...)  METALANG99_PRIV_REC_NEXT(3292, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3292(choice, ...)  METALANG99_PRIV_REC_NEXT(3293, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3293(choice, ...)  METALANG99_PRIV_REC_NEXT(3294, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3294(choice, ...)  METALANG99_PRIV_REC_NEXT(3295, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3295(choice, ...)  METALANG99_PRIV_REC_NEXT(3296, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3296(choice, ...)  METALANG99_PRIV_REC_NEXT(3297, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3297(choice, ...)  METALANG99_PRIV_REC_NEXT(3298, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3298(choice, ...)  METALANG99_PRIV_REC_NEXT(3299, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3299(choice, ...)  METALANG99_PRIV_REC_NEXT(3300, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3300(choice, ...)  METALANG99_PRIV_REC_NEXT(3301, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3301(choice, ...)  METALANG99_PRIV_REC_NEXT(3302, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3302(choice, ...)  METALANG99_PRIV_REC_NEXT(3303, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3303(choice, ...)  METALANG99_PRIV_REC_NEXT(3304, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3304(choice, ...)  METALANG99_PRIV_REC_NEXT(3305, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3305(choice, ...)  METALANG99_PRIV_REC_NEXT(3306, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3306(choice, ...)  METALANG99_PRIV_REC_NEXT(3307, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3307(choice, ...)  METALANG99_PRIV_REC_NEXT(3308, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3308(choice, ...)  METALANG99_PRIV_REC_NEXT(3309, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3309(choice, ...)  METALANG99_PRIV_REC_NEXT(3310, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3310(choice, ...)  METALANG99_PRIV_REC_NEXT(3311, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3311(choice, ...)  METALANG99_PRIV_REC_NEXT(3312, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3312(choice, ...)  METALANG99_PRIV_REC_NEXT(3313, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3313(choice, ...)  METALANG99_PRIV_REC_NEXT(3314, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3314(choice, ...)  METALANG99_PRIV_REC_NEXT(3315, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3315(choice, ...)  METALANG99_PRIV_REC_NEXT(3316, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3316(choice, ...)  METALANG99_PRIV_REC_NEXT(3317, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3317(choice, ...)  METALANG99_PRIV_REC_NEXT(3318, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3318(choice, ...)  METALANG99_PRIV_REC_NEXT(3319, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3319(choice, ...)  METALANG99_PRIV_REC_NEXT(3320, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3320(choice, ...)  METALANG99_PRIV_REC_NEXT(3321, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3321(choice, ...)  METALANG99_PRIV_REC_NEXT(3322, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3322(choice, ...)  METALANG99_PRIV_REC_NEXT(3323, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3323(choice, ...)  METALANG99_PRIV_REC_NEXT(3324, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3324(choice, ...)  METALANG99_PRIV_REC_NEXT(3325, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3325(choice, ...)  METALANG99_PRIV_REC_NEXT(3326, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3326(choice, ...)  METALANG99_PRIV_REC_NEXT(3327, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3327(choice, ...)  METALANG99_PRIV_REC_NEXT(3328, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3328(choice, ...)  METALANG99_PRIV_REC_NEXT(3329, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3329(choice, ...)  METALANG99_PRIV_REC_NEXT(3330, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3330(choice, ...)  METALANG99_PRIV_REC_NEXT(3331, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3331(choice, ...)  METALANG99_PRIV_REC_NEXT(3332, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3332(choice, ...)  METALANG99_PRIV_REC_NEXT(3333, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3333(choice, ...)  METALANG99_PRIV_REC_NEXT(3334, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3334(choice, ...)  METALANG99_PRIV_REC_NEXT(3335, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3335(choice, ...)  METALANG99_PRIV_REC_NEXT(3336, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3336(choice, ...)  METALANG99_PRIV_REC_NEXT(3337, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3337(choice, ...)  METALANG99_PRIV_REC_NEXT(3338, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3338(choice, ...)  METALANG99_PRIV_REC_NEXT(3339, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3339(choice, ...)  METALANG99_PRIV_REC_NEXT(3340, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3340(choice, ...)  METALANG99_PRIV_REC_NEXT(3341, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3341(choice, ...)  METALANG99_PRIV_REC_NEXT(3342, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3342(choice, ...)  METALANG99_PRIV_REC_NEXT(3343, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3343(choice, ...)  METALANG99_PRIV_REC_NEXT(3344, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3344(choice, ...)  METALANG99_PRIV_REC_NEXT(3345, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3345(choice, ...)  METALANG99_PRIV_REC_NEXT(3346, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3346(choice, ...)  METALANG99_PRIV_REC_NEXT(3347, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3347(choice, ...)  METALANG99_PRIV_REC_NEXT(3348, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3348(choice, ...)  METALANG99_PRIV_REC_NEXT(3349, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3349(choice, ...)  METALANG99_PRIV_REC_NEXT(3350, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3350(choice, ...)  METALANG99_PRIV_REC_NEXT(3351, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3351(choice, ...)  METALANG99_PRIV_REC_NEXT(3352, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3352(choice, ...)  METALANG99_PRIV_REC_NEXT(3353, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3353(choice, ...)  METALANG99_PRIV_REC_NEXT(3354, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3354(choice, ...)  METALANG99_PRIV_REC_NEXT(3355, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3355(choice, ...)  METALANG99_PRIV_REC_NEXT(3356, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3356(choice, ...)  METALANG99_PRIV_REC_NEXT(3357, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3357(choice, ...)  METALANG99_PRIV_REC_NEXT(3358, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3358(choice, ...)  METALANG99_PRIV_REC_NEXT(3359, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3359(choice, ...)  METALANG99_PRIV_REC_NEXT(3360, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3360(choice, ...)  METALANG99_PRIV_REC_NEXT(3361, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3361(choice, ...)  METALANG99_PRIV_REC_NEXT(3362, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3362(choice, ...)  METALANG99_PRIV_REC_NEXT(3363, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3363(choice, ...)  METALANG99_PRIV_REC_NEXT(3364, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3364(choice, ...)  METALANG99_PRIV_REC_NEXT(3365, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3365(choice, ...)  METALANG99_PRIV_REC_NEXT(3366, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3366(choice, ...)  METALANG99_PRIV_REC_NEXT(3367, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3367(choice, ...)  METALANG99_PRIV_REC_NEXT(3368, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3368(choice, ...)  METALANG99_PRIV_REC_NEXT(3369, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3369(choice, ...)  METALANG99_PRIV_REC_NEXT(3370, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3370(choice, ...)  METALANG99_PRIV_REC_NEXT(3371, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3371(choice, ...)  METALANG99_PRIV_REC_NEXT(3372, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3372(choice, ...)  METALANG99_PRIV_REC_NEXT(3373, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3373(choice, ...)  METALANG99_PRIV_REC_NEXT(3374, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3374(choice, ...)  METALANG99_PRIV_REC_NEXT(3375, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3375(choice, ...)  METALANG99_PRIV_REC_NEXT(3376, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3376(choice, ...)  METALANG99_PRIV_REC_NEXT(3377, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3377(choice, ...)  METALANG99_PRIV_REC_NEXT(3378, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3378(choice, ...)  METALANG99_PRIV_REC_NEXT(3379, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3379(choice, ...)  METALANG99_PRIV_REC_NEXT(3380, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3380(choice, ...)  METALANG99_PRIV_REC_NEXT(3381, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3381(choice, ...)  METALANG99_PRIV_REC_NEXT(3382, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3382(choice, ...)  METALANG99_PRIV_REC_NEXT(3383, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3383(choice, ...)  METALANG99_PRIV_REC_NEXT(3384, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3384(choice, ...)  METALANG99_PRIV_REC_NEXT(3385, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3385(choice, ...)  METALANG99_PRIV_REC_NEXT(3386, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3386(choice, ...)  METALANG99_PRIV_REC_NEXT(3387, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3387(choice, ...)  METALANG99_PRIV_REC_NEXT(3388, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3388(choice, ...)  METALANG99_PRIV_REC_NEXT(3389, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3389(choice, ...)  METALANG99_PRIV_REC_NEXT(3390, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3390(choice, ...)  METALANG99_PRIV_REC_NEXT(3391, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3391(choice, ...)  METALANG99_PRIV_REC_NEXT(3392, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3392(choice, ...)  METALANG99_PRIV_REC_NEXT(3393, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3393(choice, ...)  METALANG99_PRIV_REC_NEXT(3394, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3394(choice, ...)  METALANG99_PRIV_REC_NEXT(3395, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3395(choice, ...)  METALANG99_PRIV_REC_NEXT(3396, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3396(choice, ...)  METALANG99_PRIV_REC_NEXT(3397, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3397(choice, ...)  METALANG99_PRIV_REC_NEXT(3398, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3398(choice, ...)  METALANG99_PRIV_REC_NEXT(3399, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3399(choice, ...)  METALANG99_PRIV_REC_NEXT(3400, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3400(choice, ...)  METALANG99_PRIV_REC_NEXT(3401, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3401(choice, ...)  METALANG99_PRIV_REC_NEXT(3402, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3402(choice, ...)  METALANG99_PRIV_REC_NEXT(3403, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3403(choice, ...)  METALANG99_PRIV_REC_NEXT(3404, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3404(choice, ...)  METALANG99_PRIV_REC_NEXT(3405, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3405(choice, ...)  METALANG99_PRIV_REC_NEXT(3406, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3406(choice, ...)  METALANG99_PRIV_REC_NEXT(3407, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3407(choice, ...)  METALANG99_PRIV_REC_NEXT(3408, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3408(choice, ...)  METALANG99_PRIV_REC_NEXT(3409, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3409(choice, ...)  METALANG99_PRIV_REC_NEXT(3410, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3410(choice, ...)  METALANG99_PRIV_REC_NEXT(3411, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3411(choice, ...)  METALANG99_PRIV_REC_NEXT(3412, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3412(choice, ...)  METALANG99_PRIV_REC_NEXT(3413, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3413(choice, ...)  METALANG99_PRIV_REC_NEXT(3414, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3414(choice, ...)  METALANG99_PRIV_REC_NEXT(3415, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3415(choice, ...)  METALANG99_PRIV_REC_NEXT(3416, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3416(choice, ...)  METALANG99_PRIV_REC_NEXT(3417, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3417(choice, ...)  METALANG99_PRIV_REC_NEXT(3418, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3418(choice, ...)  METALANG99_PRIV_REC_NEXT(3419, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3419(choice, ...)  METALANG99_PRIV_REC_NEXT(3420, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3420(choice, ...)  METALANG99_PRIV_REC_NEXT(3421, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3421(choice, ...)  METALANG99_PRIV_REC_NEXT(3422, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3422(choice, ...)  METALANG99_PRIV_REC_NEXT(3423, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3423(choice, ...)  METALANG99_PRIV_REC_NEXT(3424, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3424(choice, ...)  METALANG99_PRIV_REC_NEXT(3425, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3425(choice, ...)  METALANG99_PRIV_REC_NEXT(3426, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3426(choice, ...)  METALANG99_PRIV_REC_NEXT(3427, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3427(choice, ...)  METALANG99_PRIV_REC_NEXT(3428, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3428(choice, ...)  METALANG99_PRIV_REC_NEXT(3429, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3429(choice, ...)  METALANG99_PRIV_REC_NEXT(3430, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3430(choice, ...)  METALANG99_PRIV_REC_NEXT(3431, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3431(choice, ...)  METALANG99_PRIV_REC_NEXT(3432, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3432(choice, ...)  METALANG99_PRIV_REC_NEXT(3433, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3433(choice, ...)  METALANG99_PRIV_REC_NEXT(3434, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3434(choice, ...)  METALANG99_PRIV_REC_NEXT(3435, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3435(choice, ...)  METALANG99_PRIV_REC_NEXT(3436, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3436(choice, ...)  METALANG99_PRIV_REC_NEXT(3437, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3437(choice, ...)  METALANG99_PRIV_REC_NEXT(3438, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3438(choice, ...)  METALANG99_PRIV_REC_NEXT(3439, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3439(choice, ...)  METALANG99_PRIV_REC_NEXT(3440, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3440(choice, ...)  METALANG99_PRIV_REC_NEXT(3441, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3441(choice, ...)  METALANG99_PRIV_REC_NEXT(3442, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3442(choice, ...)  METALANG99_PRIV_REC_NEXT(3443, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3443(choice, ...)  METALANG99_PRIV_REC_NEXT(3444, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3444(choice, ...)  METALANG99_PRIV_REC_NEXT(3445, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3445(choice, ...)  METALANG99_PRIV_REC_NEXT(3446, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3446(choice, ...)  METALANG99_PRIV_REC_NEXT(3447, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3447(choice, ...)  METALANG99_PRIV_REC_NEXT(3448, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3448(choice, ...)  METALANG99_PRIV_REC_NEXT(3449, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3449(choice, ...)  METALANG99_PRIV_REC_NEXT(3450, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3450(choice, ...)  METALANG99_PRIV_REC_NEXT(3451, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3451(choice, ...)  METALANG99_PRIV_REC_NEXT(3452, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3452(choice, ...)  METALANG99_PRIV_REC_NEXT(3453, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3453(choice, ...)  METALANG99_PRIV_REC_NEXT(3454, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3454(choice, ...)  METALANG99_PRIV_REC_NEXT(3455, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3455(choice, ...)  METALANG99_PRIV_REC_NEXT(3456, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3456(choice, ...)  METALANG99_PRIV_REC_NEXT(3457, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3457(choice, ...)  METALANG99_PRIV_REC_NEXT(3458, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3458(choice, ...)  METALANG99_PRIV_REC_NEXT(3459, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3459(choice, ...)  METALANG99_PRIV_REC_NEXT(3460, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3460(choice, ...)  METALANG99_PRIV_REC_NEXT(3461, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3461(choice, ...)  METALANG99_PRIV_REC_NEXT(3462, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3462(choice, ...)  METALANG99_PRIV_REC_NEXT(3463, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3463(choice, ...)  METALANG99_PRIV_REC_NEXT(3464, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3464(choice, ...)  METALANG99_PRIV_REC_NEXT(3465, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3465(choice, ...)  METALANG99_PRIV_REC_NEXT(3466, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3466(choice, ...)  METALANG99_PRIV_REC_NEXT(3467, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3467(choice, ...)  METALANG99_PRIV_REC_NEXT(3468, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3468(choice, ...)  METALANG99_PRIV_REC_NEXT(3469, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3469(choice, ...)  METALANG99_PRIV_REC_NEXT(3470, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3470(choice, ...)  METALANG99_PRIV_REC_NEXT(3471, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3471(choice, ...)  METALANG99_PRIV_REC_NEXT(3472, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3472(choice, ...)  METALANG99_PRIV_REC_NEXT(3473, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3473(choice, ...)  METALANG99_PRIV_REC_NEXT(3474, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3474(choice, ...)  METALANG99_PRIV_REC_NEXT(3475, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3475(choice, ...)  METALANG99_PRIV_REC_NEXT(3476, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3476(choice, ...)  METALANG99_PRIV_REC_NEXT(3477, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3477(choice, ...)  METALANG99_PRIV_REC_NEXT(3478, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3478(choice, ...)  METALANG99_PRIV_REC_NEXT(3479, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3479(choice, ...)  METALANG99_PRIV_REC_NEXT(3480, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3480(choice, ...)  METALANG99_PRIV_REC_NEXT(3481, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3481(choice, ...)  METALANG99_PRIV_REC_NEXT(3482, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3482(choice, ...)  METALANG99_PRIV_REC_NEXT(3483, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3483(choice, ...)  METALANG99_PRIV_REC_NEXT(3484, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3484(choice, ...)  METALANG99_PRIV_REC_NEXT(3485, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3485(choice, ...)  METALANG99_PRIV_REC_NEXT(3486, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3486(choice, ...)  METALANG99_PRIV_REC_NEXT(3487, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3487(choice, ...)  METALANG99_PRIV_REC_NEXT(3488, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3488(choice, ...)  METALANG99_PRIV_REC_NEXT(3489, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3489(choice, ...)  METALANG99_PRIV_REC_NEXT(3490, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3490(choice, ...)  METALANG99_PRIV_REC_NEXT(3491, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3491(choice, ...)  METALANG99_PRIV_REC_NEXT(3492, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3492(choice, ...)  METALANG99_PRIV_REC_NEXT(3493, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3493(choice, ...)  METALANG99_PRIV_REC_NEXT(3494, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3494(choice, ...)  METALANG99_PRIV_REC_NEXT(3495, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3495(choice, ...)  METALANG99_PRIV_REC_NEXT(3496, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3496(choice, ...)  METALANG99_PRIV_REC_NEXT(3497, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3497(choice, ...)  METALANG99_PRIV_REC_NEXT(3498, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3498(choice, ...)  METALANG99_PRIV_REC_NEXT(3499, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3499(choice, ...)  METALANG99_PRIV_REC_NEXT(3500, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3500(choice, ...)  METALANG99_PRIV_REC_NEXT(3501, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3501(choice, ...)  METALANG99_PRIV_REC_NEXT(3502, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3502(choice, ...)  METALANG99_PRIV_REC_NEXT(3503, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3503(choice, ...)  METALANG99_PRIV_REC_NEXT(3504, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3504(choice, ...)  METALANG99_PRIV_REC_NEXT(3505, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3505(choice, ...)  METALANG99_PRIV_REC_NEXT(3506, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3506(choice, ...)  METALANG99_PRIV_REC_NEXT(3507, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3507(choice, ...)  METALANG99_PRIV_REC_NEXT(3508, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3508(choice, ...)  METALANG99_PRIV_REC_NEXT(3509, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3509(choice, ...)  METALANG99_PRIV_REC_NEXT(3510, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3510(choice, ...)  METALANG99_PRIV_REC_NEXT(3511, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3511(choice, ...)  METALANG99_PRIV_REC_NEXT(3512, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3512(choice, ...)  METALANG99_PRIV_REC_NEXT(3513, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3513(choice, ...)  METALANG99_PRIV_REC_NEXT(3514, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3514(choice, ...)  METALANG99_PRIV_REC_NEXT(3515, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3515(choice, ...)  METALANG99_PRIV_REC_NEXT(3516, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3516(choice, ...)  METALANG99_PRIV_REC_NEXT(3517, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3517(choice, ...)  METALANG99_PRIV_REC_NEXT(3518, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3518(choice, ...)  METALANG99_PRIV_REC_NEXT(3519, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3519(choice, ...)  METALANG99_PRIV_REC_NEXT(3520, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3520(choice, ...)  METALANG99_PRIV_REC_NEXT(3521, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3521(choice, ...)  METALANG99_PRIV_REC_NEXT(3522, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3522(choice, ...)  METALANG99_PRIV_REC_NEXT(3523, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3523(choice, ...)  METALANG99_PRIV_REC_NEXT(3524, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3524(choice, ...)  METALANG99_PRIV_REC_NEXT(3525, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3525(choice, ...)  METALANG99_PRIV_REC_NEXT(3526, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3526(choice, ...)  METALANG99_PRIV_REC_NEXT(3527, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3527(choice, ...)  METALANG99_PRIV_REC_NEXT(3528, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3528(choice, ...)  METALANG99_PRIV_REC_NEXT(3529, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3529(choice, ...)  METALANG99_PRIV_REC_NEXT(3530, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3530(choice, ...)  METALANG99_PRIV_REC_NEXT(3531, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3531(choice, ...)  METALANG99_PRIV_REC_NEXT(3532, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3532(choice, ...)  METALANG99_PRIV_REC_NEXT(3533, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3533(choice, ...)  METALANG99_PRIV_REC_NEXT(3534, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3534(choice, ...)  METALANG99_PRIV_REC_NEXT(3535, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3535(choice, ...)  METALANG99_PRIV_REC_NEXT(3536, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3536(choice, ...)  METALANG99_PRIV_REC_NEXT(3537, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3537(choice, ...)  METALANG99_PRIV_REC_NEXT(3538, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3538(choice, ...)  METALANG99_PRIV_REC_NEXT(3539, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3539(choice, ...)  METALANG99_PRIV_REC_NEXT(3540, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3540(choice, ...)  METALANG99_PRIV_REC_NEXT(3541, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3541(choice, ...)  METALANG99_PRIV_REC_NEXT(3542, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3542(choice, ...)  METALANG99_PRIV_REC_NEXT(3543, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3543(choice, ...)  METALANG99_PRIV_REC_NEXT(3544, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3544(choice, ...)  METALANG99_PRIV_REC_NEXT(3545, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3545(choice, ...)  METALANG99_PRIV_REC_NEXT(3546, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3546(choice, ...)  METALANG99_PRIV_REC_NEXT(3547, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3547(choice, ...)  METALANG99_PRIV_REC_NEXT(3548, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3548(choice, ...)  METALANG99_PRIV_REC_NEXT(3549, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3549(choice, ...)  METALANG99_PRIV_REC_NEXT(3550, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3550(choice, ...)  METALANG99_PRIV_REC_NEXT(3551, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3551(choice, ...)  METALANG99_PRIV_REC_NEXT(3552, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3552(choice, ...)  METALANG99_PRIV_REC_NEXT(3553, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3553(choice, ...)  METALANG99_PRIV_REC_NEXT(3554, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3554(choice, ...)  METALANG99_PRIV_REC_NEXT(3555, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3555(choice, ...)  METALANG99_PRIV_REC_NEXT(3556, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3556(choice, ...)  METALANG99_PRIV_REC_NEXT(3557, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3557(choice, ...)  METALANG99_PRIV_REC_NEXT(3558, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3558(choice, ...)  METALANG99_PRIV_REC_NEXT(3559, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3559(choice, ...)  METALANG99_PRIV_REC_NEXT(3560, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3560(choice, ...)  METALANG99_PRIV_REC_NEXT(3561, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3561(choice, ...)  METALANG99_PRIV_REC_NEXT(3562, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3562(choice, ...)  METALANG99_PRIV_REC_NEXT(3563, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3563(choice, ...)  METALANG99_PRIV_REC_NEXT(3564, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3564(choice, ...)  METALANG99_PRIV_REC_NEXT(3565, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3565(choice, ...)  METALANG99_PRIV_REC_NEXT(3566, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3566(choice, ...)  METALANG99_PRIV_REC_NEXT(3567, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3567(choice, ...)  METALANG99_PRIV_REC_NEXT(3568, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3568(choice, ...)  METALANG99_PRIV_REC_NEXT(3569, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3569(choice, ...)  METALANG99_PRIV_REC_NEXT(3570, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3570(choice, ...)  METALANG99_PRIV_REC_NEXT(3571, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3571(choice, ...)  METALANG99_PRIV_REC_NEXT(3572, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3572(choice, ...)  METALANG99_PRIV_REC_NEXT(3573, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3573(choice, ...)  METALANG99_PRIV_REC_NEXT(3574, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3574(choice, ...)  METALANG99_PRIV_REC_NEXT(3575, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3575(choice, ...)  METALANG99_PRIV_REC_NEXT(3576, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3576(choice, ...)  METALANG99_PRIV_REC_NEXT(3577, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3577(choice, ...)  METALANG99_PRIV_REC_NEXT(3578, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3578(choice, ...)  METALANG99_PRIV_REC_NEXT(3579, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3579(choice, ...)  METALANG99_PRIV_REC_NEXT(3580, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3580(choice, ...)  METALANG99_PRIV_REC_NEXT(3581, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3581(choice, ...)  METALANG99_PRIV_REC_NEXT(3582, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3582(choice, ...)  METALANG99_PRIV_REC_NEXT(3583, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3583(choice, ...)  METALANG99_PRIV_REC_NEXT(3584, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3584(choice, ...)  METALANG99_PRIV_REC_NEXT(3585, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3585(choice, ...)  METALANG99_PRIV_REC_NEXT(3586, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3586(choice, ...)  METALANG99_PRIV_REC_NEXT(3587, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3587(choice, ...)  METALANG99_PRIV_REC_NEXT(3588, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3588(choice, ...)  METALANG99_PRIV_REC_NEXT(3589, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3589(choice, ...)  METALANG99_PRIV_REC_NEXT(3590, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3590(choice, ...)  METALANG99_PRIV_REC_NEXT(3591, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3591(choice, ...)  METALANG99_PRIV_REC_NEXT(3592, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3592(choice, ...)  METALANG99_PRIV_REC_NEXT(3593, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3593(choice, ...)  METALANG99_PRIV_REC_NEXT(3594, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3594(choice, ...)  METALANG99_PRIV_REC_NEXT(3595, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3595(choice, ...)  METALANG99_PRIV_REC_NEXT(3596, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3596(choice, ...)  METALANG99_PRIV_REC_NEXT(3597, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3597(choice, ...)  METALANG99_PRIV_REC_NEXT(3598, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3598(choice, ...)  METALANG99_PRIV_REC_NEXT(3599, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3599(choice, ...)  METALANG99_PRIV_REC_NEXT(3600, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3600(choice, ...)  METALANG99_PRIV_REC_NEXT(3601, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3601(choice, ...)  METALANG99_PRIV_REC_NEXT(3602, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3602(choice, ...)  METALANG99_PRIV_REC_NEXT(3603, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3603(choice, ...)  METALANG99_PRIV_REC_NEXT(3604, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3604(choice, ...)  METALANG99_PRIV_REC_NEXT(3605, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3605(choice, ...)  METALANG99_PRIV_REC_NEXT(3606, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3606(choice, ...)  METALANG99_PRIV_REC_NEXT(3607, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3607(choice, ...)  METALANG99_PRIV_REC_NEXT(3608, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3608(choice, ...)  METALANG99_PRIV_REC_NEXT(3609, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3609(choice, ...)  METALANG99_PRIV_REC_NEXT(3610, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3610(choice, ...)  METALANG99_PRIV_REC_NEXT(3611, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3611(choice, ...)  METALANG99_PRIV_REC_NEXT(3612, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3612(choice, ...)  METALANG99_PRIV_REC_NEXT(3613, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3613(choice, ...)  METALANG99_PRIV_REC_NEXT(3614, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3614(choice, ...)  METALANG99_PRIV_REC_NEXT(3615, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3615(choice, ...)  METALANG99_PRIV_REC_NEXT(3616, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3616(choice, ...)  METALANG99_PRIV_REC_NEXT(3617, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3617(choice, ...)  METALANG99_PRIV_REC_NEXT(3618, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3618(choice, ...)  METALANG99_PRIV_REC_NEXT(3619, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3619(choice, ...)  METALANG99_PRIV_REC_NEXT(3620, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3620(choice, ...)  METALANG99_PRIV_REC_NEXT(3621, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3621(choice, ...)  METALANG99_PRIV_REC_NEXT(3622, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3622(choice, ...)  METALANG99_PRIV_REC_NEXT(3623, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3623(choice, ...)  METALANG99_PRIV_REC_NEXT(3624, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3624(choice, ...)  METALANG99_PRIV_REC_NEXT(3625, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3625(choice, ...)  METALANG99_PRIV_REC_NEXT(3626, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3626(choice, ...)  METALANG99_PRIV_REC_NEXT(3627, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3627(choice, ...)  METALANG99_PRIV_REC_NEXT(3628, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3628(choice, ...)  METALANG99_PRIV_REC_NEXT(3629, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3629(choice, ...)  METALANG99_PRIV_REC_NEXT(3630, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3630(choice, ...)  METALANG99_PRIV_REC_NEXT(3631, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3631(choice, ...)  METALANG99_PRIV_REC_NEXT(3632, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3632(choice, ...)  METALANG99_PRIV_REC_NEXT(3633, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3633(choice, ...)  METALANG99_PRIV_REC_NEXT(3634, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3634(choice, ...)  METALANG99_PRIV_REC_NEXT(3635, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3635(choice, ...)  METALANG99_PRIV_REC_NEXT(3636, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3636(choice, ...)  METALANG99_PRIV_REC_NEXT(3637, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3637(choice, ...)  METALANG99_PRIV_REC_NEXT(3638, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3638(choice, ...)  METALANG99_PRIV_REC_NEXT(3639, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3639(choice, ...)  METALANG99_PRIV_REC_NEXT(3640, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3640(choice, ...)  METALANG99_PRIV_REC_NEXT(3641, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3641(choice, ...)  METALANG99_PRIV_REC_NEXT(3642, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3642(choice, ...)  METALANG99_PRIV_REC_NEXT(3643, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3643(choice, ...)  METALANG99_PRIV_REC_NEXT(3644, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3644(choice, ...)  METALANG99_PRIV_REC_NEXT(3645, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3645(choice, ...)  METALANG99_PRIV_REC_NEXT(3646, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3646(choice, ...)  METALANG99_PRIV_REC_NEXT(3647, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3647(choice, ...)  METALANG99_PRIV_REC_NEXT(3648, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3648(choice, ...)  METALANG99_PRIV_REC_NEXT(3649, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3649(choice, ...)  METALANG99_PRIV_REC_NEXT(3650, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3650(choice, ...)  METALANG99_PRIV_REC_NEXT(3651, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3651(choice, ...)  METALANG99_PRIV_REC_NEXT(3652, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3652(choice, ...)  METALANG99_PRIV_REC_NEXT(3653, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3653(choice, ...)  METALANG99_PRIV_REC_NEXT(3654, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3654(choice, ...)  METALANG99_PRIV_REC_NEXT(3655, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3655(choice, ...)  METALANG99_PRIV_REC_NEXT(3656, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3656(choice, ...)  METALANG99_PRIV_REC_NEXT(3657, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3657(choice, ...)  METALANG99_PRIV_REC_NEXT(3658, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3658(choice, ...)  METALANG99_PRIV_REC_NEXT(3659, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3659(choice, ...)  METALANG99_PRIV_REC_NEXT(3660, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3660(choice, ...)  METALANG99_PRIV_REC_NEXT(3661, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3661(choice, ...)  METALANG99_PRIV_REC_NEXT(3662, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3662(choice, ...)  METALANG99_PRIV_REC_NEXT(3663, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3663(choice, ...)  METALANG99_PRIV_REC_NEXT(3664, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3664(choice, ...)  METALANG99_PRIV_REC_NEXT(3665, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3665(choice, ...)  METALANG99_PRIV_REC_NEXT(3666, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3666(choice, ...)  METALANG99_PRIV_REC_NEXT(3667, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3667(choice, ...)  METALANG99_PRIV_REC_NEXT(3668, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3668(choice, ...)  METALANG99_PRIV_REC_NEXT(3669, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3669(choice, ...)  METALANG99_PRIV_REC_NEXT(3670, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3670(choice, ...)  METALANG99_PRIV_REC_NEXT(3671, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3671(choice, ...)  METALANG99_PRIV_REC_NEXT(3672, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3672(choice, ...)  METALANG99_PRIV_REC_NEXT(3673, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3673(choice, ...)  METALANG99_PRIV_REC_NEXT(3674, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3674(choice, ...)  METALANG99_PRIV_REC_NEXT(3675, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3675(choice, ...)  METALANG99_PRIV_REC_NEXT(3676, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3676(choice, ...)  METALANG99_PRIV_REC_NEXT(3677, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3677(choice, ...)  METALANG99_PRIV_REC_NEXT(3678, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3678(choice, ...)  METALANG99_PRIV_REC_NEXT(3679, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3679(choice, ...)  METALANG99_PRIV_REC_NEXT(3680, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3680(choice, ...)  METALANG99_PRIV_REC_NEXT(3681, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3681(choice, ...)  METALANG99_PRIV_REC_NEXT(3682, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3682(choice, ...)  METALANG99_PRIV_REC_NEXT(3683, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3683(choice, ...)  METALANG99_PRIV_REC_NEXT(3684, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3684(choice, ...)  METALANG99_PRIV_REC_NEXT(3685, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3685(choice, ...)  METALANG99_PRIV_REC_NEXT(3686, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3686(choice, ...)  METALANG99_PRIV_REC_NEXT(3687, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3687(choice, ...)  METALANG99_PRIV_REC_NEXT(3688, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3688(choice, ...)  METALANG99_PRIV_REC_NEXT(3689, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3689(choice, ...)  METALANG99_PRIV_REC_NEXT(3690, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3690(choice, ...)  METALANG99_PRIV_REC_NEXT(3691, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3691(choice, ...)  METALANG99_PRIV_REC_NEXT(3692, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3692(choice, ...)  METALANG99_PRIV_REC_NEXT(3693, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3693(choice, ...)  METALANG99_PRIV_REC_NEXT(3694, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3694(choice, ...)  METALANG99_PRIV_REC_NEXT(3695, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3695(choice, ...)  METALANG99_PRIV_REC_NEXT(3696, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3696(choice, ...)  METALANG99_PRIV_REC_NEXT(3697, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3697(choice, ...)  METALANG99_PRIV_REC_NEXT(3698, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3698(choice, ...)  METALANG99_PRIV_REC_NEXT(3699, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3699(choice, ...)  METALANG99_PRIV_REC_NEXT(3700, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3700(choice, ...)  METALANG99_PRIV_REC_NEXT(3701, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3701(choice, ...)  METALANG99_PRIV_REC_NEXT(3702, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3702(choice, ...)  METALANG99_PRIV_REC_NEXT(3703, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3703(choice, ...)  METALANG99_PRIV_REC_NEXT(3704, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3704(choice, ...)  METALANG99_PRIV_REC_NEXT(3705, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3705(choice, ...)  METALANG99_PRIV_REC_NEXT(3706, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3706(choice, ...)  METALANG99_PRIV_REC_NEXT(3707, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3707(choice, ...)  METALANG99_PRIV_REC_NEXT(3708, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3708(choice, ...)  METALANG99_PRIV_REC_NEXT(3709, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3709(choice, ...)  METALANG99_PRIV_REC_NEXT(3710, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3710(choice, ...)  METALANG99_PRIV_REC_NEXT(3711, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3711(choice, ...)  METALANG99_PRIV_REC_NEXT(3712, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3712(choice, ...)  METALANG99_PRIV_REC_NEXT(3713, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3713(choice, ...)  METALANG99_PRIV_REC_NEXT(3714, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3714(choice, ...)  METALANG99_PRIV_REC_NEXT(3715, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3715(choice, ...)  METALANG99_PRIV_REC_NEXT(3716, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3716(choice, ...)  METALANG99_PRIV_REC_NEXT(3717, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3717(choice, ...)  METALANG99_PRIV_REC_NEXT(3718, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3718(choice, ...)  METALANG99_PRIV_REC_NEXT(3719, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3719(choice, ...)  METALANG99_PRIV_REC_NEXT(3720, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3720(choice, ...)  METALANG99_PRIV_REC_NEXT(3721, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3721(choice, ...)  METALANG99_PRIV_REC_NEXT(3722, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3722(choice, ...)  METALANG99_PRIV_REC_NEXT(3723, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3723(choice, ...)  METALANG99_PRIV_REC_NEXT(3724, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3724(choice, ...)  METALANG99_PRIV_REC_NEXT(3725, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3725(choice, ...)  METALANG99_PRIV_REC_NEXT(3726, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3726(choice, ...)  METALANG99_PRIV_REC_NEXT(3727, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3727(choice, ...)  METALANG99_PRIV_REC_NEXT(3728, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3728(choice, ...)  METALANG99_PRIV_REC_NEXT(3729, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3729(choice, ...)  METALANG99_PRIV_REC_NEXT(3730, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3730(choice, ...)  METALANG99_PRIV_REC_NEXT(3731, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3731(choice, ...)  METALANG99_PRIV_REC_NEXT(3732, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3732(choice, ...)  METALANG99_PRIV_REC_NEXT(3733, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3733(choice, ...)  METALANG99_PRIV_REC_NEXT(3734, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3734(choice, ...)  METALANG99_PRIV_REC_NEXT(3735, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3735(choice, ...)  METALANG99_PRIV_REC_NEXT(3736, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3736(choice, ...)  METALANG99_PRIV_REC_NEXT(3737, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3737(choice, ...)  METALANG99_PRIV_REC_NEXT(3738, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3738(choice, ...)  METALANG99_PRIV_REC_NEXT(3739, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3739(choice, ...)  METALANG99_PRIV_REC_NEXT(3740, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3740(choice, ...)  METALANG99_PRIV_REC_NEXT(3741, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3741(choice, ...)  METALANG99_PRIV_REC_NEXT(3742, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3742(choice, ...)  METALANG99_PRIV_REC_NEXT(3743, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3743(choice, ...)  METALANG99_PRIV_REC_NEXT(3744, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3744(choice, ...)  METALANG99_PRIV_REC_NEXT(3745, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3745(choice, ...)  METALANG99_PRIV_REC_NEXT(3746, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3746(choice, ...)  METALANG99_PRIV_REC_NEXT(3747, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3747(choice, ...)  METALANG99_PRIV_REC_NEXT(3748, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3748(choice, ...)  METALANG99_PRIV_REC_NEXT(3749, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3749(choice, ...)  METALANG99_PRIV_REC_NEXT(3750, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3750(choice, ...)  METALANG99_PRIV_REC_NEXT(3751, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3751(choice, ...)  METALANG99_PRIV_REC_NEXT(3752, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3752(choice, ...)  METALANG99_PRIV_REC_NEXT(3753, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3753(choice, ...)  METALANG99_PRIV_REC_NEXT(3754, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3754(choice, ...)  METALANG99_PRIV_REC_NEXT(3755, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3755(choice, ...)  METALANG99_PRIV_REC_NEXT(3756, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3756(choice, ...)  METALANG99_PRIV_REC_NEXT(3757, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3757(choice, ...)  METALANG99_PRIV_REC_NEXT(3758, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3758(choice, ...)  METALANG99_PRIV_REC_NEXT(3759, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3759(choice, ...)  METALANG99_PRIV_REC_NEXT(3760, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3760(choice, ...)  METALANG99_PRIV_REC_NEXT(3761, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3761(choice, ...)  METALANG99_PRIV_REC_NEXT(3762, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3762(choice, ...)  METALANG99_PRIV_REC_NEXT(3763, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3763(choice, ...)  METALANG99_PRIV_REC_NEXT(3764, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3764(choice, ...)  METALANG99_PRIV_REC_NEXT(3765, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3765(choice, ...)  METALANG99_PRIV_REC_NEXT(3766, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3766(choice, ...)  METALANG99_PRIV_REC_NEXT(3767, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3767(choice, ...)  METALANG99_PRIV_REC_NEXT(3768, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3768(choice, ...)  METALANG99_PRIV_REC_NEXT(3769, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3769(choice, ...)  METALANG99_PRIV_REC_NEXT(3770, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3770(choice, ...)  METALANG99_PRIV_REC_NEXT(3771, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3771(choice, ...)  METALANG99_PRIV_REC_NEXT(3772, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3772(choice, ...)  METALANG99_PRIV_REC_NEXT(3773, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3773(choice, ...)  METALANG99_PRIV_REC_NEXT(3774, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3774(choice, ...)  METALANG99_PRIV_REC_NEXT(3775, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3775(choice, ...)  METALANG99_PRIV_REC_NEXT(3776, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3776(choice, ...)  METALANG99_PRIV_REC_NEXT(3777, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3777(choice, ...)  METALANG99_PRIV_REC_NEXT(3778, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3778(choice, ...)  METALANG99_PRIV_REC_NEXT(3779, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3779(choice, ...)  METALANG99_PRIV_REC_NEXT(3780, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3780(choice, ...)  METALANG99_PRIV_REC_NEXT(3781, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3781(choice, ...)  METALANG99_PRIV_REC_NEXT(3782, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3782(choice, ...)  METALANG99_PRIV_REC_NEXT(3783, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3783(choice, ...)  METALANG99_PRIV_REC_NEXT(3784, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3784(choice, ...)  METALANG99_PRIV_REC_NEXT(3785, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3785(choice, ...)  METALANG99_PRIV_REC_NEXT(3786, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3786(choice, ...)  METALANG99_PRIV_REC_NEXT(3787, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3787(choice, ...)  METALANG99_PRIV_REC_NEXT(3788, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3788(choice, ...)  METALANG99_PRIV_REC_NEXT(3789, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3789(choice, ...)  METALANG99_PRIV_REC_NEXT(3790, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3790(choice, ...)  METALANG99_PRIV_REC_NEXT(3791, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3791(choice, ...)  METALANG99_PRIV_REC_NEXT(3792, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3792(choice, ...)  METALANG99_PRIV_REC_NEXT(3793, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3793(choice, ...)  METALANG99_PRIV_REC_NEXT(3794, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3794(choice, ...)  METALANG99_PRIV_REC_NEXT(3795, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3795(choice, ...)  METALANG99_PRIV_REC_NEXT(3796, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3796(choice, ...)  METALANG99_PRIV_REC_NEXT(3797, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3797(choice, ...)  METALANG99_PRIV_REC_NEXT(3798, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3798(choice, ...)  METALANG99_PRIV_REC_NEXT(3799, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3799(choice, ...)  METALANG99_PRIV_REC_NEXT(3800, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3800(choice, ...)  METALANG99_PRIV_REC_NEXT(3801, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3801(choice, ...)  METALANG99_PRIV_REC_NEXT(3802, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3802(choice, ...)  METALANG99_PRIV_REC_NEXT(3803, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3803(choice, ...)  METALANG99_PRIV_REC_NEXT(3804, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3804(choice, ...)  METALANG99_PRIV_REC_NEXT(3805, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3805(choice, ...)  METALANG99_PRIV_REC_NEXT(3806, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3806(choice, ...)  METALANG99_PRIV_REC_NEXT(3807, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3807(choice, ...)  METALANG99_PRIV_REC_NEXT(3808, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3808(choice, ...)  METALANG99_PRIV_REC_NEXT(3809, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3809(choice, ...)  METALANG99_PRIV_REC_NEXT(3810, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3810(choice, ...)  METALANG99_PRIV_REC_NEXT(3811, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3811(choice, ...)  METALANG99_PRIV_REC_NEXT(3812, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3812(choice, ...)  METALANG99_PRIV_REC_NEXT(3813, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3813(choice, ...)  METALANG99_PRIV_REC_NEXT(3814, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3814(choice, ...)  METALANG99_PRIV_REC_NEXT(3815, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3815(choice, ...)  METALANG99_PRIV_REC_NEXT(3816, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3816(choice, ...)  METALANG99_PRIV_REC_NEXT(3817, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3817(choice, ...)  METALANG99_PRIV_REC_NEXT(3818, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3818(choice, ...)  METALANG99_PRIV_REC_NEXT(3819, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3819(choice, ...)  METALANG99_PRIV_REC_NEXT(3820, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3820(choice, ...)  METALANG99_PRIV_REC_NEXT(3821, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3821(choice, ...)  METALANG99_PRIV_REC_NEXT(3822, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3822(choice, ...)  METALANG99_PRIV_REC_NEXT(3823, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3823(choice, ...)  METALANG99_PRIV_REC_NEXT(3824, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3824(choice, ...)  METALANG99_PRIV_REC_NEXT(3825, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3825(choice, ...)  METALANG99_PRIV_REC_NEXT(3826, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3826(choice, ...)  METALANG99_PRIV_REC_NEXT(3827, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3827(choice, ...)  METALANG99_PRIV_REC_NEXT(3828, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3828(choice, ...)  METALANG99_PRIV_REC_NEXT(3829, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3829(choice, ...)  METALANG99_PRIV_REC_NEXT(3830, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3830(choice, ...)  METALANG99_PRIV_REC_NEXT(3831, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3831(choice, ...)  METALANG99_PRIV_REC_NEXT(3832, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3832(choice, ...)  METALANG99_PRIV_REC_NEXT(3833, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3833(choice, ...)  METALANG99_PRIV_REC_NEXT(3834, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3834(choice, ...)  METALANG99_PRIV_REC_NEXT(3835, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3835(choice, ...)  METALANG99_PRIV_REC_NEXT(3836, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3836(choice, ...)  METALANG99_PRIV_REC_NEXT(3837, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3837(choice, ...)  METALANG99_PRIV_REC_NEXT(3838, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3838(choice, ...)  METALANG99_PRIV_REC_NEXT(3839, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3839(choice, ...)  METALANG99_PRIV_REC_NEXT(3840, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3840(choice, ...)  METALANG99_PRIV_REC_NEXT(3841, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3841(choice, ...)  METALANG99_PRIV_REC_NEXT(3842, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3842(choice, ...)  METALANG99_PRIV_REC_NEXT(3843, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3843(choice, ...)  METALANG99_PRIV_REC_NEXT(3844, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3844(choice, ...)  METALANG99_PRIV_REC_NEXT(3845, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3845(choice, ...)  METALANG99_PRIV_REC_NEXT(3846, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3846(choice, ...)  METALANG99_PRIV_REC_NEXT(3847, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3847(choice, ...)  METALANG99_PRIV_REC_NEXT(3848, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3848(choice, ...)  METALANG99_PRIV_REC_NEXT(3849, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3849(choice, ...)  METALANG99_PRIV_REC_NEXT(3850, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3850(choice, ...)  METALANG99_PRIV_REC_NEXT(3851, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3851(choice, ...)  METALANG99_PRIV_REC_NEXT(3852, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3852(choice, ...)  METALANG99_PRIV_REC_NEXT(3853, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3853(choice, ...)  METALANG99_PRIV_REC_NEXT(3854, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3854(choice, ...)  METALANG99_PRIV_REC_NEXT(3855, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3855(choice, ...)  METALANG99_PRIV_REC_NEXT(3856, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3856(choice, ...)  METALANG99_PRIV_REC_NEXT(3857, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3857(choice, ...)  METALANG99_PRIV_REC_NEXT(3858, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3858(choice, ...)  METALANG99_PRIV_REC_NEXT(3859, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3859(choice, ...)  METALANG99_PRIV_REC_NEXT(3860, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3860(choice, ...)  METALANG99_PRIV_REC_NEXT(3861, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3861(choice, ...)  METALANG99_PRIV_REC_NEXT(3862, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3862(choice, ...)  METALANG99_PRIV_REC_NEXT(3863, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3863(choice, ...)  METALANG99_PRIV_REC_NEXT(3864, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3864(choice, ...)  METALANG99_PRIV_REC_NEXT(3865, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3865(choice, ...)  METALANG99_PRIV_REC_NEXT(3866, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3866(choice, ...)  METALANG99_PRIV_REC_NEXT(3867, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3867(choice, ...)  METALANG99_PRIV_REC_NEXT(3868, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3868(choice, ...)  METALANG99_PRIV_REC_NEXT(3869, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3869(choice, ...)  METALANG99_PRIV_REC_NEXT(3870, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3870(choice, ...)  METALANG99_PRIV_REC_NEXT(3871, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3871(choice, ...)  METALANG99_PRIV_REC_NEXT(3872, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3872(choice, ...)  METALANG99_PRIV_REC_NEXT(3873, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3873(choice, ...)  METALANG99_PRIV_REC_NEXT(3874, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3874(choice, ...)  METALANG99_PRIV_REC_NEXT(3875, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3875(choice, ...)  METALANG99_PRIV_REC_NEXT(3876, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3876(choice, ...)  METALANG99_PRIV_REC_NEXT(3877, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3877(choice, ...)  METALANG99_PRIV_REC_NEXT(3878, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3878(choice, ...)  METALANG99_PRIV_REC_NEXT(3879, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3879(choice, ...)  METALANG99_PRIV_REC_NEXT(3880, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3880(choice, ...)  METALANG99_PRIV_REC_NEXT(3881, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3881(choice, ...)  METALANG99_PRIV_REC_NEXT(3882, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3882(choice, ...)  METALANG99_PRIV_REC_NEXT(3883, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3883(choice, ...)  METALANG99_PRIV_REC_NEXT(3884, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3884(choice, ...)  METALANG99_PRIV_REC_NEXT(3885, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3885(choice, ...)  METALANG99_PRIV_REC_NEXT(3886, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3886(choice, ...)  METALANG99_PRIV_REC_NEXT(3887, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3887(choice, ...)  METALANG99_PRIV_REC_NEXT(3888, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3888(choice, ...)  METALANG99_PRIV_REC_NEXT(3889, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3889(choice, ...)  METALANG99_PRIV_REC_NEXT(3890, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3890(choice, ...)  METALANG99_PRIV_REC_NEXT(3891, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3891(choice, ...)  METALANG99_PRIV_REC_NEXT(3892, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3892(choice, ...)  METALANG99_PRIV_REC_NEXT(3893, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3893(choice, ...)  METALANG99_PRIV_REC_NEXT(3894, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3894(choice, ...)  METALANG99_PRIV_REC_NEXT(3895, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3895(choice, ...)  METALANG99_PRIV_REC_NEXT(3896, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3896(choice, ...)  METALANG99_PRIV_REC_NEXT(3897, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3897(choice, ...)  METALANG99_PRIV_REC_NEXT(3898, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3898(choice, ...)  METALANG99_PRIV_REC_NEXT(3899, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3899(choice, ...)  METALANG99_PRIV_REC_NEXT(3900, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3900(choice, ...)  METALANG99_PRIV_REC_NEXT(3901, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3901(choice, ...)  METALANG99_PRIV_REC_NEXT(3902, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3902(choice, ...)  METALANG99_PRIV_REC_NEXT(3903, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3903(choice, ...)  METALANG99_PRIV_REC_NEXT(3904, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3904(choice, ...)  METALANG99_PRIV_REC_NEXT(3905, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3905(choice, ...)  METALANG99_PRIV_REC_NEXT(3906, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3906(choice, ...)  METALANG99_PRIV_REC_NEXT(3907, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3907(choice, ...)  METALANG99_PRIV_REC_NEXT(3908, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3908(choice, ...)  METALANG99_PRIV_REC_NEXT(3909, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3909(choice, ...)  METALANG99_PRIV_REC_NEXT(3910, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3910(choice, ...)  METALANG99_PRIV_REC_NEXT(3911, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3911(choice, ...)  METALANG99_PRIV_REC_NEXT(3912, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3912(choice, ...)  METALANG99_PRIV_REC_NEXT(3913, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3913(choice, ...)  METALANG99_PRIV_REC_NEXT(3914, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3914(choice, ...)  METALANG99_PRIV_REC_NEXT(3915, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3915(choice, ...)  METALANG99_PRIV_REC_NEXT(3916, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3916(choice, ...)  METALANG99_PRIV_REC_NEXT(3917, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3917(choice, ...)  METALANG99_PRIV_REC_NEXT(3918, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3918(choice, ...)  METALANG99_PRIV_REC_NEXT(3919, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3919(choice, ...)  METALANG99_PRIV_REC_NEXT(3920, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3920(choice, ...)  METALANG99_PRIV_REC_NEXT(3921, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3921(choice, ...)  METALANG99_PRIV_REC_NEXT(3922, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3922(choice, ...)  METALANG99_PRIV_REC_NEXT(3923, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3923(choice, ...)  METALANG99_PRIV_REC_NEXT(3924, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3924(choice, ...)  METALANG99_PRIV_REC_NEXT(3925, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3925(choice, ...)  METALANG99_PRIV_REC_NEXT(3926, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3926(choice, ...)  METALANG99_PRIV_REC_NEXT(3927, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3927(choice, ...)  METALANG99_PRIV_REC_NEXT(3928, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3928(choice, ...)  METALANG99_PRIV_REC_NEXT(3929, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3929(choice, ...)  METALANG99_PRIV_REC_NEXT(3930, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3930(choice, ...)  METALANG99_PRIV_REC_NEXT(3931, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3931(choice, ...)  METALANG99_PRIV_REC_NEXT(3932, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3932(choice, ...)  METALANG99_PRIV_REC_NEXT(3933, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3933(choice, ...)  METALANG99_PRIV_REC_NEXT(3934, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3934(choice, ...)  METALANG99_PRIV_REC_NEXT(3935, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3935(choice, ...)  METALANG99_PRIV_REC_NEXT(3936, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3936(choice, ...)  METALANG99_PRIV_REC_NEXT(3937, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3937(choice, ...)  METALANG99_PRIV_REC_NEXT(3938, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3938(choice, ...)  METALANG99_PRIV_REC_NEXT(3939, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3939(choice, ...)  METALANG99_PRIV_REC_NEXT(3940, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3940(choice, ...)  METALANG99_PRIV_REC_NEXT(3941, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3941(choice, ...)  METALANG99_PRIV_REC_NEXT(3942, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3942(choice, ...)  METALANG99_PRIV_REC_NEXT(3943, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3943(choice, ...)  METALANG99_PRIV_REC_NEXT(3944, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3944(choice, ...)  METALANG99_PRIV_REC_NEXT(3945, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3945(choice, ...)  METALANG99_PRIV_REC_NEXT(3946, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3946(choice, ...)  METALANG99_PRIV_REC_NEXT(3947, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3947(choice, ...)  METALANG99_PRIV_REC_NEXT(3948, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3948(choice, ...)  METALANG99_PRIV_REC_NEXT(3949, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3949(choice, ...)  METALANG99_PRIV_REC_NEXT(3950, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3950(choice, ...)  METALANG99_PRIV_REC_NEXT(3951, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3951(choice, ...)  METALANG99_PRIV_REC_NEXT(3952, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3952(choice, ...)  METALANG99_PRIV_REC_NEXT(3953, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3953(choice, ...)  METALANG99_PRIV_REC_NEXT(3954, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3954(choice, ...)  METALANG99_PRIV_REC_NEXT(3955, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3955(choice, ...)  METALANG99_PRIV_REC_NEXT(3956, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3956(choice, ...)  METALANG99_PRIV_REC_NEXT(3957, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3957(choice, ...)  METALANG99_PRIV_REC_NEXT(3958, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3958(choice, ...)  METALANG99_PRIV_REC_NEXT(3959, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3959(choice, ...)  METALANG99_PRIV_REC_NEXT(3960, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3960(choice, ...)  METALANG99_PRIV_REC_NEXT(3961, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3961(choice, ...)  METALANG99_PRIV_REC_NEXT(3962, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3962(choice, ...)  METALANG99_PRIV_REC_NEXT(3963, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3963(choice, ...)  METALANG99_PRIV_REC_NEXT(3964, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3964(choice, ...)  METALANG99_PRIV_REC_NEXT(3965, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3965(choice, ...)  METALANG99_PRIV_REC_NEXT(3966, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3966(choice, ...)  METALANG99_PRIV_REC_NEXT(3967, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3967(choice, ...)  METALANG99_PRIV_REC_NEXT(3968, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3968(choice, ...)  METALANG99_PRIV_REC_NEXT(3969, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3969(choice, ...)  METALANG99_PRIV_REC_NEXT(3970, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3970(choice, ...)  METALANG99_PRIV_REC_NEXT(3971, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3971(choice, ...)  METALANG99_PRIV_REC_NEXT(3972, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3972(choice, ...)  METALANG99_PRIV_REC_NEXT(3973, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3973(choice, ...)  METALANG99_PRIV_REC_NEXT(3974, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3974(choice, ...)  METALANG99_PRIV_REC_NEXT(3975, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3975(choice, ...)  METALANG99_PRIV_REC_NEXT(3976, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3976(choice, ...)  METALANG99_PRIV_REC_NEXT(3977, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3977(choice, ...)  METALANG99_PRIV_REC_NEXT(3978, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3978(choice, ...)  METALANG99_PRIV_REC_NEXT(3979, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3979(choice, ...)  METALANG99_PRIV_REC_NEXT(3980, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3980(choice, ...)  METALANG99_PRIV_REC_NEXT(3981, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3981(choice, ...)  METALANG99_PRIV_REC_NEXT(3982, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3982(choice, ...)  METALANG99_PRIV_REC_NEXT(3983, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3983(choice, ...)  METALANG99_PRIV_REC_NEXT(3984, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3984(choice, ...)  METALANG99_PRIV_REC_NEXT(3985, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3985(choice, ...)  METALANG99_PRIV_REC_NEXT(3986, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3986(choice, ...)  METALANG99_PRIV_REC_NEXT(3987, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3987(choice, ...)  METALANG99_PRIV_REC_NEXT(3988, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3988(choice, ...)  METALANG99_PRIV_REC_NEXT(3989, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3989(choice, ...)  METALANG99_PRIV_REC_NEXT(3990, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3990(choice, ...)  METALANG99_PRIV_REC_NEXT(3991, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3991(choice, ...)  METALANG99_PRIV_REC_NEXT(3992, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3992(choice, ...)  METALANG99_PRIV_REC_NEXT(3993, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3993(choice, ...)  METALANG99_PRIV_REC_NEXT(3994, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3994(choice, ...)  METALANG99_PRIV_REC_NEXT(3995, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3995(choice, ...)  METALANG99_PRIV_REC_NEXT(3996, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3996(choice, ...)  METALANG99_PRIV_REC_NEXT(3997, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3997(choice, ...)  METALANG99_PRIV_REC_NEXT(3998, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3998(choice, ...)  METALANG99_PRIV_REC_NEXT(3999, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_3999(choice, ...)  METALANG99_PRIV_REC_NEXT(4000, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4000(choice, ...)  METALANG99_PRIV_REC_NEXT(4001, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4001(choice, ...)  METALANG99_PRIV_REC_NEXT(4002, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4002(choice, ...)  METALANG99_PRIV_REC_NEXT(4003, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4003(choice, ...)  METALANG99_PRIV_REC_NEXT(4004, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4004(choice, ...)  METALANG99_PRIV_REC_NEXT(4005, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4005(choice, ...)  METALANG99_PRIV_REC_NEXT(4006, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4006(choice, ...)  METALANG99_PRIV_REC_NEXT(4007, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4007(choice, ...)  METALANG99_PRIV_REC_NEXT(4008, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4008(choice, ...)  METALANG99_PRIV_REC_NEXT(4009, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4009(choice, ...)  METALANG99_PRIV_REC_NEXT(4010, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4010(choice, ...)  METALANG99_PRIV_REC_NEXT(4011, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4011(choice, ...)  METALANG99_PRIV_REC_NEXT(4012, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4012(choice, ...)  METALANG99_PRIV_REC_NEXT(4013, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4013(choice, ...)  METALANG99_PRIV_REC_NEXT(4014, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4014(choice, ...)  METALANG99_PRIV_REC_NEXT(4015, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4015(choice, ...)  METALANG99_PRIV_REC_NEXT(4016, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4016(choice, ...)  METALANG99_PRIV_REC_NEXT(4017, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4017(choice, ...)  METALANG99_PRIV_REC_NEXT(4018, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4018(choice, ...)  METALANG99_PRIV_REC_NEXT(4019, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4019(choice, ...)  METALANG99_PRIV_REC_NEXT(4020, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4020(choice, ...)  METALANG99_PRIV_REC_NEXT(4021, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4021(choice, ...)  METALANG99_PRIV_REC_NEXT(4022, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4022(choice, ...)  METALANG99_PRIV_REC_NEXT(4023, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4023(choice, ...)  METALANG99_PRIV_REC_NEXT(4024, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4024(choice, ...)  METALANG99_PRIV_REC_NEXT(4025, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4025(choice, ...)  METALANG99_PRIV_REC_NEXT(4026, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4026(choice, ...)  METALANG99_PRIV_REC_NEXT(4027, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4027(choice, ...)  METALANG99_PRIV_REC_NEXT(4028, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4028(choice, ...)  METALANG99_PRIV_REC_NEXT(4029, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4029(choice, ...)  METALANG99_PRIV_REC_NEXT(4030, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4030(choice, ...)  METALANG99_PRIV_REC_NEXT(4031, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4031(choice, ...)  METALANG99_PRIV_REC_NEXT(4032, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4032(choice, ...)  METALANG99_PRIV_REC_NEXT(4033, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4033(choice, ...)  METALANG99_PRIV_REC_NEXT(4034, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4034(choice, ...)  METALANG99_PRIV_REC_NEXT(4035, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4035(choice, ...)  METALANG99_PRIV_REC_NEXT(4036, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4036(choice, ...)  METALANG99_PRIV_REC_NEXT(4037, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4037(choice, ...)  METALANG99_PRIV_REC_NEXT(4038, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4038(choice, ...)  METALANG99_PRIV_REC_NEXT(4039, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4039(choice, ...)  METALANG99_PRIV_REC_NEXT(4040, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4040(choice, ...)  METALANG99_PRIV_REC_NEXT(4041, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4041(choice, ...)  METALANG99_PRIV_REC_NEXT(4042, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4042(choice, ...)  METALANG99_PRIV_REC_NEXT(4043, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4043(choice, ...)  METALANG99_PRIV_REC_NEXT(4044, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4044(choice, ...)  METALANG99_PRIV_REC_NEXT(4045, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4045(choice, ...)  METALANG99_PRIV_REC_NEXT(4046, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4046(choice, ...)  METALANG99_PRIV_REC_NEXT(4047, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4047(choice, ...)  METALANG99_PRIV_REC_NEXT(4048, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4048(choice, ...)  METALANG99_PRIV_REC_NEXT(4049, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4049(choice, ...)  METALANG99_PRIV_REC_NEXT(4050, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4050(choice, ...)  METALANG99_PRIV_REC_NEXT(4051, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4051(choice, ...)  METALANG99_PRIV_REC_NEXT(4052, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4052(choice, ...)  METALANG99_PRIV_REC_NEXT(4053, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4053(choice, ...)  METALANG99_PRIV_REC_NEXT(4054, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4054(choice, ...)  METALANG99_PRIV_REC_NEXT(4055, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4055(choice, ...)  METALANG99_PRIV_REC_NEXT(4056, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4056(choice, ...)  METALANG99_PRIV_REC_NEXT(4057, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4057(choice, ...)  METALANG99_PRIV_REC_NEXT(4058, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4058(choice, ...)  METALANG99_PRIV_REC_NEXT(4059, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4059(choice, ...)  METALANG99_PRIV_REC_NEXT(4060, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4060(choice, ...)  METALANG99_PRIV_REC_NEXT(4061, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4061(choice, ...)  METALANG99_PRIV_REC_NEXT(4062, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4062(choice, ...)  METALANG99_PRIV_REC_NEXT(4063, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4063(choice, ...)  METALANG99_PRIV_REC_NEXT(4064, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4064(choice, ...)  METALANG99_PRIV_REC_NEXT(4065, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4065(choice, ...)  METALANG99_PRIV_REC_NEXT(4066, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4066(choice, ...)  METALANG99_PRIV_REC_NEXT(4067, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4067(choice, ...)  METALANG99_PRIV_REC_NEXT(4068, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4068(choice, ...)  METALANG99_PRIV_REC_NEXT(4069, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4069(choice, ...)  METALANG99_PRIV_REC_NEXT(4070, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4070(choice, ...)  METALANG99_PRIV_REC_NEXT(4071, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4071(choice, ...)  METALANG99_PRIV_REC_NEXT(4072, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4072(choice, ...)  METALANG99_PRIV_REC_NEXT(4073, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4073(choice, ...)  METALANG99_PRIV_REC_NEXT(4074, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4074(choice, ...)  METALANG99_PRIV_REC_NEXT(4075, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4075(choice, ...)  METALANG99_PRIV_REC_NEXT(4076, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4076(choice, ...)  METALANG99_PRIV_REC_NEXT(4077, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4077(choice, ...)  METALANG99_PRIV_REC_NEXT(4078, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4078(choice, ...)  METALANG99_PRIV_REC_NEXT(4079, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4079(choice, ...)  METALANG99_PRIV_REC_NEXT(4080, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4080(choice, ...)  METALANG99_PRIV_REC_NEXT(4081, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4081(choice, ...)  METALANG99_PRIV_REC_NEXT(4082, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4082(choice, ...)  METALANG99_PRIV_REC_NEXT(4083, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4083(choice, ...)  METALANG99_PRIV_REC_NEXT(4084, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4084(choice, ...)  METALANG99_PRIV_REC_NEXT(4085, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4085(choice, ...)  METALANG99_PRIV_REC_NEXT(4086, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4086(choice, ...)  METALANG99_PRIV_REC_NEXT(4087, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4087(choice, ...)  METALANG99_PRIV_REC_NEXT(4088, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4088(choice, ...)  METALANG99_PRIV_REC_NEXT(4089, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4089(choice, ...)  METALANG99_PRIV_REC_NEXT(4090, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4090(choice, ...)  METALANG99_PRIV_REC_NEXT(4091, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4091(choice, ...)  METALANG99_PRIV_REC_NEXT(4092, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4092(choice, ...)  METALANG99_PRIV_REC_NEXT(4093, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4093(choice, ...)  METALANG99_PRIV_REC_NEXT(4094, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4094(choice, ...)  METALANG99_PRIV_REC_NEXT(4095, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4095(choice, ...)  METALANG99_PRIV_REC_NEXT(4096, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4096(choice, ...)  METALANG99_PRIV_REC_NEXT(4097, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4097(choice, ...)  METALANG99_PRIV_REC_NEXT(4098, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4098(choice, ...)  METALANG99_PRIV_REC_NEXT(4099, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4099(choice, ...)  METALANG99_PRIV_REC_NEXT(4100, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4100(choice, ...)  METALANG99_PRIV_REC_NEXT(4101, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4101(choice, ...)  METALANG99_PRIV_REC_NEXT(4102, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4102(choice, ...)  METALANG99_PRIV_REC_NEXT(4103, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4103(choice, ...)  METALANG99_PRIV_REC_NEXT(4104, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4104(choice, ...)  METALANG99_PRIV_REC_NEXT(4105, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4105(choice, ...)  METALANG99_PRIV_REC_NEXT(4106, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4106(choice, ...)  METALANG99_PRIV_REC_NEXT(4107, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4107(choice, ...)  METALANG99_PRIV_REC_NEXT(4108, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4108(choice, ...)  METALANG99_PRIV_REC_NEXT(4109, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4109(choice, ...)  METALANG99_PRIV_REC_NEXT(4110, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4110(choice, ...)  METALANG99_PRIV_REC_NEXT(4111, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4111(choice, ...)  METALANG99_PRIV_REC_NEXT(4112, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4112(choice, ...)  METALANG99_PRIV_REC_NEXT(4113, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4113(choice, ...)  METALANG99_PRIV_REC_NEXT(4114, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4114(choice, ...)  METALANG99_PRIV_REC_NEXT(4115, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4115(choice, ...)  METALANG99_PRIV_REC_NEXT(4116, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4116(choice, ...)  METALANG99_PRIV_REC_NEXT(4117, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4117(choice, ...)  METALANG99_PRIV_REC_NEXT(4118, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4118(choice, ...)  METALANG99_PRIV_REC_NEXT(4119, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4119(choice, ...)  METALANG99_PRIV_REC_NEXT(4120, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4120(choice, ...)  METALANG99_PRIV_REC_NEXT(4121, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4121(choice, ...)  METALANG99_PRIV_REC_NEXT(4122, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4122(choice, ...)  METALANG99_PRIV_REC_NEXT(4123, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4123(choice, ...)  METALANG99_PRIV_REC_NEXT(4124, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4124(choice, ...)  METALANG99_PRIV_REC_NEXT(4125, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4125(choice, ...)  METALANG99_PRIV_REC_NEXT(4126, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4126(choice, ...)  METALANG99_PRIV_REC_NEXT(4127, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4127(choice, ...)  METALANG99_PRIV_REC_NEXT(4128, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4128(choice, ...)  METALANG99_PRIV_REC_NEXT(4129, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4129(choice, ...)  METALANG99_PRIV_REC_NEXT(4130, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4130(choice, ...)  METALANG99_PRIV_REC_NEXT(4131, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4131(choice, ...)  METALANG99_PRIV_REC_NEXT(4132, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4132(choice, ...)  METALANG99_PRIV_REC_NEXT(4133, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4133(choice, ...)  METALANG99_PRIV_REC_NEXT(4134, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4134(choice, ...)  METALANG99_PRIV_REC_NEXT(4135, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4135(choice, ...)  METALANG99_PRIV_REC_NEXT(4136, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4136(choice, ...)  METALANG99_PRIV_REC_NEXT(4137, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4137(choice, ...)  METALANG99_PRIV_REC_NEXT(4138, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4138(choice, ...)  METALANG99_PRIV_REC_NEXT(4139, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4139(choice, ...)  METALANG99_PRIV_REC_NEXT(4140, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4140(choice, ...)  METALANG99_PRIV_REC_NEXT(4141, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4141(choice, ...)  METALANG99_PRIV_REC_NEXT(4142, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4142(choice, ...)  METALANG99_PRIV_REC_NEXT(4143, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4143(choice, ...)  METALANG99_PRIV_REC_NEXT(4144, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4144(choice, ...)  METALANG99_PRIV_REC_NEXT(4145, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4145(choice, ...)  METALANG99_PRIV_REC_NEXT(4146, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4146(choice, ...)  METALANG99_PRIV_REC_NEXT(4147, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4147(choice, ...)  METALANG99_PRIV_REC_NEXT(4148, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4148(choice, ...)  METALANG99_PRIV_REC_NEXT(4149, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4149(choice, ...)  METALANG99_PRIV_REC_NEXT(4150, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4150(choice, ...)  METALANG99_PRIV_REC_NEXT(4151, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4151(choice, ...)  METALANG99_PRIV_REC_NEXT(4152, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4152(choice, ...)  METALANG99_PRIV_REC_NEXT(4153, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4153(choice, ...)  METALANG99_PRIV_REC_NEXT(4154, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4154(choice, ...)  METALANG99_PRIV_REC_NEXT(4155, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4155(choice, ...)  METALANG99_PRIV_REC_NEXT(4156, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4156(choice, ...)  METALANG99_PRIV_REC_NEXT(4157, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4157(choice, ...)  METALANG99_PRIV_REC_NEXT(4158, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4158(choice, ...)  METALANG99_PRIV_REC_NEXT(4159, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4159(choice, ...)  METALANG99_PRIV_REC_NEXT(4160, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4160(choice, ...)  METALANG99_PRIV_REC_NEXT(4161, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4161(choice, ...)  METALANG99_PRIV_REC_NEXT(4162, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4162(choice, ...)  METALANG99_PRIV_REC_NEXT(4163, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4163(choice, ...)  METALANG99_PRIV_REC_NEXT(4164, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4164(choice, ...)  METALANG99_PRIV_REC_NEXT(4165, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4165(choice, ...)  METALANG99_PRIV_REC_NEXT(4166, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4166(choice, ...)  METALANG99_PRIV_REC_NEXT(4167, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4167(choice, ...)  METALANG99_PRIV_REC_NEXT(4168, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4168(choice, ...)  METALANG99_PRIV_REC_NEXT(4169, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4169(choice, ...)  METALANG99_PRIV_REC_NEXT(4170, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4170(choice, ...)  METALANG99_PRIV_REC_NEXT(4171, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4171(choice, ...)  METALANG99_PRIV_REC_NEXT(4172, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4172(choice, ...)  METALANG99_PRIV_REC_NEXT(4173, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4173(choice, ...)  METALANG99_PRIV_REC_NEXT(4174, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4174(choice, ...)  METALANG99_PRIV_REC_NEXT(4175, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4175(choice, ...)  METALANG99_PRIV_REC_NEXT(4176, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4176(choice, ...)  METALANG99_PRIV_REC_NEXT(4177, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4177(choice, ...)  METALANG99_PRIV_REC_NEXT(4178, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4178(choice, ...)  METALANG99_PRIV_REC_NEXT(4179, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4179(choice, ...)  METALANG99_PRIV_REC_NEXT(4180, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4180(choice, ...)  METALANG99_PRIV_REC_NEXT(4181, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4181(choice, ...)  METALANG99_PRIV_REC_NEXT(4182, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4182(choice, ...)  METALANG99_PRIV_REC_NEXT(4183, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4183(choice, ...)  METALANG99_PRIV_REC_NEXT(4184, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4184(choice, ...)  METALANG99_PRIV_REC_NEXT(4185, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4185(choice, ...)  METALANG99_PRIV_REC_NEXT(4186, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4186(choice, ...)  METALANG99_PRIV_REC_NEXT(4187, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4187(choice, ...)  METALANG99_PRIV_REC_NEXT(4188, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4188(choice, ...)  METALANG99_PRIV_REC_NEXT(4189, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4189(choice, ...)  METALANG99_PRIV_REC_NEXT(4190, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4190(choice, ...)  METALANG99_PRIV_REC_NEXT(4191, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4191(choice, ...)  METALANG99_PRIV_REC_NEXT(4192, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4192(choice, ...)  METALANG99_PRIV_REC_NEXT(4193, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4193(choice, ...)  METALANG99_PRIV_REC_NEXT(4194, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4194(choice, ...)  METALANG99_PRIV_REC_NEXT(4195, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4195(choice, ...)  METALANG99_PRIV_REC_NEXT(4196, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4196(choice, ...)  METALANG99_PRIV_REC_NEXT(4197, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4197(choice, ...)  METALANG99_PRIV_REC_NEXT(4198, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4198(choice, ...)  METALANG99_PRIV_REC_NEXT(4199, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4199(choice, ...)  METALANG99_PRIV_REC_NEXT(4200, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4200(choice, ...)  METALANG99_PRIV_REC_NEXT(4201, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4201(choice, ...)  METALANG99_PRIV_REC_NEXT(4202, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4202(choice, ...)  METALANG99_PRIV_REC_NEXT(4203, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4203(choice, ...)  METALANG99_PRIV_REC_NEXT(4204, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4204(choice, ...)  METALANG99_PRIV_REC_NEXT(4205, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4205(choice, ...)  METALANG99_PRIV_REC_NEXT(4206, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4206(choice, ...)  METALANG99_PRIV_REC_NEXT(4207, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4207(choice, ...)  METALANG99_PRIV_REC_NEXT(4208, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4208(choice, ...)  METALANG99_PRIV_REC_NEXT(4209, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4209(choice, ...)  METALANG99_PRIV_REC_NEXT(4210, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4210(choice, ...)  METALANG99_PRIV_REC_NEXT(4211, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4211(choice, ...)  METALANG99_PRIV_REC_NEXT(4212, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4212(choice, ...)  METALANG99_PRIV_REC_NEXT(4213, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4213(choice, ...)  METALANG99_PRIV_REC_NEXT(4214, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4214(choice, ...)  METALANG99_PRIV_REC_NEXT(4215, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4215(choice, ...)  METALANG99_PRIV_REC_NEXT(4216, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4216(choice, ...)  METALANG99_PRIV_REC_NEXT(4217, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4217(choice, ...)  METALANG99_PRIV_REC_NEXT(4218, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4218(choice, ...)  METALANG99_PRIV_REC_NEXT(4219, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4219(choice, ...)  METALANG99_PRIV_REC_NEXT(4220, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4220(choice, ...)  METALANG99_PRIV_REC_NEXT(4221, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4221(choice, ...)  METALANG99_PRIV_REC_NEXT(4222, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4222(choice, ...)  METALANG99_PRIV_REC_NEXT(4223, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4223(choice, ...)  METALANG99_PRIV_REC_NEXT(4224, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4224(choice, ...)  METALANG99_PRIV_REC_NEXT(4225, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4225(choice, ...)  METALANG99_PRIV_REC_NEXT(4226, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4226(choice, ...)  METALANG99_PRIV_REC_NEXT(4227, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4227(choice, ...)  METALANG99_PRIV_REC_NEXT(4228, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4228(choice, ...)  METALANG99_PRIV_REC_NEXT(4229, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4229(choice, ...)  METALANG99_PRIV_REC_NEXT(4230, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4230(choice, ...)  METALANG99_PRIV_REC_NEXT(4231, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4231(choice, ...)  METALANG99_PRIV_REC_NEXT(4232, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4232(choice, ...)  METALANG99_PRIV_REC_NEXT(4233, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4233(choice, ...)  METALANG99_PRIV_REC_NEXT(4234, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4234(choice, ...)  METALANG99_PRIV_REC_NEXT(4235, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4235(choice, ...)  METALANG99_PRIV_REC_NEXT(4236, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4236(choice, ...)  METALANG99_PRIV_REC_NEXT(4237, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4237(choice, ...)  METALANG99_PRIV_REC_NEXT(4238, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4238(choice, ...)  METALANG99_PRIV_REC_NEXT(4239, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4239(choice, ...)  METALANG99_PRIV_REC_NEXT(4240, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4240(choice, ...)  METALANG99_PRIV_REC_NEXT(4241, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4241(choice, ...)  METALANG99_PRIV_REC_NEXT(4242, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4242(choice, ...)  METALANG99_PRIV_REC_NEXT(4243, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4243(choice, ...)  METALANG99_PRIV_REC_NEXT(4244, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4244(choice, ...)  METALANG99_PRIV_REC_NEXT(4245, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4245(choice, ...)  METALANG99_PRIV_REC_NEXT(4246, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4246(choice, ...)  METALANG99_PRIV_REC_NEXT(4247, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4247(choice, ...)  METALANG99_PRIV_REC_NEXT(4248, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4248(choice, ...)  METALANG99_PRIV_REC_NEXT(4249, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4249(choice, ...)  METALANG99_PRIV_REC_NEXT(4250, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4250(choice, ...)  METALANG99_PRIV_REC_NEXT(4251, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4251(choice, ...)  METALANG99_PRIV_REC_NEXT(4252, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4252(choice, ...)  METALANG99_PRIV_REC_NEXT(4253, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4253(choice, ...)  METALANG99_PRIV_REC_NEXT(4254, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4254(choice, ...)  METALANG99_PRIV_REC_NEXT(4255, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4255(choice, ...)  METALANG99_PRIV_REC_NEXT(4256, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4256(choice, ...)  METALANG99_PRIV_REC_NEXT(4257, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4257(choice, ...)  METALANG99_PRIV_REC_NEXT(4258, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4258(choice, ...)  METALANG99_PRIV_REC_NEXT(4259, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4259(choice, ...)  METALANG99_PRIV_REC_NEXT(4260, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4260(choice, ...)  METALANG99_PRIV_REC_NEXT(4261, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4261(choice, ...)  METALANG99_PRIV_REC_NEXT(4262, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4262(choice, ...)  METALANG99_PRIV_REC_NEXT(4263, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4263(choice, ...)  METALANG99_PRIV_REC_NEXT(4264, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4264(choice, ...)  METALANG99_PRIV_REC_NEXT(4265, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4265(choice, ...)  METALANG99_PRIV_REC_NEXT(4266, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4266(choice, ...)  METALANG99_PRIV_REC_NEXT(4267, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4267(choice, ...)  METALANG99_PRIV_REC_NEXT(4268, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4268(choice, ...)  METALANG99_PRIV_REC_NEXT(4269, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4269(choice, ...)  METALANG99_PRIV_REC_NEXT(4270, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4270(choice, ...)  METALANG99_PRIV_REC_NEXT(4271, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4271(choice, ...)  METALANG99_PRIV_REC_NEXT(4272, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4272(choice, ...)  METALANG99_PRIV_REC_NEXT(4273, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4273(choice, ...)  METALANG99_PRIV_REC_NEXT(4274, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4274(choice, ...)  METALANG99_PRIV_REC_NEXT(4275, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4275(choice, ...)  METALANG99_PRIV_REC_NEXT(4276, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4276(choice, ...)  METALANG99_PRIV_REC_NEXT(4277, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4277(choice, ...)  METALANG99_PRIV_REC_NEXT(4278, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4278(choice, ...)  METALANG99_PRIV_REC_NEXT(4279, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4279(choice, ...)  METALANG99_PRIV_REC_NEXT(4280, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4280(choice, ...)  METALANG99_PRIV_REC_NEXT(4281, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4281(choice, ...)  METALANG99_PRIV_REC_NEXT(4282, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4282(choice, ...)  METALANG99_PRIV_REC_NEXT(4283, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4283(choice, ...)  METALANG99_PRIV_REC_NEXT(4284, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4284(choice, ...)  METALANG99_PRIV_REC_NEXT(4285, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4285(choice, ...)  METALANG99_PRIV_REC_NEXT(4286, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4286(choice, ...)  METALANG99_PRIV_REC_NEXT(4287, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4287(choice, ...)  METALANG99_PRIV_REC_NEXT(4288, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4288(choice, ...)  METALANG99_PRIV_REC_NEXT(4289, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4289(choice, ...)  METALANG99_PRIV_REC_NEXT(4290, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4290(choice, ...)  METALANG99_PRIV_REC_NEXT(4291, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4291(choice, ...)  METALANG99_PRIV_REC_NEXT(4292, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4292(choice, ...)  METALANG99_PRIV_REC_NEXT(4293, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4293(choice, ...)  METALANG99_PRIV_REC_NEXT(4294, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4294(choice, ...)  METALANG99_PRIV_REC_NEXT(4295, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4295(choice, ...)  METALANG99_PRIV_REC_NEXT(4296, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4296(choice, ...)  METALANG99_PRIV_REC_NEXT(4297, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4297(choice, ...)  METALANG99_PRIV_REC_NEXT(4298, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4298(choice, ...)  METALANG99_PRIV_REC_NEXT(4299, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4299(choice, ...)  METALANG99_PRIV_REC_NEXT(4300, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4300(choice, ...)  METALANG99_PRIV_REC_NEXT(4301, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4301(choice, ...)  METALANG99_PRIV_REC_NEXT(4302, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4302(choice, ...)  METALANG99_PRIV_REC_NEXT(4303, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4303(choice, ...)  METALANG99_PRIV_REC_NEXT(4304, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4304(choice, ...)  METALANG99_PRIV_REC_NEXT(4305, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4305(choice, ...)  METALANG99_PRIV_REC_NEXT(4306, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4306(choice, ...)  METALANG99_PRIV_REC_NEXT(4307, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4307(choice, ...)  METALANG99_PRIV_REC_NEXT(4308, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4308(choice, ...)  METALANG99_PRIV_REC_NEXT(4309, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4309(choice, ...)  METALANG99_PRIV_REC_NEXT(4310, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4310(choice, ...)  METALANG99_PRIV_REC_NEXT(4311, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4311(choice, ...)  METALANG99_PRIV_REC_NEXT(4312, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4312(choice, ...)  METALANG99_PRIV_REC_NEXT(4313, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4313(choice, ...)  METALANG99_PRIV_REC_NEXT(4314, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4314(choice, ...)  METALANG99_PRIV_REC_NEXT(4315, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4315(choice, ...)  METALANG99_PRIV_REC_NEXT(4316, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4316(choice, ...)  METALANG99_PRIV_REC_NEXT(4317, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4317(choice, ...)  METALANG99_PRIV_REC_NEXT(4318, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4318(choice, ...)  METALANG99_PRIV_REC_NEXT(4319, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4319(choice, ...)  METALANG99_PRIV_REC_NEXT(4320, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4320(choice, ...)  METALANG99_PRIV_REC_NEXT(4321, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4321(choice, ...)  METALANG99_PRIV_REC_NEXT(4322, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4322(choice, ...)  METALANG99_PRIV_REC_NEXT(4323, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4323(choice, ...)  METALANG99_PRIV_REC_NEXT(4324, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4324(choice, ...)  METALANG99_PRIV_REC_NEXT(4325, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4325(choice, ...)  METALANG99_PRIV_REC_NEXT(4326, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4326(choice, ...)  METALANG99_PRIV_REC_NEXT(4327, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4327(choice, ...)  METALANG99_PRIV_REC_NEXT(4328, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4328(choice, ...)  METALANG99_PRIV_REC_NEXT(4329, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4329(choice, ...)  METALANG99_PRIV_REC_NEXT(4330, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4330(choice, ...)  METALANG99_PRIV_REC_NEXT(4331, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4331(choice, ...)  METALANG99_PRIV_REC_NEXT(4332, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4332(choice, ...)  METALANG99_PRIV_REC_NEXT(4333, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4333(choice, ...)  METALANG99_PRIV_REC_NEXT(4334, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4334(choice, ...)  METALANG99_PRIV_REC_NEXT(4335, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4335(choice, ...)  METALANG99_PRIV_REC_NEXT(4336, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4336(choice, ...)  METALANG99_PRIV_REC_NEXT(4337, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4337(choice, ...)  METALANG99_PRIV_REC_NEXT(4338, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4338(choice, ...)  METALANG99_PRIV_REC_NEXT(4339, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4339(choice, ...)  METALANG99_PRIV_REC_NEXT(4340, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4340(choice, ...)  METALANG99_PRIV_REC_NEXT(4341, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4341(choice, ...)  METALANG99_PRIV_REC_NEXT(4342, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4342(choice, ...)  METALANG99_PRIV_REC_NEXT(4343, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4343(choice, ...)  METALANG99_PRIV_REC_NEXT(4344, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4344(choice, ...)  METALANG99_PRIV_REC_NEXT(4345, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4345(choice, ...)  METALANG99_PRIV_REC_NEXT(4346, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4346(choice, ...)  METALANG99_PRIV_REC_NEXT(4347, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4347(choice, ...)  METALANG99_PRIV_REC_NEXT(4348, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4348(choice, ...)  METALANG99_PRIV_REC_NEXT(4349, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4349(choice, ...)  METALANG99_PRIV_REC_NEXT(4350, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4350(choice, ...)  METALANG99_PRIV_REC_NEXT(4351, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4351(choice, ...)  METALANG99_PRIV_REC_NEXT(4352, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4352(choice, ...)  METALANG99_PRIV_REC_NEXT(4353, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4353(choice, ...)  METALANG99_PRIV_REC_NEXT(4354, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4354(choice, ...)  METALANG99_PRIV_REC_NEXT(4355, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4355(choice, ...)  METALANG99_PRIV_REC_NEXT(4356, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4356(choice, ...)  METALANG99_PRIV_REC_NEXT(4357, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4357(choice, ...)  METALANG99_PRIV_REC_NEXT(4358, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4358(choice, ...)  METALANG99_PRIV_REC_NEXT(4359, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4359(choice, ...)  METALANG99_PRIV_REC_NEXT(4360, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4360(choice, ...)  METALANG99_PRIV_REC_NEXT(4361, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4361(choice, ...)  METALANG99_PRIV_REC_NEXT(4362, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4362(choice, ...)  METALANG99_PRIV_REC_NEXT(4363, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4363(choice, ...)  METALANG99_PRIV_REC_NEXT(4364, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4364(choice, ...)  METALANG99_PRIV_REC_NEXT(4365, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4365(choice, ...)  METALANG99_PRIV_REC_NEXT(4366, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4366(choice, ...)  METALANG99_PRIV_REC_NEXT(4367, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4367(choice, ...)  METALANG99_PRIV_REC_NEXT(4368, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4368(choice, ...)  METALANG99_PRIV_REC_NEXT(4369, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4369(choice, ...)  METALANG99_PRIV_REC_NEXT(4370, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4370(choice, ...)  METALANG99_PRIV_REC_NEXT(4371, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4371(choice, ...)  METALANG99_PRIV_REC_NEXT(4372, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4372(choice, ...)  METALANG99_PRIV_REC_NEXT(4373, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4373(choice, ...)  METALANG99_PRIV_REC_NEXT(4374, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4374(choice, ...)  METALANG99_PRIV_REC_NEXT(4375, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4375(choice, ...)  METALANG99_PRIV_REC_NEXT(4376, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4376(choice, ...)  METALANG99_PRIV_REC_NEXT(4377, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4377(choice, ...)  METALANG99_PRIV_REC_NEXT(4378, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4378(choice, ...)  METALANG99_PRIV_REC_NEXT(4379, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4379(choice, ...)  METALANG99_PRIV_REC_NEXT(4380, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4380(choice, ...)  METALANG99_PRIV_REC_NEXT(4381, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4381(choice, ...)  METALANG99_PRIV_REC_NEXT(4382, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4382(choice, ...)  METALANG99_PRIV_REC_NEXT(4383, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4383(choice, ...)  METALANG99_PRIV_REC_NEXT(4384, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4384(choice, ...)  METALANG99_PRIV_REC_NEXT(4385, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4385(choice, ...)  METALANG99_PRIV_REC_NEXT(4386, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4386(choice, ...)  METALANG99_PRIV_REC_NEXT(4387, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4387(choice, ...)  METALANG99_PRIV_REC_NEXT(4388, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4388(choice, ...)  METALANG99_PRIV_REC_NEXT(4389, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4389(choice, ...)  METALANG99_PRIV_REC_NEXT(4390, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4390(choice, ...)  METALANG99_PRIV_REC_NEXT(4391, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4391(choice, ...)  METALANG99_PRIV_REC_NEXT(4392, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4392(choice, ...)  METALANG99_PRIV_REC_NEXT(4393, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4393(choice, ...)  METALANG99_PRIV_REC_NEXT(4394, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4394(choice, ...)  METALANG99_PRIV_REC_NEXT(4395, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4395(choice, ...)  METALANG99_PRIV_REC_NEXT(4396, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4396(choice, ...)  METALANG99_PRIV_REC_NEXT(4397, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4397(choice, ...)  METALANG99_PRIV_REC_NEXT(4398, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4398(choice, ...)  METALANG99_PRIV_REC_NEXT(4399, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4399(choice, ...)  METALANG99_PRIV_REC_NEXT(4400, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4400(choice, ...)  METALANG99_PRIV_REC_NEXT(4401, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4401(choice, ...)  METALANG99_PRIV_REC_NEXT(4402, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4402(choice, ...)  METALANG99_PRIV_REC_NEXT(4403, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4403(choice, ...)  METALANG99_PRIV_REC_NEXT(4404, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4404(choice, ...)  METALANG99_PRIV_REC_NEXT(4405, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4405(choice, ...)  METALANG99_PRIV_REC_NEXT(4406, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4406(choice, ...)  METALANG99_PRIV_REC_NEXT(4407, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4407(choice, ...)  METALANG99_PRIV_REC_NEXT(4408, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4408(choice, ...)  METALANG99_PRIV_REC_NEXT(4409, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4409(choice, ...)  METALANG99_PRIV_REC_NEXT(4410, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4410(choice, ...)  METALANG99_PRIV_REC_NEXT(4411, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4411(choice, ...)  METALANG99_PRIV_REC_NEXT(4412, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4412(choice, ...)  METALANG99_PRIV_REC_NEXT(4413, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4413(choice, ...)  METALANG99_PRIV_REC_NEXT(4414, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4414(choice, ...)  METALANG99_PRIV_REC_NEXT(4415, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4415(choice, ...)  METALANG99_PRIV_REC_NEXT(4416, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4416(choice, ...)  METALANG99_PRIV_REC_NEXT(4417, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4417(choice, ...)  METALANG99_PRIV_REC_NEXT(4418, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4418(choice, ...)  METALANG99_PRIV_REC_NEXT(4419, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4419(choice, ...)  METALANG99_PRIV_REC_NEXT(4420, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4420(choice, ...)  METALANG99_PRIV_REC_NEXT(4421, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4421(choice, ...)  METALANG99_PRIV_REC_NEXT(4422, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4422(choice, ...)  METALANG99_PRIV_REC_NEXT(4423, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4423(choice, ...)  METALANG99_PRIV_REC_NEXT(4424, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4424(choice, ...)  METALANG99_PRIV_REC_NEXT(4425, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4425(choice, ...)  METALANG99_PRIV_REC_NEXT(4426, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4426(choice, ...)  METALANG99_PRIV_REC_NEXT(4427, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4427(choice, ...)  METALANG99_PRIV_REC_NEXT(4428, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4428(choice, ...)  METALANG99_PRIV_REC_NEXT(4429, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4429(choice, ...)  METALANG99_PRIV_REC_NEXT(4430, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4430(choice, ...)  METALANG99_PRIV_REC_NEXT(4431, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4431(choice, ...)  METALANG99_PRIV_REC_NEXT(4432, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4432(choice, ...)  METALANG99_PRIV_REC_NEXT(4433, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4433(choice, ...)  METALANG99_PRIV_REC_NEXT(4434, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4434(choice, ...)  METALANG99_PRIV_REC_NEXT(4435, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4435(choice, ...)  METALANG99_PRIV_REC_NEXT(4436, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4436(choice, ...)  METALANG99_PRIV_REC_NEXT(4437, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4437(choice, ...)  METALANG99_PRIV_REC_NEXT(4438, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4438(choice, ...)  METALANG99_PRIV_REC_NEXT(4439, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4439(choice, ...)  METALANG99_PRIV_REC_NEXT(4440, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4440(choice, ...)  METALANG99_PRIV_REC_NEXT(4441, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4441(choice, ...)  METALANG99_PRIV_REC_NEXT(4442, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4442(choice, ...)  METALANG99_PRIV_REC_NEXT(4443, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4443(choice, ...)  METALANG99_PRIV_REC_NEXT(4444, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4444(choice, ...)  METALANG99_PRIV_REC_NEXT(4445, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4445(choice, ...)  METALANG99_PRIV_REC_NEXT(4446, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4446(choice, ...)  METALANG99_PRIV_REC_NEXT(4447, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4447(choice, ...)  METALANG99_PRIV_REC_NEXT(4448, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4448(choice, ...)  METALANG99_PRIV_REC_NEXT(4449, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4449(choice, ...)  METALANG99_PRIV_REC_NEXT(4450, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4450(choice, ...)  METALANG99_PRIV_REC_NEXT(4451, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4451(choice, ...)  METALANG99_PRIV_REC_NEXT(4452, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4452(choice, ...)  METALANG99_PRIV_REC_NEXT(4453, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4453(choice, ...)  METALANG99_PRIV_REC_NEXT(4454, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4454(choice, ...)  METALANG99_PRIV_REC_NEXT(4455, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4455(choice, ...)  METALANG99_PRIV_REC_NEXT(4456, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4456(choice, ...)  METALANG99_PRIV_REC_NEXT(4457, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4457(choice, ...)  METALANG99_PRIV_REC_NEXT(4458, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4458(choice, ...)  METALANG99_PRIV_REC_NEXT(4459, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4459(choice, ...)  METALANG99_PRIV_REC_NEXT(4460, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4460(choice, ...)  METALANG99_PRIV_REC_NEXT(4461, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4461(choice, ...)  METALANG99_PRIV_REC_NEXT(4462, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4462(choice, ...)  METALANG99_PRIV_REC_NEXT(4463, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4463(choice, ...)  METALANG99_PRIV_REC_NEXT(4464, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4464(choice, ...)  METALANG99_PRIV_REC_NEXT(4465, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4465(choice, ...)  METALANG99_PRIV_REC_NEXT(4466, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4466(choice, ...)  METALANG99_PRIV_REC_NEXT(4467, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4467(choice, ...)  METALANG99_PRIV_REC_NEXT(4468, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4468(choice, ...)  METALANG99_PRIV_REC_NEXT(4469, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4469(choice, ...)  METALANG99_PRIV_REC_NEXT(4470, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4470(choice, ...)  METALANG99_PRIV_REC_NEXT(4471, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4471(choice, ...)  METALANG99_PRIV_REC_NEXT(4472, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4472(choice, ...)  METALANG99_PRIV_REC_NEXT(4473, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4473(choice, ...)  METALANG99_PRIV_REC_NEXT(4474, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4474(choice, ...)  METALANG99_PRIV_REC_NEXT(4475, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4475(choice, ...)  METALANG99_PRIV_REC_NEXT(4476, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4476(choice, ...)  METALANG99_PRIV_REC_NEXT(4477, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4477(choice, ...)  METALANG99_PRIV_REC_NEXT(4478, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4478(choice, ...)  METALANG99_PRIV_REC_NEXT(4479, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4479(choice, ...)  METALANG99_PRIV_REC_NEXT(4480, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4480(choice, ...)  METALANG99_PRIV_REC_NEXT(4481, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4481(choice, ...)  METALANG99_PRIV_REC_NEXT(4482, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4482(choice, ...)  METALANG99_PRIV_REC_NEXT(4483, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4483(choice, ...)  METALANG99_PRIV_REC_NEXT(4484, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4484(choice, ...)  METALANG99_PRIV_REC_NEXT(4485, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4485(choice, ...)  METALANG99_PRIV_REC_NEXT(4486, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4486(choice, ...)  METALANG99_PRIV_REC_NEXT(4487, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4487(choice, ...)  METALANG99_PRIV_REC_NEXT(4488, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4488(choice, ...)  METALANG99_PRIV_REC_NEXT(4489, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4489(choice, ...)  METALANG99_PRIV_REC_NEXT(4490, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4490(choice, ...)  METALANG99_PRIV_REC_NEXT(4491, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4491(choice, ...)  METALANG99_PRIV_REC_NEXT(4492, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4492(choice, ...)  METALANG99_PRIV_REC_NEXT(4493, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4493(choice, ...)  METALANG99_PRIV_REC_NEXT(4494, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4494(choice, ...)  METALANG99_PRIV_REC_NEXT(4495, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4495(choice, ...)  METALANG99_PRIV_REC_NEXT(4496, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4496(choice, ...)  METALANG99_PRIV_REC_NEXT(4497, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4497(choice, ...)  METALANG99_PRIV_REC_NEXT(4498, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4498(choice, ...)  METALANG99_PRIV_REC_NEXT(4499, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4499(choice, ...)  METALANG99_PRIV_REC_NEXT(4500, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4500(choice, ...)  METALANG99_PRIV_REC_NEXT(4501, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4501(choice, ...)  METALANG99_PRIV_REC_NEXT(4502, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4502(choice, ...)  METALANG99_PRIV_REC_NEXT(4503, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4503(choice, ...)  METALANG99_PRIV_REC_NEXT(4504, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4504(choice, ...)  METALANG99_PRIV_REC_NEXT(4505, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4505(choice, ...)  METALANG99_PRIV_REC_NEXT(4506, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4506(choice, ...)  METALANG99_PRIV_REC_NEXT(4507, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4507(choice, ...)  METALANG99_PRIV_REC_NEXT(4508, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4508(choice, ...)  METALANG99_PRIV_REC_NEXT(4509, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4509(choice, ...)  METALANG99_PRIV_REC_NEXT(4510, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4510(choice, ...)  METALANG99_PRIV_REC_NEXT(4511, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4511(choice, ...)  METALANG99_PRIV_REC_NEXT(4512, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4512(choice, ...)  METALANG99_PRIV_REC_NEXT(4513, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4513(choice, ...)  METALANG99_PRIV_REC_NEXT(4514, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4514(choice, ...)  METALANG99_PRIV_REC_NEXT(4515, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4515(choice, ...)  METALANG99_PRIV_REC_NEXT(4516, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4516(choice, ...)  METALANG99_PRIV_REC_NEXT(4517, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4517(choice, ...)  METALANG99_PRIV_REC_NEXT(4518, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4518(choice, ...)  METALANG99_PRIV_REC_NEXT(4519, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4519(choice, ...)  METALANG99_PRIV_REC_NEXT(4520, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4520(choice, ...)  METALANG99_PRIV_REC_NEXT(4521, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4521(choice, ...)  METALANG99_PRIV_REC_NEXT(4522, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4522(choice, ...)  METALANG99_PRIV_REC_NEXT(4523, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4523(choice, ...)  METALANG99_PRIV_REC_NEXT(4524, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4524(choice, ...)  METALANG99_PRIV_REC_NEXT(4525, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4525(choice, ...)  METALANG99_PRIV_REC_NEXT(4526, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4526(choice, ...)  METALANG99_PRIV_REC_NEXT(4527, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4527(choice, ...)  METALANG99_PRIV_REC_NEXT(4528, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4528(choice, ...)  METALANG99_PRIV_REC_NEXT(4529, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4529(choice, ...)  METALANG99_PRIV_REC_NEXT(4530, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4530(choice, ...)  METALANG99_PRIV_REC_NEXT(4531, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4531(choice, ...)  METALANG99_PRIV_REC_NEXT(4532, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4532(choice, ...)  METALANG99_PRIV_REC_NEXT(4533, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4533(choice, ...)  METALANG99_PRIV_REC_NEXT(4534, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4534(choice, ...)  METALANG99_PRIV_REC_NEXT(4535, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4535(choice, ...)  METALANG99_PRIV_REC_NEXT(4536, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4536(choice, ...)  METALANG99_PRIV_REC_NEXT(4537, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4537(choice, ...)  METALANG99_PRIV_REC_NEXT(4538, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4538(choice, ...)  METALANG99_PRIV_REC_NEXT(4539, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4539(choice, ...)  METALANG99_PRIV_REC_NEXT(4540, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4540(choice, ...)  METALANG99_PRIV_REC_NEXT(4541, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4541(choice, ...)  METALANG99_PRIV_REC_NEXT(4542, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4542(choice, ...)  METALANG99_PRIV_REC_NEXT(4543, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4543(choice, ...)  METALANG99_PRIV_REC_NEXT(4544, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4544(choice, ...)  METALANG99_PRIV_REC_NEXT(4545, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4545(choice, ...)  METALANG99_PRIV_REC_NEXT(4546, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4546(choice, ...)  METALANG99_PRIV_REC_NEXT(4547, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4547(choice, ...)  METALANG99_PRIV_REC_NEXT(4548, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4548(choice, ...)  METALANG99_PRIV_REC_NEXT(4549, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4549(choice, ...)  METALANG99_PRIV_REC_NEXT(4550, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4550(choice, ...)  METALANG99_PRIV_REC_NEXT(4551, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4551(choice, ...)  METALANG99_PRIV_REC_NEXT(4552, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4552(choice, ...)  METALANG99_PRIV_REC_NEXT(4553, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4553(choice, ...)  METALANG99_PRIV_REC_NEXT(4554, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4554(choice, ...)  METALANG99_PRIV_REC_NEXT(4555, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4555(choice, ...)  METALANG99_PRIV_REC_NEXT(4556, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4556(choice, ...)  METALANG99_PRIV_REC_NEXT(4557, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4557(choice, ...)  METALANG99_PRIV_REC_NEXT(4558, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4558(choice, ...)  METALANG99_PRIV_REC_NEXT(4559, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4559(choice, ...)  METALANG99_PRIV_REC_NEXT(4560, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4560(choice, ...)  METALANG99_PRIV_REC_NEXT(4561, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4561(choice, ...)  METALANG99_PRIV_REC_NEXT(4562, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4562(choice, ...)  METALANG99_PRIV_REC_NEXT(4563, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4563(choice, ...)  METALANG99_PRIV_REC_NEXT(4564, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4564(choice, ...)  METALANG99_PRIV_REC_NEXT(4565, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4565(choice, ...)  METALANG99_PRIV_REC_NEXT(4566, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4566(choice, ...)  METALANG99_PRIV_REC_NEXT(4567, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4567(choice, ...)  METALANG99_PRIV_REC_NEXT(4568, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4568(choice, ...)  METALANG99_PRIV_REC_NEXT(4569, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4569(choice, ...)  METALANG99_PRIV_REC_NEXT(4570, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4570(choice, ...)  METALANG99_PRIV_REC_NEXT(4571, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4571(choice, ...)  METALANG99_PRIV_REC_NEXT(4572, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4572(choice, ...)  METALANG99_PRIV_REC_NEXT(4573, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4573(choice, ...)  METALANG99_PRIV_REC_NEXT(4574, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4574(choice, ...)  METALANG99_PRIV_REC_NEXT(4575, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4575(choice, ...)  METALANG99_PRIV_REC_NEXT(4576, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4576(choice, ...)  METALANG99_PRIV_REC_NEXT(4577, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4577(choice, ...)  METALANG99_PRIV_REC_NEXT(4578, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4578(choice, ...)  METALANG99_PRIV_REC_NEXT(4579, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4579(choice, ...)  METALANG99_PRIV_REC_NEXT(4580, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4580(choice, ...)  METALANG99_PRIV_REC_NEXT(4581, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4581(choice, ...)  METALANG99_PRIV_REC_NEXT(4582, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4582(choice, ...)  METALANG99_PRIV_REC_NEXT(4583, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4583(choice, ...)  METALANG99_PRIV_REC_NEXT(4584, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4584(choice, ...)  METALANG99_PRIV_REC_NEXT(4585, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4585(choice, ...)  METALANG99_PRIV_REC_NEXT(4586, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4586(choice, ...)  METALANG99_PRIV_REC_NEXT(4587, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4587(choice, ...)  METALANG99_PRIV_REC_NEXT(4588, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4588(choice, ...)  METALANG99_PRIV_REC_NEXT(4589, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4589(choice, ...)  METALANG99_PRIV_REC_NEXT(4590, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4590(choice, ...)  METALANG99_PRIV_REC_NEXT(4591, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4591(choice, ...)  METALANG99_PRIV_REC_NEXT(4592, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4592(choice, ...)  METALANG99_PRIV_REC_NEXT(4593, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4593(choice, ...)  METALANG99_PRIV_REC_NEXT(4594, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4594(choice, ...)  METALANG99_PRIV_REC_NEXT(4595, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4595(choice, ...)  METALANG99_PRIV_REC_NEXT(4596, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4596(choice, ...)  METALANG99_PRIV_REC_NEXT(4597, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4597(choice, ...)  METALANG99_PRIV_REC_NEXT(4598, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4598(choice, ...)  METALANG99_PRIV_REC_NEXT(4599, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4599(choice, ...)  METALANG99_PRIV_REC_NEXT(4600, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4600(choice, ...)  METALANG99_PRIV_REC_NEXT(4601, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4601(choice, ...)  METALANG99_PRIV_REC_NEXT(4602, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4602(choice, ...)  METALANG99_PRIV_REC_NEXT(4603, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4603(choice, ...)  METALANG99_PRIV_REC_NEXT(4604, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4604(choice, ...)  METALANG99_PRIV_REC_NEXT(4605, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4605(choice, ...)  METALANG99_PRIV_REC_NEXT(4606, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4606(choice, ...)  METALANG99_PRIV_REC_NEXT(4607, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4607(choice, ...)  METALANG99_PRIV_REC_NEXT(4608, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4608(choice, ...)  METALANG99_PRIV_REC_NEXT(4609, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4609(choice, ...)  METALANG99_PRIV_REC_NEXT(4610, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4610(choice, ...)  METALANG99_PRIV_REC_NEXT(4611, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4611(choice, ...)  METALANG99_PRIV_REC_NEXT(4612, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4612(choice, ...)  METALANG99_PRIV_REC_NEXT(4613, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4613(choice, ...)  METALANG99_PRIV_REC_NEXT(4614, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4614(choice, ...)  METALANG99_PRIV_REC_NEXT(4615, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4615(choice, ...)  METALANG99_PRIV_REC_NEXT(4616, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4616(choice, ...)  METALANG99_PRIV_REC_NEXT(4617, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4617(choice, ...)  METALANG99_PRIV_REC_NEXT(4618, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4618(choice, ...)  METALANG99_PRIV_REC_NEXT(4619, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4619(choice, ...)  METALANG99_PRIV_REC_NEXT(4620, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4620(choice, ...)  METALANG99_PRIV_REC_NEXT(4621, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4621(choice, ...)  METALANG99_PRIV_REC_NEXT(4622, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4622(choice, ...)  METALANG99_PRIV_REC_NEXT(4623, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4623(choice, ...)  METALANG99_PRIV_REC_NEXT(4624, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4624(choice, ...)  METALANG99_PRIV_REC_NEXT(4625, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4625(choice, ...)  METALANG99_PRIV_REC_NEXT(4626, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4626(choice, ...)  METALANG99_PRIV_REC_NEXT(4627, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4627(choice, ...)  METALANG99_PRIV_REC_NEXT(4628, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4628(choice, ...)  METALANG99_PRIV_REC_NEXT(4629, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4629(choice, ...)  METALANG99_PRIV_REC_NEXT(4630, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4630(choice, ...)  METALANG99_PRIV_REC_NEXT(4631, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4631(choice, ...)  METALANG99_PRIV_REC_NEXT(4632, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4632(choice, ...)  METALANG99_PRIV_REC_NEXT(4633, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4633(choice, ...)  METALANG99_PRIV_REC_NEXT(4634, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4634(choice, ...)  METALANG99_PRIV_REC_NEXT(4635, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4635(choice, ...)  METALANG99_PRIV_REC_NEXT(4636, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4636(choice, ...)  METALANG99_PRIV_REC_NEXT(4637, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4637(choice, ...)  METALANG99_PRIV_REC_NEXT(4638, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4638(choice, ...)  METALANG99_PRIV_REC_NEXT(4639, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4639(choice, ...)  METALANG99_PRIV_REC_NEXT(4640, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4640(choice, ...)  METALANG99_PRIV_REC_NEXT(4641, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4641(choice, ...)  METALANG99_PRIV_REC_NEXT(4642, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4642(choice, ...)  METALANG99_PRIV_REC_NEXT(4643, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4643(choice, ...)  METALANG99_PRIV_REC_NEXT(4644, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4644(choice, ...)  METALANG99_PRIV_REC_NEXT(4645, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4645(choice, ...)  METALANG99_PRIV_REC_NEXT(4646, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4646(choice, ...)  METALANG99_PRIV_REC_NEXT(4647, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4647(choice, ...)  METALANG99_PRIV_REC_NEXT(4648, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4648(choice, ...)  METALANG99_PRIV_REC_NEXT(4649, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4649(choice, ...)  METALANG99_PRIV_REC_NEXT(4650, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4650(choice, ...)  METALANG99_PRIV_REC_NEXT(4651, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4651(choice, ...)  METALANG99_PRIV_REC_NEXT(4652, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4652(choice, ...)  METALANG99_PRIV_REC_NEXT(4653, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4653(choice, ...)  METALANG99_PRIV_REC_NEXT(4654, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4654(choice, ...)  METALANG99_PRIV_REC_NEXT(4655, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4655(choice, ...)  METALANG99_PRIV_REC_NEXT(4656, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4656(choice, ...)  METALANG99_PRIV_REC_NEXT(4657, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4657(choice, ...)  METALANG99_PRIV_REC_NEXT(4658, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4658(choice, ...)  METALANG99_PRIV_REC_NEXT(4659, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4659(choice, ...)  METALANG99_PRIV_REC_NEXT(4660, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4660(choice, ...)  METALANG99_PRIV_REC_NEXT(4661, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4661(choice, ...)  METALANG99_PRIV_REC_NEXT(4662, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4662(choice, ...)  METALANG99_PRIV_REC_NEXT(4663, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4663(choice, ...)  METALANG99_PRIV_REC_NEXT(4664, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4664(choice, ...)  METALANG99_PRIV_REC_NEXT(4665, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4665(choice, ...)  METALANG99_PRIV_REC_NEXT(4666, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4666(choice, ...)  METALANG99_PRIV_REC_NEXT(4667, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4667(choice, ...)  METALANG99_PRIV_REC_NEXT(4668, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4668(choice, ...)  METALANG99_PRIV_REC_NEXT(4669, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4669(choice, ...)  METALANG99_PRIV_REC_NEXT(4670, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4670(choice, ...)  METALANG99_PRIV_REC_NEXT(4671, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4671(choice, ...)  METALANG99_PRIV_REC_NEXT(4672, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4672(choice, ...)  METALANG99_PRIV_REC_NEXT(4673, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4673(choice, ...)  METALANG99_PRIV_REC_NEXT(4674, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4674(choice, ...)  METALANG99_PRIV_REC_NEXT(4675, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4675(choice, ...)  METALANG99_PRIV_REC_NEXT(4676, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4676(choice, ...)  METALANG99_PRIV_REC_NEXT(4677, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4677(choice, ...)  METALANG99_PRIV_REC_NEXT(4678, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4678(choice, ...)  METALANG99_PRIV_REC_NEXT(4679, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4679(choice, ...)  METALANG99_PRIV_REC_NEXT(4680, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4680(choice, ...)  METALANG99_PRIV_REC_NEXT(4681, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4681(choice, ...)  METALANG99_PRIV_REC_NEXT(4682, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4682(choice, ...)  METALANG99_PRIV_REC_NEXT(4683, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4683(choice, ...)  METALANG99_PRIV_REC_NEXT(4684, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4684(choice, ...)  METALANG99_PRIV_REC_NEXT(4685, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4685(choice, ...)  METALANG99_PRIV_REC_NEXT(4686, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4686(choice, ...)  METALANG99_PRIV_REC_NEXT(4687, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4687(choice, ...)  METALANG99_PRIV_REC_NEXT(4688, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4688(choice, ...)  METALANG99_PRIV_REC_NEXT(4689, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4689(choice, ...)  METALANG99_PRIV_REC_NEXT(4690, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4690(choice, ...)  METALANG99_PRIV_REC_NEXT(4691, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4691(choice, ...)  METALANG99_PRIV_REC_NEXT(4692, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4692(choice, ...)  METALANG99_PRIV_REC_NEXT(4693, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4693(choice, ...)  METALANG99_PRIV_REC_NEXT(4694, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4694(choice, ...)  METALANG99_PRIV_REC_NEXT(4695, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4695(choice, ...)  METALANG99_PRIV_REC_NEXT(4696, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4696(choice, ...)  METALANG99_PRIV_REC_NEXT(4697, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4697(choice, ...)  METALANG99_PRIV_REC_NEXT(4698, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4698(choice, ...)  METALANG99_PRIV_REC_NEXT(4699, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4699(choice, ...)  METALANG99_PRIV_REC_NEXT(4700, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4700(choice, ...)  METALANG99_PRIV_REC_NEXT(4701, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4701(choice, ...)  METALANG99_PRIV_REC_NEXT(4702, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4702(choice, ...)  METALANG99_PRIV_REC_NEXT(4703, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4703(choice, ...)  METALANG99_PRIV_REC_NEXT(4704, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4704(choice, ...)  METALANG99_PRIV_REC_NEXT(4705, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4705(choice, ...)  METALANG99_PRIV_REC_NEXT(4706, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4706(choice, ...)  METALANG99_PRIV_REC_NEXT(4707, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4707(choice, ...)  METALANG99_PRIV_REC_NEXT(4708, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4708(choice, ...)  METALANG99_PRIV_REC_NEXT(4709, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4709(choice, ...)  METALANG99_PRIV_REC_NEXT(4710, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4710(choice, ...)  METALANG99_PRIV_REC_NEXT(4711, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4711(choice, ...)  METALANG99_PRIV_REC_NEXT(4712, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4712(choice, ...)  METALANG99_PRIV_REC_NEXT(4713, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4713(choice, ...)  METALANG99_PRIV_REC_NEXT(4714, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4714(choice, ...)  METALANG99_PRIV_REC_NEXT(4715, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4715(choice, ...)  METALANG99_PRIV_REC_NEXT(4716, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4716(choice, ...)  METALANG99_PRIV_REC_NEXT(4717, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4717(choice, ...)  METALANG99_PRIV_REC_NEXT(4718, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4718(choice, ...)  METALANG99_PRIV_REC_NEXT(4719, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4719(choice, ...)  METALANG99_PRIV_REC_NEXT(4720, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4720(choice, ...)  METALANG99_PRIV_REC_NEXT(4721, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4721(choice, ...)  METALANG99_PRIV_REC_NEXT(4722, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4722(choice, ...)  METALANG99_PRIV_REC_NEXT(4723, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4723(choice, ...)  METALANG99_PRIV_REC_NEXT(4724, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4724(choice, ...)  METALANG99_PRIV_REC_NEXT(4725, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4725(choice, ...)  METALANG99_PRIV_REC_NEXT(4726, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4726(choice, ...)  METALANG99_PRIV_REC_NEXT(4727, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4727(choice, ...)  METALANG99_PRIV_REC_NEXT(4728, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4728(choice, ...)  METALANG99_PRIV_REC_NEXT(4729, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4729(choice, ...)  METALANG99_PRIV_REC_NEXT(4730, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4730(choice, ...)  METALANG99_PRIV_REC_NEXT(4731, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4731(choice, ...)  METALANG99_PRIV_REC_NEXT(4732, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4732(choice, ...)  METALANG99_PRIV_REC_NEXT(4733, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4733(choice, ...)  METALANG99_PRIV_REC_NEXT(4734, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4734(choice, ...)  METALANG99_PRIV_REC_NEXT(4735, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4735(choice, ...)  METALANG99_PRIV_REC_NEXT(4736, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4736(choice, ...)  METALANG99_PRIV_REC_NEXT(4737, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4737(choice, ...)  METALANG99_PRIV_REC_NEXT(4738, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4738(choice, ...)  METALANG99_PRIV_REC_NEXT(4739, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4739(choice, ...)  METALANG99_PRIV_REC_NEXT(4740, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4740(choice, ...)  METALANG99_PRIV_REC_NEXT(4741, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4741(choice, ...)  METALANG99_PRIV_REC_NEXT(4742, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4742(choice, ...)  METALANG99_PRIV_REC_NEXT(4743, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4743(choice, ...)  METALANG99_PRIV_REC_NEXT(4744, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4744(choice, ...)  METALANG99_PRIV_REC_NEXT(4745, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4745(choice, ...)  METALANG99_PRIV_REC_NEXT(4746, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4746(choice, ...)  METALANG99_PRIV_REC_NEXT(4747, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4747(choice, ...)  METALANG99_PRIV_REC_NEXT(4748, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4748(choice, ...)  METALANG99_PRIV_REC_NEXT(4749, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4749(choice, ...)  METALANG99_PRIV_REC_NEXT(4750, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4750(choice, ...)  METALANG99_PRIV_REC_NEXT(4751, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4751(choice, ...)  METALANG99_PRIV_REC_NEXT(4752, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4752(choice, ...)  METALANG99_PRIV_REC_NEXT(4753, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4753(choice, ...)  METALANG99_PRIV_REC_NEXT(4754, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4754(choice, ...)  METALANG99_PRIV_REC_NEXT(4755, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4755(choice, ...)  METALANG99_PRIV_REC_NEXT(4756, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4756(choice, ...)  METALANG99_PRIV_REC_NEXT(4757, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4757(choice, ...)  METALANG99_PRIV_REC_NEXT(4758, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4758(choice, ...)  METALANG99_PRIV_REC_NEXT(4759, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4759(choice, ...)  METALANG99_PRIV_REC_NEXT(4760, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4760(choice, ...)  METALANG99_PRIV_REC_NEXT(4761, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4761(choice, ...)  METALANG99_PRIV_REC_NEXT(4762, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4762(choice, ...)  METALANG99_PRIV_REC_NEXT(4763, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4763(choice, ...)  METALANG99_PRIV_REC_NEXT(4764, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4764(choice, ...)  METALANG99_PRIV_REC_NEXT(4765, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4765(choice, ...)  METALANG99_PRIV_REC_NEXT(4766, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4766(choice, ...)  METALANG99_PRIV_REC_NEXT(4767, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4767(choice, ...)  METALANG99_PRIV_REC_NEXT(4768, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4768(choice, ...)  METALANG99_PRIV_REC_NEXT(4769, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4769(choice, ...)  METALANG99_PRIV_REC_NEXT(4770, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4770(choice, ...)  METALANG99_PRIV_REC_NEXT(4771, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4771(choice, ...)  METALANG99_PRIV_REC_NEXT(4772, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4772(choice, ...)  METALANG99_PRIV_REC_NEXT(4773, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4773(choice, ...)  METALANG99_PRIV_REC_NEXT(4774, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4774(choice, ...)  METALANG99_PRIV_REC_NEXT(4775, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4775(choice, ...)  METALANG99_PRIV_REC_NEXT(4776, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4776(choice, ...)  METALANG99_PRIV_REC_NEXT(4777, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4777(choice, ...)  METALANG99_PRIV_REC_NEXT(4778, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4778(choice, ...)  METALANG99_PRIV_REC_NEXT(4779, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4779(choice, ...)  METALANG99_PRIV_REC_NEXT(4780, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4780(choice, ...)  METALANG99_PRIV_REC_NEXT(4781, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4781(choice, ...)  METALANG99_PRIV_REC_NEXT(4782, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4782(choice, ...)  METALANG99_PRIV_REC_NEXT(4783, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4783(choice, ...)  METALANG99_PRIV_REC_NEXT(4784, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4784(choice, ...)  METALANG99_PRIV_REC_NEXT(4785, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4785(choice, ...)  METALANG99_PRIV_REC_NEXT(4786, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4786(choice, ...)  METALANG99_PRIV_REC_NEXT(4787, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4787(choice, ...)  METALANG99_PRIV_REC_NEXT(4788, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4788(choice, ...)  METALANG99_PRIV_REC_NEXT(4789, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4789(choice, ...)  METALANG99_PRIV_REC_NEXT(4790, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4790(choice, ...)  METALANG99_PRIV_REC_NEXT(4791, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4791(choice, ...)  METALANG99_PRIV_REC_NEXT(4792, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4792(choice, ...)  METALANG99_PRIV_REC_NEXT(4793, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4793(choice, ...)  METALANG99_PRIV_REC_NEXT(4794, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4794(choice, ...)  METALANG99_PRIV_REC_NEXT(4795, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4795(choice, ...)  METALANG99_PRIV_REC_NEXT(4796, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4796(choice, ...)  METALANG99_PRIV_REC_NEXT(4797, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4797(choice, ...)  METALANG99_PRIV_REC_NEXT(4798, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4798(choice, ...)  METALANG99_PRIV_REC_NEXT(4799, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4799(choice, ...)  METALANG99_PRIV_REC_NEXT(4800, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4800(choice, ...)  METALANG99_PRIV_REC_NEXT(4801, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4801(choice, ...)  METALANG99_PRIV_REC_NEXT(4802, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4802(choice, ...)  METALANG99_PRIV_REC_NEXT(4803, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4803(choice, ...)  METALANG99_PRIV_REC_NEXT(4804, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4804(choice, ...)  METALANG99_PRIV_REC_NEXT(4805, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4805(choice, ...)  METALANG99_PRIV_REC_NEXT(4806, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4806(choice, ...)  METALANG99_PRIV_REC_NEXT(4807, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4807(choice, ...)  METALANG99_PRIV_REC_NEXT(4808, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4808(choice, ...)  METALANG99_PRIV_REC_NEXT(4809, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4809(choice, ...)  METALANG99_PRIV_REC_NEXT(4810, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4810(choice, ...)  METALANG99_PRIV_REC_NEXT(4811, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4811(choice, ...)  METALANG99_PRIV_REC_NEXT(4812, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4812(choice, ...)  METALANG99_PRIV_REC_NEXT(4813, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4813(choice, ...)  METALANG99_PRIV_REC_NEXT(4814, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4814(choice, ...)  METALANG99_PRIV_REC_NEXT(4815, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4815(choice, ...)  METALANG99_PRIV_REC_NEXT(4816, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4816(choice, ...)  METALANG99_PRIV_REC_NEXT(4817, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4817(choice, ...)  METALANG99_PRIV_REC_NEXT(4818, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4818(choice, ...)  METALANG99_PRIV_REC_NEXT(4819, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4819(choice, ...)  METALANG99_PRIV_REC_NEXT(4820, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4820(choice, ...)  METALANG99_PRIV_REC_NEXT(4821, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4821(choice, ...)  METALANG99_PRIV_REC_NEXT(4822, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4822(choice, ...)  METALANG99_PRIV_REC_NEXT(4823, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4823(choice, ...)  METALANG99_PRIV_REC_NEXT(4824, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4824(choice, ...)  METALANG99_PRIV_REC_NEXT(4825, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4825(choice, ...)  METALANG99_PRIV_REC_NEXT(4826, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4826(choice, ...)  METALANG99_PRIV_REC_NEXT(4827, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4827(choice, ...)  METALANG99_PRIV_REC_NEXT(4828, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4828(choice, ...)  METALANG99_PRIV_REC_NEXT(4829, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4829(choice, ...)  METALANG99_PRIV_REC_NEXT(4830, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4830(choice, ...)  METALANG99_PRIV_REC_NEXT(4831, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4831(choice, ...)  METALANG99_PRIV_REC_NEXT(4832, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4832(choice, ...)  METALANG99_PRIV_REC_NEXT(4833, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4833(choice, ...)  METALANG99_PRIV_REC_NEXT(4834, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4834(choice, ...)  METALANG99_PRIV_REC_NEXT(4835, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4835(choice, ...)  METALANG99_PRIV_REC_NEXT(4836, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4836(choice, ...)  METALANG99_PRIV_REC_NEXT(4837, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4837(choice, ...)  METALANG99_PRIV_REC_NEXT(4838, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4838(choice, ...)  METALANG99_PRIV_REC_NEXT(4839, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4839(choice, ...)  METALANG99_PRIV_REC_NEXT(4840, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4840(choice, ...)  METALANG99_PRIV_REC_NEXT(4841, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4841(choice, ...)  METALANG99_PRIV_REC_NEXT(4842, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4842(choice, ...)  METALANG99_PRIV_REC_NEXT(4843, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4843(choice, ...)  METALANG99_PRIV_REC_NEXT(4844, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4844(choice, ...)  METALANG99_PRIV_REC_NEXT(4845, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4845(choice, ...)  METALANG99_PRIV_REC_NEXT(4846, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4846(choice, ...)  METALANG99_PRIV_REC_NEXT(4847, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4847(choice, ...)  METALANG99_PRIV_REC_NEXT(4848, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4848(choice, ...)  METALANG99_PRIV_REC_NEXT(4849, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4849(choice, ...)  METALANG99_PRIV_REC_NEXT(4850, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4850(choice, ...)  METALANG99_PRIV_REC_NEXT(4851, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4851(choice, ...)  METALANG99_PRIV_REC_NEXT(4852, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4852(choice, ...)  METALANG99_PRIV_REC_NEXT(4853, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4853(choice, ...)  METALANG99_PRIV_REC_NEXT(4854, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4854(choice, ...)  METALANG99_PRIV_REC_NEXT(4855, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4855(choice, ...)  METALANG99_PRIV_REC_NEXT(4856, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4856(choice, ...)  METALANG99_PRIV_REC_NEXT(4857, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4857(choice, ...)  METALANG99_PRIV_REC_NEXT(4858, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4858(choice, ...)  METALANG99_PRIV_REC_NEXT(4859, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4859(choice, ...)  METALANG99_PRIV_REC_NEXT(4860, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4860(choice, ...)  METALANG99_PRIV_REC_NEXT(4861, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4861(choice, ...)  METALANG99_PRIV_REC_NEXT(4862, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4862(choice, ...)  METALANG99_PRIV_REC_NEXT(4863, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4863(choice, ...)  METALANG99_PRIV_REC_NEXT(4864, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4864(choice, ...)  METALANG99_PRIV_REC_NEXT(4865, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4865(choice, ...)  METALANG99_PRIV_REC_NEXT(4866, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4866(choice, ...)  METALANG99_PRIV_REC_NEXT(4867, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4867(choice, ...)  METALANG99_PRIV_REC_NEXT(4868, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4868(choice, ...)  METALANG99_PRIV_REC_NEXT(4869, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4869(choice, ...)  METALANG99_PRIV_REC_NEXT(4870, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4870(choice, ...)  METALANG99_PRIV_REC_NEXT(4871, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4871(choice, ...)  METALANG99_PRIV_REC_NEXT(4872, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4872(choice, ...)  METALANG99_PRIV_REC_NEXT(4873, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4873(choice, ...)  METALANG99_PRIV_REC_NEXT(4874, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4874(choice, ...)  METALANG99_PRIV_REC_NEXT(4875, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4875(choice, ...)  METALANG99_PRIV_REC_NEXT(4876, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4876(choice, ...)  METALANG99_PRIV_REC_NEXT(4877, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4877(choice, ...)  METALANG99_PRIV_REC_NEXT(4878, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4878(choice, ...)  METALANG99_PRIV_REC_NEXT(4879, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4879(choice, ...)  METALANG99_PRIV_REC_NEXT(4880, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4880(choice, ...)  METALANG99_PRIV_REC_NEXT(4881, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4881(choice, ...)  METALANG99_PRIV_REC_NEXT(4882, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4882(choice, ...)  METALANG99_PRIV_REC_NEXT(4883, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4883(choice, ...)  METALANG99_PRIV_REC_NEXT(4884, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4884(choice, ...)  METALANG99_PRIV_REC_NEXT(4885, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4885(choice, ...)  METALANG99_PRIV_REC_NEXT(4886, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4886(choice, ...)  METALANG99_PRIV_REC_NEXT(4887, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4887(choice, ...)  METALANG99_PRIV_REC_NEXT(4888, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4888(choice, ...)  METALANG99_PRIV_REC_NEXT(4889, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4889(choice, ...)  METALANG99_PRIV_REC_NEXT(4890, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4890(choice, ...)  METALANG99_PRIV_REC_NEXT(4891, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4891(choice, ...)  METALANG99_PRIV_REC_NEXT(4892, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4892(choice, ...)  METALANG99_PRIV_REC_NEXT(4893, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4893(choice, ...)  METALANG99_PRIV_REC_NEXT(4894, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4894(choice, ...)  METALANG99_PRIV_REC_NEXT(4895, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4895(choice, ...)  METALANG99_PRIV_REC_NEXT(4896, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4896(choice, ...)  METALANG99_PRIV_REC_NEXT(4897, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4897(choice, ...)  METALANG99_PRIV_REC_NEXT(4898, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4898(choice, ...)  METALANG99_PRIV_REC_NEXT(4899, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4899(choice, ...)  METALANG99_PRIV_REC_NEXT(4900, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4900(choice, ...)  METALANG99_PRIV_REC_NEXT(4901, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4901(choice, ...)  METALANG99_PRIV_REC_NEXT(4902, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4902(choice, ...)  METALANG99_PRIV_REC_NEXT(4903, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4903(choice, ...)  METALANG99_PRIV_REC_NEXT(4904, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4904(choice, ...)  METALANG99_PRIV_REC_NEXT(4905, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4905(choice, ...)  METALANG99_PRIV_REC_NEXT(4906, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4906(choice, ...)  METALANG99_PRIV_REC_NEXT(4907, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4907(choice, ...)  METALANG99_PRIV_REC_NEXT(4908, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4908(choice, ...)  METALANG99_PRIV_REC_NEXT(4909, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4909(choice, ...)  METALANG99_PRIV_REC_NEXT(4910, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4910(choice, ...)  METALANG99_PRIV_REC_NEXT(4911, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4911(choice, ...)  METALANG99_PRIV_REC_NEXT(4912, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4912(choice, ...)  METALANG99_PRIV_REC_NEXT(4913, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4913(choice, ...)  METALANG99_PRIV_REC_NEXT(4914, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4914(choice, ...)  METALANG99_PRIV_REC_NEXT(4915, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4915(choice, ...)  METALANG99_PRIV_REC_NEXT(4916, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4916(choice, ...)  METALANG99_PRIV_REC_NEXT(4917, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4917(choice, ...)  METALANG99_PRIV_REC_NEXT(4918, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4918(choice, ...)  METALANG99_PRIV_REC_NEXT(4919, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4919(choice, ...)  METALANG99_PRIV_REC_NEXT(4920, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4920(choice, ...)  METALANG99_PRIV_REC_NEXT(4921, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4921(choice, ...)  METALANG99_PRIV_REC_NEXT(4922, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4922(choice, ...)  METALANG99_PRIV_REC_NEXT(4923, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4923(choice, ...)  METALANG99_PRIV_REC_NEXT(4924, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4924(choice, ...)  METALANG99_PRIV_REC_NEXT(4925, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4925(choice, ...)  METALANG99_PRIV_REC_NEXT(4926, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4926(choice, ...)  METALANG99_PRIV_REC_NEXT(4927, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4927(choice, ...)  METALANG99_PRIV_REC_NEXT(4928, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4928(choice, ...)  METALANG99_PRIV_REC_NEXT(4929, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4929(choice, ...)  METALANG99_PRIV_REC_NEXT(4930, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4930(choice, ...)  METALANG99_PRIV_REC_NEXT(4931, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4931(choice, ...)  METALANG99_PRIV_REC_NEXT(4932, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4932(choice, ...)  METALANG99_PRIV_REC_NEXT(4933, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4933(choice, ...)  METALANG99_PRIV_REC_NEXT(4934, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4934(choice, ...)  METALANG99_PRIV_REC_NEXT(4935, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4935(choice, ...)  METALANG99_PRIV_REC_NEXT(4936, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4936(choice, ...)  METALANG99_PRIV_REC_NEXT(4937, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4937(choice, ...)  METALANG99_PRIV_REC_NEXT(4938, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4938(choice, ...)  METALANG99_PRIV_REC_NEXT(4939, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4939(choice, ...)  METALANG99_PRIV_REC_NEXT(4940, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4940(choice, ...)  METALANG99_PRIV_REC_NEXT(4941, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4941(choice, ...)  METALANG99_PRIV_REC_NEXT(4942, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4942(choice, ...)  METALANG99_PRIV_REC_NEXT(4943, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4943(choice, ...)  METALANG99_PRIV_REC_NEXT(4944, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4944(choice, ...)  METALANG99_PRIV_REC_NEXT(4945, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4945(choice, ...)  METALANG99_PRIV_REC_NEXT(4946, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4946(choice, ...)  METALANG99_PRIV_REC_NEXT(4947, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4947(choice, ...)  METALANG99_PRIV_REC_NEXT(4948, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4948(choice, ...)  METALANG99_PRIV_REC_NEXT(4949, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4949(choice, ...)  METALANG99_PRIV_REC_NEXT(4950, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4950(choice, ...)  METALANG99_PRIV_REC_NEXT(4951, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4951(choice, ...)  METALANG99_PRIV_REC_NEXT(4952, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4952(choice, ...)  METALANG99_PRIV_REC_NEXT(4953, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4953(choice, ...)  METALANG99_PRIV_REC_NEXT(4954, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4954(choice, ...)  METALANG99_PRIV_REC_NEXT(4955, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4955(choice, ...)  METALANG99_PRIV_REC_NEXT(4956, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4956(choice, ...)  METALANG99_PRIV_REC_NEXT(4957, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4957(choice, ...)  METALANG99_PRIV_REC_NEXT(4958, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4958(choice, ...)  METALANG99_PRIV_REC_NEXT(4959, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4959(choice, ...)  METALANG99_PRIV_REC_NEXT(4960, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4960(choice, ...)  METALANG99_PRIV_REC_NEXT(4961, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4961(choice, ...)  METALANG99_PRIV_REC_NEXT(4962, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4962(choice, ...)  METALANG99_PRIV_REC_NEXT(4963, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4963(choice, ...)  METALANG99_PRIV_REC_NEXT(4964, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4964(choice, ...)  METALANG99_PRIV_REC_NEXT(4965, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4965(choice, ...)  METALANG99_PRIV_REC_NEXT(4966, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4966(choice, ...)  METALANG99_PRIV_REC_NEXT(4967, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4967(choice, ...)  METALANG99_PRIV_REC_NEXT(4968, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4968(choice, ...)  METALANG99_PRIV_REC_NEXT(4969, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4969(choice, ...)  METALANG99_PRIV_REC_NEXT(4970, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4970(choice, ...)  METALANG99_PRIV_REC_NEXT(4971, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4971(choice, ...)  METALANG99_PRIV_REC_NEXT(4972, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4972(choice, ...)  METALANG99_PRIV_REC_NEXT(4973, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4973(choice, ...)  METALANG99_PRIV_REC_NEXT(4974, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4974(choice, ...)  METALANG99_PRIV_REC_NEXT(4975, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4975(choice, ...)  METALANG99_PRIV_REC_NEXT(4976, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4976(choice, ...)  METALANG99_PRIV_REC_NEXT(4977, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4977(choice, ...)  METALANG99_PRIV_REC_NEXT(4978, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4978(choice, ...)  METALANG99_PRIV_REC_NEXT(4979, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4979(choice, ...)  METALANG99_PRIV_REC_NEXT(4980, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4980(choice, ...)  METALANG99_PRIV_REC_NEXT(4981, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4981(choice, ...)  METALANG99_PRIV_REC_NEXT(4982, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4982(choice, ...)  METALANG99_PRIV_REC_NEXT(4983, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4983(choice, ...)  METALANG99_PRIV_REC_NEXT(4984, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4984(choice, ...)  METALANG99_PRIV_REC_NEXT(4985, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4985(choice, ...)  METALANG99_PRIV_REC_NEXT(4986, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4986(choice, ...)  METALANG99_PRIV_REC_NEXT(4987, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4987(choice, ...)  METALANG99_PRIV_REC_NEXT(4988, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4988(choice, ...)  METALANG99_PRIV_REC_NEXT(4989, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4989(choice, ...)  METALANG99_PRIV_REC_NEXT(4990, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4990(choice, ...)  METALANG99_PRIV_REC_NEXT(4991, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4991(choice, ...)  METALANG99_PRIV_REC_NEXT(4992, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4992(choice, ...)  METALANG99_PRIV_REC_NEXT(4993, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4993(choice, ...)  METALANG99_PRIV_REC_NEXT(4994, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4994(choice, ...)  METALANG99_PRIV_REC_NEXT(4995, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4995(choice, ...)  METALANG99_PRIV_REC_NEXT(4996, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4996(choice, ...)  METALANG99_PRIV_REC_NEXT(4997, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4997(choice, ...)  METALANG99_PRIV_REC_NEXT(4998, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4998(choice, ...)  METALANG99_PRIV_REC_NEXT(4999, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_4999(choice, ...)  METALANG99_PRIV_REC_NEXT(5000, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5000(choice, ...)  METALANG99_PRIV_REC_NEXT(5001, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5001(choice, ...)  METALANG99_PRIV_REC_NEXT(5002, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5002(choice, ...)  METALANG99_PRIV_REC_NEXT(5003, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5003(choice, ...)  METALANG99_PRIV_REC_NEXT(5004, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5004(choice, ...)  METALANG99_PRIV_REC_NEXT(5005, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5005(choice, ...)  METALANG99_PRIV_REC_NEXT(5006, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5006(choice, ...)  METALANG99_PRIV_REC_NEXT(5007, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5007(choice, ...)  METALANG99_PRIV_REC_NEXT(5008, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5008(choice, ...)  METALANG99_PRIV_REC_NEXT(5009, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5009(choice, ...)  METALANG99_PRIV_REC_NEXT(5010, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5010(choice, ...)  METALANG99_PRIV_REC_NEXT(5011, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5011(choice, ...)  METALANG99_PRIV_REC_NEXT(5012, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5012(choice, ...)  METALANG99_PRIV_REC_NEXT(5013, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5013(choice, ...)  METALANG99_PRIV_REC_NEXT(5014, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5014(choice, ...)  METALANG99_PRIV_REC_NEXT(5015, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5015(choice, ...)  METALANG99_PRIV_REC_NEXT(5016, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5016(choice, ...)  METALANG99_PRIV_REC_NEXT(5017, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5017(choice, ...)  METALANG99_PRIV_REC_NEXT(5018, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5018(choice, ...)  METALANG99_PRIV_REC_NEXT(5019, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5019(choice, ...)  METALANG99_PRIV_REC_NEXT(5020, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5020(choice, ...)  METALANG99_PRIV_REC_NEXT(5021, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5021(choice, ...)  METALANG99_PRIV_REC_NEXT(5022, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5022(choice, ...)  METALANG99_PRIV_REC_NEXT(5023, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5023(choice, ...)  METALANG99_PRIV_REC_NEXT(5024, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5024(choice, ...)  METALANG99_PRIV_REC_NEXT(5025, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5025(choice, ...)  METALANG99_PRIV_REC_NEXT(5026, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5026(choice, ...)  METALANG99_PRIV_REC_NEXT(5027, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5027(choice, ...)  METALANG99_PRIV_REC_NEXT(5028, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5028(choice, ...)  METALANG99_PRIV_REC_NEXT(5029, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5029(choice, ...)  METALANG99_PRIV_REC_NEXT(5030, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5030(choice, ...)  METALANG99_PRIV_REC_NEXT(5031, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5031(choice, ...)  METALANG99_PRIV_REC_NEXT(5032, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5032(choice, ...)  METALANG99_PRIV_REC_NEXT(5033, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5033(choice, ...)  METALANG99_PRIV_REC_NEXT(5034, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5034(choice, ...)  METALANG99_PRIV_REC_NEXT(5035, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5035(choice, ...)  METALANG99_PRIV_REC_NEXT(5036, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5036(choice, ...)  METALANG99_PRIV_REC_NEXT(5037, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5037(choice, ...)  METALANG99_PRIV_REC_NEXT(5038, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5038(choice, ...)  METALANG99_PRIV_REC_NEXT(5039, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5039(choice, ...)  METALANG99_PRIV_REC_NEXT(5040, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5040(choice, ...)  METALANG99_PRIV_REC_NEXT(5041, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5041(choice, ...)  METALANG99_PRIV_REC_NEXT(5042, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5042(choice, ...)  METALANG99_PRIV_REC_NEXT(5043, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5043(choice, ...)  METALANG99_PRIV_REC_NEXT(5044, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5044(choice, ...)  METALANG99_PRIV_REC_NEXT(5045, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5045(choice, ...)  METALANG99_PRIV_REC_NEXT(5046, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5046(choice, ...)  METALANG99_PRIV_REC_NEXT(5047, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5047(choice, ...)  METALANG99_PRIV_REC_NEXT(5048, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5048(choice, ...)  METALANG99_PRIV_REC_NEXT(5049, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5049(choice, ...)  METALANG99_PRIV_REC_NEXT(5050, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5050(choice, ...)  METALANG99_PRIV_REC_NEXT(5051, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5051(choice, ...)  METALANG99_PRIV_REC_NEXT(5052, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5052(choice, ...)  METALANG99_PRIV_REC_NEXT(5053, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5053(choice, ...)  METALANG99_PRIV_REC_NEXT(5054, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5054(choice, ...)  METALANG99_PRIV_REC_NEXT(5055, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5055(choice, ...)  METALANG99_PRIV_REC_NEXT(5056, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5056(choice, ...)  METALANG99_PRIV_REC_NEXT(5057, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5057(choice, ...)  METALANG99_PRIV_REC_NEXT(5058, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5058(choice, ...)  METALANG99_PRIV_REC_NEXT(5059, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5059(choice, ...)  METALANG99_PRIV_REC_NEXT(5060, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5060(choice, ...)  METALANG99_PRIV_REC_NEXT(5061, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5061(choice, ...)  METALANG99_PRIV_REC_NEXT(5062, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5062(choice, ...)  METALANG99_PRIV_REC_NEXT(5063, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5063(choice, ...)  METALANG99_PRIV_REC_NEXT(5064, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5064(choice, ...)  METALANG99_PRIV_REC_NEXT(5065, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5065(choice, ...)  METALANG99_PRIV_REC_NEXT(5066, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5066(choice, ...)  METALANG99_PRIV_REC_NEXT(5067, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5067(choice, ...)  METALANG99_PRIV_REC_NEXT(5068, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5068(choice, ...)  METALANG99_PRIV_REC_NEXT(5069, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5069(choice, ...)  METALANG99_PRIV_REC_NEXT(5070, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5070(choice, ...)  METALANG99_PRIV_REC_NEXT(5071, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5071(choice, ...)  METALANG99_PRIV_REC_NEXT(5072, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5072(choice, ...)  METALANG99_PRIV_REC_NEXT(5073, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5073(choice, ...)  METALANG99_PRIV_REC_NEXT(5074, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5074(choice, ...)  METALANG99_PRIV_REC_NEXT(5075, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5075(choice, ...)  METALANG99_PRIV_REC_NEXT(5076, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5076(choice, ...)  METALANG99_PRIV_REC_NEXT(5077, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5077(choice, ...)  METALANG99_PRIV_REC_NEXT(5078, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5078(choice, ...)  METALANG99_PRIV_REC_NEXT(5079, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5079(choice, ...)  METALANG99_PRIV_REC_NEXT(5080, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5080(choice, ...)  METALANG99_PRIV_REC_NEXT(5081, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5081(choice, ...)  METALANG99_PRIV_REC_NEXT(5082, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5082(choice, ...)  METALANG99_PRIV_REC_NEXT(5083, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5083(choice, ...)  METALANG99_PRIV_REC_NEXT(5084, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5084(choice, ...)  METALANG99_PRIV_REC_NEXT(5085, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5085(choice, ...)  METALANG99_PRIV_REC_NEXT(5086, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5086(choice, ...)  METALANG99_PRIV_REC_NEXT(5087, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5087(choice, ...)  METALANG99_PRIV_REC_NEXT(5088, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5088(choice, ...)  METALANG99_PRIV_REC_NEXT(5089, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5089(choice, ...)  METALANG99_PRIV_REC_NEXT(5090, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5090(choice, ...)  METALANG99_PRIV_REC_NEXT(5091, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5091(choice, ...)  METALANG99_PRIV_REC_NEXT(5092, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5092(choice, ...)  METALANG99_PRIV_REC_NEXT(5093, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5093(choice, ...)  METALANG99_PRIV_REC_NEXT(5094, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5094(choice, ...)  METALANG99_PRIV_REC_NEXT(5095, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5095(choice, ...)  METALANG99_PRIV_REC_NEXT(5096, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5096(choice, ...)  METALANG99_PRIV_REC_NEXT(5097, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5097(choice, ...)  METALANG99_PRIV_REC_NEXT(5098, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5098(choice, ...)  METALANG99_PRIV_REC_NEXT(5099, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5099(choice, ...)  METALANG99_PRIV_REC_NEXT(5100, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5100(choice, ...)  METALANG99_PRIV_REC_NEXT(5101, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5101(choice, ...)  METALANG99_PRIV_REC_NEXT(5102, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5102(choice, ...)  METALANG99_PRIV_REC_NEXT(5103, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5103(choice, ...)  METALANG99_PRIV_REC_NEXT(5104, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5104(choice, ...)  METALANG99_PRIV_REC_NEXT(5105, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5105(choice, ...)  METALANG99_PRIV_REC_NEXT(5106, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5106(choice, ...)  METALANG99_PRIV_REC_NEXT(5107, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5107(choice, ...)  METALANG99_PRIV_REC_NEXT(5108, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5108(choice, ...)  METALANG99_PRIV_REC_NEXT(5109, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5109(choice, ...)  METALANG99_PRIV_REC_NEXT(5110, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5110(choice, ...)  METALANG99_PRIV_REC_NEXT(5111, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5111(choice, ...)  METALANG99_PRIV_REC_NEXT(5112, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5112(choice, ...)  METALANG99_PRIV_REC_NEXT(5113, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5113(choice, ...)  METALANG99_PRIV_REC_NEXT(5114, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5114(choice, ...)  METALANG99_PRIV_REC_NEXT(5115, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5115(choice, ...)  METALANG99_PRIV_REC_NEXT(5116, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5116(choice, ...)  METALANG99_PRIV_REC_NEXT(5117, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5117(choice, ...)  METALANG99_PRIV_REC_NEXT(5118, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5118(choice, ...)  METALANG99_PRIV_REC_NEXT(5119, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5119(choice, ...)  METALANG99_PRIV_REC_NEXT(5120, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5120(choice, ...)  METALANG99_PRIV_REC_NEXT(5121, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5121(choice, ...)  METALANG99_PRIV_REC_NEXT(5122, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5122(choice, ...)  METALANG99_PRIV_REC_NEXT(5123, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5123(choice, ...)  METALANG99_PRIV_REC_NEXT(5124, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5124(choice, ...)  METALANG99_PRIV_REC_NEXT(5125, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5125(choice, ...)  METALANG99_PRIV_REC_NEXT(5126, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5126(choice, ...)  METALANG99_PRIV_REC_NEXT(5127, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5127(choice, ...)  METALANG99_PRIV_REC_NEXT(5128, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5128(choice, ...)  METALANG99_PRIV_REC_NEXT(5129, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5129(choice, ...)  METALANG99_PRIV_REC_NEXT(5130, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5130(choice, ...)  METALANG99_PRIV_REC_NEXT(5131, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5131(choice, ...)  METALANG99_PRIV_REC_NEXT(5132, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5132(choice, ...)  METALANG99_PRIV_REC_NEXT(5133, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5133(choice, ...)  METALANG99_PRIV_REC_NEXT(5134, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5134(choice, ...)  METALANG99_PRIV_REC_NEXT(5135, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5135(choice, ...)  METALANG99_PRIV_REC_NEXT(5136, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5136(choice, ...)  METALANG99_PRIV_REC_NEXT(5137, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5137(choice, ...)  METALANG99_PRIV_REC_NEXT(5138, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5138(choice, ...)  METALANG99_PRIV_REC_NEXT(5139, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5139(choice, ...)  METALANG99_PRIV_REC_NEXT(5140, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5140(choice, ...)  METALANG99_PRIV_REC_NEXT(5141, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5141(choice, ...)  METALANG99_PRIV_REC_NEXT(5142, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5142(choice, ...)  METALANG99_PRIV_REC_NEXT(5143, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5143(choice, ...)  METALANG99_PRIV_REC_NEXT(5144, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5144(choice, ...)  METALANG99_PRIV_REC_NEXT(5145, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5145(choice, ...)  METALANG99_PRIV_REC_NEXT(5146, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5146(choice, ...)  METALANG99_PRIV_REC_NEXT(5147, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5147(choice, ...)  METALANG99_PRIV_REC_NEXT(5148, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5148(choice, ...)  METALANG99_PRIV_REC_NEXT(5149, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5149(choice, ...)  METALANG99_PRIV_REC_NEXT(5150, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5150(choice, ...)  METALANG99_PRIV_REC_NEXT(5151, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5151(choice, ...)  METALANG99_PRIV_REC_NEXT(5152, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5152(choice, ...)  METALANG99_PRIV_REC_NEXT(5153, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5153(choice, ...)  METALANG99_PRIV_REC_NEXT(5154, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5154(choice, ...)  METALANG99_PRIV_REC_NEXT(5155, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5155(choice, ...)  METALANG99_PRIV_REC_NEXT(5156, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5156(choice, ...)  METALANG99_PRIV_REC_NEXT(5157, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5157(choice, ...)  METALANG99_PRIV_REC_NEXT(5158, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5158(choice, ...)  METALANG99_PRIV_REC_NEXT(5159, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5159(choice, ...)  METALANG99_PRIV_REC_NEXT(5160, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5160(choice, ...)  METALANG99_PRIV_REC_NEXT(5161, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5161(choice, ...)  METALANG99_PRIV_REC_NEXT(5162, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5162(choice, ...)  METALANG99_PRIV_REC_NEXT(5163, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5163(choice, ...)  METALANG99_PRIV_REC_NEXT(5164, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5164(choice, ...)  METALANG99_PRIV_REC_NEXT(5165, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5165(choice, ...)  METALANG99_PRIV_REC_NEXT(5166, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5166(choice, ...)  METALANG99_PRIV_REC_NEXT(5167, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5167(choice, ...)  METALANG99_PRIV_REC_NEXT(5168, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5168(choice, ...)  METALANG99_PRIV_REC_NEXT(5169, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5169(choice, ...)  METALANG99_PRIV_REC_NEXT(5170, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5170(choice, ...)  METALANG99_PRIV_REC_NEXT(5171, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5171(choice, ...)  METALANG99_PRIV_REC_NEXT(5172, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5172(choice, ...)  METALANG99_PRIV_REC_NEXT(5173, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5173(choice, ...)  METALANG99_PRIV_REC_NEXT(5174, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5174(choice, ...)  METALANG99_PRIV_REC_NEXT(5175, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5175(choice, ...)  METALANG99_PRIV_REC_NEXT(5176, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5176(choice, ...)  METALANG99_PRIV_REC_NEXT(5177, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5177(choice, ...)  METALANG99_PRIV_REC_NEXT(5178, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5178(choice, ...)  METALANG99_PRIV_REC_NEXT(5179, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5179(choice, ...)  METALANG99_PRIV_REC_NEXT(5180, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5180(choice, ...)  METALANG99_PRIV_REC_NEXT(5181, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5181(choice, ...)  METALANG99_PRIV_REC_NEXT(5182, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5182(choice, ...)  METALANG99_PRIV_REC_NEXT(5183, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5183(choice, ...)  METALANG99_PRIV_REC_NEXT(5184, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5184(choice, ...)  METALANG99_PRIV_REC_NEXT(5185, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5185(choice, ...)  METALANG99_PRIV_REC_NEXT(5186, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5186(choice, ...)  METALANG99_PRIV_REC_NEXT(5187, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5187(choice, ...)  METALANG99_PRIV_REC_NEXT(5188, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5188(choice, ...)  METALANG99_PRIV_REC_NEXT(5189, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5189(choice, ...)  METALANG99_PRIV_REC_NEXT(5190, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5190(choice, ...)  METALANG99_PRIV_REC_NEXT(5191, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5191(choice, ...)  METALANG99_PRIV_REC_NEXT(5192, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5192(choice, ...)  METALANG99_PRIV_REC_NEXT(5193, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5193(choice, ...)  METALANG99_PRIV_REC_NEXT(5194, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5194(choice, ...)  METALANG99_PRIV_REC_NEXT(5195, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5195(choice, ...)  METALANG99_PRIV_REC_NEXT(5196, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5196(choice, ...)  METALANG99_PRIV_REC_NEXT(5197, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5197(choice, ...)  METALANG99_PRIV_REC_NEXT(5198, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5198(choice, ...)  METALANG99_PRIV_REC_NEXT(5199, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5199(choice, ...)  METALANG99_PRIV_REC_NEXT(5200, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5200(choice, ...)  METALANG99_PRIV_REC_NEXT(5201, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5201(choice, ...)  METALANG99_PRIV_REC_NEXT(5202, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5202(choice, ...)  METALANG99_PRIV_REC_NEXT(5203, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5203(choice, ...)  METALANG99_PRIV_REC_NEXT(5204, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5204(choice, ...)  METALANG99_PRIV_REC_NEXT(5205, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5205(choice, ...)  METALANG99_PRIV_REC_NEXT(5206, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5206(choice, ...)  METALANG99_PRIV_REC_NEXT(5207, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5207(choice, ...)  METALANG99_PRIV_REC_NEXT(5208, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5208(choice, ...)  METALANG99_PRIV_REC_NEXT(5209, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5209(choice, ...)  METALANG99_PRIV_REC_NEXT(5210, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5210(choice, ...)  METALANG99_PRIV_REC_NEXT(5211, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5211(choice, ...)  METALANG99_PRIV_REC_NEXT(5212, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5212(choice, ...)  METALANG99_PRIV_REC_NEXT(5213, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5213(choice, ...)  METALANG99_PRIV_REC_NEXT(5214, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5214(choice, ...)  METALANG99_PRIV_REC_NEXT(5215, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5215(choice, ...)  METALANG99_PRIV_REC_NEXT(5216, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5216(choice, ...)  METALANG99_PRIV_REC_NEXT(5217, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5217(choice, ...)  METALANG99_PRIV_REC_NEXT(5218, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5218(choice, ...)  METALANG99_PRIV_REC_NEXT(5219, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5219(choice, ...)  METALANG99_PRIV_REC_NEXT(5220, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5220(choice, ...)  METALANG99_PRIV_REC_NEXT(5221, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5221(choice, ...)  METALANG99_PRIV_REC_NEXT(5222, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5222(choice, ...)  METALANG99_PRIV_REC_NEXT(5223, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5223(choice, ...)  METALANG99_PRIV_REC_NEXT(5224, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5224(choice, ...)  METALANG99_PRIV_REC_NEXT(5225, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5225(choice, ...)  METALANG99_PRIV_REC_NEXT(5226, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5226(choice, ...)  METALANG99_PRIV_REC_NEXT(5227, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5227(choice, ...)  METALANG99_PRIV_REC_NEXT(5228, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5228(choice, ...)  METALANG99_PRIV_REC_NEXT(5229, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5229(choice, ...)  METALANG99_PRIV_REC_NEXT(5230, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5230(choice, ...)  METALANG99_PRIV_REC_NEXT(5231, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5231(choice, ...)  METALANG99_PRIV_REC_NEXT(5232, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5232(choice, ...)  METALANG99_PRIV_REC_NEXT(5233, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5233(choice, ...)  METALANG99_PRIV_REC_NEXT(5234, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5234(choice, ...)  METALANG99_PRIV_REC_NEXT(5235, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5235(choice, ...)  METALANG99_PRIV_REC_NEXT(5236, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5236(choice, ...)  METALANG99_PRIV_REC_NEXT(5237, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5237(choice, ...)  METALANG99_PRIV_REC_NEXT(5238, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5238(choice, ...)  METALANG99_PRIV_REC_NEXT(5239, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5239(choice, ...)  METALANG99_PRIV_REC_NEXT(5240, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5240(choice, ...)  METALANG99_PRIV_REC_NEXT(5241, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5241(choice, ...)  METALANG99_PRIV_REC_NEXT(5242, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5242(choice, ...)  METALANG99_PRIV_REC_NEXT(5243, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5243(choice, ...)  METALANG99_PRIV_REC_NEXT(5244, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5244(choice, ...)  METALANG99_PRIV_REC_NEXT(5245, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5245(choice, ...)  METALANG99_PRIV_REC_NEXT(5246, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5246(choice, ...)  METALANG99_PRIV_REC_NEXT(5247, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5247(choice, ...)  METALANG99_PRIV_REC_NEXT(5248, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5248(choice, ...)  METALANG99_PRIV_REC_NEXT(5249, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5249(choice, ...)  METALANG99_PRIV_REC_NEXT(5250, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5250(choice, ...)  METALANG99_PRIV_REC_NEXT(5251, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5251(choice, ...)  METALANG99_PRIV_REC_NEXT(5252, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5252(choice, ...)  METALANG99_PRIV_REC_NEXT(5253, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5253(choice, ...)  METALANG99_PRIV_REC_NEXT(5254, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5254(choice, ...)  METALANG99_PRIV_REC_NEXT(5255, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5255(choice, ...)  METALANG99_PRIV_REC_NEXT(5256, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5256(choice, ...)  METALANG99_PRIV_REC_NEXT(5257, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5257(choice, ...)  METALANG99_PRIV_REC_NEXT(5258, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5258(choice, ...)  METALANG99_PRIV_REC_NEXT(5259, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5259(choice, ...)  METALANG99_PRIV_REC_NEXT(5260, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5260(choice, ...)  METALANG99_PRIV_REC_NEXT(5261, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5261(choice, ...)  METALANG99_PRIV_REC_NEXT(5262, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5262(choice, ...)  METALANG99_PRIV_REC_NEXT(5263, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5263(choice, ...)  METALANG99_PRIV_REC_NEXT(5264, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5264(choice, ...)  METALANG99_PRIV_REC_NEXT(5265, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5265(choice, ...)  METALANG99_PRIV_REC_NEXT(5266, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5266(choice, ...)  METALANG99_PRIV_REC_NEXT(5267, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5267(choice, ...)  METALANG99_PRIV_REC_NEXT(5268, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5268(choice, ...)  METALANG99_PRIV_REC_NEXT(5269, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5269(choice, ...)  METALANG99_PRIV_REC_NEXT(5270, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5270(choice, ...)  METALANG99_PRIV_REC_NEXT(5271, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5271(choice, ...)  METALANG99_PRIV_REC_NEXT(5272, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5272(choice, ...)  METALANG99_PRIV_REC_NEXT(5273, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5273(choice, ...)  METALANG99_PRIV_REC_NEXT(5274, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5274(choice, ...)  METALANG99_PRIV_REC_NEXT(5275, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5275(choice, ...)  METALANG99_PRIV_REC_NEXT(5276, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5276(choice, ...)  METALANG99_PRIV_REC_NEXT(5277, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5277(choice, ...)  METALANG99_PRIV_REC_NEXT(5278, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5278(choice, ...)  METALANG99_PRIV_REC_NEXT(5279, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5279(choice, ...)  METALANG99_PRIV_REC_NEXT(5280, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5280(choice, ...)  METALANG99_PRIV_REC_NEXT(5281, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5281(choice, ...)  METALANG99_PRIV_REC_NEXT(5282, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5282(choice, ...)  METALANG99_PRIV_REC_NEXT(5283, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5283(choice, ...)  METALANG99_PRIV_REC_NEXT(5284, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5284(choice, ...)  METALANG99_PRIV_REC_NEXT(5285, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5285(choice, ...)  METALANG99_PRIV_REC_NEXT(5286, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5286(choice, ...)  METALANG99_PRIV_REC_NEXT(5287, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5287(choice, ...)  METALANG99_PRIV_REC_NEXT(5288, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5288(choice, ...)  METALANG99_PRIV_REC_NEXT(5289, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5289(choice, ...)  METALANG99_PRIV_REC_NEXT(5290, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5290(choice, ...)  METALANG99_PRIV_REC_NEXT(5291, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5291(choice, ...)  METALANG99_PRIV_REC_NEXT(5292, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5292(choice, ...)  METALANG99_PRIV_REC_NEXT(5293, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5293(choice, ...)  METALANG99_PRIV_REC_NEXT(5294, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5294(choice, ...)  METALANG99_PRIV_REC_NEXT(5295, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5295(choice, ...)  METALANG99_PRIV_REC_NEXT(5296, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5296(choice, ...)  METALANG99_PRIV_REC_NEXT(5297, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5297(choice, ...)  METALANG99_PRIV_REC_NEXT(5298, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5298(choice, ...)  METALANG99_PRIV_REC_NEXT(5299, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5299(choice, ...)  METALANG99_PRIV_REC_NEXT(5300, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5300(choice, ...)  METALANG99_PRIV_REC_NEXT(5301, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5301(choice, ...)  METALANG99_PRIV_REC_NEXT(5302, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5302(choice, ...)  METALANG99_PRIV_REC_NEXT(5303, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5303(choice, ...)  METALANG99_PRIV_REC_NEXT(5304, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5304(choice, ...)  METALANG99_PRIV_REC_NEXT(5305, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5305(choice, ...)  METALANG99_PRIV_REC_NEXT(5306, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5306(choice, ...)  METALANG99_PRIV_REC_NEXT(5307, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5307(choice, ...)  METALANG99_PRIV_REC_NEXT(5308, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5308(choice, ...)  METALANG99_PRIV_REC_NEXT(5309, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5309(choice, ...)  METALANG99_PRIV_REC_NEXT(5310, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5310(choice, ...)  METALANG99_PRIV_REC_NEXT(5311, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5311(choice, ...)  METALANG99_PRIV_REC_NEXT(5312, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5312(choice, ...)  METALANG99_PRIV_REC_NEXT(5313, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5313(choice, ...)  METALANG99_PRIV_REC_NEXT(5314, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5314(choice, ...)  METALANG99_PRIV_REC_NEXT(5315, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5315(choice, ...)  METALANG99_PRIV_REC_NEXT(5316, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5316(choice, ...)  METALANG99_PRIV_REC_NEXT(5317, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5317(choice, ...)  METALANG99_PRIV_REC_NEXT(5318, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5318(choice, ...)  METALANG99_PRIV_REC_NEXT(5319, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5319(choice, ...)  METALANG99_PRIV_REC_NEXT(5320, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5320(choice, ...)  METALANG99_PRIV_REC_NEXT(5321, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5321(choice, ...)  METALANG99_PRIV_REC_NEXT(5322, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5322(choice, ...)  METALANG99_PRIV_REC_NEXT(5323, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5323(choice, ...)  METALANG99_PRIV_REC_NEXT(5324, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5324(choice, ...)  METALANG99_PRIV_REC_NEXT(5325, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5325(choice, ...)  METALANG99_PRIV_REC_NEXT(5326, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5326(choice, ...)  METALANG99_PRIV_REC_NEXT(5327, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5327(choice, ...)  METALANG99_PRIV_REC_NEXT(5328, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5328(choice, ...)  METALANG99_PRIV_REC_NEXT(5329, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5329(choice, ...)  METALANG99_PRIV_REC_NEXT(5330, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5330(choice, ...)  METALANG99_PRIV_REC_NEXT(5331, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5331(choice, ...)  METALANG99_PRIV_REC_NEXT(5332, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5332(choice, ...)  METALANG99_PRIV_REC_NEXT(5333, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5333(choice, ...)  METALANG99_PRIV_REC_NEXT(5334, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5334(choice, ...)  METALANG99_PRIV_REC_NEXT(5335, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5335(choice, ...)  METALANG99_PRIV_REC_NEXT(5336, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5336(choice, ...)  METALANG99_PRIV_REC_NEXT(5337, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5337(choice, ...)  METALANG99_PRIV_REC_NEXT(5338, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5338(choice, ...)  METALANG99_PRIV_REC_NEXT(5339, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5339(choice, ...)  METALANG99_PRIV_REC_NEXT(5340, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5340(choice, ...)  METALANG99_PRIV_REC_NEXT(5341, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5341(choice, ...)  METALANG99_PRIV_REC_NEXT(5342, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5342(choice, ...)  METALANG99_PRIV_REC_NEXT(5343, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5343(choice, ...)  METALANG99_PRIV_REC_NEXT(5344, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5344(choice, ...)  METALANG99_PRIV_REC_NEXT(5345, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5345(choice, ...)  METALANG99_PRIV_REC_NEXT(5346, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5346(choice, ...)  METALANG99_PRIV_REC_NEXT(5347, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5347(choice, ...)  METALANG99_PRIV_REC_NEXT(5348, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5348(choice, ...)  METALANG99_PRIV_REC_NEXT(5349, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5349(choice, ...)  METALANG99_PRIV_REC_NEXT(5350, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5350(choice, ...)  METALANG99_PRIV_REC_NEXT(5351, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5351(choice, ...)  METALANG99_PRIV_REC_NEXT(5352, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5352(choice, ...)  METALANG99_PRIV_REC_NEXT(5353, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5353(choice, ...)  METALANG99_PRIV_REC_NEXT(5354, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5354(choice, ...)  METALANG99_PRIV_REC_NEXT(5355, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5355(choice, ...)  METALANG99_PRIV_REC_NEXT(5356, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5356(choice, ...)  METALANG99_PRIV_REC_NEXT(5357, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5357(choice, ...)  METALANG99_PRIV_REC_NEXT(5358, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5358(choice, ...)  METALANG99_PRIV_REC_NEXT(5359, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5359(choice, ...)  METALANG99_PRIV_REC_NEXT(5360, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5360(choice, ...)  METALANG99_PRIV_REC_NEXT(5361, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5361(choice, ...)  METALANG99_PRIV_REC_NEXT(5362, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5362(choice, ...)  METALANG99_PRIV_REC_NEXT(5363, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5363(choice, ...)  METALANG99_PRIV_REC_NEXT(5364, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5364(choice, ...)  METALANG99_PRIV_REC_NEXT(5365, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5365(choice, ...)  METALANG99_PRIV_REC_NEXT(5366, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5366(choice, ...)  METALANG99_PRIV_REC_NEXT(5367, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5367(choice, ...)  METALANG99_PRIV_REC_NEXT(5368, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5368(choice, ...)  METALANG99_PRIV_REC_NEXT(5369, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5369(choice, ...)  METALANG99_PRIV_REC_NEXT(5370, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5370(choice, ...)  METALANG99_PRIV_REC_NEXT(5371, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5371(choice, ...)  METALANG99_PRIV_REC_NEXT(5372, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5372(choice, ...)  METALANG99_PRIV_REC_NEXT(5373, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5373(choice, ...)  METALANG99_PRIV_REC_NEXT(5374, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5374(choice, ...)  METALANG99_PRIV_REC_NEXT(5375, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5375(choice, ...)  METALANG99_PRIV_REC_NEXT(5376, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5376(choice, ...)  METALANG99_PRIV_REC_NEXT(5377, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5377(choice, ...)  METALANG99_PRIV_REC_NEXT(5378, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5378(choice, ...)  METALANG99_PRIV_REC_NEXT(5379, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5379(choice, ...)  METALANG99_PRIV_REC_NEXT(5380, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5380(choice, ...)  METALANG99_PRIV_REC_NEXT(5381, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5381(choice, ...)  METALANG99_PRIV_REC_NEXT(5382, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5382(choice, ...)  METALANG99_PRIV_REC_NEXT(5383, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5383(choice, ...)  METALANG99_PRIV_REC_NEXT(5384, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5384(choice, ...)  METALANG99_PRIV_REC_NEXT(5385, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5385(choice, ...)  METALANG99_PRIV_REC_NEXT(5386, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5386(choice, ...)  METALANG99_PRIV_REC_NEXT(5387, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5387(choice, ...)  METALANG99_PRIV_REC_NEXT(5388, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5388(choice, ...)  METALANG99_PRIV_REC_NEXT(5389, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5389(choice, ...)  METALANG99_PRIV_REC_NEXT(5390, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5390(choice, ...)  METALANG99_PRIV_REC_NEXT(5391, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5391(choice, ...)  METALANG99_PRIV_REC_NEXT(5392, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5392(choice, ...)  METALANG99_PRIV_REC_NEXT(5393, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5393(choice, ...)  METALANG99_PRIV_REC_NEXT(5394, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5394(choice, ...)  METALANG99_PRIV_REC_NEXT(5395, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5395(choice, ...)  METALANG99_PRIV_REC_NEXT(5396, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5396(choice, ...)  METALANG99_PRIV_REC_NEXT(5397, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5397(choice, ...)  METALANG99_PRIV_REC_NEXT(5398, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5398(choice, ...)  METALANG99_PRIV_REC_NEXT(5399, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5399(choice, ...)  METALANG99_PRIV_REC_NEXT(5400, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5400(choice, ...)  METALANG99_PRIV_REC_NEXT(5401, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5401(choice, ...)  METALANG99_PRIV_REC_NEXT(5402, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5402(choice, ...)  METALANG99_PRIV_REC_NEXT(5403, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5403(choice, ...)  METALANG99_PRIV_REC_NEXT(5404, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5404(choice, ...)  METALANG99_PRIV_REC_NEXT(5405, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5405(choice, ...)  METALANG99_PRIV_REC_NEXT(5406, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5406(choice, ...)  METALANG99_PRIV_REC_NEXT(5407, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5407(choice, ...)  METALANG99_PRIV_REC_NEXT(5408, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5408(choice, ...)  METALANG99_PRIV_REC_NEXT(5409, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5409(choice, ...)  METALANG99_PRIV_REC_NEXT(5410, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5410(choice, ...)  METALANG99_PRIV_REC_NEXT(5411, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5411(choice, ...)  METALANG99_PRIV_REC_NEXT(5412, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5412(choice, ...)  METALANG99_PRIV_REC_NEXT(5413, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5413(choice, ...)  METALANG99_PRIV_REC_NEXT(5414, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5414(choice, ...)  METALANG99_PRIV_REC_NEXT(5415, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5415(choice, ...)  METALANG99_PRIV_REC_NEXT(5416, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5416(choice, ...)  METALANG99_PRIV_REC_NEXT(5417, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5417(choice, ...)  METALANG99_PRIV_REC_NEXT(5418, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5418(choice, ...)  METALANG99_PRIV_REC_NEXT(5419, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5419(choice, ...)  METALANG99_PRIV_REC_NEXT(5420, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5420(choice, ...)  METALANG99_PRIV_REC_NEXT(5421, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5421(choice, ...)  METALANG99_PRIV_REC_NEXT(5422, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5422(choice, ...)  METALANG99_PRIV_REC_NEXT(5423, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5423(choice, ...)  METALANG99_PRIV_REC_NEXT(5424, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5424(choice, ...)  METALANG99_PRIV_REC_NEXT(5425, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5425(choice, ...)  METALANG99_PRIV_REC_NEXT(5426, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5426(choice, ...)  METALANG99_PRIV_REC_NEXT(5427, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5427(choice, ...)  METALANG99_PRIV_REC_NEXT(5428, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5428(choice, ...)  METALANG99_PRIV_REC_NEXT(5429, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5429(choice, ...)  METALANG99_PRIV_REC_NEXT(5430, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5430(choice, ...)  METALANG99_PRIV_REC_NEXT(5431, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5431(choice, ...)  METALANG99_PRIV_REC_NEXT(5432, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5432(choice, ...)  METALANG99_PRIV_REC_NEXT(5433, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5433(choice, ...)  METALANG99_PRIV_REC_NEXT(5434, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5434(choice, ...)  METALANG99_PRIV_REC_NEXT(5435, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5435(choice, ...)  METALANG99_PRIV_REC_NEXT(5436, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5436(choice, ...)  METALANG99_PRIV_REC_NEXT(5437, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5437(choice, ...)  METALANG99_PRIV_REC_NEXT(5438, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5438(choice, ...)  METALANG99_PRIV_REC_NEXT(5439, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5439(choice, ...)  METALANG99_PRIV_REC_NEXT(5440, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5440(choice, ...)  METALANG99_PRIV_REC_NEXT(5441, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5441(choice, ...)  METALANG99_PRIV_REC_NEXT(5442, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5442(choice, ...)  METALANG99_PRIV_REC_NEXT(5443, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5443(choice, ...)  METALANG99_PRIV_REC_NEXT(5444, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5444(choice, ...)  METALANG99_PRIV_REC_NEXT(5445, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5445(choice, ...)  METALANG99_PRIV_REC_NEXT(5446, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5446(choice, ...)  METALANG99_PRIV_REC_NEXT(5447, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5447(choice, ...)  METALANG99_PRIV_REC_NEXT(5448, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5448(choice, ...)  METALANG99_PRIV_REC_NEXT(5449, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5449(choice, ...)  METALANG99_PRIV_REC_NEXT(5450, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5450(choice, ...)  METALANG99_PRIV_REC_NEXT(5451, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5451(choice, ...)  METALANG99_PRIV_REC_NEXT(5452, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5452(choice, ...)  METALANG99_PRIV_REC_NEXT(5453, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5453(choice, ...)  METALANG99_PRIV_REC_NEXT(5454, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5454(choice, ...)  METALANG99_PRIV_REC_NEXT(5455, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5455(choice, ...)  METALANG99_PRIV_REC_NEXT(5456, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5456(choice, ...)  METALANG99_PRIV_REC_NEXT(5457, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5457(choice, ...)  METALANG99_PRIV_REC_NEXT(5458, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5458(choice, ...)  METALANG99_PRIV_REC_NEXT(5459, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5459(choice, ...)  METALANG99_PRIV_REC_NEXT(5460, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5460(choice, ...)  METALANG99_PRIV_REC_NEXT(5461, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5461(choice, ...)  METALANG99_PRIV_REC_NEXT(5462, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5462(choice, ...)  METALANG99_PRIV_REC_NEXT(5463, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5463(choice, ...)  METALANG99_PRIV_REC_NEXT(5464, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5464(choice, ...)  METALANG99_PRIV_REC_NEXT(5465, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5465(choice, ...)  METALANG99_PRIV_REC_NEXT(5466, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5466(choice, ...)  METALANG99_PRIV_REC_NEXT(5467, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5467(choice, ...)  METALANG99_PRIV_REC_NEXT(5468, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5468(choice, ...)  METALANG99_PRIV_REC_NEXT(5469, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5469(choice, ...)  METALANG99_PRIV_REC_NEXT(5470, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5470(choice, ...)  METALANG99_PRIV_REC_NEXT(5471, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5471(choice, ...)  METALANG99_PRIV_REC_NEXT(5472, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5472(choice, ...)  METALANG99_PRIV_REC_NEXT(5473, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5473(choice, ...)  METALANG99_PRIV_REC_NEXT(5474, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5474(choice, ...)  METALANG99_PRIV_REC_NEXT(5475, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5475(choice, ...)  METALANG99_PRIV_REC_NEXT(5476, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5476(choice, ...)  METALANG99_PRIV_REC_NEXT(5477, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5477(choice, ...)  METALANG99_PRIV_REC_NEXT(5478, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5478(choice, ...)  METALANG99_PRIV_REC_NEXT(5479, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5479(choice, ...)  METALANG99_PRIV_REC_NEXT(5480, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5480(choice, ...)  METALANG99_PRIV_REC_NEXT(5481, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5481(choice, ...)  METALANG99_PRIV_REC_NEXT(5482, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5482(choice, ...)  METALANG99_PRIV_REC_NEXT(5483, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5483(choice, ...)  METALANG99_PRIV_REC_NEXT(5484, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5484(choice, ...)  METALANG99_PRIV_REC_NEXT(5485, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5485(choice, ...)  METALANG99_PRIV_REC_NEXT(5486, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5486(choice, ...)  METALANG99_PRIV_REC_NEXT(5487, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5487(choice, ...)  METALANG99_PRIV_REC_NEXT(5488, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5488(choice, ...)  METALANG99_PRIV_REC_NEXT(5489, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5489(choice, ...)  METALANG99_PRIV_REC_NEXT(5490, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5490(choice, ...)  METALANG99_PRIV_REC_NEXT(5491, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5491(choice, ...)  METALANG99_PRIV_REC_NEXT(5492, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5492(choice, ...)  METALANG99_PRIV_REC_NEXT(5493, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5493(choice, ...)  METALANG99_PRIV_REC_NEXT(5494, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5494(choice, ...)  METALANG99_PRIV_REC_NEXT(5495, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5495(choice, ...)  METALANG99_PRIV_REC_NEXT(5496, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5496(choice, ...)  METALANG99_PRIV_REC_NEXT(5497, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5497(choice, ...)  METALANG99_PRIV_REC_NEXT(5498, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5498(choice, ...)  METALANG99_PRIV_REC_NEXT(5499, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5499(choice, ...)  METALANG99_PRIV_REC_NEXT(5500, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5500(choice, ...)  METALANG99_PRIV_REC_NEXT(5501, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5501(choice, ...)  METALANG99_PRIV_REC_NEXT(5502, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5502(choice, ...)  METALANG99_PRIV_REC_NEXT(5503, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5503(choice, ...)  METALANG99_PRIV_REC_NEXT(5504, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5504(choice, ...)  METALANG99_PRIV_REC_NEXT(5505, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5505(choice, ...)  METALANG99_PRIV_REC_NEXT(5506, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5506(choice, ...)  METALANG99_PRIV_REC_NEXT(5507, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5507(choice, ...)  METALANG99_PRIV_REC_NEXT(5508, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5508(choice, ...)  METALANG99_PRIV_REC_NEXT(5509, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5509(choice, ...)  METALANG99_PRIV_REC_NEXT(5510, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5510(choice, ...)  METALANG99_PRIV_REC_NEXT(5511, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5511(choice, ...)  METALANG99_PRIV_REC_NEXT(5512, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5512(choice, ...)  METALANG99_PRIV_REC_NEXT(5513, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5513(choice, ...)  METALANG99_PRIV_REC_NEXT(5514, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5514(choice, ...)  METALANG99_PRIV_REC_NEXT(5515, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5515(choice, ...)  METALANG99_PRIV_REC_NEXT(5516, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5516(choice, ...)  METALANG99_PRIV_REC_NEXT(5517, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5517(choice, ...)  METALANG99_PRIV_REC_NEXT(5518, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5518(choice, ...)  METALANG99_PRIV_REC_NEXT(5519, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5519(choice, ...)  METALANG99_PRIV_REC_NEXT(5520, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5520(choice, ...)  METALANG99_PRIV_REC_NEXT(5521, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5521(choice, ...)  METALANG99_PRIV_REC_NEXT(5522, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5522(choice, ...)  METALANG99_PRIV_REC_NEXT(5523, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5523(choice, ...)  METALANG99_PRIV_REC_NEXT(5524, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5524(choice, ...)  METALANG99_PRIV_REC_NEXT(5525, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5525(choice, ...)  METALANG99_PRIV_REC_NEXT(5526, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5526(choice, ...)  METALANG99_PRIV_REC_NEXT(5527, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5527(choice, ...)  METALANG99_PRIV_REC_NEXT(5528, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5528(choice, ...)  METALANG99_PRIV_REC_NEXT(5529, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5529(choice, ...)  METALANG99_PRIV_REC_NEXT(5530, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5530(choice, ...)  METALANG99_PRIV_REC_NEXT(5531, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5531(choice, ...)  METALANG99_PRIV_REC_NEXT(5532, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5532(choice, ...)  METALANG99_PRIV_REC_NEXT(5533, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5533(choice, ...)  METALANG99_PRIV_REC_NEXT(5534, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5534(choice, ...)  METALANG99_PRIV_REC_NEXT(5535, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5535(choice, ...)  METALANG99_PRIV_REC_NEXT(5536, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5536(choice, ...)  METALANG99_PRIV_REC_NEXT(5537, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5537(choice, ...)  METALANG99_PRIV_REC_NEXT(5538, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5538(choice, ...)  METALANG99_PRIV_REC_NEXT(5539, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5539(choice, ...)  METALANG99_PRIV_REC_NEXT(5540, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5540(choice, ...)  METALANG99_PRIV_REC_NEXT(5541, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5541(choice, ...)  METALANG99_PRIV_REC_NEXT(5542, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5542(choice, ...)  METALANG99_PRIV_REC_NEXT(5543, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5543(choice, ...)  METALANG99_PRIV_REC_NEXT(5544, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5544(choice, ...)  METALANG99_PRIV_REC_NEXT(5545, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5545(choice, ...)  METALANG99_PRIV_REC_NEXT(5546, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5546(choice, ...)  METALANG99_PRIV_REC_NEXT(5547, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5547(choice, ...)  METALANG99_PRIV_REC_NEXT(5548, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5548(choice, ...)  METALANG99_PRIV_REC_NEXT(5549, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5549(choice, ...)  METALANG99_PRIV_REC_NEXT(5550, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5550(choice, ...)  METALANG99_PRIV_REC_NEXT(5551, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5551(choice, ...)  METALANG99_PRIV_REC_NEXT(5552, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5552(choice, ...)  METALANG99_PRIV_REC_NEXT(5553, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5553(choice, ...)  METALANG99_PRIV_REC_NEXT(5554, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5554(choice, ...)  METALANG99_PRIV_REC_NEXT(5555, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5555(choice, ...)  METALANG99_PRIV_REC_NEXT(5556, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5556(choice, ...)  METALANG99_PRIV_REC_NEXT(5557, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5557(choice, ...)  METALANG99_PRIV_REC_NEXT(5558, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5558(choice, ...)  METALANG99_PRIV_REC_NEXT(5559, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5559(choice, ...)  METALANG99_PRIV_REC_NEXT(5560, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5560(choice, ...)  METALANG99_PRIV_REC_NEXT(5561, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5561(choice, ...)  METALANG99_PRIV_REC_NEXT(5562, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5562(choice, ...)  METALANG99_PRIV_REC_NEXT(5563, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5563(choice, ...)  METALANG99_PRIV_REC_NEXT(5564, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5564(choice, ...)  METALANG99_PRIV_REC_NEXT(5565, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5565(choice, ...)  METALANG99_PRIV_REC_NEXT(5566, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5566(choice, ...)  METALANG99_PRIV_REC_NEXT(5567, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5567(choice, ...)  METALANG99_PRIV_REC_NEXT(5568, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5568(choice, ...)  METALANG99_PRIV_REC_NEXT(5569, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5569(choice, ...)  METALANG99_PRIV_REC_NEXT(5570, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5570(choice, ...)  METALANG99_PRIV_REC_NEXT(5571, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5571(choice, ...)  METALANG99_PRIV_REC_NEXT(5572, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5572(choice, ...)  METALANG99_PRIV_REC_NEXT(5573, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5573(choice, ...)  METALANG99_PRIV_REC_NEXT(5574, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5574(choice, ...)  METALANG99_PRIV_REC_NEXT(5575, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5575(choice, ...)  METALANG99_PRIV_REC_NEXT(5576, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5576(choice, ...)  METALANG99_PRIV_REC_NEXT(5577, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5577(choice, ...)  METALANG99_PRIV_REC_NEXT(5578, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5578(choice, ...)  METALANG99_PRIV_REC_NEXT(5579, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5579(choice, ...)  METALANG99_PRIV_REC_NEXT(5580, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5580(choice, ...)  METALANG99_PRIV_REC_NEXT(5581, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5581(choice, ...)  METALANG99_PRIV_REC_NEXT(5582, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5582(choice, ...)  METALANG99_PRIV_REC_NEXT(5583, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5583(choice, ...)  METALANG99_PRIV_REC_NEXT(5584, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5584(choice, ...)  METALANG99_PRIV_REC_NEXT(5585, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5585(choice, ...)  METALANG99_PRIV_REC_NEXT(5586, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5586(choice, ...)  METALANG99_PRIV_REC_NEXT(5587, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5587(choice, ...)  METALANG99_PRIV_REC_NEXT(5588, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5588(choice, ...)  METALANG99_PRIV_REC_NEXT(5589, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5589(choice, ...)  METALANG99_PRIV_REC_NEXT(5590, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5590(choice, ...)  METALANG99_PRIV_REC_NEXT(5591, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5591(choice, ...)  METALANG99_PRIV_REC_NEXT(5592, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5592(choice, ...)  METALANG99_PRIV_REC_NEXT(5593, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5593(choice, ...)  METALANG99_PRIV_REC_NEXT(5594, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5594(choice, ...)  METALANG99_PRIV_REC_NEXT(5595, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5595(choice, ...)  METALANG99_PRIV_REC_NEXT(5596, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5596(choice, ...)  METALANG99_PRIV_REC_NEXT(5597, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5597(choice, ...)  METALANG99_PRIV_REC_NEXT(5598, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5598(choice, ...)  METALANG99_PRIV_REC_NEXT(5599, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5599(choice, ...)  METALANG99_PRIV_REC_NEXT(5600, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5600(choice, ...)  METALANG99_PRIV_REC_NEXT(5601, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5601(choice, ...)  METALANG99_PRIV_REC_NEXT(5602, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5602(choice, ...)  METALANG99_PRIV_REC_NEXT(5603, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5603(choice, ...)  METALANG99_PRIV_REC_NEXT(5604, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5604(choice, ...)  METALANG99_PRIV_REC_NEXT(5605, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5605(choice, ...)  METALANG99_PRIV_REC_NEXT(5606, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5606(choice, ...)  METALANG99_PRIV_REC_NEXT(5607, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5607(choice, ...)  METALANG99_PRIV_REC_NEXT(5608, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5608(choice, ...)  METALANG99_PRIV_REC_NEXT(5609, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5609(choice, ...)  METALANG99_PRIV_REC_NEXT(5610, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5610(choice, ...)  METALANG99_PRIV_REC_NEXT(5611, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5611(choice, ...)  METALANG99_PRIV_REC_NEXT(5612, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5612(choice, ...)  METALANG99_PRIV_REC_NEXT(5613, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5613(choice, ...)  METALANG99_PRIV_REC_NEXT(5614, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5614(choice, ...)  METALANG99_PRIV_REC_NEXT(5615, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5615(choice, ...)  METALANG99_PRIV_REC_NEXT(5616, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5616(choice, ...)  METALANG99_PRIV_REC_NEXT(5617, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5617(choice, ...)  METALANG99_PRIV_REC_NEXT(5618, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5618(choice, ...)  METALANG99_PRIV_REC_NEXT(5619, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5619(choice, ...)  METALANG99_PRIV_REC_NEXT(5620, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5620(choice, ...)  METALANG99_PRIV_REC_NEXT(5621, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5621(choice, ...)  METALANG99_PRIV_REC_NEXT(5622, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5622(choice, ...)  METALANG99_PRIV_REC_NEXT(5623, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5623(choice, ...)  METALANG99_PRIV_REC_NEXT(5624, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5624(choice, ...)  METALANG99_PRIV_REC_NEXT(5625, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5625(choice, ...)  METALANG99_PRIV_REC_NEXT(5626, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5626(choice, ...)  METALANG99_PRIV_REC_NEXT(5627, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5627(choice, ...)  METALANG99_PRIV_REC_NEXT(5628, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5628(choice, ...)  METALANG99_PRIV_REC_NEXT(5629, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5629(choice, ...)  METALANG99_PRIV_REC_NEXT(5630, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5630(choice, ...)  METALANG99_PRIV_REC_NEXT(5631, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5631(choice, ...)  METALANG99_PRIV_REC_NEXT(5632, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5632(choice, ...)  METALANG99_PRIV_REC_NEXT(5633, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5633(choice, ...)  METALANG99_PRIV_REC_NEXT(5634, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5634(choice, ...)  METALANG99_PRIV_REC_NEXT(5635, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5635(choice, ...)  METALANG99_PRIV_REC_NEXT(5636, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5636(choice, ...)  METALANG99_PRIV_REC_NEXT(5637, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5637(choice, ...)  METALANG99_PRIV_REC_NEXT(5638, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5638(choice, ...)  METALANG99_PRIV_REC_NEXT(5639, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5639(choice, ...)  METALANG99_PRIV_REC_NEXT(5640, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5640(choice, ...)  METALANG99_PRIV_REC_NEXT(5641, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5641(choice, ...)  METALANG99_PRIV_REC_NEXT(5642, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5642(choice, ...)  METALANG99_PRIV_REC_NEXT(5643, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5643(choice, ...)  METALANG99_PRIV_REC_NEXT(5644, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5644(choice, ...)  METALANG99_PRIV_REC_NEXT(5645, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5645(choice, ...)  METALANG99_PRIV_REC_NEXT(5646, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5646(choice, ...)  METALANG99_PRIV_REC_NEXT(5647, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5647(choice, ...)  METALANG99_PRIV_REC_NEXT(5648, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5648(choice, ...)  METALANG99_PRIV_REC_NEXT(5649, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5649(choice, ...)  METALANG99_PRIV_REC_NEXT(5650, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5650(choice, ...)  METALANG99_PRIV_REC_NEXT(5651, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5651(choice, ...)  METALANG99_PRIV_REC_NEXT(5652, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5652(choice, ...)  METALANG99_PRIV_REC_NEXT(5653, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5653(choice, ...)  METALANG99_PRIV_REC_NEXT(5654, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5654(choice, ...)  METALANG99_PRIV_REC_NEXT(5655, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5655(choice, ...)  METALANG99_PRIV_REC_NEXT(5656, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5656(choice, ...)  METALANG99_PRIV_REC_NEXT(5657, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5657(choice, ...)  METALANG99_PRIV_REC_NEXT(5658, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5658(choice, ...)  METALANG99_PRIV_REC_NEXT(5659, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5659(choice, ...)  METALANG99_PRIV_REC_NEXT(5660, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5660(choice, ...)  METALANG99_PRIV_REC_NEXT(5661, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5661(choice, ...)  METALANG99_PRIV_REC_NEXT(5662, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5662(choice, ...)  METALANG99_PRIV_REC_NEXT(5663, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5663(choice, ...)  METALANG99_PRIV_REC_NEXT(5664, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5664(choice, ...)  METALANG99_PRIV_REC_NEXT(5665, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5665(choice, ...)  METALANG99_PRIV_REC_NEXT(5666, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5666(choice, ...)  METALANG99_PRIV_REC_NEXT(5667, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5667(choice, ...)  METALANG99_PRIV_REC_NEXT(5668, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5668(choice, ...)  METALANG99_PRIV_REC_NEXT(5669, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5669(choice, ...)  METALANG99_PRIV_REC_NEXT(5670, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5670(choice, ...)  METALANG99_PRIV_REC_NEXT(5671, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5671(choice, ...)  METALANG99_PRIV_REC_NEXT(5672, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5672(choice, ...)  METALANG99_PRIV_REC_NEXT(5673, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5673(choice, ...)  METALANG99_PRIV_REC_NEXT(5674, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5674(choice, ...)  METALANG99_PRIV_REC_NEXT(5675, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5675(choice, ...)  METALANG99_PRIV_REC_NEXT(5676, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5676(choice, ...)  METALANG99_PRIV_REC_NEXT(5677, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5677(choice, ...)  METALANG99_PRIV_REC_NEXT(5678, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5678(choice, ...)  METALANG99_PRIV_REC_NEXT(5679, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5679(choice, ...)  METALANG99_PRIV_REC_NEXT(5680, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5680(choice, ...)  METALANG99_PRIV_REC_NEXT(5681, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5681(choice, ...)  METALANG99_PRIV_REC_NEXT(5682, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5682(choice, ...)  METALANG99_PRIV_REC_NEXT(5683, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5683(choice, ...)  METALANG99_PRIV_REC_NEXT(5684, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5684(choice, ...)  METALANG99_PRIV_REC_NEXT(5685, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5685(choice, ...)  METALANG99_PRIV_REC_NEXT(5686, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5686(choice, ...)  METALANG99_PRIV_REC_NEXT(5687, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5687(choice, ...)  METALANG99_PRIV_REC_NEXT(5688, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5688(choice, ...)  METALANG99_PRIV_REC_NEXT(5689, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5689(choice, ...)  METALANG99_PRIV_REC_NEXT(5690, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5690(choice, ...)  METALANG99_PRIV_REC_NEXT(5691, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5691(choice, ...)  METALANG99_PRIV_REC_NEXT(5692, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5692(choice, ...)  METALANG99_PRIV_REC_NEXT(5693, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5693(choice, ...)  METALANG99_PRIV_REC_NEXT(5694, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5694(choice, ...)  METALANG99_PRIV_REC_NEXT(5695, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5695(choice, ...)  METALANG99_PRIV_REC_NEXT(5696, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5696(choice, ...)  METALANG99_PRIV_REC_NEXT(5697, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5697(choice, ...)  METALANG99_PRIV_REC_NEXT(5698, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5698(choice, ...)  METALANG99_PRIV_REC_NEXT(5699, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5699(choice, ...)  METALANG99_PRIV_REC_NEXT(5700, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5700(choice, ...)  METALANG99_PRIV_REC_NEXT(5701, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5701(choice, ...)  METALANG99_PRIV_REC_NEXT(5702, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5702(choice, ...)  METALANG99_PRIV_REC_NEXT(5703, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5703(choice, ...)  METALANG99_PRIV_REC_NEXT(5704, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5704(choice, ...)  METALANG99_PRIV_REC_NEXT(5705, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5705(choice, ...)  METALANG99_PRIV_REC_NEXT(5706, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5706(choice, ...)  METALANG99_PRIV_REC_NEXT(5707, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5707(choice, ...)  METALANG99_PRIV_REC_NEXT(5708, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5708(choice, ...)  METALANG99_PRIV_REC_NEXT(5709, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5709(choice, ...)  METALANG99_PRIV_REC_NEXT(5710, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5710(choice, ...)  METALANG99_PRIV_REC_NEXT(5711, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5711(choice, ...)  METALANG99_PRIV_REC_NEXT(5712, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5712(choice, ...)  METALANG99_PRIV_REC_NEXT(5713, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5713(choice, ...)  METALANG99_PRIV_REC_NEXT(5714, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5714(choice, ...)  METALANG99_PRIV_REC_NEXT(5715, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5715(choice, ...)  METALANG99_PRIV_REC_NEXT(5716, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5716(choice, ...)  METALANG99_PRIV_REC_NEXT(5717, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5717(choice, ...)  METALANG99_PRIV_REC_NEXT(5718, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5718(choice, ...)  METALANG99_PRIV_REC_NEXT(5719, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5719(choice, ...)  METALANG99_PRIV_REC_NEXT(5720, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5720(choice, ...)  METALANG99_PRIV_REC_NEXT(5721, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5721(choice, ...)  METALANG99_PRIV_REC_NEXT(5722, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5722(choice, ...)  METALANG99_PRIV_REC_NEXT(5723, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5723(choice, ...)  METALANG99_PRIV_REC_NEXT(5724, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5724(choice, ...)  METALANG99_PRIV_REC_NEXT(5725, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5725(choice, ...)  METALANG99_PRIV_REC_NEXT(5726, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5726(choice, ...)  METALANG99_PRIV_REC_NEXT(5727, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5727(choice, ...)  METALANG99_PRIV_REC_NEXT(5728, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5728(choice, ...)  METALANG99_PRIV_REC_NEXT(5729, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5729(choice, ...)  METALANG99_PRIV_REC_NEXT(5730, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5730(choice, ...)  METALANG99_PRIV_REC_NEXT(5731, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5731(choice, ...)  METALANG99_PRIV_REC_NEXT(5732, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5732(choice, ...)  METALANG99_PRIV_REC_NEXT(5733, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5733(choice, ...)  METALANG99_PRIV_REC_NEXT(5734, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5734(choice, ...)  METALANG99_PRIV_REC_NEXT(5735, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5735(choice, ...)  METALANG99_PRIV_REC_NEXT(5736, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5736(choice, ...)  METALANG99_PRIV_REC_NEXT(5737, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5737(choice, ...)  METALANG99_PRIV_REC_NEXT(5738, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5738(choice, ...)  METALANG99_PRIV_REC_NEXT(5739, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5739(choice, ...)  METALANG99_PRIV_REC_NEXT(5740, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5740(choice, ...)  METALANG99_PRIV_REC_NEXT(5741, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5741(choice, ...)  METALANG99_PRIV_REC_NEXT(5742, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5742(choice, ...)  METALANG99_PRIV_REC_NEXT(5743, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5743(choice, ...)  METALANG99_PRIV_REC_NEXT(5744, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5744(choice, ...)  METALANG99_PRIV_REC_NEXT(5745, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5745(choice, ...)  METALANG99_PRIV_REC_NEXT(5746, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5746(choice, ...)  METALANG99_PRIV_REC_NEXT(5747, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5747(choice, ...)  METALANG99_PRIV_REC_NEXT(5748, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5748(choice, ...)  METALANG99_PRIV_REC_NEXT(5749, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5749(choice, ...)  METALANG99_PRIV_REC_NEXT(5750, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5750(choice, ...)  METALANG99_PRIV_REC_NEXT(5751, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5751(choice, ...)  METALANG99_PRIV_REC_NEXT(5752, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5752(choice, ...)  METALANG99_PRIV_REC_NEXT(5753, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5753(choice, ...)  METALANG99_PRIV_REC_NEXT(5754, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5754(choice, ...)  METALANG99_PRIV_REC_NEXT(5755, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5755(choice, ...)  METALANG99_PRIV_REC_NEXT(5756, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5756(choice, ...)  METALANG99_PRIV_REC_NEXT(5757, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5757(choice, ...)  METALANG99_PRIV_REC_NEXT(5758, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5758(choice, ...)  METALANG99_PRIV_REC_NEXT(5759, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5759(choice, ...)  METALANG99_PRIV_REC_NEXT(5760, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5760(choice, ...)  METALANG99_PRIV_REC_NEXT(5761, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5761(choice, ...)  METALANG99_PRIV_REC_NEXT(5762, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5762(choice, ...)  METALANG99_PRIV_REC_NEXT(5763, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5763(choice, ...)  METALANG99_PRIV_REC_NEXT(5764, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5764(choice, ...)  METALANG99_PRIV_REC_NEXT(5765, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5765(choice, ...)  METALANG99_PRIV_REC_NEXT(5766, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5766(choice, ...)  METALANG99_PRIV_REC_NEXT(5767, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5767(choice, ...)  METALANG99_PRIV_REC_NEXT(5768, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5768(choice, ...)  METALANG99_PRIV_REC_NEXT(5769, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5769(choice, ...)  METALANG99_PRIV_REC_NEXT(5770, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5770(choice, ...)  METALANG99_PRIV_REC_NEXT(5771, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5771(choice, ...)  METALANG99_PRIV_REC_NEXT(5772, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5772(choice, ...)  METALANG99_PRIV_REC_NEXT(5773, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5773(choice, ...)  METALANG99_PRIV_REC_NEXT(5774, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5774(choice, ...)  METALANG99_PRIV_REC_NEXT(5775, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5775(choice, ...)  METALANG99_PRIV_REC_NEXT(5776, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5776(choice, ...)  METALANG99_PRIV_REC_NEXT(5777, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5777(choice, ...)  METALANG99_PRIV_REC_NEXT(5778, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5778(choice, ...)  METALANG99_PRIV_REC_NEXT(5779, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5779(choice, ...)  METALANG99_PRIV_REC_NEXT(5780, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5780(choice, ...)  METALANG99_PRIV_REC_NEXT(5781, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5781(choice, ...)  METALANG99_PRIV_REC_NEXT(5782, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5782(choice, ...)  METALANG99_PRIV_REC_NEXT(5783, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5783(choice, ...)  METALANG99_PRIV_REC_NEXT(5784, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5784(choice, ...)  METALANG99_PRIV_REC_NEXT(5785, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5785(choice, ...)  METALANG99_PRIV_REC_NEXT(5786, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5786(choice, ...)  METALANG99_PRIV_REC_NEXT(5787, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5787(choice, ...)  METALANG99_PRIV_REC_NEXT(5788, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5788(choice, ...)  METALANG99_PRIV_REC_NEXT(5789, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5789(choice, ...)  METALANG99_PRIV_REC_NEXT(5790, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5790(choice, ...)  METALANG99_PRIV_REC_NEXT(5791, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5791(choice, ...)  METALANG99_PRIV_REC_NEXT(5792, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5792(choice, ...)  METALANG99_PRIV_REC_NEXT(5793, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5793(choice, ...)  METALANG99_PRIV_REC_NEXT(5794, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5794(choice, ...)  METALANG99_PRIV_REC_NEXT(5795, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5795(choice, ...)  METALANG99_PRIV_REC_NEXT(5796, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5796(choice, ...)  METALANG99_PRIV_REC_NEXT(5797, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5797(choice, ...)  METALANG99_PRIV_REC_NEXT(5798, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5798(choice, ...)  METALANG99_PRIV_REC_NEXT(5799, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5799(choice, ...)  METALANG99_PRIV_REC_NEXT(5800, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5800(choice, ...)  METALANG99_PRIV_REC_NEXT(5801, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5801(choice, ...)  METALANG99_PRIV_REC_NEXT(5802, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5802(choice, ...)  METALANG99_PRIV_REC_NEXT(5803, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5803(choice, ...)  METALANG99_PRIV_REC_NEXT(5804, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5804(choice, ...)  METALANG99_PRIV_REC_NEXT(5805, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5805(choice, ...)  METALANG99_PRIV_REC_NEXT(5806, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5806(choice, ...)  METALANG99_PRIV_REC_NEXT(5807, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5807(choice, ...)  METALANG99_PRIV_REC_NEXT(5808, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5808(choice, ...)  METALANG99_PRIV_REC_NEXT(5809, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5809(choice, ...)  METALANG99_PRIV_REC_NEXT(5810, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5810(choice, ...)  METALANG99_PRIV_REC_NEXT(5811, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5811(choice, ...)  METALANG99_PRIV_REC_NEXT(5812, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5812(choice, ...)  METALANG99_PRIV_REC_NEXT(5813, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5813(choice, ...)  METALANG99_PRIV_REC_NEXT(5814, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5814(choice, ...)  METALANG99_PRIV_REC_NEXT(5815, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5815(choice, ...)  METALANG99_PRIV_REC_NEXT(5816, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5816(choice, ...)  METALANG99_PRIV_REC_NEXT(5817, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5817(choice, ...)  METALANG99_PRIV_REC_NEXT(5818, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5818(choice, ...)  METALANG99_PRIV_REC_NEXT(5819, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5819(choice, ...)  METALANG99_PRIV_REC_NEXT(5820, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5820(choice, ...)  METALANG99_PRIV_REC_NEXT(5821, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5821(choice, ...)  METALANG99_PRIV_REC_NEXT(5822, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5822(choice, ...)  METALANG99_PRIV_REC_NEXT(5823, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5823(choice, ...)  METALANG99_PRIV_REC_NEXT(5824, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5824(choice, ...)  METALANG99_PRIV_REC_NEXT(5825, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5825(choice, ...)  METALANG99_PRIV_REC_NEXT(5826, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5826(choice, ...)  METALANG99_PRIV_REC_NEXT(5827, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5827(choice, ...)  METALANG99_PRIV_REC_NEXT(5828, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5828(choice, ...)  METALANG99_PRIV_REC_NEXT(5829, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5829(choice, ...)  METALANG99_PRIV_REC_NEXT(5830, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5830(choice, ...)  METALANG99_PRIV_REC_NEXT(5831, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5831(choice, ...)  METALANG99_PRIV_REC_NEXT(5832, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5832(choice, ...)  METALANG99_PRIV_REC_NEXT(5833, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5833(choice, ...)  METALANG99_PRIV_REC_NEXT(5834, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5834(choice, ...)  METALANG99_PRIV_REC_NEXT(5835, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5835(choice, ...)  METALANG99_PRIV_REC_NEXT(5836, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5836(choice, ...)  METALANG99_PRIV_REC_NEXT(5837, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5837(choice, ...)  METALANG99_PRIV_REC_NEXT(5838, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5838(choice, ...)  METALANG99_PRIV_REC_NEXT(5839, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5839(choice, ...)  METALANG99_PRIV_REC_NEXT(5840, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5840(choice, ...)  METALANG99_PRIV_REC_NEXT(5841, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5841(choice, ...)  METALANG99_PRIV_REC_NEXT(5842, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5842(choice, ...)  METALANG99_PRIV_REC_NEXT(5843, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5843(choice, ...)  METALANG99_PRIV_REC_NEXT(5844, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5844(choice, ...)  METALANG99_PRIV_REC_NEXT(5845, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5845(choice, ...)  METALANG99_PRIV_REC_NEXT(5846, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5846(choice, ...)  METALANG99_PRIV_REC_NEXT(5847, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5847(choice, ...)  METALANG99_PRIV_REC_NEXT(5848, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5848(choice, ...)  METALANG99_PRIV_REC_NEXT(5849, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5849(choice, ...)  METALANG99_PRIV_REC_NEXT(5850, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5850(choice, ...)  METALANG99_PRIV_REC_NEXT(5851, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5851(choice, ...)  METALANG99_PRIV_REC_NEXT(5852, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5852(choice, ...)  METALANG99_PRIV_REC_NEXT(5853, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5853(choice, ...)  METALANG99_PRIV_REC_NEXT(5854, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5854(choice, ...)  METALANG99_PRIV_REC_NEXT(5855, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5855(choice, ...)  METALANG99_PRIV_REC_NEXT(5856, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5856(choice, ...)  METALANG99_PRIV_REC_NEXT(5857, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5857(choice, ...)  METALANG99_PRIV_REC_NEXT(5858, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5858(choice, ...)  METALANG99_PRIV_REC_NEXT(5859, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5859(choice, ...)  METALANG99_PRIV_REC_NEXT(5860, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5860(choice, ...)  METALANG99_PRIV_REC_NEXT(5861, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5861(choice, ...)  METALANG99_PRIV_REC_NEXT(5862, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5862(choice, ...)  METALANG99_PRIV_REC_NEXT(5863, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5863(choice, ...)  METALANG99_PRIV_REC_NEXT(5864, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5864(choice, ...)  METALANG99_PRIV_REC_NEXT(5865, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5865(choice, ...)  METALANG99_PRIV_REC_NEXT(5866, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5866(choice, ...)  METALANG99_PRIV_REC_NEXT(5867, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5867(choice, ...)  METALANG99_PRIV_REC_NEXT(5868, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5868(choice, ...)  METALANG99_PRIV_REC_NEXT(5869, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5869(choice, ...)  METALANG99_PRIV_REC_NEXT(5870, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5870(choice, ...)  METALANG99_PRIV_REC_NEXT(5871, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5871(choice, ...)  METALANG99_PRIV_REC_NEXT(5872, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5872(choice, ...)  METALANG99_PRIV_REC_NEXT(5873, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5873(choice, ...)  METALANG99_PRIV_REC_NEXT(5874, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5874(choice, ...)  METALANG99_PRIV_REC_NEXT(5875, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5875(choice, ...)  METALANG99_PRIV_REC_NEXT(5876, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5876(choice, ...)  METALANG99_PRIV_REC_NEXT(5877, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5877(choice, ...)  METALANG99_PRIV_REC_NEXT(5878, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5878(choice, ...)  METALANG99_PRIV_REC_NEXT(5879, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5879(choice, ...)  METALANG99_PRIV_REC_NEXT(5880, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5880(choice, ...)  METALANG99_PRIV_REC_NEXT(5881, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5881(choice, ...)  METALANG99_PRIV_REC_NEXT(5882, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5882(choice, ...)  METALANG99_PRIV_REC_NEXT(5883, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5883(choice, ...)  METALANG99_PRIV_REC_NEXT(5884, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5884(choice, ...)  METALANG99_PRIV_REC_NEXT(5885, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5885(choice, ...)  METALANG99_PRIV_REC_NEXT(5886, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5886(choice, ...)  METALANG99_PRIV_REC_NEXT(5887, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5887(choice, ...)  METALANG99_PRIV_REC_NEXT(5888, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5888(choice, ...)  METALANG99_PRIV_REC_NEXT(5889, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5889(choice, ...)  METALANG99_PRIV_REC_NEXT(5890, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5890(choice, ...)  METALANG99_PRIV_REC_NEXT(5891, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5891(choice, ...)  METALANG99_PRIV_REC_NEXT(5892, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5892(choice, ...)  METALANG99_PRIV_REC_NEXT(5893, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5893(choice, ...)  METALANG99_PRIV_REC_NEXT(5894, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5894(choice, ...)  METALANG99_PRIV_REC_NEXT(5895, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5895(choice, ...)  METALANG99_PRIV_REC_NEXT(5896, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5896(choice, ...)  METALANG99_PRIV_REC_NEXT(5897, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5897(choice, ...)  METALANG99_PRIV_REC_NEXT(5898, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5898(choice, ...)  METALANG99_PRIV_REC_NEXT(5899, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5899(choice, ...)  METALANG99_PRIV_REC_NEXT(5900, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5900(choice, ...)  METALANG99_PRIV_REC_NEXT(5901, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5901(choice, ...)  METALANG99_PRIV_REC_NEXT(5902, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5902(choice, ...)  METALANG99_PRIV_REC_NEXT(5903, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5903(choice, ...)  METALANG99_PRIV_REC_NEXT(5904, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5904(choice, ...)  METALANG99_PRIV_REC_NEXT(5905, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5905(choice, ...)  METALANG99_PRIV_REC_NEXT(5906, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5906(choice, ...)  METALANG99_PRIV_REC_NEXT(5907, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5907(choice, ...)  METALANG99_PRIV_REC_NEXT(5908, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5908(choice, ...)  METALANG99_PRIV_REC_NEXT(5909, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5909(choice, ...)  METALANG99_PRIV_REC_NEXT(5910, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5910(choice, ...)  METALANG99_PRIV_REC_NEXT(5911, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5911(choice, ...)  METALANG99_PRIV_REC_NEXT(5912, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5912(choice, ...)  METALANG99_PRIV_REC_NEXT(5913, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5913(choice, ...)  METALANG99_PRIV_REC_NEXT(5914, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5914(choice, ...)  METALANG99_PRIV_REC_NEXT(5915, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5915(choice, ...)  METALANG99_PRIV_REC_NEXT(5916, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5916(choice, ...)  METALANG99_PRIV_REC_NEXT(5917, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5917(choice, ...)  METALANG99_PRIV_REC_NEXT(5918, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5918(choice, ...)  METALANG99_PRIV_REC_NEXT(5919, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5919(choice, ...)  METALANG99_PRIV_REC_NEXT(5920, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5920(choice, ...)  METALANG99_PRIV_REC_NEXT(5921, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5921(choice, ...)  METALANG99_PRIV_REC_NEXT(5922, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5922(choice, ...)  METALANG99_PRIV_REC_NEXT(5923, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5923(choice, ...)  METALANG99_PRIV_REC_NEXT(5924, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5924(choice, ...)  METALANG99_PRIV_REC_NEXT(5925, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5925(choice, ...)  METALANG99_PRIV_REC_NEXT(5926, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5926(choice, ...)  METALANG99_PRIV_REC_NEXT(5927, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5927(choice, ...)  METALANG99_PRIV_REC_NEXT(5928, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5928(choice, ...)  METALANG99_PRIV_REC_NEXT(5929, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5929(choice, ...)  METALANG99_PRIV_REC_NEXT(5930, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5930(choice, ...)  METALANG99_PRIV_REC_NEXT(5931, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5931(choice, ...)  METALANG99_PRIV_REC_NEXT(5932, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5932(choice, ...)  METALANG99_PRIV_REC_NEXT(5933, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5933(choice, ...)  METALANG99_PRIV_REC_NEXT(5934, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5934(choice, ...)  METALANG99_PRIV_REC_NEXT(5935, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5935(choice, ...)  METALANG99_PRIV_REC_NEXT(5936, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5936(choice, ...)  METALANG99_PRIV_REC_NEXT(5937, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5937(choice, ...)  METALANG99_PRIV_REC_NEXT(5938, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5938(choice, ...)  METALANG99_PRIV_REC_NEXT(5939, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5939(choice, ...)  METALANG99_PRIV_REC_NEXT(5940, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5940(choice, ...)  METALANG99_PRIV_REC_NEXT(5941, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5941(choice, ...)  METALANG99_PRIV_REC_NEXT(5942, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5942(choice, ...)  METALANG99_PRIV_REC_NEXT(5943, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5943(choice, ...)  METALANG99_PRIV_REC_NEXT(5944, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5944(choice, ...)  METALANG99_PRIV_REC_NEXT(5945, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5945(choice, ...)  METALANG99_PRIV_REC_NEXT(5946, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5946(choice, ...)  METALANG99_PRIV_REC_NEXT(5947, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5947(choice, ...)  METALANG99_PRIV_REC_NEXT(5948, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5948(choice, ...)  METALANG99_PRIV_REC_NEXT(5949, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5949(choice, ...)  METALANG99_PRIV_REC_NEXT(5950, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5950(choice, ...)  METALANG99_PRIV_REC_NEXT(5951, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5951(choice, ...)  METALANG99_PRIV_REC_NEXT(5952, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5952(choice, ...)  METALANG99_PRIV_REC_NEXT(5953, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5953(choice, ...)  METALANG99_PRIV_REC_NEXT(5954, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5954(choice, ...)  METALANG99_PRIV_REC_NEXT(5955, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5955(choice, ...)  METALANG99_PRIV_REC_NEXT(5956, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5956(choice, ...)  METALANG99_PRIV_REC_NEXT(5957, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5957(choice, ...)  METALANG99_PRIV_REC_NEXT(5958, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5958(choice, ...)  METALANG99_PRIV_REC_NEXT(5959, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5959(choice, ...)  METALANG99_PRIV_REC_NEXT(5960, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5960(choice, ...)  METALANG99_PRIV_REC_NEXT(5961, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5961(choice, ...)  METALANG99_PRIV_REC_NEXT(5962, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5962(choice, ...)  METALANG99_PRIV_REC_NEXT(5963, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5963(choice, ...)  METALANG99_PRIV_REC_NEXT(5964, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5964(choice, ...)  METALANG99_PRIV_REC_NEXT(5965, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5965(choice, ...)  METALANG99_PRIV_REC_NEXT(5966, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5966(choice, ...)  METALANG99_PRIV_REC_NEXT(5967, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5967(choice, ...)  METALANG99_PRIV_REC_NEXT(5968, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5968(choice, ...)  METALANG99_PRIV_REC_NEXT(5969, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5969(choice, ...)  METALANG99_PRIV_REC_NEXT(5970, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5970(choice, ...)  METALANG99_PRIV_REC_NEXT(5971, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5971(choice, ...)  METALANG99_PRIV_REC_NEXT(5972, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5972(choice, ...)  METALANG99_PRIV_REC_NEXT(5973, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5973(choice, ...)  METALANG99_PRIV_REC_NEXT(5974, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5974(choice, ...)  METALANG99_PRIV_REC_NEXT(5975, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5975(choice, ...)  METALANG99_PRIV_REC_NEXT(5976, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5976(choice, ...)  METALANG99_PRIV_REC_NEXT(5977, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5977(choice, ...)  METALANG99_PRIV_REC_NEXT(5978, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5978(choice, ...)  METALANG99_PRIV_REC_NEXT(5979, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5979(choice, ...)  METALANG99_PRIV_REC_NEXT(5980, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5980(choice, ...)  METALANG99_PRIV_REC_NEXT(5981, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5981(choice, ...)  METALANG99_PRIV_REC_NEXT(5982, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5982(choice, ...)  METALANG99_PRIV_REC_NEXT(5983, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5983(choice, ...)  METALANG99_PRIV_REC_NEXT(5984, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5984(choice, ...)  METALANG99_PRIV_REC_NEXT(5985, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5985(choice, ...)  METALANG99_PRIV_REC_NEXT(5986, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5986(choice, ...)  METALANG99_PRIV_REC_NEXT(5987, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5987(choice, ...)  METALANG99_PRIV_REC_NEXT(5988, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5988(choice, ...)  METALANG99_PRIV_REC_NEXT(5989, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5989(choice, ...)  METALANG99_PRIV_REC_NEXT(5990, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5990(choice, ...)  METALANG99_PRIV_REC_NEXT(5991, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5991(choice, ...)  METALANG99_PRIV_REC_NEXT(5992, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5992(choice, ...)  METALANG99_PRIV_REC_NEXT(5993, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5993(choice, ...)  METALANG99_PRIV_REC_NEXT(5994, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5994(choice, ...)  METALANG99_PRIV_REC_NEXT(5995, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5995(choice, ...)  METALANG99_PRIV_REC_NEXT(5996, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5996(choice, ...)  METALANG99_PRIV_REC_NEXT(5997, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5997(choice, ...)  METALANG99_PRIV_REC_NEXT(5998, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5998(choice, ...)  METALANG99_PRIV_REC_NEXT(5999, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_5999(choice, ...)  METALANG99_PRIV_REC_NEXT(6000, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6000(choice, ...)  METALANG99_PRIV_REC_NEXT(6001, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6001(choice, ...)  METALANG99_PRIV_REC_NEXT(6002, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6002(choice, ...)  METALANG99_PRIV_REC_NEXT(6003, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6003(choice, ...)  METALANG99_PRIV_REC_NEXT(6004, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6004(choice, ...)  METALANG99_PRIV_REC_NEXT(6005, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6005(choice, ...)  METALANG99_PRIV_REC_NEXT(6006, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6006(choice, ...)  METALANG99_PRIV_REC_NEXT(6007, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6007(choice, ...)  METALANG99_PRIV_REC_NEXT(6008, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6008(choice, ...)  METALANG99_PRIV_REC_NEXT(6009, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6009(choice, ...)  METALANG99_PRIV_REC_NEXT(6010, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6010(choice, ...)  METALANG99_PRIV_REC_NEXT(6011, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6011(choice, ...)  METALANG99_PRIV_REC_NEXT(6012, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6012(choice, ...)  METALANG99_PRIV_REC_NEXT(6013, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6013(choice, ...)  METALANG99_PRIV_REC_NEXT(6014, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6014(choice, ...)  METALANG99_PRIV_REC_NEXT(6015, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6015(choice, ...)  METALANG99_PRIV_REC_NEXT(6016, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6016(choice, ...)  METALANG99_PRIV_REC_NEXT(6017, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6017(choice, ...)  METALANG99_PRIV_REC_NEXT(6018, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6018(choice, ...)  METALANG99_PRIV_REC_NEXT(6019, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6019(choice, ...)  METALANG99_PRIV_REC_NEXT(6020, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6020(choice, ...)  METALANG99_PRIV_REC_NEXT(6021, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6021(choice, ...)  METALANG99_PRIV_REC_NEXT(6022, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6022(choice, ...)  METALANG99_PRIV_REC_NEXT(6023, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6023(choice, ...)  METALANG99_PRIV_REC_NEXT(6024, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6024(choice, ...)  METALANG99_PRIV_REC_NEXT(6025, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6025(choice, ...)  METALANG99_PRIV_REC_NEXT(6026, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6026(choice, ...)  METALANG99_PRIV_REC_NEXT(6027, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6027(choice, ...)  METALANG99_PRIV_REC_NEXT(6028, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6028(choice, ...)  METALANG99_PRIV_REC_NEXT(6029, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6029(choice, ...)  METALANG99_PRIV_REC_NEXT(6030, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6030(choice, ...)  METALANG99_PRIV_REC_NEXT(6031, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6031(choice, ...)  METALANG99_PRIV_REC_NEXT(6032, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6032(choice, ...)  METALANG99_PRIV_REC_NEXT(6033, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6033(choice, ...)  METALANG99_PRIV_REC_NEXT(6034, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6034(choice, ...)  METALANG99_PRIV_REC_NEXT(6035, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6035(choice, ...)  METALANG99_PRIV_REC_NEXT(6036, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6036(choice, ...)  METALANG99_PRIV_REC_NEXT(6037, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6037(choice, ...)  METALANG99_PRIV_REC_NEXT(6038, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6038(choice, ...)  METALANG99_PRIV_REC_NEXT(6039, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6039(choice, ...)  METALANG99_PRIV_REC_NEXT(6040, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6040(choice, ...)  METALANG99_PRIV_REC_NEXT(6041, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6041(choice, ...)  METALANG99_PRIV_REC_NEXT(6042, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6042(choice, ...)  METALANG99_PRIV_REC_NEXT(6043, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6043(choice, ...)  METALANG99_PRIV_REC_NEXT(6044, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6044(choice, ...)  METALANG99_PRIV_REC_NEXT(6045, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6045(choice, ...)  METALANG99_PRIV_REC_NEXT(6046, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6046(choice, ...)  METALANG99_PRIV_REC_NEXT(6047, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6047(choice, ...)  METALANG99_PRIV_REC_NEXT(6048, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6048(choice, ...)  METALANG99_PRIV_REC_NEXT(6049, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6049(choice, ...)  METALANG99_PRIV_REC_NEXT(6050, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6050(choice, ...)  METALANG99_PRIV_REC_NEXT(6051, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6051(choice, ...)  METALANG99_PRIV_REC_NEXT(6052, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6052(choice, ...)  METALANG99_PRIV_REC_NEXT(6053, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6053(choice, ...)  METALANG99_PRIV_REC_NEXT(6054, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6054(choice, ...)  METALANG99_PRIV_REC_NEXT(6055, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6055(choice, ...)  METALANG99_PRIV_REC_NEXT(6056, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6056(choice, ...)  METALANG99_PRIV_REC_NEXT(6057, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6057(choice, ...)  METALANG99_PRIV_REC_NEXT(6058, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6058(choice, ...)  METALANG99_PRIV_REC_NEXT(6059, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6059(choice, ...)  METALANG99_PRIV_REC_NEXT(6060, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6060(choice, ...)  METALANG99_PRIV_REC_NEXT(6061, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6061(choice, ...)  METALANG99_PRIV_REC_NEXT(6062, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6062(choice, ...)  METALANG99_PRIV_REC_NEXT(6063, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6063(choice, ...)  METALANG99_PRIV_REC_NEXT(6064, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6064(choice, ...)  METALANG99_PRIV_REC_NEXT(6065, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6065(choice, ...)  METALANG99_PRIV_REC_NEXT(6066, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6066(choice, ...)  METALANG99_PRIV_REC_NEXT(6067, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6067(choice, ...)  METALANG99_PRIV_REC_NEXT(6068, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6068(choice, ...)  METALANG99_PRIV_REC_NEXT(6069, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6069(choice, ...)  METALANG99_PRIV_REC_NEXT(6070, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6070(choice, ...)  METALANG99_PRIV_REC_NEXT(6071, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6071(choice, ...)  METALANG99_PRIV_REC_NEXT(6072, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6072(choice, ...)  METALANG99_PRIV_REC_NEXT(6073, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6073(choice, ...)  METALANG99_PRIV_REC_NEXT(6074, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6074(choice, ...)  METALANG99_PRIV_REC_NEXT(6075, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6075(choice, ...)  METALANG99_PRIV_REC_NEXT(6076, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6076(choice, ...)  METALANG99_PRIV_REC_NEXT(6077, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6077(choice, ...)  METALANG99_PRIV_REC_NEXT(6078, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6078(choice, ...)  METALANG99_PRIV_REC_NEXT(6079, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6079(choice, ...)  METALANG99_PRIV_REC_NEXT(6080, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6080(choice, ...)  METALANG99_PRIV_REC_NEXT(6081, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6081(choice, ...)  METALANG99_PRIV_REC_NEXT(6082, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6082(choice, ...)  METALANG99_PRIV_REC_NEXT(6083, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6083(choice, ...)  METALANG99_PRIV_REC_NEXT(6084, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6084(choice, ...)  METALANG99_PRIV_REC_NEXT(6085, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6085(choice, ...)  METALANG99_PRIV_REC_NEXT(6086, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6086(choice, ...)  METALANG99_PRIV_REC_NEXT(6087, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6087(choice, ...)  METALANG99_PRIV_REC_NEXT(6088, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6088(choice, ...)  METALANG99_PRIV_REC_NEXT(6089, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6089(choice, ...)  METALANG99_PRIV_REC_NEXT(6090, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6090(choice, ...)  METALANG99_PRIV_REC_NEXT(6091, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6091(choice, ...)  METALANG99_PRIV_REC_NEXT(6092, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6092(choice, ...)  METALANG99_PRIV_REC_NEXT(6093, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6093(choice, ...)  METALANG99_PRIV_REC_NEXT(6094, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6094(choice, ...)  METALANG99_PRIV_REC_NEXT(6095, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6095(choice, ...)  METALANG99_PRIV_REC_NEXT(6096, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6096(choice, ...)  METALANG99_PRIV_REC_NEXT(6097, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6097(choice, ...)  METALANG99_PRIV_REC_NEXT(6098, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6098(choice, ...)  METALANG99_PRIV_REC_NEXT(6099, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6099(choice, ...)  METALANG99_PRIV_REC_NEXT(6100, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6100(choice, ...)  METALANG99_PRIV_REC_NEXT(6101, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6101(choice, ...)  METALANG99_PRIV_REC_NEXT(6102, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6102(choice, ...)  METALANG99_PRIV_REC_NEXT(6103, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6103(choice, ...)  METALANG99_PRIV_REC_NEXT(6104, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6104(choice, ...)  METALANG99_PRIV_REC_NEXT(6105, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6105(choice, ...)  METALANG99_PRIV_REC_NEXT(6106, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6106(choice, ...)  METALANG99_PRIV_REC_NEXT(6107, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6107(choice, ...)  METALANG99_PRIV_REC_NEXT(6108, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6108(choice, ...)  METALANG99_PRIV_REC_NEXT(6109, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6109(choice, ...)  METALANG99_PRIV_REC_NEXT(6110, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6110(choice, ...)  METALANG99_PRIV_REC_NEXT(6111, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6111(choice, ...)  METALANG99_PRIV_REC_NEXT(6112, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6112(choice, ...)  METALANG99_PRIV_REC_NEXT(6113, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6113(choice, ...)  METALANG99_PRIV_REC_NEXT(6114, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6114(choice, ...)  METALANG99_PRIV_REC_NEXT(6115, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6115(choice, ...)  METALANG99_PRIV_REC_NEXT(6116, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6116(choice, ...)  METALANG99_PRIV_REC_NEXT(6117, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6117(choice, ...)  METALANG99_PRIV_REC_NEXT(6118, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6118(choice, ...)  METALANG99_PRIV_REC_NEXT(6119, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6119(choice, ...)  METALANG99_PRIV_REC_NEXT(6120, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6120(choice, ...)  METALANG99_PRIV_REC_NEXT(6121, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6121(choice, ...)  METALANG99_PRIV_REC_NEXT(6122, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6122(choice, ...)  METALANG99_PRIV_REC_NEXT(6123, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6123(choice, ...)  METALANG99_PRIV_REC_NEXT(6124, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6124(choice, ...)  METALANG99_PRIV_REC_NEXT(6125, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6125(choice, ...)  METALANG99_PRIV_REC_NEXT(6126, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6126(choice, ...)  METALANG99_PRIV_REC_NEXT(6127, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6127(choice, ...)  METALANG99_PRIV_REC_NEXT(6128, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6128(choice, ...)  METALANG99_PRIV_REC_NEXT(6129, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6129(choice, ...)  METALANG99_PRIV_REC_NEXT(6130, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6130(choice, ...)  METALANG99_PRIV_REC_NEXT(6131, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6131(choice, ...)  METALANG99_PRIV_REC_NEXT(6132, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6132(choice, ...)  METALANG99_PRIV_REC_NEXT(6133, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6133(choice, ...)  METALANG99_PRIV_REC_NEXT(6134, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6134(choice, ...)  METALANG99_PRIV_REC_NEXT(6135, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6135(choice, ...)  METALANG99_PRIV_REC_NEXT(6136, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6136(choice, ...)  METALANG99_PRIV_REC_NEXT(6137, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6137(choice, ...)  METALANG99_PRIV_REC_NEXT(6138, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6138(choice, ...)  METALANG99_PRIV_REC_NEXT(6139, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6139(choice, ...)  METALANG99_PRIV_REC_NEXT(6140, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6140(choice, ...)  METALANG99_PRIV_REC_NEXT(6141, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6141(choice, ...)  METALANG99_PRIV_REC_NEXT(6142, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6142(choice, ...)  METALANG99_PRIV_REC_NEXT(6143, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6143(choice, ...)  METALANG99_PRIV_REC_NEXT(6144, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6144(choice, ...)  METALANG99_PRIV_REC_NEXT(6145, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6145(choice, ...)  METALANG99_PRIV_REC_NEXT(6146, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6146(choice, ...)  METALANG99_PRIV_REC_NEXT(6147, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6147(choice, ...)  METALANG99_PRIV_REC_NEXT(6148, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6148(choice, ...)  METALANG99_PRIV_REC_NEXT(6149, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6149(choice, ...)  METALANG99_PRIV_REC_NEXT(6150, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6150(choice, ...)  METALANG99_PRIV_REC_NEXT(6151, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6151(choice, ...)  METALANG99_PRIV_REC_NEXT(6152, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6152(choice, ...)  METALANG99_PRIV_REC_NEXT(6153, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6153(choice, ...)  METALANG99_PRIV_REC_NEXT(6154, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6154(choice, ...)  METALANG99_PRIV_REC_NEXT(6155, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6155(choice, ...)  METALANG99_PRIV_REC_NEXT(6156, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6156(choice, ...)  METALANG99_PRIV_REC_NEXT(6157, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6157(choice, ...)  METALANG99_PRIV_REC_NEXT(6158, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6158(choice, ...)  METALANG99_PRIV_REC_NEXT(6159, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6159(choice, ...)  METALANG99_PRIV_REC_NEXT(6160, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6160(choice, ...)  METALANG99_PRIV_REC_NEXT(6161, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6161(choice, ...)  METALANG99_PRIV_REC_NEXT(6162, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6162(choice, ...)  METALANG99_PRIV_REC_NEXT(6163, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6163(choice, ...)  METALANG99_PRIV_REC_NEXT(6164, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6164(choice, ...)  METALANG99_PRIV_REC_NEXT(6165, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6165(choice, ...)  METALANG99_PRIV_REC_NEXT(6166, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6166(choice, ...)  METALANG99_PRIV_REC_NEXT(6167, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6167(choice, ...)  METALANG99_PRIV_REC_NEXT(6168, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6168(choice, ...)  METALANG99_PRIV_REC_NEXT(6169, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6169(choice, ...)  METALANG99_PRIV_REC_NEXT(6170, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6170(choice, ...)  METALANG99_PRIV_REC_NEXT(6171, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6171(choice, ...)  METALANG99_PRIV_REC_NEXT(6172, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6172(choice, ...)  METALANG99_PRIV_REC_NEXT(6173, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6173(choice, ...)  METALANG99_PRIV_REC_NEXT(6174, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6174(choice, ...)  METALANG99_PRIV_REC_NEXT(6175, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6175(choice, ...)  METALANG99_PRIV_REC_NEXT(6176, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6176(choice, ...)  METALANG99_PRIV_REC_NEXT(6177, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6177(choice, ...)  METALANG99_PRIV_REC_NEXT(6178, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6178(choice, ...)  METALANG99_PRIV_REC_NEXT(6179, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6179(choice, ...)  METALANG99_PRIV_REC_NEXT(6180, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6180(choice, ...)  METALANG99_PRIV_REC_NEXT(6181, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6181(choice, ...)  METALANG99_PRIV_REC_NEXT(6182, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6182(choice, ...)  METALANG99_PRIV_REC_NEXT(6183, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6183(choice, ...)  METALANG99_PRIV_REC_NEXT(6184, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6184(choice, ...)  METALANG99_PRIV_REC_NEXT(6185, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6185(choice, ...)  METALANG99_PRIV_REC_NEXT(6186, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6186(choice, ...)  METALANG99_PRIV_REC_NEXT(6187, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6187(choice, ...)  METALANG99_PRIV_REC_NEXT(6188, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6188(choice, ...)  METALANG99_PRIV_REC_NEXT(6189, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6189(choice, ...)  METALANG99_PRIV_REC_NEXT(6190, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6190(choice, ...)  METALANG99_PRIV_REC_NEXT(6191, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6191(choice, ...)  METALANG99_PRIV_REC_NEXT(6192, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6192(choice, ...)  METALANG99_PRIV_REC_NEXT(6193, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6193(choice, ...)  METALANG99_PRIV_REC_NEXT(6194, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6194(choice, ...)  METALANG99_PRIV_REC_NEXT(6195, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6195(choice, ...)  METALANG99_PRIV_REC_NEXT(6196, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6196(choice, ...)  METALANG99_PRIV_REC_NEXT(6197, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6197(choice, ...)  METALANG99_PRIV_REC_NEXT(6198, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6198(choice, ...)  METALANG99_PRIV_REC_NEXT(6199, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6199(choice, ...)  METALANG99_PRIV_REC_NEXT(6200, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6200(choice, ...)  METALANG99_PRIV_REC_NEXT(6201, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6201(choice, ...)  METALANG99_PRIV_REC_NEXT(6202, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6202(choice, ...)  METALANG99_PRIV_REC_NEXT(6203, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6203(choice, ...)  METALANG99_PRIV_REC_NEXT(6204, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6204(choice, ...)  METALANG99_PRIV_REC_NEXT(6205, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6205(choice, ...)  METALANG99_PRIV_REC_NEXT(6206, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6206(choice, ...)  METALANG99_PRIV_REC_NEXT(6207, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6207(choice, ...)  METALANG99_PRIV_REC_NEXT(6208, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6208(choice, ...)  METALANG99_PRIV_REC_NEXT(6209, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6209(choice, ...)  METALANG99_PRIV_REC_NEXT(6210, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6210(choice, ...)  METALANG99_PRIV_REC_NEXT(6211, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6211(choice, ...)  METALANG99_PRIV_REC_NEXT(6212, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6212(choice, ...)  METALANG99_PRIV_REC_NEXT(6213, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6213(choice, ...)  METALANG99_PRIV_REC_NEXT(6214, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6214(choice, ...)  METALANG99_PRIV_REC_NEXT(6215, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6215(choice, ...)  METALANG99_PRIV_REC_NEXT(6216, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6216(choice, ...)  METALANG99_PRIV_REC_NEXT(6217, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6217(choice, ...)  METALANG99_PRIV_REC_NEXT(6218, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6218(choice, ...)  METALANG99_PRIV_REC_NEXT(6219, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6219(choice, ...)  METALANG99_PRIV_REC_NEXT(6220, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6220(choice, ...)  METALANG99_PRIV_REC_NEXT(6221, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6221(choice, ...)  METALANG99_PRIV_REC_NEXT(6222, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6222(choice, ...)  METALANG99_PRIV_REC_NEXT(6223, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6223(choice, ...)  METALANG99_PRIV_REC_NEXT(6224, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6224(choice, ...)  METALANG99_PRIV_REC_NEXT(6225, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6225(choice, ...)  METALANG99_PRIV_REC_NEXT(6226, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6226(choice, ...)  METALANG99_PRIV_REC_NEXT(6227, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6227(choice, ...)  METALANG99_PRIV_REC_NEXT(6228, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6228(choice, ...)  METALANG99_PRIV_REC_NEXT(6229, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6229(choice, ...)  METALANG99_PRIV_REC_NEXT(6230, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6230(choice, ...)  METALANG99_PRIV_REC_NEXT(6231, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6231(choice, ...)  METALANG99_PRIV_REC_NEXT(6232, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6232(choice, ...)  METALANG99_PRIV_REC_NEXT(6233, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6233(choice, ...)  METALANG99_PRIV_REC_NEXT(6234, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6234(choice, ...)  METALANG99_PRIV_REC_NEXT(6235, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6235(choice, ...)  METALANG99_PRIV_REC_NEXT(6236, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6236(choice, ...)  METALANG99_PRIV_REC_NEXT(6237, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6237(choice, ...)  METALANG99_PRIV_REC_NEXT(6238, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6238(choice, ...)  METALANG99_PRIV_REC_NEXT(6239, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6239(choice, ...)  METALANG99_PRIV_REC_NEXT(6240, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6240(choice, ...)  METALANG99_PRIV_REC_NEXT(6241, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6241(choice, ...)  METALANG99_PRIV_REC_NEXT(6242, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6242(choice, ...)  METALANG99_PRIV_REC_NEXT(6243, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6243(choice, ...)  METALANG99_PRIV_REC_NEXT(6244, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6244(choice, ...)  METALANG99_PRIV_REC_NEXT(6245, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6245(choice, ...)  METALANG99_PRIV_REC_NEXT(6246, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6246(choice, ...)  METALANG99_PRIV_REC_NEXT(6247, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6247(choice, ...)  METALANG99_PRIV_REC_NEXT(6248, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6248(choice, ...)  METALANG99_PRIV_REC_NEXT(6249, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6249(choice, ...)  METALANG99_PRIV_REC_NEXT(6250, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6250(choice, ...)  METALANG99_PRIV_REC_NEXT(6251, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6251(choice, ...)  METALANG99_PRIV_REC_NEXT(6252, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6252(choice, ...)  METALANG99_PRIV_REC_NEXT(6253, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6253(choice, ...)  METALANG99_PRIV_REC_NEXT(6254, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6254(choice, ...)  METALANG99_PRIV_REC_NEXT(6255, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6255(choice, ...)  METALANG99_PRIV_REC_NEXT(6256, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6256(choice, ...)  METALANG99_PRIV_REC_NEXT(6257, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6257(choice, ...)  METALANG99_PRIV_REC_NEXT(6258, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6258(choice, ...)  METALANG99_PRIV_REC_NEXT(6259, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6259(choice, ...)  METALANG99_PRIV_REC_NEXT(6260, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6260(choice, ...)  METALANG99_PRIV_REC_NEXT(6261, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6261(choice, ...)  METALANG99_PRIV_REC_NEXT(6262, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6262(choice, ...)  METALANG99_PRIV_REC_NEXT(6263, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6263(choice, ...)  METALANG99_PRIV_REC_NEXT(6264, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6264(choice, ...)  METALANG99_PRIV_REC_NEXT(6265, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6265(choice, ...)  METALANG99_PRIV_REC_NEXT(6266, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6266(choice, ...)  METALANG99_PRIV_REC_NEXT(6267, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6267(choice, ...)  METALANG99_PRIV_REC_NEXT(6268, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6268(choice, ...)  METALANG99_PRIV_REC_NEXT(6269, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6269(choice, ...)  METALANG99_PRIV_REC_NEXT(6270, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6270(choice, ...)  METALANG99_PRIV_REC_NEXT(6271, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6271(choice, ...)  METALANG99_PRIV_REC_NEXT(6272, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6272(choice, ...)  METALANG99_PRIV_REC_NEXT(6273, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6273(choice, ...)  METALANG99_PRIV_REC_NEXT(6274, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6274(choice, ...)  METALANG99_PRIV_REC_NEXT(6275, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6275(choice, ...)  METALANG99_PRIV_REC_NEXT(6276, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6276(choice, ...)  METALANG99_PRIV_REC_NEXT(6277, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6277(choice, ...)  METALANG99_PRIV_REC_NEXT(6278, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6278(choice, ...)  METALANG99_PRIV_REC_NEXT(6279, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6279(choice, ...)  METALANG99_PRIV_REC_NEXT(6280, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6280(choice, ...)  METALANG99_PRIV_REC_NEXT(6281, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6281(choice, ...)  METALANG99_PRIV_REC_NEXT(6282, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6282(choice, ...)  METALANG99_PRIV_REC_NEXT(6283, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6283(choice, ...)  METALANG99_PRIV_REC_NEXT(6284, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6284(choice, ...)  METALANG99_PRIV_REC_NEXT(6285, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6285(choice, ...)  METALANG99_PRIV_REC_NEXT(6286, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6286(choice, ...)  METALANG99_PRIV_REC_NEXT(6287, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6287(choice, ...)  METALANG99_PRIV_REC_NEXT(6288, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6288(choice, ...)  METALANG99_PRIV_REC_NEXT(6289, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6289(choice, ...)  METALANG99_PRIV_REC_NEXT(6290, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6290(choice, ...)  METALANG99_PRIV_REC_NEXT(6291, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6291(choice, ...)  METALANG99_PRIV_REC_NEXT(6292, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6292(choice, ...)  METALANG99_PRIV_REC_NEXT(6293, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6293(choice, ...)  METALANG99_PRIV_REC_NEXT(6294, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6294(choice, ...)  METALANG99_PRIV_REC_NEXT(6295, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6295(choice, ...)  METALANG99_PRIV_REC_NEXT(6296, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6296(choice, ...)  METALANG99_PRIV_REC_NEXT(6297, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6297(choice, ...)  METALANG99_PRIV_REC_NEXT(6298, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6298(choice, ...)  METALANG99_PRIV_REC_NEXT(6299, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6299(choice, ...)  METALANG99_PRIV_REC_NEXT(6300, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6300(choice, ...)  METALANG99_PRIV_REC_NEXT(6301, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6301(choice, ...)  METALANG99_PRIV_REC_NEXT(6302, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6302(choice, ...)  METALANG99_PRIV_REC_NEXT(6303, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6303(choice, ...)  METALANG99_PRIV_REC_NEXT(6304, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6304(choice, ...)  METALANG99_PRIV_REC_NEXT(6305, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6305(choice, ...)  METALANG99_PRIV_REC_NEXT(6306, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6306(choice, ...)  METALANG99_PRIV_REC_NEXT(6307, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6307(choice, ...)  METALANG99_PRIV_REC_NEXT(6308, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6308(choice, ...)  METALANG99_PRIV_REC_NEXT(6309, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6309(choice, ...)  METALANG99_PRIV_REC_NEXT(6310, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6310(choice, ...)  METALANG99_PRIV_REC_NEXT(6311, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6311(choice, ...)  METALANG99_PRIV_REC_NEXT(6312, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6312(choice, ...)  METALANG99_PRIV_REC_NEXT(6313, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6313(choice, ...)  METALANG99_PRIV_REC_NEXT(6314, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6314(choice, ...)  METALANG99_PRIV_REC_NEXT(6315, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6315(choice, ...)  METALANG99_PRIV_REC_NEXT(6316, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6316(choice, ...)  METALANG99_PRIV_REC_NEXT(6317, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6317(choice, ...)  METALANG99_PRIV_REC_NEXT(6318, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6318(choice, ...)  METALANG99_PRIV_REC_NEXT(6319, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6319(choice, ...)  METALANG99_PRIV_REC_NEXT(6320, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6320(choice, ...)  METALANG99_PRIV_REC_NEXT(6321, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6321(choice, ...)  METALANG99_PRIV_REC_NEXT(6322, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6322(choice, ...)  METALANG99_PRIV_REC_NEXT(6323, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6323(choice, ...)  METALANG99_PRIV_REC_NEXT(6324, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6324(choice, ...)  METALANG99_PRIV_REC_NEXT(6325, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6325(choice, ...)  METALANG99_PRIV_REC_NEXT(6326, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6326(choice, ...)  METALANG99_PRIV_REC_NEXT(6327, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6327(choice, ...)  METALANG99_PRIV_REC_NEXT(6328, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6328(choice, ...)  METALANG99_PRIV_REC_NEXT(6329, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6329(choice, ...)  METALANG99_PRIV_REC_NEXT(6330, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6330(choice, ...)  METALANG99_PRIV_REC_NEXT(6331, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6331(choice, ...)  METALANG99_PRIV_REC_NEXT(6332, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6332(choice, ...)  METALANG99_PRIV_REC_NEXT(6333, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6333(choice, ...)  METALANG99_PRIV_REC_NEXT(6334, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6334(choice, ...)  METALANG99_PRIV_REC_NEXT(6335, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6335(choice, ...)  METALANG99_PRIV_REC_NEXT(6336, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6336(choice, ...)  METALANG99_PRIV_REC_NEXT(6337, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6337(choice, ...)  METALANG99_PRIV_REC_NEXT(6338, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6338(choice, ...)  METALANG99_PRIV_REC_NEXT(6339, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6339(choice, ...)  METALANG99_PRIV_REC_NEXT(6340, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6340(choice, ...)  METALANG99_PRIV_REC_NEXT(6341, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6341(choice, ...)  METALANG99_PRIV_REC_NEXT(6342, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6342(choice, ...)  METALANG99_PRIV_REC_NEXT(6343, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6343(choice, ...)  METALANG99_PRIV_REC_NEXT(6344, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6344(choice, ...)  METALANG99_PRIV_REC_NEXT(6345, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6345(choice, ...)  METALANG99_PRIV_REC_NEXT(6346, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6346(choice, ...)  METALANG99_PRIV_REC_NEXT(6347, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6347(choice, ...)  METALANG99_PRIV_REC_NEXT(6348, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6348(choice, ...)  METALANG99_PRIV_REC_NEXT(6349, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6349(choice, ...)  METALANG99_PRIV_REC_NEXT(6350, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6350(choice, ...)  METALANG99_PRIV_REC_NEXT(6351, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6351(choice, ...)  METALANG99_PRIV_REC_NEXT(6352, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6352(choice, ...)  METALANG99_PRIV_REC_NEXT(6353, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6353(choice, ...)  METALANG99_PRIV_REC_NEXT(6354, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6354(choice, ...)  METALANG99_PRIV_REC_NEXT(6355, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6355(choice, ...)  METALANG99_PRIV_REC_NEXT(6356, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6356(choice, ...)  METALANG99_PRIV_REC_NEXT(6357, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6357(choice, ...)  METALANG99_PRIV_REC_NEXT(6358, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6358(choice, ...)  METALANG99_PRIV_REC_NEXT(6359, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6359(choice, ...)  METALANG99_PRIV_REC_NEXT(6360, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6360(choice, ...)  METALANG99_PRIV_REC_NEXT(6361, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6361(choice, ...)  METALANG99_PRIV_REC_NEXT(6362, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6362(choice, ...)  METALANG99_PRIV_REC_NEXT(6363, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6363(choice, ...)  METALANG99_PRIV_REC_NEXT(6364, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6364(choice, ...)  METALANG99_PRIV_REC_NEXT(6365, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6365(choice, ...)  METALANG99_PRIV_REC_NEXT(6366, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6366(choice, ...)  METALANG99_PRIV_REC_NEXT(6367, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6367(choice, ...)  METALANG99_PRIV_REC_NEXT(6368, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6368(choice, ...)  METALANG99_PRIV_REC_NEXT(6369, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6369(choice, ...)  METALANG99_PRIV_REC_NEXT(6370, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6370(choice, ...)  METALANG99_PRIV_REC_NEXT(6371, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6371(choice, ...)  METALANG99_PRIV_REC_NEXT(6372, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6372(choice, ...)  METALANG99_PRIV_REC_NEXT(6373, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6373(choice, ...)  METALANG99_PRIV_REC_NEXT(6374, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6374(choice, ...)  METALANG99_PRIV_REC_NEXT(6375, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6375(choice, ...)  METALANG99_PRIV_REC_NEXT(6376, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6376(choice, ...)  METALANG99_PRIV_REC_NEXT(6377, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6377(choice, ...)  METALANG99_PRIV_REC_NEXT(6378, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6378(choice, ...)  METALANG99_PRIV_REC_NEXT(6379, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6379(choice, ...)  METALANG99_PRIV_REC_NEXT(6380, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6380(choice, ...)  METALANG99_PRIV_REC_NEXT(6381, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6381(choice, ...)  METALANG99_PRIV_REC_NEXT(6382, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6382(choice, ...)  METALANG99_PRIV_REC_NEXT(6383, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6383(choice, ...)  METALANG99_PRIV_REC_NEXT(6384, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6384(choice, ...)  METALANG99_PRIV_REC_NEXT(6385, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6385(choice, ...)  METALANG99_PRIV_REC_NEXT(6386, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6386(choice, ...)  METALANG99_PRIV_REC_NEXT(6387, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6387(choice, ...)  METALANG99_PRIV_REC_NEXT(6388, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6388(choice, ...)  METALANG99_PRIV_REC_NEXT(6389, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6389(choice, ...)  METALANG99_PRIV_REC_NEXT(6390, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6390(choice, ...)  METALANG99_PRIV_REC_NEXT(6391, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6391(choice, ...)  METALANG99_PRIV_REC_NEXT(6392, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6392(choice, ...)  METALANG99_PRIV_REC_NEXT(6393, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6393(choice, ...)  METALANG99_PRIV_REC_NEXT(6394, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6394(choice, ...)  METALANG99_PRIV_REC_NEXT(6395, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6395(choice, ...)  METALANG99_PRIV_REC_NEXT(6396, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6396(choice, ...)  METALANG99_PRIV_REC_NEXT(6397, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6397(choice, ...)  METALANG99_PRIV_REC_NEXT(6398, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6398(choice, ...)  METALANG99_PRIV_REC_NEXT(6399, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6399(choice, ...)  METALANG99_PRIV_REC_NEXT(6400, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6400(choice, ...)  METALANG99_PRIV_REC_NEXT(6401, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6401(choice, ...)  METALANG99_PRIV_REC_NEXT(6402, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6402(choice, ...)  METALANG99_PRIV_REC_NEXT(6403, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6403(choice, ...)  METALANG99_PRIV_REC_NEXT(6404, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6404(choice, ...)  METALANG99_PRIV_REC_NEXT(6405, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6405(choice, ...)  METALANG99_PRIV_REC_NEXT(6406, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6406(choice, ...)  METALANG99_PRIV_REC_NEXT(6407, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6407(choice, ...)  METALANG99_PRIV_REC_NEXT(6408, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6408(choice, ...)  METALANG99_PRIV_REC_NEXT(6409, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6409(choice, ...)  METALANG99_PRIV_REC_NEXT(6410, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6410(choice, ...)  METALANG99_PRIV_REC_NEXT(6411, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6411(choice, ...)  METALANG99_PRIV_REC_NEXT(6412, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6412(choice, ...)  METALANG99_PRIV_REC_NEXT(6413, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6413(choice, ...)  METALANG99_PRIV_REC_NEXT(6414, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6414(choice, ...)  METALANG99_PRIV_REC_NEXT(6415, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6415(choice, ...)  METALANG99_PRIV_REC_NEXT(6416, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6416(choice, ...)  METALANG99_PRIV_REC_NEXT(6417, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6417(choice, ...)  METALANG99_PRIV_REC_NEXT(6418, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6418(choice, ...)  METALANG99_PRIV_REC_NEXT(6419, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6419(choice, ...)  METALANG99_PRIV_REC_NEXT(6420, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6420(choice, ...)  METALANG99_PRIV_REC_NEXT(6421, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6421(choice, ...)  METALANG99_PRIV_REC_NEXT(6422, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6422(choice, ...)  METALANG99_PRIV_REC_NEXT(6423, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6423(choice, ...)  METALANG99_PRIV_REC_NEXT(6424, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6424(choice, ...)  METALANG99_PRIV_REC_NEXT(6425, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6425(choice, ...)  METALANG99_PRIV_REC_NEXT(6426, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6426(choice, ...)  METALANG99_PRIV_REC_NEXT(6427, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6427(choice, ...)  METALANG99_PRIV_REC_NEXT(6428, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6428(choice, ...)  METALANG99_PRIV_REC_NEXT(6429, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6429(choice, ...)  METALANG99_PRIV_REC_NEXT(6430, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6430(choice, ...)  METALANG99_PRIV_REC_NEXT(6431, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6431(choice, ...)  METALANG99_PRIV_REC_NEXT(6432, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6432(choice, ...)  METALANG99_PRIV_REC_NEXT(6433, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6433(choice, ...)  METALANG99_PRIV_REC_NEXT(6434, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6434(choice, ...)  METALANG99_PRIV_REC_NEXT(6435, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6435(choice, ...)  METALANG99_PRIV_REC_NEXT(6436, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6436(choice, ...)  METALANG99_PRIV_REC_NEXT(6437, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6437(choice, ...)  METALANG99_PRIV_REC_NEXT(6438, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6438(choice, ...)  METALANG99_PRIV_REC_NEXT(6439, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6439(choice, ...)  METALANG99_PRIV_REC_NEXT(6440, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6440(choice, ...)  METALANG99_PRIV_REC_NEXT(6441, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6441(choice, ...)  METALANG99_PRIV_REC_NEXT(6442, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6442(choice, ...)  METALANG99_PRIV_REC_NEXT(6443, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6443(choice, ...)  METALANG99_PRIV_REC_NEXT(6444, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6444(choice, ...)  METALANG99_PRIV_REC_NEXT(6445, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6445(choice, ...)  METALANG99_PRIV_REC_NEXT(6446, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6446(choice, ...)  METALANG99_PRIV_REC_NEXT(6447, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6447(choice, ...)  METALANG99_PRIV_REC_NEXT(6448, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6448(choice, ...)  METALANG99_PRIV_REC_NEXT(6449, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6449(choice, ...)  METALANG99_PRIV_REC_NEXT(6450, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6450(choice, ...)  METALANG99_PRIV_REC_NEXT(6451, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6451(choice, ...)  METALANG99_PRIV_REC_NEXT(6452, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6452(choice, ...)  METALANG99_PRIV_REC_NEXT(6453, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6453(choice, ...)  METALANG99_PRIV_REC_NEXT(6454, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6454(choice, ...)  METALANG99_PRIV_REC_NEXT(6455, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6455(choice, ...)  METALANG99_PRIV_REC_NEXT(6456, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6456(choice, ...)  METALANG99_PRIV_REC_NEXT(6457, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6457(choice, ...)  METALANG99_PRIV_REC_NEXT(6458, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6458(choice, ...)  METALANG99_PRIV_REC_NEXT(6459, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6459(choice, ...)  METALANG99_PRIV_REC_NEXT(6460, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6460(choice, ...)  METALANG99_PRIV_REC_NEXT(6461, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6461(choice, ...)  METALANG99_PRIV_REC_NEXT(6462, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6462(choice, ...)  METALANG99_PRIV_REC_NEXT(6463, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6463(choice, ...)  METALANG99_PRIV_REC_NEXT(6464, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6464(choice, ...)  METALANG99_PRIV_REC_NEXT(6465, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6465(choice, ...)  METALANG99_PRIV_REC_NEXT(6466, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6466(choice, ...)  METALANG99_PRIV_REC_NEXT(6467, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6467(choice, ...)  METALANG99_PRIV_REC_NEXT(6468, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6468(choice, ...)  METALANG99_PRIV_REC_NEXT(6469, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6469(choice, ...)  METALANG99_PRIV_REC_NEXT(6470, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6470(choice, ...)  METALANG99_PRIV_REC_NEXT(6471, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6471(choice, ...)  METALANG99_PRIV_REC_NEXT(6472, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6472(choice, ...)  METALANG99_PRIV_REC_NEXT(6473, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6473(choice, ...)  METALANG99_PRIV_REC_NEXT(6474, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6474(choice, ...)  METALANG99_PRIV_REC_NEXT(6475, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6475(choice, ...)  METALANG99_PRIV_REC_NEXT(6476, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6476(choice, ...)  METALANG99_PRIV_REC_NEXT(6477, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6477(choice, ...)  METALANG99_PRIV_REC_NEXT(6478, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6478(choice, ...)  METALANG99_PRIV_REC_NEXT(6479, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6479(choice, ...)  METALANG99_PRIV_REC_NEXT(6480, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6480(choice, ...)  METALANG99_PRIV_REC_NEXT(6481, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6481(choice, ...)  METALANG99_PRIV_REC_NEXT(6482, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6482(choice, ...)  METALANG99_PRIV_REC_NEXT(6483, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6483(choice, ...)  METALANG99_PRIV_REC_NEXT(6484, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6484(choice, ...)  METALANG99_PRIV_REC_NEXT(6485, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6485(choice, ...)  METALANG99_PRIV_REC_NEXT(6486, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6486(choice, ...)  METALANG99_PRIV_REC_NEXT(6487, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6487(choice, ...)  METALANG99_PRIV_REC_NEXT(6488, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6488(choice, ...)  METALANG99_PRIV_REC_NEXT(6489, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6489(choice, ...)  METALANG99_PRIV_REC_NEXT(6490, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6490(choice, ...)  METALANG99_PRIV_REC_NEXT(6491, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6491(choice, ...)  METALANG99_PRIV_REC_NEXT(6492, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6492(choice, ...)  METALANG99_PRIV_REC_NEXT(6493, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6493(choice, ...)  METALANG99_PRIV_REC_NEXT(6494, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6494(choice, ...)  METALANG99_PRIV_REC_NEXT(6495, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6495(choice, ...)  METALANG99_PRIV_REC_NEXT(6496, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6496(choice, ...)  METALANG99_PRIV_REC_NEXT(6497, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6497(choice, ...)  METALANG99_PRIV_REC_NEXT(6498, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6498(choice, ...)  METALANG99_PRIV_REC_NEXT(6499, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6499(choice, ...)  METALANG99_PRIV_REC_NEXT(6500, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6500(choice, ...)  METALANG99_PRIV_REC_NEXT(6501, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6501(choice, ...)  METALANG99_PRIV_REC_NEXT(6502, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6502(choice, ...)  METALANG99_PRIV_REC_NEXT(6503, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6503(choice, ...)  METALANG99_PRIV_REC_NEXT(6504, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6504(choice, ...)  METALANG99_PRIV_REC_NEXT(6505, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6505(choice, ...)  METALANG99_PRIV_REC_NEXT(6506, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6506(choice, ...)  METALANG99_PRIV_REC_NEXT(6507, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6507(choice, ...)  METALANG99_PRIV_REC_NEXT(6508, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6508(choice, ...)  METALANG99_PRIV_REC_NEXT(6509, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6509(choice, ...)  METALANG99_PRIV_REC_NEXT(6510, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6510(choice, ...)  METALANG99_PRIV_REC_NEXT(6511, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6511(choice, ...)  METALANG99_PRIV_REC_NEXT(6512, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6512(choice, ...)  METALANG99_PRIV_REC_NEXT(6513, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6513(choice, ...)  METALANG99_PRIV_REC_NEXT(6514, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6514(choice, ...)  METALANG99_PRIV_REC_NEXT(6515, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6515(choice, ...)  METALANG99_PRIV_REC_NEXT(6516, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6516(choice, ...)  METALANG99_PRIV_REC_NEXT(6517, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6517(choice, ...)  METALANG99_PRIV_REC_NEXT(6518, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6518(choice, ...)  METALANG99_PRIV_REC_NEXT(6519, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6519(choice, ...)  METALANG99_PRIV_REC_NEXT(6520, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6520(choice, ...)  METALANG99_PRIV_REC_NEXT(6521, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6521(choice, ...)  METALANG99_PRIV_REC_NEXT(6522, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6522(choice, ...)  METALANG99_PRIV_REC_NEXT(6523, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6523(choice, ...)  METALANG99_PRIV_REC_NEXT(6524, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6524(choice, ...)  METALANG99_PRIV_REC_NEXT(6525, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6525(choice, ...)  METALANG99_PRIV_REC_NEXT(6526, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6526(choice, ...)  METALANG99_PRIV_REC_NEXT(6527, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6527(choice, ...)  METALANG99_PRIV_REC_NEXT(6528, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6528(choice, ...)  METALANG99_PRIV_REC_NEXT(6529, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6529(choice, ...)  METALANG99_PRIV_REC_NEXT(6530, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6530(choice, ...)  METALANG99_PRIV_REC_NEXT(6531, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6531(choice, ...)  METALANG99_PRIV_REC_NEXT(6532, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6532(choice, ...)  METALANG99_PRIV_REC_NEXT(6533, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6533(choice, ...)  METALANG99_PRIV_REC_NEXT(6534, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6534(choice, ...)  METALANG99_PRIV_REC_NEXT(6535, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6535(choice, ...)  METALANG99_PRIV_REC_NEXT(6536, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6536(choice, ...)  METALANG99_PRIV_REC_NEXT(6537, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6537(choice, ...)  METALANG99_PRIV_REC_NEXT(6538, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6538(choice, ...)  METALANG99_PRIV_REC_NEXT(6539, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6539(choice, ...)  METALANG99_PRIV_REC_NEXT(6540, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6540(choice, ...)  METALANG99_PRIV_REC_NEXT(6541, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6541(choice, ...)  METALANG99_PRIV_REC_NEXT(6542, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6542(choice, ...)  METALANG99_PRIV_REC_NEXT(6543, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6543(choice, ...)  METALANG99_PRIV_REC_NEXT(6544, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6544(choice, ...)  METALANG99_PRIV_REC_NEXT(6545, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6545(choice, ...)  METALANG99_PRIV_REC_NEXT(6546, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6546(choice, ...)  METALANG99_PRIV_REC_NEXT(6547, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6547(choice, ...)  METALANG99_PRIV_REC_NEXT(6548, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6548(choice, ...)  METALANG99_PRIV_REC_NEXT(6549, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6549(choice, ...)  METALANG99_PRIV_REC_NEXT(6550, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6550(choice, ...)  METALANG99_PRIV_REC_NEXT(6551, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6551(choice, ...)  METALANG99_PRIV_REC_NEXT(6552, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6552(choice, ...)  METALANG99_PRIV_REC_NEXT(6553, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6553(choice, ...)  METALANG99_PRIV_REC_NEXT(6554, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6554(choice, ...)  METALANG99_PRIV_REC_NEXT(6555, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6555(choice, ...)  METALANG99_PRIV_REC_NEXT(6556, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6556(choice, ...)  METALANG99_PRIV_REC_NEXT(6557, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6557(choice, ...)  METALANG99_PRIV_REC_NEXT(6558, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6558(choice, ...)  METALANG99_PRIV_REC_NEXT(6559, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6559(choice, ...)  METALANG99_PRIV_REC_NEXT(6560, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6560(choice, ...)  METALANG99_PRIV_REC_NEXT(6561, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6561(choice, ...)  METALANG99_PRIV_REC_NEXT(6562, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6562(choice, ...)  METALANG99_PRIV_REC_NEXT(6563, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6563(choice, ...)  METALANG99_PRIV_REC_NEXT(6564, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6564(choice, ...)  METALANG99_PRIV_REC_NEXT(6565, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6565(choice, ...)  METALANG99_PRIV_REC_NEXT(6566, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6566(choice, ...)  METALANG99_PRIV_REC_NEXT(6567, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6567(choice, ...)  METALANG99_PRIV_REC_NEXT(6568, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6568(choice, ...)  METALANG99_PRIV_REC_NEXT(6569, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6569(choice, ...)  METALANG99_PRIV_REC_NEXT(6570, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6570(choice, ...)  METALANG99_PRIV_REC_NEXT(6571, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6571(choice, ...)  METALANG99_PRIV_REC_NEXT(6572, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6572(choice, ...)  METALANG99_PRIV_REC_NEXT(6573, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6573(choice, ...)  METALANG99_PRIV_REC_NEXT(6574, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6574(choice, ...)  METALANG99_PRIV_REC_NEXT(6575, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6575(choice, ...)  METALANG99_PRIV_REC_NEXT(6576, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6576(choice, ...)  METALANG99_PRIV_REC_NEXT(6577, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6577(choice, ...)  METALANG99_PRIV_REC_NEXT(6578, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6578(choice, ...)  METALANG99_PRIV_REC_NEXT(6579, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6579(choice, ...)  METALANG99_PRIV_REC_NEXT(6580, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6580(choice, ...)  METALANG99_PRIV_REC_NEXT(6581, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6581(choice, ...)  METALANG99_PRIV_REC_NEXT(6582, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6582(choice, ...)  METALANG99_PRIV_REC_NEXT(6583, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6583(choice, ...)  METALANG99_PRIV_REC_NEXT(6584, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6584(choice, ...)  METALANG99_PRIV_REC_NEXT(6585, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6585(choice, ...)  METALANG99_PRIV_REC_NEXT(6586, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6586(choice, ...)  METALANG99_PRIV_REC_NEXT(6587, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6587(choice, ...)  METALANG99_PRIV_REC_NEXT(6588, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6588(choice, ...)  METALANG99_PRIV_REC_NEXT(6589, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6589(choice, ...)  METALANG99_PRIV_REC_NEXT(6590, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6590(choice, ...)  METALANG99_PRIV_REC_NEXT(6591, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6591(choice, ...)  METALANG99_PRIV_REC_NEXT(6592, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6592(choice, ...)  METALANG99_PRIV_REC_NEXT(6593, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6593(choice, ...)  METALANG99_PRIV_REC_NEXT(6594, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6594(choice, ...)  METALANG99_PRIV_REC_NEXT(6595, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6595(choice, ...)  METALANG99_PRIV_REC_NEXT(6596, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6596(choice, ...)  METALANG99_PRIV_REC_NEXT(6597, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6597(choice, ...)  METALANG99_PRIV_REC_NEXT(6598, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6598(choice, ...)  METALANG99_PRIV_REC_NEXT(6599, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6599(choice, ...)  METALANG99_PRIV_REC_NEXT(6600, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6600(choice, ...)  METALANG99_PRIV_REC_NEXT(6601, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6601(choice, ...)  METALANG99_PRIV_REC_NEXT(6602, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6602(choice, ...)  METALANG99_PRIV_REC_NEXT(6603, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6603(choice, ...)  METALANG99_PRIV_REC_NEXT(6604, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6604(choice, ...)  METALANG99_PRIV_REC_NEXT(6605, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6605(choice, ...)  METALANG99_PRIV_REC_NEXT(6606, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6606(choice, ...)  METALANG99_PRIV_REC_NEXT(6607, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6607(choice, ...)  METALANG99_PRIV_REC_NEXT(6608, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6608(choice, ...)  METALANG99_PRIV_REC_NEXT(6609, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6609(choice, ...)  METALANG99_PRIV_REC_NEXT(6610, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6610(choice, ...)  METALANG99_PRIV_REC_NEXT(6611, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6611(choice, ...)  METALANG99_PRIV_REC_NEXT(6612, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6612(choice, ...)  METALANG99_PRIV_REC_NEXT(6613, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6613(choice, ...)  METALANG99_PRIV_REC_NEXT(6614, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6614(choice, ...)  METALANG99_PRIV_REC_NEXT(6615, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6615(choice, ...)  METALANG99_PRIV_REC_NEXT(6616, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6616(choice, ...)  METALANG99_PRIV_REC_NEXT(6617, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6617(choice, ...)  METALANG99_PRIV_REC_NEXT(6618, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6618(choice, ...)  METALANG99_PRIV_REC_NEXT(6619, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6619(choice, ...)  METALANG99_PRIV_REC_NEXT(6620, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6620(choice, ...)  METALANG99_PRIV_REC_NEXT(6621, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6621(choice, ...)  METALANG99_PRIV_REC_NEXT(6622, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6622(choice, ...)  METALANG99_PRIV_REC_NEXT(6623, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6623(choice, ...)  METALANG99_PRIV_REC_NEXT(6624, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6624(choice, ...)  METALANG99_PRIV_REC_NEXT(6625, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6625(choice, ...)  METALANG99_PRIV_REC_NEXT(6626, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6626(choice, ...)  METALANG99_PRIV_REC_NEXT(6627, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6627(choice, ...)  METALANG99_PRIV_REC_NEXT(6628, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6628(choice, ...)  METALANG99_PRIV_REC_NEXT(6629, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6629(choice, ...)  METALANG99_PRIV_REC_NEXT(6630, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6630(choice, ...)  METALANG99_PRIV_REC_NEXT(6631, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6631(choice, ...)  METALANG99_PRIV_REC_NEXT(6632, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6632(choice, ...)  METALANG99_PRIV_REC_NEXT(6633, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6633(choice, ...)  METALANG99_PRIV_REC_NEXT(6634, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6634(choice, ...)  METALANG99_PRIV_REC_NEXT(6635, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6635(choice, ...)  METALANG99_PRIV_REC_NEXT(6636, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6636(choice, ...)  METALANG99_PRIV_REC_NEXT(6637, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6637(choice, ...)  METALANG99_PRIV_REC_NEXT(6638, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6638(choice, ...)  METALANG99_PRIV_REC_NEXT(6639, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6639(choice, ...)  METALANG99_PRIV_REC_NEXT(6640, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6640(choice, ...)  METALANG99_PRIV_REC_NEXT(6641, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6641(choice, ...)  METALANG99_PRIV_REC_NEXT(6642, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6642(choice, ...)  METALANG99_PRIV_REC_NEXT(6643, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6643(choice, ...)  METALANG99_PRIV_REC_NEXT(6644, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6644(choice, ...)  METALANG99_PRIV_REC_NEXT(6645, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6645(choice, ...)  METALANG99_PRIV_REC_NEXT(6646, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6646(choice, ...)  METALANG99_PRIV_REC_NEXT(6647, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6647(choice, ...)  METALANG99_PRIV_REC_NEXT(6648, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6648(choice, ...)  METALANG99_PRIV_REC_NEXT(6649, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6649(choice, ...)  METALANG99_PRIV_REC_NEXT(6650, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6650(choice, ...)  METALANG99_PRIV_REC_NEXT(6651, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6651(choice, ...)  METALANG99_PRIV_REC_NEXT(6652, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6652(choice, ...)  METALANG99_PRIV_REC_NEXT(6653, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6653(choice, ...)  METALANG99_PRIV_REC_NEXT(6654, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6654(choice, ...)  METALANG99_PRIV_REC_NEXT(6655, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6655(choice, ...)  METALANG99_PRIV_REC_NEXT(6656, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6656(choice, ...)  METALANG99_PRIV_REC_NEXT(6657, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6657(choice, ...)  METALANG99_PRIV_REC_NEXT(6658, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6658(choice, ...)  METALANG99_PRIV_REC_NEXT(6659, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6659(choice, ...)  METALANG99_PRIV_REC_NEXT(6660, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6660(choice, ...)  METALANG99_PRIV_REC_NEXT(6661, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6661(choice, ...)  METALANG99_PRIV_REC_NEXT(6662, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6662(choice, ...)  METALANG99_PRIV_REC_NEXT(6663, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6663(choice, ...)  METALANG99_PRIV_REC_NEXT(6664, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6664(choice, ...)  METALANG99_PRIV_REC_NEXT(6665, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6665(choice, ...)  METALANG99_PRIV_REC_NEXT(6666, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6666(choice, ...)  METALANG99_PRIV_REC_NEXT(6667, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6667(choice, ...)  METALANG99_PRIV_REC_NEXT(6668, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6668(choice, ...)  METALANG99_PRIV_REC_NEXT(6669, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6669(choice, ...)  METALANG99_PRIV_REC_NEXT(6670, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6670(choice, ...)  METALANG99_PRIV_REC_NEXT(6671, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6671(choice, ...)  METALANG99_PRIV_REC_NEXT(6672, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6672(choice, ...)  METALANG99_PRIV_REC_NEXT(6673, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6673(choice, ...)  METALANG99_PRIV_REC_NEXT(6674, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6674(choice, ...)  METALANG99_PRIV_REC_NEXT(6675, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6675(choice, ...)  METALANG99_PRIV_REC_NEXT(6676, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6676(choice, ...)  METALANG99_PRIV_REC_NEXT(6677, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6677(choice, ...)  METALANG99_PRIV_REC_NEXT(6678, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6678(choice, ...)  METALANG99_PRIV_REC_NEXT(6679, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6679(choice, ...)  METALANG99_PRIV_REC_NEXT(6680, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6680(choice, ...)  METALANG99_PRIV_REC_NEXT(6681, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6681(choice, ...)  METALANG99_PRIV_REC_NEXT(6682, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6682(choice, ...)  METALANG99_PRIV_REC_NEXT(6683, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6683(choice, ...)  METALANG99_PRIV_REC_NEXT(6684, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6684(choice, ...)  METALANG99_PRIV_REC_NEXT(6685, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6685(choice, ...)  METALANG99_PRIV_REC_NEXT(6686, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6686(choice, ...)  METALANG99_PRIV_REC_NEXT(6687, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6687(choice, ...)  METALANG99_PRIV_REC_NEXT(6688, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6688(choice, ...)  METALANG99_PRIV_REC_NEXT(6689, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6689(choice, ...)  METALANG99_PRIV_REC_NEXT(6690, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6690(choice, ...)  METALANG99_PRIV_REC_NEXT(6691, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6691(choice, ...)  METALANG99_PRIV_REC_NEXT(6692, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6692(choice, ...)  METALANG99_PRIV_REC_NEXT(6693, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6693(choice, ...)  METALANG99_PRIV_REC_NEXT(6694, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6694(choice, ...)  METALANG99_PRIV_REC_NEXT(6695, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6695(choice, ...)  METALANG99_PRIV_REC_NEXT(6696, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6696(choice, ...)  METALANG99_PRIV_REC_NEXT(6697, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6697(choice, ...)  METALANG99_PRIV_REC_NEXT(6698, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6698(choice, ...)  METALANG99_PRIV_REC_NEXT(6699, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6699(choice, ...)  METALANG99_PRIV_REC_NEXT(6700, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6700(choice, ...)  METALANG99_PRIV_REC_NEXT(6701, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6701(choice, ...)  METALANG99_PRIV_REC_NEXT(6702, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6702(choice, ...)  METALANG99_PRIV_REC_NEXT(6703, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6703(choice, ...)  METALANG99_PRIV_REC_NEXT(6704, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6704(choice, ...)  METALANG99_PRIV_REC_NEXT(6705, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6705(choice, ...)  METALANG99_PRIV_REC_NEXT(6706, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6706(choice, ...)  METALANG99_PRIV_REC_NEXT(6707, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6707(choice, ...)  METALANG99_PRIV_REC_NEXT(6708, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6708(choice, ...)  METALANG99_PRIV_REC_NEXT(6709, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6709(choice, ...)  METALANG99_PRIV_REC_NEXT(6710, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6710(choice, ...)  METALANG99_PRIV_REC_NEXT(6711, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6711(choice, ...)  METALANG99_PRIV_REC_NEXT(6712, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6712(choice, ...)  METALANG99_PRIV_REC_NEXT(6713, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6713(choice, ...)  METALANG99_PRIV_REC_NEXT(6714, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6714(choice, ...)  METALANG99_PRIV_REC_NEXT(6715, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6715(choice, ...)  METALANG99_PRIV_REC_NEXT(6716, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6716(choice, ...)  METALANG99_PRIV_REC_NEXT(6717, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6717(choice, ...)  METALANG99_PRIV_REC_NEXT(6718, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6718(choice, ...)  METALANG99_PRIV_REC_NEXT(6719, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6719(choice, ...)  METALANG99_PRIV_REC_NEXT(6720, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6720(choice, ...)  METALANG99_PRIV_REC_NEXT(6721, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6721(choice, ...)  METALANG99_PRIV_REC_NEXT(6722, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6722(choice, ...)  METALANG99_PRIV_REC_NEXT(6723, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6723(choice, ...)  METALANG99_PRIV_REC_NEXT(6724, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6724(choice, ...)  METALANG99_PRIV_REC_NEXT(6725, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6725(choice, ...)  METALANG99_PRIV_REC_NEXT(6726, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6726(choice, ...)  METALANG99_PRIV_REC_NEXT(6727, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6727(choice, ...)  METALANG99_PRIV_REC_NEXT(6728, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6728(choice, ...)  METALANG99_PRIV_REC_NEXT(6729, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6729(choice, ...)  METALANG99_PRIV_REC_NEXT(6730, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6730(choice, ...)  METALANG99_PRIV_REC_NEXT(6731, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6731(choice, ...)  METALANG99_PRIV_REC_NEXT(6732, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6732(choice, ...)  METALANG99_PRIV_REC_NEXT(6733, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6733(choice, ...)  METALANG99_PRIV_REC_NEXT(6734, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6734(choice, ...)  METALANG99_PRIV_REC_NEXT(6735, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6735(choice, ...)  METALANG99_PRIV_REC_NEXT(6736, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6736(choice, ...)  METALANG99_PRIV_REC_NEXT(6737, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6737(choice, ...)  METALANG99_PRIV_REC_NEXT(6738, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6738(choice, ...)  METALANG99_PRIV_REC_NEXT(6739, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6739(choice, ...)  METALANG99_PRIV_REC_NEXT(6740, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6740(choice, ...)  METALANG99_PRIV_REC_NEXT(6741, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6741(choice, ...)  METALANG99_PRIV_REC_NEXT(6742, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6742(choice, ...)  METALANG99_PRIV_REC_NEXT(6743, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6743(choice, ...)  METALANG99_PRIV_REC_NEXT(6744, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6744(choice, ...)  METALANG99_PRIV_REC_NEXT(6745, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6745(choice, ...)  METALANG99_PRIV_REC_NEXT(6746, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6746(choice, ...)  METALANG99_PRIV_REC_NEXT(6747, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6747(choice, ...)  METALANG99_PRIV_REC_NEXT(6748, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6748(choice, ...)  METALANG99_PRIV_REC_NEXT(6749, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6749(choice, ...)  METALANG99_PRIV_REC_NEXT(6750, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6750(choice, ...)  METALANG99_PRIV_REC_NEXT(6751, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6751(choice, ...)  METALANG99_PRIV_REC_NEXT(6752, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6752(choice, ...)  METALANG99_PRIV_REC_NEXT(6753, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6753(choice, ...)  METALANG99_PRIV_REC_NEXT(6754, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6754(choice, ...)  METALANG99_PRIV_REC_NEXT(6755, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6755(choice, ...)  METALANG99_PRIV_REC_NEXT(6756, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6756(choice, ...)  METALANG99_PRIV_REC_NEXT(6757, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6757(choice, ...)  METALANG99_PRIV_REC_NEXT(6758, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6758(choice, ...)  METALANG99_PRIV_REC_NEXT(6759, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6759(choice, ...)  METALANG99_PRIV_REC_NEXT(6760, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6760(choice, ...)  METALANG99_PRIV_REC_NEXT(6761, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6761(choice, ...)  METALANG99_PRIV_REC_NEXT(6762, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6762(choice, ...)  METALANG99_PRIV_REC_NEXT(6763, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6763(choice, ...)  METALANG99_PRIV_REC_NEXT(6764, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6764(choice, ...)  METALANG99_PRIV_REC_NEXT(6765, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6765(choice, ...)  METALANG99_PRIV_REC_NEXT(6766, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6766(choice, ...)  METALANG99_PRIV_REC_NEXT(6767, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6767(choice, ...)  METALANG99_PRIV_REC_NEXT(6768, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6768(choice, ...)  METALANG99_PRIV_REC_NEXT(6769, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6769(choice, ...)  METALANG99_PRIV_REC_NEXT(6770, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6770(choice, ...)  METALANG99_PRIV_REC_NEXT(6771, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6771(choice, ...)  METALANG99_PRIV_REC_NEXT(6772, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6772(choice, ...)  METALANG99_PRIV_REC_NEXT(6773, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6773(choice, ...)  METALANG99_PRIV_REC_NEXT(6774, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6774(choice, ...)  METALANG99_PRIV_REC_NEXT(6775, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6775(choice, ...)  METALANG99_PRIV_REC_NEXT(6776, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6776(choice, ...)  METALANG99_PRIV_REC_NEXT(6777, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6777(choice, ...)  METALANG99_PRIV_REC_NEXT(6778, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6778(choice, ...)  METALANG99_PRIV_REC_NEXT(6779, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6779(choice, ...)  METALANG99_PRIV_REC_NEXT(6780, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6780(choice, ...)  METALANG99_PRIV_REC_NEXT(6781, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6781(choice, ...)  METALANG99_PRIV_REC_NEXT(6782, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6782(choice, ...)  METALANG99_PRIV_REC_NEXT(6783, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6783(choice, ...)  METALANG99_PRIV_REC_NEXT(6784, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6784(choice, ...)  METALANG99_PRIV_REC_NEXT(6785, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6785(choice, ...)  METALANG99_PRIV_REC_NEXT(6786, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6786(choice, ...)  METALANG99_PRIV_REC_NEXT(6787, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6787(choice, ...)  METALANG99_PRIV_REC_NEXT(6788, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6788(choice, ...)  METALANG99_PRIV_REC_NEXT(6789, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6789(choice, ...)  METALANG99_PRIV_REC_NEXT(6790, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6790(choice, ...)  METALANG99_PRIV_REC_NEXT(6791, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6791(choice, ...)  METALANG99_PRIV_REC_NEXT(6792, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6792(choice, ...)  METALANG99_PRIV_REC_NEXT(6793, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6793(choice, ...)  METALANG99_PRIV_REC_NEXT(6794, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6794(choice, ...)  METALANG99_PRIV_REC_NEXT(6795, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6795(choice, ...)  METALANG99_PRIV_REC_NEXT(6796, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6796(choice, ...)  METALANG99_PRIV_REC_NEXT(6797, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6797(choice, ...)  METALANG99_PRIV_REC_NEXT(6798, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6798(choice, ...)  METALANG99_PRIV_REC_NEXT(6799, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6799(choice, ...)  METALANG99_PRIV_REC_NEXT(6800, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6800(choice, ...)  METALANG99_PRIV_REC_NEXT(6801, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6801(choice, ...)  METALANG99_PRIV_REC_NEXT(6802, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6802(choice, ...)  METALANG99_PRIV_REC_NEXT(6803, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6803(choice, ...)  METALANG99_PRIV_REC_NEXT(6804, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6804(choice, ...)  METALANG99_PRIV_REC_NEXT(6805, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6805(choice, ...)  METALANG99_PRIV_REC_NEXT(6806, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6806(choice, ...)  METALANG99_PRIV_REC_NEXT(6807, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6807(choice, ...)  METALANG99_PRIV_REC_NEXT(6808, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6808(choice, ...)  METALANG99_PRIV_REC_NEXT(6809, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6809(choice, ...)  METALANG99_PRIV_REC_NEXT(6810, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6810(choice, ...)  METALANG99_PRIV_REC_NEXT(6811, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6811(choice, ...)  METALANG99_PRIV_REC_NEXT(6812, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6812(choice, ...)  METALANG99_PRIV_REC_NEXT(6813, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6813(choice, ...)  METALANG99_PRIV_REC_NEXT(6814, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6814(choice, ...)  METALANG99_PRIV_REC_NEXT(6815, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6815(choice, ...)  METALANG99_PRIV_REC_NEXT(6816, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6816(choice, ...)  METALANG99_PRIV_REC_NEXT(6817, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6817(choice, ...)  METALANG99_PRIV_REC_NEXT(6818, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6818(choice, ...)  METALANG99_PRIV_REC_NEXT(6819, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6819(choice, ...)  METALANG99_PRIV_REC_NEXT(6820, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6820(choice, ...)  METALANG99_PRIV_REC_NEXT(6821, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6821(choice, ...)  METALANG99_PRIV_REC_NEXT(6822, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6822(choice, ...)  METALANG99_PRIV_REC_NEXT(6823, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6823(choice, ...)  METALANG99_PRIV_REC_NEXT(6824, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6824(choice, ...)  METALANG99_PRIV_REC_NEXT(6825, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6825(choice, ...)  METALANG99_PRIV_REC_NEXT(6826, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6826(choice, ...)  METALANG99_PRIV_REC_NEXT(6827, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6827(choice, ...)  METALANG99_PRIV_REC_NEXT(6828, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6828(choice, ...)  METALANG99_PRIV_REC_NEXT(6829, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6829(choice, ...)  METALANG99_PRIV_REC_NEXT(6830, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6830(choice, ...)  METALANG99_PRIV_REC_NEXT(6831, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6831(choice, ...)  METALANG99_PRIV_REC_NEXT(6832, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6832(choice, ...)  METALANG99_PRIV_REC_NEXT(6833, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6833(choice, ...)  METALANG99_PRIV_REC_NEXT(6834, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6834(choice, ...)  METALANG99_PRIV_REC_NEXT(6835, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6835(choice, ...)  METALANG99_PRIV_REC_NEXT(6836, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6836(choice, ...)  METALANG99_PRIV_REC_NEXT(6837, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6837(choice, ...)  METALANG99_PRIV_REC_NEXT(6838, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6838(choice, ...)  METALANG99_PRIV_REC_NEXT(6839, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6839(choice, ...)  METALANG99_PRIV_REC_NEXT(6840, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6840(choice, ...)  METALANG99_PRIV_REC_NEXT(6841, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6841(choice, ...)  METALANG99_PRIV_REC_NEXT(6842, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6842(choice, ...)  METALANG99_PRIV_REC_NEXT(6843, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6843(choice, ...)  METALANG99_PRIV_REC_NEXT(6844, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6844(choice, ...)  METALANG99_PRIV_REC_NEXT(6845, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6845(choice, ...)  METALANG99_PRIV_REC_NEXT(6846, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6846(choice, ...)  METALANG99_PRIV_REC_NEXT(6847, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6847(choice, ...)  METALANG99_PRIV_REC_NEXT(6848, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6848(choice, ...)  METALANG99_PRIV_REC_NEXT(6849, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6849(choice, ...)  METALANG99_PRIV_REC_NEXT(6850, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6850(choice, ...)  METALANG99_PRIV_REC_NEXT(6851, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6851(choice, ...)  METALANG99_PRIV_REC_NEXT(6852, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6852(choice, ...)  METALANG99_PRIV_REC_NEXT(6853, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6853(choice, ...)  METALANG99_PRIV_REC_NEXT(6854, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6854(choice, ...)  METALANG99_PRIV_REC_NEXT(6855, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6855(choice, ...)  METALANG99_PRIV_REC_NEXT(6856, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6856(choice, ...)  METALANG99_PRIV_REC_NEXT(6857, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6857(choice, ...)  METALANG99_PRIV_REC_NEXT(6858, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6858(choice, ...)  METALANG99_PRIV_REC_NEXT(6859, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6859(choice, ...)  METALANG99_PRIV_REC_NEXT(6860, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6860(choice, ...)  METALANG99_PRIV_REC_NEXT(6861, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6861(choice, ...)  METALANG99_PRIV_REC_NEXT(6862, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6862(choice, ...)  METALANG99_PRIV_REC_NEXT(6863, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6863(choice, ...)  METALANG99_PRIV_REC_NEXT(6864, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6864(choice, ...)  METALANG99_PRIV_REC_NEXT(6865, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6865(choice, ...)  METALANG99_PRIV_REC_NEXT(6866, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6866(choice, ...)  METALANG99_PRIV_REC_NEXT(6867, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6867(choice, ...)  METALANG99_PRIV_REC_NEXT(6868, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6868(choice, ...)  METALANG99_PRIV_REC_NEXT(6869, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6869(choice, ...)  METALANG99_PRIV_REC_NEXT(6870, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6870(choice, ...)  METALANG99_PRIV_REC_NEXT(6871, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6871(choice, ...)  METALANG99_PRIV_REC_NEXT(6872, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6872(choice, ...)  METALANG99_PRIV_REC_NEXT(6873, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6873(choice, ...)  METALANG99_PRIV_REC_NEXT(6874, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6874(choice, ...)  METALANG99_PRIV_REC_NEXT(6875, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6875(choice, ...)  METALANG99_PRIV_REC_NEXT(6876, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6876(choice, ...)  METALANG99_PRIV_REC_NEXT(6877, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6877(choice, ...)  METALANG99_PRIV_REC_NEXT(6878, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6878(choice, ...)  METALANG99_PRIV_REC_NEXT(6879, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6879(choice, ...)  METALANG99_PRIV_REC_NEXT(6880, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6880(choice, ...)  METALANG99_PRIV_REC_NEXT(6881, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6881(choice, ...)  METALANG99_PRIV_REC_NEXT(6882, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6882(choice, ...)  METALANG99_PRIV_REC_NEXT(6883, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6883(choice, ...)  METALANG99_PRIV_REC_NEXT(6884, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6884(choice, ...)  METALANG99_PRIV_REC_NEXT(6885, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6885(choice, ...)  METALANG99_PRIV_REC_NEXT(6886, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6886(choice, ...)  METALANG99_PRIV_REC_NEXT(6887, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6887(choice, ...)  METALANG99_PRIV_REC_NEXT(6888, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6888(choice, ...)  METALANG99_PRIV_REC_NEXT(6889, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6889(choice, ...)  METALANG99_PRIV_REC_NEXT(6890, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6890(choice, ...)  METALANG99_PRIV_REC_NEXT(6891, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6891(choice, ...)  METALANG99_PRIV_REC_NEXT(6892, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6892(choice, ...)  METALANG99_PRIV_REC_NEXT(6893, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6893(choice, ...)  METALANG99_PRIV_REC_NEXT(6894, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6894(choice, ...)  METALANG99_PRIV_REC_NEXT(6895, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6895(choice, ...)  METALANG99_PRIV_REC_NEXT(6896, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6896(choice, ...)  METALANG99_PRIV_REC_NEXT(6897, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6897(choice, ...)  METALANG99_PRIV_REC_NEXT(6898, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6898(choice, ...)  METALANG99_PRIV_REC_NEXT(6899, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6899(choice, ...)  METALANG99_PRIV_REC_NEXT(6900, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6900(choice, ...)  METALANG99_PRIV_REC_NEXT(6901, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6901(choice, ...)  METALANG99_PRIV_REC_NEXT(6902, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6902(choice, ...)  METALANG99_PRIV_REC_NEXT(6903, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6903(choice, ...)  METALANG99_PRIV_REC_NEXT(6904, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6904(choice, ...)  METALANG99_PRIV_REC_NEXT(6905, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6905(choice, ...)  METALANG99_PRIV_REC_NEXT(6906, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6906(choice, ...)  METALANG99_PRIV_REC_NEXT(6907, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6907(choice, ...)  METALANG99_PRIV_REC_NEXT(6908, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6908(choice, ...)  METALANG99_PRIV_REC_NEXT(6909, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6909(choice, ...)  METALANG99_PRIV_REC_NEXT(6910, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6910(choice, ...)  METALANG99_PRIV_REC_NEXT(6911, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6911(choice, ...)  METALANG99_PRIV_REC_NEXT(6912, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6912(choice, ...)  METALANG99_PRIV_REC_NEXT(6913, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6913(choice, ...)  METALANG99_PRIV_REC_NEXT(6914, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6914(choice, ...)  METALANG99_PRIV_REC_NEXT(6915, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6915(choice, ...)  METALANG99_PRIV_REC_NEXT(6916, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6916(choice, ...)  METALANG99_PRIV_REC_NEXT(6917, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6917(choice, ...)  METALANG99_PRIV_REC_NEXT(6918, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6918(choice, ...)  METALANG99_PRIV_REC_NEXT(6919, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6919(choice, ...)  METALANG99_PRIV_REC_NEXT(6920, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6920(choice, ...)  METALANG99_PRIV_REC_NEXT(6921, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6921(choice, ...)  METALANG99_PRIV_REC_NEXT(6922, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6922(choice, ...)  METALANG99_PRIV_REC_NEXT(6923, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6923(choice, ...)  METALANG99_PRIV_REC_NEXT(6924, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6924(choice, ...)  METALANG99_PRIV_REC_NEXT(6925, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6925(choice, ...)  METALANG99_PRIV_REC_NEXT(6926, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6926(choice, ...)  METALANG99_PRIV_REC_NEXT(6927, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6927(choice, ...)  METALANG99_PRIV_REC_NEXT(6928, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6928(choice, ...)  METALANG99_PRIV_REC_NEXT(6929, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6929(choice, ...)  METALANG99_PRIV_REC_NEXT(6930, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6930(choice, ...)  METALANG99_PRIV_REC_NEXT(6931, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6931(choice, ...)  METALANG99_PRIV_REC_NEXT(6932, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6932(choice, ...)  METALANG99_PRIV_REC_NEXT(6933, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6933(choice, ...)  METALANG99_PRIV_REC_NEXT(6934, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6934(choice, ...)  METALANG99_PRIV_REC_NEXT(6935, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6935(choice, ...)  METALANG99_PRIV_REC_NEXT(6936, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6936(choice, ...)  METALANG99_PRIV_REC_NEXT(6937, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6937(choice, ...)  METALANG99_PRIV_REC_NEXT(6938, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6938(choice, ...)  METALANG99_PRIV_REC_NEXT(6939, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6939(choice, ...)  METALANG99_PRIV_REC_NEXT(6940, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6940(choice, ...)  METALANG99_PRIV_REC_NEXT(6941, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6941(choice, ...)  METALANG99_PRIV_REC_NEXT(6942, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6942(choice, ...)  METALANG99_PRIV_REC_NEXT(6943, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6943(choice, ...)  METALANG99_PRIV_REC_NEXT(6944, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6944(choice, ...)  METALANG99_PRIV_REC_NEXT(6945, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6945(choice, ...)  METALANG99_PRIV_REC_NEXT(6946, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6946(choice, ...)  METALANG99_PRIV_REC_NEXT(6947, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6947(choice, ...)  METALANG99_PRIV_REC_NEXT(6948, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6948(choice, ...)  METALANG99_PRIV_REC_NEXT(6949, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6949(choice, ...)  METALANG99_PRIV_REC_NEXT(6950, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6950(choice, ...)  METALANG99_PRIV_REC_NEXT(6951, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6951(choice, ...)  METALANG99_PRIV_REC_NEXT(6952, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6952(choice, ...)  METALANG99_PRIV_REC_NEXT(6953, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6953(choice, ...)  METALANG99_PRIV_REC_NEXT(6954, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6954(choice, ...)  METALANG99_PRIV_REC_NEXT(6955, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6955(choice, ...)  METALANG99_PRIV_REC_NEXT(6956, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6956(choice, ...)  METALANG99_PRIV_REC_NEXT(6957, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6957(choice, ...)  METALANG99_PRIV_REC_NEXT(6958, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6958(choice, ...)  METALANG99_PRIV_REC_NEXT(6959, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6959(choice, ...)  METALANG99_PRIV_REC_NEXT(6960, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6960(choice, ...)  METALANG99_PRIV_REC_NEXT(6961, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6961(choice, ...)  METALANG99_PRIV_REC_NEXT(6962, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6962(choice, ...)  METALANG99_PRIV_REC_NEXT(6963, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6963(choice, ...)  METALANG99_PRIV_REC_NEXT(6964, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6964(choice, ...)  METALANG99_PRIV_REC_NEXT(6965, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6965(choice, ...)  METALANG99_PRIV_REC_NEXT(6966, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6966(choice, ...)  METALANG99_PRIV_REC_NEXT(6967, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6967(choice, ...)  METALANG99_PRIV_REC_NEXT(6968, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6968(choice, ...)  METALANG99_PRIV_REC_NEXT(6969, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6969(choice, ...)  METALANG99_PRIV_REC_NEXT(6970, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6970(choice, ...)  METALANG99_PRIV_REC_NEXT(6971, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6971(choice, ...)  METALANG99_PRIV_REC_NEXT(6972, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6972(choice, ...)  METALANG99_PRIV_REC_NEXT(6973, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6973(choice, ...)  METALANG99_PRIV_REC_NEXT(6974, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6974(choice, ...)  METALANG99_PRIV_REC_NEXT(6975, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6975(choice, ...)  METALANG99_PRIV_REC_NEXT(6976, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6976(choice, ...)  METALANG99_PRIV_REC_NEXT(6977, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6977(choice, ...)  METALANG99_PRIV_REC_NEXT(6978, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6978(choice, ...)  METALANG99_PRIV_REC_NEXT(6979, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6979(choice, ...)  METALANG99_PRIV_REC_NEXT(6980, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6980(choice, ...)  METALANG99_PRIV_REC_NEXT(6981, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6981(choice, ...)  METALANG99_PRIV_REC_NEXT(6982, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6982(choice, ...)  METALANG99_PRIV_REC_NEXT(6983, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6983(choice, ...)  METALANG99_PRIV_REC_NEXT(6984, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6984(choice, ...)  METALANG99_PRIV_REC_NEXT(6985, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6985(choice, ...)  METALANG99_PRIV_REC_NEXT(6986, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6986(choice, ...)  METALANG99_PRIV_REC_NEXT(6987, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6987(choice, ...)  METALANG99_PRIV_REC_NEXT(6988, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6988(choice, ...)  METALANG99_PRIV_REC_NEXT(6989, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6989(choice, ...)  METALANG99_PRIV_REC_NEXT(6990, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6990(choice, ...)  METALANG99_PRIV_REC_NEXT(6991, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6991(choice, ...)  METALANG99_PRIV_REC_NEXT(6992, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6992(choice, ...)  METALANG99_PRIV_REC_NEXT(6993, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6993(choice, ...)  METALANG99_PRIV_REC_NEXT(6994, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6994(choice, ...)  METALANG99_PRIV_REC_NEXT(6995, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6995(choice, ...)  METALANG99_PRIV_REC_NEXT(6996, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6996(choice, ...)  METALANG99_PRIV_REC_NEXT(6997, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6997(choice, ...)  METALANG99_PRIV_REC_NEXT(6998, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6998(choice, ...)  METALANG99_PRIV_REC_NEXT(6999, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_6999(choice, ...)  METALANG99_PRIV_REC_NEXT(7000, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7000(choice, ...)  METALANG99_PRIV_REC_NEXT(7001, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7001(choice, ...)  METALANG99_PRIV_REC_NEXT(7002, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7002(choice, ...)  METALANG99_PRIV_REC_NEXT(7003, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7003(choice, ...)  METALANG99_PRIV_REC_NEXT(7004, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7004(choice, ...)  METALANG99_PRIV_REC_NEXT(7005, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7005(choice, ...)  METALANG99_PRIV_REC_NEXT(7006, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7006(choice, ...)  METALANG99_PRIV_REC_NEXT(7007, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7007(choice, ...)  METALANG99_PRIV_REC_NEXT(7008, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7008(choice, ...)  METALANG99_PRIV_REC_NEXT(7009, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7009(choice, ...)  METALANG99_PRIV_REC_NEXT(7010, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7010(choice, ...)  METALANG99_PRIV_REC_NEXT(7011, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7011(choice, ...)  METALANG99_PRIV_REC_NEXT(7012, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7012(choice, ...)  METALANG99_PRIV_REC_NEXT(7013, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7013(choice, ...)  METALANG99_PRIV_REC_NEXT(7014, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7014(choice, ...)  METALANG99_PRIV_REC_NEXT(7015, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7015(choice, ...)  METALANG99_PRIV_REC_NEXT(7016, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7016(choice, ...)  METALANG99_PRIV_REC_NEXT(7017, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7017(choice, ...)  METALANG99_PRIV_REC_NEXT(7018, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7018(choice, ...)  METALANG99_PRIV_REC_NEXT(7019, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7019(choice, ...)  METALANG99_PRIV_REC_NEXT(7020, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7020(choice, ...)  METALANG99_PRIV_REC_NEXT(7021, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7021(choice, ...)  METALANG99_PRIV_REC_NEXT(7022, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7022(choice, ...)  METALANG99_PRIV_REC_NEXT(7023, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7023(choice, ...)  METALANG99_PRIV_REC_NEXT(7024, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7024(choice, ...)  METALANG99_PRIV_REC_NEXT(7025, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7025(choice, ...)  METALANG99_PRIV_REC_NEXT(7026, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7026(choice, ...)  METALANG99_PRIV_REC_NEXT(7027, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7027(choice, ...)  METALANG99_PRIV_REC_NEXT(7028, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7028(choice, ...)  METALANG99_PRIV_REC_NEXT(7029, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7029(choice, ...)  METALANG99_PRIV_REC_NEXT(7030, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7030(choice, ...)  METALANG99_PRIV_REC_NEXT(7031, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7031(choice, ...)  METALANG99_PRIV_REC_NEXT(7032, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7032(choice, ...)  METALANG99_PRIV_REC_NEXT(7033, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7033(choice, ...)  METALANG99_PRIV_REC_NEXT(7034, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7034(choice, ...)  METALANG99_PRIV_REC_NEXT(7035, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7035(choice, ...)  METALANG99_PRIV_REC_NEXT(7036, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7036(choice, ...)  METALANG99_PRIV_REC_NEXT(7037, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7037(choice, ...)  METALANG99_PRIV_REC_NEXT(7038, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7038(choice, ...)  METALANG99_PRIV_REC_NEXT(7039, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7039(choice, ...)  METALANG99_PRIV_REC_NEXT(7040, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7040(choice, ...)  METALANG99_PRIV_REC_NEXT(7041, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7041(choice, ...)  METALANG99_PRIV_REC_NEXT(7042, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7042(choice, ...)  METALANG99_PRIV_REC_NEXT(7043, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7043(choice, ...)  METALANG99_PRIV_REC_NEXT(7044, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7044(choice, ...)  METALANG99_PRIV_REC_NEXT(7045, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7045(choice, ...)  METALANG99_PRIV_REC_NEXT(7046, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7046(choice, ...)  METALANG99_PRIV_REC_NEXT(7047, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7047(choice, ...)  METALANG99_PRIV_REC_NEXT(7048, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7048(choice, ...)  METALANG99_PRIV_REC_NEXT(7049, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7049(choice, ...)  METALANG99_PRIV_REC_NEXT(7050, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7050(choice, ...)  METALANG99_PRIV_REC_NEXT(7051, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7051(choice, ...)  METALANG99_PRIV_REC_NEXT(7052, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7052(choice, ...)  METALANG99_PRIV_REC_NEXT(7053, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7053(choice, ...)  METALANG99_PRIV_REC_NEXT(7054, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7054(choice, ...)  METALANG99_PRIV_REC_NEXT(7055, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7055(choice, ...)  METALANG99_PRIV_REC_NEXT(7056, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7056(choice, ...)  METALANG99_PRIV_REC_NEXT(7057, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7057(choice, ...)  METALANG99_PRIV_REC_NEXT(7058, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7058(choice, ...)  METALANG99_PRIV_REC_NEXT(7059, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7059(choice, ...)  METALANG99_PRIV_REC_NEXT(7060, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7060(choice, ...)  METALANG99_PRIV_REC_NEXT(7061, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7061(choice, ...)  METALANG99_PRIV_REC_NEXT(7062, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7062(choice, ...)  METALANG99_PRIV_REC_NEXT(7063, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7063(choice, ...)  METALANG99_PRIV_REC_NEXT(7064, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7064(choice, ...)  METALANG99_PRIV_REC_NEXT(7065, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7065(choice, ...)  METALANG99_PRIV_REC_NEXT(7066, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7066(choice, ...)  METALANG99_PRIV_REC_NEXT(7067, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7067(choice, ...)  METALANG99_PRIV_REC_NEXT(7068, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7068(choice, ...)  METALANG99_PRIV_REC_NEXT(7069, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7069(choice, ...)  METALANG99_PRIV_REC_NEXT(7070, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7070(choice, ...)  METALANG99_PRIV_REC_NEXT(7071, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7071(choice, ...)  METALANG99_PRIV_REC_NEXT(7072, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7072(choice, ...)  METALANG99_PRIV_REC_NEXT(7073, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7073(choice, ...)  METALANG99_PRIV_REC_NEXT(7074, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7074(choice, ...)  METALANG99_PRIV_REC_NEXT(7075, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7075(choice, ...)  METALANG99_PRIV_REC_NEXT(7076, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7076(choice, ...)  METALANG99_PRIV_REC_NEXT(7077, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7077(choice, ...)  METALANG99_PRIV_REC_NEXT(7078, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7078(choice, ...)  METALANG99_PRIV_REC_NEXT(7079, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7079(choice, ...)  METALANG99_PRIV_REC_NEXT(7080, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7080(choice, ...)  METALANG99_PRIV_REC_NEXT(7081, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7081(choice, ...)  METALANG99_PRIV_REC_NEXT(7082, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7082(choice, ...)  METALANG99_PRIV_REC_NEXT(7083, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7083(choice, ...)  METALANG99_PRIV_REC_NEXT(7084, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7084(choice, ...)  METALANG99_PRIV_REC_NEXT(7085, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7085(choice, ...)  METALANG99_PRIV_REC_NEXT(7086, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7086(choice, ...)  METALANG99_PRIV_REC_NEXT(7087, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7087(choice, ...)  METALANG99_PRIV_REC_NEXT(7088, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7088(choice, ...)  METALANG99_PRIV_REC_NEXT(7089, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7089(choice, ...)  METALANG99_PRIV_REC_NEXT(7090, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7090(choice, ...)  METALANG99_PRIV_REC_NEXT(7091, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7091(choice, ...)  METALANG99_PRIV_REC_NEXT(7092, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7092(choice, ...)  METALANG99_PRIV_REC_NEXT(7093, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7093(choice, ...)  METALANG99_PRIV_REC_NEXT(7094, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7094(choice, ...)  METALANG99_PRIV_REC_NEXT(7095, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7095(choice, ...)  METALANG99_PRIV_REC_NEXT(7096, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7096(choice, ...)  METALANG99_PRIV_REC_NEXT(7097, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7097(choice, ...)  METALANG99_PRIV_REC_NEXT(7098, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7098(choice, ...)  METALANG99_PRIV_REC_NEXT(7099, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7099(choice, ...)  METALANG99_PRIV_REC_NEXT(7100, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7100(choice, ...)  METALANG99_PRIV_REC_NEXT(7101, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7101(choice, ...)  METALANG99_PRIV_REC_NEXT(7102, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7102(choice, ...)  METALANG99_PRIV_REC_NEXT(7103, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7103(choice, ...)  METALANG99_PRIV_REC_NEXT(7104, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7104(choice, ...)  METALANG99_PRIV_REC_NEXT(7105, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7105(choice, ...)  METALANG99_PRIV_REC_NEXT(7106, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7106(choice, ...)  METALANG99_PRIV_REC_NEXT(7107, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7107(choice, ...)  METALANG99_PRIV_REC_NEXT(7108, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7108(choice, ...)  METALANG99_PRIV_REC_NEXT(7109, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7109(choice, ...)  METALANG99_PRIV_REC_NEXT(7110, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7110(choice, ...)  METALANG99_PRIV_REC_NEXT(7111, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7111(choice, ...)  METALANG99_PRIV_REC_NEXT(7112, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7112(choice, ...)  METALANG99_PRIV_REC_NEXT(7113, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7113(choice, ...)  METALANG99_PRIV_REC_NEXT(7114, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7114(choice, ...)  METALANG99_PRIV_REC_NEXT(7115, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7115(choice, ...)  METALANG99_PRIV_REC_NEXT(7116, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7116(choice, ...)  METALANG99_PRIV_REC_NEXT(7117, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7117(choice, ...)  METALANG99_PRIV_REC_NEXT(7118, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7118(choice, ...)  METALANG99_PRIV_REC_NEXT(7119, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7119(choice, ...)  METALANG99_PRIV_REC_NEXT(7120, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7120(choice, ...)  METALANG99_PRIV_REC_NEXT(7121, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7121(choice, ...)  METALANG99_PRIV_REC_NEXT(7122, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7122(choice, ...)  METALANG99_PRIV_REC_NEXT(7123, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7123(choice, ...)  METALANG99_PRIV_REC_NEXT(7124, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7124(choice, ...)  METALANG99_PRIV_REC_NEXT(7125, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7125(choice, ...)  METALANG99_PRIV_REC_NEXT(7126, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7126(choice, ...)  METALANG99_PRIV_REC_NEXT(7127, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7127(choice, ...)  METALANG99_PRIV_REC_NEXT(7128, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7128(choice, ...)  METALANG99_PRIV_REC_NEXT(7129, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7129(choice, ...)  METALANG99_PRIV_REC_NEXT(7130, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7130(choice, ...)  METALANG99_PRIV_REC_NEXT(7131, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7131(choice, ...)  METALANG99_PRIV_REC_NEXT(7132, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7132(choice, ...)  METALANG99_PRIV_REC_NEXT(7133, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7133(choice, ...)  METALANG99_PRIV_REC_NEXT(7134, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7134(choice, ...)  METALANG99_PRIV_REC_NEXT(7135, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7135(choice, ...)  METALANG99_PRIV_REC_NEXT(7136, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7136(choice, ...)  METALANG99_PRIV_REC_NEXT(7137, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7137(choice, ...)  METALANG99_PRIV_REC_NEXT(7138, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7138(choice, ...)  METALANG99_PRIV_REC_NEXT(7139, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7139(choice, ...)  METALANG99_PRIV_REC_NEXT(7140, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7140(choice, ...)  METALANG99_PRIV_REC_NEXT(7141, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7141(choice, ...)  METALANG99_PRIV_REC_NEXT(7142, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7142(choice, ...)  METALANG99_PRIV_REC_NEXT(7143, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7143(choice, ...)  METALANG99_PRIV_REC_NEXT(7144, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7144(choice, ...)  METALANG99_PRIV_REC_NEXT(7145, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7145(choice, ...)  METALANG99_PRIV_REC_NEXT(7146, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7146(choice, ...)  METALANG99_PRIV_REC_NEXT(7147, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7147(choice, ...)  METALANG99_PRIV_REC_NEXT(7148, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7148(choice, ...)  METALANG99_PRIV_REC_NEXT(7149, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7149(choice, ...)  METALANG99_PRIV_REC_NEXT(7150, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7150(choice, ...)  METALANG99_PRIV_REC_NEXT(7151, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7151(choice, ...)  METALANG99_PRIV_REC_NEXT(7152, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7152(choice, ...)  METALANG99_PRIV_REC_NEXT(7153, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7153(choice, ...)  METALANG99_PRIV_REC_NEXT(7154, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7154(choice, ...)  METALANG99_PRIV_REC_NEXT(7155, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7155(choice, ...)  METALANG99_PRIV_REC_NEXT(7156, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7156(choice, ...)  METALANG99_PRIV_REC_NEXT(7157, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7157(choice, ...)  METALANG99_PRIV_REC_NEXT(7158, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7158(choice, ...)  METALANG99_PRIV_REC_NEXT(7159, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7159(choice, ...)  METALANG99_PRIV_REC_NEXT(7160, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7160(choice, ...)  METALANG99_PRIV_REC_NEXT(7161, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7161(choice, ...)  METALANG99_PRIV_REC_NEXT(7162, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7162(choice, ...)  METALANG99_PRIV_REC_NEXT(7163, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7163(choice, ...)  METALANG99_PRIV_REC_NEXT(7164, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7164(choice, ...)  METALANG99_PRIV_REC_NEXT(7165, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7165(choice, ...)  METALANG99_PRIV_REC_NEXT(7166, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7166(choice, ...)  METALANG99_PRIV_REC_NEXT(7167, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7167(choice, ...)  METALANG99_PRIV_REC_NEXT(7168, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7168(choice, ...)  METALANG99_PRIV_REC_NEXT(7169, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7169(choice, ...)  METALANG99_PRIV_REC_NEXT(7170, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7170(choice, ...)  METALANG99_PRIV_REC_NEXT(7171, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7171(choice, ...)  METALANG99_PRIV_REC_NEXT(7172, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7172(choice, ...)  METALANG99_PRIV_REC_NEXT(7173, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7173(choice, ...)  METALANG99_PRIV_REC_NEXT(7174, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7174(choice, ...)  METALANG99_PRIV_REC_NEXT(7175, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7175(choice, ...)  METALANG99_PRIV_REC_NEXT(7176, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7176(choice, ...)  METALANG99_PRIV_REC_NEXT(7177, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7177(choice, ...)  METALANG99_PRIV_REC_NEXT(7178, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7178(choice, ...)  METALANG99_PRIV_REC_NEXT(7179, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7179(choice, ...)  METALANG99_PRIV_REC_NEXT(7180, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7180(choice, ...)  METALANG99_PRIV_REC_NEXT(7181, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7181(choice, ...)  METALANG99_PRIV_REC_NEXT(7182, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7182(choice, ...)  METALANG99_PRIV_REC_NEXT(7183, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7183(choice, ...)  METALANG99_PRIV_REC_NEXT(7184, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7184(choice, ...)  METALANG99_PRIV_REC_NEXT(7185, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7185(choice, ...)  METALANG99_PRIV_REC_NEXT(7186, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7186(choice, ...)  METALANG99_PRIV_REC_NEXT(7187, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7187(choice, ...)  METALANG99_PRIV_REC_NEXT(7188, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7188(choice, ...)  METALANG99_PRIV_REC_NEXT(7189, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7189(choice, ...)  METALANG99_PRIV_REC_NEXT(7190, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7190(choice, ...)  METALANG99_PRIV_REC_NEXT(7191, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7191(choice, ...)  METALANG99_PRIV_REC_NEXT(7192, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7192(choice, ...)  METALANG99_PRIV_REC_NEXT(7193, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7193(choice, ...)  METALANG99_PRIV_REC_NEXT(7194, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7194(choice, ...)  METALANG99_PRIV_REC_NEXT(7195, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7195(choice, ...)  METALANG99_PRIV_REC_NEXT(7196, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7196(choice, ...)  METALANG99_PRIV_REC_NEXT(7197, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7197(choice, ...)  METALANG99_PRIV_REC_NEXT(7198, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7198(choice, ...)  METALANG99_PRIV_REC_NEXT(7199, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7199(choice, ...)  METALANG99_PRIV_REC_NEXT(7200, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7200(choice, ...)  METALANG99_PRIV_REC_NEXT(7201, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7201(choice, ...)  METALANG99_PRIV_REC_NEXT(7202, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7202(choice, ...)  METALANG99_PRIV_REC_NEXT(7203, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7203(choice, ...)  METALANG99_PRIV_REC_NEXT(7204, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7204(choice, ...)  METALANG99_PRIV_REC_NEXT(7205, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7205(choice, ...)  METALANG99_PRIV_REC_NEXT(7206, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7206(choice, ...)  METALANG99_PRIV_REC_NEXT(7207, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7207(choice, ...)  METALANG99_PRIV_REC_NEXT(7208, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7208(choice, ...)  METALANG99_PRIV_REC_NEXT(7209, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7209(choice, ...)  METALANG99_PRIV_REC_NEXT(7210, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7210(choice, ...)  METALANG99_PRIV_REC_NEXT(7211, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7211(choice, ...)  METALANG99_PRIV_REC_NEXT(7212, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7212(choice, ...)  METALANG99_PRIV_REC_NEXT(7213, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7213(choice, ...)  METALANG99_PRIV_REC_NEXT(7214, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7214(choice, ...)  METALANG99_PRIV_REC_NEXT(7215, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7215(choice, ...)  METALANG99_PRIV_REC_NEXT(7216, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7216(choice, ...)  METALANG99_PRIV_REC_NEXT(7217, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7217(choice, ...)  METALANG99_PRIV_REC_NEXT(7218, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7218(choice, ...)  METALANG99_PRIV_REC_NEXT(7219, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7219(choice, ...)  METALANG99_PRIV_REC_NEXT(7220, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7220(choice, ...)  METALANG99_PRIV_REC_NEXT(7221, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7221(choice, ...)  METALANG99_PRIV_REC_NEXT(7222, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7222(choice, ...)  METALANG99_PRIV_REC_NEXT(7223, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7223(choice, ...)  METALANG99_PRIV_REC_NEXT(7224, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7224(choice, ...)  METALANG99_PRIV_REC_NEXT(7225, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7225(choice, ...)  METALANG99_PRIV_REC_NEXT(7226, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7226(choice, ...)  METALANG99_PRIV_REC_NEXT(7227, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7227(choice, ...)  METALANG99_PRIV_REC_NEXT(7228, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7228(choice, ...)  METALANG99_PRIV_REC_NEXT(7229, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7229(choice, ...)  METALANG99_PRIV_REC_NEXT(7230, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7230(choice, ...)  METALANG99_PRIV_REC_NEXT(7231, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7231(choice, ...)  METALANG99_PRIV_REC_NEXT(7232, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7232(choice, ...)  METALANG99_PRIV_REC_NEXT(7233, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7233(choice, ...)  METALANG99_PRIV_REC_NEXT(7234, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7234(choice, ...)  METALANG99_PRIV_REC_NEXT(7235, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7235(choice, ...)  METALANG99_PRIV_REC_NEXT(7236, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7236(choice, ...)  METALANG99_PRIV_REC_NEXT(7237, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7237(choice, ...)  METALANG99_PRIV_REC_NEXT(7238, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7238(choice, ...)  METALANG99_PRIV_REC_NEXT(7239, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7239(choice, ...)  METALANG99_PRIV_REC_NEXT(7240, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7240(choice, ...)  METALANG99_PRIV_REC_NEXT(7241, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7241(choice, ...)  METALANG99_PRIV_REC_NEXT(7242, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7242(choice, ...)  METALANG99_PRIV_REC_NEXT(7243, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7243(choice, ...)  METALANG99_PRIV_REC_NEXT(7244, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7244(choice, ...)  METALANG99_PRIV_REC_NEXT(7245, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7245(choice, ...)  METALANG99_PRIV_REC_NEXT(7246, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7246(choice, ...)  METALANG99_PRIV_REC_NEXT(7247, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7247(choice, ...)  METALANG99_PRIV_REC_NEXT(7248, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7248(choice, ...)  METALANG99_PRIV_REC_NEXT(7249, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7249(choice, ...)  METALANG99_PRIV_REC_NEXT(7250, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7250(choice, ...)  METALANG99_PRIV_REC_NEXT(7251, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7251(choice, ...)  METALANG99_PRIV_REC_NEXT(7252, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7252(choice, ...)  METALANG99_PRIV_REC_NEXT(7253, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7253(choice, ...)  METALANG99_PRIV_REC_NEXT(7254, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7254(choice, ...)  METALANG99_PRIV_REC_NEXT(7255, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7255(choice, ...)  METALANG99_PRIV_REC_NEXT(7256, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7256(choice, ...)  METALANG99_PRIV_REC_NEXT(7257, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7257(choice, ...)  METALANG99_PRIV_REC_NEXT(7258, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7258(choice, ...)  METALANG99_PRIV_REC_NEXT(7259, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7259(choice, ...)  METALANG99_PRIV_REC_NEXT(7260, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7260(choice, ...)  METALANG99_PRIV_REC_NEXT(7261, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7261(choice, ...)  METALANG99_PRIV_REC_NEXT(7262, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7262(choice, ...)  METALANG99_PRIV_REC_NEXT(7263, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7263(choice, ...)  METALANG99_PRIV_REC_NEXT(7264, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7264(choice, ...)  METALANG99_PRIV_REC_NEXT(7265, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7265(choice, ...)  METALANG99_PRIV_REC_NEXT(7266, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7266(choice, ...)  METALANG99_PRIV_REC_NEXT(7267, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7267(choice, ...)  METALANG99_PRIV_REC_NEXT(7268, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7268(choice, ...)  METALANG99_PRIV_REC_NEXT(7269, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7269(choice, ...)  METALANG99_PRIV_REC_NEXT(7270, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7270(choice, ...)  METALANG99_PRIV_REC_NEXT(7271, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7271(choice, ...)  METALANG99_PRIV_REC_NEXT(7272, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7272(choice, ...)  METALANG99_PRIV_REC_NEXT(7273, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7273(choice, ...)  METALANG99_PRIV_REC_NEXT(7274, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7274(choice, ...)  METALANG99_PRIV_REC_NEXT(7275, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7275(choice, ...)  METALANG99_PRIV_REC_NEXT(7276, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7276(choice, ...)  METALANG99_PRIV_REC_NEXT(7277, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7277(choice, ...)  METALANG99_PRIV_REC_NEXT(7278, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7278(choice, ...)  METALANG99_PRIV_REC_NEXT(7279, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7279(choice, ...)  METALANG99_PRIV_REC_NEXT(7280, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7280(choice, ...)  METALANG99_PRIV_REC_NEXT(7281, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7281(choice, ...)  METALANG99_PRIV_REC_NEXT(7282, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7282(choice, ...)  METALANG99_PRIV_REC_NEXT(7283, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7283(choice, ...)  METALANG99_PRIV_REC_NEXT(7284, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7284(choice, ...)  METALANG99_PRIV_REC_NEXT(7285, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7285(choice, ...)  METALANG99_PRIV_REC_NEXT(7286, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7286(choice, ...)  METALANG99_PRIV_REC_NEXT(7287, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7287(choice, ...)  METALANG99_PRIV_REC_NEXT(7288, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7288(choice, ...)  METALANG99_PRIV_REC_NEXT(7289, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7289(choice, ...)  METALANG99_PRIV_REC_NEXT(7290, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7290(choice, ...)  METALANG99_PRIV_REC_NEXT(7291, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7291(choice, ...)  METALANG99_PRIV_REC_NEXT(7292, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7292(choice, ...)  METALANG99_PRIV_REC_NEXT(7293, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7293(choice, ...)  METALANG99_PRIV_REC_NEXT(7294, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7294(choice, ...)  METALANG99_PRIV_REC_NEXT(7295, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7295(choice, ...)  METALANG99_PRIV_REC_NEXT(7296, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7296(choice, ...)  METALANG99_PRIV_REC_NEXT(7297, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7297(choice, ...)  METALANG99_PRIV_REC_NEXT(7298, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7298(choice, ...)  METALANG99_PRIV_REC_NEXT(7299, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7299(choice, ...)  METALANG99_PRIV_REC_NEXT(7300, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7300(choice, ...)  METALANG99_PRIV_REC_NEXT(7301, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7301(choice, ...)  METALANG99_PRIV_REC_NEXT(7302, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7302(choice, ...)  METALANG99_PRIV_REC_NEXT(7303, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7303(choice, ...)  METALANG99_PRIV_REC_NEXT(7304, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7304(choice, ...)  METALANG99_PRIV_REC_NEXT(7305, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7305(choice, ...)  METALANG99_PRIV_REC_NEXT(7306, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7306(choice, ...)  METALANG99_PRIV_REC_NEXT(7307, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7307(choice, ...)  METALANG99_PRIV_REC_NEXT(7308, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7308(choice, ...)  METALANG99_PRIV_REC_NEXT(7309, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7309(choice, ...)  METALANG99_PRIV_REC_NEXT(7310, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7310(choice, ...)  METALANG99_PRIV_REC_NEXT(7311, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7311(choice, ...)  METALANG99_PRIV_REC_NEXT(7312, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7312(choice, ...)  METALANG99_PRIV_REC_NEXT(7313, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7313(choice, ...)  METALANG99_PRIV_REC_NEXT(7314, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7314(choice, ...)  METALANG99_PRIV_REC_NEXT(7315, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7315(choice, ...)  METALANG99_PRIV_REC_NEXT(7316, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7316(choice, ...)  METALANG99_PRIV_REC_NEXT(7317, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7317(choice, ...)  METALANG99_PRIV_REC_NEXT(7318, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7318(choice, ...)  METALANG99_PRIV_REC_NEXT(7319, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7319(choice, ...)  METALANG99_PRIV_REC_NEXT(7320, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7320(choice, ...)  METALANG99_PRIV_REC_NEXT(7321, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7321(choice, ...)  METALANG99_PRIV_REC_NEXT(7322, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7322(choice, ...)  METALANG99_PRIV_REC_NEXT(7323, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7323(choice, ...)  METALANG99_PRIV_REC_NEXT(7324, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7324(choice, ...)  METALANG99_PRIV_REC_NEXT(7325, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7325(choice, ...)  METALANG99_PRIV_REC_NEXT(7326, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7326(choice, ...)  METALANG99_PRIV_REC_NEXT(7327, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7327(choice, ...)  METALANG99_PRIV_REC_NEXT(7328, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7328(choice, ...)  METALANG99_PRIV_REC_NEXT(7329, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7329(choice, ...)  METALANG99_PRIV_REC_NEXT(7330, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7330(choice, ...)  METALANG99_PRIV_REC_NEXT(7331, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7331(choice, ...)  METALANG99_PRIV_REC_NEXT(7332, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7332(choice, ...)  METALANG99_PRIV_REC_NEXT(7333, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7333(choice, ...)  METALANG99_PRIV_REC_NEXT(7334, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7334(choice, ...)  METALANG99_PRIV_REC_NEXT(7335, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7335(choice, ...)  METALANG99_PRIV_REC_NEXT(7336, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7336(choice, ...)  METALANG99_PRIV_REC_NEXT(7337, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7337(choice, ...)  METALANG99_PRIV_REC_NEXT(7338, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7338(choice, ...)  METALANG99_PRIV_REC_NEXT(7339, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7339(choice, ...)  METALANG99_PRIV_REC_NEXT(7340, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7340(choice, ...)  METALANG99_PRIV_REC_NEXT(7341, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7341(choice, ...)  METALANG99_PRIV_REC_NEXT(7342, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7342(choice, ...)  METALANG99_PRIV_REC_NEXT(7343, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7343(choice, ...)  METALANG99_PRIV_REC_NEXT(7344, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7344(choice, ...)  METALANG99_PRIV_REC_NEXT(7345, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7345(choice, ...)  METALANG99_PRIV_REC_NEXT(7346, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7346(choice, ...)  METALANG99_PRIV_REC_NEXT(7347, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7347(choice, ...)  METALANG99_PRIV_REC_NEXT(7348, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7348(choice, ...)  METALANG99_PRIV_REC_NEXT(7349, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7349(choice, ...)  METALANG99_PRIV_REC_NEXT(7350, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7350(choice, ...)  METALANG99_PRIV_REC_NEXT(7351, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7351(choice, ...)  METALANG99_PRIV_REC_NEXT(7352, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7352(choice, ...)  METALANG99_PRIV_REC_NEXT(7353, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7353(choice, ...)  METALANG99_PRIV_REC_NEXT(7354, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7354(choice, ...)  METALANG99_PRIV_REC_NEXT(7355, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7355(choice, ...)  METALANG99_PRIV_REC_NEXT(7356, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7356(choice, ...)  METALANG99_PRIV_REC_NEXT(7357, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7357(choice, ...)  METALANG99_PRIV_REC_NEXT(7358, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7358(choice, ...)  METALANG99_PRIV_REC_NEXT(7359, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7359(choice, ...)  METALANG99_PRIV_REC_NEXT(7360, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7360(choice, ...)  METALANG99_PRIV_REC_NEXT(7361, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7361(choice, ...)  METALANG99_PRIV_REC_NEXT(7362, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7362(choice, ...)  METALANG99_PRIV_REC_NEXT(7363, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7363(choice, ...)  METALANG99_PRIV_REC_NEXT(7364, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7364(choice, ...)  METALANG99_PRIV_REC_NEXT(7365, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7365(choice, ...)  METALANG99_PRIV_REC_NEXT(7366, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7366(choice, ...)  METALANG99_PRIV_REC_NEXT(7367, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7367(choice, ...)  METALANG99_PRIV_REC_NEXT(7368, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7368(choice, ...)  METALANG99_PRIV_REC_NEXT(7369, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7369(choice, ...)  METALANG99_PRIV_REC_NEXT(7370, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7370(choice, ...)  METALANG99_PRIV_REC_NEXT(7371, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7371(choice, ...)  METALANG99_PRIV_REC_NEXT(7372, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7372(choice, ...)  METALANG99_PRIV_REC_NEXT(7373, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7373(choice, ...)  METALANG99_PRIV_REC_NEXT(7374, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7374(choice, ...)  METALANG99_PRIV_REC_NEXT(7375, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7375(choice, ...)  METALANG99_PRIV_REC_NEXT(7376, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7376(choice, ...)  METALANG99_PRIV_REC_NEXT(7377, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7377(choice, ...)  METALANG99_PRIV_REC_NEXT(7378, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7378(choice, ...)  METALANG99_PRIV_REC_NEXT(7379, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7379(choice, ...)  METALANG99_PRIV_REC_NEXT(7380, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7380(choice, ...)  METALANG99_PRIV_REC_NEXT(7381, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7381(choice, ...)  METALANG99_PRIV_REC_NEXT(7382, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7382(choice, ...)  METALANG99_PRIV_REC_NEXT(7383, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7383(choice, ...)  METALANG99_PRIV_REC_NEXT(7384, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7384(choice, ...)  METALANG99_PRIV_REC_NEXT(7385, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7385(choice, ...)  METALANG99_PRIV_REC_NEXT(7386, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7386(choice, ...)  METALANG99_PRIV_REC_NEXT(7387, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7387(choice, ...)  METALANG99_PRIV_REC_NEXT(7388, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7388(choice, ...)  METALANG99_PRIV_REC_NEXT(7389, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7389(choice, ...)  METALANG99_PRIV_REC_NEXT(7390, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7390(choice, ...)  METALANG99_PRIV_REC_NEXT(7391, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7391(choice, ...)  METALANG99_PRIV_REC_NEXT(7392, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7392(choice, ...)  METALANG99_PRIV_REC_NEXT(7393, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7393(choice, ...)  METALANG99_PRIV_REC_NEXT(7394, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7394(choice, ...)  METALANG99_PRIV_REC_NEXT(7395, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7395(choice, ...)  METALANG99_PRIV_REC_NEXT(7396, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7396(choice, ...)  METALANG99_PRIV_REC_NEXT(7397, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7397(choice, ...)  METALANG99_PRIV_REC_NEXT(7398, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7398(choice, ...)  METALANG99_PRIV_REC_NEXT(7399, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7399(choice, ...)  METALANG99_PRIV_REC_NEXT(7400, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7400(choice, ...)  METALANG99_PRIV_REC_NEXT(7401, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7401(choice, ...)  METALANG99_PRIV_REC_NEXT(7402, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7402(choice, ...)  METALANG99_PRIV_REC_NEXT(7403, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7403(choice, ...)  METALANG99_PRIV_REC_NEXT(7404, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7404(choice, ...)  METALANG99_PRIV_REC_NEXT(7405, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7405(choice, ...)  METALANG99_PRIV_REC_NEXT(7406, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7406(choice, ...)  METALANG99_PRIV_REC_NEXT(7407, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7407(choice, ...)  METALANG99_PRIV_REC_NEXT(7408, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7408(choice, ...)  METALANG99_PRIV_REC_NEXT(7409, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7409(choice, ...)  METALANG99_PRIV_REC_NEXT(7410, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7410(choice, ...)  METALANG99_PRIV_REC_NEXT(7411, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7411(choice, ...)  METALANG99_PRIV_REC_NEXT(7412, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7412(choice, ...)  METALANG99_PRIV_REC_NEXT(7413, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7413(choice, ...)  METALANG99_PRIV_REC_NEXT(7414, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7414(choice, ...)  METALANG99_PRIV_REC_NEXT(7415, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7415(choice, ...)  METALANG99_PRIV_REC_NEXT(7416, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7416(choice, ...)  METALANG99_PRIV_REC_NEXT(7417, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7417(choice, ...)  METALANG99_PRIV_REC_NEXT(7418, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7418(choice, ...)  METALANG99_PRIV_REC_NEXT(7419, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7419(choice, ...)  METALANG99_PRIV_REC_NEXT(7420, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7420(choice, ...)  METALANG99_PRIV_REC_NEXT(7421, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7421(choice, ...)  METALANG99_PRIV_REC_NEXT(7422, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7422(choice, ...)  METALANG99_PRIV_REC_NEXT(7423, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7423(choice, ...)  METALANG99_PRIV_REC_NEXT(7424, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7424(choice, ...)  METALANG99_PRIV_REC_NEXT(7425, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7425(choice, ...)  METALANG99_PRIV_REC_NEXT(7426, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7426(choice, ...)  METALANG99_PRIV_REC_NEXT(7427, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7427(choice, ...)  METALANG99_PRIV_REC_NEXT(7428, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7428(choice, ...)  METALANG99_PRIV_REC_NEXT(7429, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7429(choice, ...)  METALANG99_PRIV_REC_NEXT(7430, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7430(choice, ...)  METALANG99_PRIV_REC_NEXT(7431, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7431(choice, ...)  METALANG99_PRIV_REC_NEXT(7432, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7432(choice, ...)  METALANG99_PRIV_REC_NEXT(7433, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7433(choice, ...)  METALANG99_PRIV_REC_NEXT(7434, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7434(choice, ...)  METALANG99_PRIV_REC_NEXT(7435, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7435(choice, ...)  METALANG99_PRIV_REC_NEXT(7436, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7436(choice, ...)  METALANG99_PRIV_REC_NEXT(7437, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7437(choice, ...)  METALANG99_PRIV_REC_NEXT(7438, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7438(choice, ...)  METALANG99_PRIV_REC_NEXT(7439, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7439(choice, ...)  METALANG99_PRIV_REC_NEXT(7440, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7440(choice, ...)  METALANG99_PRIV_REC_NEXT(7441, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7441(choice, ...)  METALANG99_PRIV_REC_NEXT(7442, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7442(choice, ...)  METALANG99_PRIV_REC_NEXT(7443, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7443(choice, ...)  METALANG99_PRIV_REC_NEXT(7444, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7444(choice, ...)  METALANG99_PRIV_REC_NEXT(7445, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7445(choice, ...)  METALANG99_PRIV_REC_NEXT(7446, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7446(choice, ...)  METALANG99_PRIV_REC_NEXT(7447, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7447(choice, ...)  METALANG99_PRIV_REC_NEXT(7448, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7448(choice, ...)  METALANG99_PRIV_REC_NEXT(7449, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7449(choice, ...)  METALANG99_PRIV_REC_NEXT(7450, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7450(choice, ...)  METALANG99_PRIV_REC_NEXT(7451, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7451(choice, ...)  METALANG99_PRIV_REC_NEXT(7452, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7452(choice, ...)  METALANG99_PRIV_REC_NEXT(7453, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7453(choice, ...)  METALANG99_PRIV_REC_NEXT(7454, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7454(choice, ...)  METALANG99_PRIV_REC_NEXT(7455, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7455(choice, ...)  METALANG99_PRIV_REC_NEXT(7456, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7456(choice, ...)  METALANG99_PRIV_REC_NEXT(7457, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7457(choice, ...)  METALANG99_PRIV_REC_NEXT(7458, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7458(choice, ...)  METALANG99_PRIV_REC_NEXT(7459, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7459(choice, ...)  METALANG99_PRIV_REC_NEXT(7460, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7460(choice, ...)  METALANG99_PRIV_REC_NEXT(7461, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7461(choice, ...)  METALANG99_PRIV_REC_NEXT(7462, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7462(choice, ...)  METALANG99_PRIV_REC_NEXT(7463, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7463(choice, ...)  METALANG99_PRIV_REC_NEXT(7464, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7464(choice, ...)  METALANG99_PRIV_REC_NEXT(7465, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7465(choice, ...)  METALANG99_PRIV_REC_NEXT(7466, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7466(choice, ...)  METALANG99_PRIV_REC_NEXT(7467, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7467(choice, ...)  METALANG99_PRIV_REC_NEXT(7468, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7468(choice, ...)  METALANG99_PRIV_REC_NEXT(7469, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7469(choice, ...)  METALANG99_PRIV_REC_NEXT(7470, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7470(choice, ...)  METALANG99_PRIV_REC_NEXT(7471, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7471(choice, ...)  METALANG99_PRIV_REC_NEXT(7472, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7472(choice, ...)  METALANG99_PRIV_REC_NEXT(7473, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7473(choice, ...)  METALANG99_PRIV_REC_NEXT(7474, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7474(choice, ...)  METALANG99_PRIV_REC_NEXT(7475, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7475(choice, ...)  METALANG99_PRIV_REC_NEXT(7476, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7476(choice, ...)  METALANG99_PRIV_REC_NEXT(7477, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7477(choice, ...)  METALANG99_PRIV_REC_NEXT(7478, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7478(choice, ...)  METALANG99_PRIV_REC_NEXT(7479, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7479(choice, ...)  METALANG99_PRIV_REC_NEXT(7480, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7480(choice, ...)  METALANG99_PRIV_REC_NEXT(7481, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7481(choice, ...)  METALANG99_PRIV_REC_NEXT(7482, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7482(choice, ...)  METALANG99_PRIV_REC_NEXT(7483, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7483(choice, ...)  METALANG99_PRIV_REC_NEXT(7484, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7484(choice, ...)  METALANG99_PRIV_REC_NEXT(7485, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7485(choice, ...)  METALANG99_PRIV_REC_NEXT(7486, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7486(choice, ...)  METALANG99_PRIV_REC_NEXT(7487, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7487(choice, ...)  METALANG99_PRIV_REC_NEXT(7488, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7488(choice, ...)  METALANG99_PRIV_REC_NEXT(7489, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7489(choice, ...)  METALANG99_PRIV_REC_NEXT(7490, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7490(choice, ...)  METALANG99_PRIV_REC_NEXT(7491, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7491(choice, ...)  METALANG99_PRIV_REC_NEXT(7492, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7492(choice, ...)  METALANG99_PRIV_REC_NEXT(7493, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7493(choice, ...)  METALANG99_PRIV_REC_NEXT(7494, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7494(choice, ...)  METALANG99_PRIV_REC_NEXT(7495, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7495(choice, ...)  METALANG99_PRIV_REC_NEXT(7496, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7496(choice, ...)  METALANG99_PRIV_REC_NEXT(7497, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7497(choice, ...)  METALANG99_PRIV_REC_NEXT(7498, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7498(choice, ...)  METALANG99_PRIV_REC_NEXT(7499, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7499(choice, ...)  METALANG99_PRIV_REC_NEXT(7500, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7500(choice, ...)  METALANG99_PRIV_REC_NEXT(7501, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7501(choice, ...)  METALANG99_PRIV_REC_NEXT(7502, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7502(choice, ...)  METALANG99_PRIV_REC_NEXT(7503, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7503(choice, ...)  METALANG99_PRIV_REC_NEXT(7504, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7504(choice, ...)  METALANG99_PRIV_REC_NEXT(7505, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7505(choice, ...)  METALANG99_PRIV_REC_NEXT(7506, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7506(choice, ...)  METALANG99_PRIV_REC_NEXT(7507, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7507(choice, ...)  METALANG99_PRIV_REC_NEXT(7508, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7508(choice, ...)  METALANG99_PRIV_REC_NEXT(7509, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7509(choice, ...)  METALANG99_PRIV_REC_NEXT(7510, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7510(choice, ...)  METALANG99_PRIV_REC_NEXT(7511, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7511(choice, ...)  METALANG99_PRIV_REC_NEXT(7512, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7512(choice, ...)  METALANG99_PRIV_REC_NEXT(7513, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7513(choice, ...)  METALANG99_PRIV_REC_NEXT(7514, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7514(choice, ...)  METALANG99_PRIV_REC_NEXT(7515, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7515(choice, ...)  METALANG99_PRIV_REC_NEXT(7516, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7516(choice, ...)  METALANG99_PRIV_REC_NEXT(7517, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7517(choice, ...)  METALANG99_PRIV_REC_NEXT(7518, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7518(choice, ...)  METALANG99_PRIV_REC_NEXT(7519, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7519(choice, ...)  METALANG99_PRIV_REC_NEXT(7520, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7520(choice, ...)  METALANG99_PRIV_REC_NEXT(7521, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7521(choice, ...)  METALANG99_PRIV_REC_NEXT(7522, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7522(choice, ...)  METALANG99_PRIV_REC_NEXT(7523, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7523(choice, ...)  METALANG99_PRIV_REC_NEXT(7524, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7524(choice, ...)  METALANG99_PRIV_REC_NEXT(7525, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7525(choice, ...)  METALANG99_PRIV_REC_NEXT(7526, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7526(choice, ...)  METALANG99_PRIV_REC_NEXT(7527, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7527(choice, ...)  METALANG99_PRIV_REC_NEXT(7528, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7528(choice, ...)  METALANG99_PRIV_REC_NEXT(7529, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7529(choice, ...)  METALANG99_PRIV_REC_NEXT(7530, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7530(choice, ...)  METALANG99_PRIV_REC_NEXT(7531, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7531(choice, ...)  METALANG99_PRIV_REC_NEXT(7532, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7532(choice, ...)  METALANG99_PRIV_REC_NEXT(7533, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7533(choice, ...)  METALANG99_PRIV_REC_NEXT(7534, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7534(choice, ...)  METALANG99_PRIV_REC_NEXT(7535, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7535(choice, ...)  METALANG99_PRIV_REC_NEXT(7536, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7536(choice, ...)  METALANG99_PRIV_REC_NEXT(7537, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7537(choice, ...)  METALANG99_PRIV_REC_NEXT(7538, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7538(choice, ...)  METALANG99_PRIV_REC_NEXT(7539, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7539(choice, ...)  METALANG99_PRIV_REC_NEXT(7540, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7540(choice, ...)  METALANG99_PRIV_REC_NEXT(7541, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7541(choice, ...)  METALANG99_PRIV_REC_NEXT(7542, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7542(choice, ...)  METALANG99_PRIV_REC_NEXT(7543, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7543(choice, ...)  METALANG99_PRIV_REC_NEXT(7544, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7544(choice, ...)  METALANG99_PRIV_REC_NEXT(7545, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7545(choice, ...)  METALANG99_PRIV_REC_NEXT(7546, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7546(choice, ...)  METALANG99_PRIV_REC_NEXT(7547, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7547(choice, ...)  METALANG99_PRIV_REC_NEXT(7548, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7548(choice, ...)  METALANG99_PRIV_REC_NEXT(7549, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7549(choice, ...)  METALANG99_PRIV_REC_NEXT(7550, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7550(choice, ...)  METALANG99_PRIV_REC_NEXT(7551, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7551(choice, ...)  METALANG99_PRIV_REC_NEXT(7552, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7552(choice, ...)  METALANG99_PRIV_REC_NEXT(7553, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7553(choice, ...)  METALANG99_PRIV_REC_NEXT(7554, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7554(choice, ...)  METALANG99_PRIV_REC_NEXT(7555, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7555(choice, ...)  METALANG99_PRIV_REC_NEXT(7556, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7556(choice, ...)  METALANG99_PRIV_REC_NEXT(7557, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7557(choice, ...)  METALANG99_PRIV_REC_NEXT(7558, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7558(choice, ...)  METALANG99_PRIV_REC_NEXT(7559, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7559(choice, ...)  METALANG99_PRIV_REC_NEXT(7560, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7560(choice, ...)  METALANG99_PRIV_REC_NEXT(7561, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7561(choice, ...)  METALANG99_PRIV_REC_NEXT(7562, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7562(choice, ...)  METALANG99_PRIV_REC_NEXT(7563, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7563(choice, ...)  METALANG99_PRIV_REC_NEXT(7564, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7564(choice, ...)  METALANG99_PRIV_REC_NEXT(7565, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7565(choice, ...)  METALANG99_PRIV_REC_NEXT(7566, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7566(choice, ...)  METALANG99_PRIV_REC_NEXT(7567, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7567(choice, ...)  METALANG99_PRIV_REC_NEXT(7568, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7568(choice, ...)  METALANG99_PRIV_REC_NEXT(7569, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7569(choice, ...)  METALANG99_PRIV_REC_NEXT(7570, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7570(choice, ...)  METALANG99_PRIV_REC_NEXT(7571, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7571(choice, ...)  METALANG99_PRIV_REC_NEXT(7572, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7572(choice, ...)  METALANG99_PRIV_REC_NEXT(7573, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7573(choice, ...)  METALANG99_PRIV_REC_NEXT(7574, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7574(choice, ...)  METALANG99_PRIV_REC_NEXT(7575, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7575(choice, ...)  METALANG99_PRIV_REC_NEXT(7576, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7576(choice, ...)  METALANG99_PRIV_REC_NEXT(7577, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7577(choice, ...)  METALANG99_PRIV_REC_NEXT(7578, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7578(choice, ...)  METALANG99_PRIV_REC_NEXT(7579, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7579(choice, ...)  METALANG99_PRIV_REC_NEXT(7580, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7580(choice, ...)  METALANG99_PRIV_REC_NEXT(7581, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7581(choice, ...)  METALANG99_PRIV_REC_NEXT(7582, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7582(choice, ...)  METALANG99_PRIV_REC_NEXT(7583, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7583(choice, ...)  METALANG99_PRIV_REC_NEXT(7584, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7584(choice, ...)  METALANG99_PRIV_REC_NEXT(7585, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7585(choice, ...)  METALANG99_PRIV_REC_NEXT(7586, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7586(choice, ...)  METALANG99_PRIV_REC_NEXT(7587, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7587(choice, ...)  METALANG99_PRIV_REC_NEXT(7588, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7588(choice, ...)  METALANG99_PRIV_REC_NEXT(7589, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7589(choice, ...)  METALANG99_PRIV_REC_NEXT(7590, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7590(choice, ...)  METALANG99_PRIV_REC_NEXT(7591, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7591(choice, ...)  METALANG99_PRIV_REC_NEXT(7592, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7592(choice, ...)  METALANG99_PRIV_REC_NEXT(7593, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7593(choice, ...)  METALANG99_PRIV_REC_NEXT(7594, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7594(choice, ...)  METALANG99_PRIV_REC_NEXT(7595, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7595(choice, ...)  METALANG99_PRIV_REC_NEXT(7596, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7596(choice, ...)  METALANG99_PRIV_REC_NEXT(7597, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7597(choice, ...)  METALANG99_PRIV_REC_NEXT(7598, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7598(choice, ...)  METALANG99_PRIV_REC_NEXT(7599, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7599(choice, ...)  METALANG99_PRIV_REC_NEXT(7600, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7600(choice, ...)  METALANG99_PRIV_REC_NEXT(7601, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7601(choice, ...)  METALANG99_PRIV_REC_NEXT(7602, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7602(choice, ...)  METALANG99_PRIV_REC_NEXT(7603, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7603(choice, ...)  METALANG99_PRIV_REC_NEXT(7604, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7604(choice, ...)  METALANG99_PRIV_REC_NEXT(7605, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7605(choice, ...)  METALANG99_PRIV_REC_NEXT(7606, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7606(choice, ...)  METALANG99_PRIV_REC_NEXT(7607, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7607(choice, ...)  METALANG99_PRIV_REC_NEXT(7608, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7608(choice, ...)  METALANG99_PRIV_REC_NEXT(7609, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7609(choice, ...)  METALANG99_PRIV_REC_NEXT(7610, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7610(choice, ...)  METALANG99_PRIV_REC_NEXT(7611, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7611(choice, ...)  METALANG99_PRIV_REC_NEXT(7612, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7612(choice, ...)  METALANG99_PRIV_REC_NEXT(7613, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7613(choice, ...)  METALANG99_PRIV_REC_NEXT(7614, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7614(choice, ...)  METALANG99_PRIV_REC_NEXT(7615, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7615(choice, ...)  METALANG99_PRIV_REC_NEXT(7616, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7616(choice, ...)  METALANG99_PRIV_REC_NEXT(7617, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7617(choice, ...)  METALANG99_PRIV_REC_NEXT(7618, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7618(choice, ...)  METALANG99_PRIV_REC_NEXT(7619, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7619(choice, ...)  METALANG99_PRIV_REC_NEXT(7620, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7620(choice, ...)  METALANG99_PRIV_REC_NEXT(7621, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7621(choice, ...)  METALANG99_PRIV_REC_NEXT(7622, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7622(choice, ...)  METALANG99_PRIV_REC_NEXT(7623, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7623(choice, ...)  METALANG99_PRIV_REC_NEXT(7624, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7624(choice, ...)  METALANG99_PRIV_REC_NEXT(7625, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7625(choice, ...)  METALANG99_PRIV_REC_NEXT(7626, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7626(choice, ...)  METALANG99_PRIV_REC_NEXT(7627, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7627(choice, ...)  METALANG99_PRIV_REC_NEXT(7628, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7628(choice, ...)  METALANG99_PRIV_REC_NEXT(7629, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7629(choice, ...)  METALANG99_PRIV_REC_NEXT(7630, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7630(choice, ...)  METALANG99_PRIV_REC_NEXT(7631, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7631(choice, ...)  METALANG99_PRIV_REC_NEXT(7632, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7632(choice, ...)  METALANG99_PRIV_REC_NEXT(7633, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7633(choice, ...)  METALANG99_PRIV_REC_NEXT(7634, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7634(choice, ...)  METALANG99_PRIV_REC_NEXT(7635, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7635(choice, ...)  METALANG99_PRIV_REC_NEXT(7636, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7636(choice, ...)  METALANG99_PRIV_REC_NEXT(7637, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7637(choice, ...)  METALANG99_PRIV_REC_NEXT(7638, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7638(choice, ...)  METALANG99_PRIV_REC_NEXT(7639, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7639(choice, ...)  METALANG99_PRIV_REC_NEXT(7640, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7640(choice, ...)  METALANG99_PRIV_REC_NEXT(7641, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7641(choice, ...)  METALANG99_PRIV_REC_NEXT(7642, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7642(choice, ...)  METALANG99_PRIV_REC_NEXT(7643, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7643(choice, ...)  METALANG99_PRIV_REC_NEXT(7644, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7644(choice, ...)  METALANG99_PRIV_REC_NEXT(7645, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7645(choice, ...)  METALANG99_PRIV_REC_NEXT(7646, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7646(choice, ...)  METALANG99_PRIV_REC_NEXT(7647, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7647(choice, ...)  METALANG99_PRIV_REC_NEXT(7648, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7648(choice, ...)  METALANG99_PRIV_REC_NEXT(7649, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7649(choice, ...)  METALANG99_PRIV_REC_NEXT(7650, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7650(choice, ...)  METALANG99_PRIV_REC_NEXT(7651, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7651(choice, ...)  METALANG99_PRIV_REC_NEXT(7652, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7652(choice, ...)  METALANG99_PRIV_REC_NEXT(7653, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7653(choice, ...)  METALANG99_PRIV_REC_NEXT(7654, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7654(choice, ...)  METALANG99_PRIV_REC_NEXT(7655, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7655(choice, ...)  METALANG99_PRIV_REC_NEXT(7656, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7656(choice, ...)  METALANG99_PRIV_REC_NEXT(7657, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7657(choice, ...)  METALANG99_PRIV_REC_NEXT(7658, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7658(choice, ...)  METALANG99_PRIV_REC_NEXT(7659, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7659(choice, ...)  METALANG99_PRIV_REC_NEXT(7660, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7660(choice, ...)  METALANG99_PRIV_REC_NEXT(7661, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7661(choice, ...)  METALANG99_PRIV_REC_NEXT(7662, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7662(choice, ...)  METALANG99_PRIV_REC_NEXT(7663, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7663(choice, ...)  METALANG99_PRIV_REC_NEXT(7664, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7664(choice, ...)  METALANG99_PRIV_REC_NEXT(7665, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7665(choice, ...)  METALANG99_PRIV_REC_NEXT(7666, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7666(choice, ...)  METALANG99_PRIV_REC_NEXT(7667, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7667(choice, ...)  METALANG99_PRIV_REC_NEXT(7668, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7668(choice, ...)  METALANG99_PRIV_REC_NEXT(7669, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7669(choice, ...)  METALANG99_PRIV_REC_NEXT(7670, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7670(choice, ...)  METALANG99_PRIV_REC_NEXT(7671, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7671(choice, ...)  METALANG99_PRIV_REC_NEXT(7672, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7672(choice, ...)  METALANG99_PRIV_REC_NEXT(7673, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7673(choice, ...)  METALANG99_PRIV_REC_NEXT(7674, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7674(choice, ...)  METALANG99_PRIV_REC_NEXT(7675, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7675(choice, ...)  METALANG99_PRIV_REC_NEXT(7676, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7676(choice, ...)  METALANG99_PRIV_REC_NEXT(7677, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7677(choice, ...)  METALANG99_PRIV_REC_NEXT(7678, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7678(choice, ...)  METALANG99_PRIV_REC_NEXT(7679, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7679(choice, ...)  METALANG99_PRIV_REC_NEXT(7680, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7680(choice, ...)  METALANG99_PRIV_REC_NEXT(7681, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7681(choice, ...)  METALANG99_PRIV_REC_NEXT(7682, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7682(choice, ...)  METALANG99_PRIV_REC_NEXT(7683, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7683(choice, ...)  METALANG99_PRIV_REC_NEXT(7684, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7684(choice, ...)  METALANG99_PRIV_REC_NEXT(7685, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7685(choice, ...)  METALANG99_PRIV_REC_NEXT(7686, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7686(choice, ...)  METALANG99_PRIV_REC_NEXT(7687, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7687(choice, ...)  METALANG99_PRIV_REC_NEXT(7688, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7688(choice, ...)  METALANG99_PRIV_REC_NEXT(7689, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7689(choice, ...)  METALANG99_PRIV_REC_NEXT(7690, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7690(choice, ...)  METALANG99_PRIV_REC_NEXT(7691, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7691(choice, ...)  METALANG99_PRIV_REC_NEXT(7692, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7692(choice, ...)  METALANG99_PRIV_REC_NEXT(7693, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7693(choice, ...)  METALANG99_PRIV_REC_NEXT(7694, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7694(choice, ...)  METALANG99_PRIV_REC_NEXT(7695, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7695(choice, ...)  METALANG99_PRIV_REC_NEXT(7696, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7696(choice, ...)  METALANG99_PRIV_REC_NEXT(7697, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7697(choice, ...)  METALANG99_PRIV_REC_NEXT(7698, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7698(choice, ...)  METALANG99_PRIV_REC_NEXT(7699, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7699(choice, ...)  METALANG99_PRIV_REC_NEXT(7700, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7700(choice, ...)  METALANG99_PRIV_REC_NEXT(7701, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7701(choice, ...)  METALANG99_PRIV_REC_NEXT(7702, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7702(choice, ...)  METALANG99_PRIV_REC_NEXT(7703, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7703(choice, ...)  METALANG99_PRIV_REC_NEXT(7704, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7704(choice, ...)  METALANG99_PRIV_REC_NEXT(7705, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7705(choice, ...)  METALANG99_PRIV_REC_NEXT(7706, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7706(choice, ...)  METALANG99_PRIV_REC_NEXT(7707, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7707(choice, ...)  METALANG99_PRIV_REC_NEXT(7708, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7708(choice, ...)  METALANG99_PRIV_REC_NEXT(7709, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7709(choice, ...)  METALANG99_PRIV_REC_NEXT(7710, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7710(choice, ...)  METALANG99_PRIV_REC_NEXT(7711, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7711(choice, ...)  METALANG99_PRIV_REC_NEXT(7712, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7712(choice, ...)  METALANG99_PRIV_REC_NEXT(7713, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7713(choice, ...)  METALANG99_PRIV_REC_NEXT(7714, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7714(choice, ...)  METALANG99_PRIV_REC_NEXT(7715, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7715(choice, ...)  METALANG99_PRIV_REC_NEXT(7716, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7716(choice, ...)  METALANG99_PRIV_REC_NEXT(7717, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7717(choice, ...)  METALANG99_PRIV_REC_NEXT(7718, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7718(choice, ...)  METALANG99_PRIV_REC_NEXT(7719, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7719(choice, ...)  METALANG99_PRIV_REC_NEXT(7720, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7720(choice, ...)  METALANG99_PRIV_REC_NEXT(7721, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7721(choice, ...)  METALANG99_PRIV_REC_NEXT(7722, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7722(choice, ...)  METALANG99_PRIV_REC_NEXT(7723, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7723(choice, ...)  METALANG99_PRIV_REC_NEXT(7724, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7724(choice, ...)  METALANG99_PRIV_REC_NEXT(7725, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7725(choice, ...)  METALANG99_PRIV_REC_NEXT(7726, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7726(choice, ...)  METALANG99_PRIV_REC_NEXT(7727, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7727(choice, ...)  METALANG99_PRIV_REC_NEXT(7728, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7728(choice, ...)  METALANG99_PRIV_REC_NEXT(7729, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7729(choice, ...)  METALANG99_PRIV_REC_NEXT(7730, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7730(choice, ...)  METALANG99_PRIV_REC_NEXT(7731, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7731(choice, ...)  METALANG99_PRIV_REC_NEXT(7732, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7732(choice, ...)  METALANG99_PRIV_REC_NEXT(7733, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7733(choice, ...)  METALANG99_PRIV_REC_NEXT(7734, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7734(choice, ...)  METALANG99_PRIV_REC_NEXT(7735, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7735(choice, ...)  METALANG99_PRIV_REC_NEXT(7736, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7736(choice, ...)  METALANG99_PRIV_REC_NEXT(7737, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7737(choice, ...)  METALANG99_PRIV_REC_NEXT(7738, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7738(choice, ...)  METALANG99_PRIV_REC_NEXT(7739, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7739(choice, ...)  METALANG99_PRIV_REC_NEXT(7740, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7740(choice, ...)  METALANG99_PRIV_REC_NEXT(7741, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7741(choice, ...)  METALANG99_PRIV_REC_NEXT(7742, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7742(choice, ...)  METALANG99_PRIV_REC_NEXT(7743, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7743(choice, ...)  METALANG99_PRIV_REC_NEXT(7744, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7744(choice, ...)  METALANG99_PRIV_REC_NEXT(7745, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7745(choice, ...)  METALANG99_PRIV_REC_NEXT(7746, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7746(choice, ...)  METALANG99_PRIV_REC_NEXT(7747, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7747(choice, ...)  METALANG99_PRIV_REC_NEXT(7748, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7748(choice, ...)  METALANG99_PRIV_REC_NEXT(7749, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7749(choice, ...)  METALANG99_PRIV_REC_NEXT(7750, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7750(choice, ...)  METALANG99_PRIV_REC_NEXT(7751, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7751(choice, ...)  METALANG99_PRIV_REC_NEXT(7752, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7752(choice, ...)  METALANG99_PRIV_REC_NEXT(7753, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7753(choice, ...)  METALANG99_PRIV_REC_NEXT(7754, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7754(choice, ...)  METALANG99_PRIV_REC_NEXT(7755, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7755(choice, ...)  METALANG99_PRIV_REC_NEXT(7756, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7756(choice, ...)  METALANG99_PRIV_REC_NEXT(7757, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7757(choice, ...)  METALANG99_PRIV_REC_NEXT(7758, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7758(choice, ...)  METALANG99_PRIV_REC_NEXT(7759, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7759(choice, ...)  METALANG99_PRIV_REC_NEXT(7760, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7760(choice, ...)  METALANG99_PRIV_REC_NEXT(7761, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7761(choice, ...)  METALANG99_PRIV_REC_NEXT(7762, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7762(choice, ...)  METALANG99_PRIV_REC_NEXT(7763, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7763(choice, ...)  METALANG99_PRIV_REC_NEXT(7764, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7764(choice, ...)  METALANG99_PRIV_REC_NEXT(7765, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7765(choice, ...)  METALANG99_PRIV_REC_NEXT(7766, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7766(choice, ...)  METALANG99_PRIV_REC_NEXT(7767, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7767(choice, ...)  METALANG99_PRIV_REC_NEXT(7768, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7768(choice, ...)  METALANG99_PRIV_REC_NEXT(7769, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7769(choice, ...)  METALANG99_PRIV_REC_NEXT(7770, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7770(choice, ...)  METALANG99_PRIV_REC_NEXT(7771, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7771(choice, ...)  METALANG99_PRIV_REC_NEXT(7772, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7772(choice, ...)  METALANG99_PRIV_REC_NEXT(7773, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7773(choice, ...)  METALANG99_PRIV_REC_NEXT(7774, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7774(choice, ...)  METALANG99_PRIV_REC_NEXT(7775, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7775(choice, ...)  METALANG99_PRIV_REC_NEXT(7776, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7776(choice, ...)  METALANG99_PRIV_REC_NEXT(7777, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7777(choice, ...)  METALANG99_PRIV_REC_NEXT(7778, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7778(choice, ...)  METALANG99_PRIV_REC_NEXT(7779, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7779(choice, ...)  METALANG99_PRIV_REC_NEXT(7780, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7780(choice, ...)  METALANG99_PRIV_REC_NEXT(7781, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7781(choice, ...)  METALANG99_PRIV_REC_NEXT(7782, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7782(choice, ...)  METALANG99_PRIV_REC_NEXT(7783, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7783(choice, ...)  METALANG99_PRIV_REC_NEXT(7784, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7784(choice, ...)  METALANG99_PRIV_REC_NEXT(7785, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7785(choice, ...)  METALANG99_PRIV_REC_NEXT(7786, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7786(choice, ...)  METALANG99_PRIV_REC_NEXT(7787, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7787(choice, ...)  METALANG99_PRIV_REC_NEXT(7788, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7788(choice, ...)  METALANG99_PRIV_REC_NEXT(7789, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7789(choice, ...)  METALANG99_PRIV_REC_NEXT(7790, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7790(choice, ...)  METALANG99_PRIV_REC_NEXT(7791, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7791(choice, ...)  METALANG99_PRIV_REC_NEXT(7792, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7792(choice, ...)  METALANG99_PRIV_REC_NEXT(7793, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7793(choice, ...)  METALANG99_PRIV_REC_NEXT(7794, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7794(choice, ...)  METALANG99_PRIV_REC_NEXT(7795, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7795(choice, ...)  METALANG99_PRIV_REC_NEXT(7796, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7796(choice, ...)  METALANG99_PRIV_REC_NEXT(7797, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7797(choice, ...)  METALANG99_PRIV_REC_NEXT(7798, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7798(choice, ...)  METALANG99_PRIV_REC_NEXT(7799, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7799(choice, ...)  METALANG99_PRIV_REC_NEXT(7800, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7800(choice, ...)  METALANG99_PRIV_REC_NEXT(7801, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7801(choice, ...)  METALANG99_PRIV_REC_NEXT(7802, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7802(choice, ...)  METALANG99_PRIV_REC_NEXT(7803, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7803(choice, ...)  METALANG99_PRIV_REC_NEXT(7804, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7804(choice, ...)  METALANG99_PRIV_REC_NEXT(7805, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7805(choice, ...)  METALANG99_PRIV_REC_NEXT(7806, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7806(choice, ...)  METALANG99_PRIV_REC_NEXT(7807, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7807(choice, ...)  METALANG99_PRIV_REC_NEXT(7808, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7808(choice, ...)  METALANG99_PRIV_REC_NEXT(7809, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7809(choice, ...)  METALANG99_PRIV_REC_NEXT(7810, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7810(choice, ...)  METALANG99_PRIV_REC_NEXT(7811, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7811(choice, ...)  METALANG99_PRIV_REC_NEXT(7812, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7812(choice, ...)  METALANG99_PRIV_REC_NEXT(7813, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7813(choice, ...)  METALANG99_PRIV_REC_NEXT(7814, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7814(choice, ...)  METALANG99_PRIV_REC_NEXT(7815, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7815(choice, ...)  METALANG99_PRIV_REC_NEXT(7816, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7816(choice, ...)  METALANG99_PRIV_REC_NEXT(7817, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7817(choice, ...)  METALANG99_PRIV_REC_NEXT(7818, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7818(choice, ...)  METALANG99_PRIV_REC_NEXT(7819, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7819(choice, ...)  METALANG99_PRIV_REC_NEXT(7820, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7820(choice, ...)  METALANG99_PRIV_REC_NEXT(7821, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7821(choice, ...)  METALANG99_PRIV_REC_NEXT(7822, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7822(choice, ...)  METALANG99_PRIV_REC_NEXT(7823, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7823(choice, ...)  METALANG99_PRIV_REC_NEXT(7824, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7824(choice, ...)  METALANG99_PRIV_REC_NEXT(7825, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7825(choice, ...)  METALANG99_PRIV_REC_NEXT(7826, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7826(choice, ...)  METALANG99_PRIV_REC_NEXT(7827, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7827(choice, ...)  METALANG99_PRIV_REC_NEXT(7828, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7828(choice, ...)  METALANG99_PRIV_REC_NEXT(7829, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7829(choice, ...)  METALANG99_PRIV_REC_NEXT(7830, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7830(choice, ...)  METALANG99_PRIV_REC_NEXT(7831, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7831(choice, ...)  METALANG99_PRIV_REC_NEXT(7832, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7832(choice, ...)  METALANG99_PRIV_REC_NEXT(7833, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7833(choice, ...)  METALANG99_PRIV_REC_NEXT(7834, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7834(choice, ...)  METALANG99_PRIV_REC_NEXT(7835, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7835(choice, ...)  METALANG99_PRIV_REC_NEXT(7836, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7836(choice, ...)  METALANG99_PRIV_REC_NEXT(7837, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7837(choice, ...)  METALANG99_PRIV_REC_NEXT(7838, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7838(choice, ...)  METALANG99_PRIV_REC_NEXT(7839, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7839(choice, ...)  METALANG99_PRIV_REC_NEXT(7840, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7840(choice, ...)  METALANG99_PRIV_REC_NEXT(7841, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7841(choice, ...)  METALANG99_PRIV_REC_NEXT(7842, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7842(choice, ...)  METALANG99_PRIV_REC_NEXT(7843, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7843(choice, ...)  METALANG99_PRIV_REC_NEXT(7844, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7844(choice, ...)  METALANG99_PRIV_REC_NEXT(7845, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7845(choice, ...)  METALANG99_PRIV_REC_NEXT(7846, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7846(choice, ...)  METALANG99_PRIV_REC_NEXT(7847, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7847(choice, ...)  METALANG99_PRIV_REC_NEXT(7848, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7848(choice, ...)  METALANG99_PRIV_REC_NEXT(7849, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7849(choice, ...)  METALANG99_PRIV_REC_NEXT(7850, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7850(choice, ...)  METALANG99_PRIV_REC_NEXT(7851, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7851(choice, ...)  METALANG99_PRIV_REC_NEXT(7852, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7852(choice, ...)  METALANG99_PRIV_REC_NEXT(7853, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7853(choice, ...)  METALANG99_PRIV_REC_NEXT(7854, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7854(choice, ...)  METALANG99_PRIV_REC_NEXT(7855, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7855(choice, ...)  METALANG99_PRIV_REC_NEXT(7856, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7856(choice, ...)  METALANG99_PRIV_REC_NEXT(7857, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7857(choice, ...)  METALANG99_PRIV_REC_NEXT(7858, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7858(choice, ...)  METALANG99_PRIV_REC_NEXT(7859, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7859(choice, ...)  METALANG99_PRIV_REC_NEXT(7860, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7860(choice, ...)  METALANG99_PRIV_REC_NEXT(7861, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7861(choice, ...)  METALANG99_PRIV_REC_NEXT(7862, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7862(choice, ...)  METALANG99_PRIV_REC_NEXT(7863, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7863(choice, ...)  METALANG99_PRIV_REC_NEXT(7864, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7864(choice, ...)  METALANG99_PRIV_REC_NEXT(7865, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7865(choice, ...)  METALANG99_PRIV_REC_NEXT(7866, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7866(choice, ...)  METALANG99_PRIV_REC_NEXT(7867, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7867(choice, ...)  METALANG99_PRIV_REC_NEXT(7868, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7868(choice, ...)  METALANG99_PRIV_REC_NEXT(7869, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7869(choice, ...)  METALANG99_PRIV_REC_NEXT(7870, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7870(choice, ...)  METALANG99_PRIV_REC_NEXT(7871, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7871(choice, ...)  METALANG99_PRIV_REC_NEXT(7872, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7872(choice, ...)  METALANG99_PRIV_REC_NEXT(7873, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7873(choice, ...)  METALANG99_PRIV_REC_NEXT(7874, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7874(choice, ...)  METALANG99_PRIV_REC_NEXT(7875, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7875(choice, ...)  METALANG99_PRIV_REC_NEXT(7876, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7876(choice, ...)  METALANG99_PRIV_REC_NEXT(7877, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7877(choice, ...)  METALANG99_PRIV_REC_NEXT(7878, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7878(choice, ...)  METALANG99_PRIV_REC_NEXT(7879, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7879(choice, ...)  METALANG99_PRIV_REC_NEXT(7880, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7880(choice, ...)  METALANG99_PRIV_REC_NEXT(7881, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7881(choice, ...)  METALANG99_PRIV_REC_NEXT(7882, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7882(choice, ...)  METALANG99_PRIV_REC_NEXT(7883, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7883(choice, ...)  METALANG99_PRIV_REC_NEXT(7884, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7884(choice, ...)  METALANG99_PRIV_REC_NEXT(7885, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7885(choice, ...)  METALANG99_PRIV_REC_NEXT(7886, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7886(choice, ...)  METALANG99_PRIV_REC_NEXT(7887, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7887(choice, ...)  METALANG99_PRIV_REC_NEXT(7888, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7888(choice, ...)  METALANG99_PRIV_REC_NEXT(7889, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7889(choice, ...)  METALANG99_PRIV_REC_NEXT(7890, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7890(choice, ...)  METALANG99_PRIV_REC_NEXT(7891, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7891(choice, ...)  METALANG99_PRIV_REC_NEXT(7892, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7892(choice, ...)  METALANG99_PRIV_REC_NEXT(7893, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7893(choice, ...)  METALANG99_PRIV_REC_NEXT(7894, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7894(choice, ...)  METALANG99_PRIV_REC_NEXT(7895, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7895(choice, ...)  METALANG99_PRIV_REC_NEXT(7896, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7896(choice, ...)  METALANG99_PRIV_REC_NEXT(7897, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7897(choice, ...)  METALANG99_PRIV_REC_NEXT(7898, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7898(choice, ...)  METALANG99_PRIV_REC_NEXT(7899, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7899(choice, ...)  METALANG99_PRIV_REC_NEXT(7900, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7900(choice, ...)  METALANG99_PRIV_REC_NEXT(7901, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7901(choice, ...)  METALANG99_PRIV_REC_NEXT(7902, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7902(choice, ...)  METALANG99_PRIV_REC_NEXT(7903, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7903(choice, ...)  METALANG99_PRIV_REC_NEXT(7904, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7904(choice, ...)  METALANG99_PRIV_REC_NEXT(7905, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7905(choice, ...)  METALANG99_PRIV_REC_NEXT(7906, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7906(choice, ...)  METALANG99_PRIV_REC_NEXT(7907, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7907(choice, ...)  METALANG99_PRIV_REC_NEXT(7908, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7908(choice, ...)  METALANG99_PRIV_REC_NEXT(7909, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7909(choice, ...)  METALANG99_PRIV_REC_NEXT(7910, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7910(choice, ...)  METALANG99_PRIV_REC_NEXT(7911, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7911(choice, ...)  METALANG99_PRIV_REC_NEXT(7912, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7912(choice, ...)  METALANG99_PRIV_REC_NEXT(7913, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7913(choice, ...)  METALANG99_PRIV_REC_NEXT(7914, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7914(choice, ...)  METALANG99_PRIV_REC_NEXT(7915, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7915(choice, ...)  METALANG99_PRIV_REC_NEXT(7916, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7916(choice, ...)  METALANG99_PRIV_REC_NEXT(7917, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7917(choice, ...)  METALANG99_PRIV_REC_NEXT(7918, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7918(choice, ...)  METALANG99_PRIV_REC_NEXT(7919, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7919(choice, ...)  METALANG99_PRIV_REC_NEXT(7920, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7920(choice, ...)  METALANG99_PRIV_REC_NEXT(7921, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7921(choice, ...)  METALANG99_PRIV_REC_NEXT(7922, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7922(choice, ...)  METALANG99_PRIV_REC_NEXT(7923, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7923(choice, ...)  METALANG99_PRIV_REC_NEXT(7924, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7924(choice, ...)  METALANG99_PRIV_REC_NEXT(7925, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7925(choice, ...)  METALANG99_PRIV_REC_NEXT(7926, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7926(choice, ...)  METALANG99_PRIV_REC_NEXT(7927, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7927(choice, ...)  METALANG99_PRIV_REC_NEXT(7928, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7928(choice, ...)  METALANG99_PRIV_REC_NEXT(7929, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7929(choice, ...)  METALANG99_PRIV_REC_NEXT(7930, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7930(choice, ...)  METALANG99_PRIV_REC_NEXT(7931, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7931(choice, ...)  METALANG99_PRIV_REC_NEXT(7932, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7932(choice, ...)  METALANG99_PRIV_REC_NEXT(7933, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7933(choice, ...)  METALANG99_PRIV_REC_NEXT(7934, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7934(choice, ...)  METALANG99_PRIV_REC_NEXT(7935, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7935(choice, ...)  METALANG99_PRIV_REC_NEXT(7936, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7936(choice, ...)  METALANG99_PRIV_REC_NEXT(7937, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7937(choice, ...)  METALANG99_PRIV_REC_NEXT(7938, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7938(choice, ...)  METALANG99_PRIV_REC_NEXT(7939, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7939(choice, ...)  METALANG99_PRIV_REC_NEXT(7940, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7940(choice, ...)  METALANG99_PRIV_REC_NEXT(7941, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7941(choice, ...)  METALANG99_PRIV_REC_NEXT(7942, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7942(choice, ...)  METALANG99_PRIV_REC_NEXT(7943, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7943(choice, ...)  METALANG99_PRIV_REC_NEXT(7944, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7944(choice, ...)  METALANG99_PRIV_REC_NEXT(7945, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7945(choice, ...)  METALANG99_PRIV_REC_NEXT(7946, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7946(choice, ...)  METALANG99_PRIV_REC_NEXT(7947, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7947(choice, ...)  METALANG99_PRIV_REC_NEXT(7948, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7948(choice, ...)  METALANG99_PRIV_REC_NEXT(7949, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7949(choice, ...)  METALANG99_PRIV_REC_NEXT(7950, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7950(choice, ...)  METALANG99_PRIV_REC_NEXT(7951, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7951(choice, ...)  METALANG99_PRIV_REC_NEXT(7952, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7952(choice, ...)  METALANG99_PRIV_REC_NEXT(7953, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7953(choice, ...)  METALANG99_PRIV_REC_NEXT(7954, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7954(choice, ...)  METALANG99_PRIV_REC_NEXT(7955, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7955(choice, ...)  METALANG99_PRIV_REC_NEXT(7956, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7956(choice, ...)  METALANG99_PRIV_REC_NEXT(7957, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7957(choice, ...)  METALANG99_PRIV_REC_NEXT(7958, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7958(choice, ...)  METALANG99_PRIV_REC_NEXT(7959, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7959(choice, ...)  METALANG99_PRIV_REC_NEXT(7960, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7960(choice, ...)  METALANG99_PRIV_REC_NEXT(7961, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7961(choice, ...)  METALANG99_PRIV_REC_NEXT(7962, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7962(choice, ...)  METALANG99_PRIV_REC_NEXT(7963, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7963(choice, ...)  METALANG99_PRIV_REC_NEXT(7964, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7964(choice, ...)  METALANG99_PRIV_REC_NEXT(7965, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7965(choice, ...)  METALANG99_PRIV_REC_NEXT(7966, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7966(choice, ...)  METALANG99_PRIV_REC_NEXT(7967, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7967(choice, ...)  METALANG99_PRIV_REC_NEXT(7968, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7968(choice, ...)  METALANG99_PRIV_REC_NEXT(7969, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7969(choice, ...)  METALANG99_PRIV_REC_NEXT(7970, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7970(choice, ...)  METALANG99_PRIV_REC_NEXT(7971, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7971(choice, ...)  METALANG99_PRIV_REC_NEXT(7972, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7972(choice, ...)  METALANG99_PRIV_REC_NEXT(7973, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7973(choice, ...)  METALANG99_PRIV_REC_NEXT(7974, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7974(choice, ...)  METALANG99_PRIV_REC_NEXT(7975, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7975(choice, ...)  METALANG99_PRIV_REC_NEXT(7976, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7976(choice, ...)  METALANG99_PRIV_REC_NEXT(7977, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7977(choice, ...)  METALANG99_PRIV_REC_NEXT(7978, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7978(choice, ...)  METALANG99_PRIV_REC_NEXT(7979, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7979(choice, ...)  METALANG99_PRIV_REC_NEXT(7980, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7980(choice, ...)  METALANG99_PRIV_REC_NEXT(7981, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7981(choice, ...)  METALANG99_PRIV_REC_NEXT(7982, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7982(choice, ...)  METALANG99_PRIV_REC_NEXT(7983, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7983(choice, ...)  METALANG99_PRIV_REC_NEXT(7984, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7984(choice, ...)  METALANG99_PRIV_REC_NEXT(7985, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7985(choice, ...)  METALANG99_PRIV_REC_NEXT(7986, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7986(choice, ...)  METALANG99_PRIV_REC_NEXT(7987, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7987(choice, ...)  METALANG99_PRIV_REC_NEXT(7988, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7988(choice, ...)  METALANG99_PRIV_REC_NEXT(7989, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7989(choice, ...)  METALANG99_PRIV_REC_NEXT(7990, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7990(choice, ...)  METALANG99_PRIV_REC_NEXT(7991, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7991(choice, ...)  METALANG99_PRIV_REC_NEXT(7992, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7992(choice, ...)  METALANG99_PRIV_REC_NEXT(7993, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7993(choice, ...)  METALANG99_PRIV_REC_NEXT(7994, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7994(choice, ...)  METALANG99_PRIV_REC_NEXT(7995, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7995(choice, ...)  METALANG99_PRIV_REC_NEXT(7996, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7996(choice, ...)  METALANG99_PRIV_REC_NEXT(7997, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7997(choice, ...)  METALANG99_PRIV_REC_NEXT(7998, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7998(choice, ...)  METALANG99_PRIV_REC_NEXT(7999, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_7999(choice, ...)  METALANG99_PRIV_REC_NEXT(8000, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8000(choice, ...)  METALANG99_PRIV_REC_NEXT(8001, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8001(choice, ...)  METALANG99_PRIV_REC_NEXT(8002, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8002(choice, ...)  METALANG99_PRIV_REC_NEXT(8003, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8003(choice, ...)  METALANG99_PRIV_REC_NEXT(8004, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8004(choice, ...)  METALANG99_PRIV_REC_NEXT(8005, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8005(choice, ...)  METALANG99_PRIV_REC_NEXT(8006, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8006(choice, ...)  METALANG99_PRIV_REC_NEXT(8007, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8007(choice, ...)  METALANG99_PRIV_REC_NEXT(8008, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8008(choice, ...)  METALANG99_PRIV_REC_NEXT(8009, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8009(choice, ...)  METALANG99_PRIV_REC_NEXT(8010, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8010(choice, ...)  METALANG99_PRIV_REC_NEXT(8011, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8011(choice, ...)  METALANG99_PRIV_REC_NEXT(8012, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8012(choice, ...)  METALANG99_PRIV_REC_NEXT(8013, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8013(choice, ...)  METALANG99_PRIV_REC_NEXT(8014, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8014(choice, ...)  METALANG99_PRIV_REC_NEXT(8015, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8015(choice, ...)  METALANG99_PRIV_REC_NEXT(8016, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8016(choice, ...)  METALANG99_PRIV_REC_NEXT(8017, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8017(choice, ...)  METALANG99_PRIV_REC_NEXT(8018, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8018(choice, ...)  METALANG99_PRIV_REC_NEXT(8019, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8019(choice, ...)  METALANG99_PRIV_REC_NEXT(8020, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8020(choice, ...)  METALANG99_PRIV_REC_NEXT(8021, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8021(choice, ...)  METALANG99_PRIV_REC_NEXT(8022, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8022(choice, ...)  METALANG99_PRIV_REC_NEXT(8023, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8023(choice, ...)  METALANG99_PRIV_REC_NEXT(8024, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8024(choice, ...)  METALANG99_PRIV_REC_NEXT(8025, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8025(choice, ...)  METALANG99_PRIV_REC_NEXT(8026, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8026(choice, ...)  METALANG99_PRIV_REC_NEXT(8027, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8027(choice, ...)  METALANG99_PRIV_REC_NEXT(8028, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8028(choice, ...)  METALANG99_PRIV_REC_NEXT(8029, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8029(choice, ...)  METALANG99_PRIV_REC_NEXT(8030, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8030(choice, ...)  METALANG99_PRIV_REC_NEXT(8031, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8031(choice, ...)  METALANG99_PRIV_REC_NEXT(8032, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8032(choice, ...)  METALANG99_PRIV_REC_NEXT(8033, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8033(choice, ...)  METALANG99_PRIV_REC_NEXT(8034, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8034(choice, ...)  METALANG99_PRIV_REC_NEXT(8035, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8035(choice, ...)  METALANG99_PRIV_REC_NEXT(8036, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8036(choice, ...)  METALANG99_PRIV_REC_NEXT(8037, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8037(choice, ...)  METALANG99_PRIV_REC_NEXT(8038, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8038(choice, ...)  METALANG99_PRIV_REC_NEXT(8039, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8039(choice, ...)  METALANG99_PRIV_REC_NEXT(8040, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8040(choice, ...)  METALANG99_PRIV_REC_NEXT(8041, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8041(choice, ...)  METALANG99_PRIV_REC_NEXT(8042, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8042(choice, ...)  METALANG99_PRIV_REC_NEXT(8043, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8043(choice, ...)  METALANG99_PRIV_REC_NEXT(8044, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8044(choice, ...)  METALANG99_PRIV_REC_NEXT(8045, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8045(choice, ...)  METALANG99_PRIV_REC_NEXT(8046, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8046(choice, ...)  METALANG99_PRIV_REC_NEXT(8047, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8047(choice, ...)  METALANG99_PRIV_REC_NEXT(8048, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8048(choice, ...)  METALANG99_PRIV_REC_NEXT(8049, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8049(choice, ...)  METALANG99_PRIV_REC_NEXT(8050, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8050(choice, ...)  METALANG99_PRIV_REC_NEXT(8051, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8051(choice, ...)  METALANG99_PRIV_REC_NEXT(8052, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8052(choice, ...)  METALANG99_PRIV_REC_NEXT(8053, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8053(choice, ...)  METALANG99_PRIV_REC_NEXT(8054, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8054(choice, ...)  METALANG99_PRIV_REC_NEXT(8055, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8055(choice, ...)  METALANG99_PRIV_REC_NEXT(8056, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8056(choice, ...)  METALANG99_PRIV_REC_NEXT(8057, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8057(choice, ...)  METALANG99_PRIV_REC_NEXT(8058, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8058(choice, ...)  METALANG99_PRIV_REC_NEXT(8059, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8059(choice, ...)  METALANG99_PRIV_REC_NEXT(8060, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8060(choice, ...)  METALANG99_PRIV_REC_NEXT(8061, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8061(choice, ...)  METALANG99_PRIV_REC_NEXT(8062, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8062(choice, ...)  METALANG99_PRIV_REC_NEXT(8063, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8063(choice, ...)  METALANG99_PRIV_REC_NEXT(8064, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8064(choice, ...)  METALANG99_PRIV_REC_NEXT(8065, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8065(choice, ...)  METALANG99_PRIV_REC_NEXT(8066, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8066(choice, ...)  METALANG99_PRIV_REC_NEXT(8067, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8067(choice, ...)  METALANG99_PRIV_REC_NEXT(8068, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8068(choice, ...)  METALANG99_PRIV_REC_NEXT(8069, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8069(choice, ...)  METALANG99_PRIV_REC_NEXT(8070, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8070(choice, ...)  METALANG99_PRIV_REC_NEXT(8071, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8071(choice, ...)  METALANG99_PRIV_REC_NEXT(8072, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8072(choice, ...)  METALANG99_PRIV_REC_NEXT(8073, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8073(choice, ...)  METALANG99_PRIV_REC_NEXT(8074, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8074(choice, ...)  METALANG99_PRIV_REC_NEXT(8075, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8075(choice, ...)  METALANG99_PRIV_REC_NEXT(8076, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8076(choice, ...)  METALANG99_PRIV_REC_NEXT(8077, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8077(choice, ...)  METALANG99_PRIV_REC_NEXT(8078, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8078(choice, ...)  METALANG99_PRIV_REC_NEXT(8079, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8079(choice, ...)  METALANG99_PRIV_REC_NEXT(8080, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8080(choice, ...)  METALANG99_PRIV_REC_NEXT(8081, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8081(choice, ...)  METALANG99_PRIV_REC_NEXT(8082, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8082(choice, ...)  METALANG99_PRIV_REC_NEXT(8083, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8083(choice, ...)  METALANG99_PRIV_REC_NEXT(8084, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8084(choice, ...)  METALANG99_PRIV_REC_NEXT(8085, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8085(choice, ...)  METALANG99_PRIV_REC_NEXT(8086, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8086(choice, ...)  METALANG99_PRIV_REC_NEXT(8087, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8087(choice, ...)  METALANG99_PRIV_REC_NEXT(8088, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8088(choice, ...)  METALANG99_PRIV_REC_NEXT(8089, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8089(choice, ...)  METALANG99_PRIV_REC_NEXT(8090, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8090(choice, ...)  METALANG99_PRIV_REC_NEXT(8091, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8091(choice, ...)  METALANG99_PRIV_REC_NEXT(8092, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8092(choice, ...)  METALANG99_PRIV_REC_NEXT(8093, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8093(choice, ...)  METALANG99_PRIV_REC_NEXT(8094, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8094(choice, ...)  METALANG99_PRIV_REC_NEXT(8095, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8095(choice, ...)  METALANG99_PRIV_REC_NEXT(8096, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8096(choice, ...)  METALANG99_PRIV_REC_NEXT(8097, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8097(choice, ...)  METALANG99_PRIV_REC_NEXT(8098, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8098(choice, ...)  METALANG99_PRIV_REC_NEXT(8099, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8099(choice, ...)  METALANG99_PRIV_REC_NEXT(8100, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8100(choice, ...)  METALANG99_PRIV_REC_NEXT(8101, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8101(choice, ...)  METALANG99_PRIV_REC_NEXT(8102, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8102(choice, ...)  METALANG99_PRIV_REC_NEXT(8103, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8103(choice, ...)  METALANG99_PRIV_REC_NEXT(8104, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8104(choice, ...)  METALANG99_PRIV_REC_NEXT(8105, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8105(choice, ...)  METALANG99_PRIV_REC_NEXT(8106, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8106(choice, ...)  METALANG99_PRIV_REC_NEXT(8107, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8107(choice, ...)  METALANG99_PRIV_REC_NEXT(8108, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8108(choice, ...)  METALANG99_PRIV_REC_NEXT(8109, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8109(choice, ...)  METALANG99_PRIV_REC_NEXT(8110, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8110(choice, ...)  METALANG99_PRIV_REC_NEXT(8111, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8111(choice, ...)  METALANG99_PRIV_REC_NEXT(8112, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8112(choice, ...)  METALANG99_PRIV_REC_NEXT(8113, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8113(choice, ...)  METALANG99_PRIV_REC_NEXT(8114, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8114(choice, ...)  METALANG99_PRIV_REC_NEXT(8115, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8115(choice, ...)  METALANG99_PRIV_REC_NEXT(8116, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8116(choice, ...)  METALANG99_PRIV_REC_NEXT(8117, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8117(choice, ...)  METALANG99_PRIV_REC_NEXT(8118, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8118(choice, ...)  METALANG99_PRIV_REC_NEXT(8119, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8119(choice, ...)  METALANG99_PRIV_REC_NEXT(8120, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8120(choice, ...)  METALANG99_PRIV_REC_NEXT(8121, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8121(choice, ...)  METALANG99_PRIV_REC_NEXT(8122, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8122(choice, ...)  METALANG99_PRIV_REC_NEXT(8123, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8123(choice, ...)  METALANG99_PRIV_REC_NEXT(8124, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8124(choice, ...)  METALANG99_PRIV_REC_NEXT(8125, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8125(choice, ...)  METALANG99_PRIV_REC_NEXT(8126, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8126(choice, ...)  METALANG99_PRIV_REC_NEXT(8127, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8127(choice, ...)  METALANG99_PRIV_REC_NEXT(8128, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8128(choice, ...)  METALANG99_PRIV_REC_NEXT(8129, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8129(choice, ...)  METALANG99_PRIV_REC_NEXT(8130, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8130(choice, ...)  METALANG99_PRIV_REC_NEXT(8131, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8131(choice, ...)  METALANG99_PRIV_REC_NEXT(8132, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8132(choice, ...)  METALANG99_PRIV_REC_NEXT(8133, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8133(choice, ...)  METALANG99_PRIV_REC_NEXT(8134, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8134(choice, ...)  METALANG99_PRIV_REC_NEXT(8135, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8135(choice, ...)  METALANG99_PRIV_REC_NEXT(8136, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8136(choice, ...)  METALANG99_PRIV_REC_NEXT(8137, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8137(choice, ...)  METALANG99_PRIV_REC_NEXT(8138, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8138(choice, ...)  METALANG99_PRIV_REC_NEXT(8139, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8139(choice, ...)  METALANG99_PRIV_REC_NEXT(8140, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8140(choice, ...)  METALANG99_PRIV_REC_NEXT(8141, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8141(choice, ...)  METALANG99_PRIV_REC_NEXT(8142, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8142(choice, ...)  METALANG99_PRIV_REC_NEXT(8143, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8143(choice, ...)  METALANG99_PRIV_REC_NEXT(8144, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8144(choice, ...)  METALANG99_PRIV_REC_NEXT(8145, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8145(choice, ...)  METALANG99_PRIV_REC_NEXT(8146, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8146(choice, ...)  METALANG99_PRIV_REC_NEXT(8147, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8147(choice, ...)  METALANG99_PRIV_REC_NEXT(8148, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8148(choice, ...)  METALANG99_PRIV_REC_NEXT(8149, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8149(choice, ...)  METALANG99_PRIV_REC_NEXT(8150, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8150(choice, ...)  METALANG99_PRIV_REC_NEXT(8151, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8151(choice, ...)  METALANG99_PRIV_REC_NEXT(8152, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8152(choice, ...)  METALANG99_PRIV_REC_NEXT(8153, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8153(choice, ...)  METALANG99_PRIV_REC_NEXT(8154, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8154(choice, ...)  METALANG99_PRIV_REC_NEXT(8155, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8155(choice, ...)  METALANG99_PRIV_REC_NEXT(8156, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8156(choice, ...)  METALANG99_PRIV_REC_NEXT(8157, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8157(choice, ...)  METALANG99_PRIV_REC_NEXT(8158, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8158(choice, ...)  METALANG99_PRIV_REC_NEXT(8159, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8159(choice, ...)  METALANG99_PRIV_REC_NEXT(8160, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8160(choice, ...)  METALANG99_PRIV_REC_NEXT(8161, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8161(choice, ...)  METALANG99_PRIV_REC_NEXT(8162, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8162(choice, ...)  METALANG99_PRIV_REC_NEXT(8163, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8163(choice, ...)  METALANG99_PRIV_REC_NEXT(8164, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8164(choice, ...)  METALANG99_PRIV_REC_NEXT(8165, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8165(choice, ...)  METALANG99_PRIV_REC_NEXT(8166, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8166(choice, ...)  METALANG99_PRIV_REC_NEXT(8167, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8167(choice, ...)  METALANG99_PRIV_REC_NEXT(8168, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8168(choice, ...)  METALANG99_PRIV_REC_NEXT(8169, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8169(choice, ...)  METALANG99_PRIV_REC_NEXT(8170, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8170(choice, ...)  METALANG99_PRIV_REC_NEXT(8171, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8171(choice, ...)  METALANG99_PRIV_REC_NEXT(8172, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8172(choice, ...)  METALANG99_PRIV_REC_NEXT(8173, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8173(choice, ...)  METALANG99_PRIV_REC_NEXT(8174, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8174(choice, ...)  METALANG99_PRIV_REC_NEXT(8175, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8175(choice, ...)  METALANG99_PRIV_REC_NEXT(8176, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8176(choice, ...)  METALANG99_PRIV_REC_NEXT(8177, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8177(choice, ...)  METALANG99_PRIV_REC_NEXT(8178, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8178(choice, ...)  METALANG99_PRIV_REC_NEXT(8179, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8179(choice, ...)  METALANG99_PRIV_REC_NEXT(8180, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8180(choice, ...)  METALANG99_PRIV_REC_NEXT(8181, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8181(choice, ...)  METALANG99_PRIV_REC_NEXT(8182, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8182(choice, ...)  METALANG99_PRIV_REC_NEXT(8183, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8183(choice, ...)  METALANG99_PRIV_REC_NEXT(8184, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8184(choice, ...)  METALANG99_PRIV_REC_NEXT(8185, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8185(choice, ...)  METALANG99_PRIV_REC_NEXT(8186, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8186(choice, ...)  METALANG99_PRIV_REC_NEXT(8187, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8187(choice, ...)  METALANG99_PRIV_REC_NEXT(8188, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8188(choice, ...)  METALANG99_PRIV_REC_NEXT(8189, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8189(choice, ...)  METALANG99_PRIV_REC_NEXT(8190, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8190(choice, ...)  METALANG99_PRIV_REC_NEXT(8191, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8191(choice, ...)  METALANG99_PRIV_REC_NEXT(8192, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8192(choice, ...)  METALANG99_PRIV_REC_NEXT(8193, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8193(choice, ...)  METALANG99_PRIV_REC_NEXT(8194, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8194(choice, ...)  METALANG99_PRIV_REC_NEXT(8195, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8195(choice, ...)  METALANG99_PRIV_REC_NEXT(8196, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8196(choice, ...)  METALANG99_PRIV_REC_NEXT(8197, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8197(choice, ...)  METALANG99_PRIV_REC_NEXT(8198, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8198(choice, ...)  METALANG99_PRIV_REC_NEXT(8199, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8199(choice, ...)  METALANG99_PRIV_REC_NEXT(8200, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8200(choice, ...)  METALANG99_PRIV_REC_NEXT(8201, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8201(choice, ...)  METALANG99_PRIV_REC_NEXT(8202, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8202(choice, ...)  METALANG99_PRIV_REC_NEXT(8203, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8203(choice, ...)  METALANG99_PRIV_REC_NEXT(8204, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8204(choice, ...)  METALANG99_PRIV_REC_NEXT(8205, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8205(choice, ...)  METALANG99_PRIV_REC_NEXT(8206, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8206(choice, ...)  METALANG99_PRIV_REC_NEXT(8207, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8207(choice, ...)  METALANG99_PRIV_REC_NEXT(8208, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8208(choice, ...)  METALANG99_PRIV_REC_NEXT(8209, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8209(choice, ...)  METALANG99_PRIV_REC_NEXT(8210, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8210(choice, ...)  METALANG99_PRIV_REC_NEXT(8211, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8211(choice, ...)  METALANG99_PRIV_REC_NEXT(8212, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8212(choice, ...)  METALANG99_PRIV_REC_NEXT(8213, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8213(choice, ...)  METALANG99_PRIV_REC_NEXT(8214, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8214(choice, ...)  METALANG99_PRIV_REC_NEXT(8215, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8215(choice, ...)  METALANG99_PRIV_REC_NEXT(8216, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8216(choice, ...)  METALANG99_PRIV_REC_NEXT(8217, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8217(choice, ...)  METALANG99_PRIV_REC_NEXT(8218, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8218(choice, ...)  METALANG99_PRIV_REC_NEXT(8219, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8219(choice, ...)  METALANG99_PRIV_REC_NEXT(8220, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8220(choice, ...)  METALANG99_PRIV_REC_NEXT(8221, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8221(choice, ...)  METALANG99_PRIV_REC_NEXT(8222, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8222(choice, ...)  METALANG99_PRIV_REC_NEXT(8223, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8223(choice, ...)  METALANG99_PRIV_REC_NEXT(8224, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8224(choice, ...)  METALANG99_PRIV_REC_NEXT(8225, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8225(choice, ...)  METALANG99_PRIV_REC_NEXT(8226, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8226(choice, ...)  METALANG99_PRIV_REC_NEXT(8227, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8227(choice, ...)  METALANG99_PRIV_REC_NEXT(8228, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8228(choice, ...)  METALANG99_PRIV_REC_NEXT(8229, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8229(choice, ...)  METALANG99_PRIV_REC_NEXT(8230, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8230(choice, ...)  METALANG99_PRIV_REC_NEXT(8231, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8231(choice, ...)  METALANG99_PRIV_REC_NEXT(8232, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8232(choice, ...)  METALANG99_PRIV_REC_NEXT(8233, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8233(choice, ...)  METALANG99_PRIV_REC_NEXT(8234, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8234(choice, ...)  METALANG99_PRIV_REC_NEXT(8235, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8235(choice, ...)  METALANG99_PRIV_REC_NEXT(8236, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8236(choice, ...)  METALANG99_PRIV_REC_NEXT(8237, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8237(choice, ...)  METALANG99_PRIV_REC_NEXT(8238, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8238(choice, ...)  METALANG99_PRIV_REC_NEXT(8239, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8239(choice, ...)  METALANG99_PRIV_REC_NEXT(8240, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8240(choice, ...)  METALANG99_PRIV_REC_NEXT(8241, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8241(choice, ...)  METALANG99_PRIV_REC_NEXT(8242, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8242(choice, ...)  METALANG99_PRIV_REC_NEXT(8243, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8243(choice, ...)  METALANG99_PRIV_REC_NEXT(8244, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8244(choice, ...)  METALANG99_PRIV_REC_NEXT(8245, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8245(choice, ...)  METALANG99_PRIV_REC_NEXT(8246, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8246(choice, ...)  METALANG99_PRIV_REC_NEXT(8247, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8247(choice, ...)  METALANG99_PRIV_REC_NEXT(8248, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8248(choice, ...)  METALANG99_PRIV_REC_NEXT(8249, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8249(choice, ...)  METALANG99_PRIV_REC_NEXT(8250, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8250(choice, ...)  METALANG99_PRIV_REC_NEXT(8251, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8251(choice, ...)  METALANG99_PRIV_REC_NEXT(8252, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8252(choice, ...)  METALANG99_PRIV_REC_NEXT(8253, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8253(choice, ...)  METALANG99_PRIV_REC_NEXT(8254, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8254(choice, ...)  METALANG99_PRIV_REC_NEXT(8255, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8255(choice, ...)  METALANG99_PRIV_REC_NEXT(8256, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8256(choice, ...)  METALANG99_PRIV_REC_NEXT(8257, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8257(choice, ...)  METALANG99_PRIV_REC_NEXT(8258, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8258(choice, ...)  METALANG99_PRIV_REC_NEXT(8259, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8259(choice, ...)  METALANG99_PRIV_REC_NEXT(8260, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8260(choice, ...)  METALANG99_PRIV_REC_NEXT(8261, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8261(choice, ...)  METALANG99_PRIV_REC_NEXT(8262, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8262(choice, ...)  METALANG99_PRIV_REC_NEXT(8263, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8263(choice, ...)  METALANG99_PRIV_REC_NEXT(8264, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8264(choice, ...)  METALANG99_PRIV_REC_NEXT(8265, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8265(choice, ...)  METALANG99_PRIV_REC_NEXT(8266, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8266(choice, ...)  METALANG99_PRIV_REC_NEXT(8267, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8267(choice, ...)  METALANG99_PRIV_REC_NEXT(8268, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8268(choice, ...)  METALANG99_PRIV_REC_NEXT(8269, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8269(choice, ...)  METALANG99_PRIV_REC_NEXT(8270, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8270(choice, ...)  METALANG99_PRIV_REC_NEXT(8271, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8271(choice, ...)  METALANG99_PRIV_REC_NEXT(8272, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8272(choice, ...)  METALANG99_PRIV_REC_NEXT(8273, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8273(choice, ...)  METALANG99_PRIV_REC_NEXT(8274, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8274(choice, ...)  METALANG99_PRIV_REC_NEXT(8275, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8275(choice, ...)  METALANG99_PRIV_REC_NEXT(8276, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8276(choice, ...)  METALANG99_PRIV_REC_NEXT(8277, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8277(choice, ...)  METALANG99_PRIV_REC_NEXT(8278, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8278(choice, ...)  METALANG99_PRIV_REC_NEXT(8279, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8279(choice, ...)  METALANG99_PRIV_REC_NEXT(8280, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8280(choice, ...)  METALANG99_PRIV_REC_NEXT(8281, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8281(choice, ...)  METALANG99_PRIV_REC_NEXT(8282, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8282(choice, ...)  METALANG99_PRIV_REC_NEXT(8283, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8283(choice, ...)  METALANG99_PRIV_REC_NEXT(8284, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8284(choice, ...)  METALANG99_PRIV_REC_NEXT(8285, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8285(choice, ...)  METALANG99_PRIV_REC_NEXT(8286, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8286(choice, ...)  METALANG99_PRIV_REC_NEXT(8287, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8287(choice, ...)  METALANG99_PRIV_REC_NEXT(8288, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8288(choice, ...)  METALANG99_PRIV_REC_NEXT(8289, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8289(choice, ...)  METALANG99_PRIV_REC_NEXT(8290, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8290(choice, ...)  METALANG99_PRIV_REC_NEXT(8291, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8291(choice, ...)  METALANG99_PRIV_REC_NEXT(8292, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8292(choice, ...)  METALANG99_PRIV_REC_NEXT(8293, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8293(choice, ...)  METALANG99_PRIV_REC_NEXT(8294, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8294(choice, ...)  METALANG99_PRIV_REC_NEXT(8295, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8295(choice, ...)  METALANG99_PRIV_REC_NEXT(8296, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8296(choice, ...)  METALANG99_PRIV_REC_NEXT(8297, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8297(choice, ...)  METALANG99_PRIV_REC_NEXT(8298, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8298(choice, ...)  METALANG99_PRIV_REC_NEXT(8299, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8299(choice, ...)  METALANG99_PRIV_REC_NEXT(8300, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8300(choice, ...)  METALANG99_PRIV_REC_NEXT(8301, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8301(choice, ...)  METALANG99_PRIV_REC_NEXT(8302, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8302(choice, ...)  METALANG99_PRIV_REC_NEXT(8303, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8303(choice, ...)  METALANG99_PRIV_REC_NEXT(8304, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8304(choice, ...)  METALANG99_PRIV_REC_NEXT(8305, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8305(choice, ...)  METALANG99_PRIV_REC_NEXT(8306, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8306(choice, ...)  METALANG99_PRIV_REC_NEXT(8307, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8307(choice, ...)  METALANG99_PRIV_REC_NEXT(8308, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8308(choice, ...)  METALANG99_PRIV_REC_NEXT(8309, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8309(choice, ...)  METALANG99_PRIV_REC_NEXT(8310, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8310(choice, ...)  METALANG99_PRIV_REC_NEXT(8311, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8311(choice, ...)  METALANG99_PRIV_REC_NEXT(8312, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8312(choice, ...)  METALANG99_PRIV_REC_NEXT(8313, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8313(choice, ...)  METALANG99_PRIV_REC_NEXT(8314, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8314(choice, ...)  METALANG99_PRIV_REC_NEXT(8315, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8315(choice, ...)  METALANG99_PRIV_REC_NEXT(8316, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8316(choice, ...)  METALANG99_PRIV_REC_NEXT(8317, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8317(choice, ...)  METALANG99_PRIV_REC_NEXT(8318, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8318(choice, ...)  METALANG99_PRIV_REC_NEXT(8319, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8319(choice, ...)  METALANG99_PRIV_REC_NEXT(8320, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8320(choice, ...)  METALANG99_PRIV_REC_NEXT(8321, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8321(choice, ...)  METALANG99_PRIV_REC_NEXT(8322, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8322(choice, ...)  METALANG99_PRIV_REC_NEXT(8323, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8323(choice, ...)  METALANG99_PRIV_REC_NEXT(8324, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8324(choice, ...)  METALANG99_PRIV_REC_NEXT(8325, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8325(choice, ...)  METALANG99_PRIV_REC_NEXT(8326, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8326(choice, ...)  METALANG99_PRIV_REC_NEXT(8327, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8327(choice, ...)  METALANG99_PRIV_REC_NEXT(8328, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8328(choice, ...)  METALANG99_PRIV_REC_NEXT(8329, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8329(choice, ...)  METALANG99_PRIV_REC_NEXT(8330, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8330(choice, ...)  METALANG99_PRIV_REC_NEXT(8331, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8331(choice, ...)  METALANG99_PRIV_REC_NEXT(8332, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8332(choice, ...)  METALANG99_PRIV_REC_NEXT(8333, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8333(choice, ...)  METALANG99_PRIV_REC_NEXT(8334, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8334(choice, ...)  METALANG99_PRIV_REC_NEXT(8335, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8335(choice, ...)  METALANG99_PRIV_REC_NEXT(8336, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8336(choice, ...)  METALANG99_PRIV_REC_NEXT(8337, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8337(choice, ...)  METALANG99_PRIV_REC_NEXT(8338, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8338(choice, ...)  METALANG99_PRIV_REC_NEXT(8339, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8339(choice, ...)  METALANG99_PRIV_REC_NEXT(8340, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8340(choice, ...)  METALANG99_PRIV_REC_NEXT(8341, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8341(choice, ...)  METALANG99_PRIV_REC_NEXT(8342, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8342(choice, ...)  METALANG99_PRIV_REC_NEXT(8343, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8343(choice, ...)  METALANG99_PRIV_REC_NEXT(8344, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8344(choice, ...)  METALANG99_PRIV_REC_NEXT(8345, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8345(choice, ...)  METALANG99_PRIV_REC_NEXT(8346, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8346(choice, ...)  METALANG99_PRIV_REC_NEXT(8347, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8347(choice, ...)  METALANG99_PRIV_REC_NEXT(8348, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8348(choice, ...)  METALANG99_PRIV_REC_NEXT(8349, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8349(choice, ...)  METALANG99_PRIV_REC_NEXT(8350, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8350(choice, ...)  METALANG99_PRIV_REC_NEXT(8351, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8351(choice, ...)  METALANG99_PRIV_REC_NEXT(8352, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8352(choice, ...)  METALANG99_PRIV_REC_NEXT(8353, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8353(choice, ...)  METALANG99_PRIV_REC_NEXT(8354, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8354(choice, ...)  METALANG99_PRIV_REC_NEXT(8355, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8355(choice, ...)  METALANG99_PRIV_REC_NEXT(8356, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8356(choice, ...)  METALANG99_PRIV_REC_NEXT(8357, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8357(choice, ...)  METALANG99_PRIV_REC_NEXT(8358, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8358(choice, ...)  METALANG99_PRIV_REC_NEXT(8359, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8359(choice, ...)  METALANG99_PRIV_REC_NEXT(8360, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8360(choice, ...)  METALANG99_PRIV_REC_NEXT(8361, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8361(choice, ...)  METALANG99_PRIV_REC_NEXT(8362, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8362(choice, ...)  METALANG99_PRIV_REC_NEXT(8363, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8363(choice, ...)  METALANG99_PRIV_REC_NEXT(8364, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8364(choice, ...)  METALANG99_PRIV_REC_NEXT(8365, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8365(choice, ...)  METALANG99_PRIV_REC_NEXT(8366, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8366(choice, ...)  METALANG99_PRIV_REC_NEXT(8367, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8367(choice, ...)  METALANG99_PRIV_REC_NEXT(8368, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8368(choice, ...)  METALANG99_PRIV_REC_NEXT(8369, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8369(choice, ...)  METALANG99_PRIV_REC_NEXT(8370, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8370(choice, ...)  METALANG99_PRIV_REC_NEXT(8371, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8371(choice, ...)  METALANG99_PRIV_REC_NEXT(8372, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8372(choice, ...)  METALANG99_PRIV_REC_NEXT(8373, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8373(choice, ...)  METALANG99_PRIV_REC_NEXT(8374, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8374(choice, ...)  METALANG99_PRIV_REC_NEXT(8375, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8375(choice, ...)  METALANG99_PRIV_REC_NEXT(8376, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8376(choice, ...)  METALANG99_PRIV_REC_NEXT(8377, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8377(choice, ...)  METALANG99_PRIV_REC_NEXT(8378, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8378(choice, ...)  METALANG99_PRIV_REC_NEXT(8379, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8379(choice, ...)  METALANG99_PRIV_REC_NEXT(8380, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8380(choice, ...)  METALANG99_PRIV_REC_NEXT(8381, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8381(choice, ...)  METALANG99_PRIV_REC_NEXT(8382, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8382(choice, ...)  METALANG99_PRIV_REC_NEXT(8383, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8383(choice, ...)  METALANG99_PRIV_REC_NEXT(8384, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8384(choice, ...)  METALANG99_PRIV_REC_NEXT(8385, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8385(choice, ...)  METALANG99_PRIV_REC_NEXT(8386, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8386(choice, ...)  METALANG99_PRIV_REC_NEXT(8387, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8387(choice, ...)  METALANG99_PRIV_REC_NEXT(8388, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8388(choice, ...)  METALANG99_PRIV_REC_NEXT(8389, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8389(choice, ...)  METALANG99_PRIV_REC_NEXT(8390, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8390(choice, ...)  METALANG99_PRIV_REC_NEXT(8391, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8391(choice, ...)  METALANG99_PRIV_REC_NEXT(8392, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8392(choice, ...)  METALANG99_PRIV_REC_NEXT(8393, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8393(choice, ...)  METALANG99_PRIV_REC_NEXT(8394, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8394(choice, ...)  METALANG99_PRIV_REC_NEXT(8395, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8395(choice, ...)  METALANG99_PRIV_REC_NEXT(8396, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8396(choice, ...)  METALANG99_PRIV_REC_NEXT(8397, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8397(choice, ...)  METALANG99_PRIV_REC_NEXT(8398, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8398(choice, ...)  METALANG99_PRIV_REC_NEXT(8399, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8399(choice, ...)  METALANG99_PRIV_REC_NEXT(8400, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8400(choice, ...)  METALANG99_PRIV_REC_NEXT(8401, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8401(choice, ...)  METALANG99_PRIV_REC_NEXT(8402, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8402(choice, ...)  METALANG99_PRIV_REC_NEXT(8403, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8403(choice, ...)  METALANG99_PRIV_REC_NEXT(8404, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8404(choice, ...)  METALANG99_PRIV_REC_NEXT(8405, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8405(choice, ...)  METALANG99_PRIV_REC_NEXT(8406, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8406(choice, ...)  METALANG99_PRIV_REC_NEXT(8407, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8407(choice, ...)  METALANG99_PRIV_REC_NEXT(8408, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8408(choice, ...)  METALANG99_PRIV_REC_NEXT(8409, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8409(choice, ...)  METALANG99_PRIV_REC_NEXT(8410, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8410(choice, ...)  METALANG99_PRIV_REC_NEXT(8411, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8411(choice, ...)  METALANG99_PRIV_REC_NEXT(8412, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8412(choice, ...)  METALANG99_PRIV_REC_NEXT(8413, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8413(choice, ...)  METALANG99_PRIV_REC_NEXT(8414, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8414(choice, ...)  METALANG99_PRIV_REC_NEXT(8415, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8415(choice, ...)  METALANG99_PRIV_REC_NEXT(8416, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8416(choice, ...)  METALANG99_PRIV_REC_NEXT(8417, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8417(choice, ...)  METALANG99_PRIV_REC_NEXT(8418, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8418(choice, ...)  METALANG99_PRIV_REC_NEXT(8419, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8419(choice, ...)  METALANG99_PRIV_REC_NEXT(8420, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8420(choice, ...)  METALANG99_PRIV_REC_NEXT(8421, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8421(choice, ...)  METALANG99_PRIV_REC_NEXT(8422, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8422(choice, ...)  METALANG99_PRIV_REC_NEXT(8423, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8423(choice, ...)  METALANG99_PRIV_REC_NEXT(8424, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8424(choice, ...)  METALANG99_PRIV_REC_NEXT(8425, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8425(choice, ...)  METALANG99_PRIV_REC_NEXT(8426, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8426(choice, ...)  METALANG99_PRIV_REC_NEXT(8427, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8427(choice, ...)  METALANG99_PRIV_REC_NEXT(8428, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8428(choice, ...)  METALANG99_PRIV_REC_NEXT(8429, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8429(choice, ...)  METALANG99_PRIV_REC_NEXT(8430, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8430(choice, ...)  METALANG99_PRIV_REC_NEXT(8431, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8431(choice, ...)  METALANG99_PRIV_REC_NEXT(8432, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8432(choice, ...)  METALANG99_PRIV_REC_NEXT(8433, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8433(choice, ...)  METALANG99_PRIV_REC_NEXT(8434, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8434(choice, ...)  METALANG99_PRIV_REC_NEXT(8435, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8435(choice, ...)  METALANG99_PRIV_REC_NEXT(8436, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8436(choice, ...)  METALANG99_PRIV_REC_NEXT(8437, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8437(choice, ...)  METALANG99_PRIV_REC_NEXT(8438, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8438(choice, ...)  METALANG99_PRIV_REC_NEXT(8439, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8439(choice, ...)  METALANG99_PRIV_REC_NEXT(8440, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8440(choice, ...)  METALANG99_PRIV_REC_NEXT(8441, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8441(choice, ...)  METALANG99_PRIV_REC_NEXT(8442, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8442(choice, ...)  METALANG99_PRIV_REC_NEXT(8443, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8443(choice, ...)  METALANG99_PRIV_REC_NEXT(8444, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8444(choice, ...)  METALANG99_PRIV_REC_NEXT(8445, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8445(choice, ...)  METALANG99_PRIV_REC_NEXT(8446, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8446(choice, ...)  METALANG99_PRIV_REC_NEXT(8447, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8447(choice, ...)  METALANG99_PRIV_REC_NEXT(8448, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8448(choice, ...)  METALANG99_PRIV_REC_NEXT(8449, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8449(choice, ...)  METALANG99_PRIV_REC_NEXT(8450, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8450(choice, ...)  METALANG99_PRIV_REC_NEXT(8451, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8451(choice, ...)  METALANG99_PRIV_REC_NEXT(8452, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8452(choice, ...)  METALANG99_PRIV_REC_NEXT(8453, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8453(choice, ...)  METALANG99_PRIV_REC_NEXT(8454, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8454(choice, ...)  METALANG99_PRIV_REC_NEXT(8455, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8455(choice, ...)  METALANG99_PRIV_REC_NEXT(8456, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8456(choice, ...)  METALANG99_PRIV_REC_NEXT(8457, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8457(choice, ...)  METALANG99_PRIV_REC_NEXT(8458, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8458(choice, ...)  METALANG99_PRIV_REC_NEXT(8459, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8459(choice, ...)  METALANG99_PRIV_REC_NEXT(8460, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8460(choice, ...)  METALANG99_PRIV_REC_NEXT(8461, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8461(choice, ...)  METALANG99_PRIV_REC_NEXT(8462, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8462(choice, ...)  METALANG99_PRIV_REC_NEXT(8463, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8463(choice, ...)  METALANG99_PRIV_REC_NEXT(8464, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8464(choice, ...)  METALANG99_PRIV_REC_NEXT(8465, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8465(choice, ...)  METALANG99_PRIV_REC_NEXT(8466, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8466(choice, ...)  METALANG99_PRIV_REC_NEXT(8467, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8467(choice, ...)  METALANG99_PRIV_REC_NEXT(8468, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8468(choice, ...)  METALANG99_PRIV_REC_NEXT(8469, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8469(choice, ...)  METALANG99_PRIV_REC_NEXT(8470, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8470(choice, ...)  METALANG99_PRIV_REC_NEXT(8471, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8471(choice, ...)  METALANG99_PRIV_REC_NEXT(8472, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8472(choice, ...)  METALANG99_PRIV_REC_NEXT(8473, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8473(choice, ...)  METALANG99_PRIV_REC_NEXT(8474, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8474(choice, ...)  METALANG99_PRIV_REC_NEXT(8475, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8475(choice, ...)  METALANG99_PRIV_REC_NEXT(8476, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8476(choice, ...)  METALANG99_PRIV_REC_NEXT(8477, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8477(choice, ...)  METALANG99_PRIV_REC_NEXT(8478, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8478(choice, ...)  METALANG99_PRIV_REC_NEXT(8479, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8479(choice, ...)  METALANG99_PRIV_REC_NEXT(8480, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8480(choice, ...)  METALANG99_PRIV_REC_NEXT(8481, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8481(choice, ...)  METALANG99_PRIV_REC_NEXT(8482, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8482(choice, ...)  METALANG99_PRIV_REC_NEXT(8483, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8483(choice, ...)  METALANG99_PRIV_REC_NEXT(8484, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8484(choice, ...)  METALANG99_PRIV_REC_NEXT(8485, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8485(choice, ...)  METALANG99_PRIV_REC_NEXT(8486, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8486(choice, ...)  METALANG99_PRIV_REC_NEXT(8487, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8487(choice, ...)  METALANG99_PRIV_REC_NEXT(8488, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8488(choice, ...)  METALANG99_PRIV_REC_NEXT(8489, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8489(choice, ...)  METALANG99_PRIV_REC_NEXT(8490, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8490(choice, ...)  METALANG99_PRIV_REC_NEXT(8491, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8491(choice, ...)  METALANG99_PRIV_REC_NEXT(8492, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8492(choice, ...)  METALANG99_PRIV_REC_NEXT(8493, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8493(choice, ...)  METALANG99_PRIV_REC_NEXT(8494, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8494(choice, ...)  METALANG99_PRIV_REC_NEXT(8495, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8495(choice, ...)  METALANG99_PRIV_REC_NEXT(8496, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8496(choice, ...)  METALANG99_PRIV_REC_NEXT(8497, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8497(choice, ...)  METALANG99_PRIV_REC_NEXT(8498, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8498(choice, ...)  METALANG99_PRIV_REC_NEXT(8499, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8499(choice, ...)  METALANG99_PRIV_REC_NEXT(8500, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8500(choice, ...)  METALANG99_PRIV_REC_NEXT(8501, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8501(choice, ...)  METALANG99_PRIV_REC_NEXT(8502, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8502(choice, ...)  METALANG99_PRIV_REC_NEXT(8503, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8503(choice, ...)  METALANG99_PRIV_REC_NEXT(8504, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8504(choice, ...)  METALANG99_PRIV_REC_NEXT(8505, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8505(choice, ...)  METALANG99_PRIV_REC_NEXT(8506, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8506(choice, ...)  METALANG99_PRIV_REC_NEXT(8507, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8507(choice, ...)  METALANG99_PRIV_REC_NEXT(8508, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8508(choice, ...)  METALANG99_PRIV_REC_NEXT(8509, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8509(choice, ...)  METALANG99_PRIV_REC_NEXT(8510, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8510(choice, ...)  METALANG99_PRIV_REC_NEXT(8511, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8511(choice, ...)  METALANG99_PRIV_REC_NEXT(8512, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8512(choice, ...)  METALANG99_PRIV_REC_NEXT(8513, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8513(choice, ...)  METALANG99_PRIV_REC_NEXT(8514, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8514(choice, ...)  METALANG99_PRIV_REC_NEXT(8515, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8515(choice, ...)  METALANG99_PRIV_REC_NEXT(8516, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8516(choice, ...)  METALANG99_PRIV_REC_NEXT(8517, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8517(choice, ...)  METALANG99_PRIV_REC_NEXT(8518, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8518(choice, ...)  METALANG99_PRIV_REC_NEXT(8519, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8519(choice, ...)  METALANG99_PRIV_REC_NEXT(8520, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8520(choice, ...)  METALANG99_PRIV_REC_NEXT(8521, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8521(choice, ...)  METALANG99_PRIV_REC_NEXT(8522, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8522(choice, ...)  METALANG99_PRIV_REC_NEXT(8523, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8523(choice, ...)  METALANG99_PRIV_REC_NEXT(8524, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8524(choice, ...)  METALANG99_PRIV_REC_NEXT(8525, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8525(choice, ...)  METALANG99_PRIV_REC_NEXT(8526, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8526(choice, ...)  METALANG99_PRIV_REC_NEXT(8527, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8527(choice, ...)  METALANG99_PRIV_REC_NEXT(8528, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8528(choice, ...)  METALANG99_PRIV_REC_NEXT(8529, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8529(choice, ...)  METALANG99_PRIV_REC_NEXT(8530, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8530(choice, ...)  METALANG99_PRIV_REC_NEXT(8531, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8531(choice, ...)  METALANG99_PRIV_REC_NEXT(8532, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8532(choice, ...)  METALANG99_PRIV_REC_NEXT(8533, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8533(choice, ...)  METALANG99_PRIV_REC_NEXT(8534, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8534(choice, ...)  METALANG99_PRIV_REC_NEXT(8535, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8535(choice, ...)  METALANG99_PRIV_REC_NEXT(8536, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8536(choice, ...)  METALANG99_PRIV_REC_NEXT(8537, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8537(choice, ...)  METALANG99_PRIV_REC_NEXT(8538, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8538(choice, ...)  METALANG99_PRIV_REC_NEXT(8539, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8539(choice, ...)  METALANG99_PRIV_REC_NEXT(8540, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8540(choice, ...)  METALANG99_PRIV_REC_NEXT(8541, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8541(choice, ...)  METALANG99_PRIV_REC_NEXT(8542, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8542(choice, ...)  METALANG99_PRIV_REC_NEXT(8543, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8543(choice, ...)  METALANG99_PRIV_REC_NEXT(8544, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8544(choice, ...)  METALANG99_PRIV_REC_NEXT(8545, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8545(choice, ...)  METALANG99_PRIV_REC_NEXT(8546, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8546(choice, ...)  METALANG99_PRIV_REC_NEXT(8547, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8547(choice, ...)  METALANG99_PRIV_REC_NEXT(8548, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8548(choice, ...)  METALANG99_PRIV_REC_NEXT(8549, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8549(choice, ...)  METALANG99_PRIV_REC_NEXT(8550, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8550(choice, ...)  METALANG99_PRIV_REC_NEXT(8551, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8551(choice, ...)  METALANG99_PRIV_REC_NEXT(8552, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8552(choice, ...)  METALANG99_PRIV_REC_NEXT(8553, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8553(choice, ...)  METALANG99_PRIV_REC_NEXT(8554, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8554(choice, ...)  METALANG99_PRIV_REC_NEXT(8555, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8555(choice, ...)  METALANG99_PRIV_REC_NEXT(8556, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8556(choice, ...)  METALANG99_PRIV_REC_NEXT(8557, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8557(choice, ...)  METALANG99_PRIV_REC_NEXT(8558, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8558(choice, ...)  METALANG99_PRIV_REC_NEXT(8559, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8559(choice, ...)  METALANG99_PRIV_REC_NEXT(8560, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8560(choice, ...)  METALANG99_PRIV_REC_NEXT(8561, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8561(choice, ...)  METALANG99_PRIV_REC_NEXT(8562, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8562(choice, ...)  METALANG99_PRIV_REC_NEXT(8563, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8563(choice, ...)  METALANG99_PRIV_REC_NEXT(8564, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8564(choice, ...)  METALANG99_PRIV_REC_NEXT(8565, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8565(choice, ...)  METALANG99_PRIV_REC_NEXT(8566, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8566(choice, ...)  METALANG99_PRIV_REC_NEXT(8567, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8567(choice, ...)  METALANG99_PRIV_REC_NEXT(8568, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8568(choice, ...)  METALANG99_PRIV_REC_NEXT(8569, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8569(choice, ...)  METALANG99_PRIV_REC_NEXT(8570, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8570(choice, ...)  METALANG99_PRIV_REC_NEXT(8571, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8571(choice, ...)  METALANG99_PRIV_REC_NEXT(8572, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8572(choice, ...)  METALANG99_PRIV_REC_NEXT(8573, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8573(choice, ...)  METALANG99_PRIV_REC_NEXT(8574, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8574(choice, ...)  METALANG99_PRIV_REC_NEXT(8575, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8575(choice, ...)  METALANG99_PRIV_REC_NEXT(8576, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8576(choice, ...)  METALANG99_PRIV_REC_NEXT(8577, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8577(choice, ...)  METALANG99_PRIV_REC_NEXT(8578, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8578(choice, ...)  METALANG99_PRIV_REC_NEXT(8579, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8579(choice, ...)  METALANG99_PRIV_REC_NEXT(8580, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8580(choice, ...)  METALANG99_PRIV_REC_NEXT(8581, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8581(choice, ...)  METALANG99_PRIV_REC_NEXT(8582, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8582(choice, ...)  METALANG99_PRIV_REC_NEXT(8583, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8583(choice, ...)  METALANG99_PRIV_REC_NEXT(8584, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8584(choice, ...)  METALANG99_PRIV_REC_NEXT(8585, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8585(choice, ...)  METALANG99_PRIV_REC_NEXT(8586, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8586(choice, ...)  METALANG99_PRIV_REC_NEXT(8587, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8587(choice, ...)  METALANG99_PRIV_REC_NEXT(8588, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8588(choice, ...)  METALANG99_PRIV_REC_NEXT(8589, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8589(choice, ...)  METALANG99_PRIV_REC_NEXT(8590, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8590(choice, ...)  METALANG99_PRIV_REC_NEXT(8591, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8591(choice, ...)  METALANG99_PRIV_REC_NEXT(8592, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8592(choice, ...)  METALANG99_PRIV_REC_NEXT(8593, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8593(choice, ...)  METALANG99_PRIV_REC_NEXT(8594, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8594(choice, ...)  METALANG99_PRIV_REC_NEXT(8595, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8595(choice, ...)  METALANG99_PRIV_REC_NEXT(8596, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8596(choice, ...)  METALANG99_PRIV_REC_NEXT(8597, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8597(choice, ...)  METALANG99_PRIV_REC_NEXT(8598, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8598(choice, ...)  METALANG99_PRIV_REC_NEXT(8599, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8599(choice, ...)  METALANG99_PRIV_REC_NEXT(8600, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8600(choice, ...)  METALANG99_PRIV_REC_NEXT(8601, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8601(choice, ...)  METALANG99_PRIV_REC_NEXT(8602, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8602(choice, ...)  METALANG99_PRIV_REC_NEXT(8603, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8603(choice, ...)  METALANG99_PRIV_REC_NEXT(8604, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8604(choice, ...)  METALANG99_PRIV_REC_NEXT(8605, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8605(choice, ...)  METALANG99_PRIV_REC_NEXT(8606, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8606(choice, ...)  METALANG99_PRIV_REC_NEXT(8607, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8607(choice, ...)  METALANG99_PRIV_REC_NEXT(8608, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8608(choice, ...)  METALANG99_PRIV_REC_NEXT(8609, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8609(choice, ...)  METALANG99_PRIV_REC_NEXT(8610, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8610(choice, ...)  METALANG99_PRIV_REC_NEXT(8611, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8611(choice, ...)  METALANG99_PRIV_REC_NEXT(8612, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8612(choice, ...)  METALANG99_PRIV_REC_NEXT(8613, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8613(choice, ...)  METALANG99_PRIV_REC_NEXT(8614, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8614(choice, ...)  METALANG99_PRIV_REC_NEXT(8615, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8615(choice, ...)  METALANG99_PRIV_REC_NEXT(8616, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8616(choice, ...)  METALANG99_PRIV_REC_NEXT(8617, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8617(choice, ...)  METALANG99_PRIV_REC_NEXT(8618, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8618(choice, ...)  METALANG99_PRIV_REC_NEXT(8619, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8619(choice, ...)  METALANG99_PRIV_REC_NEXT(8620, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8620(choice, ...)  METALANG99_PRIV_REC_NEXT(8621, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8621(choice, ...)  METALANG99_PRIV_REC_NEXT(8622, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8622(choice, ...)  METALANG99_PRIV_REC_NEXT(8623, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8623(choice, ...)  METALANG99_PRIV_REC_NEXT(8624, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8624(choice, ...)  METALANG99_PRIV_REC_NEXT(8625, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8625(choice, ...)  METALANG99_PRIV_REC_NEXT(8626, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8626(choice, ...)  METALANG99_PRIV_REC_NEXT(8627, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8627(choice, ...)  METALANG99_PRIV_REC_NEXT(8628, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8628(choice, ...)  METALANG99_PRIV_REC_NEXT(8629, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8629(choice, ...)  METALANG99_PRIV_REC_NEXT(8630, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8630(choice, ...)  METALANG99_PRIV_REC_NEXT(8631, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8631(choice, ...)  METALANG99_PRIV_REC_NEXT(8632, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8632(choice, ...)  METALANG99_PRIV_REC_NEXT(8633, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8633(choice, ...)  METALANG99_PRIV_REC_NEXT(8634, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8634(choice, ...)  METALANG99_PRIV_REC_NEXT(8635, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8635(choice, ...)  METALANG99_PRIV_REC_NEXT(8636, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8636(choice, ...)  METALANG99_PRIV_REC_NEXT(8637, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8637(choice, ...)  METALANG99_PRIV_REC_NEXT(8638, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8638(choice, ...)  METALANG99_PRIV_REC_NEXT(8639, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8639(choice, ...)  METALANG99_PRIV_REC_NEXT(8640, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8640(choice, ...)  METALANG99_PRIV_REC_NEXT(8641, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8641(choice, ...)  METALANG99_PRIV_REC_NEXT(8642, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8642(choice, ...)  METALANG99_PRIV_REC_NEXT(8643, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8643(choice, ...)  METALANG99_PRIV_REC_NEXT(8644, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8644(choice, ...)  METALANG99_PRIV_REC_NEXT(8645, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8645(choice, ...)  METALANG99_PRIV_REC_NEXT(8646, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8646(choice, ...)  METALANG99_PRIV_REC_NEXT(8647, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8647(choice, ...)  METALANG99_PRIV_REC_NEXT(8648, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8648(choice, ...)  METALANG99_PRIV_REC_NEXT(8649, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8649(choice, ...)  METALANG99_PRIV_REC_NEXT(8650, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8650(choice, ...)  METALANG99_PRIV_REC_NEXT(8651, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8651(choice, ...)  METALANG99_PRIV_REC_NEXT(8652, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8652(choice, ...)  METALANG99_PRIV_REC_NEXT(8653, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8653(choice, ...)  METALANG99_PRIV_REC_NEXT(8654, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8654(choice, ...)  METALANG99_PRIV_REC_NEXT(8655, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8655(choice, ...)  METALANG99_PRIV_REC_NEXT(8656, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8656(choice, ...)  METALANG99_PRIV_REC_NEXT(8657, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8657(choice, ...)  METALANG99_PRIV_REC_NEXT(8658, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8658(choice, ...)  METALANG99_PRIV_REC_NEXT(8659, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8659(choice, ...)  METALANG99_PRIV_REC_NEXT(8660, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8660(choice, ...)  METALANG99_PRIV_REC_NEXT(8661, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8661(choice, ...)  METALANG99_PRIV_REC_NEXT(8662, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8662(choice, ...)  METALANG99_PRIV_REC_NEXT(8663, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8663(choice, ...)  METALANG99_PRIV_REC_NEXT(8664, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8664(choice, ...)  METALANG99_PRIV_REC_NEXT(8665, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8665(choice, ...)  METALANG99_PRIV_REC_NEXT(8666, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8666(choice, ...)  METALANG99_PRIV_REC_NEXT(8667, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8667(choice, ...)  METALANG99_PRIV_REC_NEXT(8668, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8668(choice, ...)  METALANG99_PRIV_REC_NEXT(8669, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8669(choice, ...)  METALANG99_PRIV_REC_NEXT(8670, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8670(choice, ...)  METALANG99_PRIV_REC_NEXT(8671, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8671(choice, ...)  METALANG99_PRIV_REC_NEXT(8672, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8672(choice, ...)  METALANG99_PRIV_REC_NEXT(8673, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8673(choice, ...)  METALANG99_PRIV_REC_NEXT(8674, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8674(choice, ...)  METALANG99_PRIV_REC_NEXT(8675, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8675(choice, ...)  METALANG99_PRIV_REC_NEXT(8676, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8676(choice, ...)  METALANG99_PRIV_REC_NEXT(8677, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8677(choice, ...)  METALANG99_PRIV_REC_NEXT(8678, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8678(choice, ...)  METALANG99_PRIV_REC_NEXT(8679, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8679(choice, ...)  METALANG99_PRIV_REC_NEXT(8680, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8680(choice, ...)  METALANG99_PRIV_REC_NEXT(8681, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8681(choice, ...)  METALANG99_PRIV_REC_NEXT(8682, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8682(choice, ...)  METALANG99_PRIV_REC_NEXT(8683, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8683(choice, ...)  METALANG99_PRIV_REC_NEXT(8684, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8684(choice, ...)  METALANG99_PRIV_REC_NEXT(8685, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8685(choice, ...)  METALANG99_PRIV_REC_NEXT(8686, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8686(choice, ...)  METALANG99_PRIV_REC_NEXT(8687, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8687(choice, ...)  METALANG99_PRIV_REC_NEXT(8688, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8688(choice, ...)  METALANG99_PRIV_REC_NEXT(8689, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8689(choice, ...)  METALANG99_PRIV_REC_NEXT(8690, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8690(choice, ...)  METALANG99_PRIV_REC_NEXT(8691, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8691(choice, ...)  METALANG99_PRIV_REC_NEXT(8692, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8692(choice, ...)  METALANG99_PRIV_REC_NEXT(8693, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8693(choice, ...)  METALANG99_PRIV_REC_NEXT(8694, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8694(choice, ...)  METALANG99_PRIV_REC_NEXT(8695, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8695(choice, ...)  METALANG99_PRIV_REC_NEXT(8696, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8696(choice, ...)  METALANG99_PRIV_REC_NEXT(8697, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8697(choice, ...)  METALANG99_PRIV_REC_NEXT(8698, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8698(choice, ...)  METALANG99_PRIV_REC_NEXT(8699, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8699(choice, ...)  METALANG99_PRIV_REC_NEXT(8700, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8700(choice, ...)  METALANG99_PRIV_REC_NEXT(8701, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8701(choice, ...)  METALANG99_PRIV_REC_NEXT(8702, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8702(choice, ...)  METALANG99_PRIV_REC_NEXT(8703, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8703(choice, ...)  METALANG99_PRIV_REC_NEXT(8704, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8704(choice, ...)  METALANG99_PRIV_REC_NEXT(8705, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8705(choice, ...)  METALANG99_PRIV_REC_NEXT(8706, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8706(choice, ...)  METALANG99_PRIV_REC_NEXT(8707, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8707(choice, ...)  METALANG99_PRIV_REC_NEXT(8708, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8708(choice, ...)  METALANG99_PRIV_REC_NEXT(8709, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8709(choice, ...)  METALANG99_PRIV_REC_NEXT(8710, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8710(choice, ...)  METALANG99_PRIV_REC_NEXT(8711, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8711(choice, ...)  METALANG99_PRIV_REC_NEXT(8712, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8712(choice, ...)  METALANG99_PRIV_REC_NEXT(8713, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8713(choice, ...)  METALANG99_PRIV_REC_NEXT(8714, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8714(choice, ...)  METALANG99_PRIV_REC_NEXT(8715, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8715(choice, ...)  METALANG99_PRIV_REC_NEXT(8716, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8716(choice, ...)  METALANG99_PRIV_REC_NEXT(8717, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8717(choice, ...)  METALANG99_PRIV_REC_NEXT(8718, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8718(choice, ...)  METALANG99_PRIV_REC_NEXT(8719, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8719(choice, ...)  METALANG99_PRIV_REC_NEXT(8720, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8720(choice, ...)  METALANG99_PRIV_REC_NEXT(8721, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8721(choice, ...)  METALANG99_PRIV_REC_NEXT(8722, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8722(choice, ...)  METALANG99_PRIV_REC_NEXT(8723, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8723(choice, ...)  METALANG99_PRIV_REC_NEXT(8724, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8724(choice, ...)  METALANG99_PRIV_REC_NEXT(8725, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8725(choice, ...)  METALANG99_PRIV_REC_NEXT(8726, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8726(choice, ...)  METALANG99_PRIV_REC_NEXT(8727, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8727(choice, ...)  METALANG99_PRIV_REC_NEXT(8728, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8728(choice, ...)  METALANG99_PRIV_REC_NEXT(8729, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8729(choice, ...)  METALANG99_PRIV_REC_NEXT(8730, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8730(choice, ...)  METALANG99_PRIV_REC_NEXT(8731, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8731(choice, ...)  METALANG99_PRIV_REC_NEXT(8732, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8732(choice, ...)  METALANG99_PRIV_REC_NEXT(8733, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8733(choice, ...)  METALANG99_PRIV_REC_NEXT(8734, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8734(choice, ...)  METALANG99_PRIV_REC_NEXT(8735, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8735(choice, ...)  METALANG99_PRIV_REC_NEXT(8736, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8736(choice, ...)  METALANG99_PRIV_REC_NEXT(8737, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8737(choice, ...)  METALANG99_PRIV_REC_NEXT(8738, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8738(choice, ...)  METALANG99_PRIV_REC_NEXT(8739, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8739(choice, ...)  METALANG99_PRIV_REC_NEXT(8740, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8740(choice, ...)  METALANG99_PRIV_REC_NEXT(8741, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8741(choice, ...)  METALANG99_PRIV_REC_NEXT(8742, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8742(choice, ...)  METALANG99_PRIV_REC_NEXT(8743, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8743(choice, ...)  METALANG99_PRIV_REC_NEXT(8744, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8744(choice, ...)  METALANG99_PRIV_REC_NEXT(8745, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8745(choice, ...)  METALANG99_PRIV_REC_NEXT(8746, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8746(choice, ...)  METALANG99_PRIV_REC_NEXT(8747, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8747(choice, ...)  METALANG99_PRIV_REC_NEXT(8748, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8748(choice, ...)  METALANG99_PRIV_REC_NEXT(8749, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8749(choice, ...)  METALANG99_PRIV_REC_NEXT(8750, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8750(choice, ...)  METALANG99_PRIV_REC_NEXT(8751, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8751(choice, ...)  METALANG99_PRIV_REC_NEXT(8752, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8752(choice, ...)  METALANG99_PRIV_REC_NEXT(8753, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8753(choice, ...)  METALANG99_PRIV_REC_NEXT(8754, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8754(choice, ...)  METALANG99_PRIV_REC_NEXT(8755, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8755(choice, ...)  METALANG99_PRIV_REC_NEXT(8756, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8756(choice, ...)  METALANG99_PRIV_REC_NEXT(8757, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8757(choice, ...)  METALANG99_PRIV_REC_NEXT(8758, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8758(choice, ...)  METALANG99_PRIV_REC_NEXT(8759, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8759(choice, ...)  METALANG99_PRIV_REC_NEXT(8760, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8760(choice, ...)  METALANG99_PRIV_REC_NEXT(8761, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8761(choice, ...)  METALANG99_PRIV_REC_NEXT(8762, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8762(choice, ...)  METALANG99_PRIV_REC_NEXT(8763, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8763(choice, ...)  METALANG99_PRIV_REC_NEXT(8764, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8764(choice, ...)  METALANG99_PRIV_REC_NEXT(8765, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8765(choice, ...)  METALANG99_PRIV_REC_NEXT(8766, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8766(choice, ...)  METALANG99_PRIV_REC_NEXT(8767, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8767(choice, ...)  METALANG99_PRIV_REC_NEXT(8768, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8768(choice, ...)  METALANG99_PRIV_REC_NEXT(8769, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8769(choice, ...)  METALANG99_PRIV_REC_NEXT(8770, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8770(choice, ...)  METALANG99_PRIV_REC_NEXT(8771, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8771(choice, ...)  METALANG99_PRIV_REC_NEXT(8772, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8772(choice, ...)  METALANG99_PRIV_REC_NEXT(8773, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8773(choice, ...)  METALANG99_PRIV_REC_NEXT(8774, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8774(choice, ...)  METALANG99_PRIV_REC_NEXT(8775, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8775(choice, ...)  METALANG99_PRIV_REC_NEXT(8776, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8776(choice, ...)  METALANG99_PRIV_REC_NEXT(8777, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8777(choice, ...)  METALANG99_PRIV_REC_NEXT(8778, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8778(choice, ...)  METALANG99_PRIV_REC_NEXT(8779, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8779(choice, ...)  METALANG99_PRIV_REC_NEXT(8780, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8780(choice, ...)  METALANG99_PRIV_REC_NEXT(8781, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8781(choice, ...)  METALANG99_PRIV_REC_NEXT(8782, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8782(choice, ...)  METALANG99_PRIV_REC_NEXT(8783, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8783(choice, ...)  METALANG99_PRIV_REC_NEXT(8784, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8784(choice, ...)  METALANG99_PRIV_REC_NEXT(8785, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8785(choice, ...)  METALANG99_PRIV_REC_NEXT(8786, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8786(choice, ...)  METALANG99_PRIV_REC_NEXT(8787, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8787(choice, ...)  METALANG99_PRIV_REC_NEXT(8788, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8788(choice, ...)  METALANG99_PRIV_REC_NEXT(8789, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8789(choice, ...)  METALANG99_PRIV_REC_NEXT(8790, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8790(choice, ...)  METALANG99_PRIV_REC_NEXT(8791, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8791(choice, ...)  METALANG99_PRIV_REC_NEXT(8792, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8792(choice, ...)  METALANG99_PRIV_REC_NEXT(8793, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8793(choice, ...)  METALANG99_PRIV_REC_NEXT(8794, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8794(choice, ...)  METALANG99_PRIV_REC_NEXT(8795, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8795(choice, ...)  METALANG99_PRIV_REC_NEXT(8796, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8796(choice, ...)  METALANG99_PRIV_REC_NEXT(8797, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8797(choice, ...)  METALANG99_PRIV_REC_NEXT(8798, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8798(choice, ...)  METALANG99_PRIV_REC_NEXT(8799, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8799(choice, ...)  METALANG99_PRIV_REC_NEXT(8800, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8800(choice, ...)  METALANG99_PRIV_REC_NEXT(8801, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8801(choice, ...)  METALANG99_PRIV_REC_NEXT(8802, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8802(choice, ...)  METALANG99_PRIV_REC_NEXT(8803, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8803(choice, ...)  METALANG99_PRIV_REC_NEXT(8804, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8804(choice, ...)  METALANG99_PRIV_REC_NEXT(8805, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8805(choice, ...)  METALANG99_PRIV_REC_NEXT(8806, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8806(choice, ...)  METALANG99_PRIV_REC_NEXT(8807, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8807(choice, ...)  METALANG99_PRIV_REC_NEXT(8808, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8808(choice, ...)  METALANG99_PRIV_REC_NEXT(8809, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8809(choice, ...)  METALANG99_PRIV_REC_NEXT(8810, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8810(choice, ...)  METALANG99_PRIV_REC_NEXT(8811, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8811(choice, ...)  METALANG99_PRIV_REC_NEXT(8812, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8812(choice, ...)  METALANG99_PRIV_REC_NEXT(8813, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8813(choice, ...)  METALANG99_PRIV_REC_NEXT(8814, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8814(choice, ...)  METALANG99_PRIV_REC_NEXT(8815, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8815(choice, ...)  METALANG99_PRIV_REC_NEXT(8816, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8816(choice, ...)  METALANG99_PRIV_REC_NEXT(8817, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8817(choice, ...)  METALANG99_PRIV_REC_NEXT(8818, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8818(choice, ...)  METALANG99_PRIV_REC_NEXT(8819, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8819(choice, ...)  METALANG99_PRIV_REC_NEXT(8820, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8820(choice, ...)  METALANG99_PRIV_REC_NEXT(8821, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8821(choice, ...)  METALANG99_PRIV_REC_NEXT(8822, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8822(choice, ...)  METALANG99_PRIV_REC_NEXT(8823, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8823(choice, ...)  METALANG99_PRIV_REC_NEXT(8824, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8824(choice, ...)  METALANG99_PRIV_REC_NEXT(8825, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8825(choice, ...)  METALANG99_PRIV_REC_NEXT(8826, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8826(choice, ...)  METALANG99_PRIV_REC_NEXT(8827, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8827(choice, ...)  METALANG99_PRIV_REC_NEXT(8828, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8828(choice, ...)  METALANG99_PRIV_REC_NEXT(8829, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8829(choice, ...)  METALANG99_PRIV_REC_NEXT(8830, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8830(choice, ...)  METALANG99_PRIV_REC_NEXT(8831, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8831(choice, ...)  METALANG99_PRIV_REC_NEXT(8832, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8832(choice, ...)  METALANG99_PRIV_REC_NEXT(8833, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8833(choice, ...)  METALANG99_PRIV_REC_NEXT(8834, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8834(choice, ...)  METALANG99_PRIV_REC_NEXT(8835, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8835(choice, ...)  METALANG99_PRIV_REC_NEXT(8836, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8836(choice, ...)  METALANG99_PRIV_REC_NEXT(8837, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8837(choice, ...)  METALANG99_PRIV_REC_NEXT(8838, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8838(choice, ...)  METALANG99_PRIV_REC_NEXT(8839, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8839(choice, ...)  METALANG99_PRIV_REC_NEXT(8840, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8840(choice, ...)  METALANG99_PRIV_REC_NEXT(8841, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8841(choice, ...)  METALANG99_PRIV_REC_NEXT(8842, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8842(choice, ...)  METALANG99_PRIV_REC_NEXT(8843, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8843(choice, ...)  METALANG99_PRIV_REC_NEXT(8844, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8844(choice, ...)  METALANG99_PRIV_REC_NEXT(8845, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8845(choice, ...)  METALANG99_PRIV_REC_NEXT(8846, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8846(choice, ...)  METALANG99_PRIV_REC_NEXT(8847, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8847(choice, ...)  METALANG99_PRIV_REC_NEXT(8848, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8848(choice, ...)  METALANG99_PRIV_REC_NEXT(8849, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8849(choice, ...)  METALANG99_PRIV_REC_NEXT(8850, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8850(choice, ...)  METALANG99_PRIV_REC_NEXT(8851, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8851(choice, ...)  METALANG99_PRIV_REC_NEXT(8852, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8852(choice, ...)  METALANG99_PRIV_REC_NEXT(8853, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8853(choice, ...)  METALANG99_PRIV_REC_NEXT(8854, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8854(choice, ...)  METALANG99_PRIV_REC_NEXT(8855, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8855(choice, ...)  METALANG99_PRIV_REC_NEXT(8856, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8856(choice, ...)  METALANG99_PRIV_REC_NEXT(8857, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8857(choice, ...)  METALANG99_PRIV_REC_NEXT(8858, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8858(choice, ...)  METALANG99_PRIV_REC_NEXT(8859, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8859(choice, ...)  METALANG99_PRIV_REC_NEXT(8860, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8860(choice, ...)  METALANG99_PRIV_REC_NEXT(8861, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8861(choice, ...)  METALANG99_PRIV_REC_NEXT(8862, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8862(choice, ...)  METALANG99_PRIV_REC_NEXT(8863, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8863(choice, ...)  METALANG99_PRIV_REC_NEXT(8864, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8864(choice, ...)  METALANG99_PRIV_REC_NEXT(8865, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8865(choice, ...)  METALANG99_PRIV_REC_NEXT(8866, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8866(choice, ...)  METALANG99_PRIV_REC_NEXT(8867, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8867(choice, ...)  METALANG99_PRIV_REC_NEXT(8868, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8868(choice, ...)  METALANG99_PRIV_REC_NEXT(8869, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8869(choice, ...)  METALANG99_PRIV_REC_NEXT(8870, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8870(choice, ...)  METALANG99_PRIV_REC_NEXT(8871, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8871(choice, ...)  METALANG99_PRIV_REC_NEXT(8872, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8872(choice, ...)  METALANG99_PRIV_REC_NEXT(8873, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8873(choice, ...)  METALANG99_PRIV_REC_NEXT(8874, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8874(choice, ...)  METALANG99_PRIV_REC_NEXT(8875, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8875(choice, ...)  METALANG99_PRIV_REC_NEXT(8876, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8876(choice, ...)  METALANG99_PRIV_REC_NEXT(8877, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8877(choice, ...)  METALANG99_PRIV_REC_NEXT(8878, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8878(choice, ...)  METALANG99_PRIV_REC_NEXT(8879, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8879(choice, ...)  METALANG99_PRIV_REC_NEXT(8880, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8880(choice, ...)  METALANG99_PRIV_REC_NEXT(8881, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8881(choice, ...)  METALANG99_PRIV_REC_NEXT(8882, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8882(choice, ...)  METALANG99_PRIV_REC_NEXT(8883, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8883(choice, ...)  METALANG99_PRIV_REC_NEXT(8884, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8884(choice, ...)  METALANG99_PRIV_REC_NEXT(8885, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8885(choice, ...)  METALANG99_PRIV_REC_NEXT(8886, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8886(choice, ...)  METALANG99_PRIV_REC_NEXT(8887, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8887(choice, ...)  METALANG99_PRIV_REC_NEXT(8888, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8888(choice, ...)  METALANG99_PRIV_REC_NEXT(8889, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8889(choice, ...)  METALANG99_PRIV_REC_NEXT(8890, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8890(choice, ...)  METALANG99_PRIV_REC_NEXT(8891, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8891(choice, ...)  METALANG99_PRIV_REC_NEXT(8892, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8892(choice, ...)  METALANG99_PRIV_REC_NEXT(8893, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8893(choice, ...)  METALANG99_PRIV_REC_NEXT(8894, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8894(choice, ...)  METALANG99_PRIV_REC_NEXT(8895, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8895(choice, ...)  METALANG99_PRIV_REC_NEXT(8896, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8896(choice, ...)  METALANG99_PRIV_REC_NEXT(8897, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8897(choice, ...)  METALANG99_PRIV_REC_NEXT(8898, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8898(choice, ...)  METALANG99_PRIV_REC_NEXT(8899, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8899(choice, ...)  METALANG99_PRIV_REC_NEXT(8900, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8900(choice, ...)  METALANG99_PRIV_REC_NEXT(8901, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8901(choice, ...)  METALANG99_PRIV_REC_NEXT(8902, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8902(choice, ...)  METALANG99_PRIV_REC_NEXT(8903, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8903(choice, ...)  METALANG99_PRIV_REC_NEXT(8904, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8904(choice, ...)  METALANG99_PRIV_REC_NEXT(8905, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8905(choice, ...)  METALANG99_PRIV_REC_NEXT(8906, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8906(choice, ...)  METALANG99_PRIV_REC_NEXT(8907, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8907(choice, ...)  METALANG99_PRIV_REC_NEXT(8908, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8908(choice, ...)  METALANG99_PRIV_REC_NEXT(8909, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8909(choice, ...)  METALANG99_PRIV_REC_NEXT(8910, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8910(choice, ...)  METALANG99_PRIV_REC_NEXT(8911, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8911(choice, ...)  METALANG99_PRIV_REC_NEXT(8912, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8912(choice, ...)  METALANG99_PRIV_REC_NEXT(8913, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8913(choice, ...)  METALANG99_PRIV_REC_NEXT(8914, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8914(choice, ...)  METALANG99_PRIV_REC_NEXT(8915, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8915(choice, ...)  METALANG99_PRIV_REC_NEXT(8916, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8916(choice, ...)  METALANG99_PRIV_REC_NEXT(8917, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8917(choice, ...)  METALANG99_PRIV_REC_NEXT(8918, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8918(choice, ...)  METALANG99_PRIV_REC_NEXT(8919, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8919(choice, ...)  METALANG99_PRIV_REC_NEXT(8920, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8920(choice, ...)  METALANG99_PRIV_REC_NEXT(8921, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8921(choice, ...)  METALANG99_PRIV_REC_NEXT(8922, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8922(choice, ...)  METALANG99_PRIV_REC_NEXT(8923, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8923(choice, ...)  METALANG99_PRIV_REC_NEXT(8924, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8924(choice, ...)  METALANG99_PRIV_REC_NEXT(8925, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8925(choice, ...)  METALANG99_PRIV_REC_NEXT(8926, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8926(choice, ...)  METALANG99_PRIV_REC_NEXT(8927, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8927(choice, ...)  METALANG99_PRIV_REC_NEXT(8928, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8928(choice, ...)  METALANG99_PRIV_REC_NEXT(8929, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8929(choice, ...)  METALANG99_PRIV_REC_NEXT(8930, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8930(choice, ...)  METALANG99_PRIV_REC_NEXT(8931, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8931(choice, ...)  METALANG99_PRIV_REC_NEXT(8932, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8932(choice, ...)  METALANG99_PRIV_REC_NEXT(8933, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8933(choice, ...)  METALANG99_PRIV_REC_NEXT(8934, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8934(choice, ...)  METALANG99_PRIV_REC_NEXT(8935, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8935(choice, ...)  METALANG99_PRIV_REC_NEXT(8936, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8936(choice, ...)  METALANG99_PRIV_REC_NEXT(8937, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8937(choice, ...)  METALANG99_PRIV_REC_NEXT(8938, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8938(choice, ...)  METALANG99_PRIV_REC_NEXT(8939, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8939(choice, ...)  METALANG99_PRIV_REC_NEXT(8940, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8940(choice, ...)  METALANG99_PRIV_REC_NEXT(8941, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8941(choice, ...)  METALANG99_PRIV_REC_NEXT(8942, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8942(choice, ...)  METALANG99_PRIV_REC_NEXT(8943, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8943(choice, ...)  METALANG99_PRIV_REC_NEXT(8944, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8944(choice, ...)  METALANG99_PRIV_REC_NEXT(8945, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8945(choice, ...)  METALANG99_PRIV_REC_NEXT(8946, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8946(choice, ...)  METALANG99_PRIV_REC_NEXT(8947, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8947(choice, ...)  METALANG99_PRIV_REC_NEXT(8948, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8948(choice, ...)  METALANG99_PRIV_REC_NEXT(8949, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8949(choice, ...)  METALANG99_PRIV_REC_NEXT(8950, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8950(choice, ...)  METALANG99_PRIV_REC_NEXT(8951, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8951(choice, ...)  METALANG99_PRIV_REC_NEXT(8952, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8952(choice, ...)  METALANG99_PRIV_REC_NEXT(8953, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8953(choice, ...)  METALANG99_PRIV_REC_NEXT(8954, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8954(choice, ...)  METALANG99_PRIV_REC_NEXT(8955, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8955(choice, ...)  METALANG99_PRIV_REC_NEXT(8956, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8956(choice, ...)  METALANG99_PRIV_REC_NEXT(8957, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8957(choice, ...)  METALANG99_PRIV_REC_NEXT(8958, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8958(choice, ...)  METALANG99_PRIV_REC_NEXT(8959, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8959(choice, ...)  METALANG99_PRIV_REC_NEXT(8960, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8960(choice, ...)  METALANG99_PRIV_REC_NEXT(8961, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8961(choice, ...)  METALANG99_PRIV_REC_NEXT(8962, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8962(choice, ...)  METALANG99_PRIV_REC_NEXT(8963, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8963(choice, ...)  METALANG99_PRIV_REC_NEXT(8964, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8964(choice, ...)  METALANG99_PRIV_REC_NEXT(8965, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8965(choice, ...)  METALANG99_PRIV_REC_NEXT(8966, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8966(choice, ...)  METALANG99_PRIV_REC_NEXT(8967, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8967(choice, ...)  METALANG99_PRIV_REC_NEXT(8968, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8968(choice, ...)  METALANG99_PRIV_REC_NEXT(8969, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8969(choice, ...)  METALANG99_PRIV_REC_NEXT(8970, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8970(choice, ...)  METALANG99_PRIV_REC_NEXT(8971, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8971(choice, ...)  METALANG99_PRIV_REC_NEXT(8972, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8972(choice, ...)  METALANG99_PRIV_REC_NEXT(8973, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8973(choice, ...)  METALANG99_PRIV_REC_NEXT(8974, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8974(choice, ...)  METALANG99_PRIV_REC_NEXT(8975, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8975(choice, ...)  METALANG99_PRIV_REC_NEXT(8976, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8976(choice, ...)  METALANG99_PRIV_REC_NEXT(8977, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8977(choice, ...)  METALANG99_PRIV_REC_NEXT(8978, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8978(choice, ...)  METALANG99_PRIV_REC_NEXT(8979, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8979(choice, ...)  METALANG99_PRIV_REC_NEXT(8980, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8980(choice, ...)  METALANG99_PRIV_REC_NEXT(8981, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8981(choice, ...)  METALANG99_PRIV_REC_NEXT(8982, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8982(choice, ...)  METALANG99_PRIV_REC_NEXT(8983, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8983(choice, ...)  METALANG99_PRIV_REC_NEXT(8984, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8984(choice, ...)  METALANG99_PRIV_REC_NEXT(8985, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8985(choice, ...)  METALANG99_PRIV_REC_NEXT(8986, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8986(choice, ...)  METALANG99_PRIV_REC_NEXT(8987, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8987(choice, ...)  METALANG99_PRIV_REC_NEXT(8988, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8988(choice, ...)  METALANG99_PRIV_REC_NEXT(8989, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8989(choice, ...)  METALANG99_PRIV_REC_NEXT(8990, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8990(choice, ...)  METALANG99_PRIV_REC_NEXT(8991, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8991(choice, ...)  METALANG99_PRIV_REC_NEXT(8992, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8992(choice, ...)  METALANG99_PRIV_REC_NEXT(8993, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8993(choice, ...)  METALANG99_PRIV_REC_NEXT(8994, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8994(choice, ...)  METALANG99_PRIV_REC_NEXT(8995, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8995(choice, ...)  METALANG99_PRIV_REC_NEXT(8996, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8996(choice, ...)  METALANG99_PRIV_REC_NEXT(8997, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8997(choice, ...)  METALANG99_PRIV_REC_NEXT(8998, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8998(choice, ...)  METALANG99_PRIV_REC_NEXT(8999, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_8999(choice, ...)  METALANG99_PRIV_REC_NEXT(9000, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9000(choice, ...)  METALANG99_PRIV_REC_NEXT(9001, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9001(choice, ...)  METALANG99_PRIV_REC_NEXT(9002, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9002(choice, ...)  METALANG99_PRIV_REC_NEXT(9003, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9003(choice, ...)  METALANG99_PRIV_REC_NEXT(9004, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9004(choice, ...)  METALANG99_PRIV_REC_NEXT(9005, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9005(choice, ...)  METALANG99_PRIV_REC_NEXT(9006, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9006(choice, ...)  METALANG99_PRIV_REC_NEXT(9007, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9007(choice, ...)  METALANG99_PRIV_REC_NEXT(9008, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9008(choice, ...)  METALANG99_PRIV_REC_NEXT(9009, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9009(choice, ...)  METALANG99_PRIV_REC_NEXT(9010, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9010(choice, ...)  METALANG99_PRIV_REC_NEXT(9011, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9011(choice, ...)  METALANG99_PRIV_REC_NEXT(9012, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9012(choice, ...)  METALANG99_PRIV_REC_NEXT(9013, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9013(choice, ...)  METALANG99_PRIV_REC_NEXT(9014, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9014(choice, ...)  METALANG99_PRIV_REC_NEXT(9015, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9015(choice, ...)  METALANG99_PRIV_REC_NEXT(9016, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9016(choice, ...)  METALANG99_PRIV_REC_NEXT(9017, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9017(choice, ...)  METALANG99_PRIV_REC_NEXT(9018, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9018(choice, ...)  METALANG99_PRIV_REC_NEXT(9019, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9019(choice, ...)  METALANG99_PRIV_REC_NEXT(9020, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9020(choice, ...)  METALANG99_PRIV_REC_NEXT(9021, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9021(choice, ...)  METALANG99_PRIV_REC_NEXT(9022, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9022(choice, ...)  METALANG99_PRIV_REC_NEXT(9023, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9023(choice, ...)  METALANG99_PRIV_REC_NEXT(9024, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9024(choice, ...)  METALANG99_PRIV_REC_NEXT(9025, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9025(choice, ...)  METALANG99_PRIV_REC_NEXT(9026, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9026(choice, ...)  METALANG99_PRIV_REC_NEXT(9027, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9027(choice, ...)  METALANG99_PRIV_REC_NEXT(9028, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9028(choice, ...)  METALANG99_PRIV_REC_NEXT(9029, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9029(choice, ...)  METALANG99_PRIV_REC_NEXT(9030, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9030(choice, ...)  METALANG99_PRIV_REC_NEXT(9031, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9031(choice, ...)  METALANG99_PRIV_REC_NEXT(9032, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9032(choice, ...)  METALANG99_PRIV_REC_NEXT(9033, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9033(choice, ...)  METALANG99_PRIV_REC_NEXT(9034, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9034(choice, ...)  METALANG99_PRIV_REC_NEXT(9035, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9035(choice, ...)  METALANG99_PRIV_REC_NEXT(9036, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9036(choice, ...)  METALANG99_PRIV_REC_NEXT(9037, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9037(choice, ...)  METALANG99_PRIV_REC_NEXT(9038, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9038(choice, ...)  METALANG99_PRIV_REC_NEXT(9039, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9039(choice, ...)  METALANG99_PRIV_REC_NEXT(9040, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9040(choice, ...)  METALANG99_PRIV_REC_NEXT(9041, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9041(choice, ...)  METALANG99_PRIV_REC_NEXT(9042, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9042(choice, ...)  METALANG99_PRIV_REC_NEXT(9043, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9043(choice, ...)  METALANG99_PRIV_REC_NEXT(9044, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9044(choice, ...)  METALANG99_PRIV_REC_NEXT(9045, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9045(choice, ...)  METALANG99_PRIV_REC_NEXT(9046, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9046(choice, ...)  METALANG99_PRIV_REC_NEXT(9047, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9047(choice, ...)  METALANG99_PRIV_REC_NEXT(9048, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9048(choice, ...)  METALANG99_PRIV_REC_NEXT(9049, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9049(choice, ...)  METALANG99_PRIV_REC_NEXT(9050, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9050(choice, ...)  METALANG99_PRIV_REC_NEXT(9051, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9051(choice, ...)  METALANG99_PRIV_REC_NEXT(9052, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9052(choice, ...)  METALANG99_PRIV_REC_NEXT(9053, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9053(choice, ...)  METALANG99_PRIV_REC_NEXT(9054, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9054(choice, ...)  METALANG99_PRIV_REC_NEXT(9055, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9055(choice, ...)  METALANG99_PRIV_REC_NEXT(9056, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9056(choice, ...)  METALANG99_PRIV_REC_NEXT(9057, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9057(choice, ...)  METALANG99_PRIV_REC_NEXT(9058, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9058(choice, ...)  METALANG99_PRIV_REC_NEXT(9059, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9059(choice, ...)  METALANG99_PRIV_REC_NEXT(9060, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9060(choice, ...)  METALANG99_PRIV_REC_NEXT(9061, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9061(choice, ...)  METALANG99_PRIV_REC_NEXT(9062, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9062(choice, ...)  METALANG99_PRIV_REC_NEXT(9063, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9063(choice, ...)  METALANG99_PRIV_REC_NEXT(9064, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9064(choice, ...)  METALANG99_PRIV_REC_NEXT(9065, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9065(choice, ...)  METALANG99_PRIV_REC_NEXT(9066, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9066(choice, ...)  METALANG99_PRIV_REC_NEXT(9067, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9067(choice, ...)  METALANG99_PRIV_REC_NEXT(9068, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9068(choice, ...)  METALANG99_PRIV_REC_NEXT(9069, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9069(choice, ...)  METALANG99_PRIV_REC_NEXT(9070, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9070(choice, ...)  METALANG99_PRIV_REC_NEXT(9071, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9071(choice, ...)  METALANG99_PRIV_REC_NEXT(9072, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9072(choice, ...)  METALANG99_PRIV_REC_NEXT(9073, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9073(choice, ...)  METALANG99_PRIV_REC_NEXT(9074, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9074(choice, ...)  METALANG99_PRIV_REC_NEXT(9075, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9075(choice, ...)  METALANG99_PRIV_REC_NEXT(9076, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9076(choice, ...)  METALANG99_PRIV_REC_NEXT(9077, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9077(choice, ...)  METALANG99_PRIV_REC_NEXT(9078, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9078(choice, ...)  METALANG99_PRIV_REC_NEXT(9079, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9079(choice, ...)  METALANG99_PRIV_REC_NEXT(9080, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9080(choice, ...)  METALANG99_PRIV_REC_NEXT(9081, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9081(choice, ...)  METALANG99_PRIV_REC_NEXT(9082, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9082(choice, ...)  METALANG99_PRIV_REC_NEXT(9083, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9083(choice, ...)  METALANG99_PRIV_REC_NEXT(9084, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9084(choice, ...)  METALANG99_PRIV_REC_NEXT(9085, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9085(choice, ...)  METALANG99_PRIV_REC_NEXT(9086, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9086(choice, ...)  METALANG99_PRIV_REC_NEXT(9087, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9087(choice, ...)  METALANG99_PRIV_REC_NEXT(9088, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9088(choice, ...)  METALANG99_PRIV_REC_NEXT(9089, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9089(choice, ...)  METALANG99_PRIV_REC_NEXT(9090, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9090(choice, ...)  METALANG99_PRIV_REC_NEXT(9091, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9091(choice, ...)  METALANG99_PRIV_REC_NEXT(9092, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9092(choice, ...)  METALANG99_PRIV_REC_NEXT(9093, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9093(choice, ...)  METALANG99_PRIV_REC_NEXT(9094, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9094(choice, ...)  METALANG99_PRIV_REC_NEXT(9095, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9095(choice, ...)  METALANG99_PRIV_REC_NEXT(9096, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9096(choice, ...)  METALANG99_PRIV_REC_NEXT(9097, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9097(choice, ...)  METALANG99_PRIV_REC_NEXT(9098, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9098(choice, ...)  METALANG99_PRIV_REC_NEXT(9099, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9099(choice, ...)  METALANG99_PRIV_REC_NEXT(9100, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9100(choice, ...)  METALANG99_PRIV_REC_NEXT(9101, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9101(choice, ...)  METALANG99_PRIV_REC_NEXT(9102, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9102(choice, ...)  METALANG99_PRIV_REC_NEXT(9103, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9103(choice, ...)  METALANG99_PRIV_REC_NEXT(9104, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9104(choice, ...)  METALANG99_PRIV_REC_NEXT(9105, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9105(choice, ...)  METALANG99_PRIV_REC_NEXT(9106, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9106(choice, ...)  METALANG99_PRIV_REC_NEXT(9107, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9107(choice, ...)  METALANG99_PRIV_REC_NEXT(9108, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9108(choice, ...)  METALANG99_PRIV_REC_NEXT(9109, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9109(choice, ...)  METALANG99_PRIV_REC_NEXT(9110, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9110(choice, ...)  METALANG99_PRIV_REC_NEXT(9111, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9111(choice, ...)  METALANG99_PRIV_REC_NEXT(9112, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9112(choice, ...)  METALANG99_PRIV_REC_NEXT(9113, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9113(choice, ...)  METALANG99_PRIV_REC_NEXT(9114, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9114(choice, ...)  METALANG99_PRIV_REC_NEXT(9115, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9115(choice, ...)  METALANG99_PRIV_REC_NEXT(9116, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9116(choice, ...)  METALANG99_PRIV_REC_NEXT(9117, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9117(choice, ...)  METALANG99_PRIV_REC_NEXT(9118, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9118(choice, ...)  METALANG99_PRIV_REC_NEXT(9119, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9119(choice, ...)  METALANG99_PRIV_REC_NEXT(9120, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9120(choice, ...)  METALANG99_PRIV_REC_NEXT(9121, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9121(choice, ...)  METALANG99_PRIV_REC_NEXT(9122, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9122(choice, ...)  METALANG99_PRIV_REC_NEXT(9123, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9123(choice, ...)  METALANG99_PRIV_REC_NEXT(9124, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9124(choice, ...)  METALANG99_PRIV_REC_NEXT(9125, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9125(choice, ...)  METALANG99_PRIV_REC_NEXT(9126, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9126(choice, ...)  METALANG99_PRIV_REC_NEXT(9127, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9127(choice, ...)  METALANG99_PRIV_REC_NEXT(9128, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9128(choice, ...)  METALANG99_PRIV_REC_NEXT(9129, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9129(choice, ...)  METALANG99_PRIV_REC_NEXT(9130, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9130(choice, ...)  METALANG99_PRIV_REC_NEXT(9131, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9131(choice, ...)  METALANG99_PRIV_REC_NEXT(9132, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9132(choice, ...)  METALANG99_PRIV_REC_NEXT(9133, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9133(choice, ...)  METALANG99_PRIV_REC_NEXT(9134, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9134(choice, ...)  METALANG99_PRIV_REC_NEXT(9135, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9135(choice, ...)  METALANG99_PRIV_REC_NEXT(9136, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9136(choice, ...)  METALANG99_PRIV_REC_NEXT(9137, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9137(choice, ...)  METALANG99_PRIV_REC_NEXT(9138, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9138(choice, ...)  METALANG99_PRIV_REC_NEXT(9139, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9139(choice, ...)  METALANG99_PRIV_REC_NEXT(9140, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9140(choice, ...)  METALANG99_PRIV_REC_NEXT(9141, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9141(choice, ...)  METALANG99_PRIV_REC_NEXT(9142, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9142(choice, ...)  METALANG99_PRIV_REC_NEXT(9143, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9143(choice, ...)  METALANG99_PRIV_REC_NEXT(9144, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9144(choice, ...)  METALANG99_PRIV_REC_NEXT(9145, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9145(choice, ...)  METALANG99_PRIV_REC_NEXT(9146, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9146(choice, ...)  METALANG99_PRIV_REC_NEXT(9147, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9147(choice, ...)  METALANG99_PRIV_REC_NEXT(9148, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9148(choice, ...)  METALANG99_PRIV_REC_NEXT(9149, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9149(choice, ...)  METALANG99_PRIV_REC_NEXT(9150, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9150(choice, ...)  METALANG99_PRIV_REC_NEXT(9151, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9151(choice, ...)  METALANG99_PRIV_REC_NEXT(9152, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9152(choice, ...)  METALANG99_PRIV_REC_NEXT(9153, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9153(choice, ...)  METALANG99_PRIV_REC_NEXT(9154, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9154(choice, ...)  METALANG99_PRIV_REC_NEXT(9155, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9155(choice, ...)  METALANG99_PRIV_REC_NEXT(9156, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9156(choice, ...)  METALANG99_PRIV_REC_NEXT(9157, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9157(choice, ...)  METALANG99_PRIV_REC_NEXT(9158, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9158(choice, ...)  METALANG99_PRIV_REC_NEXT(9159, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9159(choice, ...)  METALANG99_PRIV_REC_NEXT(9160, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9160(choice, ...)  METALANG99_PRIV_REC_NEXT(9161, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9161(choice, ...)  METALANG99_PRIV_REC_NEXT(9162, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9162(choice, ...)  METALANG99_PRIV_REC_NEXT(9163, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9163(choice, ...)  METALANG99_PRIV_REC_NEXT(9164, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9164(choice, ...)  METALANG99_PRIV_REC_NEXT(9165, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9165(choice, ...)  METALANG99_PRIV_REC_NEXT(9166, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9166(choice, ...)  METALANG99_PRIV_REC_NEXT(9167, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9167(choice, ...)  METALANG99_PRIV_REC_NEXT(9168, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9168(choice, ...)  METALANG99_PRIV_REC_NEXT(9169, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9169(choice, ...)  METALANG99_PRIV_REC_NEXT(9170, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9170(choice, ...)  METALANG99_PRIV_REC_NEXT(9171, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9171(choice, ...)  METALANG99_PRIV_REC_NEXT(9172, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9172(choice, ...)  METALANG99_PRIV_REC_NEXT(9173, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9173(choice, ...)  METALANG99_PRIV_REC_NEXT(9174, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9174(choice, ...)  METALANG99_PRIV_REC_NEXT(9175, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9175(choice, ...)  METALANG99_PRIV_REC_NEXT(9176, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9176(choice, ...)  METALANG99_PRIV_REC_NEXT(9177, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9177(choice, ...)  METALANG99_PRIV_REC_NEXT(9178, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9178(choice, ...)  METALANG99_PRIV_REC_NEXT(9179, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9179(choice, ...)  METALANG99_PRIV_REC_NEXT(9180, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9180(choice, ...)  METALANG99_PRIV_REC_NEXT(9181, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9181(choice, ...)  METALANG99_PRIV_REC_NEXT(9182, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9182(choice, ...)  METALANG99_PRIV_REC_NEXT(9183, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9183(choice, ...)  METALANG99_PRIV_REC_NEXT(9184, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9184(choice, ...)  METALANG99_PRIV_REC_NEXT(9185, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9185(choice, ...)  METALANG99_PRIV_REC_NEXT(9186, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9186(choice, ...)  METALANG99_PRIV_REC_NEXT(9187, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9187(choice, ...)  METALANG99_PRIV_REC_NEXT(9188, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9188(choice, ...)  METALANG99_PRIV_REC_NEXT(9189, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9189(choice, ...)  METALANG99_PRIV_REC_NEXT(9190, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9190(choice, ...)  METALANG99_PRIV_REC_NEXT(9191, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9191(choice, ...)  METALANG99_PRIV_REC_NEXT(9192, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9192(choice, ...)  METALANG99_PRIV_REC_NEXT(9193, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9193(choice, ...)  METALANG99_PRIV_REC_NEXT(9194, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9194(choice, ...)  METALANG99_PRIV_REC_NEXT(9195, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9195(choice, ...)  METALANG99_PRIV_REC_NEXT(9196, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9196(choice, ...)  METALANG99_PRIV_REC_NEXT(9197, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9197(choice, ...)  METALANG99_PRIV_REC_NEXT(9198, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9198(choice, ...)  METALANG99_PRIV_REC_NEXT(9199, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9199(choice, ...)  METALANG99_PRIV_REC_NEXT(9200, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9200(choice, ...)  METALANG99_PRIV_REC_NEXT(9201, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9201(choice, ...)  METALANG99_PRIV_REC_NEXT(9202, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9202(choice, ...)  METALANG99_PRIV_REC_NEXT(9203, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9203(choice, ...)  METALANG99_PRIV_REC_NEXT(9204, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9204(choice, ...)  METALANG99_PRIV_REC_NEXT(9205, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9205(choice, ...)  METALANG99_PRIV_REC_NEXT(9206, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9206(choice, ...)  METALANG99_PRIV_REC_NEXT(9207, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9207(choice, ...)  METALANG99_PRIV_REC_NEXT(9208, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9208(choice, ...)  METALANG99_PRIV_REC_NEXT(9209, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9209(choice, ...)  METALANG99_PRIV_REC_NEXT(9210, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9210(choice, ...)  METALANG99_PRIV_REC_NEXT(9211, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9211(choice, ...)  METALANG99_PRIV_REC_NEXT(9212, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9212(choice, ...)  METALANG99_PRIV_REC_NEXT(9213, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9213(choice, ...)  METALANG99_PRIV_REC_NEXT(9214, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9214(choice, ...)  METALANG99_PRIV_REC_NEXT(9215, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9215(choice, ...)  METALANG99_PRIV_REC_NEXT(9216, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9216(choice, ...)  METALANG99_PRIV_REC_NEXT(9217, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9217(choice, ...)  METALANG99_PRIV_REC_NEXT(9218, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9218(choice, ...)  METALANG99_PRIV_REC_NEXT(9219, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9219(choice, ...)  METALANG99_PRIV_REC_NEXT(9220, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9220(choice, ...)  METALANG99_PRIV_REC_NEXT(9221, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9221(choice, ...)  METALANG99_PRIV_REC_NEXT(9222, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9222(choice, ...)  METALANG99_PRIV_REC_NEXT(9223, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9223(choice, ...)  METALANG99_PRIV_REC_NEXT(9224, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9224(choice, ...)  METALANG99_PRIV_REC_NEXT(9225, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9225(choice, ...)  METALANG99_PRIV_REC_NEXT(9226, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9226(choice, ...)  METALANG99_PRIV_REC_NEXT(9227, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9227(choice, ...)  METALANG99_PRIV_REC_NEXT(9228, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9228(choice, ...)  METALANG99_PRIV_REC_NEXT(9229, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9229(choice, ...)  METALANG99_PRIV_REC_NEXT(9230, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9230(choice, ...)  METALANG99_PRIV_REC_NEXT(9231, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9231(choice, ...)  METALANG99_PRIV_REC_NEXT(9232, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9232(choice, ...)  METALANG99_PRIV_REC_NEXT(9233, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9233(choice, ...)  METALANG99_PRIV_REC_NEXT(9234, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9234(choice, ...)  METALANG99_PRIV_REC_NEXT(9235, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9235(choice, ...)  METALANG99_PRIV_REC_NEXT(9236, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9236(choice, ...)  METALANG99_PRIV_REC_NEXT(9237, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9237(choice, ...)  METALANG99_PRIV_REC_NEXT(9238, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9238(choice, ...)  METALANG99_PRIV_REC_NEXT(9239, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9239(choice, ...)  METALANG99_PRIV_REC_NEXT(9240, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9240(choice, ...)  METALANG99_PRIV_REC_NEXT(9241, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9241(choice, ...)  METALANG99_PRIV_REC_NEXT(9242, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9242(choice, ...)  METALANG99_PRIV_REC_NEXT(9243, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9243(choice, ...)  METALANG99_PRIV_REC_NEXT(9244, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9244(choice, ...)  METALANG99_PRIV_REC_NEXT(9245, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9245(choice, ...)  METALANG99_PRIV_REC_NEXT(9246, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9246(choice, ...)  METALANG99_PRIV_REC_NEXT(9247, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9247(choice, ...)  METALANG99_PRIV_REC_NEXT(9248, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9248(choice, ...)  METALANG99_PRIV_REC_NEXT(9249, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9249(choice, ...)  METALANG99_PRIV_REC_NEXT(9250, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9250(choice, ...)  METALANG99_PRIV_REC_NEXT(9251, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9251(choice, ...)  METALANG99_PRIV_REC_NEXT(9252, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9252(choice, ...)  METALANG99_PRIV_REC_NEXT(9253, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9253(choice, ...)  METALANG99_PRIV_REC_NEXT(9254, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9254(choice, ...)  METALANG99_PRIV_REC_NEXT(9255, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9255(choice, ...)  METALANG99_PRIV_REC_NEXT(9256, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9256(choice, ...)  METALANG99_PRIV_REC_NEXT(9257, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9257(choice, ...)  METALANG99_PRIV_REC_NEXT(9258, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9258(choice, ...)  METALANG99_PRIV_REC_NEXT(9259, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9259(choice, ...)  METALANG99_PRIV_REC_NEXT(9260, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9260(choice, ...)  METALANG99_PRIV_REC_NEXT(9261, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9261(choice, ...)  METALANG99_PRIV_REC_NEXT(9262, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9262(choice, ...)  METALANG99_PRIV_REC_NEXT(9263, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9263(choice, ...)  METALANG99_PRIV_REC_NEXT(9264, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9264(choice, ...)  METALANG99_PRIV_REC_NEXT(9265, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9265(choice, ...)  METALANG99_PRIV_REC_NEXT(9266, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9266(choice, ...)  METALANG99_PRIV_REC_NEXT(9267, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9267(choice, ...)  METALANG99_PRIV_REC_NEXT(9268, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9268(choice, ...)  METALANG99_PRIV_REC_NEXT(9269, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9269(choice, ...)  METALANG99_PRIV_REC_NEXT(9270, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9270(choice, ...)  METALANG99_PRIV_REC_NEXT(9271, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9271(choice, ...)  METALANG99_PRIV_REC_NEXT(9272, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9272(choice, ...)  METALANG99_PRIV_REC_NEXT(9273, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9273(choice, ...)  METALANG99_PRIV_REC_NEXT(9274, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9274(choice, ...)  METALANG99_PRIV_REC_NEXT(9275, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9275(choice, ...)  METALANG99_PRIV_REC_NEXT(9276, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9276(choice, ...)  METALANG99_PRIV_REC_NEXT(9277, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9277(choice, ...)  METALANG99_PRIV_REC_NEXT(9278, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9278(choice, ...)  METALANG99_PRIV_REC_NEXT(9279, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9279(choice, ...)  METALANG99_PRIV_REC_NEXT(9280, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9280(choice, ...)  METALANG99_PRIV_REC_NEXT(9281, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9281(choice, ...)  METALANG99_PRIV_REC_NEXT(9282, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9282(choice, ...)  METALANG99_PRIV_REC_NEXT(9283, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9283(choice, ...)  METALANG99_PRIV_REC_NEXT(9284, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9284(choice, ...)  METALANG99_PRIV_REC_NEXT(9285, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9285(choice, ...)  METALANG99_PRIV_REC_NEXT(9286, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9286(choice, ...)  METALANG99_PRIV_REC_NEXT(9287, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9287(choice, ...)  METALANG99_PRIV_REC_NEXT(9288, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9288(choice, ...)  METALANG99_PRIV_REC_NEXT(9289, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9289(choice, ...)  METALANG99_PRIV_REC_NEXT(9290, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9290(choice, ...)  METALANG99_PRIV_REC_NEXT(9291, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9291(choice, ...)  METALANG99_PRIV_REC_NEXT(9292, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9292(choice, ...)  METALANG99_PRIV_REC_NEXT(9293, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9293(choice, ...)  METALANG99_PRIV_REC_NEXT(9294, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9294(choice, ...)  METALANG99_PRIV_REC_NEXT(9295, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9295(choice, ...)  METALANG99_PRIV_REC_NEXT(9296, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9296(choice, ...)  METALANG99_PRIV_REC_NEXT(9297, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9297(choice, ...)  METALANG99_PRIV_REC_NEXT(9298, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9298(choice, ...)  METALANG99_PRIV_REC_NEXT(9299, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9299(choice, ...)  METALANG99_PRIV_REC_NEXT(9300, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9300(choice, ...)  METALANG99_PRIV_REC_NEXT(9301, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9301(choice, ...)  METALANG99_PRIV_REC_NEXT(9302, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9302(choice, ...)  METALANG99_PRIV_REC_NEXT(9303, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9303(choice, ...)  METALANG99_PRIV_REC_NEXT(9304, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9304(choice, ...)  METALANG99_PRIV_REC_NEXT(9305, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9305(choice, ...)  METALANG99_PRIV_REC_NEXT(9306, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9306(choice, ...)  METALANG99_PRIV_REC_NEXT(9307, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9307(choice, ...)  METALANG99_PRIV_REC_NEXT(9308, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9308(choice, ...)  METALANG99_PRIV_REC_NEXT(9309, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9309(choice, ...)  METALANG99_PRIV_REC_NEXT(9310, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9310(choice, ...)  METALANG99_PRIV_REC_NEXT(9311, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9311(choice, ...)  METALANG99_PRIV_REC_NEXT(9312, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9312(choice, ...)  METALANG99_PRIV_REC_NEXT(9313, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9313(choice, ...)  METALANG99_PRIV_REC_NEXT(9314, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9314(choice, ...)  METALANG99_PRIV_REC_NEXT(9315, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9315(choice, ...)  METALANG99_PRIV_REC_NEXT(9316, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9316(choice, ...)  METALANG99_PRIV_REC_NEXT(9317, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9317(choice, ...)  METALANG99_PRIV_REC_NEXT(9318, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9318(choice, ...)  METALANG99_PRIV_REC_NEXT(9319, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9319(choice, ...)  METALANG99_PRIV_REC_NEXT(9320, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9320(choice, ...)  METALANG99_PRIV_REC_NEXT(9321, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9321(choice, ...)  METALANG99_PRIV_REC_NEXT(9322, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9322(choice, ...)  METALANG99_PRIV_REC_NEXT(9323, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9323(choice, ...)  METALANG99_PRIV_REC_NEXT(9324, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9324(choice, ...)  METALANG99_PRIV_REC_NEXT(9325, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9325(choice, ...)  METALANG99_PRIV_REC_NEXT(9326, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9326(choice, ...)  METALANG99_PRIV_REC_NEXT(9327, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9327(choice, ...)  METALANG99_PRIV_REC_NEXT(9328, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9328(choice, ...)  METALANG99_PRIV_REC_NEXT(9329, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9329(choice, ...)  METALANG99_PRIV_REC_NEXT(9330, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9330(choice, ...)  METALANG99_PRIV_REC_NEXT(9331, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9331(choice, ...)  METALANG99_PRIV_REC_NEXT(9332, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9332(choice, ...)  METALANG99_PRIV_REC_NEXT(9333, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9333(choice, ...)  METALANG99_PRIV_REC_NEXT(9334, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9334(choice, ...)  METALANG99_PRIV_REC_NEXT(9335, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9335(choice, ...)  METALANG99_PRIV_REC_NEXT(9336, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9336(choice, ...)  METALANG99_PRIV_REC_NEXT(9337, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9337(choice, ...)  METALANG99_PRIV_REC_NEXT(9338, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9338(choice, ...)  METALANG99_PRIV_REC_NEXT(9339, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9339(choice, ...)  METALANG99_PRIV_REC_NEXT(9340, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9340(choice, ...)  METALANG99_PRIV_REC_NEXT(9341, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9341(choice, ...)  METALANG99_PRIV_REC_NEXT(9342, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9342(choice, ...)  METALANG99_PRIV_REC_NEXT(9343, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9343(choice, ...)  METALANG99_PRIV_REC_NEXT(9344, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9344(choice, ...)  METALANG99_PRIV_REC_NEXT(9345, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9345(choice, ...)  METALANG99_PRIV_REC_NEXT(9346, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9346(choice, ...)  METALANG99_PRIV_REC_NEXT(9347, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9347(choice, ...)  METALANG99_PRIV_REC_NEXT(9348, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9348(choice, ...)  METALANG99_PRIV_REC_NEXT(9349, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9349(choice, ...)  METALANG99_PRIV_REC_NEXT(9350, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9350(choice, ...)  METALANG99_PRIV_REC_NEXT(9351, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9351(choice, ...)  METALANG99_PRIV_REC_NEXT(9352, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9352(choice, ...)  METALANG99_PRIV_REC_NEXT(9353, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9353(choice, ...)  METALANG99_PRIV_REC_NEXT(9354, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9354(choice, ...)  METALANG99_PRIV_REC_NEXT(9355, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9355(choice, ...)  METALANG99_PRIV_REC_NEXT(9356, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9356(choice, ...)  METALANG99_PRIV_REC_NEXT(9357, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9357(choice, ...)  METALANG99_PRIV_REC_NEXT(9358, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9358(choice, ...)  METALANG99_PRIV_REC_NEXT(9359, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9359(choice, ...)  METALANG99_PRIV_REC_NEXT(9360, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9360(choice, ...)  METALANG99_PRIV_REC_NEXT(9361, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9361(choice, ...)  METALANG99_PRIV_REC_NEXT(9362, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9362(choice, ...)  METALANG99_PRIV_REC_NEXT(9363, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9363(choice, ...)  METALANG99_PRIV_REC_NEXT(9364, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9364(choice, ...)  METALANG99_PRIV_REC_NEXT(9365, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9365(choice, ...)  METALANG99_PRIV_REC_NEXT(9366, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9366(choice, ...)  METALANG99_PRIV_REC_NEXT(9367, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9367(choice, ...)  METALANG99_PRIV_REC_NEXT(9368, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9368(choice, ...)  METALANG99_PRIV_REC_NEXT(9369, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9369(choice, ...)  METALANG99_PRIV_REC_NEXT(9370, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9370(choice, ...)  METALANG99_PRIV_REC_NEXT(9371, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9371(choice, ...)  METALANG99_PRIV_REC_NEXT(9372, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9372(choice, ...)  METALANG99_PRIV_REC_NEXT(9373, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9373(choice, ...)  METALANG99_PRIV_REC_NEXT(9374, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9374(choice, ...)  METALANG99_PRIV_REC_NEXT(9375, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9375(choice, ...)  METALANG99_PRIV_REC_NEXT(9376, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9376(choice, ...)  METALANG99_PRIV_REC_NEXT(9377, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9377(choice, ...)  METALANG99_PRIV_REC_NEXT(9378, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9378(choice, ...)  METALANG99_PRIV_REC_NEXT(9379, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9379(choice, ...)  METALANG99_PRIV_REC_NEXT(9380, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9380(choice, ...)  METALANG99_PRIV_REC_NEXT(9381, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9381(choice, ...)  METALANG99_PRIV_REC_NEXT(9382, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9382(choice, ...)  METALANG99_PRIV_REC_NEXT(9383, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9383(choice, ...)  METALANG99_PRIV_REC_NEXT(9384, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9384(choice, ...)  METALANG99_PRIV_REC_NEXT(9385, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9385(choice, ...)  METALANG99_PRIV_REC_NEXT(9386, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9386(choice, ...)  METALANG99_PRIV_REC_NEXT(9387, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9387(choice, ...)  METALANG99_PRIV_REC_NEXT(9388, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9388(choice, ...)  METALANG99_PRIV_REC_NEXT(9389, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9389(choice, ...)  METALANG99_PRIV_REC_NEXT(9390, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9390(choice, ...)  METALANG99_PRIV_REC_NEXT(9391, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9391(choice, ...)  METALANG99_PRIV_REC_NEXT(9392, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9392(choice, ...)  METALANG99_PRIV_REC_NEXT(9393, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9393(choice, ...)  METALANG99_PRIV_REC_NEXT(9394, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9394(choice, ...)  METALANG99_PRIV_REC_NEXT(9395, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9395(choice, ...)  METALANG99_PRIV_REC_NEXT(9396, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9396(choice, ...)  METALANG99_PRIV_REC_NEXT(9397, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9397(choice, ...)  METALANG99_PRIV_REC_NEXT(9398, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9398(choice, ...)  METALANG99_PRIV_REC_NEXT(9399, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9399(choice, ...)  METALANG99_PRIV_REC_NEXT(9400, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9400(choice, ...)  METALANG99_PRIV_REC_NEXT(9401, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9401(choice, ...)  METALANG99_PRIV_REC_NEXT(9402, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9402(choice, ...)  METALANG99_PRIV_REC_NEXT(9403, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9403(choice, ...)  METALANG99_PRIV_REC_NEXT(9404, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9404(choice, ...)  METALANG99_PRIV_REC_NEXT(9405, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9405(choice, ...)  METALANG99_PRIV_REC_NEXT(9406, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9406(choice, ...)  METALANG99_PRIV_REC_NEXT(9407, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9407(choice, ...)  METALANG99_PRIV_REC_NEXT(9408, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9408(choice, ...)  METALANG99_PRIV_REC_NEXT(9409, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9409(choice, ...)  METALANG99_PRIV_REC_NEXT(9410, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9410(choice, ...)  METALANG99_PRIV_REC_NEXT(9411, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9411(choice, ...)  METALANG99_PRIV_REC_NEXT(9412, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9412(choice, ...)  METALANG99_PRIV_REC_NEXT(9413, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9413(choice, ...)  METALANG99_PRIV_REC_NEXT(9414, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9414(choice, ...)  METALANG99_PRIV_REC_NEXT(9415, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9415(choice, ...)  METALANG99_PRIV_REC_NEXT(9416, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9416(choice, ...)  METALANG99_PRIV_REC_NEXT(9417, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9417(choice, ...)  METALANG99_PRIV_REC_NEXT(9418, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9418(choice, ...)  METALANG99_PRIV_REC_NEXT(9419, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9419(choice, ...)  METALANG99_PRIV_REC_NEXT(9420, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9420(choice, ...)  METALANG99_PRIV_REC_NEXT(9421, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9421(choice, ...)  METALANG99_PRIV_REC_NEXT(9422, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9422(choice, ...)  METALANG99_PRIV_REC_NEXT(9423, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9423(choice, ...)  METALANG99_PRIV_REC_NEXT(9424, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9424(choice, ...)  METALANG99_PRIV_REC_NEXT(9425, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9425(choice, ...)  METALANG99_PRIV_REC_NEXT(9426, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9426(choice, ...)  METALANG99_PRIV_REC_NEXT(9427, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9427(choice, ...)  METALANG99_PRIV_REC_NEXT(9428, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9428(choice, ...)  METALANG99_PRIV_REC_NEXT(9429, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9429(choice, ...)  METALANG99_PRIV_REC_NEXT(9430, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9430(choice, ...)  METALANG99_PRIV_REC_NEXT(9431, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9431(choice, ...)  METALANG99_PRIV_REC_NEXT(9432, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9432(choice, ...)  METALANG99_PRIV_REC_NEXT(9433, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9433(choice, ...)  METALANG99_PRIV_REC_NEXT(9434, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9434(choice, ...)  METALANG99_PRIV_REC_NEXT(9435, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9435(choice, ...)  METALANG99_PRIV_REC_NEXT(9436, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9436(choice, ...)  METALANG99_PRIV_REC_NEXT(9437, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9437(choice, ...)  METALANG99_PRIV_REC_NEXT(9438, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9438(choice, ...)  METALANG99_PRIV_REC_NEXT(9439, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9439(choice, ...)  METALANG99_PRIV_REC_NEXT(9440, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9440(choice, ...)  METALANG99_PRIV_REC_NEXT(9441, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9441(choice, ...)  METALANG99_PRIV_REC_NEXT(9442, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9442(choice, ...)  METALANG99_PRIV_REC_NEXT(9443, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9443(choice, ...)  METALANG99_PRIV_REC_NEXT(9444, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9444(choice, ...)  METALANG99_PRIV_REC_NEXT(9445, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9445(choice, ...)  METALANG99_PRIV_REC_NEXT(9446, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9446(choice, ...)  METALANG99_PRIV_REC_NEXT(9447, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9447(choice, ...)  METALANG99_PRIV_REC_NEXT(9448, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9448(choice, ...)  METALANG99_PRIV_REC_NEXT(9449, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9449(choice, ...)  METALANG99_PRIV_REC_NEXT(9450, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9450(choice, ...)  METALANG99_PRIV_REC_NEXT(9451, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9451(choice, ...)  METALANG99_PRIV_REC_NEXT(9452, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9452(choice, ...)  METALANG99_PRIV_REC_NEXT(9453, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9453(choice, ...)  METALANG99_PRIV_REC_NEXT(9454, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9454(choice, ...)  METALANG99_PRIV_REC_NEXT(9455, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9455(choice, ...)  METALANG99_PRIV_REC_NEXT(9456, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9456(choice, ...)  METALANG99_PRIV_REC_NEXT(9457, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9457(choice, ...)  METALANG99_PRIV_REC_NEXT(9458, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9458(choice, ...)  METALANG99_PRIV_REC_NEXT(9459, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9459(choice, ...)  METALANG99_PRIV_REC_NEXT(9460, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9460(choice, ...)  METALANG99_PRIV_REC_NEXT(9461, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9461(choice, ...)  METALANG99_PRIV_REC_NEXT(9462, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9462(choice, ...)  METALANG99_PRIV_REC_NEXT(9463, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9463(choice, ...)  METALANG99_PRIV_REC_NEXT(9464, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9464(choice, ...)  METALANG99_PRIV_REC_NEXT(9465, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9465(choice, ...)  METALANG99_PRIV_REC_NEXT(9466, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9466(choice, ...)  METALANG99_PRIV_REC_NEXT(9467, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9467(choice, ...)  METALANG99_PRIV_REC_NEXT(9468, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9468(choice, ...)  METALANG99_PRIV_REC_NEXT(9469, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9469(choice, ...)  METALANG99_PRIV_REC_NEXT(9470, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9470(choice, ...)  METALANG99_PRIV_REC_NEXT(9471, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9471(choice, ...)  METALANG99_PRIV_REC_NEXT(9472, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9472(choice, ...)  METALANG99_PRIV_REC_NEXT(9473, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9473(choice, ...)  METALANG99_PRIV_REC_NEXT(9474, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9474(choice, ...)  METALANG99_PRIV_REC_NEXT(9475, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9475(choice, ...)  METALANG99_PRIV_REC_NEXT(9476, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9476(choice, ...)  METALANG99_PRIV_REC_NEXT(9477, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9477(choice, ...)  METALANG99_PRIV_REC_NEXT(9478, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9478(choice, ...)  METALANG99_PRIV_REC_NEXT(9479, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9479(choice, ...)  METALANG99_PRIV_REC_NEXT(9480, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9480(choice, ...)  METALANG99_PRIV_REC_NEXT(9481, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9481(choice, ...)  METALANG99_PRIV_REC_NEXT(9482, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9482(choice, ...)  METALANG99_PRIV_REC_NEXT(9483, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9483(choice, ...)  METALANG99_PRIV_REC_NEXT(9484, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9484(choice, ...)  METALANG99_PRIV_REC_NEXT(9485, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9485(choice, ...)  METALANG99_PRIV_REC_NEXT(9486, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9486(choice, ...)  METALANG99_PRIV_REC_NEXT(9487, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9487(choice, ...)  METALANG99_PRIV_REC_NEXT(9488, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9488(choice, ...)  METALANG99_PRIV_REC_NEXT(9489, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9489(choice, ...)  METALANG99_PRIV_REC_NEXT(9490, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9490(choice, ...)  METALANG99_PRIV_REC_NEXT(9491, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9491(choice, ...)  METALANG99_PRIV_REC_NEXT(9492, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9492(choice, ...)  METALANG99_PRIV_REC_NEXT(9493, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9493(choice, ...)  METALANG99_PRIV_REC_NEXT(9494, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9494(choice, ...)  METALANG99_PRIV_REC_NEXT(9495, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9495(choice, ...)  METALANG99_PRIV_REC_NEXT(9496, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9496(choice, ...)  METALANG99_PRIV_REC_NEXT(9497, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9497(choice, ...)  METALANG99_PRIV_REC_NEXT(9498, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9498(choice, ...)  METALANG99_PRIV_REC_NEXT(9499, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9499(choice, ...)  METALANG99_PRIV_REC_NEXT(9500, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9500(choice, ...)  METALANG99_PRIV_REC_NEXT(9501, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9501(choice, ...)  METALANG99_PRIV_REC_NEXT(9502, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9502(choice, ...)  METALANG99_PRIV_REC_NEXT(9503, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9503(choice, ...)  METALANG99_PRIV_REC_NEXT(9504, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9504(choice, ...)  METALANG99_PRIV_REC_NEXT(9505, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9505(choice, ...)  METALANG99_PRIV_REC_NEXT(9506, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9506(choice, ...)  METALANG99_PRIV_REC_NEXT(9507, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9507(choice, ...)  METALANG99_PRIV_REC_NEXT(9508, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9508(choice, ...)  METALANG99_PRIV_REC_NEXT(9509, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9509(choice, ...)  METALANG99_PRIV_REC_NEXT(9510, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9510(choice, ...)  METALANG99_PRIV_REC_NEXT(9511, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9511(choice, ...)  METALANG99_PRIV_REC_NEXT(9512, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9512(choice, ...)  METALANG99_PRIV_REC_NEXT(9513, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9513(choice, ...)  METALANG99_PRIV_REC_NEXT(9514, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9514(choice, ...)  METALANG99_PRIV_REC_NEXT(9515, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9515(choice, ...)  METALANG99_PRIV_REC_NEXT(9516, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9516(choice, ...)  METALANG99_PRIV_REC_NEXT(9517, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9517(choice, ...)  METALANG99_PRIV_REC_NEXT(9518, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9518(choice, ...)  METALANG99_PRIV_REC_NEXT(9519, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9519(choice, ...)  METALANG99_PRIV_REC_NEXT(9520, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9520(choice, ...)  METALANG99_PRIV_REC_NEXT(9521, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9521(choice, ...)  METALANG99_PRIV_REC_NEXT(9522, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9522(choice, ...)  METALANG99_PRIV_REC_NEXT(9523, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9523(choice, ...)  METALANG99_PRIV_REC_NEXT(9524, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9524(choice, ...)  METALANG99_PRIV_REC_NEXT(9525, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9525(choice, ...)  METALANG99_PRIV_REC_NEXT(9526, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9526(choice, ...)  METALANG99_PRIV_REC_NEXT(9527, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9527(choice, ...)  METALANG99_PRIV_REC_NEXT(9528, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9528(choice, ...)  METALANG99_PRIV_REC_NEXT(9529, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9529(choice, ...)  METALANG99_PRIV_REC_NEXT(9530, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9530(choice, ...)  METALANG99_PRIV_REC_NEXT(9531, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9531(choice, ...)  METALANG99_PRIV_REC_NEXT(9532, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9532(choice, ...)  METALANG99_PRIV_REC_NEXT(9533, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9533(choice, ...)  METALANG99_PRIV_REC_NEXT(9534, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9534(choice, ...)  METALANG99_PRIV_REC_NEXT(9535, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9535(choice, ...)  METALANG99_PRIV_REC_NEXT(9536, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9536(choice, ...)  METALANG99_PRIV_REC_NEXT(9537, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9537(choice, ...)  METALANG99_PRIV_REC_NEXT(9538, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9538(choice, ...)  METALANG99_PRIV_REC_NEXT(9539, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9539(choice, ...)  METALANG99_PRIV_REC_NEXT(9540, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9540(choice, ...)  METALANG99_PRIV_REC_NEXT(9541, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9541(choice, ...)  METALANG99_PRIV_REC_NEXT(9542, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9542(choice, ...)  METALANG99_PRIV_REC_NEXT(9543, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9543(choice, ...)  METALANG99_PRIV_REC_NEXT(9544, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9544(choice, ...)  METALANG99_PRIV_REC_NEXT(9545, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9545(choice, ...)  METALANG99_PRIV_REC_NEXT(9546, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9546(choice, ...)  METALANG99_PRIV_REC_NEXT(9547, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9547(choice, ...)  METALANG99_PRIV_REC_NEXT(9548, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9548(choice, ...)  METALANG99_PRIV_REC_NEXT(9549, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9549(choice, ...)  METALANG99_PRIV_REC_NEXT(9550, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9550(choice, ...)  METALANG99_PRIV_REC_NEXT(9551, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9551(choice, ...)  METALANG99_PRIV_REC_NEXT(9552, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9552(choice, ...)  METALANG99_PRIV_REC_NEXT(9553, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9553(choice, ...)  METALANG99_PRIV_REC_NEXT(9554, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9554(choice, ...)  METALANG99_PRIV_REC_NEXT(9555, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9555(choice, ...)  METALANG99_PRIV_REC_NEXT(9556, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9556(choice, ...)  METALANG99_PRIV_REC_NEXT(9557, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9557(choice, ...)  METALANG99_PRIV_REC_NEXT(9558, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9558(choice, ...)  METALANG99_PRIV_REC_NEXT(9559, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9559(choice, ...)  METALANG99_PRIV_REC_NEXT(9560, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9560(choice, ...)  METALANG99_PRIV_REC_NEXT(9561, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9561(choice, ...)  METALANG99_PRIV_REC_NEXT(9562, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9562(choice, ...)  METALANG99_PRIV_REC_NEXT(9563, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9563(choice, ...)  METALANG99_PRIV_REC_NEXT(9564, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9564(choice, ...)  METALANG99_PRIV_REC_NEXT(9565, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9565(choice, ...)  METALANG99_PRIV_REC_NEXT(9566, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9566(choice, ...)  METALANG99_PRIV_REC_NEXT(9567, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9567(choice, ...)  METALANG99_PRIV_REC_NEXT(9568, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9568(choice, ...)  METALANG99_PRIV_REC_NEXT(9569, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9569(choice, ...)  METALANG99_PRIV_REC_NEXT(9570, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9570(choice, ...)  METALANG99_PRIV_REC_NEXT(9571, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9571(choice, ...)  METALANG99_PRIV_REC_NEXT(9572, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9572(choice, ...)  METALANG99_PRIV_REC_NEXT(9573, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9573(choice, ...)  METALANG99_PRIV_REC_NEXT(9574, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9574(choice, ...)  METALANG99_PRIV_REC_NEXT(9575, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9575(choice, ...)  METALANG99_PRIV_REC_NEXT(9576, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9576(choice, ...)  METALANG99_PRIV_REC_NEXT(9577, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9577(choice, ...)  METALANG99_PRIV_REC_NEXT(9578, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9578(choice, ...)  METALANG99_PRIV_REC_NEXT(9579, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9579(choice, ...)  METALANG99_PRIV_REC_NEXT(9580, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9580(choice, ...)  METALANG99_PRIV_REC_NEXT(9581, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9581(choice, ...)  METALANG99_PRIV_REC_NEXT(9582, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9582(choice, ...)  METALANG99_PRIV_REC_NEXT(9583, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9583(choice, ...)  METALANG99_PRIV_REC_NEXT(9584, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9584(choice, ...)  METALANG99_PRIV_REC_NEXT(9585, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9585(choice, ...)  METALANG99_PRIV_REC_NEXT(9586, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9586(choice, ...)  METALANG99_PRIV_REC_NEXT(9587, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9587(choice, ...)  METALANG99_PRIV_REC_NEXT(9588, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9588(choice, ...)  METALANG99_PRIV_REC_NEXT(9589, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9589(choice, ...)  METALANG99_PRIV_REC_NEXT(9590, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9590(choice, ...)  METALANG99_PRIV_REC_NEXT(9591, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9591(choice, ...)  METALANG99_PRIV_REC_NEXT(9592, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9592(choice, ...)  METALANG99_PRIV_REC_NEXT(9593, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9593(choice, ...)  METALANG99_PRIV_REC_NEXT(9594, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9594(choice, ...)  METALANG99_PRIV_REC_NEXT(9595, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9595(choice, ...)  METALANG99_PRIV_REC_NEXT(9596, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9596(choice, ...)  METALANG99_PRIV_REC_NEXT(9597, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9597(choice, ...)  METALANG99_PRIV_REC_NEXT(9598, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9598(choice, ...)  METALANG99_PRIV_REC_NEXT(9599, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9599(choice, ...)  METALANG99_PRIV_REC_NEXT(9600, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9600(choice, ...)  METALANG99_PRIV_REC_NEXT(9601, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9601(choice, ...)  METALANG99_PRIV_REC_NEXT(9602, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9602(choice, ...)  METALANG99_PRIV_REC_NEXT(9603, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9603(choice, ...)  METALANG99_PRIV_REC_NEXT(9604, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9604(choice, ...)  METALANG99_PRIV_REC_NEXT(9605, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9605(choice, ...)  METALANG99_PRIV_REC_NEXT(9606, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9606(choice, ...)  METALANG99_PRIV_REC_NEXT(9607, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9607(choice, ...)  METALANG99_PRIV_REC_NEXT(9608, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9608(choice, ...)  METALANG99_PRIV_REC_NEXT(9609, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9609(choice, ...)  METALANG99_PRIV_REC_NEXT(9610, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9610(choice, ...)  METALANG99_PRIV_REC_NEXT(9611, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9611(choice, ...)  METALANG99_PRIV_REC_NEXT(9612, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9612(choice, ...)  METALANG99_PRIV_REC_NEXT(9613, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9613(choice, ...)  METALANG99_PRIV_REC_NEXT(9614, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9614(choice, ...)  METALANG99_PRIV_REC_NEXT(9615, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9615(choice, ...)  METALANG99_PRIV_REC_NEXT(9616, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9616(choice, ...)  METALANG99_PRIV_REC_NEXT(9617, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9617(choice, ...)  METALANG99_PRIV_REC_NEXT(9618, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9618(choice, ...)  METALANG99_PRIV_REC_NEXT(9619, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9619(choice, ...)  METALANG99_PRIV_REC_NEXT(9620, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9620(choice, ...)  METALANG99_PRIV_REC_NEXT(9621, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9621(choice, ...)  METALANG99_PRIV_REC_NEXT(9622, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9622(choice, ...)  METALANG99_PRIV_REC_NEXT(9623, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9623(choice, ...)  METALANG99_PRIV_REC_NEXT(9624, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9624(choice, ...)  METALANG99_PRIV_REC_NEXT(9625, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9625(choice, ...)  METALANG99_PRIV_REC_NEXT(9626, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9626(choice, ...)  METALANG99_PRIV_REC_NEXT(9627, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9627(choice, ...)  METALANG99_PRIV_REC_NEXT(9628, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9628(choice, ...)  METALANG99_PRIV_REC_NEXT(9629, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9629(choice, ...)  METALANG99_PRIV_REC_NEXT(9630, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9630(choice, ...)  METALANG99_PRIV_REC_NEXT(9631, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9631(choice, ...)  METALANG99_PRIV_REC_NEXT(9632, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9632(choice, ...)  METALANG99_PRIV_REC_NEXT(9633, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9633(choice, ...)  METALANG99_PRIV_REC_NEXT(9634, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9634(choice, ...)  METALANG99_PRIV_REC_NEXT(9635, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9635(choice, ...)  METALANG99_PRIV_REC_NEXT(9636, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9636(choice, ...)  METALANG99_PRIV_REC_NEXT(9637, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9637(choice, ...)  METALANG99_PRIV_REC_NEXT(9638, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9638(choice, ...)  METALANG99_PRIV_REC_NEXT(9639, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9639(choice, ...)  METALANG99_PRIV_REC_NEXT(9640, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9640(choice, ...)  METALANG99_PRIV_REC_NEXT(9641, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9641(choice, ...)  METALANG99_PRIV_REC_NEXT(9642, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9642(choice, ...)  METALANG99_PRIV_REC_NEXT(9643, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9643(choice, ...)  METALANG99_PRIV_REC_NEXT(9644, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9644(choice, ...)  METALANG99_PRIV_REC_NEXT(9645, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9645(choice, ...)  METALANG99_PRIV_REC_NEXT(9646, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9646(choice, ...)  METALANG99_PRIV_REC_NEXT(9647, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9647(choice, ...)  METALANG99_PRIV_REC_NEXT(9648, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9648(choice, ...)  METALANG99_PRIV_REC_NEXT(9649, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9649(choice, ...)  METALANG99_PRIV_REC_NEXT(9650, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9650(choice, ...)  METALANG99_PRIV_REC_NEXT(9651, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9651(choice, ...)  METALANG99_PRIV_REC_NEXT(9652, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9652(choice, ...)  METALANG99_PRIV_REC_NEXT(9653, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9653(choice, ...)  METALANG99_PRIV_REC_NEXT(9654, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9654(choice, ...)  METALANG99_PRIV_REC_NEXT(9655, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9655(choice, ...)  METALANG99_PRIV_REC_NEXT(9656, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9656(choice, ...)  METALANG99_PRIV_REC_NEXT(9657, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9657(choice, ...)  METALANG99_PRIV_REC_NEXT(9658, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9658(choice, ...)  METALANG99_PRIV_REC_NEXT(9659, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9659(choice, ...)  METALANG99_PRIV_REC_NEXT(9660, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9660(choice, ...)  METALANG99_PRIV_REC_NEXT(9661, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9661(choice, ...)  METALANG99_PRIV_REC_NEXT(9662, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9662(choice, ...)  METALANG99_PRIV_REC_NEXT(9663, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9663(choice, ...)  METALANG99_PRIV_REC_NEXT(9664, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9664(choice, ...)  METALANG99_PRIV_REC_NEXT(9665, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9665(choice, ...)  METALANG99_PRIV_REC_NEXT(9666, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9666(choice, ...)  METALANG99_PRIV_REC_NEXT(9667, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9667(choice, ...)  METALANG99_PRIV_REC_NEXT(9668, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9668(choice, ...)  METALANG99_PRIV_REC_NEXT(9669, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9669(choice, ...)  METALANG99_PRIV_REC_NEXT(9670, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9670(choice, ...)  METALANG99_PRIV_REC_NEXT(9671, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9671(choice, ...)  METALANG99_PRIV_REC_NEXT(9672, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9672(choice, ...)  METALANG99_PRIV_REC_NEXT(9673, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9673(choice, ...)  METALANG99_PRIV_REC_NEXT(9674, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9674(choice, ...)  METALANG99_PRIV_REC_NEXT(9675, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9675(choice, ...)  METALANG99_PRIV_REC_NEXT(9676, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9676(choice, ...)  METALANG99_PRIV_REC_NEXT(9677, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9677(choice, ...)  METALANG99_PRIV_REC_NEXT(9678, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9678(choice, ...)  METALANG99_PRIV_REC_NEXT(9679, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9679(choice, ...)  METALANG99_PRIV_REC_NEXT(9680, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9680(choice, ...)  METALANG99_PRIV_REC_NEXT(9681, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9681(choice, ...)  METALANG99_PRIV_REC_NEXT(9682, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9682(choice, ...)  METALANG99_PRIV_REC_NEXT(9683, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9683(choice, ...)  METALANG99_PRIV_REC_NEXT(9684, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9684(choice, ...)  METALANG99_PRIV_REC_NEXT(9685, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9685(choice, ...)  METALANG99_PRIV_REC_NEXT(9686, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9686(choice, ...)  METALANG99_PRIV_REC_NEXT(9687, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9687(choice, ...)  METALANG99_PRIV_REC_NEXT(9688, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9688(choice, ...)  METALANG99_PRIV_REC_NEXT(9689, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9689(choice, ...)  METALANG99_PRIV_REC_NEXT(9690, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9690(choice, ...)  METALANG99_PRIV_REC_NEXT(9691, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9691(choice, ...)  METALANG99_PRIV_REC_NEXT(9692, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9692(choice, ...)  METALANG99_PRIV_REC_NEXT(9693, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9693(choice, ...)  METALANG99_PRIV_REC_NEXT(9694, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9694(choice, ...)  METALANG99_PRIV_REC_NEXT(9695, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9695(choice, ...)  METALANG99_PRIV_REC_NEXT(9696, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9696(choice, ...)  METALANG99_PRIV_REC_NEXT(9697, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9697(choice, ...)  METALANG99_PRIV_REC_NEXT(9698, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9698(choice, ...)  METALANG99_PRIV_REC_NEXT(9699, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9699(choice, ...)  METALANG99_PRIV_REC_NEXT(9700, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9700(choice, ...)  METALANG99_PRIV_REC_NEXT(9701, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9701(choice, ...)  METALANG99_PRIV_REC_NEXT(9702, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9702(choice, ...)  METALANG99_PRIV_REC_NEXT(9703, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9703(choice, ...)  METALANG99_PRIV_REC_NEXT(9704, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9704(choice, ...)  METALANG99_PRIV_REC_NEXT(9705, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9705(choice, ...)  METALANG99_PRIV_REC_NEXT(9706, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9706(choice, ...)  METALANG99_PRIV_REC_NEXT(9707, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9707(choice, ...)  METALANG99_PRIV_REC_NEXT(9708, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9708(choice, ...)  METALANG99_PRIV_REC_NEXT(9709, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9709(choice, ...)  METALANG99_PRIV_REC_NEXT(9710, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9710(choice, ...)  METALANG99_PRIV_REC_NEXT(9711, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9711(choice, ...)  METALANG99_PRIV_REC_NEXT(9712, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9712(choice, ...)  METALANG99_PRIV_REC_NEXT(9713, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9713(choice, ...)  METALANG99_PRIV_REC_NEXT(9714, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9714(choice, ...)  METALANG99_PRIV_REC_NEXT(9715, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9715(choice, ...)  METALANG99_PRIV_REC_NEXT(9716, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9716(choice, ...)  METALANG99_PRIV_REC_NEXT(9717, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9717(choice, ...)  METALANG99_PRIV_REC_NEXT(9718, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9718(choice, ...)  METALANG99_PRIV_REC_NEXT(9719, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9719(choice, ...)  METALANG99_PRIV_REC_NEXT(9720, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9720(choice, ...)  METALANG99_PRIV_REC_NEXT(9721, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9721(choice, ...)  METALANG99_PRIV_REC_NEXT(9722, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9722(choice, ...)  METALANG99_PRIV_REC_NEXT(9723, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9723(choice, ...)  METALANG99_PRIV_REC_NEXT(9724, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9724(choice, ...)  METALANG99_PRIV_REC_NEXT(9725, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9725(choice, ...)  METALANG99_PRIV_REC_NEXT(9726, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9726(choice, ...)  METALANG99_PRIV_REC_NEXT(9727, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9727(choice, ...)  METALANG99_PRIV_REC_NEXT(9728, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9728(choice, ...)  METALANG99_PRIV_REC_NEXT(9729, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9729(choice, ...)  METALANG99_PRIV_REC_NEXT(9730, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9730(choice, ...)  METALANG99_PRIV_REC_NEXT(9731, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9731(choice, ...)  METALANG99_PRIV_REC_NEXT(9732, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9732(choice, ...)  METALANG99_PRIV_REC_NEXT(9733, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9733(choice, ...)  METALANG99_PRIV_REC_NEXT(9734, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9734(choice, ...)  METALANG99_PRIV_REC_NEXT(9735, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9735(choice, ...)  METALANG99_PRIV_REC_NEXT(9736, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9736(choice, ...)  METALANG99_PRIV_REC_NEXT(9737, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9737(choice, ...)  METALANG99_PRIV_REC_NEXT(9738, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9738(choice, ...)  METALANG99_PRIV_REC_NEXT(9739, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9739(choice, ...)  METALANG99_PRIV_REC_NEXT(9740, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9740(choice, ...)  METALANG99_PRIV_REC_NEXT(9741, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9741(choice, ...)  METALANG99_PRIV_REC_NEXT(9742, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9742(choice, ...)  METALANG99_PRIV_REC_NEXT(9743, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9743(choice, ...)  METALANG99_PRIV_REC_NEXT(9744, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9744(choice, ...)  METALANG99_PRIV_REC_NEXT(9745, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9745(choice, ...)  METALANG99_PRIV_REC_NEXT(9746, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9746(choice, ...)  METALANG99_PRIV_REC_NEXT(9747, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9747(choice, ...)  METALANG99_PRIV_REC_NEXT(9748, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9748(choice, ...)  METALANG99_PRIV_REC_NEXT(9749, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9749(choice, ...)  METALANG99_PRIV_REC_NEXT(9750, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9750(choice, ...)  METALANG99_PRIV_REC_NEXT(9751, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9751(choice, ...)  METALANG99_PRIV_REC_NEXT(9752, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9752(choice, ...)  METALANG99_PRIV_REC_NEXT(9753, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9753(choice, ...)  METALANG99_PRIV_REC_NEXT(9754, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9754(choice, ...)  METALANG99_PRIV_REC_NEXT(9755, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9755(choice, ...)  METALANG99_PRIV_REC_NEXT(9756, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9756(choice, ...)  METALANG99_PRIV_REC_NEXT(9757, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9757(choice, ...)  METALANG99_PRIV_REC_NEXT(9758, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9758(choice, ...)  METALANG99_PRIV_REC_NEXT(9759, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9759(choice, ...)  METALANG99_PRIV_REC_NEXT(9760, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9760(choice, ...)  METALANG99_PRIV_REC_NEXT(9761, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9761(choice, ...)  METALANG99_PRIV_REC_NEXT(9762, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9762(choice, ...)  METALANG99_PRIV_REC_NEXT(9763, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9763(choice, ...)  METALANG99_PRIV_REC_NEXT(9764, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9764(choice, ...)  METALANG99_PRIV_REC_NEXT(9765, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9765(choice, ...)  METALANG99_PRIV_REC_NEXT(9766, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9766(choice, ...)  METALANG99_PRIV_REC_NEXT(9767, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9767(choice, ...)  METALANG99_PRIV_REC_NEXT(9768, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9768(choice, ...)  METALANG99_PRIV_REC_NEXT(9769, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9769(choice, ...)  METALANG99_PRIV_REC_NEXT(9770, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9770(choice, ...)  METALANG99_PRIV_REC_NEXT(9771, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9771(choice, ...)  METALANG99_PRIV_REC_NEXT(9772, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9772(choice, ...)  METALANG99_PRIV_REC_NEXT(9773, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9773(choice, ...)  METALANG99_PRIV_REC_NEXT(9774, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9774(choice, ...)  METALANG99_PRIV_REC_NEXT(9775, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9775(choice, ...)  METALANG99_PRIV_REC_NEXT(9776, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9776(choice, ...)  METALANG99_PRIV_REC_NEXT(9777, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9777(choice, ...)  METALANG99_PRIV_REC_NEXT(9778, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9778(choice, ...)  METALANG99_PRIV_REC_NEXT(9779, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9779(choice, ...)  METALANG99_PRIV_REC_NEXT(9780, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9780(choice, ...)  METALANG99_PRIV_REC_NEXT(9781, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9781(choice, ...)  METALANG99_PRIV_REC_NEXT(9782, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9782(choice, ...)  METALANG99_PRIV_REC_NEXT(9783, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9783(choice, ...)  METALANG99_PRIV_REC_NEXT(9784, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9784(choice, ...)  METALANG99_PRIV_REC_NEXT(9785, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9785(choice, ...)  METALANG99_PRIV_REC_NEXT(9786, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9786(choice, ...)  METALANG99_PRIV_REC_NEXT(9787, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9787(choice, ...)  METALANG99_PRIV_REC_NEXT(9788, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9788(choice, ...)  METALANG99_PRIV_REC_NEXT(9789, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9789(choice, ...)  METALANG99_PRIV_REC_NEXT(9790, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9790(choice, ...)  METALANG99_PRIV_REC_NEXT(9791, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9791(choice, ...)  METALANG99_PRIV_REC_NEXT(9792, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9792(choice, ...)  METALANG99_PRIV_REC_NEXT(9793, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9793(choice, ...)  METALANG99_PRIV_REC_NEXT(9794, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9794(choice, ...)  METALANG99_PRIV_REC_NEXT(9795, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9795(choice, ...)  METALANG99_PRIV_REC_NEXT(9796, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9796(choice, ...)  METALANG99_PRIV_REC_NEXT(9797, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9797(choice, ...)  METALANG99_PRIV_REC_NEXT(9798, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9798(choice, ...)  METALANG99_PRIV_REC_NEXT(9799, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9799(choice, ...)  METALANG99_PRIV_REC_NEXT(9800, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9800(choice, ...)  METALANG99_PRIV_REC_NEXT(9801, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9801(choice, ...)  METALANG99_PRIV_REC_NEXT(9802, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9802(choice, ...)  METALANG99_PRIV_REC_NEXT(9803, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9803(choice, ...)  METALANG99_PRIV_REC_NEXT(9804, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9804(choice, ...)  METALANG99_PRIV_REC_NEXT(9805, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9805(choice, ...)  METALANG99_PRIV_REC_NEXT(9806, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9806(choice, ...)  METALANG99_PRIV_REC_NEXT(9807, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9807(choice, ...)  METALANG99_PRIV_REC_NEXT(9808, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9808(choice, ...)  METALANG99_PRIV_REC_NEXT(9809, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9809(choice, ...)  METALANG99_PRIV_REC_NEXT(9810, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9810(choice, ...)  METALANG99_PRIV_REC_NEXT(9811, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9811(choice, ...)  METALANG99_PRIV_REC_NEXT(9812, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9812(choice, ...)  METALANG99_PRIV_REC_NEXT(9813, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9813(choice, ...)  METALANG99_PRIV_REC_NEXT(9814, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9814(choice, ...)  METALANG99_PRIV_REC_NEXT(9815, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9815(choice, ...)  METALANG99_PRIV_REC_NEXT(9816, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9816(choice, ...)  METALANG99_PRIV_REC_NEXT(9817, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9817(choice, ...)  METALANG99_PRIV_REC_NEXT(9818, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9818(choice, ...)  METALANG99_PRIV_REC_NEXT(9819, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9819(choice, ...)  METALANG99_PRIV_REC_NEXT(9820, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9820(choice, ...)  METALANG99_PRIV_REC_NEXT(9821, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9821(choice, ...)  METALANG99_PRIV_REC_NEXT(9822, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9822(choice, ...)  METALANG99_PRIV_REC_NEXT(9823, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9823(choice, ...)  METALANG99_PRIV_REC_NEXT(9824, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9824(choice, ...)  METALANG99_PRIV_REC_NEXT(9825, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9825(choice, ...)  METALANG99_PRIV_REC_NEXT(9826, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9826(choice, ...)  METALANG99_PRIV_REC_NEXT(9827, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9827(choice, ...)  METALANG99_PRIV_REC_NEXT(9828, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9828(choice, ...)  METALANG99_PRIV_REC_NEXT(9829, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9829(choice, ...)  METALANG99_PRIV_REC_NEXT(9830, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9830(choice, ...)  METALANG99_PRIV_REC_NEXT(9831, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9831(choice, ...)  METALANG99_PRIV_REC_NEXT(9832, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9832(choice, ...)  METALANG99_PRIV_REC_NEXT(9833, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9833(choice, ...)  METALANG99_PRIV_REC_NEXT(9834, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9834(choice, ...)  METALANG99_PRIV_REC_NEXT(9835, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9835(choice, ...)  METALANG99_PRIV_REC_NEXT(9836, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9836(choice, ...)  METALANG99_PRIV_REC_NEXT(9837, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9837(choice, ...)  METALANG99_PRIV_REC_NEXT(9838, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9838(choice, ...)  METALANG99_PRIV_REC_NEXT(9839, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9839(choice, ...)  METALANG99_PRIV_REC_NEXT(9840, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9840(choice, ...)  METALANG99_PRIV_REC_NEXT(9841, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9841(choice, ...)  METALANG99_PRIV_REC_NEXT(9842, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9842(choice, ...)  METALANG99_PRIV_REC_NEXT(9843, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9843(choice, ...)  METALANG99_PRIV_REC_NEXT(9844, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9844(choice, ...)  METALANG99_PRIV_REC_NEXT(9845, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9845(choice, ...)  METALANG99_PRIV_REC_NEXT(9846, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9846(choice, ...)  METALANG99_PRIV_REC_NEXT(9847, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9847(choice, ...)  METALANG99_PRIV_REC_NEXT(9848, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9848(choice, ...)  METALANG99_PRIV_REC_NEXT(9849, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9849(choice, ...)  METALANG99_PRIV_REC_NEXT(9850, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9850(choice, ...)  METALANG99_PRIV_REC_NEXT(9851, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9851(choice, ...)  METALANG99_PRIV_REC_NEXT(9852, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9852(choice, ...)  METALANG99_PRIV_REC_NEXT(9853, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9853(choice, ...)  METALANG99_PRIV_REC_NEXT(9854, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9854(choice, ...)  METALANG99_PRIV_REC_NEXT(9855, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9855(choice, ...)  METALANG99_PRIV_REC_NEXT(9856, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9856(choice, ...)  METALANG99_PRIV_REC_NEXT(9857, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9857(choice, ...)  METALANG99_PRIV_REC_NEXT(9858, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9858(choice, ...)  METALANG99_PRIV_REC_NEXT(9859, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9859(choice, ...)  METALANG99_PRIV_REC_NEXT(9860, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9860(choice, ...)  METALANG99_PRIV_REC_NEXT(9861, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9861(choice, ...)  METALANG99_PRIV_REC_NEXT(9862, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9862(choice, ...)  METALANG99_PRIV_REC_NEXT(9863, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9863(choice, ...)  METALANG99_PRIV_REC_NEXT(9864, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9864(choice, ...)  METALANG99_PRIV_REC_NEXT(9865, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9865(choice, ...)  METALANG99_PRIV_REC_NEXT(9866, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9866(choice, ...)  METALANG99_PRIV_REC_NEXT(9867, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9867(choice, ...)  METALANG99_PRIV_REC_NEXT(9868, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9868(choice, ...)  METALANG99_PRIV_REC_NEXT(9869, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9869(choice, ...)  METALANG99_PRIV_REC_NEXT(9870, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9870(choice, ...)  METALANG99_PRIV_REC_NEXT(9871, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9871(choice, ...)  METALANG99_PRIV_REC_NEXT(9872, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9872(choice, ...)  METALANG99_PRIV_REC_NEXT(9873, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9873(choice, ...)  METALANG99_PRIV_REC_NEXT(9874, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9874(choice, ...)  METALANG99_PRIV_REC_NEXT(9875, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9875(choice, ...)  METALANG99_PRIV_REC_NEXT(9876, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9876(choice, ...)  METALANG99_PRIV_REC_NEXT(9877, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9877(choice, ...)  METALANG99_PRIV_REC_NEXT(9878, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9878(choice, ...)  METALANG99_PRIV_REC_NEXT(9879, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9879(choice, ...)  METALANG99_PRIV_REC_NEXT(9880, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9880(choice, ...)  METALANG99_PRIV_REC_NEXT(9881, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9881(choice, ...)  METALANG99_PRIV_REC_NEXT(9882, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9882(choice, ...)  METALANG99_PRIV_REC_NEXT(9883, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9883(choice, ...)  METALANG99_PRIV_REC_NEXT(9884, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9884(choice, ...)  METALANG99_PRIV_REC_NEXT(9885, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9885(choice, ...)  METALANG99_PRIV_REC_NEXT(9886, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9886(choice, ...)  METALANG99_PRIV_REC_NEXT(9887, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9887(choice, ...)  METALANG99_PRIV_REC_NEXT(9888, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9888(choice, ...)  METALANG99_PRIV_REC_NEXT(9889, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9889(choice, ...)  METALANG99_PRIV_REC_NEXT(9890, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9890(choice, ...)  METALANG99_PRIV_REC_NEXT(9891, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9891(choice, ...)  METALANG99_PRIV_REC_NEXT(9892, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9892(choice, ...)  METALANG99_PRIV_REC_NEXT(9893, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9893(choice, ...)  METALANG99_PRIV_REC_NEXT(9894, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9894(choice, ...)  METALANG99_PRIV_REC_NEXT(9895, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9895(choice, ...)  METALANG99_PRIV_REC_NEXT(9896, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9896(choice, ...)  METALANG99_PRIV_REC_NEXT(9897, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9897(choice, ...)  METALANG99_PRIV_REC_NEXT(9898, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9898(choice, ...)  METALANG99_PRIV_REC_NEXT(9899, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9899(choice, ...)  METALANG99_PRIV_REC_NEXT(9900, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9900(choice, ...)  METALANG99_PRIV_REC_NEXT(9901, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9901(choice, ...)  METALANG99_PRIV_REC_NEXT(9902, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9902(choice, ...)  METALANG99_PRIV_REC_NEXT(9903, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9903(choice, ...)  METALANG99_PRIV_REC_NEXT(9904, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9904(choice, ...)  METALANG99_PRIV_REC_NEXT(9905, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9905(choice, ...)  METALANG99_PRIV_REC_NEXT(9906, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9906(choice, ...)  METALANG99_PRIV_REC_NEXT(9907, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9907(choice, ...)  METALANG99_PRIV_REC_NEXT(9908, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9908(choice, ...)  METALANG99_PRIV_REC_NEXT(9909, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9909(choice, ...)  METALANG99_PRIV_REC_NEXT(9910, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9910(choice, ...)  METALANG99_PRIV_REC_NEXT(9911, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9911(choice, ...)  METALANG99_PRIV_REC_NEXT(9912, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9912(choice, ...)  METALANG99_PRIV_REC_NEXT(9913, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9913(choice, ...)  METALANG99_PRIV_REC_NEXT(9914, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9914(choice, ...)  METALANG99_PRIV_REC_NEXT(9915, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9915(choice, ...)  METALANG99_PRIV_REC_NEXT(9916, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9916(choice, ...)  METALANG99_PRIV_REC_NEXT(9917, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9917(choice, ...)  METALANG99_PRIV_REC_NEXT(9918, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9918(choice, ...)  METALANG99_PRIV_REC_NEXT(9919, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9919(choice, ...)  METALANG99_PRIV_REC_NEXT(9920, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9920(choice, ...)  METALANG99_PRIV_REC_NEXT(9921, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9921(choice, ...)  METALANG99_PRIV_REC_NEXT(9922, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9922(choice, ...)  METALANG99_PRIV_REC_NEXT(9923, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9923(choice, ...)  METALANG99_PRIV_REC_NEXT(9924, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9924(choice, ...)  METALANG99_PRIV_REC_NEXT(9925, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9925(choice, ...)  METALANG99_PRIV_REC_NEXT(9926, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9926(choice, ...)  METALANG99_PRIV_REC_NEXT(9927, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9927(choice, ...)  METALANG99_PRIV_REC_NEXT(9928, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9928(choice, ...)  METALANG99_PRIV_REC_NEXT(9929, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9929(choice, ...)  METALANG99_PRIV_REC_NEXT(9930, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9930(choice, ...)  METALANG99_PRIV_REC_NEXT(9931, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9931(choice, ...)  METALANG99_PRIV_REC_NEXT(9932, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9932(choice, ...)  METALANG99_PRIV_REC_NEXT(9933, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9933(choice, ...)  METALANG99_PRIV_REC_NEXT(9934, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9934(choice, ...)  METALANG99_PRIV_REC_NEXT(9935, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9935(choice, ...)  METALANG99_PRIV_REC_NEXT(9936, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9936(choice, ...)  METALANG99_PRIV_REC_NEXT(9937, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9937(choice, ...)  METALANG99_PRIV_REC_NEXT(9938, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9938(choice, ...)  METALANG99_PRIV_REC_NEXT(9939, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9939(choice, ...)  METALANG99_PRIV_REC_NEXT(9940, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9940(choice, ...)  METALANG99_PRIV_REC_NEXT(9941, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9941(choice, ...)  METALANG99_PRIV_REC_NEXT(9942, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9942(choice, ...)  METALANG99_PRIV_REC_NEXT(9943, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9943(choice, ...)  METALANG99_PRIV_REC_NEXT(9944, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9944(choice, ...)  METALANG99_PRIV_REC_NEXT(9945, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9945(choice, ...)  METALANG99_PRIV_REC_NEXT(9946, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9946(choice, ...)  METALANG99_PRIV_REC_NEXT(9947, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9947(choice, ...)  METALANG99_PRIV_REC_NEXT(9948, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9948(choice, ...)  METALANG99_PRIV_REC_NEXT(9949, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9949(choice, ...)  METALANG99_PRIV_REC_NEXT(9950, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9950(choice, ...)  METALANG99_PRIV_REC_NEXT(9951, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9951(choice, ...)  METALANG99_PRIV_REC_NEXT(9952, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9952(choice, ...)  METALANG99_PRIV_REC_NEXT(9953, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9953(choice, ...)  METALANG99_PRIV_REC_NEXT(9954, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9954(choice, ...)  METALANG99_PRIV_REC_NEXT(9955, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9955(choice, ...)  METALANG99_PRIV_REC_NEXT(9956, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9956(choice, ...)  METALANG99_PRIV_REC_NEXT(9957, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9957(choice, ...)  METALANG99_PRIV_REC_NEXT(9958, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9958(choice, ...)  METALANG99_PRIV_REC_NEXT(9959, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9959(choice, ...)  METALANG99_PRIV_REC_NEXT(9960, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9960(choice, ...)  METALANG99_PRIV_REC_NEXT(9961, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9961(choice, ...)  METALANG99_PRIV_REC_NEXT(9962, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9962(choice, ...)  METALANG99_PRIV_REC_NEXT(9963, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9963(choice, ...)  METALANG99_PRIV_REC_NEXT(9964, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9964(choice, ...)  METALANG99_PRIV_REC_NEXT(9965, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9965(choice, ...)  METALANG99_PRIV_REC_NEXT(9966, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9966(choice, ...)  METALANG99_PRIV_REC_NEXT(9967, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9967(choice, ...)  METALANG99_PRIV_REC_NEXT(9968, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9968(choice, ...)  METALANG99_PRIV_REC_NEXT(9969, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9969(choice, ...)  METALANG99_PRIV_REC_NEXT(9970, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9970(choice, ...)  METALANG99_PRIV_REC_NEXT(9971, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9971(choice, ...)  METALANG99_PRIV_REC_NEXT(9972, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9972(choice, ...)  METALANG99_PRIV_REC_NEXT(9973, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9973(choice, ...)  METALANG99_PRIV_REC_NEXT(9974, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9974(choice, ...)  METALANG99_PRIV_REC_NEXT(9975, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9975(choice, ...)  METALANG99_PRIV_REC_NEXT(9976, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9976(choice, ...)  METALANG99_PRIV_REC_NEXT(9977, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9977(choice, ...)  METALANG99_PRIV_REC_NEXT(9978, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9978(choice, ...)  METALANG99_PRIV_REC_NEXT(9979, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9979(choice, ...)  METALANG99_PRIV_REC_NEXT(9980, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9980(choice, ...)  METALANG99_PRIV_REC_NEXT(9981, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9981(choice, ...)  METALANG99_PRIV_REC_NEXT(9982, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9982(choice, ...)  METALANG99_PRIV_REC_NEXT(9983, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9983(choice, ...)  METALANG99_PRIV_REC_NEXT(9984, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9984(choice, ...)  METALANG99_PRIV_REC_NEXT(9985, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9985(choice, ...)  METALANG99_PRIV_REC_NEXT(9986, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9986(choice, ...)  METALANG99_PRIV_REC_NEXT(9987, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9987(choice, ...)  METALANG99_PRIV_REC_NEXT(9988, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9988(choice, ...)  METALANG99_PRIV_REC_NEXT(9989, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9989(choice, ...)  METALANG99_PRIV_REC_NEXT(9990, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9990(choice, ...)  METALANG99_PRIV_REC_NEXT(9991, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9991(choice, ...)  METALANG99_PRIV_REC_NEXT(9992, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9992(choice, ...)  METALANG99_PRIV_REC_NEXT(9993, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9993(choice, ...)  METALANG99_PRIV_REC_NEXT(9994, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9994(choice, ...)  METALANG99_PRIV_REC_NEXT(9995, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9995(choice, ...)  METALANG99_PRIV_REC_NEXT(9996, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9996(choice, ...)  METALANG99_PRIV_REC_NEXT(9997, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9997(choice, ...)  METALANG99_PRIV_REC_NEXT(9998, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9998(choice, ...)  METALANG99_PRIV_REC_NEXT(9999, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_9999(choice, ...)  METALANG99_PRIV_REC_NEXT(10000, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10000(choice, ...) METALANG99_PRIV_REC_NEXT(10001, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10001(choice, ...) METALANG99_PRIV_REC_NEXT(10002, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10002(choice, ...) METALANG99_PRIV_REC_NEXT(10003, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10003(choice, ...) METALANG99_PRIV_REC_NEXT(10004, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10004(choice, ...) METALANG99_PRIV_REC_NEXT(10005, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10005(choice, ...) METALANG99_PRIV_REC_NEXT(10006, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10006(choice, ...) METALANG99_PRIV_REC_NEXT(10007, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10007(choice, ...) METALANG99_PRIV_REC_NEXT(10008, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10008(choice, ...) METALANG99_PRIV_REC_NEXT(10009, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10009(choice, ...) METALANG99_PRIV_REC_NEXT(10010, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10010(choice, ...) METALANG99_PRIV_REC_NEXT(10011, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10011(choice, ...) METALANG99_PRIV_REC_NEXT(10012, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10012(choice, ...) METALANG99_PRIV_REC_NEXT(10013, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10013(choice, ...) METALANG99_PRIV_REC_NEXT(10014, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10014(choice, ...) METALANG99_PRIV_REC_NEXT(10015, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10015(choice, ...) METALANG99_PRIV_REC_NEXT(10016, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10016(choice, ...) METALANG99_PRIV_REC_NEXT(10017, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10017(choice, ...) METALANG99_PRIV_REC_NEXT(10018, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10018(choice, ...) METALANG99_PRIV_REC_NEXT(10019, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10019(choice, ...) METALANG99_PRIV_REC_NEXT(10020, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10020(choice, ...) METALANG99_PRIV_REC_NEXT(10021, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10021(choice, ...) METALANG99_PRIV_REC_NEXT(10022, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10022(choice, ...) METALANG99_PRIV_REC_NEXT(10023, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10023(choice, ...) METALANG99_PRIV_REC_NEXT(10024, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10024(choice, ...) METALANG99_PRIV_REC_NEXT(10025, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10025(choice, ...) METALANG99_PRIV_REC_NEXT(10026, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10026(choice, ...) METALANG99_PRIV_REC_NEXT(10027, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10027(choice, ...) METALANG99_PRIV_REC_NEXT(10028, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10028(choice, ...) METALANG99_PRIV_REC_NEXT(10029, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10029(choice, ...) METALANG99_PRIV_REC_NEXT(10030, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10030(choice, ...) METALANG99_PRIV_REC_NEXT(10031, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10031(choice, ...) METALANG99_PRIV_REC_NEXT(10032, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10032(choice, ...) METALANG99_PRIV_REC_NEXT(10033, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10033(choice, ...) METALANG99_PRIV_REC_NEXT(10034, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10034(choice, ...) METALANG99_PRIV_REC_NEXT(10035, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10035(choice, ...) METALANG99_PRIV_REC_NEXT(10036, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10036(choice, ...) METALANG99_PRIV_REC_NEXT(10037, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10037(choice, ...) METALANG99_PRIV_REC_NEXT(10038, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10038(choice, ...) METALANG99_PRIV_REC_NEXT(10039, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10039(choice, ...) METALANG99_PRIV_REC_NEXT(10040, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10040(choice, ...) METALANG99_PRIV_REC_NEXT(10041, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10041(choice, ...) METALANG99_PRIV_REC_NEXT(10042, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10042(choice, ...) METALANG99_PRIV_REC_NEXT(10043, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10043(choice, ...) METALANG99_PRIV_REC_NEXT(10044, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10044(choice, ...) METALANG99_PRIV_REC_NEXT(10045, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10045(choice, ...) METALANG99_PRIV_REC_NEXT(10046, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10046(choice, ...) METALANG99_PRIV_REC_NEXT(10047, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10047(choice, ...) METALANG99_PRIV_REC_NEXT(10048, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10048(choice, ...) METALANG99_PRIV_REC_NEXT(10049, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10049(choice, ...) METALANG99_PRIV_REC_NEXT(10050, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10050(choice, ...) METALANG99_PRIV_REC_NEXT(10051, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10051(choice, ...) METALANG99_PRIV_REC_NEXT(10052, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10052(choice, ...) METALANG99_PRIV_REC_NEXT(10053, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10053(choice, ...) METALANG99_PRIV_REC_NEXT(10054, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10054(choice, ...) METALANG99_PRIV_REC_NEXT(10055, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10055(choice, ...) METALANG99_PRIV_REC_NEXT(10056, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10056(choice, ...) METALANG99_PRIV_REC_NEXT(10057, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10057(choice, ...) METALANG99_PRIV_REC_NEXT(10058, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10058(choice, ...) METALANG99_PRIV_REC_NEXT(10059, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10059(choice, ...) METALANG99_PRIV_REC_NEXT(10060, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10060(choice, ...) METALANG99_PRIV_REC_NEXT(10061, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10061(choice, ...) METALANG99_PRIV_REC_NEXT(10062, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10062(choice, ...) METALANG99_PRIV_REC_NEXT(10063, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10063(choice, ...) METALANG99_PRIV_REC_NEXT(10064, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10064(choice, ...) METALANG99_PRIV_REC_NEXT(10065, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10065(choice, ...) METALANG99_PRIV_REC_NEXT(10066, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10066(choice, ...) METALANG99_PRIV_REC_NEXT(10067, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10067(choice, ...) METALANG99_PRIV_REC_NEXT(10068, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10068(choice, ...) METALANG99_PRIV_REC_NEXT(10069, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10069(choice, ...) METALANG99_PRIV_REC_NEXT(10070, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10070(choice, ...) METALANG99_PRIV_REC_NEXT(10071, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10071(choice, ...) METALANG99_PRIV_REC_NEXT(10072, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10072(choice, ...) METALANG99_PRIV_REC_NEXT(10073, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10073(choice, ...) METALANG99_PRIV_REC_NEXT(10074, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10074(choice, ...) METALANG99_PRIV_REC_NEXT(10075, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10075(choice, ...) METALANG99_PRIV_REC_NEXT(10076, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10076(choice, ...) METALANG99_PRIV_REC_NEXT(10077, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10077(choice, ...) METALANG99_PRIV_REC_NEXT(10078, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10078(choice, ...) METALANG99_PRIV_REC_NEXT(10079, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10079(choice, ...) METALANG99_PRIV_REC_NEXT(10080, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10080(choice, ...) METALANG99_PRIV_REC_NEXT(10081, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10081(choice, ...) METALANG99_PRIV_REC_NEXT(10082, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10082(choice, ...) METALANG99_PRIV_REC_NEXT(10083, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10083(choice, ...) METALANG99_PRIV_REC_NEXT(10084, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10084(choice, ...) METALANG99_PRIV_REC_NEXT(10085, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10085(choice, ...) METALANG99_PRIV_REC_NEXT(10086, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10086(choice, ...) METALANG99_PRIV_REC_NEXT(10087, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10087(choice, ...) METALANG99_PRIV_REC_NEXT(10088, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10088(choice, ...) METALANG99_PRIV_REC_NEXT(10089, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10089(choice, ...) METALANG99_PRIV_REC_NEXT(10090, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10090(choice, ...) METALANG99_PRIV_REC_NEXT(10091, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10091(choice, ...) METALANG99_PRIV_REC_NEXT(10092, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10092(choice, ...) METALANG99_PRIV_REC_NEXT(10093, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10093(choice, ...) METALANG99_PRIV_REC_NEXT(10094, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10094(choice, ...) METALANG99_PRIV_REC_NEXT(10095, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10095(choice, ...) METALANG99_PRIV_REC_NEXT(10096, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10096(choice, ...) METALANG99_PRIV_REC_NEXT(10097, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10097(choice, ...) METALANG99_PRIV_REC_NEXT(10098, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10098(choice, ...) METALANG99_PRIV_REC_NEXT(10099, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10099(choice, ...) METALANG99_PRIV_REC_NEXT(10100, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10100(choice, ...) METALANG99_PRIV_REC_NEXT(10101, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10101(choice, ...) METALANG99_PRIV_REC_NEXT(10102, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10102(choice, ...) METALANG99_PRIV_REC_NEXT(10103, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10103(choice, ...) METALANG99_PRIV_REC_NEXT(10104, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10104(choice, ...) METALANG99_PRIV_REC_NEXT(10105, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10105(choice, ...) METALANG99_PRIV_REC_NEXT(10106, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10106(choice, ...) METALANG99_PRIV_REC_NEXT(10107, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10107(choice, ...) METALANG99_PRIV_REC_NEXT(10108, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10108(choice, ...) METALANG99_PRIV_REC_NEXT(10109, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10109(choice, ...) METALANG99_PRIV_REC_NEXT(10110, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10110(choice, ...) METALANG99_PRIV_REC_NEXT(10111, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10111(choice, ...) METALANG99_PRIV_REC_NEXT(10112, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10112(choice, ...) METALANG99_PRIV_REC_NEXT(10113, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10113(choice, ...) METALANG99_PRIV_REC_NEXT(10114, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10114(choice, ...) METALANG99_PRIV_REC_NEXT(10115, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10115(choice, ...) METALANG99_PRIV_REC_NEXT(10116, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10116(choice, ...) METALANG99_PRIV_REC_NEXT(10117, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10117(choice, ...) METALANG99_PRIV_REC_NEXT(10118, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10118(choice, ...) METALANG99_PRIV_REC_NEXT(10119, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10119(choice, ...) METALANG99_PRIV_REC_NEXT(10120, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10120(choice, ...) METALANG99_PRIV_REC_NEXT(10121, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10121(choice, ...) METALANG99_PRIV_REC_NEXT(10122, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10122(choice, ...) METALANG99_PRIV_REC_NEXT(10123, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10123(choice, ...) METALANG99_PRIV_REC_NEXT(10124, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10124(choice, ...) METALANG99_PRIV_REC_NEXT(10125, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10125(choice, ...) METALANG99_PRIV_REC_NEXT(10126, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10126(choice, ...) METALANG99_PRIV_REC_NEXT(10127, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10127(choice, ...) METALANG99_PRIV_REC_NEXT(10128, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10128(choice, ...) METALANG99_PRIV_REC_NEXT(10129, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10129(choice, ...) METALANG99_PRIV_REC_NEXT(10130, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10130(choice, ...) METALANG99_PRIV_REC_NEXT(10131, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10131(choice, ...) METALANG99_PRIV_REC_NEXT(10132, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10132(choice, ...) METALANG99_PRIV_REC_NEXT(10133, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10133(choice, ...) METALANG99_PRIV_REC_NEXT(10134, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10134(choice, ...) METALANG99_PRIV_REC_NEXT(10135, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10135(choice, ...) METALANG99_PRIV_REC_NEXT(10136, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10136(choice, ...) METALANG99_PRIV_REC_NEXT(10137, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10137(choice, ...) METALANG99_PRIV_REC_NEXT(10138, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10138(choice, ...) METALANG99_PRIV_REC_NEXT(10139, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10139(choice, ...) METALANG99_PRIV_REC_NEXT(10140, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10140(choice, ...) METALANG99_PRIV_REC_NEXT(10141, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10141(choice, ...) METALANG99_PRIV_REC_NEXT(10142, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10142(choice, ...) METALANG99_PRIV_REC_NEXT(10143, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10143(choice, ...) METALANG99_PRIV_REC_NEXT(10144, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10144(choice, ...) METALANG99_PRIV_REC_NEXT(10145, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10145(choice, ...) METALANG99_PRIV_REC_NEXT(10146, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10146(choice, ...) METALANG99_PRIV_REC_NEXT(10147, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10147(choice, ...) METALANG99_PRIV_REC_NEXT(10148, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10148(choice, ...) METALANG99_PRIV_REC_NEXT(10149, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10149(choice, ...) METALANG99_PRIV_REC_NEXT(10150, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10150(choice, ...) METALANG99_PRIV_REC_NEXT(10151, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10151(choice, ...) METALANG99_PRIV_REC_NEXT(10152, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10152(choice, ...) METALANG99_PRIV_REC_NEXT(10153, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10153(choice, ...) METALANG99_PRIV_REC_NEXT(10154, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10154(choice, ...) METALANG99_PRIV_REC_NEXT(10155, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10155(choice, ...) METALANG99_PRIV_REC_NEXT(10156, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10156(choice, ...) METALANG99_PRIV_REC_NEXT(10157, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10157(choice, ...) METALANG99_PRIV_REC_NEXT(10158, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10158(choice, ...) METALANG99_PRIV_REC_NEXT(10159, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10159(choice, ...) METALANG99_PRIV_REC_NEXT(10160, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10160(choice, ...) METALANG99_PRIV_REC_NEXT(10161, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10161(choice, ...) METALANG99_PRIV_REC_NEXT(10162, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10162(choice, ...) METALANG99_PRIV_REC_NEXT(10163, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10163(choice, ...) METALANG99_PRIV_REC_NEXT(10164, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10164(choice, ...) METALANG99_PRIV_REC_NEXT(10165, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10165(choice, ...) METALANG99_PRIV_REC_NEXT(10166, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10166(choice, ...) METALANG99_PRIV_REC_NEXT(10167, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10167(choice, ...) METALANG99_PRIV_REC_NEXT(10168, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10168(choice, ...) METALANG99_PRIV_REC_NEXT(10169, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10169(choice, ...) METALANG99_PRIV_REC_NEXT(10170, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10170(choice, ...) METALANG99_PRIV_REC_NEXT(10171, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10171(choice, ...) METALANG99_PRIV_REC_NEXT(10172, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10172(choice, ...) METALANG99_PRIV_REC_NEXT(10173, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10173(choice, ...) METALANG99_PRIV_REC_NEXT(10174, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10174(choice, ...) METALANG99_PRIV_REC_NEXT(10175, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10175(choice, ...) METALANG99_PRIV_REC_NEXT(10176, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10176(choice, ...) METALANG99_PRIV_REC_NEXT(10177, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10177(choice, ...) METALANG99_PRIV_REC_NEXT(10178, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10178(choice, ...) METALANG99_PRIV_REC_NEXT(10179, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10179(choice, ...) METALANG99_PRIV_REC_NEXT(10180, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10180(choice, ...) METALANG99_PRIV_REC_NEXT(10181, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10181(choice, ...) METALANG99_PRIV_REC_NEXT(10182, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10182(choice, ...) METALANG99_PRIV_REC_NEXT(10183, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10183(choice, ...) METALANG99_PRIV_REC_NEXT(10184, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10184(choice, ...) METALANG99_PRIV_REC_NEXT(10185, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10185(choice, ...) METALANG99_PRIV_REC_NEXT(10186, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10186(choice, ...) METALANG99_PRIV_REC_NEXT(10187, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10187(choice, ...) METALANG99_PRIV_REC_NEXT(10188, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10188(choice, ...) METALANG99_PRIV_REC_NEXT(10189, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10189(choice, ...) METALANG99_PRIV_REC_NEXT(10190, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10190(choice, ...) METALANG99_PRIV_REC_NEXT(10191, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10191(choice, ...) METALANG99_PRIV_REC_NEXT(10192, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10192(choice, ...) METALANG99_PRIV_REC_NEXT(10193, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10193(choice, ...) METALANG99_PRIV_REC_NEXT(10194, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10194(choice, ...) METALANG99_PRIV_REC_NEXT(10195, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10195(choice, ...) METALANG99_PRIV_REC_NEXT(10196, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10196(choice, ...) METALANG99_PRIV_REC_NEXT(10197, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10197(choice, ...) METALANG99_PRIV_REC_NEXT(10198, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10198(choice, ...) METALANG99_PRIV_REC_NEXT(10199, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10199(choice, ...) METALANG99_PRIV_REC_NEXT(10200, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10200(choice, ...) METALANG99_PRIV_REC_NEXT(10201, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10201(choice, ...) METALANG99_PRIV_REC_NEXT(10202, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10202(choice, ...) METALANG99_PRIV_REC_NEXT(10203, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10203(choice, ...) METALANG99_PRIV_REC_NEXT(10204, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10204(choice, ...) METALANG99_PRIV_REC_NEXT(10205, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10205(choice, ...) METALANG99_PRIV_REC_NEXT(10206, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10206(choice, ...) METALANG99_PRIV_REC_NEXT(10207, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10207(choice, ...) METALANG99_PRIV_REC_NEXT(10208, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10208(choice, ...) METALANG99_PRIV_REC_NEXT(10209, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10209(choice, ...) METALANG99_PRIV_REC_NEXT(10210, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10210(choice, ...) METALANG99_PRIV_REC_NEXT(10211, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10211(choice, ...) METALANG99_PRIV_REC_NEXT(10212, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10212(choice, ...) METALANG99_PRIV_REC_NEXT(10213, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10213(choice, ...) METALANG99_PRIV_REC_NEXT(10214, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10214(choice, ...) METALANG99_PRIV_REC_NEXT(10215, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10215(choice, ...) METALANG99_PRIV_REC_NEXT(10216, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10216(choice, ...) METALANG99_PRIV_REC_NEXT(10217, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10217(choice, ...) METALANG99_PRIV_REC_NEXT(10218, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10218(choice, ...) METALANG99_PRIV_REC_NEXT(10219, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10219(choice, ...) METALANG99_PRIV_REC_NEXT(10220, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10220(choice, ...) METALANG99_PRIV_REC_NEXT(10221, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10221(choice, ...) METALANG99_PRIV_REC_NEXT(10222, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10222(choice, ...) METALANG99_PRIV_REC_NEXT(10223, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10223(choice, ...) METALANG99_PRIV_REC_NEXT(10224, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10224(choice, ...) METALANG99_PRIV_REC_NEXT(10225, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10225(choice, ...) METALANG99_PRIV_REC_NEXT(10226, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10226(choice, ...) METALANG99_PRIV_REC_NEXT(10227, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10227(choice, ...) METALANG99_PRIV_REC_NEXT(10228, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10228(choice, ...) METALANG99_PRIV_REC_NEXT(10229, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10229(choice, ...) METALANG99_PRIV_REC_NEXT(10230, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10230(choice, ...) METALANG99_PRIV_REC_NEXT(10231, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10231(choice, ...) METALANG99_PRIV_REC_NEXT(10232, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10232(choice, ...) METALANG99_PRIV_REC_NEXT(10233, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10233(choice, ...) METALANG99_PRIV_REC_NEXT(10234, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10234(choice, ...) METALANG99_PRIV_REC_NEXT(10235, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10235(choice, ...) METALANG99_PRIV_REC_NEXT(10236, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10236(choice, ...) METALANG99_PRIV_REC_NEXT(10237, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10237(choice, ...) METALANG99_PRIV_REC_NEXT(10238, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10238(choice, ...) METALANG99_PRIV_REC_NEXT(10239, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10239(choice, ...) METALANG99_PRIV_REC_NEXT(10240, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10240(choice, ...) METALANG99_PRIV_REC_NEXT(10241, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10241(choice, ...) METALANG99_PRIV_REC_NEXT(10242, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10242(choice, ...) METALANG99_PRIV_REC_NEXT(10243, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10243(choice, ...) METALANG99_PRIV_REC_NEXT(10244, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10244(choice, ...) METALANG99_PRIV_REC_NEXT(10245, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10245(choice, ...) METALANG99_PRIV_REC_NEXT(10246, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10246(choice, ...) METALANG99_PRIV_REC_NEXT(10247, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10247(choice, ...) METALANG99_PRIV_REC_NEXT(10248, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10248(choice, ...) METALANG99_PRIV_REC_NEXT(10249, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10249(choice, ...) METALANG99_PRIV_REC_NEXT(10250, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10250(choice, ...) METALANG99_PRIV_REC_NEXT(10251, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10251(choice, ...) METALANG99_PRIV_REC_NEXT(10252, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10252(choice, ...) METALANG99_PRIV_REC_NEXT(10253, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10253(choice, ...) METALANG99_PRIV_REC_NEXT(10254, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10254(choice, ...) METALANG99_PRIV_REC_NEXT(10255, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10255(choice, ...) METALANG99_PRIV_REC_NEXT(10256, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10256(choice, ...) METALANG99_PRIV_REC_NEXT(10257, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10257(choice, ...) METALANG99_PRIV_REC_NEXT(10258, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10258(choice, ...) METALANG99_PRIV_REC_NEXT(10259, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10259(choice, ...) METALANG99_PRIV_REC_NEXT(10260, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10260(choice, ...) METALANG99_PRIV_REC_NEXT(10261, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10261(choice, ...) METALANG99_PRIV_REC_NEXT(10262, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10262(choice, ...) METALANG99_PRIV_REC_NEXT(10263, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10263(choice, ...) METALANG99_PRIV_REC_NEXT(10264, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10264(choice, ...) METALANG99_PRIV_REC_NEXT(10265, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10265(choice, ...) METALANG99_PRIV_REC_NEXT(10266, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10266(choice, ...) METALANG99_PRIV_REC_NEXT(10267, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10267(choice, ...) METALANG99_PRIV_REC_NEXT(10268, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10268(choice, ...) METALANG99_PRIV_REC_NEXT(10269, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10269(choice, ...) METALANG99_PRIV_REC_NEXT(10270, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10270(choice, ...) METALANG99_PRIV_REC_NEXT(10271, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10271(choice, ...) METALANG99_PRIV_REC_NEXT(10272, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10272(choice, ...) METALANG99_PRIV_REC_NEXT(10273, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10273(choice, ...) METALANG99_PRIV_REC_NEXT(10274, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10274(choice, ...) METALANG99_PRIV_REC_NEXT(10275, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10275(choice, ...) METALANG99_PRIV_REC_NEXT(10276, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10276(choice, ...) METALANG99_PRIV_REC_NEXT(10277, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10277(choice, ...) METALANG99_PRIV_REC_NEXT(10278, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10278(choice, ...) METALANG99_PRIV_REC_NEXT(10279, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10279(choice, ...) METALANG99_PRIV_REC_NEXT(10280, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10280(choice, ...) METALANG99_PRIV_REC_NEXT(10281, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10281(choice, ...) METALANG99_PRIV_REC_NEXT(10282, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10282(choice, ...) METALANG99_PRIV_REC_NEXT(10283, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10283(choice, ...) METALANG99_PRIV_REC_NEXT(10284, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10284(choice, ...) METALANG99_PRIV_REC_NEXT(10285, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10285(choice, ...) METALANG99_PRIV_REC_NEXT(10286, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10286(choice, ...) METALANG99_PRIV_REC_NEXT(10287, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10287(choice, ...) METALANG99_PRIV_REC_NEXT(10288, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10288(choice, ...) METALANG99_PRIV_REC_NEXT(10289, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10289(choice, ...) METALANG99_PRIV_REC_NEXT(10290, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10290(choice, ...) METALANG99_PRIV_REC_NEXT(10291, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10291(choice, ...) METALANG99_PRIV_REC_NEXT(10292, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10292(choice, ...) METALANG99_PRIV_REC_NEXT(10293, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10293(choice, ...) METALANG99_PRIV_REC_NEXT(10294, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10294(choice, ...) METALANG99_PRIV_REC_NEXT(10295, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10295(choice, ...) METALANG99_PRIV_REC_NEXT(10296, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10296(choice, ...) METALANG99_PRIV_REC_NEXT(10297, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10297(choice, ...) METALANG99_PRIV_REC_NEXT(10298, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10298(choice, ...) METALANG99_PRIV_REC_NEXT(10299, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10299(choice, ...) METALANG99_PRIV_REC_NEXT(10300, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10300(choice, ...) METALANG99_PRIV_REC_NEXT(10301, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10301(choice, ...) METALANG99_PRIV_REC_NEXT(10302, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10302(choice, ...) METALANG99_PRIV_REC_NEXT(10303, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10303(choice, ...) METALANG99_PRIV_REC_NEXT(10304, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10304(choice, ...) METALANG99_PRIV_REC_NEXT(10305, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10305(choice, ...) METALANG99_PRIV_REC_NEXT(10306, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10306(choice, ...) METALANG99_PRIV_REC_NEXT(10307, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10307(choice, ...) METALANG99_PRIV_REC_NEXT(10308, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10308(choice, ...) METALANG99_PRIV_REC_NEXT(10309, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10309(choice, ...) METALANG99_PRIV_REC_NEXT(10310, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10310(choice, ...) METALANG99_PRIV_REC_NEXT(10311, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10311(choice, ...) METALANG99_PRIV_REC_NEXT(10312, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10312(choice, ...) METALANG99_PRIV_REC_NEXT(10313, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10313(choice, ...) METALANG99_PRIV_REC_NEXT(10314, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10314(choice, ...) METALANG99_PRIV_REC_NEXT(10315, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10315(choice, ...) METALANG99_PRIV_REC_NEXT(10316, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10316(choice, ...) METALANG99_PRIV_REC_NEXT(10317, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10317(choice, ...) METALANG99_PRIV_REC_NEXT(10318, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10318(choice, ...) METALANG99_PRIV_REC_NEXT(10319, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10319(choice, ...) METALANG99_PRIV_REC_NEXT(10320, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10320(choice, ...) METALANG99_PRIV_REC_NEXT(10321, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10321(choice, ...) METALANG99_PRIV_REC_NEXT(10322, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10322(choice, ...) METALANG99_PRIV_REC_NEXT(10323, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10323(choice, ...) METALANG99_PRIV_REC_NEXT(10324, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10324(choice, ...) METALANG99_PRIV_REC_NEXT(10325, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10325(choice, ...) METALANG99_PRIV_REC_NEXT(10326, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10326(choice, ...) METALANG99_PRIV_REC_NEXT(10327, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10327(choice, ...) METALANG99_PRIV_REC_NEXT(10328, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10328(choice, ...) METALANG99_PRIV_REC_NEXT(10329, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10329(choice, ...) METALANG99_PRIV_REC_NEXT(10330, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10330(choice, ...) METALANG99_PRIV_REC_NEXT(10331, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10331(choice, ...) METALANG99_PRIV_REC_NEXT(10332, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10332(choice, ...) METALANG99_PRIV_REC_NEXT(10333, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10333(choice, ...) METALANG99_PRIV_REC_NEXT(10334, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10334(choice, ...) METALANG99_PRIV_REC_NEXT(10335, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10335(choice, ...) METALANG99_PRIV_REC_NEXT(10336, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10336(choice, ...) METALANG99_PRIV_REC_NEXT(10337, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10337(choice, ...) METALANG99_PRIV_REC_NEXT(10338, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10338(choice, ...) METALANG99_PRIV_REC_NEXT(10339, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10339(choice, ...) METALANG99_PRIV_REC_NEXT(10340, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10340(choice, ...) METALANG99_PRIV_REC_NEXT(10341, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10341(choice, ...) METALANG99_PRIV_REC_NEXT(10342, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10342(choice, ...) METALANG99_PRIV_REC_NEXT(10343, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10343(choice, ...) METALANG99_PRIV_REC_NEXT(10344, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10344(choice, ...) METALANG99_PRIV_REC_NEXT(10345, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10345(choice, ...) METALANG99_PRIV_REC_NEXT(10346, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10346(choice, ...) METALANG99_PRIV_REC_NEXT(10347, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10347(choice, ...) METALANG99_PRIV_REC_NEXT(10348, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10348(choice, ...) METALANG99_PRIV_REC_NEXT(10349, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10349(choice, ...) METALANG99_PRIV_REC_NEXT(10350, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10350(choice, ...) METALANG99_PRIV_REC_NEXT(10351, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10351(choice, ...) METALANG99_PRIV_REC_NEXT(10352, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10352(choice, ...) METALANG99_PRIV_REC_NEXT(10353, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10353(choice, ...) METALANG99_PRIV_REC_NEXT(10354, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10354(choice, ...) METALANG99_PRIV_REC_NEXT(10355, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10355(choice, ...) METALANG99_PRIV_REC_NEXT(10356, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10356(choice, ...) METALANG99_PRIV_REC_NEXT(10357, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10357(choice, ...) METALANG99_PRIV_REC_NEXT(10358, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10358(choice, ...) METALANG99_PRIV_REC_NEXT(10359, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10359(choice, ...) METALANG99_PRIV_REC_NEXT(10360, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10360(choice, ...) METALANG99_PRIV_REC_NEXT(10361, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10361(choice, ...) METALANG99_PRIV_REC_NEXT(10362, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10362(choice, ...) METALANG99_PRIV_REC_NEXT(10363, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10363(choice, ...) METALANG99_PRIV_REC_NEXT(10364, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10364(choice, ...) METALANG99_PRIV_REC_NEXT(10365, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10365(choice, ...) METALANG99_PRIV_REC_NEXT(10366, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10366(choice, ...) METALANG99_PRIV_REC_NEXT(10367, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10367(choice, ...) METALANG99_PRIV_REC_NEXT(10368, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10368(choice, ...) METALANG99_PRIV_REC_NEXT(10369, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10369(choice, ...) METALANG99_PRIV_REC_NEXT(10370, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10370(choice, ...) METALANG99_PRIV_REC_NEXT(10371, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10371(choice, ...) METALANG99_PRIV_REC_NEXT(10372, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10372(choice, ...) METALANG99_PRIV_REC_NEXT(10373, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10373(choice, ...) METALANG99_PRIV_REC_NEXT(10374, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10374(choice, ...) METALANG99_PRIV_REC_NEXT(10375, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10375(choice, ...) METALANG99_PRIV_REC_NEXT(10376, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10376(choice, ...) METALANG99_PRIV_REC_NEXT(10377, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10377(choice, ...) METALANG99_PRIV_REC_NEXT(10378, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10378(choice, ...) METALANG99_PRIV_REC_NEXT(10379, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10379(choice, ...) METALANG99_PRIV_REC_NEXT(10380, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10380(choice, ...) METALANG99_PRIV_REC_NEXT(10381, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10381(choice, ...) METALANG99_PRIV_REC_NEXT(10382, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10382(choice, ...) METALANG99_PRIV_REC_NEXT(10383, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10383(choice, ...) METALANG99_PRIV_REC_NEXT(10384, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10384(choice, ...) METALANG99_PRIV_REC_NEXT(10385, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10385(choice, ...) METALANG99_PRIV_REC_NEXT(10386, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10386(choice, ...) METALANG99_PRIV_REC_NEXT(10387, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10387(choice, ...) METALANG99_PRIV_REC_NEXT(10388, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10388(choice, ...) METALANG99_PRIV_REC_NEXT(10389, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10389(choice, ...) METALANG99_PRIV_REC_NEXT(10390, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10390(choice, ...) METALANG99_PRIV_REC_NEXT(10391, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10391(choice, ...) METALANG99_PRIV_REC_NEXT(10392, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10392(choice, ...) METALANG99_PRIV_REC_NEXT(10393, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10393(choice, ...) METALANG99_PRIV_REC_NEXT(10394, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10394(choice, ...) METALANG99_PRIV_REC_NEXT(10395, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10395(choice, ...) METALANG99_PRIV_REC_NEXT(10396, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10396(choice, ...) METALANG99_PRIV_REC_NEXT(10397, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10397(choice, ...) METALANG99_PRIV_REC_NEXT(10398, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10398(choice, ...) METALANG99_PRIV_REC_NEXT(10399, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10399(choice, ...) METALANG99_PRIV_REC_NEXT(10400, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10400(choice, ...) METALANG99_PRIV_REC_NEXT(10401, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10401(choice, ...) METALANG99_PRIV_REC_NEXT(10402, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10402(choice, ...) METALANG99_PRIV_REC_NEXT(10403, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10403(choice, ...) METALANG99_PRIV_REC_NEXT(10404, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10404(choice, ...) METALANG99_PRIV_REC_NEXT(10405, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10405(choice, ...) METALANG99_PRIV_REC_NEXT(10406, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10406(choice, ...) METALANG99_PRIV_REC_NEXT(10407, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10407(choice, ...) METALANG99_PRIV_REC_NEXT(10408, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10408(choice, ...) METALANG99_PRIV_REC_NEXT(10409, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10409(choice, ...) METALANG99_PRIV_REC_NEXT(10410, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10410(choice, ...) METALANG99_PRIV_REC_NEXT(10411, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10411(choice, ...) METALANG99_PRIV_REC_NEXT(10412, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10412(choice, ...) METALANG99_PRIV_REC_NEXT(10413, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10413(choice, ...) METALANG99_PRIV_REC_NEXT(10414, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10414(choice, ...) METALANG99_PRIV_REC_NEXT(10415, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10415(choice, ...) METALANG99_PRIV_REC_NEXT(10416, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10416(choice, ...) METALANG99_PRIV_REC_NEXT(10417, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10417(choice, ...) METALANG99_PRIV_REC_NEXT(10418, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10418(choice, ...) METALANG99_PRIV_REC_NEXT(10419, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10419(choice, ...) METALANG99_PRIV_REC_NEXT(10420, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10420(choice, ...) METALANG99_PRIV_REC_NEXT(10421, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10421(choice, ...) METALANG99_PRIV_REC_NEXT(10422, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10422(choice, ...) METALANG99_PRIV_REC_NEXT(10423, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10423(choice, ...) METALANG99_PRIV_REC_NEXT(10424, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10424(choice, ...) METALANG99_PRIV_REC_NEXT(10425, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10425(choice, ...) METALANG99_PRIV_REC_NEXT(10426, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10426(choice, ...) METALANG99_PRIV_REC_NEXT(10427, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10427(choice, ...) METALANG99_PRIV_REC_NEXT(10428, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10428(choice, ...) METALANG99_PRIV_REC_NEXT(10429, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10429(choice, ...) METALANG99_PRIV_REC_NEXT(10430, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10430(choice, ...) METALANG99_PRIV_REC_NEXT(10431, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10431(choice, ...) METALANG99_PRIV_REC_NEXT(10432, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10432(choice, ...) METALANG99_PRIV_REC_NEXT(10433, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10433(choice, ...) METALANG99_PRIV_REC_NEXT(10434, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10434(choice, ...) METALANG99_PRIV_REC_NEXT(10435, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10435(choice, ...) METALANG99_PRIV_REC_NEXT(10436, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10436(choice, ...) METALANG99_PRIV_REC_NEXT(10437, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10437(choice, ...) METALANG99_PRIV_REC_NEXT(10438, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10438(choice, ...) METALANG99_PRIV_REC_NEXT(10439, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10439(choice, ...) METALANG99_PRIV_REC_NEXT(10440, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10440(choice, ...) METALANG99_PRIV_REC_NEXT(10441, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10441(choice, ...) METALANG99_PRIV_REC_NEXT(10442, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10442(choice, ...) METALANG99_PRIV_REC_NEXT(10443, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10443(choice, ...) METALANG99_PRIV_REC_NEXT(10444, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10444(choice, ...) METALANG99_PRIV_REC_NEXT(10445, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10445(choice, ...) METALANG99_PRIV_REC_NEXT(10446, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10446(choice, ...) METALANG99_PRIV_REC_NEXT(10447, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10447(choice, ...) METALANG99_PRIV_REC_NEXT(10448, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10448(choice, ...) METALANG99_PRIV_REC_NEXT(10449, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10449(choice, ...) METALANG99_PRIV_REC_NEXT(10450, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10450(choice, ...) METALANG99_PRIV_REC_NEXT(10451, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10451(choice, ...) METALANG99_PRIV_REC_NEXT(10452, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10452(choice, ...) METALANG99_PRIV_REC_NEXT(10453, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10453(choice, ...) METALANG99_PRIV_REC_NEXT(10454, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10454(choice, ...) METALANG99_PRIV_REC_NEXT(10455, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10455(choice, ...) METALANG99_PRIV_REC_NEXT(10456, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10456(choice, ...) METALANG99_PRIV_REC_NEXT(10457, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10457(choice, ...) METALANG99_PRIV_REC_NEXT(10458, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10458(choice, ...) METALANG99_PRIV_REC_NEXT(10459, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10459(choice, ...) METALANG99_PRIV_REC_NEXT(10460, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10460(choice, ...) METALANG99_PRIV_REC_NEXT(10461, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10461(choice, ...) METALANG99_PRIV_REC_NEXT(10462, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10462(choice, ...) METALANG99_PRIV_REC_NEXT(10463, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10463(choice, ...) METALANG99_PRIV_REC_NEXT(10464, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10464(choice, ...) METALANG99_PRIV_REC_NEXT(10465, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10465(choice, ...) METALANG99_PRIV_REC_NEXT(10466, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10466(choice, ...) METALANG99_PRIV_REC_NEXT(10467, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10467(choice, ...) METALANG99_PRIV_REC_NEXT(10468, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10468(choice, ...) METALANG99_PRIV_REC_NEXT(10469, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10469(choice, ...) METALANG99_PRIV_REC_NEXT(10470, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10470(choice, ...) METALANG99_PRIV_REC_NEXT(10471, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10471(choice, ...) METALANG99_PRIV_REC_NEXT(10472, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10472(choice, ...) METALANG99_PRIV_REC_NEXT(10473, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10473(choice, ...) METALANG99_PRIV_REC_NEXT(10474, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10474(choice, ...) METALANG99_PRIV_REC_NEXT(10475, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10475(choice, ...) METALANG99_PRIV_REC_NEXT(10476, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10476(choice, ...) METALANG99_PRIV_REC_NEXT(10477, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10477(choice, ...) METALANG99_PRIV_REC_NEXT(10478, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10478(choice, ...) METALANG99_PRIV_REC_NEXT(10479, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10479(choice, ...) METALANG99_PRIV_REC_NEXT(10480, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10480(choice, ...) METALANG99_PRIV_REC_NEXT(10481, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10481(choice, ...) METALANG99_PRIV_REC_NEXT(10482, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10482(choice, ...) METALANG99_PRIV_REC_NEXT(10483, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10483(choice, ...) METALANG99_PRIV_REC_NEXT(10484, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10484(choice, ...) METALANG99_PRIV_REC_NEXT(10485, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10485(choice, ...) METALANG99_PRIV_REC_NEXT(10486, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10486(choice, ...) METALANG99_PRIV_REC_NEXT(10487, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10487(choice, ...) METALANG99_PRIV_REC_NEXT(10488, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10488(choice, ...) METALANG99_PRIV_REC_NEXT(10489, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10489(choice, ...) METALANG99_PRIV_REC_NEXT(10490, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10490(choice, ...) METALANG99_PRIV_REC_NEXT(10491, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10491(choice, ...) METALANG99_PRIV_REC_NEXT(10492, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10492(choice, ...) METALANG99_PRIV_REC_NEXT(10493, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10493(choice, ...) METALANG99_PRIV_REC_NEXT(10494, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10494(choice, ...) METALANG99_PRIV_REC_NEXT(10495, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10495(choice, ...) METALANG99_PRIV_REC_NEXT(10496, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10496(choice, ...) METALANG99_PRIV_REC_NEXT(10497, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10497(choice, ...) METALANG99_PRIV_REC_NEXT(10498, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10498(choice, ...) METALANG99_PRIV_REC_NEXT(10499, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10499(choice, ...) METALANG99_PRIV_REC_NEXT(10500, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10500(choice, ...) METALANG99_PRIV_REC_NEXT(10501, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10501(choice, ...) METALANG99_PRIV_REC_NEXT(10502, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10502(choice, ...) METALANG99_PRIV_REC_NEXT(10503, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10503(choice, ...) METALANG99_PRIV_REC_NEXT(10504, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10504(choice, ...) METALANG99_PRIV_REC_NEXT(10505, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10505(choice, ...) METALANG99_PRIV_REC_NEXT(10506, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10506(choice, ...) METALANG99_PRIV_REC_NEXT(10507, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10507(choice, ...) METALANG99_PRIV_REC_NEXT(10508, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10508(choice, ...) METALANG99_PRIV_REC_NEXT(10509, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10509(choice, ...) METALANG99_PRIV_REC_NEXT(10510, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10510(choice, ...) METALANG99_PRIV_REC_NEXT(10511, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10511(choice, ...) METALANG99_PRIV_REC_NEXT(10512, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10512(choice, ...) METALANG99_PRIV_REC_NEXT(10513, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10513(choice, ...) METALANG99_PRIV_REC_NEXT(10514, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10514(choice, ...) METALANG99_PRIV_REC_NEXT(10515, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10515(choice, ...) METALANG99_PRIV_REC_NEXT(10516, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10516(choice, ...) METALANG99_PRIV_REC_NEXT(10517, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10517(choice, ...) METALANG99_PRIV_REC_NEXT(10518, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10518(choice, ...) METALANG99_PRIV_REC_NEXT(10519, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10519(choice, ...) METALANG99_PRIV_REC_NEXT(10520, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10520(choice, ...) METALANG99_PRIV_REC_NEXT(10521, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10521(choice, ...) METALANG99_PRIV_REC_NEXT(10522, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10522(choice, ...) METALANG99_PRIV_REC_NEXT(10523, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10523(choice, ...) METALANG99_PRIV_REC_NEXT(10524, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10524(choice, ...) METALANG99_PRIV_REC_NEXT(10525, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10525(choice, ...) METALANG99_PRIV_REC_NEXT(10526, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10526(choice, ...) METALANG99_PRIV_REC_NEXT(10527, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10527(choice, ...) METALANG99_PRIV_REC_NEXT(10528, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10528(choice, ...) METALANG99_PRIV_REC_NEXT(10529, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10529(choice, ...) METALANG99_PRIV_REC_NEXT(10530, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10530(choice, ...) METALANG99_PRIV_REC_NEXT(10531, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10531(choice, ...) METALANG99_PRIV_REC_NEXT(10532, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10532(choice, ...) METALANG99_PRIV_REC_NEXT(10533, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10533(choice, ...) METALANG99_PRIV_REC_NEXT(10534, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10534(choice, ...) METALANG99_PRIV_REC_NEXT(10535, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10535(choice, ...) METALANG99_PRIV_REC_NEXT(10536, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10536(choice, ...) METALANG99_PRIV_REC_NEXT(10537, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10537(choice, ...) METALANG99_PRIV_REC_NEXT(10538, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10538(choice, ...) METALANG99_PRIV_REC_NEXT(10539, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10539(choice, ...) METALANG99_PRIV_REC_NEXT(10540, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10540(choice, ...) METALANG99_PRIV_REC_NEXT(10541, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10541(choice, ...) METALANG99_PRIV_REC_NEXT(10542, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10542(choice, ...) METALANG99_PRIV_REC_NEXT(10543, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10543(choice, ...) METALANG99_PRIV_REC_NEXT(10544, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10544(choice, ...) METALANG99_PRIV_REC_NEXT(10545, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10545(choice, ...) METALANG99_PRIV_REC_NEXT(10546, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10546(choice, ...) METALANG99_PRIV_REC_NEXT(10547, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10547(choice, ...) METALANG99_PRIV_REC_NEXT(10548, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10548(choice, ...) METALANG99_PRIV_REC_NEXT(10549, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10549(choice, ...) METALANG99_PRIV_REC_NEXT(10550, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10550(choice, ...) METALANG99_PRIV_REC_NEXT(10551, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10551(choice, ...) METALANG99_PRIV_REC_NEXT(10552, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10552(choice, ...) METALANG99_PRIV_REC_NEXT(10553, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10553(choice, ...) METALANG99_PRIV_REC_NEXT(10554, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10554(choice, ...) METALANG99_PRIV_REC_NEXT(10555, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10555(choice, ...) METALANG99_PRIV_REC_NEXT(10556, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10556(choice, ...) METALANG99_PRIV_REC_NEXT(10557, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10557(choice, ...) METALANG99_PRIV_REC_NEXT(10558, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10558(choice, ...) METALANG99_PRIV_REC_NEXT(10559, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10559(choice, ...) METALANG99_PRIV_REC_NEXT(10560, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10560(choice, ...) METALANG99_PRIV_REC_NEXT(10561, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10561(choice, ...) METALANG99_PRIV_REC_NEXT(10562, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10562(choice, ...) METALANG99_PRIV_REC_NEXT(10563, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10563(choice, ...) METALANG99_PRIV_REC_NEXT(10564, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10564(choice, ...) METALANG99_PRIV_REC_NEXT(10565, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10565(choice, ...) METALANG99_PRIV_REC_NEXT(10566, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10566(choice, ...) METALANG99_PRIV_REC_NEXT(10567, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10567(choice, ...) METALANG99_PRIV_REC_NEXT(10568, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10568(choice, ...) METALANG99_PRIV_REC_NEXT(10569, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10569(choice, ...) METALANG99_PRIV_REC_NEXT(10570, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10570(choice, ...) METALANG99_PRIV_REC_NEXT(10571, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10571(choice, ...) METALANG99_PRIV_REC_NEXT(10572, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10572(choice, ...) METALANG99_PRIV_REC_NEXT(10573, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10573(choice, ...) METALANG99_PRIV_REC_NEXT(10574, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10574(choice, ...) METALANG99_PRIV_REC_NEXT(10575, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10575(choice, ...) METALANG99_PRIV_REC_NEXT(10576, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10576(choice, ...) METALANG99_PRIV_REC_NEXT(10577, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10577(choice, ...) METALANG99_PRIV_REC_NEXT(10578, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10578(choice, ...) METALANG99_PRIV_REC_NEXT(10579, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10579(choice, ...) METALANG99_PRIV_REC_NEXT(10580, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10580(choice, ...) METALANG99_PRIV_REC_NEXT(10581, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10581(choice, ...) METALANG99_PRIV_REC_NEXT(10582, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10582(choice, ...) METALANG99_PRIV_REC_NEXT(10583, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10583(choice, ...) METALANG99_PRIV_REC_NEXT(10584, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10584(choice, ...) METALANG99_PRIV_REC_NEXT(10585, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10585(choice, ...) METALANG99_PRIV_REC_NEXT(10586, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10586(choice, ...) METALANG99_PRIV_REC_NEXT(10587, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10587(choice, ...) METALANG99_PRIV_REC_NEXT(10588, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10588(choice, ...) METALANG99_PRIV_REC_NEXT(10589, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10589(choice, ...) METALANG99_PRIV_REC_NEXT(10590, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10590(choice, ...) METALANG99_PRIV_REC_NEXT(10591, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10591(choice, ...) METALANG99_PRIV_REC_NEXT(10592, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10592(choice, ...) METALANG99_PRIV_REC_NEXT(10593, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10593(choice, ...) METALANG99_PRIV_REC_NEXT(10594, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10594(choice, ...) METALANG99_PRIV_REC_NEXT(10595, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10595(choice, ...) METALANG99_PRIV_REC_NEXT(10596, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10596(choice, ...) METALANG99_PRIV_REC_NEXT(10597, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10597(choice, ...) METALANG99_PRIV_REC_NEXT(10598, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10598(choice, ...) METALANG99_PRIV_REC_NEXT(10599, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10599(choice, ...) METALANG99_PRIV_REC_NEXT(10600, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10600(choice, ...) METALANG99_PRIV_REC_NEXT(10601, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10601(choice, ...) METALANG99_PRIV_REC_NEXT(10602, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10602(choice, ...) METALANG99_PRIV_REC_NEXT(10603, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10603(choice, ...) METALANG99_PRIV_REC_NEXT(10604, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10604(choice, ...) METALANG99_PRIV_REC_NEXT(10605, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10605(choice, ...) METALANG99_PRIV_REC_NEXT(10606, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10606(choice, ...) METALANG99_PRIV_REC_NEXT(10607, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10607(choice, ...) METALANG99_PRIV_REC_NEXT(10608, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10608(choice, ...) METALANG99_PRIV_REC_NEXT(10609, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10609(choice, ...) METALANG99_PRIV_REC_NEXT(10610, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10610(choice, ...) METALANG99_PRIV_REC_NEXT(10611, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10611(choice, ...) METALANG99_PRIV_REC_NEXT(10612, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10612(choice, ...) METALANG99_PRIV_REC_NEXT(10613, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10613(choice, ...) METALANG99_PRIV_REC_NEXT(10614, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10614(choice, ...) METALANG99_PRIV_REC_NEXT(10615, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10615(choice, ...) METALANG99_PRIV_REC_NEXT(10616, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10616(choice, ...) METALANG99_PRIV_REC_NEXT(10617, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10617(choice, ...) METALANG99_PRIV_REC_NEXT(10618, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10618(choice, ...) METALANG99_PRIV_REC_NEXT(10619, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10619(choice, ...) METALANG99_PRIV_REC_NEXT(10620, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10620(choice, ...) METALANG99_PRIV_REC_NEXT(10621, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10621(choice, ...) METALANG99_PRIV_REC_NEXT(10622, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10622(choice, ...) METALANG99_PRIV_REC_NEXT(10623, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10623(choice, ...) METALANG99_PRIV_REC_NEXT(10624, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10624(choice, ...) METALANG99_PRIV_REC_NEXT(10625, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10625(choice, ...) METALANG99_PRIV_REC_NEXT(10626, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10626(choice, ...) METALANG99_PRIV_REC_NEXT(10627, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10627(choice, ...) METALANG99_PRIV_REC_NEXT(10628, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10628(choice, ...) METALANG99_PRIV_REC_NEXT(10629, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10629(choice, ...) METALANG99_PRIV_REC_NEXT(10630, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10630(choice, ...) METALANG99_PRIV_REC_NEXT(10631, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10631(choice, ...) METALANG99_PRIV_REC_NEXT(10632, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10632(choice, ...) METALANG99_PRIV_REC_NEXT(10633, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10633(choice, ...) METALANG99_PRIV_REC_NEXT(10634, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10634(choice, ...) METALANG99_PRIV_REC_NEXT(10635, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10635(choice, ...) METALANG99_PRIV_REC_NEXT(10636, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10636(choice, ...) METALANG99_PRIV_REC_NEXT(10637, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10637(choice, ...) METALANG99_PRIV_REC_NEXT(10638, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10638(choice, ...) METALANG99_PRIV_REC_NEXT(10639, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10639(choice, ...) METALANG99_PRIV_REC_NEXT(10640, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10640(choice, ...) METALANG99_PRIV_REC_NEXT(10641, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10641(choice, ...) METALANG99_PRIV_REC_NEXT(10642, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10642(choice, ...) METALANG99_PRIV_REC_NEXT(10643, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10643(choice, ...) METALANG99_PRIV_REC_NEXT(10644, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10644(choice, ...) METALANG99_PRIV_REC_NEXT(10645, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10645(choice, ...) METALANG99_PRIV_REC_NEXT(10646, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10646(choice, ...) METALANG99_PRIV_REC_NEXT(10647, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10647(choice, ...) METALANG99_PRIV_REC_NEXT(10648, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10648(choice, ...) METALANG99_PRIV_REC_NEXT(10649, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10649(choice, ...) METALANG99_PRIV_REC_NEXT(10650, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10650(choice, ...) METALANG99_PRIV_REC_NEXT(10651, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10651(choice, ...) METALANG99_PRIV_REC_NEXT(10652, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10652(choice, ...) METALANG99_PRIV_REC_NEXT(10653, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10653(choice, ...) METALANG99_PRIV_REC_NEXT(10654, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10654(choice, ...) METALANG99_PRIV_REC_NEXT(10655, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10655(choice, ...) METALANG99_PRIV_REC_NEXT(10656, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10656(choice, ...) METALANG99_PRIV_REC_NEXT(10657, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10657(choice, ...) METALANG99_PRIV_REC_NEXT(10658, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10658(choice, ...) METALANG99_PRIV_REC_NEXT(10659, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10659(choice, ...) METALANG99_PRIV_REC_NEXT(10660, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10660(choice, ...) METALANG99_PRIV_REC_NEXT(10661, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10661(choice, ...) METALANG99_PRIV_REC_NEXT(10662, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10662(choice, ...) METALANG99_PRIV_REC_NEXT(10663, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10663(choice, ...) METALANG99_PRIV_REC_NEXT(10664, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10664(choice, ...) METALANG99_PRIV_REC_NEXT(10665, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10665(choice, ...) METALANG99_PRIV_REC_NEXT(10666, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10666(choice, ...) METALANG99_PRIV_REC_NEXT(10667, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10667(choice, ...) METALANG99_PRIV_REC_NEXT(10668, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10668(choice, ...) METALANG99_PRIV_REC_NEXT(10669, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10669(choice, ...) METALANG99_PRIV_REC_NEXT(10670, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10670(choice, ...) METALANG99_PRIV_REC_NEXT(10671, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10671(choice, ...) METALANG99_PRIV_REC_NEXT(10672, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10672(choice, ...) METALANG99_PRIV_REC_NEXT(10673, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10673(choice, ...) METALANG99_PRIV_REC_NEXT(10674, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10674(choice, ...) METALANG99_PRIV_REC_NEXT(10675, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10675(choice, ...) METALANG99_PRIV_REC_NEXT(10676, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10676(choice, ...) METALANG99_PRIV_REC_NEXT(10677, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10677(choice, ...) METALANG99_PRIV_REC_NEXT(10678, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10678(choice, ...) METALANG99_PRIV_REC_NEXT(10679, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10679(choice, ...) METALANG99_PRIV_REC_NEXT(10680, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10680(choice, ...) METALANG99_PRIV_REC_NEXT(10681, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10681(choice, ...) METALANG99_PRIV_REC_NEXT(10682, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10682(choice, ...) METALANG99_PRIV_REC_NEXT(10683, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10683(choice, ...) METALANG99_PRIV_REC_NEXT(10684, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10684(choice, ...) METALANG99_PRIV_REC_NEXT(10685, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10685(choice, ...) METALANG99_PRIV_REC_NEXT(10686, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10686(choice, ...) METALANG99_PRIV_REC_NEXT(10687, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10687(choice, ...) METALANG99_PRIV_REC_NEXT(10688, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10688(choice, ...) METALANG99_PRIV_REC_NEXT(10689, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10689(choice, ...) METALANG99_PRIV_REC_NEXT(10690, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10690(choice, ...) METALANG99_PRIV_REC_NEXT(10691, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10691(choice, ...) METALANG99_PRIV_REC_NEXT(10692, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10692(choice, ...) METALANG99_PRIV_REC_NEXT(10693, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10693(choice, ...) METALANG99_PRIV_REC_NEXT(10694, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10694(choice, ...) METALANG99_PRIV_REC_NEXT(10695, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10695(choice, ...) METALANG99_PRIV_REC_NEXT(10696, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10696(choice, ...) METALANG99_PRIV_REC_NEXT(10697, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10697(choice, ...) METALANG99_PRIV_REC_NEXT(10698, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10698(choice, ...) METALANG99_PRIV_REC_NEXT(10699, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10699(choice, ...) METALANG99_PRIV_REC_NEXT(10700, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10700(choice, ...) METALANG99_PRIV_REC_NEXT(10701, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10701(choice, ...) METALANG99_PRIV_REC_NEXT(10702, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10702(choice, ...) METALANG99_PRIV_REC_NEXT(10703, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10703(choice, ...) METALANG99_PRIV_REC_NEXT(10704, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10704(choice, ...) METALANG99_PRIV_REC_NEXT(10705, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10705(choice, ...) METALANG99_PRIV_REC_NEXT(10706, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10706(choice, ...) METALANG99_PRIV_REC_NEXT(10707, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10707(choice, ...) METALANG99_PRIV_REC_NEXT(10708, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10708(choice, ...) METALANG99_PRIV_REC_NEXT(10709, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10709(choice, ...) METALANG99_PRIV_REC_NEXT(10710, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10710(choice, ...) METALANG99_PRIV_REC_NEXT(10711, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10711(choice, ...) METALANG99_PRIV_REC_NEXT(10712, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10712(choice, ...) METALANG99_PRIV_REC_NEXT(10713, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10713(choice, ...) METALANG99_PRIV_REC_NEXT(10714, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10714(choice, ...) METALANG99_PRIV_REC_NEXT(10715, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10715(choice, ...) METALANG99_PRIV_REC_NEXT(10716, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10716(choice, ...) METALANG99_PRIV_REC_NEXT(10717, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10717(choice, ...) METALANG99_PRIV_REC_NEXT(10718, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10718(choice, ...) METALANG99_PRIV_REC_NEXT(10719, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10719(choice, ...) METALANG99_PRIV_REC_NEXT(10720, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10720(choice, ...) METALANG99_PRIV_REC_NEXT(10721, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10721(choice, ...) METALANG99_PRIV_REC_NEXT(10722, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10722(choice, ...) METALANG99_PRIV_REC_NEXT(10723, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10723(choice, ...) METALANG99_PRIV_REC_NEXT(10724, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10724(choice, ...) METALANG99_PRIV_REC_NEXT(10725, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10725(choice, ...) METALANG99_PRIV_REC_NEXT(10726, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10726(choice, ...) METALANG99_PRIV_REC_NEXT(10727, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10727(choice, ...) METALANG99_PRIV_REC_NEXT(10728, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10728(choice, ...) METALANG99_PRIV_REC_NEXT(10729, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10729(choice, ...) METALANG99_PRIV_REC_NEXT(10730, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10730(choice, ...) METALANG99_PRIV_REC_NEXT(10731, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10731(choice, ...) METALANG99_PRIV_REC_NEXT(10732, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10732(choice, ...) METALANG99_PRIV_REC_NEXT(10733, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10733(choice, ...) METALANG99_PRIV_REC_NEXT(10734, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10734(choice, ...) METALANG99_PRIV_REC_NEXT(10735, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10735(choice, ...) METALANG99_PRIV_REC_NEXT(10736, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10736(choice, ...) METALANG99_PRIV_REC_NEXT(10737, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10737(choice, ...) METALANG99_PRIV_REC_NEXT(10738, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10738(choice, ...) METALANG99_PRIV_REC_NEXT(10739, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10739(choice, ...) METALANG99_PRIV_REC_NEXT(10740, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10740(choice, ...) METALANG99_PRIV_REC_NEXT(10741, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10741(choice, ...) METALANG99_PRIV_REC_NEXT(10742, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10742(choice, ...) METALANG99_PRIV_REC_NEXT(10743, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10743(choice, ...) METALANG99_PRIV_REC_NEXT(10744, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10744(choice, ...) METALANG99_PRIV_REC_NEXT(10745, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10745(choice, ...) METALANG99_PRIV_REC_NEXT(10746, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10746(choice, ...) METALANG99_PRIV_REC_NEXT(10747, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10747(choice, ...) METALANG99_PRIV_REC_NEXT(10748, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10748(choice, ...) METALANG99_PRIV_REC_NEXT(10749, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10749(choice, ...) METALANG99_PRIV_REC_NEXT(10750, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10750(choice, ...) METALANG99_PRIV_REC_NEXT(10751, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10751(choice, ...) METALANG99_PRIV_REC_NEXT(10752, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10752(choice, ...) METALANG99_PRIV_REC_NEXT(10753, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10753(choice, ...) METALANG99_PRIV_REC_NEXT(10754, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10754(choice, ...) METALANG99_PRIV_REC_NEXT(10755, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10755(choice, ...) METALANG99_PRIV_REC_NEXT(10756, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10756(choice, ...) METALANG99_PRIV_REC_NEXT(10757, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10757(choice, ...) METALANG99_PRIV_REC_NEXT(10758, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10758(choice, ...) METALANG99_PRIV_REC_NEXT(10759, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10759(choice, ...) METALANG99_PRIV_REC_NEXT(10760, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10760(choice, ...) METALANG99_PRIV_REC_NEXT(10761, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10761(choice, ...) METALANG99_PRIV_REC_NEXT(10762, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10762(choice, ...) METALANG99_PRIV_REC_NEXT(10763, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10763(choice, ...) METALANG99_PRIV_REC_NEXT(10764, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10764(choice, ...) METALANG99_PRIV_REC_NEXT(10765, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10765(choice, ...) METALANG99_PRIV_REC_NEXT(10766, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10766(choice, ...) METALANG99_PRIV_REC_NEXT(10767, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10767(choice, ...) METALANG99_PRIV_REC_NEXT(10768, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10768(choice, ...) METALANG99_PRIV_REC_NEXT(10769, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10769(choice, ...) METALANG99_PRIV_REC_NEXT(10770, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10770(choice, ...) METALANG99_PRIV_REC_NEXT(10771, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10771(choice, ...) METALANG99_PRIV_REC_NEXT(10772, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10772(choice, ...) METALANG99_PRIV_REC_NEXT(10773, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10773(choice, ...) METALANG99_PRIV_REC_NEXT(10774, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10774(choice, ...) METALANG99_PRIV_REC_NEXT(10775, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10775(choice, ...) METALANG99_PRIV_REC_NEXT(10776, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10776(choice, ...) METALANG99_PRIV_REC_NEXT(10777, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10777(choice, ...) METALANG99_PRIV_REC_NEXT(10778, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10778(choice, ...) METALANG99_PRIV_REC_NEXT(10779, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10779(choice, ...) METALANG99_PRIV_REC_NEXT(10780, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10780(choice, ...) METALANG99_PRIV_REC_NEXT(10781, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10781(choice, ...) METALANG99_PRIV_REC_NEXT(10782, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10782(choice, ...) METALANG99_PRIV_REC_NEXT(10783, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10783(choice, ...) METALANG99_PRIV_REC_NEXT(10784, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10784(choice, ...) METALANG99_PRIV_REC_NEXT(10785, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10785(choice, ...) METALANG99_PRIV_REC_NEXT(10786, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10786(choice, ...) METALANG99_PRIV_REC_NEXT(10787, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10787(choice, ...) METALANG99_PRIV_REC_NEXT(10788, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10788(choice, ...) METALANG99_PRIV_REC_NEXT(10789, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10789(choice, ...) METALANG99_PRIV_REC_NEXT(10790, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10790(choice, ...) METALANG99_PRIV_REC_NEXT(10791, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10791(choice, ...) METALANG99_PRIV_REC_NEXT(10792, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10792(choice, ...) METALANG99_PRIV_REC_NEXT(10793, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10793(choice, ...) METALANG99_PRIV_REC_NEXT(10794, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10794(choice, ...) METALANG99_PRIV_REC_NEXT(10795, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10795(choice, ...) METALANG99_PRIV_REC_NEXT(10796, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10796(choice, ...) METALANG99_PRIV_REC_NEXT(10797, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10797(choice, ...) METALANG99_PRIV_REC_NEXT(10798, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10798(choice, ...) METALANG99_PRIV_REC_NEXT(10799, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10799(choice, ...) METALANG99_PRIV_REC_NEXT(10800, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10800(choice, ...) METALANG99_PRIV_REC_NEXT(10801, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10801(choice, ...) METALANG99_PRIV_REC_NEXT(10802, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10802(choice, ...) METALANG99_PRIV_REC_NEXT(10803, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10803(choice, ...) METALANG99_PRIV_REC_NEXT(10804, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10804(choice, ...) METALANG99_PRIV_REC_NEXT(10805, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10805(choice, ...) METALANG99_PRIV_REC_NEXT(10806, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10806(choice, ...) METALANG99_PRIV_REC_NEXT(10807, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10807(choice, ...) METALANG99_PRIV_REC_NEXT(10808, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10808(choice, ...) METALANG99_PRIV_REC_NEXT(10809, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10809(choice, ...) METALANG99_PRIV_REC_NEXT(10810, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10810(choice, ...) METALANG99_PRIV_REC_NEXT(10811, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10811(choice, ...) METALANG99_PRIV_REC_NEXT(10812, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10812(choice, ...) METALANG99_PRIV_REC_NEXT(10813, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10813(choice, ...) METALANG99_PRIV_REC_NEXT(10814, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10814(choice, ...) METALANG99_PRIV_REC_NEXT(10815, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10815(choice, ...) METALANG99_PRIV_REC_NEXT(10816, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10816(choice, ...) METALANG99_PRIV_REC_NEXT(10817, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10817(choice, ...) METALANG99_PRIV_REC_NEXT(10818, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10818(choice, ...) METALANG99_PRIV_REC_NEXT(10819, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10819(choice, ...) METALANG99_PRIV_REC_NEXT(10820, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10820(choice, ...) METALANG99_PRIV_REC_NEXT(10821, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10821(choice, ...) METALANG99_PRIV_REC_NEXT(10822, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10822(choice, ...) METALANG99_PRIV_REC_NEXT(10823, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10823(choice, ...) METALANG99_PRIV_REC_NEXT(10824, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10824(choice, ...) METALANG99_PRIV_REC_NEXT(10825, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10825(choice, ...) METALANG99_PRIV_REC_NEXT(10826, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10826(choice, ...) METALANG99_PRIV_REC_NEXT(10827, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10827(choice, ...) METALANG99_PRIV_REC_NEXT(10828, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10828(choice, ...) METALANG99_PRIV_REC_NEXT(10829, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10829(choice, ...) METALANG99_PRIV_REC_NEXT(10830, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10830(choice, ...) METALANG99_PRIV_REC_NEXT(10831, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10831(choice, ...) METALANG99_PRIV_REC_NEXT(10832, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10832(choice, ...) METALANG99_PRIV_REC_NEXT(10833, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10833(choice, ...) METALANG99_PRIV_REC_NEXT(10834, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10834(choice, ...) METALANG99_PRIV_REC_NEXT(10835, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10835(choice, ...) METALANG99_PRIV_REC_NEXT(10836, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10836(choice, ...) METALANG99_PRIV_REC_NEXT(10837, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10837(choice, ...) METALANG99_PRIV_REC_NEXT(10838, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10838(choice, ...) METALANG99_PRIV_REC_NEXT(10839, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10839(choice, ...) METALANG99_PRIV_REC_NEXT(10840, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10840(choice, ...) METALANG99_PRIV_REC_NEXT(10841, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10841(choice, ...) METALANG99_PRIV_REC_NEXT(10842, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10842(choice, ...) METALANG99_PRIV_REC_NEXT(10843, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10843(choice, ...) METALANG99_PRIV_REC_NEXT(10844, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10844(choice, ...) METALANG99_PRIV_REC_NEXT(10845, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10845(choice, ...) METALANG99_PRIV_REC_NEXT(10846, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10846(choice, ...) METALANG99_PRIV_REC_NEXT(10847, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10847(choice, ...) METALANG99_PRIV_REC_NEXT(10848, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10848(choice, ...) METALANG99_PRIV_REC_NEXT(10849, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10849(choice, ...) METALANG99_PRIV_REC_NEXT(10850, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10850(choice, ...) METALANG99_PRIV_REC_NEXT(10851, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10851(choice, ...) METALANG99_PRIV_REC_NEXT(10852, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10852(choice, ...) METALANG99_PRIV_REC_NEXT(10853, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10853(choice, ...) METALANG99_PRIV_REC_NEXT(10854, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10854(choice, ...) METALANG99_PRIV_REC_NEXT(10855, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10855(choice, ...) METALANG99_PRIV_REC_NEXT(10856, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10856(choice, ...) METALANG99_PRIV_REC_NEXT(10857, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10857(choice, ...) METALANG99_PRIV_REC_NEXT(10858, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10858(choice, ...) METALANG99_PRIV_REC_NEXT(10859, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10859(choice, ...) METALANG99_PRIV_REC_NEXT(10860, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10860(choice, ...) METALANG99_PRIV_REC_NEXT(10861, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10861(choice, ...) METALANG99_PRIV_REC_NEXT(10862, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10862(choice, ...) METALANG99_PRIV_REC_NEXT(10863, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10863(choice, ...) METALANG99_PRIV_REC_NEXT(10864, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10864(choice, ...) METALANG99_PRIV_REC_NEXT(10865, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10865(choice, ...) METALANG99_PRIV_REC_NEXT(10866, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10866(choice, ...) METALANG99_PRIV_REC_NEXT(10867, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10867(choice, ...) METALANG99_PRIV_REC_NEXT(10868, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10868(choice, ...) METALANG99_PRIV_REC_NEXT(10869, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10869(choice, ...) METALANG99_PRIV_REC_NEXT(10870, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10870(choice, ...) METALANG99_PRIV_REC_NEXT(10871, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10871(choice, ...) METALANG99_PRIV_REC_NEXT(10872, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10872(choice, ...) METALANG99_PRIV_REC_NEXT(10873, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10873(choice, ...) METALANG99_PRIV_REC_NEXT(10874, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10874(choice, ...) METALANG99_PRIV_REC_NEXT(10875, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10875(choice, ...) METALANG99_PRIV_REC_NEXT(10876, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10876(choice, ...) METALANG99_PRIV_REC_NEXT(10877, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10877(choice, ...) METALANG99_PRIV_REC_NEXT(10878, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10878(choice, ...) METALANG99_PRIV_REC_NEXT(10879, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10879(choice, ...) METALANG99_PRIV_REC_NEXT(10880, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10880(choice, ...) METALANG99_PRIV_REC_NEXT(10881, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10881(choice, ...) METALANG99_PRIV_REC_NEXT(10882, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10882(choice, ...) METALANG99_PRIV_REC_NEXT(10883, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10883(choice, ...) METALANG99_PRIV_REC_NEXT(10884, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10884(choice, ...) METALANG99_PRIV_REC_NEXT(10885, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10885(choice, ...) METALANG99_PRIV_REC_NEXT(10886, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10886(choice, ...) METALANG99_PRIV_REC_NEXT(10887, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10887(choice, ...) METALANG99_PRIV_REC_NEXT(10888, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10888(choice, ...) METALANG99_PRIV_REC_NEXT(10889, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10889(choice, ...) METALANG99_PRIV_REC_NEXT(10890, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10890(choice, ...) METALANG99_PRIV_REC_NEXT(10891, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10891(choice, ...) METALANG99_PRIV_REC_NEXT(10892, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10892(choice, ...) METALANG99_PRIV_REC_NEXT(10893, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10893(choice, ...) METALANG99_PRIV_REC_NEXT(10894, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10894(choice, ...) METALANG99_PRIV_REC_NEXT(10895, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10895(choice, ...) METALANG99_PRIV_REC_NEXT(10896, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10896(choice, ...) METALANG99_PRIV_REC_NEXT(10897, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10897(choice, ...) METALANG99_PRIV_REC_NEXT(10898, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10898(choice, ...) METALANG99_PRIV_REC_NEXT(10899, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10899(choice, ...) METALANG99_PRIV_REC_NEXT(10900, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10900(choice, ...) METALANG99_PRIV_REC_NEXT(10901, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10901(choice, ...) METALANG99_PRIV_REC_NEXT(10902, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10902(choice, ...) METALANG99_PRIV_REC_NEXT(10903, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10903(choice, ...) METALANG99_PRIV_REC_NEXT(10904, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10904(choice, ...) METALANG99_PRIV_REC_NEXT(10905, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10905(choice, ...) METALANG99_PRIV_REC_NEXT(10906, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10906(choice, ...) METALANG99_PRIV_REC_NEXT(10907, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10907(choice, ...) METALANG99_PRIV_REC_NEXT(10908, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10908(choice, ...) METALANG99_PRIV_REC_NEXT(10909, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10909(choice, ...) METALANG99_PRIV_REC_NEXT(10910, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10910(choice, ...) METALANG99_PRIV_REC_NEXT(10911, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10911(choice, ...) METALANG99_PRIV_REC_NEXT(10912, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10912(choice, ...) METALANG99_PRIV_REC_NEXT(10913, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10913(choice, ...) METALANG99_PRIV_REC_NEXT(10914, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10914(choice, ...) METALANG99_PRIV_REC_NEXT(10915, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10915(choice, ...) METALANG99_PRIV_REC_NEXT(10916, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10916(choice, ...) METALANG99_PRIV_REC_NEXT(10917, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10917(choice, ...) METALANG99_PRIV_REC_NEXT(10918, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10918(choice, ...) METALANG99_PRIV_REC_NEXT(10919, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10919(choice, ...) METALANG99_PRIV_REC_NEXT(10920, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10920(choice, ...) METALANG99_PRIV_REC_NEXT(10921, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10921(choice, ...) METALANG99_PRIV_REC_NEXT(10922, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10922(choice, ...) METALANG99_PRIV_REC_NEXT(10923, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10923(choice, ...) METALANG99_PRIV_REC_NEXT(10924, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10924(choice, ...) METALANG99_PRIV_REC_NEXT(10925, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10925(choice, ...) METALANG99_PRIV_REC_NEXT(10926, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10926(choice, ...) METALANG99_PRIV_REC_NEXT(10927, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10927(choice, ...) METALANG99_PRIV_REC_NEXT(10928, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10928(choice, ...) METALANG99_PRIV_REC_NEXT(10929, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10929(choice, ...) METALANG99_PRIV_REC_NEXT(10930, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10930(choice, ...) METALANG99_PRIV_REC_NEXT(10931, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10931(choice, ...) METALANG99_PRIV_REC_NEXT(10932, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10932(choice, ...) METALANG99_PRIV_REC_NEXT(10933, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10933(choice, ...) METALANG99_PRIV_REC_NEXT(10934, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10934(choice, ...) METALANG99_PRIV_REC_NEXT(10935, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10935(choice, ...) METALANG99_PRIV_REC_NEXT(10936, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10936(choice, ...) METALANG99_PRIV_REC_NEXT(10937, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10937(choice, ...) METALANG99_PRIV_REC_NEXT(10938, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10938(choice, ...) METALANG99_PRIV_REC_NEXT(10939, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10939(choice, ...) METALANG99_PRIV_REC_NEXT(10940, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10940(choice, ...) METALANG99_PRIV_REC_NEXT(10941, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10941(choice, ...) METALANG99_PRIV_REC_NEXT(10942, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10942(choice, ...) METALANG99_PRIV_REC_NEXT(10943, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10943(choice, ...) METALANG99_PRIV_REC_NEXT(10944, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10944(choice, ...) METALANG99_PRIV_REC_NEXT(10945, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10945(choice, ...) METALANG99_PRIV_REC_NEXT(10946, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10946(choice, ...) METALANG99_PRIV_REC_NEXT(10947, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10947(choice, ...) METALANG99_PRIV_REC_NEXT(10948, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10948(choice, ...) METALANG99_PRIV_REC_NEXT(10949, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10949(choice, ...) METALANG99_PRIV_REC_NEXT(10950, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10950(choice, ...) METALANG99_PRIV_REC_NEXT(10951, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10951(choice, ...) METALANG99_PRIV_REC_NEXT(10952, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10952(choice, ...) METALANG99_PRIV_REC_NEXT(10953, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10953(choice, ...) METALANG99_PRIV_REC_NEXT(10954, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10954(choice, ...) METALANG99_PRIV_REC_NEXT(10955, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10955(choice, ...) METALANG99_PRIV_REC_NEXT(10956, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10956(choice, ...) METALANG99_PRIV_REC_NEXT(10957, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10957(choice, ...) METALANG99_PRIV_REC_NEXT(10958, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10958(choice, ...) METALANG99_PRIV_REC_NEXT(10959, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10959(choice, ...) METALANG99_PRIV_REC_NEXT(10960, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10960(choice, ...) METALANG99_PRIV_REC_NEXT(10961, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10961(choice, ...) METALANG99_PRIV_REC_NEXT(10962, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10962(choice, ...) METALANG99_PRIV_REC_NEXT(10963, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10963(choice, ...) METALANG99_PRIV_REC_NEXT(10964, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10964(choice, ...) METALANG99_PRIV_REC_NEXT(10965, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10965(choice, ...) METALANG99_PRIV_REC_NEXT(10966, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10966(choice, ...) METALANG99_PRIV_REC_NEXT(10967, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10967(choice, ...) METALANG99_PRIV_REC_NEXT(10968, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10968(choice, ...) METALANG99_PRIV_REC_NEXT(10969, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10969(choice, ...) METALANG99_PRIV_REC_NEXT(10970, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10970(choice, ...) METALANG99_PRIV_REC_NEXT(10971, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10971(choice, ...) METALANG99_PRIV_REC_NEXT(10972, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10972(choice, ...) METALANG99_PRIV_REC_NEXT(10973, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10973(choice, ...) METALANG99_PRIV_REC_NEXT(10974, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10974(choice, ...) METALANG99_PRIV_REC_NEXT(10975, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10975(choice, ...) METALANG99_PRIV_REC_NEXT(10976, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10976(choice, ...) METALANG99_PRIV_REC_NEXT(10977, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10977(choice, ...) METALANG99_PRIV_REC_NEXT(10978, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10978(choice, ...) METALANG99_PRIV_REC_NEXT(10979, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10979(choice, ...) METALANG99_PRIV_REC_NEXT(10980, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10980(choice, ...) METALANG99_PRIV_REC_NEXT(10981, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10981(choice, ...) METALANG99_PRIV_REC_NEXT(10982, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10982(choice, ...) METALANG99_PRIV_REC_NEXT(10983, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10983(choice, ...) METALANG99_PRIV_REC_NEXT(10984, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10984(choice, ...) METALANG99_PRIV_REC_NEXT(10985, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10985(choice, ...) METALANG99_PRIV_REC_NEXT(10986, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10986(choice, ...) METALANG99_PRIV_REC_NEXT(10987, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10987(choice, ...) METALANG99_PRIV_REC_NEXT(10988, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10988(choice, ...) METALANG99_PRIV_REC_NEXT(10989, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10989(choice, ...) METALANG99_PRIV_REC_NEXT(10990, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10990(choice, ...) METALANG99_PRIV_REC_NEXT(10991, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10991(choice, ...) METALANG99_PRIV_REC_NEXT(10992, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10992(choice, ...) METALANG99_PRIV_REC_NEXT(10993, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10993(choice, ...) METALANG99_PRIV_REC_NEXT(10994, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10994(choice, ...) METALANG99_PRIV_REC_NEXT(10995, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10995(choice, ...) METALANG99_PRIV_REC_NEXT(10996, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10996(choice, ...) METALANG99_PRIV_REC_NEXT(10997, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10997(choice, ...) METALANG99_PRIV_REC_NEXT(10998, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10998(choice, ...) METALANG99_PRIV_REC_NEXT(10999, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_10999(choice, ...) METALANG99_PRIV_REC_NEXT(11000, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11000(choice, ...) METALANG99_PRIV_REC_NEXT(11001, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11001(choice, ...) METALANG99_PRIV_REC_NEXT(11002, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11002(choice, ...) METALANG99_PRIV_REC_NEXT(11003, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11003(choice, ...) METALANG99_PRIV_REC_NEXT(11004, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11004(choice, ...) METALANG99_PRIV_REC_NEXT(11005, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11005(choice, ...) METALANG99_PRIV_REC_NEXT(11006, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11006(choice, ...) METALANG99_PRIV_REC_NEXT(11007, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11007(choice, ...) METALANG99_PRIV_REC_NEXT(11008, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11008(choice, ...) METALANG99_PRIV_REC_NEXT(11009, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11009(choice, ...) METALANG99_PRIV_REC_NEXT(11010, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11010(choice, ...) METALANG99_PRIV_REC_NEXT(11011, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11011(choice, ...) METALANG99_PRIV_REC_NEXT(11012, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11012(choice, ...) METALANG99_PRIV_REC_NEXT(11013, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11013(choice, ...) METALANG99_PRIV_REC_NEXT(11014, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11014(choice, ...) METALANG99_PRIV_REC_NEXT(11015, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11015(choice, ...) METALANG99_PRIV_REC_NEXT(11016, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11016(choice, ...) METALANG99_PRIV_REC_NEXT(11017, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11017(choice, ...) METALANG99_PRIV_REC_NEXT(11018, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11018(choice, ...) METALANG99_PRIV_REC_NEXT(11019, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11019(choice, ...) METALANG99_PRIV_REC_NEXT(11020, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11020(choice, ...) METALANG99_PRIV_REC_NEXT(11021, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11021(choice, ...) METALANG99_PRIV_REC_NEXT(11022, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11022(choice, ...) METALANG99_PRIV_REC_NEXT(11023, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11023(choice, ...) METALANG99_PRIV_REC_NEXT(11024, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11024(choice, ...) METALANG99_PRIV_REC_NEXT(11025, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11025(choice, ...) METALANG99_PRIV_REC_NEXT(11026, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11026(choice, ...) METALANG99_PRIV_REC_NEXT(11027, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11027(choice, ...) METALANG99_PRIV_REC_NEXT(11028, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11028(choice, ...) METALANG99_PRIV_REC_NEXT(11029, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11029(choice, ...) METALANG99_PRIV_REC_NEXT(11030, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11030(choice, ...) METALANG99_PRIV_REC_NEXT(11031, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11031(choice, ...) METALANG99_PRIV_REC_NEXT(11032, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11032(choice, ...) METALANG99_PRIV_REC_NEXT(11033, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11033(choice, ...) METALANG99_PRIV_REC_NEXT(11034, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11034(choice, ...) METALANG99_PRIV_REC_NEXT(11035, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11035(choice, ...) METALANG99_PRIV_REC_NEXT(11036, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11036(choice, ...) METALANG99_PRIV_REC_NEXT(11037, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11037(choice, ...) METALANG99_PRIV_REC_NEXT(11038, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11038(choice, ...) METALANG99_PRIV_REC_NEXT(11039, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11039(choice, ...) METALANG99_PRIV_REC_NEXT(11040, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11040(choice, ...) METALANG99_PRIV_REC_NEXT(11041, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11041(choice, ...) METALANG99_PRIV_REC_NEXT(11042, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11042(choice, ...) METALANG99_PRIV_REC_NEXT(11043, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11043(choice, ...) METALANG99_PRIV_REC_NEXT(11044, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11044(choice, ...) METALANG99_PRIV_REC_NEXT(11045, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11045(choice, ...) METALANG99_PRIV_REC_NEXT(11046, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11046(choice, ...) METALANG99_PRIV_REC_NEXT(11047, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11047(choice, ...) METALANG99_PRIV_REC_NEXT(11048, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11048(choice, ...) METALANG99_PRIV_REC_NEXT(11049, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11049(choice, ...) METALANG99_PRIV_REC_NEXT(11050, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11050(choice, ...) METALANG99_PRIV_REC_NEXT(11051, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11051(choice, ...) METALANG99_PRIV_REC_NEXT(11052, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11052(choice, ...) METALANG99_PRIV_REC_NEXT(11053, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11053(choice, ...) METALANG99_PRIV_REC_NEXT(11054, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11054(choice, ...) METALANG99_PRIV_REC_NEXT(11055, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11055(choice, ...) METALANG99_PRIV_REC_NEXT(11056, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11056(choice, ...) METALANG99_PRIV_REC_NEXT(11057, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11057(choice, ...) METALANG99_PRIV_REC_NEXT(11058, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11058(choice, ...) METALANG99_PRIV_REC_NEXT(11059, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11059(choice, ...) METALANG99_PRIV_REC_NEXT(11060, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11060(choice, ...) METALANG99_PRIV_REC_NEXT(11061, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11061(choice, ...) METALANG99_PRIV_REC_NEXT(11062, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11062(choice, ...) METALANG99_PRIV_REC_NEXT(11063, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11063(choice, ...) METALANG99_PRIV_REC_NEXT(11064, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11064(choice, ...) METALANG99_PRIV_REC_NEXT(11065, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11065(choice, ...) METALANG99_PRIV_REC_NEXT(11066, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11066(choice, ...) METALANG99_PRIV_REC_NEXT(11067, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11067(choice, ...) METALANG99_PRIV_REC_NEXT(11068, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11068(choice, ...) METALANG99_PRIV_REC_NEXT(11069, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11069(choice, ...) METALANG99_PRIV_REC_NEXT(11070, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11070(choice, ...) METALANG99_PRIV_REC_NEXT(11071, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11071(choice, ...) METALANG99_PRIV_REC_NEXT(11072, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11072(choice, ...) METALANG99_PRIV_REC_NEXT(11073, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11073(choice, ...) METALANG99_PRIV_REC_NEXT(11074, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11074(choice, ...) METALANG99_PRIV_REC_NEXT(11075, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11075(choice, ...) METALANG99_PRIV_REC_NEXT(11076, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11076(choice, ...) METALANG99_PRIV_REC_NEXT(11077, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11077(choice, ...) METALANG99_PRIV_REC_NEXT(11078, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11078(choice, ...) METALANG99_PRIV_REC_NEXT(11079, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11079(choice, ...) METALANG99_PRIV_REC_NEXT(11080, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11080(choice, ...) METALANG99_PRIV_REC_NEXT(11081, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11081(choice, ...) METALANG99_PRIV_REC_NEXT(11082, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11082(choice, ...) METALANG99_PRIV_REC_NEXT(11083, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11083(choice, ...) METALANG99_PRIV_REC_NEXT(11084, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11084(choice, ...) METALANG99_PRIV_REC_NEXT(11085, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11085(choice, ...) METALANG99_PRIV_REC_NEXT(11086, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11086(choice, ...) METALANG99_PRIV_REC_NEXT(11087, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11087(choice, ...) METALANG99_PRIV_REC_NEXT(11088, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11088(choice, ...) METALANG99_PRIV_REC_NEXT(11089, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11089(choice, ...) METALANG99_PRIV_REC_NEXT(11090, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11090(choice, ...) METALANG99_PRIV_REC_NEXT(11091, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11091(choice, ...) METALANG99_PRIV_REC_NEXT(11092, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11092(choice, ...) METALANG99_PRIV_REC_NEXT(11093, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11093(choice, ...) METALANG99_PRIV_REC_NEXT(11094, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11094(choice, ...) METALANG99_PRIV_REC_NEXT(11095, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11095(choice, ...) METALANG99_PRIV_REC_NEXT(11096, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11096(choice, ...) METALANG99_PRIV_REC_NEXT(11097, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11097(choice, ...) METALANG99_PRIV_REC_NEXT(11098, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11098(choice, ...) METALANG99_PRIV_REC_NEXT(11099, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11099(choice, ...) METALANG99_PRIV_REC_NEXT(11100, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11100(choice, ...) METALANG99_PRIV_REC_NEXT(11101, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11101(choice, ...) METALANG99_PRIV_REC_NEXT(11102, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11102(choice, ...) METALANG99_PRIV_REC_NEXT(11103, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11103(choice, ...) METALANG99_PRIV_REC_NEXT(11104, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11104(choice, ...) METALANG99_PRIV_REC_NEXT(11105, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11105(choice, ...) METALANG99_PRIV_REC_NEXT(11106, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11106(choice, ...) METALANG99_PRIV_REC_NEXT(11107, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11107(choice, ...) METALANG99_PRIV_REC_NEXT(11108, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11108(choice, ...) METALANG99_PRIV_REC_NEXT(11109, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11109(choice, ...) METALANG99_PRIV_REC_NEXT(11110, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11110(choice, ...) METALANG99_PRIV_REC_NEXT(11111, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11111(choice, ...) METALANG99_PRIV_REC_NEXT(11112, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11112(choice, ...) METALANG99_PRIV_REC_NEXT(11113, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11113(choice, ...) METALANG99_PRIV_REC_NEXT(11114, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11114(choice, ...) METALANG99_PRIV_REC_NEXT(11115, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11115(choice, ...) METALANG99_PRIV_REC_NEXT(11116, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11116(choice, ...) METALANG99_PRIV_REC_NEXT(11117, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11117(choice, ...) METALANG99_PRIV_REC_NEXT(11118, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11118(choice, ...) METALANG99_PRIV_REC_NEXT(11119, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11119(choice, ...) METALANG99_PRIV_REC_NEXT(11120, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11120(choice, ...) METALANG99_PRIV_REC_NEXT(11121, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11121(choice, ...) METALANG99_PRIV_REC_NEXT(11122, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11122(choice, ...) METALANG99_PRIV_REC_NEXT(11123, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11123(choice, ...) METALANG99_PRIV_REC_NEXT(11124, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11124(choice, ...) METALANG99_PRIV_REC_NEXT(11125, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11125(choice, ...) METALANG99_PRIV_REC_NEXT(11126, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11126(choice, ...) METALANG99_PRIV_REC_NEXT(11127, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11127(choice, ...) METALANG99_PRIV_REC_NEXT(11128, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11128(choice, ...) METALANG99_PRIV_REC_NEXT(11129, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11129(choice, ...) METALANG99_PRIV_REC_NEXT(11130, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11130(choice, ...) METALANG99_PRIV_REC_NEXT(11131, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11131(choice, ...) METALANG99_PRIV_REC_NEXT(11132, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11132(choice, ...) METALANG99_PRIV_REC_NEXT(11133, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11133(choice, ...) METALANG99_PRIV_REC_NEXT(11134, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11134(choice, ...) METALANG99_PRIV_REC_NEXT(11135, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11135(choice, ...) METALANG99_PRIV_REC_NEXT(11136, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11136(choice, ...) METALANG99_PRIV_REC_NEXT(11137, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11137(choice, ...) METALANG99_PRIV_REC_NEXT(11138, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11138(choice, ...) METALANG99_PRIV_REC_NEXT(11139, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11139(choice, ...) METALANG99_PRIV_REC_NEXT(11140, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11140(choice, ...) METALANG99_PRIV_REC_NEXT(11141, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11141(choice, ...) METALANG99_PRIV_REC_NEXT(11142, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11142(choice, ...) METALANG99_PRIV_REC_NEXT(11143, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11143(choice, ...) METALANG99_PRIV_REC_NEXT(11144, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11144(choice, ...) METALANG99_PRIV_REC_NEXT(11145, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11145(choice, ...) METALANG99_PRIV_REC_NEXT(11146, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11146(choice, ...) METALANG99_PRIV_REC_NEXT(11147, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11147(choice, ...) METALANG99_PRIV_REC_NEXT(11148, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11148(choice, ...) METALANG99_PRIV_REC_NEXT(11149, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11149(choice, ...) METALANG99_PRIV_REC_NEXT(11150, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11150(choice, ...) METALANG99_PRIV_REC_NEXT(11151, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11151(choice, ...) METALANG99_PRIV_REC_NEXT(11152, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11152(choice, ...) METALANG99_PRIV_REC_NEXT(11153, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11153(choice, ...) METALANG99_PRIV_REC_NEXT(11154, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11154(choice, ...) METALANG99_PRIV_REC_NEXT(11155, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11155(choice, ...) METALANG99_PRIV_REC_NEXT(11156, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11156(choice, ...) METALANG99_PRIV_REC_NEXT(11157, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11157(choice, ...) METALANG99_PRIV_REC_NEXT(11158, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11158(choice, ...) METALANG99_PRIV_REC_NEXT(11159, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11159(choice, ...) METALANG99_PRIV_REC_NEXT(11160, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11160(choice, ...) METALANG99_PRIV_REC_NEXT(11161, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11161(choice, ...) METALANG99_PRIV_REC_NEXT(11162, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11162(choice, ...) METALANG99_PRIV_REC_NEXT(11163, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11163(choice, ...) METALANG99_PRIV_REC_NEXT(11164, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11164(choice, ...) METALANG99_PRIV_REC_NEXT(11165, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11165(choice, ...) METALANG99_PRIV_REC_NEXT(11166, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11166(choice, ...) METALANG99_PRIV_REC_NEXT(11167, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11167(choice, ...) METALANG99_PRIV_REC_NEXT(11168, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11168(choice, ...) METALANG99_PRIV_REC_NEXT(11169, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11169(choice, ...) METALANG99_PRIV_REC_NEXT(11170, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11170(choice, ...) METALANG99_PRIV_REC_NEXT(11171, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11171(choice, ...) METALANG99_PRIV_REC_NEXT(11172, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11172(choice, ...) METALANG99_PRIV_REC_NEXT(11173, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11173(choice, ...) METALANG99_PRIV_REC_NEXT(11174, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11174(choice, ...) METALANG99_PRIV_REC_NEXT(11175, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11175(choice, ...) METALANG99_PRIV_REC_NEXT(11176, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11176(choice, ...) METALANG99_PRIV_REC_NEXT(11177, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11177(choice, ...) METALANG99_PRIV_REC_NEXT(11178, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11178(choice, ...) METALANG99_PRIV_REC_NEXT(11179, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11179(choice, ...) METALANG99_PRIV_REC_NEXT(11180, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11180(choice, ...) METALANG99_PRIV_REC_NEXT(11181, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11181(choice, ...) METALANG99_PRIV_REC_NEXT(11182, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11182(choice, ...) METALANG99_PRIV_REC_NEXT(11183, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11183(choice, ...) METALANG99_PRIV_REC_NEXT(11184, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11184(choice, ...) METALANG99_PRIV_REC_NEXT(11185, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11185(choice, ...) METALANG99_PRIV_REC_NEXT(11186, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11186(choice, ...) METALANG99_PRIV_REC_NEXT(11187, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11187(choice, ...) METALANG99_PRIV_REC_NEXT(11188, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11188(choice, ...) METALANG99_PRIV_REC_NEXT(11189, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11189(choice, ...) METALANG99_PRIV_REC_NEXT(11190, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11190(choice, ...) METALANG99_PRIV_REC_NEXT(11191, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11191(choice, ...) METALANG99_PRIV_REC_NEXT(11192, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11192(choice, ...) METALANG99_PRIV_REC_NEXT(11193, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11193(choice, ...) METALANG99_PRIV_REC_NEXT(11194, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11194(choice, ...) METALANG99_PRIV_REC_NEXT(11195, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11195(choice, ...) METALANG99_PRIV_REC_NEXT(11196, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11196(choice, ...) METALANG99_PRIV_REC_NEXT(11197, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11197(choice, ...) METALANG99_PRIV_REC_NEXT(11198, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11198(choice, ...) METALANG99_PRIV_REC_NEXT(11199, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11199(choice, ...) METALANG99_PRIV_REC_NEXT(11200, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11200(choice, ...) METALANG99_PRIV_REC_NEXT(11201, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11201(choice, ...) METALANG99_PRIV_REC_NEXT(11202, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11202(choice, ...) METALANG99_PRIV_REC_NEXT(11203, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11203(choice, ...) METALANG99_PRIV_REC_NEXT(11204, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11204(choice, ...) METALANG99_PRIV_REC_NEXT(11205, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11205(choice, ...) METALANG99_PRIV_REC_NEXT(11206, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11206(choice, ...) METALANG99_PRIV_REC_NEXT(11207, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11207(choice, ...) METALANG99_PRIV_REC_NEXT(11208, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11208(choice, ...) METALANG99_PRIV_REC_NEXT(11209, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11209(choice, ...) METALANG99_PRIV_REC_NEXT(11210, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11210(choice, ...) METALANG99_PRIV_REC_NEXT(11211, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11211(choice, ...) METALANG99_PRIV_REC_NEXT(11212, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11212(choice, ...) METALANG99_PRIV_REC_NEXT(11213, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11213(choice, ...) METALANG99_PRIV_REC_NEXT(11214, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11214(choice, ...) METALANG99_PRIV_REC_NEXT(11215, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11215(choice, ...) METALANG99_PRIV_REC_NEXT(11216, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11216(choice, ...) METALANG99_PRIV_REC_NEXT(11217, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11217(choice, ...) METALANG99_PRIV_REC_NEXT(11218, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11218(choice, ...) METALANG99_PRIV_REC_NEXT(11219, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11219(choice, ...) METALANG99_PRIV_REC_NEXT(11220, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11220(choice, ...) METALANG99_PRIV_REC_NEXT(11221, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11221(choice, ...) METALANG99_PRIV_REC_NEXT(11222, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11222(choice, ...) METALANG99_PRIV_REC_NEXT(11223, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11223(choice, ...) METALANG99_PRIV_REC_NEXT(11224, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11224(choice, ...) METALANG99_PRIV_REC_NEXT(11225, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11225(choice, ...) METALANG99_PRIV_REC_NEXT(11226, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11226(choice, ...) METALANG99_PRIV_REC_NEXT(11227, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11227(choice, ...) METALANG99_PRIV_REC_NEXT(11228, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11228(choice, ...) METALANG99_PRIV_REC_NEXT(11229, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11229(choice, ...) METALANG99_PRIV_REC_NEXT(11230, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11230(choice, ...) METALANG99_PRIV_REC_NEXT(11231, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11231(choice, ...) METALANG99_PRIV_REC_NEXT(11232, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11232(choice, ...) METALANG99_PRIV_REC_NEXT(11233, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11233(choice, ...) METALANG99_PRIV_REC_NEXT(11234, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11234(choice, ...) METALANG99_PRIV_REC_NEXT(11235, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11235(choice, ...) METALANG99_PRIV_REC_NEXT(11236, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11236(choice, ...) METALANG99_PRIV_REC_NEXT(11237, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11237(choice, ...) METALANG99_PRIV_REC_NEXT(11238, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11238(choice, ...) METALANG99_PRIV_REC_NEXT(11239, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11239(choice, ...) METALANG99_PRIV_REC_NEXT(11240, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11240(choice, ...) METALANG99_PRIV_REC_NEXT(11241, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11241(choice, ...) METALANG99_PRIV_REC_NEXT(11242, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11242(choice, ...) METALANG99_PRIV_REC_NEXT(11243, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11243(choice, ...) METALANG99_PRIV_REC_NEXT(11244, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11244(choice, ...) METALANG99_PRIV_REC_NEXT(11245, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11245(choice, ...) METALANG99_PRIV_REC_NEXT(11246, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11246(choice, ...) METALANG99_PRIV_REC_NEXT(11247, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11247(choice, ...) METALANG99_PRIV_REC_NEXT(11248, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11248(choice, ...) METALANG99_PRIV_REC_NEXT(11249, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11249(choice, ...) METALANG99_PRIV_REC_NEXT(11250, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11250(choice, ...) METALANG99_PRIV_REC_NEXT(11251, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11251(choice, ...) METALANG99_PRIV_REC_NEXT(11252, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11252(choice, ...) METALANG99_PRIV_REC_NEXT(11253, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11253(choice, ...) METALANG99_PRIV_REC_NEXT(11254, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11254(choice, ...) METALANG99_PRIV_REC_NEXT(11255, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11255(choice, ...) METALANG99_PRIV_REC_NEXT(11256, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11256(choice, ...) METALANG99_PRIV_REC_NEXT(11257, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11257(choice, ...) METALANG99_PRIV_REC_NEXT(11258, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11258(choice, ...) METALANG99_PRIV_REC_NEXT(11259, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11259(choice, ...) METALANG99_PRIV_REC_NEXT(11260, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11260(choice, ...) METALANG99_PRIV_REC_NEXT(11261, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11261(choice, ...) METALANG99_PRIV_REC_NEXT(11262, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11262(choice, ...) METALANG99_PRIV_REC_NEXT(11263, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11263(choice, ...) METALANG99_PRIV_REC_NEXT(11264, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11264(choice, ...) METALANG99_PRIV_REC_NEXT(11265, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11265(choice, ...) METALANG99_PRIV_REC_NEXT(11266, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11266(choice, ...) METALANG99_PRIV_REC_NEXT(11267, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11267(choice, ...) METALANG99_PRIV_REC_NEXT(11268, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11268(choice, ...) METALANG99_PRIV_REC_NEXT(11269, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11269(choice, ...) METALANG99_PRIV_REC_NEXT(11270, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11270(choice, ...) METALANG99_PRIV_REC_NEXT(11271, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11271(choice, ...) METALANG99_PRIV_REC_NEXT(11272, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11272(choice, ...) METALANG99_PRIV_REC_NEXT(11273, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11273(choice, ...) METALANG99_PRIV_REC_NEXT(11274, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11274(choice, ...) METALANG99_PRIV_REC_NEXT(11275, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11275(choice, ...) METALANG99_PRIV_REC_NEXT(11276, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11276(choice, ...) METALANG99_PRIV_REC_NEXT(11277, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11277(choice, ...) METALANG99_PRIV_REC_NEXT(11278, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11278(choice, ...) METALANG99_PRIV_REC_NEXT(11279, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11279(choice, ...) METALANG99_PRIV_REC_NEXT(11280, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11280(choice, ...) METALANG99_PRIV_REC_NEXT(11281, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11281(choice, ...) METALANG99_PRIV_REC_NEXT(11282, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11282(choice, ...) METALANG99_PRIV_REC_NEXT(11283, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11283(choice, ...) METALANG99_PRIV_REC_NEXT(11284, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11284(choice, ...) METALANG99_PRIV_REC_NEXT(11285, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11285(choice, ...) METALANG99_PRIV_REC_NEXT(11286, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11286(choice, ...) METALANG99_PRIV_REC_NEXT(11287, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11287(choice, ...) METALANG99_PRIV_REC_NEXT(11288, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11288(choice, ...) METALANG99_PRIV_REC_NEXT(11289, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11289(choice, ...) METALANG99_PRIV_REC_NEXT(11290, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11290(choice, ...) METALANG99_PRIV_REC_NEXT(11291, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11291(choice, ...) METALANG99_PRIV_REC_NEXT(11292, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11292(choice, ...) METALANG99_PRIV_REC_NEXT(11293, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11293(choice, ...) METALANG99_PRIV_REC_NEXT(11294, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11294(choice, ...) METALANG99_PRIV_REC_NEXT(11295, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11295(choice, ...) METALANG99_PRIV_REC_NEXT(11296, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11296(choice, ...) METALANG99_PRIV_REC_NEXT(11297, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11297(choice, ...) METALANG99_PRIV_REC_NEXT(11298, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11298(choice, ...) METALANG99_PRIV_REC_NEXT(11299, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11299(choice, ...) METALANG99_PRIV_REC_NEXT(11300, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11300(choice, ...) METALANG99_PRIV_REC_NEXT(11301, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11301(choice, ...) METALANG99_PRIV_REC_NEXT(11302, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11302(choice, ...) METALANG99_PRIV_REC_NEXT(11303, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11303(choice, ...) METALANG99_PRIV_REC_NEXT(11304, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11304(choice, ...) METALANG99_PRIV_REC_NEXT(11305, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11305(choice, ...) METALANG99_PRIV_REC_NEXT(11306, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11306(choice, ...) METALANG99_PRIV_REC_NEXT(11307, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11307(choice, ...) METALANG99_PRIV_REC_NEXT(11308, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11308(choice, ...) METALANG99_PRIV_REC_NEXT(11309, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11309(choice, ...) METALANG99_PRIV_REC_NEXT(11310, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11310(choice, ...) METALANG99_PRIV_REC_NEXT(11311, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11311(choice, ...) METALANG99_PRIV_REC_NEXT(11312, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11312(choice, ...) METALANG99_PRIV_REC_NEXT(11313, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11313(choice, ...) METALANG99_PRIV_REC_NEXT(11314, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11314(choice, ...) METALANG99_PRIV_REC_NEXT(11315, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11315(choice, ...) METALANG99_PRIV_REC_NEXT(11316, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11316(choice, ...) METALANG99_PRIV_REC_NEXT(11317, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11317(choice, ...) METALANG99_PRIV_REC_NEXT(11318, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11318(choice, ...) METALANG99_PRIV_REC_NEXT(11319, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11319(choice, ...) METALANG99_PRIV_REC_NEXT(11320, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11320(choice, ...) METALANG99_PRIV_REC_NEXT(11321, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11321(choice, ...) METALANG99_PRIV_REC_NEXT(11322, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11322(choice, ...) METALANG99_PRIV_REC_NEXT(11323, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11323(choice, ...) METALANG99_PRIV_REC_NEXT(11324, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11324(choice, ...) METALANG99_PRIV_REC_NEXT(11325, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11325(choice, ...) METALANG99_PRIV_REC_NEXT(11326, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11326(choice, ...) METALANG99_PRIV_REC_NEXT(11327, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11327(choice, ...) METALANG99_PRIV_REC_NEXT(11328, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11328(choice, ...) METALANG99_PRIV_REC_NEXT(11329, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11329(choice, ...) METALANG99_PRIV_REC_NEXT(11330, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11330(choice, ...) METALANG99_PRIV_REC_NEXT(11331, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11331(choice, ...) METALANG99_PRIV_REC_NEXT(11332, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11332(choice, ...) METALANG99_PRIV_REC_NEXT(11333, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11333(choice, ...) METALANG99_PRIV_REC_NEXT(11334, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11334(choice, ...) METALANG99_PRIV_REC_NEXT(11335, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11335(choice, ...) METALANG99_PRIV_REC_NEXT(11336, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11336(choice, ...) METALANG99_PRIV_REC_NEXT(11337, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11337(choice, ...) METALANG99_PRIV_REC_NEXT(11338, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11338(choice, ...) METALANG99_PRIV_REC_NEXT(11339, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11339(choice, ...) METALANG99_PRIV_REC_NEXT(11340, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11340(choice, ...) METALANG99_PRIV_REC_NEXT(11341, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11341(choice, ...) METALANG99_PRIV_REC_NEXT(11342, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11342(choice, ...) METALANG99_PRIV_REC_NEXT(11343, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11343(choice, ...) METALANG99_PRIV_REC_NEXT(11344, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11344(choice, ...) METALANG99_PRIV_REC_NEXT(11345, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11345(choice, ...) METALANG99_PRIV_REC_NEXT(11346, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11346(choice, ...) METALANG99_PRIV_REC_NEXT(11347, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11347(choice, ...) METALANG99_PRIV_REC_NEXT(11348, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11348(choice, ...) METALANG99_PRIV_REC_NEXT(11349, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11349(choice, ...) METALANG99_PRIV_REC_NEXT(11350, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11350(choice, ...) METALANG99_PRIV_REC_NEXT(11351, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11351(choice, ...) METALANG99_PRIV_REC_NEXT(11352, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11352(choice, ...) METALANG99_PRIV_REC_NEXT(11353, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11353(choice, ...) METALANG99_PRIV_REC_NEXT(11354, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11354(choice, ...) METALANG99_PRIV_REC_NEXT(11355, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11355(choice, ...) METALANG99_PRIV_REC_NEXT(11356, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11356(choice, ...) METALANG99_PRIV_REC_NEXT(11357, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11357(choice, ...) METALANG99_PRIV_REC_NEXT(11358, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11358(choice, ...) METALANG99_PRIV_REC_NEXT(11359, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11359(choice, ...) METALANG99_PRIV_REC_NEXT(11360, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11360(choice, ...) METALANG99_PRIV_REC_NEXT(11361, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11361(choice, ...) METALANG99_PRIV_REC_NEXT(11362, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11362(choice, ...) METALANG99_PRIV_REC_NEXT(11363, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11363(choice, ...) METALANG99_PRIV_REC_NEXT(11364, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11364(choice, ...) METALANG99_PRIV_REC_NEXT(11365, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11365(choice, ...) METALANG99_PRIV_REC_NEXT(11366, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11366(choice, ...) METALANG99_PRIV_REC_NEXT(11367, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11367(choice, ...) METALANG99_PRIV_REC_NEXT(11368, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11368(choice, ...) METALANG99_PRIV_REC_NEXT(11369, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11369(choice, ...) METALANG99_PRIV_REC_NEXT(11370, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11370(choice, ...) METALANG99_PRIV_REC_NEXT(11371, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11371(choice, ...) METALANG99_PRIV_REC_NEXT(11372, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11372(choice, ...) METALANG99_PRIV_REC_NEXT(11373, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11373(choice, ...) METALANG99_PRIV_REC_NEXT(11374, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11374(choice, ...) METALANG99_PRIV_REC_NEXT(11375, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11375(choice, ...) METALANG99_PRIV_REC_NEXT(11376, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11376(choice, ...) METALANG99_PRIV_REC_NEXT(11377, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11377(choice, ...) METALANG99_PRIV_REC_NEXT(11378, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11378(choice, ...) METALANG99_PRIV_REC_NEXT(11379, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11379(choice, ...) METALANG99_PRIV_REC_NEXT(11380, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11380(choice, ...) METALANG99_PRIV_REC_NEXT(11381, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11381(choice, ...) METALANG99_PRIV_REC_NEXT(11382, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11382(choice, ...) METALANG99_PRIV_REC_NEXT(11383, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11383(choice, ...) METALANG99_PRIV_REC_NEXT(11384, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11384(choice, ...) METALANG99_PRIV_REC_NEXT(11385, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11385(choice, ...) METALANG99_PRIV_REC_NEXT(11386, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11386(choice, ...) METALANG99_PRIV_REC_NEXT(11387, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11387(choice, ...) METALANG99_PRIV_REC_NEXT(11388, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11388(choice, ...) METALANG99_PRIV_REC_NEXT(11389, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11389(choice, ...) METALANG99_PRIV_REC_NEXT(11390, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11390(choice, ...) METALANG99_PRIV_REC_NEXT(11391, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11391(choice, ...) METALANG99_PRIV_REC_NEXT(11392, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11392(choice, ...) METALANG99_PRIV_REC_NEXT(11393, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11393(choice, ...) METALANG99_PRIV_REC_NEXT(11394, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11394(choice, ...) METALANG99_PRIV_REC_NEXT(11395, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11395(choice, ...) METALANG99_PRIV_REC_NEXT(11396, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11396(choice, ...) METALANG99_PRIV_REC_NEXT(11397, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11397(choice, ...) METALANG99_PRIV_REC_NEXT(11398, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11398(choice, ...) METALANG99_PRIV_REC_NEXT(11399, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11399(choice, ...) METALANG99_PRIV_REC_NEXT(11400, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11400(choice, ...) METALANG99_PRIV_REC_NEXT(11401, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11401(choice, ...) METALANG99_PRIV_REC_NEXT(11402, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11402(choice, ...) METALANG99_PRIV_REC_NEXT(11403, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11403(choice, ...) METALANG99_PRIV_REC_NEXT(11404, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11404(choice, ...) METALANG99_PRIV_REC_NEXT(11405, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11405(choice, ...) METALANG99_PRIV_REC_NEXT(11406, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11406(choice, ...) METALANG99_PRIV_REC_NEXT(11407, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11407(choice, ...) METALANG99_PRIV_REC_NEXT(11408, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11408(choice, ...) METALANG99_PRIV_REC_NEXT(11409, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11409(choice, ...) METALANG99_PRIV_REC_NEXT(11410, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11410(choice, ...) METALANG99_PRIV_REC_NEXT(11411, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11411(choice, ...) METALANG99_PRIV_REC_NEXT(11412, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11412(choice, ...) METALANG99_PRIV_REC_NEXT(11413, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11413(choice, ...) METALANG99_PRIV_REC_NEXT(11414, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11414(choice, ...) METALANG99_PRIV_REC_NEXT(11415, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11415(choice, ...) METALANG99_PRIV_REC_NEXT(11416, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11416(choice, ...) METALANG99_PRIV_REC_NEXT(11417, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11417(choice, ...) METALANG99_PRIV_REC_NEXT(11418, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11418(choice, ...) METALANG99_PRIV_REC_NEXT(11419, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11419(choice, ...) METALANG99_PRIV_REC_NEXT(11420, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11420(choice, ...) METALANG99_PRIV_REC_NEXT(11421, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11421(choice, ...) METALANG99_PRIV_REC_NEXT(11422, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11422(choice, ...) METALANG99_PRIV_REC_NEXT(11423, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11423(choice, ...) METALANG99_PRIV_REC_NEXT(11424, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11424(choice, ...) METALANG99_PRIV_REC_NEXT(11425, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11425(choice, ...) METALANG99_PRIV_REC_NEXT(11426, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11426(choice, ...) METALANG99_PRIV_REC_NEXT(11427, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11427(choice, ...) METALANG99_PRIV_REC_NEXT(11428, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11428(choice, ...) METALANG99_PRIV_REC_NEXT(11429, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11429(choice, ...) METALANG99_PRIV_REC_NEXT(11430, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11430(choice, ...) METALANG99_PRIV_REC_NEXT(11431, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11431(choice, ...) METALANG99_PRIV_REC_NEXT(11432, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11432(choice, ...) METALANG99_PRIV_REC_NEXT(11433, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11433(choice, ...) METALANG99_PRIV_REC_NEXT(11434, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11434(choice, ...) METALANG99_PRIV_REC_NEXT(11435, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11435(choice, ...) METALANG99_PRIV_REC_NEXT(11436, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11436(choice, ...) METALANG99_PRIV_REC_NEXT(11437, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11437(choice, ...) METALANG99_PRIV_REC_NEXT(11438, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11438(choice, ...) METALANG99_PRIV_REC_NEXT(11439, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11439(choice, ...) METALANG99_PRIV_REC_NEXT(11440, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11440(choice, ...) METALANG99_PRIV_REC_NEXT(11441, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11441(choice, ...) METALANG99_PRIV_REC_NEXT(11442, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11442(choice, ...) METALANG99_PRIV_REC_NEXT(11443, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11443(choice, ...) METALANG99_PRIV_REC_NEXT(11444, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11444(choice, ...) METALANG99_PRIV_REC_NEXT(11445, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11445(choice, ...) METALANG99_PRIV_REC_NEXT(11446, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11446(choice, ...) METALANG99_PRIV_REC_NEXT(11447, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11447(choice, ...) METALANG99_PRIV_REC_NEXT(11448, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11448(choice, ...) METALANG99_PRIV_REC_NEXT(11449, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11449(choice, ...) METALANG99_PRIV_REC_NEXT(11450, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11450(choice, ...) METALANG99_PRIV_REC_NEXT(11451, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11451(choice, ...) METALANG99_PRIV_REC_NEXT(11452, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11452(choice, ...) METALANG99_PRIV_REC_NEXT(11453, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11453(choice, ...) METALANG99_PRIV_REC_NEXT(11454, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11454(choice, ...) METALANG99_PRIV_REC_NEXT(11455, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11455(choice, ...) METALANG99_PRIV_REC_NEXT(11456, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11456(choice, ...) METALANG99_PRIV_REC_NEXT(11457, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11457(choice, ...) METALANG99_PRIV_REC_NEXT(11458, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11458(choice, ...) METALANG99_PRIV_REC_NEXT(11459, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11459(choice, ...) METALANG99_PRIV_REC_NEXT(11460, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11460(choice, ...) METALANG99_PRIV_REC_NEXT(11461, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11461(choice, ...) METALANG99_PRIV_REC_NEXT(11462, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11462(choice, ...) METALANG99_PRIV_REC_NEXT(11463, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11463(choice, ...) METALANG99_PRIV_REC_NEXT(11464, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11464(choice, ...) METALANG99_PRIV_REC_NEXT(11465, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11465(choice, ...) METALANG99_PRIV_REC_NEXT(11466, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11466(choice, ...) METALANG99_PRIV_REC_NEXT(11467, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11467(choice, ...) METALANG99_PRIV_REC_NEXT(11468, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11468(choice, ...) METALANG99_PRIV_REC_NEXT(11469, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11469(choice, ...) METALANG99_PRIV_REC_NEXT(11470, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11470(choice, ...) METALANG99_PRIV_REC_NEXT(11471, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11471(choice, ...) METALANG99_PRIV_REC_NEXT(11472, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11472(choice, ...) METALANG99_PRIV_REC_NEXT(11473, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11473(choice, ...) METALANG99_PRIV_REC_NEXT(11474, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11474(choice, ...) METALANG99_PRIV_REC_NEXT(11475, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11475(choice, ...) METALANG99_PRIV_REC_NEXT(11476, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11476(choice, ...) METALANG99_PRIV_REC_NEXT(11477, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11477(choice, ...) METALANG99_PRIV_REC_NEXT(11478, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11478(choice, ...) METALANG99_PRIV_REC_NEXT(11479, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11479(choice, ...) METALANG99_PRIV_REC_NEXT(11480, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11480(choice, ...) METALANG99_PRIV_REC_NEXT(11481, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11481(choice, ...) METALANG99_PRIV_REC_NEXT(11482, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11482(choice, ...) METALANG99_PRIV_REC_NEXT(11483, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11483(choice, ...) METALANG99_PRIV_REC_NEXT(11484, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11484(choice, ...) METALANG99_PRIV_REC_NEXT(11485, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11485(choice, ...) METALANG99_PRIV_REC_NEXT(11486, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11486(choice, ...) METALANG99_PRIV_REC_NEXT(11487, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11487(choice, ...) METALANG99_PRIV_REC_NEXT(11488, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11488(choice, ...) METALANG99_PRIV_REC_NEXT(11489, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11489(choice, ...) METALANG99_PRIV_REC_NEXT(11490, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11490(choice, ...) METALANG99_PRIV_REC_NEXT(11491, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11491(choice, ...) METALANG99_PRIV_REC_NEXT(11492, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11492(choice, ...) METALANG99_PRIV_REC_NEXT(11493, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11493(choice, ...) METALANG99_PRIV_REC_NEXT(11494, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11494(choice, ...) METALANG99_PRIV_REC_NEXT(11495, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11495(choice, ...) METALANG99_PRIV_REC_NEXT(11496, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11496(choice, ...) METALANG99_PRIV_REC_NEXT(11497, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11497(choice, ...) METALANG99_PRIV_REC_NEXT(11498, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11498(choice, ...) METALANG99_PRIV_REC_NEXT(11499, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11499(choice, ...) METALANG99_PRIV_REC_NEXT(11500, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11500(choice, ...) METALANG99_PRIV_REC_NEXT(11501, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11501(choice, ...) METALANG99_PRIV_REC_NEXT(11502, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11502(choice, ...) METALANG99_PRIV_REC_NEXT(11503, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11503(choice, ...) METALANG99_PRIV_REC_NEXT(11504, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11504(choice, ...) METALANG99_PRIV_REC_NEXT(11505, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11505(choice, ...) METALANG99_PRIV_REC_NEXT(11506, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11506(choice, ...) METALANG99_PRIV_REC_NEXT(11507, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11507(choice, ...) METALANG99_PRIV_REC_NEXT(11508, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11508(choice, ...) METALANG99_PRIV_REC_NEXT(11509, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11509(choice, ...) METALANG99_PRIV_REC_NEXT(11510, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11510(choice, ...) METALANG99_PRIV_REC_NEXT(11511, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11511(choice, ...) METALANG99_PRIV_REC_NEXT(11512, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11512(choice, ...) METALANG99_PRIV_REC_NEXT(11513, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11513(choice, ...) METALANG99_PRIV_REC_NEXT(11514, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11514(choice, ...) METALANG99_PRIV_REC_NEXT(11515, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11515(choice, ...) METALANG99_PRIV_REC_NEXT(11516, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11516(choice, ...) METALANG99_PRIV_REC_NEXT(11517, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11517(choice, ...) METALANG99_PRIV_REC_NEXT(11518, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11518(choice, ...) METALANG99_PRIV_REC_NEXT(11519, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11519(choice, ...) METALANG99_PRIV_REC_NEXT(11520, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11520(choice, ...) METALANG99_PRIV_REC_NEXT(11521, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11521(choice, ...) METALANG99_PRIV_REC_NEXT(11522, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11522(choice, ...) METALANG99_PRIV_REC_NEXT(11523, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11523(choice, ...) METALANG99_PRIV_REC_NEXT(11524, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11524(choice, ...) METALANG99_PRIV_REC_NEXT(11525, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11525(choice, ...) METALANG99_PRIV_REC_NEXT(11526, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11526(choice, ...) METALANG99_PRIV_REC_NEXT(11527, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11527(choice, ...) METALANG99_PRIV_REC_NEXT(11528, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11528(choice, ...) METALANG99_PRIV_REC_NEXT(11529, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11529(choice, ...) METALANG99_PRIV_REC_NEXT(11530, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11530(choice, ...) METALANG99_PRIV_REC_NEXT(11531, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11531(choice, ...) METALANG99_PRIV_REC_NEXT(11532, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11532(choice, ...) METALANG99_PRIV_REC_NEXT(11533, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11533(choice, ...) METALANG99_PRIV_REC_NEXT(11534, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11534(choice, ...) METALANG99_PRIV_REC_NEXT(11535, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11535(choice, ...) METALANG99_PRIV_REC_NEXT(11536, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11536(choice, ...) METALANG99_PRIV_REC_NEXT(11537, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11537(choice, ...) METALANG99_PRIV_REC_NEXT(11538, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11538(choice, ...) METALANG99_PRIV_REC_NEXT(11539, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11539(choice, ...) METALANG99_PRIV_REC_NEXT(11540, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11540(choice, ...) METALANG99_PRIV_REC_NEXT(11541, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11541(choice, ...) METALANG99_PRIV_REC_NEXT(11542, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11542(choice, ...) METALANG99_PRIV_REC_NEXT(11543, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11543(choice, ...) METALANG99_PRIV_REC_NEXT(11544, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11544(choice, ...) METALANG99_PRIV_REC_NEXT(11545, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11545(choice, ...) METALANG99_PRIV_REC_NEXT(11546, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11546(choice, ...) METALANG99_PRIV_REC_NEXT(11547, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11547(choice, ...) METALANG99_PRIV_REC_NEXT(11548, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11548(choice, ...) METALANG99_PRIV_REC_NEXT(11549, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11549(choice, ...) METALANG99_PRIV_REC_NEXT(11550, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11550(choice, ...) METALANG99_PRIV_REC_NEXT(11551, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11551(choice, ...) METALANG99_PRIV_REC_NEXT(11552, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11552(choice, ...) METALANG99_PRIV_REC_NEXT(11553, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11553(choice, ...) METALANG99_PRIV_REC_NEXT(11554, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11554(choice, ...) METALANG99_PRIV_REC_NEXT(11555, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11555(choice, ...) METALANG99_PRIV_REC_NEXT(11556, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11556(choice, ...) METALANG99_PRIV_REC_NEXT(11557, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11557(choice, ...) METALANG99_PRIV_REC_NEXT(11558, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11558(choice, ...) METALANG99_PRIV_REC_NEXT(11559, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11559(choice, ...) METALANG99_PRIV_REC_NEXT(11560, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11560(choice, ...) METALANG99_PRIV_REC_NEXT(11561, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11561(choice, ...) METALANG99_PRIV_REC_NEXT(11562, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11562(choice, ...) METALANG99_PRIV_REC_NEXT(11563, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11563(choice, ...) METALANG99_PRIV_REC_NEXT(11564, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11564(choice, ...) METALANG99_PRIV_REC_NEXT(11565, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11565(choice, ...) METALANG99_PRIV_REC_NEXT(11566, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11566(choice, ...) METALANG99_PRIV_REC_NEXT(11567, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11567(choice, ...) METALANG99_PRIV_REC_NEXT(11568, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11568(choice, ...) METALANG99_PRIV_REC_NEXT(11569, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11569(choice, ...) METALANG99_PRIV_REC_NEXT(11570, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11570(choice, ...) METALANG99_PRIV_REC_NEXT(11571, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11571(choice, ...) METALANG99_PRIV_REC_NEXT(11572, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11572(choice, ...) METALANG99_PRIV_REC_NEXT(11573, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11573(choice, ...) METALANG99_PRIV_REC_NEXT(11574, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11574(choice, ...) METALANG99_PRIV_REC_NEXT(11575, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11575(choice, ...) METALANG99_PRIV_REC_NEXT(11576, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11576(choice, ...) METALANG99_PRIV_REC_NEXT(11577, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11577(choice, ...) METALANG99_PRIV_REC_NEXT(11578, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11578(choice, ...) METALANG99_PRIV_REC_NEXT(11579, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11579(choice, ...) METALANG99_PRIV_REC_NEXT(11580, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11580(choice, ...) METALANG99_PRIV_REC_NEXT(11581, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11581(choice, ...) METALANG99_PRIV_REC_NEXT(11582, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11582(choice, ...) METALANG99_PRIV_REC_NEXT(11583, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11583(choice, ...) METALANG99_PRIV_REC_NEXT(11584, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11584(choice, ...) METALANG99_PRIV_REC_NEXT(11585, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11585(choice, ...) METALANG99_PRIV_REC_NEXT(11586, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11586(choice, ...) METALANG99_PRIV_REC_NEXT(11587, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11587(choice, ...) METALANG99_PRIV_REC_NEXT(11588, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11588(choice, ...) METALANG99_PRIV_REC_NEXT(11589, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11589(choice, ...) METALANG99_PRIV_REC_NEXT(11590, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11590(choice, ...) METALANG99_PRIV_REC_NEXT(11591, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11591(choice, ...) METALANG99_PRIV_REC_NEXT(11592, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11592(choice, ...) METALANG99_PRIV_REC_NEXT(11593, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11593(choice, ...) METALANG99_PRIV_REC_NEXT(11594, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11594(choice, ...) METALANG99_PRIV_REC_NEXT(11595, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11595(choice, ...) METALANG99_PRIV_REC_NEXT(11596, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11596(choice, ...) METALANG99_PRIV_REC_NEXT(11597, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11597(choice, ...) METALANG99_PRIV_REC_NEXT(11598, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11598(choice, ...) METALANG99_PRIV_REC_NEXT(11599, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11599(choice, ...) METALANG99_PRIV_REC_NEXT(11600, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11600(choice, ...) METALANG99_PRIV_REC_NEXT(11601, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11601(choice, ...) METALANG99_PRIV_REC_NEXT(11602, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11602(choice, ...) METALANG99_PRIV_REC_NEXT(11603, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11603(choice, ...) METALANG99_PRIV_REC_NEXT(11604, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11604(choice, ...) METALANG99_PRIV_REC_NEXT(11605, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11605(choice, ...) METALANG99_PRIV_REC_NEXT(11606, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11606(choice, ...) METALANG99_PRIV_REC_NEXT(11607, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11607(choice, ...) METALANG99_PRIV_REC_NEXT(11608, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11608(choice, ...) METALANG99_PRIV_REC_NEXT(11609, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11609(choice, ...) METALANG99_PRIV_REC_NEXT(11610, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11610(choice, ...) METALANG99_PRIV_REC_NEXT(11611, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11611(choice, ...) METALANG99_PRIV_REC_NEXT(11612, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11612(choice, ...) METALANG99_PRIV_REC_NEXT(11613, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11613(choice, ...) METALANG99_PRIV_REC_NEXT(11614, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11614(choice, ...) METALANG99_PRIV_REC_NEXT(11615, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11615(choice, ...) METALANG99_PRIV_REC_NEXT(11616, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11616(choice, ...) METALANG99_PRIV_REC_NEXT(11617, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11617(choice, ...) METALANG99_PRIV_REC_NEXT(11618, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11618(choice, ...) METALANG99_PRIV_REC_NEXT(11619, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11619(choice, ...) METALANG99_PRIV_REC_NEXT(11620, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11620(choice, ...) METALANG99_PRIV_REC_NEXT(11621, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11621(choice, ...) METALANG99_PRIV_REC_NEXT(11622, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11622(choice, ...) METALANG99_PRIV_REC_NEXT(11623, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11623(choice, ...) METALANG99_PRIV_REC_NEXT(11624, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11624(choice, ...) METALANG99_PRIV_REC_NEXT(11625, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11625(choice, ...) METALANG99_PRIV_REC_NEXT(11626, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11626(choice, ...) METALANG99_PRIV_REC_NEXT(11627, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11627(choice, ...) METALANG99_PRIV_REC_NEXT(11628, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11628(choice, ...) METALANG99_PRIV_REC_NEXT(11629, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11629(choice, ...) METALANG99_PRIV_REC_NEXT(11630, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11630(choice, ...) METALANG99_PRIV_REC_NEXT(11631, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11631(choice, ...) METALANG99_PRIV_REC_NEXT(11632, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11632(choice, ...) METALANG99_PRIV_REC_NEXT(11633, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11633(choice, ...) METALANG99_PRIV_REC_NEXT(11634, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11634(choice, ...) METALANG99_PRIV_REC_NEXT(11635, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11635(choice, ...) METALANG99_PRIV_REC_NEXT(11636, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11636(choice, ...) METALANG99_PRIV_REC_NEXT(11637, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11637(choice, ...) METALANG99_PRIV_REC_NEXT(11638, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11638(choice, ...) METALANG99_PRIV_REC_NEXT(11639, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11639(choice, ...) METALANG99_PRIV_REC_NEXT(11640, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11640(choice, ...) METALANG99_PRIV_REC_NEXT(11641, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11641(choice, ...) METALANG99_PRIV_REC_NEXT(11642, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11642(choice, ...) METALANG99_PRIV_REC_NEXT(11643, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11643(choice, ...) METALANG99_PRIV_REC_NEXT(11644, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11644(choice, ...) METALANG99_PRIV_REC_NEXT(11645, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11645(choice, ...) METALANG99_PRIV_REC_NEXT(11646, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11646(choice, ...) METALANG99_PRIV_REC_NEXT(11647, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11647(choice, ...) METALANG99_PRIV_REC_NEXT(11648, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11648(choice, ...) METALANG99_PRIV_REC_NEXT(11649, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11649(choice, ...) METALANG99_PRIV_REC_NEXT(11650, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11650(choice, ...) METALANG99_PRIV_REC_NEXT(11651, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11651(choice, ...) METALANG99_PRIV_REC_NEXT(11652, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11652(choice, ...) METALANG99_PRIV_REC_NEXT(11653, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11653(choice, ...) METALANG99_PRIV_REC_NEXT(11654, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11654(choice, ...) METALANG99_PRIV_REC_NEXT(11655, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11655(choice, ...) METALANG99_PRIV_REC_NEXT(11656, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11656(choice, ...) METALANG99_PRIV_REC_NEXT(11657, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11657(choice, ...) METALANG99_PRIV_REC_NEXT(11658, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11658(choice, ...) METALANG99_PRIV_REC_NEXT(11659, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11659(choice, ...) METALANG99_PRIV_REC_NEXT(11660, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11660(choice, ...) METALANG99_PRIV_REC_NEXT(11661, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11661(choice, ...) METALANG99_PRIV_REC_NEXT(11662, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11662(choice, ...) METALANG99_PRIV_REC_NEXT(11663, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11663(choice, ...) METALANG99_PRIV_REC_NEXT(11664, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11664(choice, ...) METALANG99_PRIV_REC_NEXT(11665, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11665(choice, ...) METALANG99_PRIV_REC_NEXT(11666, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11666(choice, ...) METALANG99_PRIV_REC_NEXT(11667, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11667(choice, ...) METALANG99_PRIV_REC_NEXT(11668, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11668(choice, ...) METALANG99_PRIV_REC_NEXT(11669, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11669(choice, ...) METALANG99_PRIV_REC_NEXT(11670, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11670(choice, ...) METALANG99_PRIV_REC_NEXT(11671, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11671(choice, ...) METALANG99_PRIV_REC_NEXT(11672, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11672(choice, ...) METALANG99_PRIV_REC_NEXT(11673, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11673(choice, ...) METALANG99_PRIV_REC_NEXT(11674, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11674(choice, ...) METALANG99_PRIV_REC_NEXT(11675, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11675(choice, ...) METALANG99_PRIV_REC_NEXT(11676, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11676(choice, ...) METALANG99_PRIV_REC_NEXT(11677, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11677(choice, ...) METALANG99_PRIV_REC_NEXT(11678, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11678(choice, ...) METALANG99_PRIV_REC_NEXT(11679, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11679(choice, ...) METALANG99_PRIV_REC_NEXT(11680, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11680(choice, ...) METALANG99_PRIV_REC_NEXT(11681, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11681(choice, ...) METALANG99_PRIV_REC_NEXT(11682, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11682(choice, ...) METALANG99_PRIV_REC_NEXT(11683, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11683(choice, ...) METALANG99_PRIV_REC_NEXT(11684, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11684(choice, ...) METALANG99_PRIV_REC_NEXT(11685, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11685(choice, ...) METALANG99_PRIV_REC_NEXT(11686, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11686(choice, ...) METALANG99_PRIV_REC_NEXT(11687, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11687(choice, ...) METALANG99_PRIV_REC_NEXT(11688, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11688(choice, ...) METALANG99_PRIV_REC_NEXT(11689, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11689(choice, ...) METALANG99_PRIV_REC_NEXT(11690, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11690(choice, ...) METALANG99_PRIV_REC_NEXT(11691, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11691(choice, ...) METALANG99_PRIV_REC_NEXT(11692, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11692(choice, ...) METALANG99_PRIV_REC_NEXT(11693, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11693(choice, ...) METALANG99_PRIV_REC_NEXT(11694, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11694(choice, ...) METALANG99_PRIV_REC_NEXT(11695, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11695(choice, ...) METALANG99_PRIV_REC_NEXT(11696, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11696(choice, ...) METALANG99_PRIV_REC_NEXT(11697, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11697(choice, ...) METALANG99_PRIV_REC_NEXT(11698, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11698(choice, ...) METALANG99_PRIV_REC_NEXT(11699, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11699(choice, ...) METALANG99_PRIV_REC_NEXT(11700, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11700(choice, ...) METALANG99_PRIV_REC_NEXT(11701, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11701(choice, ...) METALANG99_PRIV_REC_NEXT(11702, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11702(choice, ...) METALANG99_PRIV_REC_NEXT(11703, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11703(choice, ...) METALANG99_PRIV_REC_NEXT(11704, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11704(choice, ...) METALANG99_PRIV_REC_NEXT(11705, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11705(choice, ...) METALANG99_PRIV_REC_NEXT(11706, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11706(choice, ...) METALANG99_PRIV_REC_NEXT(11707, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11707(choice, ...) METALANG99_PRIV_REC_NEXT(11708, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11708(choice, ...) METALANG99_PRIV_REC_NEXT(11709, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11709(choice, ...) METALANG99_PRIV_REC_NEXT(11710, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11710(choice, ...) METALANG99_PRIV_REC_NEXT(11711, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11711(choice, ...) METALANG99_PRIV_REC_NEXT(11712, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11712(choice, ...) METALANG99_PRIV_REC_NEXT(11713, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11713(choice, ...) METALANG99_PRIV_REC_NEXT(11714, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11714(choice, ...) METALANG99_PRIV_REC_NEXT(11715, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11715(choice, ...) METALANG99_PRIV_REC_NEXT(11716, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11716(choice, ...) METALANG99_PRIV_REC_NEXT(11717, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11717(choice, ...) METALANG99_PRIV_REC_NEXT(11718, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11718(choice, ...) METALANG99_PRIV_REC_NEXT(11719, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11719(choice, ...) METALANG99_PRIV_REC_NEXT(11720, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11720(choice, ...) METALANG99_PRIV_REC_NEXT(11721, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11721(choice, ...) METALANG99_PRIV_REC_NEXT(11722, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11722(choice, ...) METALANG99_PRIV_REC_NEXT(11723, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11723(choice, ...) METALANG99_PRIV_REC_NEXT(11724, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11724(choice, ...) METALANG99_PRIV_REC_NEXT(11725, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11725(choice, ...) METALANG99_PRIV_REC_NEXT(11726, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11726(choice, ...) METALANG99_PRIV_REC_NEXT(11727, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11727(choice, ...) METALANG99_PRIV_REC_NEXT(11728, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11728(choice, ...) METALANG99_PRIV_REC_NEXT(11729, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11729(choice, ...) METALANG99_PRIV_REC_NEXT(11730, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11730(choice, ...) METALANG99_PRIV_REC_NEXT(11731, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11731(choice, ...) METALANG99_PRIV_REC_NEXT(11732, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11732(choice, ...) METALANG99_PRIV_REC_NEXT(11733, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11733(choice, ...) METALANG99_PRIV_REC_NEXT(11734, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11734(choice, ...) METALANG99_PRIV_REC_NEXT(11735, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11735(choice, ...) METALANG99_PRIV_REC_NEXT(11736, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11736(choice, ...) METALANG99_PRIV_REC_NEXT(11737, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11737(choice, ...) METALANG99_PRIV_REC_NEXT(11738, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11738(choice, ...) METALANG99_PRIV_REC_NEXT(11739, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11739(choice, ...) METALANG99_PRIV_REC_NEXT(11740, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11740(choice, ...) METALANG99_PRIV_REC_NEXT(11741, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11741(choice, ...) METALANG99_PRIV_REC_NEXT(11742, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11742(choice, ...) METALANG99_PRIV_REC_NEXT(11743, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11743(choice, ...) METALANG99_PRIV_REC_NEXT(11744, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11744(choice, ...) METALANG99_PRIV_REC_NEXT(11745, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11745(choice, ...) METALANG99_PRIV_REC_NEXT(11746, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11746(choice, ...) METALANG99_PRIV_REC_NEXT(11747, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11747(choice, ...) METALANG99_PRIV_REC_NEXT(11748, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11748(choice, ...) METALANG99_PRIV_REC_NEXT(11749, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11749(choice, ...) METALANG99_PRIV_REC_NEXT(11750, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11750(choice, ...) METALANG99_PRIV_REC_NEXT(11751, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11751(choice, ...) METALANG99_PRIV_REC_NEXT(11752, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11752(choice, ...) METALANG99_PRIV_REC_NEXT(11753, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11753(choice, ...) METALANG99_PRIV_REC_NEXT(11754, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11754(choice, ...) METALANG99_PRIV_REC_NEXT(11755, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11755(choice, ...) METALANG99_PRIV_REC_NEXT(11756, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11756(choice, ...) METALANG99_PRIV_REC_NEXT(11757, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11757(choice, ...) METALANG99_PRIV_REC_NEXT(11758, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11758(choice, ...) METALANG99_PRIV_REC_NEXT(11759, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11759(choice, ...) METALANG99_PRIV_REC_NEXT(11760, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11760(choice, ...) METALANG99_PRIV_REC_NEXT(11761, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11761(choice, ...) METALANG99_PRIV_REC_NEXT(11762, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11762(choice, ...) METALANG99_PRIV_REC_NEXT(11763, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11763(choice, ...) METALANG99_PRIV_REC_NEXT(11764, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11764(choice, ...) METALANG99_PRIV_REC_NEXT(11765, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11765(choice, ...) METALANG99_PRIV_REC_NEXT(11766, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11766(choice, ...) METALANG99_PRIV_REC_NEXT(11767, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11767(choice, ...) METALANG99_PRIV_REC_NEXT(11768, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11768(choice, ...) METALANG99_PRIV_REC_NEXT(11769, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11769(choice, ...) METALANG99_PRIV_REC_NEXT(11770, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11770(choice, ...) METALANG99_PRIV_REC_NEXT(11771, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11771(choice, ...) METALANG99_PRIV_REC_NEXT(11772, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11772(choice, ...) METALANG99_PRIV_REC_NEXT(11773, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11773(choice, ...) METALANG99_PRIV_REC_NEXT(11774, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11774(choice, ...) METALANG99_PRIV_REC_NEXT(11775, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11775(choice, ...) METALANG99_PRIV_REC_NEXT(11776, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11776(choice, ...) METALANG99_PRIV_REC_NEXT(11777, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11777(choice, ...) METALANG99_PRIV_REC_NEXT(11778, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11778(choice, ...) METALANG99_PRIV_REC_NEXT(11779, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11779(choice, ...) METALANG99_PRIV_REC_NEXT(11780, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11780(choice, ...) METALANG99_PRIV_REC_NEXT(11781, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11781(choice, ...) METALANG99_PRIV_REC_NEXT(11782, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11782(choice, ...) METALANG99_PRIV_REC_NEXT(11783, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11783(choice, ...) METALANG99_PRIV_REC_NEXT(11784, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11784(choice, ...) METALANG99_PRIV_REC_NEXT(11785, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11785(choice, ...) METALANG99_PRIV_REC_NEXT(11786, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11786(choice, ...) METALANG99_PRIV_REC_NEXT(11787, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11787(choice, ...) METALANG99_PRIV_REC_NEXT(11788, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11788(choice, ...) METALANG99_PRIV_REC_NEXT(11789, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11789(choice, ...) METALANG99_PRIV_REC_NEXT(11790, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11790(choice, ...) METALANG99_PRIV_REC_NEXT(11791, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11791(choice, ...) METALANG99_PRIV_REC_NEXT(11792, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11792(choice, ...) METALANG99_PRIV_REC_NEXT(11793, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11793(choice, ...) METALANG99_PRIV_REC_NEXT(11794, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11794(choice, ...) METALANG99_PRIV_REC_NEXT(11795, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11795(choice, ...) METALANG99_PRIV_REC_NEXT(11796, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11796(choice, ...) METALANG99_PRIV_REC_NEXT(11797, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11797(choice, ...) METALANG99_PRIV_REC_NEXT(11798, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11798(choice, ...) METALANG99_PRIV_REC_NEXT(11799, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11799(choice, ...) METALANG99_PRIV_REC_NEXT(11800, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11800(choice, ...) METALANG99_PRIV_REC_NEXT(11801, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11801(choice, ...) METALANG99_PRIV_REC_NEXT(11802, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11802(choice, ...) METALANG99_PRIV_REC_NEXT(11803, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11803(choice, ...) METALANG99_PRIV_REC_NEXT(11804, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11804(choice, ...) METALANG99_PRIV_REC_NEXT(11805, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11805(choice, ...) METALANG99_PRIV_REC_NEXT(11806, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11806(choice, ...) METALANG99_PRIV_REC_NEXT(11807, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11807(choice, ...) METALANG99_PRIV_REC_NEXT(11808, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11808(choice, ...) METALANG99_PRIV_REC_NEXT(11809, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11809(choice, ...) METALANG99_PRIV_REC_NEXT(11810, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11810(choice, ...) METALANG99_PRIV_REC_NEXT(11811, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11811(choice, ...) METALANG99_PRIV_REC_NEXT(11812, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11812(choice, ...) METALANG99_PRIV_REC_NEXT(11813, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11813(choice, ...) METALANG99_PRIV_REC_NEXT(11814, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11814(choice, ...) METALANG99_PRIV_REC_NEXT(11815, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11815(choice, ...) METALANG99_PRIV_REC_NEXT(11816, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11816(choice, ...) METALANG99_PRIV_REC_NEXT(11817, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11817(choice, ...) METALANG99_PRIV_REC_NEXT(11818, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11818(choice, ...) METALANG99_PRIV_REC_NEXT(11819, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11819(choice, ...) METALANG99_PRIV_REC_NEXT(11820, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11820(choice, ...) METALANG99_PRIV_REC_NEXT(11821, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11821(choice, ...) METALANG99_PRIV_REC_NEXT(11822, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11822(choice, ...) METALANG99_PRIV_REC_NEXT(11823, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11823(choice, ...) METALANG99_PRIV_REC_NEXT(11824, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11824(choice, ...) METALANG99_PRIV_REC_NEXT(11825, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11825(choice, ...) METALANG99_PRIV_REC_NEXT(11826, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11826(choice, ...) METALANG99_PRIV_REC_NEXT(11827, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11827(choice, ...) METALANG99_PRIV_REC_NEXT(11828, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11828(choice, ...) METALANG99_PRIV_REC_NEXT(11829, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11829(choice, ...) METALANG99_PRIV_REC_NEXT(11830, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11830(choice, ...) METALANG99_PRIV_REC_NEXT(11831, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11831(choice, ...) METALANG99_PRIV_REC_NEXT(11832, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11832(choice, ...) METALANG99_PRIV_REC_NEXT(11833, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11833(choice, ...) METALANG99_PRIV_REC_NEXT(11834, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11834(choice, ...) METALANG99_PRIV_REC_NEXT(11835, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11835(choice, ...) METALANG99_PRIV_REC_NEXT(11836, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11836(choice, ...) METALANG99_PRIV_REC_NEXT(11837, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11837(choice, ...) METALANG99_PRIV_REC_NEXT(11838, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11838(choice, ...) METALANG99_PRIV_REC_NEXT(11839, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11839(choice, ...) METALANG99_PRIV_REC_NEXT(11840, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11840(choice, ...) METALANG99_PRIV_REC_NEXT(11841, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11841(choice, ...) METALANG99_PRIV_REC_NEXT(11842, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11842(choice, ...) METALANG99_PRIV_REC_NEXT(11843, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11843(choice, ...) METALANG99_PRIV_REC_NEXT(11844, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11844(choice, ...) METALANG99_PRIV_REC_NEXT(11845, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11845(choice, ...) METALANG99_PRIV_REC_NEXT(11846, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11846(choice, ...) METALANG99_PRIV_REC_NEXT(11847, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11847(choice, ...) METALANG99_PRIV_REC_NEXT(11848, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11848(choice, ...) METALANG99_PRIV_REC_NEXT(11849, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11849(choice, ...) METALANG99_PRIV_REC_NEXT(11850, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11850(choice, ...) METALANG99_PRIV_REC_NEXT(11851, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11851(choice, ...) METALANG99_PRIV_REC_NEXT(11852, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11852(choice, ...) METALANG99_PRIV_REC_NEXT(11853, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11853(choice, ...) METALANG99_PRIV_REC_NEXT(11854, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11854(choice, ...) METALANG99_PRIV_REC_NEXT(11855, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11855(choice, ...) METALANG99_PRIV_REC_NEXT(11856, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11856(choice, ...) METALANG99_PRIV_REC_NEXT(11857, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11857(choice, ...) METALANG99_PRIV_REC_NEXT(11858, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11858(choice, ...) METALANG99_PRIV_REC_NEXT(11859, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11859(choice, ...) METALANG99_PRIV_REC_NEXT(11860, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11860(choice, ...) METALANG99_PRIV_REC_NEXT(11861, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11861(choice, ...) METALANG99_PRIV_REC_NEXT(11862, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11862(choice, ...) METALANG99_PRIV_REC_NEXT(11863, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11863(choice, ...) METALANG99_PRIV_REC_NEXT(11864, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11864(choice, ...) METALANG99_PRIV_REC_NEXT(11865, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11865(choice, ...) METALANG99_PRIV_REC_NEXT(11866, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11866(choice, ...) METALANG99_PRIV_REC_NEXT(11867, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11867(choice, ...) METALANG99_PRIV_REC_NEXT(11868, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11868(choice, ...) METALANG99_PRIV_REC_NEXT(11869, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11869(choice, ...) METALANG99_PRIV_REC_NEXT(11870, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11870(choice, ...) METALANG99_PRIV_REC_NEXT(11871, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11871(choice, ...) METALANG99_PRIV_REC_NEXT(11872, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11872(choice, ...) METALANG99_PRIV_REC_NEXT(11873, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11873(choice, ...) METALANG99_PRIV_REC_NEXT(11874, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11874(choice, ...) METALANG99_PRIV_REC_NEXT(11875, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11875(choice, ...) METALANG99_PRIV_REC_NEXT(11876, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11876(choice, ...) METALANG99_PRIV_REC_NEXT(11877, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11877(choice, ...) METALANG99_PRIV_REC_NEXT(11878, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11878(choice, ...) METALANG99_PRIV_REC_NEXT(11879, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11879(choice, ...) METALANG99_PRIV_REC_NEXT(11880, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11880(choice, ...) METALANG99_PRIV_REC_NEXT(11881, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11881(choice, ...) METALANG99_PRIV_REC_NEXT(11882, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11882(choice, ...) METALANG99_PRIV_REC_NEXT(11883, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11883(choice, ...) METALANG99_PRIV_REC_NEXT(11884, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11884(choice, ...) METALANG99_PRIV_REC_NEXT(11885, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11885(choice, ...) METALANG99_PRIV_REC_NEXT(11886, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11886(choice, ...) METALANG99_PRIV_REC_NEXT(11887, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11887(choice, ...) METALANG99_PRIV_REC_NEXT(11888, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11888(choice, ...) METALANG99_PRIV_REC_NEXT(11889, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11889(choice, ...) METALANG99_PRIV_REC_NEXT(11890, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11890(choice, ...) METALANG99_PRIV_REC_NEXT(11891, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11891(choice, ...) METALANG99_PRIV_REC_NEXT(11892, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11892(choice, ...) METALANG99_PRIV_REC_NEXT(11893, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11893(choice, ...) METALANG99_PRIV_REC_NEXT(11894, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11894(choice, ...) METALANG99_PRIV_REC_NEXT(11895, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11895(choice, ...) METALANG99_PRIV_REC_NEXT(11896, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11896(choice, ...) METALANG99_PRIV_REC_NEXT(11897, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11897(choice, ...) METALANG99_PRIV_REC_NEXT(11898, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11898(choice, ...) METALANG99_PRIV_REC_NEXT(11899, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11899(choice, ...) METALANG99_PRIV_REC_NEXT(11900, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11900(choice, ...) METALANG99_PRIV_REC_NEXT(11901, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11901(choice, ...) METALANG99_PRIV_REC_NEXT(11902, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11902(choice, ...) METALANG99_PRIV_REC_NEXT(11903, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11903(choice, ...) METALANG99_PRIV_REC_NEXT(11904, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11904(choice, ...) METALANG99_PRIV_REC_NEXT(11905, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11905(choice, ...) METALANG99_PRIV_REC_NEXT(11906, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11906(choice, ...) METALANG99_PRIV_REC_NEXT(11907, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11907(choice, ...) METALANG99_PRIV_REC_NEXT(11908, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11908(choice, ...) METALANG99_PRIV_REC_NEXT(11909, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11909(choice, ...) METALANG99_PRIV_REC_NEXT(11910, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11910(choice, ...) METALANG99_PRIV_REC_NEXT(11911, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11911(choice, ...) METALANG99_PRIV_REC_NEXT(11912, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11912(choice, ...) METALANG99_PRIV_REC_NEXT(11913, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11913(choice, ...) METALANG99_PRIV_REC_NEXT(11914, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11914(choice, ...) METALANG99_PRIV_REC_NEXT(11915, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11915(choice, ...) METALANG99_PRIV_REC_NEXT(11916, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11916(choice, ...) METALANG99_PRIV_REC_NEXT(11917, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11917(choice, ...) METALANG99_PRIV_REC_NEXT(11918, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11918(choice, ...) METALANG99_PRIV_REC_NEXT(11919, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11919(choice, ...) METALANG99_PRIV_REC_NEXT(11920, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11920(choice, ...) METALANG99_PRIV_REC_NEXT(11921, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11921(choice, ...) METALANG99_PRIV_REC_NEXT(11922, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11922(choice, ...) METALANG99_PRIV_REC_NEXT(11923, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11923(choice, ...) METALANG99_PRIV_REC_NEXT(11924, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11924(choice, ...) METALANG99_PRIV_REC_NEXT(11925, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11925(choice, ...) METALANG99_PRIV_REC_NEXT(11926, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11926(choice, ...) METALANG99_PRIV_REC_NEXT(11927, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11927(choice, ...) METALANG99_PRIV_REC_NEXT(11928, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11928(choice, ...) METALANG99_PRIV_REC_NEXT(11929, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11929(choice, ...) METALANG99_PRIV_REC_NEXT(11930, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11930(choice, ...) METALANG99_PRIV_REC_NEXT(11931, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11931(choice, ...) METALANG99_PRIV_REC_NEXT(11932, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11932(choice, ...) METALANG99_PRIV_REC_NEXT(11933, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11933(choice, ...) METALANG99_PRIV_REC_NEXT(11934, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11934(choice, ...) METALANG99_PRIV_REC_NEXT(11935, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11935(choice, ...) METALANG99_PRIV_REC_NEXT(11936, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11936(choice, ...) METALANG99_PRIV_REC_NEXT(11937, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11937(choice, ...) METALANG99_PRIV_REC_NEXT(11938, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11938(choice, ...) METALANG99_PRIV_REC_NEXT(11939, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11939(choice, ...) METALANG99_PRIV_REC_NEXT(11940, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11940(choice, ...) METALANG99_PRIV_REC_NEXT(11941, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11941(choice, ...) METALANG99_PRIV_REC_NEXT(11942, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11942(choice, ...) METALANG99_PRIV_REC_NEXT(11943, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11943(choice, ...) METALANG99_PRIV_REC_NEXT(11944, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11944(choice, ...) METALANG99_PRIV_REC_NEXT(11945, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11945(choice, ...) METALANG99_PRIV_REC_NEXT(11946, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11946(choice, ...) METALANG99_PRIV_REC_NEXT(11947, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11947(choice, ...) METALANG99_PRIV_REC_NEXT(11948, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11948(choice, ...) METALANG99_PRIV_REC_NEXT(11949, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11949(choice, ...) METALANG99_PRIV_REC_NEXT(11950, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11950(choice, ...) METALANG99_PRIV_REC_NEXT(11951, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11951(choice, ...) METALANG99_PRIV_REC_NEXT(11952, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11952(choice, ...) METALANG99_PRIV_REC_NEXT(11953, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11953(choice, ...) METALANG99_PRIV_REC_NEXT(11954, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11954(choice, ...) METALANG99_PRIV_REC_NEXT(11955, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11955(choice, ...) METALANG99_PRIV_REC_NEXT(11956, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11956(choice, ...) METALANG99_PRIV_REC_NEXT(11957, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11957(choice, ...) METALANG99_PRIV_REC_NEXT(11958, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11958(choice, ...) METALANG99_PRIV_REC_NEXT(11959, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11959(choice, ...) METALANG99_PRIV_REC_NEXT(11960, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11960(choice, ...) METALANG99_PRIV_REC_NEXT(11961, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11961(choice, ...) METALANG99_PRIV_REC_NEXT(11962, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11962(choice, ...) METALANG99_PRIV_REC_NEXT(11963, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11963(choice, ...) METALANG99_PRIV_REC_NEXT(11964, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11964(choice, ...) METALANG99_PRIV_REC_NEXT(11965, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11965(choice, ...) METALANG99_PRIV_REC_NEXT(11966, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11966(choice, ...) METALANG99_PRIV_REC_NEXT(11967, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11967(choice, ...) METALANG99_PRIV_REC_NEXT(11968, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11968(choice, ...) METALANG99_PRIV_REC_NEXT(11969, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11969(choice, ...) METALANG99_PRIV_REC_NEXT(11970, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11970(choice, ...) METALANG99_PRIV_REC_NEXT(11971, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11971(choice, ...) METALANG99_PRIV_REC_NEXT(11972, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11972(choice, ...) METALANG99_PRIV_REC_NEXT(11973, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11973(choice, ...) METALANG99_PRIV_REC_NEXT(11974, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11974(choice, ...) METALANG99_PRIV_REC_NEXT(11975, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11975(choice, ...) METALANG99_PRIV_REC_NEXT(11976, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11976(choice, ...) METALANG99_PRIV_REC_NEXT(11977, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11977(choice, ...) METALANG99_PRIV_REC_NEXT(11978, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11978(choice, ...) METALANG99_PRIV_REC_NEXT(11979, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11979(choice, ...) METALANG99_PRIV_REC_NEXT(11980, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11980(choice, ...) METALANG99_PRIV_REC_NEXT(11981, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11981(choice, ...) METALANG99_PRIV_REC_NEXT(11982, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11982(choice, ...) METALANG99_PRIV_REC_NEXT(11983, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11983(choice, ...) METALANG99_PRIV_REC_NEXT(11984, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11984(choice, ...) METALANG99_PRIV_REC_NEXT(11985, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11985(choice, ...) METALANG99_PRIV_REC_NEXT(11986, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11986(choice, ...) METALANG99_PRIV_REC_NEXT(11987, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11987(choice, ...) METALANG99_PRIV_REC_NEXT(11988, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11988(choice, ...) METALANG99_PRIV_REC_NEXT(11989, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11989(choice, ...) METALANG99_PRIV_REC_NEXT(11990, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11990(choice, ...) METALANG99_PRIV_REC_NEXT(11991, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11991(choice, ...) METALANG99_PRIV_REC_NEXT(11992, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11992(choice, ...) METALANG99_PRIV_REC_NEXT(11993, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11993(choice, ...) METALANG99_PRIV_REC_NEXT(11994, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11994(choice, ...) METALANG99_PRIV_REC_NEXT(11995, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11995(choice, ...) METALANG99_PRIV_REC_NEXT(11996, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11996(choice, ...) METALANG99_PRIV_REC_NEXT(11997, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11997(choice, ...) METALANG99_PRIV_REC_NEXT(11998, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11998(choice, ...) METALANG99_PRIV_REC_NEXT(11999, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_11999(choice, ...) METALANG99_PRIV_REC_NEXT(12000, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12000(choice, ...) METALANG99_PRIV_REC_NEXT(12001, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12001(choice, ...) METALANG99_PRIV_REC_NEXT(12002, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12002(choice, ...) METALANG99_PRIV_REC_NEXT(12003, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12003(choice, ...) METALANG99_PRIV_REC_NEXT(12004, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12004(choice, ...) METALANG99_PRIV_REC_NEXT(12005, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12005(choice, ...) METALANG99_PRIV_REC_NEXT(12006, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12006(choice, ...) METALANG99_PRIV_REC_NEXT(12007, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12007(choice, ...) METALANG99_PRIV_REC_NEXT(12008, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12008(choice, ...) METALANG99_PRIV_REC_NEXT(12009, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12009(choice, ...) METALANG99_PRIV_REC_NEXT(12010, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12010(choice, ...) METALANG99_PRIV_REC_NEXT(12011, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12011(choice, ...) METALANG99_PRIV_REC_NEXT(12012, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12012(choice, ...) METALANG99_PRIV_REC_NEXT(12013, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12013(choice, ...) METALANG99_PRIV_REC_NEXT(12014, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12014(choice, ...) METALANG99_PRIV_REC_NEXT(12015, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12015(choice, ...) METALANG99_PRIV_REC_NEXT(12016, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12016(choice, ...) METALANG99_PRIV_REC_NEXT(12017, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12017(choice, ...) METALANG99_PRIV_REC_NEXT(12018, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12018(choice, ...) METALANG99_PRIV_REC_NEXT(12019, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12019(choice, ...) METALANG99_PRIV_REC_NEXT(12020, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12020(choice, ...) METALANG99_PRIV_REC_NEXT(12021, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12021(choice, ...) METALANG99_PRIV_REC_NEXT(12022, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12022(choice, ...) METALANG99_PRIV_REC_NEXT(12023, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12023(choice, ...) METALANG99_PRIV_REC_NEXT(12024, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12024(choice, ...) METALANG99_PRIV_REC_NEXT(12025, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12025(choice, ...) METALANG99_PRIV_REC_NEXT(12026, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12026(choice, ...) METALANG99_PRIV_REC_NEXT(12027, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12027(choice, ...) METALANG99_PRIV_REC_NEXT(12028, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12028(choice, ...) METALANG99_PRIV_REC_NEXT(12029, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12029(choice, ...) METALANG99_PRIV_REC_NEXT(12030, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12030(choice, ...) METALANG99_PRIV_REC_NEXT(12031, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12031(choice, ...) METALANG99_PRIV_REC_NEXT(12032, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12032(choice, ...) METALANG99_PRIV_REC_NEXT(12033, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12033(choice, ...) METALANG99_PRIV_REC_NEXT(12034, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12034(choice, ...) METALANG99_PRIV_REC_NEXT(12035, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12035(choice, ...) METALANG99_PRIV_REC_NEXT(12036, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12036(choice, ...) METALANG99_PRIV_REC_NEXT(12037, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12037(choice, ...) METALANG99_PRIV_REC_NEXT(12038, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12038(choice, ...) METALANG99_PRIV_REC_NEXT(12039, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12039(choice, ...) METALANG99_PRIV_REC_NEXT(12040, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12040(choice, ...) METALANG99_PRIV_REC_NEXT(12041, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12041(choice, ...) METALANG99_PRIV_REC_NEXT(12042, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12042(choice, ...) METALANG99_PRIV_REC_NEXT(12043, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12043(choice, ...) METALANG99_PRIV_REC_NEXT(12044, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12044(choice, ...) METALANG99_PRIV_REC_NEXT(12045, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12045(choice, ...) METALANG99_PRIV_REC_NEXT(12046, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12046(choice, ...) METALANG99_PRIV_REC_NEXT(12047, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12047(choice, ...) METALANG99_PRIV_REC_NEXT(12048, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12048(choice, ...) METALANG99_PRIV_REC_NEXT(12049, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12049(choice, ...) METALANG99_PRIV_REC_NEXT(12050, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12050(choice, ...) METALANG99_PRIV_REC_NEXT(12051, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12051(choice, ...) METALANG99_PRIV_REC_NEXT(12052, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12052(choice, ...) METALANG99_PRIV_REC_NEXT(12053, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12053(choice, ...) METALANG99_PRIV_REC_NEXT(12054, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12054(choice, ...) METALANG99_PRIV_REC_NEXT(12055, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12055(choice, ...) METALANG99_PRIV_REC_NEXT(12056, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12056(choice, ...) METALANG99_PRIV_REC_NEXT(12057, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12057(choice, ...) METALANG99_PRIV_REC_NEXT(12058, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12058(choice, ...) METALANG99_PRIV_REC_NEXT(12059, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12059(choice, ...) METALANG99_PRIV_REC_NEXT(12060, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12060(choice, ...) METALANG99_PRIV_REC_NEXT(12061, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12061(choice, ...) METALANG99_PRIV_REC_NEXT(12062, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12062(choice, ...) METALANG99_PRIV_REC_NEXT(12063, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12063(choice, ...) METALANG99_PRIV_REC_NEXT(12064, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12064(choice, ...) METALANG99_PRIV_REC_NEXT(12065, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12065(choice, ...) METALANG99_PRIV_REC_NEXT(12066, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12066(choice, ...) METALANG99_PRIV_REC_NEXT(12067, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12067(choice, ...) METALANG99_PRIV_REC_NEXT(12068, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12068(choice, ...) METALANG99_PRIV_REC_NEXT(12069, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12069(choice, ...) METALANG99_PRIV_REC_NEXT(12070, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12070(choice, ...) METALANG99_PRIV_REC_NEXT(12071, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12071(choice, ...) METALANG99_PRIV_REC_NEXT(12072, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12072(choice, ...) METALANG99_PRIV_REC_NEXT(12073, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12073(choice, ...) METALANG99_PRIV_REC_NEXT(12074, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12074(choice, ...) METALANG99_PRIV_REC_NEXT(12075, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12075(choice, ...) METALANG99_PRIV_REC_NEXT(12076, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12076(choice, ...) METALANG99_PRIV_REC_NEXT(12077, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12077(choice, ...) METALANG99_PRIV_REC_NEXT(12078, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12078(choice, ...) METALANG99_PRIV_REC_NEXT(12079, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12079(choice, ...) METALANG99_PRIV_REC_NEXT(12080, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12080(choice, ...) METALANG99_PRIV_REC_NEXT(12081, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12081(choice, ...) METALANG99_PRIV_REC_NEXT(12082, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12082(choice, ...) METALANG99_PRIV_REC_NEXT(12083, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12083(choice, ...) METALANG99_PRIV_REC_NEXT(12084, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12084(choice, ...) METALANG99_PRIV_REC_NEXT(12085, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12085(choice, ...) METALANG99_PRIV_REC_NEXT(12086, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12086(choice, ...) METALANG99_PRIV_REC_NEXT(12087, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12087(choice, ...) METALANG99_PRIV_REC_NEXT(12088, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12088(choice, ...) METALANG99_PRIV_REC_NEXT(12089, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12089(choice, ...) METALANG99_PRIV_REC_NEXT(12090, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12090(choice, ...) METALANG99_PRIV_REC_NEXT(12091, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12091(choice, ...) METALANG99_PRIV_REC_NEXT(12092, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12092(choice, ...) METALANG99_PRIV_REC_NEXT(12093, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12093(choice, ...) METALANG99_PRIV_REC_NEXT(12094, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12094(choice, ...) METALANG99_PRIV_REC_NEXT(12095, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12095(choice, ...) METALANG99_PRIV_REC_NEXT(12096, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12096(choice, ...) METALANG99_PRIV_REC_NEXT(12097, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12097(choice, ...) METALANG99_PRIV_REC_NEXT(12098, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12098(choice, ...) METALANG99_PRIV_REC_NEXT(12099, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12099(choice, ...) METALANG99_PRIV_REC_NEXT(12100, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12100(choice, ...) METALANG99_PRIV_REC_NEXT(12101, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12101(choice, ...) METALANG99_PRIV_REC_NEXT(12102, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12102(choice, ...) METALANG99_PRIV_REC_NEXT(12103, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12103(choice, ...) METALANG99_PRIV_REC_NEXT(12104, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12104(choice, ...) METALANG99_PRIV_REC_NEXT(12105, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12105(choice, ...) METALANG99_PRIV_REC_NEXT(12106, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12106(choice, ...) METALANG99_PRIV_REC_NEXT(12107, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12107(choice, ...) METALANG99_PRIV_REC_NEXT(12108, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12108(choice, ...) METALANG99_PRIV_REC_NEXT(12109, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12109(choice, ...) METALANG99_PRIV_REC_NEXT(12110, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12110(choice, ...) METALANG99_PRIV_REC_NEXT(12111, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12111(choice, ...) METALANG99_PRIV_REC_NEXT(12112, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12112(choice, ...) METALANG99_PRIV_REC_NEXT(12113, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12113(choice, ...) METALANG99_PRIV_REC_NEXT(12114, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12114(choice, ...) METALANG99_PRIV_REC_NEXT(12115, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12115(choice, ...) METALANG99_PRIV_REC_NEXT(12116, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12116(choice, ...) METALANG99_PRIV_REC_NEXT(12117, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12117(choice, ...) METALANG99_PRIV_REC_NEXT(12118, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12118(choice, ...) METALANG99_PRIV_REC_NEXT(12119, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12119(choice, ...) METALANG99_PRIV_REC_NEXT(12120, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12120(choice, ...) METALANG99_PRIV_REC_NEXT(12121, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12121(choice, ...) METALANG99_PRIV_REC_NEXT(12122, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12122(choice, ...) METALANG99_PRIV_REC_NEXT(12123, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12123(choice, ...) METALANG99_PRIV_REC_NEXT(12124, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12124(choice, ...) METALANG99_PRIV_REC_NEXT(12125, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12125(choice, ...) METALANG99_PRIV_REC_NEXT(12126, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12126(choice, ...) METALANG99_PRIV_REC_NEXT(12127, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12127(choice, ...) METALANG99_PRIV_REC_NEXT(12128, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12128(choice, ...) METALANG99_PRIV_REC_NEXT(12129, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12129(choice, ...) METALANG99_PRIV_REC_NEXT(12130, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12130(choice, ...) METALANG99_PRIV_REC_NEXT(12131, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12131(choice, ...) METALANG99_PRIV_REC_NEXT(12132, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12132(choice, ...) METALANG99_PRIV_REC_NEXT(12133, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12133(choice, ...) METALANG99_PRIV_REC_NEXT(12134, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12134(choice, ...) METALANG99_PRIV_REC_NEXT(12135, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12135(choice, ...) METALANG99_PRIV_REC_NEXT(12136, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12136(choice, ...) METALANG99_PRIV_REC_NEXT(12137, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12137(choice, ...) METALANG99_PRIV_REC_NEXT(12138, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12138(choice, ...) METALANG99_PRIV_REC_NEXT(12139, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12139(choice, ...) METALANG99_PRIV_REC_NEXT(12140, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12140(choice, ...) METALANG99_PRIV_REC_NEXT(12141, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12141(choice, ...) METALANG99_PRIV_REC_NEXT(12142, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12142(choice, ...) METALANG99_PRIV_REC_NEXT(12143, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12143(choice, ...) METALANG99_PRIV_REC_NEXT(12144, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12144(choice, ...) METALANG99_PRIV_REC_NEXT(12145, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12145(choice, ...) METALANG99_PRIV_REC_NEXT(12146, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12146(choice, ...) METALANG99_PRIV_REC_NEXT(12147, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12147(choice, ...) METALANG99_PRIV_REC_NEXT(12148, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12148(choice, ...) METALANG99_PRIV_REC_NEXT(12149, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12149(choice, ...) METALANG99_PRIV_REC_NEXT(12150, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12150(choice, ...) METALANG99_PRIV_REC_NEXT(12151, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12151(choice, ...) METALANG99_PRIV_REC_NEXT(12152, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12152(choice, ...) METALANG99_PRIV_REC_NEXT(12153, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12153(choice, ...) METALANG99_PRIV_REC_NEXT(12154, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12154(choice, ...) METALANG99_PRIV_REC_NEXT(12155, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12155(choice, ...) METALANG99_PRIV_REC_NEXT(12156, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12156(choice, ...) METALANG99_PRIV_REC_NEXT(12157, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12157(choice, ...) METALANG99_PRIV_REC_NEXT(12158, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12158(choice, ...) METALANG99_PRIV_REC_NEXT(12159, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12159(choice, ...) METALANG99_PRIV_REC_NEXT(12160, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12160(choice, ...) METALANG99_PRIV_REC_NEXT(12161, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12161(choice, ...) METALANG99_PRIV_REC_NEXT(12162, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12162(choice, ...) METALANG99_PRIV_REC_NEXT(12163, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12163(choice, ...) METALANG99_PRIV_REC_NEXT(12164, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12164(choice, ...) METALANG99_PRIV_REC_NEXT(12165, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12165(choice, ...) METALANG99_PRIV_REC_NEXT(12166, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12166(choice, ...) METALANG99_PRIV_REC_NEXT(12167, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12167(choice, ...) METALANG99_PRIV_REC_NEXT(12168, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12168(choice, ...) METALANG99_PRIV_REC_NEXT(12169, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12169(choice, ...) METALANG99_PRIV_REC_NEXT(12170, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12170(choice, ...) METALANG99_PRIV_REC_NEXT(12171, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12171(choice, ...) METALANG99_PRIV_REC_NEXT(12172, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12172(choice, ...) METALANG99_PRIV_REC_NEXT(12173, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12173(choice, ...) METALANG99_PRIV_REC_NEXT(12174, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12174(choice, ...) METALANG99_PRIV_REC_NEXT(12175, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12175(choice, ...) METALANG99_PRIV_REC_NEXT(12176, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12176(choice, ...) METALANG99_PRIV_REC_NEXT(12177, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12177(choice, ...) METALANG99_PRIV_REC_NEXT(12178, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12178(choice, ...) METALANG99_PRIV_REC_NEXT(12179, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12179(choice, ...) METALANG99_PRIV_REC_NEXT(12180, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12180(choice, ...) METALANG99_PRIV_REC_NEXT(12181, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12181(choice, ...) METALANG99_PRIV_REC_NEXT(12182, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12182(choice, ...) METALANG99_PRIV_REC_NEXT(12183, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12183(choice, ...) METALANG99_PRIV_REC_NEXT(12184, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12184(choice, ...) METALANG99_PRIV_REC_NEXT(12185, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12185(choice, ...) METALANG99_PRIV_REC_NEXT(12186, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12186(choice, ...) METALANG99_PRIV_REC_NEXT(12187, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12187(choice, ...) METALANG99_PRIV_REC_NEXT(12188, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12188(choice, ...) METALANG99_PRIV_REC_NEXT(12189, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12189(choice, ...) METALANG99_PRIV_REC_NEXT(12190, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12190(choice, ...) METALANG99_PRIV_REC_NEXT(12191, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12191(choice, ...) METALANG99_PRIV_REC_NEXT(12192, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12192(choice, ...) METALANG99_PRIV_REC_NEXT(12193, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12193(choice, ...) METALANG99_PRIV_REC_NEXT(12194, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12194(choice, ...) METALANG99_PRIV_REC_NEXT(12195, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12195(choice, ...) METALANG99_PRIV_REC_NEXT(12196, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12196(choice, ...) METALANG99_PRIV_REC_NEXT(12197, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12197(choice, ...) METALANG99_PRIV_REC_NEXT(12198, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12198(choice, ...) METALANG99_PRIV_REC_NEXT(12199, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12199(choice, ...) METALANG99_PRIV_REC_NEXT(12200, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12200(choice, ...) METALANG99_PRIV_REC_NEXT(12201, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12201(choice, ...) METALANG99_PRIV_REC_NEXT(12202, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12202(choice, ...) METALANG99_PRIV_REC_NEXT(12203, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12203(choice, ...) METALANG99_PRIV_REC_NEXT(12204, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12204(choice, ...) METALANG99_PRIV_REC_NEXT(12205, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12205(choice, ...) METALANG99_PRIV_REC_NEXT(12206, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12206(choice, ...) METALANG99_PRIV_REC_NEXT(12207, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12207(choice, ...) METALANG99_PRIV_REC_NEXT(12208, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12208(choice, ...) METALANG99_PRIV_REC_NEXT(12209, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12209(choice, ...) METALANG99_PRIV_REC_NEXT(12210, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12210(choice, ...) METALANG99_PRIV_REC_NEXT(12211, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12211(choice, ...) METALANG99_PRIV_REC_NEXT(12212, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12212(choice, ...) METALANG99_PRIV_REC_NEXT(12213, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12213(choice, ...) METALANG99_PRIV_REC_NEXT(12214, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12214(choice, ...) METALANG99_PRIV_REC_NEXT(12215, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12215(choice, ...) METALANG99_PRIV_REC_NEXT(12216, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12216(choice, ...) METALANG99_PRIV_REC_NEXT(12217, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12217(choice, ...) METALANG99_PRIV_REC_NEXT(12218, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12218(choice, ...) METALANG99_PRIV_REC_NEXT(12219, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12219(choice, ...) METALANG99_PRIV_REC_NEXT(12220, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12220(choice, ...) METALANG99_PRIV_REC_NEXT(12221, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12221(choice, ...) METALANG99_PRIV_REC_NEXT(12222, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12222(choice, ...) METALANG99_PRIV_REC_NEXT(12223, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12223(choice, ...) METALANG99_PRIV_REC_NEXT(12224, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12224(choice, ...) METALANG99_PRIV_REC_NEXT(12225, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12225(choice, ...) METALANG99_PRIV_REC_NEXT(12226, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12226(choice, ...) METALANG99_PRIV_REC_NEXT(12227, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12227(choice, ...) METALANG99_PRIV_REC_NEXT(12228, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12228(choice, ...) METALANG99_PRIV_REC_NEXT(12229, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12229(choice, ...) METALANG99_PRIV_REC_NEXT(12230, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12230(choice, ...) METALANG99_PRIV_REC_NEXT(12231, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12231(choice, ...) METALANG99_PRIV_REC_NEXT(12232, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12232(choice, ...) METALANG99_PRIV_REC_NEXT(12233, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12233(choice, ...) METALANG99_PRIV_REC_NEXT(12234, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12234(choice, ...) METALANG99_PRIV_REC_NEXT(12235, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12235(choice, ...) METALANG99_PRIV_REC_NEXT(12236, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12236(choice, ...) METALANG99_PRIV_REC_NEXT(12237, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12237(choice, ...) METALANG99_PRIV_REC_NEXT(12238, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12238(choice, ...) METALANG99_PRIV_REC_NEXT(12239, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12239(choice, ...) METALANG99_PRIV_REC_NEXT(12240, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12240(choice, ...) METALANG99_PRIV_REC_NEXT(12241, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12241(choice, ...) METALANG99_PRIV_REC_NEXT(12242, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12242(choice, ...) METALANG99_PRIV_REC_NEXT(12243, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12243(choice, ...) METALANG99_PRIV_REC_NEXT(12244, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12244(choice, ...) METALANG99_PRIV_REC_NEXT(12245, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12245(choice, ...) METALANG99_PRIV_REC_NEXT(12246, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12246(choice, ...) METALANG99_PRIV_REC_NEXT(12247, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12247(choice, ...) METALANG99_PRIV_REC_NEXT(12248, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12248(choice, ...) METALANG99_PRIV_REC_NEXT(12249, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12249(choice, ...) METALANG99_PRIV_REC_NEXT(12250, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12250(choice, ...) METALANG99_PRIV_REC_NEXT(12251, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12251(choice, ...) METALANG99_PRIV_REC_NEXT(12252, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12252(choice, ...) METALANG99_PRIV_REC_NEXT(12253, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12253(choice, ...) METALANG99_PRIV_REC_NEXT(12254, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12254(choice, ...) METALANG99_PRIV_REC_NEXT(12255, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12255(choice, ...) METALANG99_PRIV_REC_NEXT(12256, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12256(choice, ...) METALANG99_PRIV_REC_NEXT(12257, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12257(choice, ...) METALANG99_PRIV_REC_NEXT(12258, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12258(choice, ...) METALANG99_PRIV_REC_NEXT(12259, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12259(choice, ...) METALANG99_PRIV_REC_NEXT(12260, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12260(choice, ...) METALANG99_PRIV_REC_NEXT(12261, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12261(choice, ...) METALANG99_PRIV_REC_NEXT(12262, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12262(choice, ...) METALANG99_PRIV_REC_NEXT(12263, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12263(choice, ...) METALANG99_PRIV_REC_NEXT(12264, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12264(choice, ...) METALANG99_PRIV_REC_NEXT(12265, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12265(choice, ...) METALANG99_PRIV_REC_NEXT(12266, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12266(choice, ...) METALANG99_PRIV_REC_NEXT(12267, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12267(choice, ...) METALANG99_PRIV_REC_NEXT(12268, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12268(choice, ...) METALANG99_PRIV_REC_NEXT(12269, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12269(choice, ...) METALANG99_PRIV_REC_NEXT(12270, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12270(choice, ...) METALANG99_PRIV_REC_NEXT(12271, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12271(choice, ...) METALANG99_PRIV_REC_NEXT(12272, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12272(choice, ...) METALANG99_PRIV_REC_NEXT(12273, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12273(choice, ...) METALANG99_PRIV_REC_NEXT(12274, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12274(choice, ...) METALANG99_PRIV_REC_NEXT(12275, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12275(choice, ...) METALANG99_PRIV_REC_NEXT(12276, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12276(choice, ...) METALANG99_PRIV_REC_NEXT(12277, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12277(choice, ...) METALANG99_PRIV_REC_NEXT(12278, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12278(choice, ...) METALANG99_PRIV_REC_NEXT(12279, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12279(choice, ...) METALANG99_PRIV_REC_NEXT(12280, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12280(choice, ...) METALANG99_PRIV_REC_NEXT(12281, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12281(choice, ...) METALANG99_PRIV_REC_NEXT(12282, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12282(choice, ...) METALANG99_PRIV_REC_NEXT(12283, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12283(choice, ...) METALANG99_PRIV_REC_NEXT(12284, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12284(choice, ...) METALANG99_PRIV_REC_NEXT(12285, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12285(choice, ...) METALANG99_PRIV_REC_NEXT(12286, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12286(choice, ...) METALANG99_PRIV_REC_NEXT(12287, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12287(choice, ...) METALANG99_PRIV_REC_NEXT(12288, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12288(choice, ...) METALANG99_PRIV_REC_NEXT(12289, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12289(choice, ...) METALANG99_PRIV_REC_NEXT(12290, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12290(choice, ...) METALANG99_PRIV_REC_NEXT(12291, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12291(choice, ...) METALANG99_PRIV_REC_NEXT(12292, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12292(choice, ...) METALANG99_PRIV_REC_NEXT(12293, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12293(choice, ...) METALANG99_PRIV_REC_NEXT(12294, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12294(choice, ...) METALANG99_PRIV_REC_NEXT(12295, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12295(choice, ...) METALANG99_PRIV_REC_NEXT(12296, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12296(choice, ...) METALANG99_PRIV_REC_NEXT(12297, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12297(choice, ...) METALANG99_PRIV_REC_NEXT(12298, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12298(choice, ...) METALANG99_PRIV_REC_NEXT(12299, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12299(choice, ...) METALANG99_PRIV_REC_NEXT(12300, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12300(choice, ...) METALANG99_PRIV_REC_NEXT(12301, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12301(choice, ...) METALANG99_PRIV_REC_NEXT(12302, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12302(choice, ...) METALANG99_PRIV_REC_NEXT(12303, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12303(choice, ...) METALANG99_PRIV_REC_NEXT(12304, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12304(choice, ...) METALANG99_PRIV_REC_NEXT(12305, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12305(choice, ...) METALANG99_PRIV_REC_NEXT(12306, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12306(choice, ...) METALANG99_PRIV_REC_NEXT(12307, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12307(choice, ...) METALANG99_PRIV_REC_NEXT(12308, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12308(choice, ...) METALANG99_PRIV_REC_NEXT(12309, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12309(choice, ...) METALANG99_PRIV_REC_NEXT(12310, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12310(choice, ...) METALANG99_PRIV_REC_NEXT(12311, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12311(choice, ...) METALANG99_PRIV_REC_NEXT(12312, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12312(choice, ...) METALANG99_PRIV_REC_NEXT(12313, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12313(choice, ...) METALANG99_PRIV_REC_NEXT(12314, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12314(choice, ...) METALANG99_PRIV_REC_NEXT(12315, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12315(choice, ...) METALANG99_PRIV_REC_NEXT(12316, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12316(choice, ...) METALANG99_PRIV_REC_NEXT(12317, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12317(choice, ...) METALANG99_PRIV_REC_NEXT(12318, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12318(choice, ...) METALANG99_PRIV_REC_NEXT(12319, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12319(choice, ...) METALANG99_PRIV_REC_NEXT(12320, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12320(choice, ...) METALANG99_PRIV_REC_NEXT(12321, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12321(choice, ...) METALANG99_PRIV_REC_NEXT(12322, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12322(choice, ...) METALANG99_PRIV_REC_NEXT(12323, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12323(choice, ...) METALANG99_PRIV_REC_NEXT(12324, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12324(choice, ...) METALANG99_PRIV_REC_NEXT(12325, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12325(choice, ...) METALANG99_PRIV_REC_NEXT(12326, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12326(choice, ...) METALANG99_PRIV_REC_NEXT(12327, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12327(choice, ...) METALANG99_PRIV_REC_NEXT(12328, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12328(choice, ...) METALANG99_PRIV_REC_NEXT(12329, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12329(choice, ...) METALANG99_PRIV_REC_NEXT(12330, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12330(choice, ...) METALANG99_PRIV_REC_NEXT(12331, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12331(choice, ...) METALANG99_PRIV_REC_NEXT(12332, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12332(choice, ...) METALANG99_PRIV_REC_NEXT(12333, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12333(choice, ...) METALANG99_PRIV_REC_NEXT(12334, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12334(choice, ...) METALANG99_PRIV_REC_NEXT(12335, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12335(choice, ...) METALANG99_PRIV_REC_NEXT(12336, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12336(choice, ...) METALANG99_PRIV_REC_NEXT(12337, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12337(choice, ...) METALANG99_PRIV_REC_NEXT(12338, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12338(choice, ...) METALANG99_PRIV_REC_NEXT(12339, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12339(choice, ...) METALANG99_PRIV_REC_NEXT(12340, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12340(choice, ...) METALANG99_PRIV_REC_NEXT(12341, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12341(choice, ...) METALANG99_PRIV_REC_NEXT(12342, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12342(choice, ...) METALANG99_PRIV_REC_NEXT(12343, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12343(choice, ...) METALANG99_PRIV_REC_NEXT(12344, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12344(choice, ...) METALANG99_PRIV_REC_NEXT(12345, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12345(choice, ...) METALANG99_PRIV_REC_NEXT(12346, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12346(choice, ...) METALANG99_PRIV_REC_NEXT(12347, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12347(choice, ...) METALANG99_PRIV_REC_NEXT(12348, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12348(choice, ...) METALANG99_PRIV_REC_NEXT(12349, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12349(choice, ...) METALANG99_PRIV_REC_NEXT(12350, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12350(choice, ...) METALANG99_PRIV_REC_NEXT(12351, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12351(choice, ...) METALANG99_PRIV_REC_NEXT(12352, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12352(choice, ...) METALANG99_PRIV_REC_NEXT(12353, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12353(choice, ...) METALANG99_PRIV_REC_NEXT(12354, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12354(choice, ...) METALANG99_PRIV_REC_NEXT(12355, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12355(choice, ...) METALANG99_PRIV_REC_NEXT(12356, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12356(choice, ...) METALANG99_PRIV_REC_NEXT(12357, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12357(choice, ...) METALANG99_PRIV_REC_NEXT(12358, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12358(choice, ...) METALANG99_PRIV_REC_NEXT(12359, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12359(choice, ...) METALANG99_PRIV_REC_NEXT(12360, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12360(choice, ...) METALANG99_PRIV_REC_NEXT(12361, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12361(choice, ...) METALANG99_PRIV_REC_NEXT(12362, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12362(choice, ...) METALANG99_PRIV_REC_NEXT(12363, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12363(choice, ...) METALANG99_PRIV_REC_NEXT(12364, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12364(choice, ...) METALANG99_PRIV_REC_NEXT(12365, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12365(choice, ...) METALANG99_PRIV_REC_NEXT(12366, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12366(choice, ...) METALANG99_PRIV_REC_NEXT(12367, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12367(choice, ...) METALANG99_PRIV_REC_NEXT(12368, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12368(choice, ...) METALANG99_PRIV_REC_NEXT(12369, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12369(choice, ...) METALANG99_PRIV_REC_NEXT(12370, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12370(choice, ...) METALANG99_PRIV_REC_NEXT(12371, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12371(choice, ...) METALANG99_PRIV_REC_NEXT(12372, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12372(choice, ...) METALANG99_PRIV_REC_NEXT(12373, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12373(choice, ...) METALANG99_PRIV_REC_NEXT(12374, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12374(choice, ...) METALANG99_PRIV_REC_NEXT(12375, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12375(choice, ...) METALANG99_PRIV_REC_NEXT(12376, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12376(choice, ...) METALANG99_PRIV_REC_NEXT(12377, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12377(choice, ...) METALANG99_PRIV_REC_NEXT(12378, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12378(choice, ...) METALANG99_PRIV_REC_NEXT(12379, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12379(choice, ...) METALANG99_PRIV_REC_NEXT(12380, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12380(choice, ...) METALANG99_PRIV_REC_NEXT(12381, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12381(choice, ...) METALANG99_PRIV_REC_NEXT(12382, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12382(choice, ...) METALANG99_PRIV_REC_NEXT(12383, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12383(choice, ...) METALANG99_PRIV_REC_NEXT(12384, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12384(choice, ...) METALANG99_PRIV_REC_NEXT(12385, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12385(choice, ...) METALANG99_PRIV_REC_NEXT(12386, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12386(choice, ...) METALANG99_PRIV_REC_NEXT(12387, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12387(choice, ...) METALANG99_PRIV_REC_NEXT(12388, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12388(choice, ...) METALANG99_PRIV_REC_NEXT(12389, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12389(choice, ...) METALANG99_PRIV_REC_NEXT(12390, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12390(choice, ...) METALANG99_PRIV_REC_NEXT(12391, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12391(choice, ...) METALANG99_PRIV_REC_NEXT(12392, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12392(choice, ...) METALANG99_PRIV_REC_NEXT(12393, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12393(choice, ...) METALANG99_PRIV_REC_NEXT(12394, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12394(choice, ...) METALANG99_PRIV_REC_NEXT(12395, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12395(choice, ...) METALANG99_PRIV_REC_NEXT(12396, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12396(choice, ...) METALANG99_PRIV_REC_NEXT(12397, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12397(choice, ...) METALANG99_PRIV_REC_NEXT(12398, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12398(choice, ...) METALANG99_PRIV_REC_NEXT(12399, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12399(choice, ...) METALANG99_PRIV_REC_NEXT(12400, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12400(choice, ...) METALANG99_PRIV_REC_NEXT(12401, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12401(choice, ...) METALANG99_PRIV_REC_NEXT(12402, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12402(choice, ...) METALANG99_PRIV_REC_NEXT(12403, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12403(choice, ...) METALANG99_PRIV_REC_NEXT(12404, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12404(choice, ...) METALANG99_PRIV_REC_NEXT(12405, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12405(choice, ...) METALANG99_PRIV_REC_NEXT(12406, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12406(choice, ...) METALANG99_PRIV_REC_NEXT(12407, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12407(choice, ...) METALANG99_PRIV_REC_NEXT(12408, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12408(choice, ...) METALANG99_PRIV_REC_NEXT(12409, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12409(choice, ...) METALANG99_PRIV_REC_NEXT(12410, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12410(choice, ...) METALANG99_PRIV_REC_NEXT(12411, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12411(choice, ...) METALANG99_PRIV_REC_NEXT(12412, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12412(choice, ...) METALANG99_PRIV_REC_NEXT(12413, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12413(choice, ...) METALANG99_PRIV_REC_NEXT(12414, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12414(choice, ...) METALANG99_PRIV_REC_NEXT(12415, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12415(choice, ...) METALANG99_PRIV_REC_NEXT(12416, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12416(choice, ...) METALANG99_PRIV_REC_NEXT(12417, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12417(choice, ...) METALANG99_PRIV_REC_NEXT(12418, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12418(choice, ...) METALANG99_PRIV_REC_NEXT(12419, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12419(choice, ...) METALANG99_PRIV_REC_NEXT(12420, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12420(choice, ...) METALANG99_PRIV_REC_NEXT(12421, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12421(choice, ...) METALANG99_PRIV_REC_NEXT(12422, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12422(choice, ...) METALANG99_PRIV_REC_NEXT(12423, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12423(choice, ...) METALANG99_PRIV_REC_NEXT(12424, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12424(choice, ...) METALANG99_PRIV_REC_NEXT(12425, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12425(choice, ...) METALANG99_PRIV_REC_NEXT(12426, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12426(choice, ...) METALANG99_PRIV_REC_NEXT(12427, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12427(choice, ...) METALANG99_PRIV_REC_NEXT(12428, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12428(choice, ...) METALANG99_PRIV_REC_NEXT(12429, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12429(choice, ...) METALANG99_PRIV_REC_NEXT(12430, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12430(choice, ...) METALANG99_PRIV_REC_NEXT(12431, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12431(choice, ...) METALANG99_PRIV_REC_NEXT(12432, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12432(choice, ...) METALANG99_PRIV_REC_NEXT(12433, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12433(choice, ...) METALANG99_PRIV_REC_NEXT(12434, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12434(choice, ...) METALANG99_PRIV_REC_NEXT(12435, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12435(choice, ...) METALANG99_PRIV_REC_NEXT(12436, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12436(choice, ...) METALANG99_PRIV_REC_NEXT(12437, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12437(choice, ...) METALANG99_PRIV_REC_NEXT(12438, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12438(choice, ...) METALANG99_PRIV_REC_NEXT(12439, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12439(choice, ...) METALANG99_PRIV_REC_NEXT(12440, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12440(choice, ...) METALANG99_PRIV_REC_NEXT(12441, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12441(choice, ...) METALANG99_PRIV_REC_NEXT(12442, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12442(choice, ...) METALANG99_PRIV_REC_NEXT(12443, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12443(choice, ...) METALANG99_PRIV_REC_NEXT(12444, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12444(choice, ...) METALANG99_PRIV_REC_NEXT(12445, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12445(choice, ...) METALANG99_PRIV_REC_NEXT(12446, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12446(choice, ...) METALANG99_PRIV_REC_NEXT(12447, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12447(choice, ...) METALANG99_PRIV_REC_NEXT(12448, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12448(choice, ...) METALANG99_PRIV_REC_NEXT(12449, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12449(choice, ...) METALANG99_PRIV_REC_NEXT(12450, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12450(choice, ...) METALANG99_PRIV_REC_NEXT(12451, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12451(choice, ...) METALANG99_PRIV_REC_NEXT(12452, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12452(choice, ...) METALANG99_PRIV_REC_NEXT(12453, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12453(choice, ...) METALANG99_PRIV_REC_NEXT(12454, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12454(choice, ...) METALANG99_PRIV_REC_NEXT(12455, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12455(choice, ...) METALANG99_PRIV_REC_NEXT(12456, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12456(choice, ...) METALANG99_PRIV_REC_NEXT(12457, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12457(choice, ...) METALANG99_PRIV_REC_NEXT(12458, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12458(choice, ...) METALANG99_PRIV_REC_NEXT(12459, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12459(choice, ...) METALANG99_PRIV_REC_NEXT(12460, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12460(choice, ...) METALANG99_PRIV_REC_NEXT(12461, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12461(choice, ...) METALANG99_PRIV_REC_NEXT(12462, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12462(choice, ...) METALANG99_PRIV_REC_NEXT(12463, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12463(choice, ...) METALANG99_PRIV_REC_NEXT(12464, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12464(choice, ...) METALANG99_PRIV_REC_NEXT(12465, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12465(choice, ...) METALANG99_PRIV_REC_NEXT(12466, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12466(choice, ...) METALANG99_PRIV_REC_NEXT(12467, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12467(choice, ...) METALANG99_PRIV_REC_NEXT(12468, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12468(choice, ...) METALANG99_PRIV_REC_NEXT(12469, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12469(choice, ...) METALANG99_PRIV_REC_NEXT(12470, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12470(choice, ...) METALANG99_PRIV_REC_NEXT(12471, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12471(choice, ...) METALANG99_PRIV_REC_NEXT(12472, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12472(choice, ...) METALANG99_PRIV_REC_NEXT(12473, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12473(choice, ...) METALANG99_PRIV_REC_NEXT(12474, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12474(choice, ...) METALANG99_PRIV_REC_NEXT(12475, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12475(choice, ...) METALANG99_PRIV_REC_NEXT(12476, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12476(choice, ...) METALANG99_PRIV_REC_NEXT(12477, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12477(choice, ...) METALANG99_PRIV_REC_NEXT(12478, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12478(choice, ...) METALANG99_PRIV_REC_NEXT(12479, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12479(choice, ...) METALANG99_PRIV_REC_NEXT(12480, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12480(choice, ...) METALANG99_PRIV_REC_NEXT(12481, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12481(choice, ...) METALANG99_PRIV_REC_NEXT(12482, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12482(choice, ...) METALANG99_PRIV_REC_NEXT(12483, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12483(choice, ...) METALANG99_PRIV_REC_NEXT(12484, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12484(choice, ...) METALANG99_PRIV_REC_NEXT(12485, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12485(choice, ...) METALANG99_PRIV_REC_NEXT(12486, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12486(choice, ...) METALANG99_PRIV_REC_NEXT(12487, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12487(choice, ...) METALANG99_PRIV_REC_NEXT(12488, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12488(choice, ...) METALANG99_PRIV_REC_NEXT(12489, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12489(choice, ...) METALANG99_PRIV_REC_NEXT(12490, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12490(choice, ...) METALANG99_PRIV_REC_NEXT(12491, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12491(choice, ...) METALANG99_PRIV_REC_NEXT(12492, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12492(choice, ...) METALANG99_PRIV_REC_NEXT(12493, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12493(choice, ...) METALANG99_PRIV_REC_NEXT(12494, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12494(choice, ...) METALANG99_PRIV_REC_NEXT(12495, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12495(choice, ...) METALANG99_PRIV_REC_NEXT(12496, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12496(choice, ...) METALANG99_PRIV_REC_NEXT(12497, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12497(choice, ...) METALANG99_PRIV_REC_NEXT(12498, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12498(choice, ...) METALANG99_PRIV_REC_NEXT(12499, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12499(choice, ...) METALANG99_PRIV_REC_NEXT(12500, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12500(choice, ...) METALANG99_PRIV_REC_NEXT(12501, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12501(choice, ...) METALANG99_PRIV_REC_NEXT(12502, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12502(choice, ...) METALANG99_PRIV_REC_NEXT(12503, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12503(choice, ...) METALANG99_PRIV_REC_NEXT(12504, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12504(choice, ...) METALANG99_PRIV_REC_NEXT(12505, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12505(choice, ...) METALANG99_PRIV_REC_NEXT(12506, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12506(choice, ...) METALANG99_PRIV_REC_NEXT(12507, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12507(choice, ...) METALANG99_PRIV_REC_NEXT(12508, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12508(choice, ...) METALANG99_PRIV_REC_NEXT(12509, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12509(choice, ...) METALANG99_PRIV_REC_NEXT(12510, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12510(choice, ...) METALANG99_PRIV_REC_NEXT(12511, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12511(choice, ...) METALANG99_PRIV_REC_NEXT(12512, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12512(choice, ...) METALANG99_PRIV_REC_NEXT(12513, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12513(choice, ...) METALANG99_PRIV_REC_NEXT(12514, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12514(choice, ...) METALANG99_PRIV_REC_NEXT(12515, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12515(choice, ...) METALANG99_PRIV_REC_NEXT(12516, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12516(choice, ...) METALANG99_PRIV_REC_NEXT(12517, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12517(choice, ...) METALANG99_PRIV_REC_NEXT(12518, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12518(choice, ...) METALANG99_PRIV_REC_NEXT(12519, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12519(choice, ...) METALANG99_PRIV_REC_NEXT(12520, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12520(choice, ...) METALANG99_PRIV_REC_NEXT(12521, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12521(choice, ...) METALANG99_PRIV_REC_NEXT(12522, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12522(choice, ...) METALANG99_PRIV_REC_NEXT(12523, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12523(choice, ...) METALANG99_PRIV_REC_NEXT(12524, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12524(choice, ...) METALANG99_PRIV_REC_NEXT(12525, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12525(choice, ...) METALANG99_PRIV_REC_NEXT(12526, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12526(choice, ...) METALANG99_PRIV_REC_NEXT(12527, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12527(choice, ...) METALANG99_PRIV_REC_NEXT(12528, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12528(choice, ...) METALANG99_PRIV_REC_NEXT(12529, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12529(choice, ...) METALANG99_PRIV_REC_NEXT(12530, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12530(choice, ...) METALANG99_PRIV_REC_NEXT(12531, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12531(choice, ...) METALANG99_PRIV_REC_NEXT(12532, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12532(choice, ...) METALANG99_PRIV_REC_NEXT(12533, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12533(choice, ...) METALANG99_PRIV_REC_NEXT(12534, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12534(choice, ...) METALANG99_PRIV_REC_NEXT(12535, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12535(choice, ...) METALANG99_PRIV_REC_NEXT(12536, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12536(choice, ...) METALANG99_PRIV_REC_NEXT(12537, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12537(choice, ...) METALANG99_PRIV_REC_NEXT(12538, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12538(choice, ...) METALANG99_PRIV_REC_NEXT(12539, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12539(choice, ...) METALANG99_PRIV_REC_NEXT(12540, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12540(choice, ...) METALANG99_PRIV_REC_NEXT(12541, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12541(choice, ...) METALANG99_PRIV_REC_NEXT(12542, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12542(choice, ...) METALANG99_PRIV_REC_NEXT(12543, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12543(choice, ...) METALANG99_PRIV_REC_NEXT(12544, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12544(choice, ...) METALANG99_PRIV_REC_NEXT(12545, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12545(choice, ...) METALANG99_PRIV_REC_NEXT(12546, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12546(choice, ...) METALANG99_PRIV_REC_NEXT(12547, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12547(choice, ...) METALANG99_PRIV_REC_NEXT(12548, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12548(choice, ...) METALANG99_PRIV_REC_NEXT(12549, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12549(choice, ...) METALANG99_PRIV_REC_NEXT(12550, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12550(choice, ...) METALANG99_PRIV_REC_NEXT(12551, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12551(choice, ...) METALANG99_PRIV_REC_NEXT(12552, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12552(choice, ...) METALANG99_PRIV_REC_NEXT(12553, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12553(choice, ...) METALANG99_PRIV_REC_NEXT(12554, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12554(choice, ...) METALANG99_PRIV_REC_NEXT(12555, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12555(choice, ...) METALANG99_PRIV_REC_NEXT(12556, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12556(choice, ...) METALANG99_PRIV_REC_NEXT(12557, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12557(choice, ...) METALANG99_PRIV_REC_NEXT(12558, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12558(choice, ...) METALANG99_PRIV_REC_NEXT(12559, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12559(choice, ...) METALANG99_PRIV_REC_NEXT(12560, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12560(choice, ...) METALANG99_PRIV_REC_NEXT(12561, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12561(choice, ...) METALANG99_PRIV_REC_NEXT(12562, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12562(choice, ...) METALANG99_PRIV_REC_NEXT(12563, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12563(choice, ...) METALANG99_PRIV_REC_NEXT(12564, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12564(choice, ...) METALANG99_PRIV_REC_NEXT(12565, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12565(choice, ...) METALANG99_PRIV_REC_NEXT(12566, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12566(choice, ...) METALANG99_PRIV_REC_NEXT(12567, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12567(choice, ...) METALANG99_PRIV_REC_NEXT(12568, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12568(choice, ...) METALANG99_PRIV_REC_NEXT(12569, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12569(choice, ...) METALANG99_PRIV_REC_NEXT(12570, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12570(choice, ...) METALANG99_PRIV_REC_NEXT(12571, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12571(choice, ...) METALANG99_PRIV_REC_NEXT(12572, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12572(choice, ...) METALANG99_PRIV_REC_NEXT(12573, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12573(choice, ...) METALANG99_PRIV_REC_NEXT(12574, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12574(choice, ...) METALANG99_PRIV_REC_NEXT(12575, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12575(choice, ...) METALANG99_PRIV_REC_NEXT(12576, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12576(choice, ...) METALANG99_PRIV_REC_NEXT(12577, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12577(choice, ...) METALANG99_PRIV_REC_NEXT(12578, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12578(choice, ...) METALANG99_PRIV_REC_NEXT(12579, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12579(choice, ...) METALANG99_PRIV_REC_NEXT(12580, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12580(choice, ...) METALANG99_PRIV_REC_NEXT(12581, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12581(choice, ...) METALANG99_PRIV_REC_NEXT(12582, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12582(choice, ...) METALANG99_PRIV_REC_NEXT(12583, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12583(choice, ...) METALANG99_PRIV_REC_NEXT(12584, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12584(choice, ...) METALANG99_PRIV_REC_NEXT(12585, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12585(choice, ...) METALANG99_PRIV_REC_NEXT(12586, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12586(choice, ...) METALANG99_PRIV_REC_NEXT(12587, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12587(choice, ...) METALANG99_PRIV_REC_NEXT(12588, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12588(choice, ...) METALANG99_PRIV_REC_NEXT(12589, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12589(choice, ...) METALANG99_PRIV_REC_NEXT(12590, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12590(choice, ...) METALANG99_PRIV_REC_NEXT(12591, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12591(choice, ...) METALANG99_PRIV_REC_NEXT(12592, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12592(choice, ...) METALANG99_PRIV_REC_NEXT(12593, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12593(choice, ...) METALANG99_PRIV_REC_NEXT(12594, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12594(choice, ...) METALANG99_PRIV_REC_NEXT(12595, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12595(choice, ...) METALANG99_PRIV_REC_NEXT(12596, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12596(choice, ...) METALANG99_PRIV_REC_NEXT(12597, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12597(choice, ...) METALANG99_PRIV_REC_NEXT(12598, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12598(choice, ...) METALANG99_PRIV_REC_NEXT(12599, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12599(choice, ...) METALANG99_PRIV_REC_NEXT(12600, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12600(choice, ...) METALANG99_PRIV_REC_NEXT(12601, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12601(choice, ...) METALANG99_PRIV_REC_NEXT(12602, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12602(choice, ...) METALANG99_PRIV_REC_NEXT(12603, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12603(choice, ...) METALANG99_PRIV_REC_NEXT(12604, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12604(choice, ...) METALANG99_PRIV_REC_NEXT(12605, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12605(choice, ...) METALANG99_PRIV_REC_NEXT(12606, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12606(choice, ...) METALANG99_PRIV_REC_NEXT(12607, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12607(choice, ...) METALANG99_PRIV_REC_NEXT(12608, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12608(choice, ...) METALANG99_PRIV_REC_NEXT(12609, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12609(choice, ...) METALANG99_PRIV_REC_NEXT(12610, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12610(choice, ...) METALANG99_PRIV_REC_NEXT(12611, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12611(choice, ...) METALANG99_PRIV_REC_NEXT(12612, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12612(choice, ...) METALANG99_PRIV_REC_NEXT(12613, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12613(choice, ...) METALANG99_PRIV_REC_NEXT(12614, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12614(choice, ...) METALANG99_PRIV_REC_NEXT(12615, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12615(choice, ...) METALANG99_PRIV_REC_NEXT(12616, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12616(choice, ...) METALANG99_PRIV_REC_NEXT(12617, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12617(choice, ...) METALANG99_PRIV_REC_NEXT(12618, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12618(choice, ...) METALANG99_PRIV_REC_NEXT(12619, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12619(choice, ...) METALANG99_PRIV_REC_NEXT(12620, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12620(choice, ...) METALANG99_PRIV_REC_NEXT(12621, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12621(choice, ...) METALANG99_PRIV_REC_NEXT(12622, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12622(choice, ...) METALANG99_PRIV_REC_NEXT(12623, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12623(choice, ...) METALANG99_PRIV_REC_NEXT(12624, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12624(choice, ...) METALANG99_PRIV_REC_NEXT(12625, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12625(choice, ...) METALANG99_PRIV_REC_NEXT(12626, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12626(choice, ...) METALANG99_PRIV_REC_NEXT(12627, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12627(choice, ...) METALANG99_PRIV_REC_NEXT(12628, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12628(choice, ...) METALANG99_PRIV_REC_NEXT(12629, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12629(choice, ...) METALANG99_PRIV_REC_NEXT(12630, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12630(choice, ...) METALANG99_PRIV_REC_NEXT(12631, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12631(choice, ...) METALANG99_PRIV_REC_NEXT(12632, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12632(choice, ...) METALANG99_PRIV_REC_NEXT(12633, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12633(choice, ...) METALANG99_PRIV_REC_NEXT(12634, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12634(choice, ...) METALANG99_PRIV_REC_NEXT(12635, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12635(choice, ...) METALANG99_PRIV_REC_NEXT(12636, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12636(choice, ...) METALANG99_PRIV_REC_NEXT(12637, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12637(choice, ...) METALANG99_PRIV_REC_NEXT(12638, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12638(choice, ...) METALANG99_PRIV_REC_NEXT(12639, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12639(choice, ...) METALANG99_PRIV_REC_NEXT(12640, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12640(choice, ...) METALANG99_PRIV_REC_NEXT(12641, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12641(choice, ...) METALANG99_PRIV_REC_NEXT(12642, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12642(choice, ...) METALANG99_PRIV_REC_NEXT(12643, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12643(choice, ...) METALANG99_PRIV_REC_NEXT(12644, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12644(choice, ...) METALANG99_PRIV_REC_NEXT(12645, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12645(choice, ...) METALANG99_PRIV_REC_NEXT(12646, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12646(choice, ...) METALANG99_PRIV_REC_NEXT(12647, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12647(choice, ...) METALANG99_PRIV_REC_NEXT(12648, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12648(choice, ...) METALANG99_PRIV_REC_NEXT(12649, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12649(choice, ...) METALANG99_PRIV_REC_NEXT(12650, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12650(choice, ...) METALANG99_PRIV_REC_NEXT(12651, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12651(choice, ...) METALANG99_PRIV_REC_NEXT(12652, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12652(choice, ...) METALANG99_PRIV_REC_NEXT(12653, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12653(choice, ...) METALANG99_PRIV_REC_NEXT(12654, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12654(choice, ...) METALANG99_PRIV_REC_NEXT(12655, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12655(choice, ...) METALANG99_PRIV_REC_NEXT(12656, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12656(choice, ...) METALANG99_PRIV_REC_NEXT(12657, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12657(choice, ...) METALANG99_PRIV_REC_NEXT(12658, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12658(choice, ...) METALANG99_PRIV_REC_NEXT(12659, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12659(choice, ...) METALANG99_PRIV_REC_NEXT(12660, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12660(choice, ...) METALANG99_PRIV_REC_NEXT(12661, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12661(choice, ...) METALANG99_PRIV_REC_NEXT(12662, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12662(choice, ...) METALANG99_PRIV_REC_NEXT(12663, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12663(choice, ...) METALANG99_PRIV_REC_NEXT(12664, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12664(choice, ...) METALANG99_PRIV_REC_NEXT(12665, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12665(choice, ...) METALANG99_PRIV_REC_NEXT(12666, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12666(choice, ...) METALANG99_PRIV_REC_NEXT(12667, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12667(choice, ...) METALANG99_PRIV_REC_NEXT(12668, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12668(choice, ...) METALANG99_PRIV_REC_NEXT(12669, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12669(choice, ...) METALANG99_PRIV_REC_NEXT(12670, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12670(choice, ...) METALANG99_PRIV_REC_NEXT(12671, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12671(choice, ...) METALANG99_PRIV_REC_NEXT(12672, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12672(choice, ...) METALANG99_PRIV_REC_NEXT(12673, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12673(choice, ...) METALANG99_PRIV_REC_NEXT(12674, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12674(choice, ...) METALANG99_PRIV_REC_NEXT(12675, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12675(choice, ...) METALANG99_PRIV_REC_NEXT(12676, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12676(choice, ...) METALANG99_PRIV_REC_NEXT(12677, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12677(choice, ...) METALANG99_PRIV_REC_NEXT(12678, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12678(choice, ...) METALANG99_PRIV_REC_NEXT(12679, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12679(choice, ...) METALANG99_PRIV_REC_NEXT(12680, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12680(choice, ...) METALANG99_PRIV_REC_NEXT(12681, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12681(choice, ...) METALANG99_PRIV_REC_NEXT(12682, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12682(choice, ...) METALANG99_PRIV_REC_NEXT(12683, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12683(choice, ...) METALANG99_PRIV_REC_NEXT(12684, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12684(choice, ...) METALANG99_PRIV_REC_NEXT(12685, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12685(choice, ...) METALANG99_PRIV_REC_NEXT(12686, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12686(choice, ...) METALANG99_PRIV_REC_NEXT(12687, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12687(choice, ...) METALANG99_PRIV_REC_NEXT(12688, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12688(choice, ...) METALANG99_PRIV_REC_NEXT(12689, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12689(choice, ...) METALANG99_PRIV_REC_NEXT(12690, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12690(choice, ...) METALANG99_PRIV_REC_NEXT(12691, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12691(choice, ...) METALANG99_PRIV_REC_NEXT(12692, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12692(choice, ...) METALANG99_PRIV_REC_NEXT(12693, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12693(choice, ...) METALANG99_PRIV_REC_NEXT(12694, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12694(choice, ...) METALANG99_PRIV_REC_NEXT(12695, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12695(choice, ...) METALANG99_PRIV_REC_NEXT(12696, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12696(choice, ...) METALANG99_PRIV_REC_NEXT(12697, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12697(choice, ...) METALANG99_PRIV_REC_NEXT(12698, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12698(choice, ...) METALANG99_PRIV_REC_NEXT(12699, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12699(choice, ...) METALANG99_PRIV_REC_NEXT(12700, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12700(choice, ...) METALANG99_PRIV_REC_NEXT(12701, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12701(choice, ...) METALANG99_PRIV_REC_NEXT(12702, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12702(choice, ...) METALANG99_PRIV_REC_NEXT(12703, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12703(choice, ...) METALANG99_PRIV_REC_NEXT(12704, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12704(choice, ...) METALANG99_PRIV_REC_NEXT(12705, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12705(choice, ...) METALANG99_PRIV_REC_NEXT(12706, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12706(choice, ...) METALANG99_PRIV_REC_NEXT(12707, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12707(choice, ...) METALANG99_PRIV_REC_NEXT(12708, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12708(choice, ...) METALANG99_PRIV_REC_NEXT(12709, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12709(choice, ...) METALANG99_PRIV_REC_NEXT(12710, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12710(choice, ...) METALANG99_PRIV_REC_NEXT(12711, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12711(choice, ...) METALANG99_PRIV_REC_NEXT(12712, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12712(choice, ...) METALANG99_PRIV_REC_NEXT(12713, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12713(choice, ...) METALANG99_PRIV_REC_NEXT(12714, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12714(choice, ...) METALANG99_PRIV_REC_NEXT(12715, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12715(choice, ...) METALANG99_PRIV_REC_NEXT(12716, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12716(choice, ...) METALANG99_PRIV_REC_NEXT(12717, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12717(choice, ...) METALANG99_PRIV_REC_NEXT(12718, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12718(choice, ...) METALANG99_PRIV_REC_NEXT(12719, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12719(choice, ...) METALANG99_PRIV_REC_NEXT(12720, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12720(choice, ...) METALANG99_PRIV_REC_NEXT(12721, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12721(choice, ...) METALANG99_PRIV_REC_NEXT(12722, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12722(choice, ...) METALANG99_PRIV_REC_NEXT(12723, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12723(choice, ...) METALANG99_PRIV_REC_NEXT(12724, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12724(choice, ...) METALANG99_PRIV_REC_NEXT(12725, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12725(choice, ...) METALANG99_PRIV_REC_NEXT(12726, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12726(choice, ...) METALANG99_PRIV_REC_NEXT(12727, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12727(choice, ...) METALANG99_PRIV_REC_NEXT(12728, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12728(choice, ...) METALANG99_PRIV_REC_NEXT(12729, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12729(choice, ...) METALANG99_PRIV_REC_NEXT(12730, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12730(choice, ...) METALANG99_PRIV_REC_NEXT(12731, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12731(choice, ...) METALANG99_PRIV_REC_NEXT(12732, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12732(choice, ...) METALANG99_PRIV_REC_NEXT(12733, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12733(choice, ...) METALANG99_PRIV_REC_NEXT(12734, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12734(choice, ...) METALANG99_PRIV_REC_NEXT(12735, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12735(choice, ...) METALANG99_PRIV_REC_NEXT(12736, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12736(choice, ...) METALANG99_PRIV_REC_NEXT(12737, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12737(choice, ...) METALANG99_PRIV_REC_NEXT(12738, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12738(choice, ...) METALANG99_PRIV_REC_NEXT(12739, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12739(choice, ...) METALANG99_PRIV_REC_NEXT(12740, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12740(choice, ...) METALANG99_PRIV_REC_NEXT(12741, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12741(choice, ...) METALANG99_PRIV_REC_NEXT(12742, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12742(choice, ...) METALANG99_PRIV_REC_NEXT(12743, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12743(choice, ...) METALANG99_PRIV_REC_NEXT(12744, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12744(choice, ...) METALANG99_PRIV_REC_NEXT(12745, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12745(choice, ...) METALANG99_PRIV_REC_NEXT(12746, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12746(choice, ...) METALANG99_PRIV_REC_NEXT(12747, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12747(choice, ...) METALANG99_PRIV_REC_NEXT(12748, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12748(choice, ...) METALANG99_PRIV_REC_NEXT(12749, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12749(choice, ...) METALANG99_PRIV_REC_NEXT(12750, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12750(choice, ...) METALANG99_PRIV_REC_NEXT(12751, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12751(choice, ...) METALANG99_PRIV_REC_NEXT(12752, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12752(choice, ...) METALANG99_PRIV_REC_NEXT(12753, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12753(choice, ...) METALANG99_PRIV_REC_NEXT(12754, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12754(choice, ...) METALANG99_PRIV_REC_NEXT(12755, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12755(choice, ...) METALANG99_PRIV_REC_NEXT(12756, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12756(choice, ...) METALANG99_PRIV_REC_NEXT(12757, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12757(choice, ...) METALANG99_PRIV_REC_NEXT(12758, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12758(choice, ...) METALANG99_PRIV_REC_NEXT(12759, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12759(choice, ...) METALANG99_PRIV_REC_NEXT(12760, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12760(choice, ...) METALANG99_PRIV_REC_NEXT(12761, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12761(choice, ...) METALANG99_PRIV_REC_NEXT(12762, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12762(choice, ...) METALANG99_PRIV_REC_NEXT(12763, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12763(choice, ...) METALANG99_PRIV_REC_NEXT(12764, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12764(choice, ...) METALANG99_PRIV_REC_NEXT(12765, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12765(choice, ...) METALANG99_PRIV_REC_NEXT(12766, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12766(choice, ...) METALANG99_PRIV_REC_NEXT(12767, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12767(choice, ...) METALANG99_PRIV_REC_NEXT(12768, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12768(choice, ...) METALANG99_PRIV_REC_NEXT(12769, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12769(choice, ...) METALANG99_PRIV_REC_NEXT(12770, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12770(choice, ...) METALANG99_PRIV_REC_NEXT(12771, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12771(choice, ...) METALANG99_PRIV_REC_NEXT(12772, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12772(choice, ...) METALANG99_PRIV_REC_NEXT(12773, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12773(choice, ...) METALANG99_PRIV_REC_NEXT(12774, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12774(choice, ...) METALANG99_PRIV_REC_NEXT(12775, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12775(choice, ...) METALANG99_PRIV_REC_NEXT(12776, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12776(choice, ...) METALANG99_PRIV_REC_NEXT(12777, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12777(choice, ...) METALANG99_PRIV_REC_NEXT(12778, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12778(choice, ...) METALANG99_PRIV_REC_NEXT(12779, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12779(choice, ...) METALANG99_PRIV_REC_NEXT(12780, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12780(choice, ...) METALANG99_PRIV_REC_NEXT(12781, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12781(choice, ...) METALANG99_PRIV_REC_NEXT(12782, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12782(choice, ...) METALANG99_PRIV_REC_NEXT(12783, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12783(choice, ...) METALANG99_PRIV_REC_NEXT(12784, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12784(choice, ...) METALANG99_PRIV_REC_NEXT(12785, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12785(choice, ...) METALANG99_PRIV_REC_NEXT(12786, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12786(choice, ...) METALANG99_PRIV_REC_NEXT(12787, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12787(choice, ...) METALANG99_PRIV_REC_NEXT(12788, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12788(choice, ...) METALANG99_PRIV_REC_NEXT(12789, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12789(choice, ...) METALANG99_PRIV_REC_NEXT(12790, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12790(choice, ...) METALANG99_PRIV_REC_NEXT(12791, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12791(choice, ...) METALANG99_PRIV_REC_NEXT(12792, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12792(choice, ...) METALANG99_PRIV_REC_NEXT(12793, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12793(choice, ...) METALANG99_PRIV_REC_NEXT(12794, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12794(choice, ...) METALANG99_PRIV_REC_NEXT(12795, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12795(choice, ...) METALANG99_PRIV_REC_NEXT(12796, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12796(choice, ...) METALANG99_PRIV_REC_NEXT(12797, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12797(choice, ...) METALANG99_PRIV_REC_NEXT(12798, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12798(choice, ...) METALANG99_PRIV_REC_NEXT(12799, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12799(choice, ...) METALANG99_PRIV_REC_NEXT(12800, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12800(choice, ...) METALANG99_PRIV_REC_NEXT(12801, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12801(choice, ...) METALANG99_PRIV_REC_NEXT(12802, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12802(choice, ...) METALANG99_PRIV_REC_NEXT(12803, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12803(choice, ...) METALANG99_PRIV_REC_NEXT(12804, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12804(choice, ...) METALANG99_PRIV_REC_NEXT(12805, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12805(choice, ...) METALANG99_PRIV_REC_NEXT(12806, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12806(choice, ...) METALANG99_PRIV_REC_NEXT(12807, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12807(choice, ...) METALANG99_PRIV_REC_NEXT(12808, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12808(choice, ...) METALANG99_PRIV_REC_NEXT(12809, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12809(choice, ...) METALANG99_PRIV_REC_NEXT(12810, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12810(choice, ...) METALANG99_PRIV_REC_NEXT(12811, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12811(choice, ...) METALANG99_PRIV_REC_NEXT(12812, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12812(choice, ...) METALANG99_PRIV_REC_NEXT(12813, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12813(choice, ...) METALANG99_PRIV_REC_NEXT(12814, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12814(choice, ...) METALANG99_PRIV_REC_NEXT(12815, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12815(choice, ...) METALANG99_PRIV_REC_NEXT(12816, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12816(choice, ...) METALANG99_PRIV_REC_NEXT(12817, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12817(choice, ...) METALANG99_PRIV_REC_NEXT(12818, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12818(choice, ...) METALANG99_PRIV_REC_NEXT(12819, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12819(choice, ...) METALANG99_PRIV_REC_NEXT(12820, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12820(choice, ...) METALANG99_PRIV_REC_NEXT(12821, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12821(choice, ...) METALANG99_PRIV_REC_NEXT(12822, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12822(choice, ...) METALANG99_PRIV_REC_NEXT(12823, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12823(choice, ...) METALANG99_PRIV_REC_NEXT(12824, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12824(choice, ...) METALANG99_PRIV_REC_NEXT(12825, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12825(choice, ...) METALANG99_PRIV_REC_NEXT(12826, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12826(choice, ...) METALANG99_PRIV_REC_NEXT(12827, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12827(choice, ...) METALANG99_PRIV_REC_NEXT(12828, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12828(choice, ...) METALANG99_PRIV_REC_NEXT(12829, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12829(choice, ...) METALANG99_PRIV_REC_NEXT(12830, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12830(choice, ...) METALANG99_PRIV_REC_NEXT(12831, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12831(choice, ...) METALANG99_PRIV_REC_NEXT(12832, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12832(choice, ...) METALANG99_PRIV_REC_NEXT(12833, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12833(choice, ...) METALANG99_PRIV_REC_NEXT(12834, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12834(choice, ...) METALANG99_PRIV_REC_NEXT(12835, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12835(choice, ...) METALANG99_PRIV_REC_NEXT(12836, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12836(choice, ...) METALANG99_PRIV_REC_NEXT(12837, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12837(choice, ...) METALANG99_PRIV_REC_NEXT(12838, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12838(choice, ...) METALANG99_PRIV_REC_NEXT(12839, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12839(choice, ...) METALANG99_PRIV_REC_NEXT(12840, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12840(choice, ...) METALANG99_PRIV_REC_NEXT(12841, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12841(choice, ...) METALANG99_PRIV_REC_NEXT(12842, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12842(choice, ...) METALANG99_PRIV_REC_NEXT(12843, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12843(choice, ...) METALANG99_PRIV_REC_NEXT(12844, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12844(choice, ...) METALANG99_PRIV_REC_NEXT(12845, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12845(choice, ...) METALANG99_PRIV_REC_NEXT(12846, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12846(choice, ...) METALANG99_PRIV_REC_NEXT(12847, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12847(choice, ...) METALANG99_PRIV_REC_NEXT(12848, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12848(choice, ...) METALANG99_PRIV_REC_NEXT(12849, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12849(choice, ...) METALANG99_PRIV_REC_NEXT(12850, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12850(choice, ...) METALANG99_PRIV_REC_NEXT(12851, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12851(choice, ...) METALANG99_PRIV_REC_NEXT(12852, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12852(choice, ...) METALANG99_PRIV_REC_NEXT(12853, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12853(choice, ...) METALANG99_PRIV_REC_NEXT(12854, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12854(choice, ...) METALANG99_PRIV_REC_NEXT(12855, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12855(choice, ...) METALANG99_PRIV_REC_NEXT(12856, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12856(choice, ...) METALANG99_PRIV_REC_NEXT(12857, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12857(choice, ...) METALANG99_PRIV_REC_NEXT(12858, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12858(choice, ...) METALANG99_PRIV_REC_NEXT(12859, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12859(choice, ...) METALANG99_PRIV_REC_NEXT(12860, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12860(choice, ...) METALANG99_PRIV_REC_NEXT(12861, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12861(choice, ...) METALANG99_PRIV_REC_NEXT(12862, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12862(choice, ...) METALANG99_PRIV_REC_NEXT(12863, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12863(choice, ...) METALANG99_PRIV_REC_NEXT(12864, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12864(choice, ...) METALANG99_PRIV_REC_NEXT(12865, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12865(choice, ...) METALANG99_PRIV_REC_NEXT(12866, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12866(choice, ...) METALANG99_PRIV_REC_NEXT(12867, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12867(choice, ...) METALANG99_PRIV_REC_NEXT(12868, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12868(choice, ...) METALANG99_PRIV_REC_NEXT(12869, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12869(choice, ...) METALANG99_PRIV_REC_NEXT(12870, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12870(choice, ...) METALANG99_PRIV_REC_NEXT(12871, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12871(choice, ...) METALANG99_PRIV_REC_NEXT(12872, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12872(choice, ...) METALANG99_PRIV_REC_NEXT(12873, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12873(choice, ...) METALANG99_PRIV_REC_NEXT(12874, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12874(choice, ...) METALANG99_PRIV_REC_NEXT(12875, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12875(choice, ...) METALANG99_PRIV_REC_NEXT(12876, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12876(choice, ...) METALANG99_PRIV_REC_NEXT(12877, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12877(choice, ...) METALANG99_PRIV_REC_NEXT(12878, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12878(choice, ...) METALANG99_PRIV_REC_NEXT(12879, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12879(choice, ...) METALANG99_PRIV_REC_NEXT(12880, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12880(choice, ...) METALANG99_PRIV_REC_NEXT(12881, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12881(choice, ...) METALANG99_PRIV_REC_NEXT(12882, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12882(choice, ...) METALANG99_PRIV_REC_NEXT(12883, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12883(choice, ...) METALANG99_PRIV_REC_NEXT(12884, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12884(choice, ...) METALANG99_PRIV_REC_NEXT(12885, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12885(choice, ...) METALANG99_PRIV_REC_NEXT(12886, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12886(choice, ...) METALANG99_PRIV_REC_NEXT(12887, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12887(choice, ...) METALANG99_PRIV_REC_NEXT(12888, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12888(choice, ...) METALANG99_PRIV_REC_NEXT(12889, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12889(choice, ...) METALANG99_PRIV_REC_NEXT(12890, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12890(choice, ...) METALANG99_PRIV_REC_NEXT(12891, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12891(choice, ...) METALANG99_PRIV_REC_NEXT(12892, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12892(choice, ...) METALANG99_PRIV_REC_NEXT(12893, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12893(choice, ...) METALANG99_PRIV_REC_NEXT(12894, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12894(choice, ...) METALANG99_PRIV_REC_NEXT(12895, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12895(choice, ...) METALANG99_PRIV_REC_NEXT(12896, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12896(choice, ...) METALANG99_PRIV_REC_NEXT(12897, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12897(choice, ...) METALANG99_PRIV_REC_NEXT(12898, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12898(choice, ...) METALANG99_PRIV_REC_NEXT(12899, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12899(choice, ...) METALANG99_PRIV_REC_NEXT(12900, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12900(choice, ...) METALANG99_PRIV_REC_NEXT(12901, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12901(choice, ...) METALANG99_PRIV_REC_NEXT(12902, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12902(choice, ...) METALANG99_PRIV_REC_NEXT(12903, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12903(choice, ...) METALANG99_PRIV_REC_NEXT(12904, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12904(choice, ...) METALANG99_PRIV_REC_NEXT(12905, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12905(choice, ...) METALANG99_PRIV_REC_NEXT(12906, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12906(choice, ...) METALANG99_PRIV_REC_NEXT(12907, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12907(choice, ...) METALANG99_PRIV_REC_NEXT(12908, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12908(choice, ...) METALANG99_PRIV_REC_NEXT(12909, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12909(choice, ...) METALANG99_PRIV_REC_NEXT(12910, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12910(choice, ...) METALANG99_PRIV_REC_NEXT(12911, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12911(choice, ...) METALANG99_PRIV_REC_NEXT(12912, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12912(choice, ...) METALANG99_PRIV_REC_NEXT(12913, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12913(choice, ...) METALANG99_PRIV_REC_NEXT(12914, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12914(choice, ...) METALANG99_PRIV_REC_NEXT(12915, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12915(choice, ...) METALANG99_PRIV_REC_NEXT(12916, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12916(choice, ...) METALANG99_PRIV_REC_NEXT(12917, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12917(choice, ...) METALANG99_PRIV_REC_NEXT(12918, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12918(choice, ...) METALANG99_PRIV_REC_NEXT(12919, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12919(choice, ...) METALANG99_PRIV_REC_NEXT(12920, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12920(choice, ...) METALANG99_PRIV_REC_NEXT(12921, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12921(choice, ...) METALANG99_PRIV_REC_NEXT(12922, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12922(choice, ...) METALANG99_PRIV_REC_NEXT(12923, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12923(choice, ...) METALANG99_PRIV_REC_NEXT(12924, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12924(choice, ...) METALANG99_PRIV_REC_NEXT(12925, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12925(choice, ...) METALANG99_PRIV_REC_NEXT(12926, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12926(choice, ...) METALANG99_PRIV_REC_NEXT(12927, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12927(choice, ...) METALANG99_PRIV_REC_NEXT(12928, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12928(choice, ...) METALANG99_PRIV_REC_NEXT(12929, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12929(choice, ...) METALANG99_PRIV_REC_NEXT(12930, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12930(choice, ...) METALANG99_PRIV_REC_NEXT(12931, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12931(choice, ...) METALANG99_PRIV_REC_NEXT(12932, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12932(choice, ...) METALANG99_PRIV_REC_NEXT(12933, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12933(choice, ...) METALANG99_PRIV_REC_NEXT(12934, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12934(choice, ...) METALANG99_PRIV_REC_NEXT(12935, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12935(choice, ...) METALANG99_PRIV_REC_NEXT(12936, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12936(choice, ...) METALANG99_PRIV_REC_NEXT(12937, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12937(choice, ...) METALANG99_PRIV_REC_NEXT(12938, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12938(choice, ...) METALANG99_PRIV_REC_NEXT(12939, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12939(choice, ...) METALANG99_PRIV_REC_NEXT(12940, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12940(choice, ...) METALANG99_PRIV_REC_NEXT(12941, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12941(choice, ...) METALANG99_PRIV_REC_NEXT(12942, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12942(choice, ...) METALANG99_PRIV_REC_NEXT(12943, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12943(choice, ...) METALANG99_PRIV_REC_NEXT(12944, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12944(choice, ...) METALANG99_PRIV_REC_NEXT(12945, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12945(choice, ...) METALANG99_PRIV_REC_NEXT(12946, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12946(choice, ...) METALANG99_PRIV_REC_NEXT(12947, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12947(choice, ...) METALANG99_PRIV_REC_NEXT(12948, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12948(choice, ...) METALANG99_PRIV_REC_NEXT(12949, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12949(choice, ...) METALANG99_PRIV_REC_NEXT(12950, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12950(choice, ...) METALANG99_PRIV_REC_NEXT(12951, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12951(choice, ...) METALANG99_PRIV_REC_NEXT(12952, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12952(choice, ...) METALANG99_PRIV_REC_NEXT(12953, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12953(choice, ...) METALANG99_PRIV_REC_NEXT(12954, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12954(choice, ...) METALANG99_PRIV_REC_NEXT(12955, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12955(choice, ...) METALANG99_PRIV_REC_NEXT(12956, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12956(choice, ...) METALANG99_PRIV_REC_NEXT(12957, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12957(choice, ...) METALANG99_PRIV_REC_NEXT(12958, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12958(choice, ...) METALANG99_PRIV_REC_NEXT(12959, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12959(choice, ...) METALANG99_PRIV_REC_NEXT(12960, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12960(choice, ...) METALANG99_PRIV_REC_NEXT(12961, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12961(choice, ...) METALANG99_PRIV_REC_NEXT(12962, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12962(choice, ...) METALANG99_PRIV_REC_NEXT(12963, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12963(choice, ...) METALANG99_PRIV_REC_NEXT(12964, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12964(choice, ...) METALANG99_PRIV_REC_NEXT(12965, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12965(choice, ...) METALANG99_PRIV_REC_NEXT(12966, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12966(choice, ...) METALANG99_PRIV_REC_NEXT(12967, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12967(choice, ...) METALANG99_PRIV_REC_NEXT(12968, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12968(choice, ...) METALANG99_PRIV_REC_NEXT(12969, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12969(choice, ...) METALANG99_PRIV_REC_NEXT(12970, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12970(choice, ...) METALANG99_PRIV_REC_NEXT(12971, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12971(choice, ...) METALANG99_PRIV_REC_NEXT(12972, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12972(choice, ...) METALANG99_PRIV_REC_NEXT(12973, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12973(choice, ...) METALANG99_PRIV_REC_NEXT(12974, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12974(choice, ...) METALANG99_PRIV_REC_NEXT(12975, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12975(choice, ...) METALANG99_PRIV_REC_NEXT(12976, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12976(choice, ...) METALANG99_PRIV_REC_NEXT(12977, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12977(choice, ...) METALANG99_PRIV_REC_NEXT(12978, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12978(choice, ...) METALANG99_PRIV_REC_NEXT(12979, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12979(choice, ...) METALANG99_PRIV_REC_NEXT(12980, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12980(choice, ...) METALANG99_PRIV_REC_NEXT(12981, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12981(choice, ...) METALANG99_PRIV_REC_NEXT(12982, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12982(choice, ...) METALANG99_PRIV_REC_NEXT(12983, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12983(choice, ...) METALANG99_PRIV_REC_NEXT(12984, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12984(choice, ...) METALANG99_PRIV_REC_NEXT(12985, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12985(choice, ...) METALANG99_PRIV_REC_NEXT(12986, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12986(choice, ...) METALANG99_PRIV_REC_NEXT(12987, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12987(choice, ...) METALANG99_PRIV_REC_NEXT(12988, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12988(choice, ...) METALANG99_PRIV_REC_NEXT(12989, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12989(choice, ...) METALANG99_PRIV_REC_NEXT(12990, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12990(choice, ...) METALANG99_PRIV_REC_NEXT(12991, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12991(choice, ...) METALANG99_PRIV_REC_NEXT(12992, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12992(choice, ...) METALANG99_PRIV_REC_NEXT(12993, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12993(choice, ...) METALANG99_PRIV_REC_NEXT(12994, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12994(choice, ...) METALANG99_PRIV_REC_NEXT(12995, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12995(choice, ...) METALANG99_PRIV_REC_NEXT(12996, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12996(choice, ...) METALANG99_PRIV_REC_NEXT(12997, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12997(choice, ...) METALANG99_PRIV_REC_NEXT(12998, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12998(choice, ...) METALANG99_PRIV_REC_NEXT(12999, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_12999(choice, ...) METALANG99_PRIV_REC_NEXT(13000, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13000(choice, ...) METALANG99_PRIV_REC_NEXT(13001, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13001(choice, ...) METALANG99_PRIV_REC_NEXT(13002, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13002(choice, ...) METALANG99_PRIV_REC_NEXT(13003, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13003(choice, ...) METALANG99_PRIV_REC_NEXT(13004, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13004(choice, ...) METALANG99_PRIV_REC_NEXT(13005, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13005(choice, ...) METALANG99_PRIV_REC_NEXT(13006, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13006(choice, ...) METALANG99_PRIV_REC_NEXT(13007, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13007(choice, ...) METALANG99_PRIV_REC_NEXT(13008, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13008(choice, ...) METALANG99_PRIV_REC_NEXT(13009, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13009(choice, ...) METALANG99_PRIV_REC_NEXT(13010, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13010(choice, ...) METALANG99_PRIV_REC_NEXT(13011, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13011(choice, ...) METALANG99_PRIV_REC_NEXT(13012, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13012(choice, ...) METALANG99_PRIV_REC_NEXT(13013, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13013(choice, ...) METALANG99_PRIV_REC_NEXT(13014, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13014(choice, ...) METALANG99_PRIV_REC_NEXT(13015, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13015(choice, ...) METALANG99_PRIV_REC_NEXT(13016, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13016(choice, ...) METALANG99_PRIV_REC_NEXT(13017, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13017(choice, ...) METALANG99_PRIV_REC_NEXT(13018, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13018(choice, ...) METALANG99_PRIV_REC_NEXT(13019, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13019(choice, ...) METALANG99_PRIV_REC_NEXT(13020, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13020(choice, ...) METALANG99_PRIV_REC_NEXT(13021, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13021(choice, ...) METALANG99_PRIV_REC_NEXT(13022, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13022(choice, ...) METALANG99_PRIV_REC_NEXT(13023, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13023(choice, ...) METALANG99_PRIV_REC_NEXT(13024, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13024(choice, ...) METALANG99_PRIV_REC_NEXT(13025, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13025(choice, ...) METALANG99_PRIV_REC_NEXT(13026, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13026(choice, ...) METALANG99_PRIV_REC_NEXT(13027, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13027(choice, ...) METALANG99_PRIV_REC_NEXT(13028, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13028(choice, ...) METALANG99_PRIV_REC_NEXT(13029, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13029(choice, ...) METALANG99_PRIV_REC_NEXT(13030, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13030(choice, ...) METALANG99_PRIV_REC_NEXT(13031, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13031(choice, ...) METALANG99_PRIV_REC_NEXT(13032, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13032(choice, ...) METALANG99_PRIV_REC_NEXT(13033, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13033(choice, ...) METALANG99_PRIV_REC_NEXT(13034, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13034(choice, ...) METALANG99_PRIV_REC_NEXT(13035, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13035(choice, ...) METALANG99_PRIV_REC_NEXT(13036, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13036(choice, ...) METALANG99_PRIV_REC_NEXT(13037, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13037(choice, ...) METALANG99_PRIV_REC_NEXT(13038, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13038(choice, ...) METALANG99_PRIV_REC_NEXT(13039, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13039(choice, ...) METALANG99_PRIV_REC_NEXT(13040, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13040(choice, ...) METALANG99_PRIV_REC_NEXT(13041, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13041(choice, ...) METALANG99_PRIV_REC_NEXT(13042, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13042(choice, ...) METALANG99_PRIV_REC_NEXT(13043, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13043(choice, ...) METALANG99_PRIV_REC_NEXT(13044, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13044(choice, ...) METALANG99_PRIV_REC_NEXT(13045, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13045(choice, ...) METALANG99_PRIV_REC_NEXT(13046, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13046(choice, ...) METALANG99_PRIV_REC_NEXT(13047, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13047(choice, ...) METALANG99_PRIV_REC_NEXT(13048, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13048(choice, ...) METALANG99_PRIV_REC_NEXT(13049, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13049(choice, ...) METALANG99_PRIV_REC_NEXT(13050, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13050(choice, ...) METALANG99_PRIV_REC_NEXT(13051, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13051(choice, ...) METALANG99_PRIV_REC_NEXT(13052, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13052(choice, ...) METALANG99_PRIV_REC_NEXT(13053, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13053(choice, ...) METALANG99_PRIV_REC_NEXT(13054, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13054(choice, ...) METALANG99_PRIV_REC_NEXT(13055, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13055(choice, ...) METALANG99_PRIV_REC_NEXT(13056, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13056(choice, ...) METALANG99_PRIV_REC_NEXT(13057, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13057(choice, ...) METALANG99_PRIV_REC_NEXT(13058, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13058(choice, ...) METALANG99_PRIV_REC_NEXT(13059, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13059(choice, ...) METALANG99_PRIV_REC_NEXT(13060, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13060(choice, ...) METALANG99_PRIV_REC_NEXT(13061, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13061(choice, ...) METALANG99_PRIV_REC_NEXT(13062, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13062(choice, ...) METALANG99_PRIV_REC_NEXT(13063, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13063(choice, ...) METALANG99_PRIV_REC_NEXT(13064, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13064(choice, ...) METALANG99_PRIV_REC_NEXT(13065, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13065(choice, ...) METALANG99_PRIV_REC_NEXT(13066, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13066(choice, ...) METALANG99_PRIV_REC_NEXT(13067, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13067(choice, ...) METALANG99_PRIV_REC_NEXT(13068, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13068(choice, ...) METALANG99_PRIV_REC_NEXT(13069, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13069(choice, ...) METALANG99_PRIV_REC_NEXT(13070, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13070(choice, ...) METALANG99_PRIV_REC_NEXT(13071, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13071(choice, ...) METALANG99_PRIV_REC_NEXT(13072, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13072(choice, ...) METALANG99_PRIV_REC_NEXT(13073, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13073(choice, ...) METALANG99_PRIV_REC_NEXT(13074, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13074(choice, ...) METALANG99_PRIV_REC_NEXT(13075, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13075(choice, ...) METALANG99_PRIV_REC_NEXT(13076, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13076(choice, ...) METALANG99_PRIV_REC_NEXT(13077, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13077(choice, ...) METALANG99_PRIV_REC_NEXT(13078, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13078(choice, ...) METALANG99_PRIV_REC_NEXT(13079, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13079(choice, ...) METALANG99_PRIV_REC_NEXT(13080, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13080(choice, ...) METALANG99_PRIV_REC_NEXT(13081, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13081(choice, ...) METALANG99_PRIV_REC_NEXT(13082, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13082(choice, ...) METALANG99_PRIV_REC_NEXT(13083, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13083(choice, ...) METALANG99_PRIV_REC_NEXT(13084, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13084(choice, ...) METALANG99_PRIV_REC_NEXT(13085, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13085(choice, ...) METALANG99_PRIV_REC_NEXT(13086, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13086(choice, ...) METALANG99_PRIV_REC_NEXT(13087, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13087(choice, ...) METALANG99_PRIV_REC_NEXT(13088, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13088(choice, ...) METALANG99_PRIV_REC_NEXT(13089, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13089(choice, ...) METALANG99_PRIV_REC_NEXT(13090, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13090(choice, ...) METALANG99_PRIV_REC_NEXT(13091, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13091(choice, ...) METALANG99_PRIV_REC_NEXT(13092, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13092(choice, ...) METALANG99_PRIV_REC_NEXT(13093, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13093(choice, ...) METALANG99_PRIV_REC_NEXT(13094, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13094(choice, ...) METALANG99_PRIV_REC_NEXT(13095, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13095(choice, ...) METALANG99_PRIV_REC_NEXT(13096, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13096(choice, ...) METALANG99_PRIV_REC_NEXT(13097, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13097(choice, ...) METALANG99_PRIV_REC_NEXT(13098, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13098(choice, ...) METALANG99_PRIV_REC_NEXT(13099, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13099(choice, ...) METALANG99_PRIV_REC_NEXT(13100, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13100(choice, ...) METALANG99_PRIV_REC_NEXT(13101, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13101(choice, ...) METALANG99_PRIV_REC_NEXT(13102, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13102(choice, ...) METALANG99_PRIV_REC_NEXT(13103, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13103(choice, ...) METALANG99_PRIV_REC_NEXT(13104, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13104(choice, ...) METALANG99_PRIV_REC_NEXT(13105, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13105(choice, ...) METALANG99_PRIV_REC_NEXT(13106, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13106(choice, ...) METALANG99_PRIV_REC_NEXT(13107, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13107(choice, ...) METALANG99_PRIV_REC_NEXT(13108, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13108(choice, ...) METALANG99_PRIV_REC_NEXT(13109, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13109(choice, ...) METALANG99_PRIV_REC_NEXT(13110, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13110(choice, ...) METALANG99_PRIV_REC_NEXT(13111, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13111(choice, ...) METALANG99_PRIV_REC_NEXT(13112, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13112(choice, ...) METALANG99_PRIV_REC_NEXT(13113, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13113(choice, ...) METALANG99_PRIV_REC_NEXT(13114, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13114(choice, ...) METALANG99_PRIV_REC_NEXT(13115, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13115(choice, ...) METALANG99_PRIV_REC_NEXT(13116, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13116(choice, ...) METALANG99_PRIV_REC_NEXT(13117, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13117(choice, ...) METALANG99_PRIV_REC_NEXT(13118, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13118(choice, ...) METALANG99_PRIV_REC_NEXT(13119, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13119(choice, ...) METALANG99_PRIV_REC_NEXT(13120, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13120(choice, ...) METALANG99_PRIV_REC_NEXT(13121, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13121(choice, ...) METALANG99_PRIV_REC_NEXT(13122, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13122(choice, ...) METALANG99_PRIV_REC_NEXT(13123, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13123(choice, ...) METALANG99_PRIV_REC_NEXT(13124, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13124(choice, ...) METALANG99_PRIV_REC_NEXT(13125, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13125(choice, ...) METALANG99_PRIV_REC_NEXT(13126, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13126(choice, ...) METALANG99_PRIV_REC_NEXT(13127, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13127(choice, ...) METALANG99_PRIV_REC_NEXT(13128, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13128(choice, ...) METALANG99_PRIV_REC_NEXT(13129, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13129(choice, ...) METALANG99_PRIV_REC_NEXT(13130, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13130(choice, ...) METALANG99_PRIV_REC_NEXT(13131, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13131(choice, ...) METALANG99_PRIV_REC_NEXT(13132, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13132(choice, ...) METALANG99_PRIV_REC_NEXT(13133, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13133(choice, ...) METALANG99_PRIV_REC_NEXT(13134, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13134(choice, ...) METALANG99_PRIV_REC_NEXT(13135, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13135(choice, ...) METALANG99_PRIV_REC_NEXT(13136, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13136(choice, ...) METALANG99_PRIV_REC_NEXT(13137, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13137(choice, ...) METALANG99_PRIV_REC_NEXT(13138, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13138(choice, ...) METALANG99_PRIV_REC_NEXT(13139, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13139(choice, ...) METALANG99_PRIV_REC_NEXT(13140, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13140(choice, ...) METALANG99_PRIV_REC_NEXT(13141, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13141(choice, ...) METALANG99_PRIV_REC_NEXT(13142, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13142(choice, ...) METALANG99_PRIV_REC_NEXT(13143, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13143(choice, ...) METALANG99_PRIV_REC_NEXT(13144, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13144(choice, ...) METALANG99_PRIV_REC_NEXT(13145, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13145(choice, ...) METALANG99_PRIV_REC_NEXT(13146, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13146(choice, ...) METALANG99_PRIV_REC_NEXT(13147, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13147(choice, ...) METALANG99_PRIV_REC_NEXT(13148, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13148(choice, ...) METALANG99_PRIV_REC_NEXT(13149, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13149(choice, ...) METALANG99_PRIV_REC_NEXT(13150, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13150(choice, ...) METALANG99_PRIV_REC_NEXT(13151, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13151(choice, ...) METALANG99_PRIV_REC_NEXT(13152, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13152(choice, ...) METALANG99_PRIV_REC_NEXT(13153, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13153(choice, ...) METALANG99_PRIV_REC_NEXT(13154, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13154(choice, ...) METALANG99_PRIV_REC_NEXT(13155, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13155(choice, ...) METALANG99_PRIV_REC_NEXT(13156, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13156(choice, ...) METALANG99_PRIV_REC_NEXT(13157, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13157(choice, ...) METALANG99_PRIV_REC_NEXT(13158, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13158(choice, ...) METALANG99_PRIV_REC_NEXT(13159, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13159(choice, ...) METALANG99_PRIV_REC_NEXT(13160, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13160(choice, ...) METALANG99_PRIV_REC_NEXT(13161, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13161(choice, ...) METALANG99_PRIV_REC_NEXT(13162, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13162(choice, ...) METALANG99_PRIV_REC_NEXT(13163, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13163(choice, ...) METALANG99_PRIV_REC_NEXT(13164, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13164(choice, ...) METALANG99_PRIV_REC_NEXT(13165, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13165(choice, ...) METALANG99_PRIV_REC_NEXT(13166, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13166(choice, ...) METALANG99_PRIV_REC_NEXT(13167, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13167(choice, ...) METALANG99_PRIV_REC_NEXT(13168, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13168(choice, ...) METALANG99_PRIV_REC_NEXT(13169, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13169(choice, ...) METALANG99_PRIV_REC_NEXT(13170, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13170(choice, ...) METALANG99_PRIV_REC_NEXT(13171, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13171(choice, ...) METALANG99_PRIV_REC_NEXT(13172, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13172(choice, ...) METALANG99_PRIV_REC_NEXT(13173, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13173(choice, ...) METALANG99_PRIV_REC_NEXT(13174, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13174(choice, ...) METALANG99_PRIV_REC_NEXT(13175, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13175(choice, ...) METALANG99_PRIV_REC_NEXT(13176, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13176(choice, ...) METALANG99_PRIV_REC_NEXT(13177, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13177(choice, ...) METALANG99_PRIV_REC_NEXT(13178, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13178(choice, ...) METALANG99_PRIV_REC_NEXT(13179, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13179(choice, ...) METALANG99_PRIV_REC_NEXT(13180, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13180(choice, ...) METALANG99_PRIV_REC_NEXT(13181, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13181(choice, ...) METALANG99_PRIV_REC_NEXT(13182, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13182(choice, ...) METALANG99_PRIV_REC_NEXT(13183, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13183(choice, ...) METALANG99_PRIV_REC_NEXT(13184, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13184(choice, ...) METALANG99_PRIV_REC_NEXT(13185, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13185(choice, ...) METALANG99_PRIV_REC_NEXT(13186, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13186(choice, ...) METALANG99_PRIV_REC_NEXT(13187, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13187(choice, ...) METALANG99_PRIV_REC_NEXT(13188, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13188(choice, ...) METALANG99_PRIV_REC_NEXT(13189, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13189(choice, ...) METALANG99_PRIV_REC_NEXT(13190, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13190(choice, ...) METALANG99_PRIV_REC_NEXT(13191, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13191(choice, ...) METALANG99_PRIV_REC_NEXT(13192, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13192(choice, ...) METALANG99_PRIV_REC_NEXT(13193, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13193(choice, ...) METALANG99_PRIV_REC_NEXT(13194, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13194(choice, ...) METALANG99_PRIV_REC_NEXT(13195, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13195(choice, ...) METALANG99_PRIV_REC_NEXT(13196, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13196(choice, ...) METALANG99_PRIV_REC_NEXT(13197, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13197(choice, ...) METALANG99_PRIV_REC_NEXT(13198, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13198(choice, ...) METALANG99_PRIV_REC_NEXT(13199, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13199(choice, ...) METALANG99_PRIV_REC_NEXT(13200, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13200(choice, ...) METALANG99_PRIV_REC_NEXT(13201, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13201(choice, ...) METALANG99_PRIV_REC_NEXT(13202, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13202(choice, ...) METALANG99_PRIV_REC_NEXT(13203, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13203(choice, ...) METALANG99_PRIV_REC_NEXT(13204, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13204(choice, ...) METALANG99_PRIV_REC_NEXT(13205, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13205(choice, ...) METALANG99_PRIV_REC_NEXT(13206, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13206(choice, ...) METALANG99_PRIV_REC_NEXT(13207, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13207(choice, ...) METALANG99_PRIV_REC_NEXT(13208, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13208(choice, ...) METALANG99_PRIV_REC_NEXT(13209, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13209(choice, ...) METALANG99_PRIV_REC_NEXT(13210, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13210(choice, ...) METALANG99_PRIV_REC_NEXT(13211, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13211(choice, ...) METALANG99_PRIV_REC_NEXT(13212, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13212(choice, ...) METALANG99_PRIV_REC_NEXT(13213, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13213(choice, ...) METALANG99_PRIV_REC_NEXT(13214, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13214(choice, ...) METALANG99_PRIV_REC_NEXT(13215, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13215(choice, ...) METALANG99_PRIV_REC_NEXT(13216, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13216(choice, ...) METALANG99_PRIV_REC_NEXT(13217, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13217(choice, ...) METALANG99_PRIV_REC_NEXT(13218, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13218(choice, ...) METALANG99_PRIV_REC_NEXT(13219, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13219(choice, ...) METALANG99_PRIV_REC_NEXT(13220, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13220(choice, ...) METALANG99_PRIV_REC_NEXT(13221, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13221(choice, ...) METALANG99_PRIV_REC_NEXT(13222, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13222(choice, ...) METALANG99_PRIV_REC_NEXT(13223, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13223(choice, ...) METALANG99_PRIV_REC_NEXT(13224, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13224(choice, ...) METALANG99_PRIV_REC_NEXT(13225, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13225(choice, ...) METALANG99_PRIV_REC_NEXT(13226, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13226(choice, ...) METALANG99_PRIV_REC_NEXT(13227, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13227(choice, ...) METALANG99_PRIV_REC_NEXT(13228, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13228(choice, ...) METALANG99_PRIV_REC_NEXT(13229, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13229(choice, ...) METALANG99_PRIV_REC_NEXT(13230, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13230(choice, ...) METALANG99_PRIV_REC_NEXT(13231, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13231(choice, ...) METALANG99_PRIV_REC_NEXT(13232, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13232(choice, ...) METALANG99_PRIV_REC_NEXT(13233, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13233(choice, ...) METALANG99_PRIV_REC_NEXT(13234, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13234(choice, ...) METALANG99_PRIV_REC_NEXT(13235, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13235(choice, ...) METALANG99_PRIV_REC_NEXT(13236, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13236(choice, ...) METALANG99_PRIV_REC_NEXT(13237, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13237(choice, ...) METALANG99_PRIV_REC_NEXT(13238, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13238(choice, ...) METALANG99_PRIV_REC_NEXT(13239, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13239(choice, ...) METALANG99_PRIV_REC_NEXT(13240, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13240(choice, ...) METALANG99_PRIV_REC_NEXT(13241, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13241(choice, ...) METALANG99_PRIV_REC_NEXT(13242, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13242(choice, ...) METALANG99_PRIV_REC_NEXT(13243, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13243(choice, ...) METALANG99_PRIV_REC_NEXT(13244, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13244(choice, ...) METALANG99_PRIV_REC_NEXT(13245, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13245(choice, ...) METALANG99_PRIV_REC_NEXT(13246, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13246(choice, ...) METALANG99_PRIV_REC_NEXT(13247, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13247(choice, ...) METALANG99_PRIV_REC_NEXT(13248, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13248(choice, ...) METALANG99_PRIV_REC_NEXT(13249, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13249(choice, ...) METALANG99_PRIV_REC_NEXT(13250, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13250(choice, ...) METALANG99_PRIV_REC_NEXT(13251, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13251(choice, ...) METALANG99_PRIV_REC_NEXT(13252, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13252(choice, ...) METALANG99_PRIV_REC_NEXT(13253, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13253(choice, ...) METALANG99_PRIV_REC_NEXT(13254, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13254(choice, ...) METALANG99_PRIV_REC_NEXT(13255, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13255(choice, ...) METALANG99_PRIV_REC_NEXT(13256, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13256(choice, ...) METALANG99_PRIV_REC_NEXT(13257, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13257(choice, ...) METALANG99_PRIV_REC_NEXT(13258, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13258(choice, ...) METALANG99_PRIV_REC_NEXT(13259, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13259(choice, ...) METALANG99_PRIV_REC_NEXT(13260, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13260(choice, ...) METALANG99_PRIV_REC_NEXT(13261, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13261(choice, ...) METALANG99_PRIV_REC_NEXT(13262, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13262(choice, ...) METALANG99_PRIV_REC_NEXT(13263, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13263(choice, ...) METALANG99_PRIV_REC_NEXT(13264, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13264(choice, ...) METALANG99_PRIV_REC_NEXT(13265, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13265(choice, ...) METALANG99_PRIV_REC_NEXT(13266, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13266(choice, ...) METALANG99_PRIV_REC_NEXT(13267, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13267(choice, ...) METALANG99_PRIV_REC_NEXT(13268, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13268(choice, ...) METALANG99_PRIV_REC_NEXT(13269, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13269(choice, ...) METALANG99_PRIV_REC_NEXT(13270, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13270(choice, ...) METALANG99_PRIV_REC_NEXT(13271, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13271(choice, ...) METALANG99_PRIV_REC_NEXT(13272, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13272(choice, ...) METALANG99_PRIV_REC_NEXT(13273, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13273(choice, ...) METALANG99_PRIV_REC_NEXT(13274, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13274(choice, ...) METALANG99_PRIV_REC_NEXT(13275, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13275(choice, ...) METALANG99_PRIV_REC_NEXT(13276, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13276(choice, ...) METALANG99_PRIV_REC_NEXT(13277, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13277(choice, ...) METALANG99_PRIV_REC_NEXT(13278, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13278(choice, ...) METALANG99_PRIV_REC_NEXT(13279, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13279(choice, ...) METALANG99_PRIV_REC_NEXT(13280, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13280(choice, ...) METALANG99_PRIV_REC_NEXT(13281, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13281(choice, ...) METALANG99_PRIV_REC_NEXT(13282, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13282(choice, ...) METALANG99_PRIV_REC_NEXT(13283, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13283(choice, ...) METALANG99_PRIV_REC_NEXT(13284, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13284(choice, ...) METALANG99_PRIV_REC_NEXT(13285, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13285(choice, ...) METALANG99_PRIV_REC_NEXT(13286, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13286(choice, ...) METALANG99_PRIV_REC_NEXT(13287, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13287(choice, ...) METALANG99_PRIV_REC_NEXT(13288, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13288(choice, ...) METALANG99_PRIV_REC_NEXT(13289, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13289(choice, ...) METALANG99_PRIV_REC_NEXT(13290, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13290(choice, ...) METALANG99_PRIV_REC_NEXT(13291, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13291(choice, ...) METALANG99_PRIV_REC_NEXT(13292, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13292(choice, ...) METALANG99_PRIV_REC_NEXT(13293, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13293(choice, ...) METALANG99_PRIV_REC_NEXT(13294, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13294(choice, ...) METALANG99_PRIV_REC_NEXT(13295, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13295(choice, ...) METALANG99_PRIV_REC_NEXT(13296, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13296(choice, ...) METALANG99_PRIV_REC_NEXT(13297, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13297(choice, ...) METALANG99_PRIV_REC_NEXT(13298, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13298(choice, ...) METALANG99_PRIV_REC_NEXT(13299, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13299(choice, ...) METALANG99_PRIV_REC_NEXT(13300, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13300(choice, ...) METALANG99_PRIV_REC_NEXT(13301, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13301(choice, ...) METALANG99_PRIV_REC_NEXT(13302, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13302(choice, ...) METALANG99_PRIV_REC_NEXT(13303, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13303(choice, ...) METALANG99_PRIV_REC_NEXT(13304, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13304(choice, ...) METALANG99_PRIV_REC_NEXT(13305, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13305(choice, ...) METALANG99_PRIV_REC_NEXT(13306, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13306(choice, ...) METALANG99_PRIV_REC_NEXT(13307, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13307(choice, ...) METALANG99_PRIV_REC_NEXT(13308, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13308(choice, ...) METALANG99_PRIV_REC_NEXT(13309, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13309(choice, ...) METALANG99_PRIV_REC_NEXT(13310, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13310(choice, ...) METALANG99_PRIV_REC_NEXT(13311, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13311(choice, ...) METALANG99_PRIV_REC_NEXT(13312, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13312(choice, ...) METALANG99_PRIV_REC_NEXT(13313, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13313(choice, ...) METALANG99_PRIV_REC_NEXT(13314, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13314(choice, ...) METALANG99_PRIV_REC_NEXT(13315, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13315(choice, ...) METALANG99_PRIV_REC_NEXT(13316, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13316(choice, ...) METALANG99_PRIV_REC_NEXT(13317, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13317(choice, ...) METALANG99_PRIV_REC_NEXT(13318, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13318(choice, ...) METALANG99_PRIV_REC_NEXT(13319, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13319(choice, ...) METALANG99_PRIV_REC_NEXT(13320, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13320(choice, ...) METALANG99_PRIV_REC_NEXT(13321, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13321(choice, ...) METALANG99_PRIV_REC_NEXT(13322, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13322(choice, ...) METALANG99_PRIV_REC_NEXT(13323, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13323(choice, ...) METALANG99_PRIV_REC_NEXT(13324, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13324(choice, ...) METALANG99_PRIV_REC_NEXT(13325, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13325(choice, ...) METALANG99_PRIV_REC_NEXT(13326, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13326(choice, ...) METALANG99_PRIV_REC_NEXT(13327, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13327(choice, ...) METALANG99_PRIV_REC_NEXT(13328, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13328(choice, ...) METALANG99_PRIV_REC_NEXT(13329, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13329(choice, ...) METALANG99_PRIV_REC_NEXT(13330, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13330(choice, ...) METALANG99_PRIV_REC_NEXT(13331, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13331(choice, ...) METALANG99_PRIV_REC_NEXT(13332, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13332(choice, ...) METALANG99_PRIV_REC_NEXT(13333, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13333(choice, ...) METALANG99_PRIV_REC_NEXT(13334, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13334(choice, ...) METALANG99_PRIV_REC_NEXT(13335, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13335(choice, ...) METALANG99_PRIV_REC_NEXT(13336, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13336(choice, ...) METALANG99_PRIV_REC_NEXT(13337, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13337(choice, ...) METALANG99_PRIV_REC_NEXT(13338, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13338(choice, ...) METALANG99_PRIV_REC_NEXT(13339, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13339(choice, ...) METALANG99_PRIV_REC_NEXT(13340, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13340(choice, ...) METALANG99_PRIV_REC_NEXT(13341, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13341(choice, ...) METALANG99_PRIV_REC_NEXT(13342, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13342(choice, ...) METALANG99_PRIV_REC_NEXT(13343, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13343(choice, ...) METALANG99_PRIV_REC_NEXT(13344, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13344(choice, ...) METALANG99_PRIV_REC_NEXT(13345, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13345(choice, ...) METALANG99_PRIV_REC_NEXT(13346, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13346(choice, ...) METALANG99_PRIV_REC_NEXT(13347, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13347(choice, ...) METALANG99_PRIV_REC_NEXT(13348, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13348(choice, ...) METALANG99_PRIV_REC_NEXT(13349, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13349(choice, ...) METALANG99_PRIV_REC_NEXT(13350, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13350(choice, ...) METALANG99_PRIV_REC_NEXT(13351, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13351(choice, ...) METALANG99_PRIV_REC_NEXT(13352, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13352(choice, ...) METALANG99_PRIV_REC_NEXT(13353, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13353(choice, ...) METALANG99_PRIV_REC_NEXT(13354, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13354(choice, ...) METALANG99_PRIV_REC_NEXT(13355, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13355(choice, ...) METALANG99_PRIV_REC_NEXT(13356, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13356(choice, ...) METALANG99_PRIV_REC_NEXT(13357, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13357(choice, ...) METALANG99_PRIV_REC_NEXT(13358, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13358(choice, ...) METALANG99_PRIV_REC_NEXT(13359, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13359(choice, ...) METALANG99_PRIV_REC_NEXT(13360, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13360(choice, ...) METALANG99_PRIV_REC_NEXT(13361, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13361(choice, ...) METALANG99_PRIV_REC_NEXT(13362, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13362(choice, ...) METALANG99_PRIV_REC_NEXT(13363, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13363(choice, ...) METALANG99_PRIV_REC_NEXT(13364, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13364(choice, ...) METALANG99_PRIV_REC_NEXT(13365, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13365(choice, ...) METALANG99_PRIV_REC_NEXT(13366, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13366(choice, ...) METALANG99_PRIV_REC_NEXT(13367, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13367(choice, ...) METALANG99_PRIV_REC_NEXT(13368, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13368(choice, ...) METALANG99_PRIV_REC_NEXT(13369, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13369(choice, ...) METALANG99_PRIV_REC_NEXT(13370, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13370(choice, ...) METALANG99_PRIV_REC_NEXT(13371, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13371(choice, ...) METALANG99_PRIV_REC_NEXT(13372, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13372(choice, ...) METALANG99_PRIV_REC_NEXT(13373, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13373(choice, ...) METALANG99_PRIV_REC_NEXT(13374, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13374(choice, ...) METALANG99_PRIV_REC_NEXT(13375, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13375(choice, ...) METALANG99_PRIV_REC_NEXT(13376, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13376(choice, ...) METALANG99_PRIV_REC_NEXT(13377, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13377(choice, ...) METALANG99_PRIV_REC_NEXT(13378, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13378(choice, ...) METALANG99_PRIV_REC_NEXT(13379, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13379(choice, ...) METALANG99_PRIV_REC_NEXT(13380, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13380(choice, ...) METALANG99_PRIV_REC_NEXT(13381, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13381(choice, ...) METALANG99_PRIV_REC_NEXT(13382, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13382(choice, ...) METALANG99_PRIV_REC_NEXT(13383, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13383(choice, ...) METALANG99_PRIV_REC_NEXT(13384, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13384(choice, ...) METALANG99_PRIV_REC_NEXT(13385, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13385(choice, ...) METALANG99_PRIV_REC_NEXT(13386, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13386(choice, ...) METALANG99_PRIV_REC_NEXT(13387, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13387(choice, ...) METALANG99_PRIV_REC_NEXT(13388, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13388(choice, ...) METALANG99_PRIV_REC_NEXT(13389, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13389(choice, ...) METALANG99_PRIV_REC_NEXT(13390, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13390(choice, ...) METALANG99_PRIV_REC_NEXT(13391, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13391(choice, ...) METALANG99_PRIV_REC_NEXT(13392, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13392(choice, ...) METALANG99_PRIV_REC_NEXT(13393, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13393(choice, ...) METALANG99_PRIV_REC_NEXT(13394, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13394(choice, ...) METALANG99_PRIV_REC_NEXT(13395, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13395(choice, ...) METALANG99_PRIV_REC_NEXT(13396, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13396(choice, ...) METALANG99_PRIV_REC_NEXT(13397, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13397(choice, ...) METALANG99_PRIV_REC_NEXT(13398, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13398(choice, ...) METALANG99_PRIV_REC_NEXT(13399, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13399(choice, ...) METALANG99_PRIV_REC_NEXT(13400, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13400(choice, ...) METALANG99_PRIV_REC_NEXT(13401, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13401(choice, ...) METALANG99_PRIV_REC_NEXT(13402, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13402(choice, ...) METALANG99_PRIV_REC_NEXT(13403, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13403(choice, ...) METALANG99_PRIV_REC_NEXT(13404, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13404(choice, ...) METALANG99_PRIV_REC_NEXT(13405, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13405(choice, ...) METALANG99_PRIV_REC_NEXT(13406, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13406(choice, ...) METALANG99_PRIV_REC_NEXT(13407, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13407(choice, ...) METALANG99_PRIV_REC_NEXT(13408, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13408(choice, ...) METALANG99_PRIV_REC_NEXT(13409, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13409(choice, ...) METALANG99_PRIV_REC_NEXT(13410, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13410(choice, ...) METALANG99_PRIV_REC_NEXT(13411, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13411(choice, ...) METALANG99_PRIV_REC_NEXT(13412, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13412(choice, ...) METALANG99_PRIV_REC_NEXT(13413, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13413(choice, ...) METALANG99_PRIV_REC_NEXT(13414, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13414(choice, ...) METALANG99_PRIV_REC_NEXT(13415, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13415(choice, ...) METALANG99_PRIV_REC_NEXT(13416, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13416(choice, ...) METALANG99_PRIV_REC_NEXT(13417, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13417(choice, ...) METALANG99_PRIV_REC_NEXT(13418, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13418(choice, ...) METALANG99_PRIV_REC_NEXT(13419, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13419(choice, ...) METALANG99_PRIV_REC_NEXT(13420, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13420(choice, ...) METALANG99_PRIV_REC_NEXT(13421, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13421(choice, ...) METALANG99_PRIV_REC_NEXT(13422, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13422(choice, ...) METALANG99_PRIV_REC_NEXT(13423, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13423(choice, ...) METALANG99_PRIV_REC_NEXT(13424, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13424(choice, ...) METALANG99_PRIV_REC_NEXT(13425, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13425(choice, ...) METALANG99_PRIV_REC_NEXT(13426, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13426(choice, ...) METALANG99_PRIV_REC_NEXT(13427, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13427(choice, ...) METALANG99_PRIV_REC_NEXT(13428, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13428(choice, ...) METALANG99_PRIV_REC_NEXT(13429, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13429(choice, ...) METALANG99_PRIV_REC_NEXT(13430, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13430(choice, ...) METALANG99_PRIV_REC_NEXT(13431, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13431(choice, ...) METALANG99_PRIV_REC_NEXT(13432, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13432(choice, ...) METALANG99_PRIV_REC_NEXT(13433, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13433(choice, ...) METALANG99_PRIV_REC_NEXT(13434, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13434(choice, ...) METALANG99_PRIV_REC_NEXT(13435, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13435(choice, ...) METALANG99_PRIV_REC_NEXT(13436, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13436(choice, ...) METALANG99_PRIV_REC_NEXT(13437, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13437(choice, ...) METALANG99_PRIV_REC_NEXT(13438, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13438(choice, ...) METALANG99_PRIV_REC_NEXT(13439, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13439(choice, ...) METALANG99_PRIV_REC_NEXT(13440, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13440(choice, ...) METALANG99_PRIV_REC_NEXT(13441, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13441(choice, ...) METALANG99_PRIV_REC_NEXT(13442, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13442(choice, ...) METALANG99_PRIV_REC_NEXT(13443, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13443(choice, ...) METALANG99_PRIV_REC_NEXT(13444, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13444(choice, ...) METALANG99_PRIV_REC_NEXT(13445, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13445(choice, ...) METALANG99_PRIV_REC_NEXT(13446, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13446(choice, ...) METALANG99_PRIV_REC_NEXT(13447, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13447(choice, ...) METALANG99_PRIV_REC_NEXT(13448, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13448(choice, ...) METALANG99_PRIV_REC_NEXT(13449, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13449(choice, ...) METALANG99_PRIV_REC_NEXT(13450, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13450(choice, ...) METALANG99_PRIV_REC_NEXT(13451, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13451(choice, ...) METALANG99_PRIV_REC_NEXT(13452, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13452(choice, ...) METALANG99_PRIV_REC_NEXT(13453, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13453(choice, ...) METALANG99_PRIV_REC_NEXT(13454, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13454(choice, ...) METALANG99_PRIV_REC_NEXT(13455, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13455(choice, ...) METALANG99_PRIV_REC_NEXT(13456, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13456(choice, ...) METALANG99_PRIV_REC_NEXT(13457, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13457(choice, ...) METALANG99_PRIV_REC_NEXT(13458, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13458(choice, ...) METALANG99_PRIV_REC_NEXT(13459, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13459(choice, ...) METALANG99_PRIV_REC_NEXT(13460, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13460(choice, ...) METALANG99_PRIV_REC_NEXT(13461, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13461(choice, ...) METALANG99_PRIV_REC_NEXT(13462, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13462(choice, ...) METALANG99_PRIV_REC_NEXT(13463, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13463(choice, ...) METALANG99_PRIV_REC_NEXT(13464, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13464(choice, ...) METALANG99_PRIV_REC_NEXT(13465, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13465(choice, ...) METALANG99_PRIV_REC_NEXT(13466, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13466(choice, ...) METALANG99_PRIV_REC_NEXT(13467, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13467(choice, ...) METALANG99_PRIV_REC_NEXT(13468, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13468(choice, ...) METALANG99_PRIV_REC_NEXT(13469, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13469(choice, ...) METALANG99_PRIV_REC_NEXT(13470, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13470(choice, ...) METALANG99_PRIV_REC_NEXT(13471, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13471(choice, ...) METALANG99_PRIV_REC_NEXT(13472, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13472(choice, ...) METALANG99_PRIV_REC_NEXT(13473, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13473(choice, ...) METALANG99_PRIV_REC_NEXT(13474, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13474(choice, ...) METALANG99_PRIV_REC_NEXT(13475, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13475(choice, ...) METALANG99_PRIV_REC_NEXT(13476, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13476(choice, ...) METALANG99_PRIV_REC_NEXT(13477, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13477(choice, ...) METALANG99_PRIV_REC_NEXT(13478, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13478(choice, ...) METALANG99_PRIV_REC_NEXT(13479, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13479(choice, ...) METALANG99_PRIV_REC_NEXT(13480, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13480(choice, ...) METALANG99_PRIV_REC_NEXT(13481, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13481(choice, ...) METALANG99_PRIV_REC_NEXT(13482, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13482(choice, ...) METALANG99_PRIV_REC_NEXT(13483, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13483(choice, ...) METALANG99_PRIV_REC_NEXT(13484, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13484(choice, ...) METALANG99_PRIV_REC_NEXT(13485, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13485(choice, ...) METALANG99_PRIV_REC_NEXT(13486, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13486(choice, ...) METALANG99_PRIV_REC_NEXT(13487, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13487(choice, ...) METALANG99_PRIV_REC_NEXT(13488, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13488(choice, ...) METALANG99_PRIV_REC_NEXT(13489, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13489(choice, ...) METALANG99_PRIV_REC_NEXT(13490, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13490(choice, ...) METALANG99_PRIV_REC_NEXT(13491, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13491(choice, ...) METALANG99_PRIV_REC_NEXT(13492, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13492(choice, ...) METALANG99_PRIV_REC_NEXT(13493, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13493(choice, ...) METALANG99_PRIV_REC_NEXT(13494, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13494(choice, ...) METALANG99_PRIV_REC_NEXT(13495, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13495(choice, ...) METALANG99_PRIV_REC_NEXT(13496, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13496(choice, ...) METALANG99_PRIV_REC_NEXT(13497, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13497(choice, ...) METALANG99_PRIV_REC_NEXT(13498, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13498(choice, ...) METALANG99_PRIV_REC_NEXT(13499, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13499(choice, ...) METALANG99_PRIV_REC_NEXT(13500, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13500(choice, ...) METALANG99_PRIV_REC_NEXT(13501, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13501(choice, ...) METALANG99_PRIV_REC_NEXT(13502, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13502(choice, ...) METALANG99_PRIV_REC_NEXT(13503, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13503(choice, ...) METALANG99_PRIV_REC_NEXT(13504, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13504(choice, ...) METALANG99_PRIV_REC_NEXT(13505, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13505(choice, ...) METALANG99_PRIV_REC_NEXT(13506, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13506(choice, ...) METALANG99_PRIV_REC_NEXT(13507, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13507(choice, ...) METALANG99_PRIV_REC_NEXT(13508, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13508(choice, ...) METALANG99_PRIV_REC_NEXT(13509, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13509(choice, ...) METALANG99_PRIV_REC_NEXT(13510, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13510(choice, ...) METALANG99_PRIV_REC_NEXT(13511, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13511(choice, ...) METALANG99_PRIV_REC_NEXT(13512, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13512(choice, ...) METALANG99_PRIV_REC_NEXT(13513, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13513(choice, ...) METALANG99_PRIV_REC_NEXT(13514, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13514(choice, ...) METALANG99_PRIV_REC_NEXT(13515, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13515(choice, ...) METALANG99_PRIV_REC_NEXT(13516, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13516(choice, ...) METALANG99_PRIV_REC_NEXT(13517, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13517(choice, ...) METALANG99_PRIV_REC_NEXT(13518, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13518(choice, ...) METALANG99_PRIV_REC_NEXT(13519, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13519(choice, ...) METALANG99_PRIV_REC_NEXT(13520, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13520(choice, ...) METALANG99_PRIV_REC_NEXT(13521, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13521(choice, ...) METALANG99_PRIV_REC_NEXT(13522, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13522(choice, ...) METALANG99_PRIV_REC_NEXT(13523, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13523(choice, ...) METALANG99_PRIV_REC_NEXT(13524, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13524(choice, ...) METALANG99_PRIV_REC_NEXT(13525, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13525(choice, ...) METALANG99_PRIV_REC_NEXT(13526, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13526(choice, ...) METALANG99_PRIV_REC_NEXT(13527, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13527(choice, ...) METALANG99_PRIV_REC_NEXT(13528, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13528(choice, ...) METALANG99_PRIV_REC_NEXT(13529, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13529(choice, ...) METALANG99_PRIV_REC_NEXT(13530, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13530(choice, ...) METALANG99_PRIV_REC_NEXT(13531, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13531(choice, ...) METALANG99_PRIV_REC_NEXT(13532, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13532(choice, ...) METALANG99_PRIV_REC_NEXT(13533, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13533(choice, ...) METALANG99_PRIV_REC_NEXT(13534, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13534(choice, ...) METALANG99_PRIV_REC_NEXT(13535, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13535(choice, ...) METALANG99_PRIV_REC_NEXT(13536, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13536(choice, ...) METALANG99_PRIV_REC_NEXT(13537, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13537(choice, ...) METALANG99_PRIV_REC_NEXT(13538, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13538(choice, ...) METALANG99_PRIV_REC_NEXT(13539, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13539(choice, ...) METALANG99_PRIV_REC_NEXT(13540, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13540(choice, ...) METALANG99_PRIV_REC_NEXT(13541, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13541(choice, ...) METALANG99_PRIV_REC_NEXT(13542, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13542(choice, ...) METALANG99_PRIV_REC_NEXT(13543, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13543(choice, ...) METALANG99_PRIV_REC_NEXT(13544, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13544(choice, ...) METALANG99_PRIV_REC_NEXT(13545, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13545(choice, ...) METALANG99_PRIV_REC_NEXT(13546, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13546(choice, ...) METALANG99_PRIV_REC_NEXT(13547, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13547(choice, ...) METALANG99_PRIV_REC_NEXT(13548, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13548(choice, ...) METALANG99_PRIV_REC_NEXT(13549, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13549(choice, ...) METALANG99_PRIV_REC_NEXT(13550, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13550(choice, ...) METALANG99_PRIV_REC_NEXT(13551, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13551(choice, ...) METALANG99_PRIV_REC_NEXT(13552, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13552(choice, ...) METALANG99_PRIV_REC_NEXT(13553, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13553(choice, ...) METALANG99_PRIV_REC_NEXT(13554, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13554(choice, ...) METALANG99_PRIV_REC_NEXT(13555, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13555(choice, ...) METALANG99_PRIV_REC_NEXT(13556, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13556(choice, ...) METALANG99_PRIV_REC_NEXT(13557, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13557(choice, ...) METALANG99_PRIV_REC_NEXT(13558, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13558(choice, ...) METALANG99_PRIV_REC_NEXT(13559, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13559(choice, ...) METALANG99_PRIV_REC_NEXT(13560, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13560(choice, ...) METALANG99_PRIV_REC_NEXT(13561, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13561(choice, ...) METALANG99_PRIV_REC_NEXT(13562, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13562(choice, ...) METALANG99_PRIV_REC_NEXT(13563, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13563(choice, ...) METALANG99_PRIV_REC_NEXT(13564, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13564(choice, ...) METALANG99_PRIV_REC_NEXT(13565, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13565(choice, ...) METALANG99_PRIV_REC_NEXT(13566, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13566(choice, ...) METALANG99_PRIV_REC_NEXT(13567, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13567(choice, ...) METALANG99_PRIV_REC_NEXT(13568, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13568(choice, ...) METALANG99_PRIV_REC_NEXT(13569, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13569(choice, ...) METALANG99_PRIV_REC_NEXT(13570, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13570(choice, ...) METALANG99_PRIV_REC_NEXT(13571, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13571(choice, ...) METALANG99_PRIV_REC_NEXT(13572, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13572(choice, ...) METALANG99_PRIV_REC_NEXT(13573, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13573(choice, ...) METALANG99_PRIV_REC_NEXT(13574, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13574(choice, ...) METALANG99_PRIV_REC_NEXT(13575, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13575(choice, ...) METALANG99_PRIV_REC_NEXT(13576, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13576(choice, ...) METALANG99_PRIV_REC_NEXT(13577, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13577(choice, ...) METALANG99_PRIV_REC_NEXT(13578, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13578(choice, ...) METALANG99_PRIV_REC_NEXT(13579, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13579(choice, ...) METALANG99_PRIV_REC_NEXT(13580, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13580(choice, ...) METALANG99_PRIV_REC_NEXT(13581, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13581(choice, ...) METALANG99_PRIV_REC_NEXT(13582, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13582(choice, ...) METALANG99_PRIV_REC_NEXT(13583, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13583(choice, ...) METALANG99_PRIV_REC_NEXT(13584, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13584(choice, ...) METALANG99_PRIV_REC_NEXT(13585, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13585(choice, ...) METALANG99_PRIV_REC_NEXT(13586, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13586(choice, ...) METALANG99_PRIV_REC_NEXT(13587, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13587(choice, ...) METALANG99_PRIV_REC_NEXT(13588, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13588(choice, ...) METALANG99_PRIV_REC_NEXT(13589, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13589(choice, ...) METALANG99_PRIV_REC_NEXT(13590, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13590(choice, ...) METALANG99_PRIV_REC_NEXT(13591, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13591(choice, ...) METALANG99_PRIV_REC_NEXT(13592, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13592(choice, ...) METALANG99_PRIV_REC_NEXT(13593, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13593(choice, ...) METALANG99_PRIV_REC_NEXT(13594, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13594(choice, ...) METALANG99_PRIV_REC_NEXT(13595, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13595(choice, ...) METALANG99_PRIV_REC_NEXT(13596, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13596(choice, ...) METALANG99_PRIV_REC_NEXT(13597, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13597(choice, ...) METALANG99_PRIV_REC_NEXT(13598, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13598(choice, ...) METALANG99_PRIV_REC_NEXT(13599, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13599(choice, ...) METALANG99_PRIV_REC_NEXT(13600, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13600(choice, ...) METALANG99_PRIV_REC_NEXT(13601, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13601(choice, ...) METALANG99_PRIV_REC_NEXT(13602, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13602(choice, ...) METALANG99_PRIV_REC_NEXT(13603, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13603(choice, ...) METALANG99_PRIV_REC_NEXT(13604, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13604(choice, ...) METALANG99_PRIV_REC_NEXT(13605, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13605(choice, ...) METALANG99_PRIV_REC_NEXT(13606, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13606(choice, ...) METALANG99_PRIV_REC_NEXT(13607, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13607(choice, ...) METALANG99_PRIV_REC_NEXT(13608, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13608(choice, ...) METALANG99_PRIV_REC_NEXT(13609, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13609(choice, ...) METALANG99_PRIV_REC_NEXT(13610, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13610(choice, ...) METALANG99_PRIV_REC_NEXT(13611, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13611(choice, ...) METALANG99_PRIV_REC_NEXT(13612, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13612(choice, ...) METALANG99_PRIV_REC_NEXT(13613, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13613(choice, ...) METALANG99_PRIV_REC_NEXT(13614, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13614(choice, ...) METALANG99_PRIV_REC_NEXT(13615, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13615(choice, ...) METALANG99_PRIV_REC_NEXT(13616, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13616(choice, ...) METALANG99_PRIV_REC_NEXT(13617, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13617(choice, ...) METALANG99_PRIV_REC_NEXT(13618, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13618(choice, ...) METALANG99_PRIV_REC_NEXT(13619, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13619(choice, ...) METALANG99_PRIV_REC_NEXT(13620, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13620(choice, ...) METALANG99_PRIV_REC_NEXT(13621, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13621(choice, ...) METALANG99_PRIV_REC_NEXT(13622, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13622(choice, ...) METALANG99_PRIV_REC_NEXT(13623, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13623(choice, ...) METALANG99_PRIV_REC_NEXT(13624, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13624(choice, ...) METALANG99_PRIV_REC_NEXT(13625, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13625(choice, ...) METALANG99_PRIV_REC_NEXT(13626, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13626(choice, ...) METALANG99_PRIV_REC_NEXT(13627, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13627(choice, ...) METALANG99_PRIV_REC_NEXT(13628, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13628(choice, ...) METALANG99_PRIV_REC_NEXT(13629, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13629(choice, ...) METALANG99_PRIV_REC_NEXT(13630, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13630(choice, ...) METALANG99_PRIV_REC_NEXT(13631, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13631(choice, ...) METALANG99_PRIV_REC_NEXT(13632, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13632(choice, ...) METALANG99_PRIV_REC_NEXT(13633, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13633(choice, ...) METALANG99_PRIV_REC_NEXT(13634, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13634(choice, ...) METALANG99_PRIV_REC_NEXT(13635, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13635(choice, ...) METALANG99_PRIV_REC_NEXT(13636, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13636(choice, ...) METALANG99_PRIV_REC_NEXT(13637, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13637(choice, ...) METALANG99_PRIV_REC_NEXT(13638, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13638(choice, ...) METALANG99_PRIV_REC_NEXT(13639, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13639(choice, ...) METALANG99_PRIV_REC_NEXT(13640, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13640(choice, ...) METALANG99_PRIV_REC_NEXT(13641, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13641(choice, ...) METALANG99_PRIV_REC_NEXT(13642, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13642(choice, ...) METALANG99_PRIV_REC_NEXT(13643, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13643(choice, ...) METALANG99_PRIV_REC_NEXT(13644, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13644(choice, ...) METALANG99_PRIV_REC_NEXT(13645, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13645(choice, ...) METALANG99_PRIV_REC_NEXT(13646, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13646(choice, ...) METALANG99_PRIV_REC_NEXT(13647, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13647(choice, ...) METALANG99_PRIV_REC_NEXT(13648, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13648(choice, ...) METALANG99_PRIV_REC_NEXT(13649, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13649(choice, ...) METALANG99_PRIV_REC_NEXT(13650, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13650(choice, ...) METALANG99_PRIV_REC_NEXT(13651, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13651(choice, ...) METALANG99_PRIV_REC_NEXT(13652, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13652(choice, ...) METALANG99_PRIV_REC_NEXT(13653, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13653(choice, ...) METALANG99_PRIV_REC_NEXT(13654, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13654(choice, ...) METALANG99_PRIV_REC_NEXT(13655, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13655(choice, ...) METALANG99_PRIV_REC_NEXT(13656, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13656(choice, ...) METALANG99_PRIV_REC_NEXT(13657, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13657(choice, ...) METALANG99_PRIV_REC_NEXT(13658, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13658(choice, ...) METALANG99_PRIV_REC_NEXT(13659, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13659(choice, ...) METALANG99_PRIV_REC_NEXT(13660, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13660(choice, ...) METALANG99_PRIV_REC_NEXT(13661, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13661(choice, ...) METALANG99_PRIV_REC_NEXT(13662, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13662(choice, ...) METALANG99_PRIV_REC_NEXT(13663, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13663(choice, ...) METALANG99_PRIV_REC_NEXT(13664, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13664(choice, ...) METALANG99_PRIV_REC_NEXT(13665, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13665(choice, ...) METALANG99_PRIV_REC_NEXT(13666, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13666(choice, ...) METALANG99_PRIV_REC_NEXT(13667, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13667(choice, ...) METALANG99_PRIV_REC_NEXT(13668, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13668(choice, ...) METALANG99_PRIV_REC_NEXT(13669, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13669(choice, ...) METALANG99_PRIV_REC_NEXT(13670, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13670(choice, ...) METALANG99_PRIV_REC_NEXT(13671, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13671(choice, ...) METALANG99_PRIV_REC_NEXT(13672, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13672(choice, ...) METALANG99_PRIV_REC_NEXT(13673, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13673(choice, ...) METALANG99_PRIV_REC_NEXT(13674, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13674(choice, ...) METALANG99_PRIV_REC_NEXT(13675, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13675(choice, ...) METALANG99_PRIV_REC_NEXT(13676, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13676(choice, ...) METALANG99_PRIV_REC_NEXT(13677, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13677(choice, ...) METALANG99_PRIV_REC_NEXT(13678, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13678(choice, ...) METALANG99_PRIV_REC_NEXT(13679, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13679(choice, ...) METALANG99_PRIV_REC_NEXT(13680, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13680(choice, ...) METALANG99_PRIV_REC_NEXT(13681, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13681(choice, ...) METALANG99_PRIV_REC_NEXT(13682, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13682(choice, ...) METALANG99_PRIV_REC_NEXT(13683, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13683(choice, ...) METALANG99_PRIV_REC_NEXT(13684, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13684(choice, ...) METALANG99_PRIV_REC_NEXT(13685, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13685(choice, ...) METALANG99_PRIV_REC_NEXT(13686, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13686(choice, ...) METALANG99_PRIV_REC_NEXT(13687, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13687(choice, ...) METALANG99_PRIV_REC_NEXT(13688, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13688(choice, ...) METALANG99_PRIV_REC_NEXT(13689, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13689(choice, ...) METALANG99_PRIV_REC_NEXT(13690, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13690(choice, ...) METALANG99_PRIV_REC_NEXT(13691, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13691(choice, ...) METALANG99_PRIV_REC_NEXT(13692, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13692(choice, ...) METALANG99_PRIV_REC_NEXT(13693, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13693(choice, ...) METALANG99_PRIV_REC_NEXT(13694, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13694(choice, ...) METALANG99_PRIV_REC_NEXT(13695, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13695(choice, ...) METALANG99_PRIV_REC_NEXT(13696, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13696(choice, ...) METALANG99_PRIV_REC_NEXT(13697, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13697(choice, ...) METALANG99_PRIV_REC_NEXT(13698, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13698(choice, ...) METALANG99_PRIV_REC_NEXT(13699, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13699(choice, ...) METALANG99_PRIV_REC_NEXT(13700, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13700(choice, ...) METALANG99_PRIV_REC_NEXT(13701, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13701(choice, ...) METALANG99_PRIV_REC_NEXT(13702, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13702(choice, ...) METALANG99_PRIV_REC_NEXT(13703, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13703(choice, ...) METALANG99_PRIV_REC_NEXT(13704, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13704(choice, ...) METALANG99_PRIV_REC_NEXT(13705, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13705(choice, ...) METALANG99_PRIV_REC_NEXT(13706, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13706(choice, ...) METALANG99_PRIV_REC_NEXT(13707, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13707(choice, ...) METALANG99_PRIV_REC_NEXT(13708, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13708(choice, ...) METALANG99_PRIV_REC_NEXT(13709, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13709(choice, ...) METALANG99_PRIV_REC_NEXT(13710, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13710(choice, ...) METALANG99_PRIV_REC_NEXT(13711, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13711(choice, ...) METALANG99_PRIV_REC_NEXT(13712, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13712(choice, ...) METALANG99_PRIV_REC_NEXT(13713, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13713(choice, ...) METALANG99_PRIV_REC_NEXT(13714, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13714(choice, ...) METALANG99_PRIV_REC_NEXT(13715, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13715(choice, ...) METALANG99_PRIV_REC_NEXT(13716, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13716(choice, ...) METALANG99_PRIV_REC_NEXT(13717, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13717(choice, ...) METALANG99_PRIV_REC_NEXT(13718, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13718(choice, ...) METALANG99_PRIV_REC_NEXT(13719, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13719(choice, ...) METALANG99_PRIV_REC_NEXT(13720, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13720(choice, ...) METALANG99_PRIV_REC_NEXT(13721, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13721(choice, ...) METALANG99_PRIV_REC_NEXT(13722, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13722(choice, ...) METALANG99_PRIV_REC_NEXT(13723, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13723(choice, ...) METALANG99_PRIV_REC_NEXT(13724, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13724(choice, ...) METALANG99_PRIV_REC_NEXT(13725, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13725(choice, ...) METALANG99_PRIV_REC_NEXT(13726, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13726(choice, ...) METALANG99_PRIV_REC_NEXT(13727, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13727(choice, ...) METALANG99_PRIV_REC_NEXT(13728, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13728(choice, ...) METALANG99_PRIV_REC_NEXT(13729, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13729(choice, ...) METALANG99_PRIV_REC_NEXT(13730, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13730(choice, ...) METALANG99_PRIV_REC_NEXT(13731, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13731(choice, ...) METALANG99_PRIV_REC_NEXT(13732, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13732(choice, ...) METALANG99_PRIV_REC_NEXT(13733, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13733(choice, ...) METALANG99_PRIV_REC_NEXT(13734, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13734(choice, ...) METALANG99_PRIV_REC_NEXT(13735, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13735(choice, ...) METALANG99_PRIV_REC_NEXT(13736, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13736(choice, ...) METALANG99_PRIV_REC_NEXT(13737, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13737(choice, ...) METALANG99_PRIV_REC_NEXT(13738, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13738(choice, ...) METALANG99_PRIV_REC_NEXT(13739, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13739(choice, ...) METALANG99_PRIV_REC_NEXT(13740, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13740(choice, ...) METALANG99_PRIV_REC_NEXT(13741, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13741(choice, ...) METALANG99_PRIV_REC_NEXT(13742, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13742(choice, ...) METALANG99_PRIV_REC_NEXT(13743, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13743(choice, ...) METALANG99_PRIV_REC_NEXT(13744, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13744(choice, ...) METALANG99_PRIV_REC_NEXT(13745, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13745(choice, ...) METALANG99_PRIV_REC_NEXT(13746, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13746(choice, ...) METALANG99_PRIV_REC_NEXT(13747, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13747(choice, ...) METALANG99_PRIV_REC_NEXT(13748, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13748(choice, ...) METALANG99_PRIV_REC_NEXT(13749, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13749(choice, ...) METALANG99_PRIV_REC_NEXT(13750, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13750(choice, ...) METALANG99_PRIV_REC_NEXT(13751, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13751(choice, ...) METALANG99_PRIV_REC_NEXT(13752, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13752(choice, ...) METALANG99_PRIV_REC_NEXT(13753, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13753(choice, ...) METALANG99_PRIV_REC_NEXT(13754, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13754(choice, ...) METALANG99_PRIV_REC_NEXT(13755, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13755(choice, ...) METALANG99_PRIV_REC_NEXT(13756, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13756(choice, ...) METALANG99_PRIV_REC_NEXT(13757, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13757(choice, ...) METALANG99_PRIV_REC_NEXT(13758, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13758(choice, ...) METALANG99_PRIV_REC_NEXT(13759, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13759(choice, ...) METALANG99_PRIV_REC_NEXT(13760, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13760(choice, ...) METALANG99_PRIV_REC_NEXT(13761, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13761(choice, ...) METALANG99_PRIV_REC_NEXT(13762, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13762(choice, ...) METALANG99_PRIV_REC_NEXT(13763, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13763(choice, ...) METALANG99_PRIV_REC_NEXT(13764, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13764(choice, ...) METALANG99_PRIV_REC_NEXT(13765, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13765(choice, ...) METALANG99_PRIV_REC_NEXT(13766, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13766(choice, ...) METALANG99_PRIV_REC_NEXT(13767, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13767(choice, ...) METALANG99_PRIV_REC_NEXT(13768, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13768(choice, ...) METALANG99_PRIV_REC_NEXT(13769, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13769(choice, ...) METALANG99_PRIV_REC_NEXT(13770, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13770(choice, ...) METALANG99_PRIV_REC_NEXT(13771, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13771(choice, ...) METALANG99_PRIV_REC_NEXT(13772, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13772(choice, ...) METALANG99_PRIV_REC_NEXT(13773, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13773(choice, ...) METALANG99_PRIV_REC_NEXT(13774, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13774(choice, ...) METALANG99_PRIV_REC_NEXT(13775, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13775(choice, ...) METALANG99_PRIV_REC_NEXT(13776, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13776(choice, ...) METALANG99_PRIV_REC_NEXT(13777, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13777(choice, ...) METALANG99_PRIV_REC_NEXT(13778, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13778(choice, ...) METALANG99_PRIV_REC_NEXT(13779, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13779(choice, ...) METALANG99_PRIV_REC_NEXT(13780, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13780(choice, ...) METALANG99_PRIV_REC_NEXT(13781, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13781(choice, ...) METALANG99_PRIV_REC_NEXT(13782, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13782(choice, ...) METALANG99_PRIV_REC_NEXT(13783, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13783(choice, ...) METALANG99_PRIV_REC_NEXT(13784, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13784(choice, ...) METALANG99_PRIV_REC_NEXT(13785, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13785(choice, ...) METALANG99_PRIV_REC_NEXT(13786, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13786(choice, ...) METALANG99_PRIV_REC_NEXT(13787, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13787(choice, ...) METALANG99_PRIV_REC_NEXT(13788, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13788(choice, ...) METALANG99_PRIV_REC_NEXT(13789, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13789(choice, ...) METALANG99_PRIV_REC_NEXT(13790, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13790(choice, ...) METALANG99_PRIV_REC_NEXT(13791, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13791(choice, ...) METALANG99_PRIV_REC_NEXT(13792, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13792(choice, ...) METALANG99_PRIV_REC_NEXT(13793, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13793(choice, ...) METALANG99_PRIV_REC_NEXT(13794, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13794(choice, ...) METALANG99_PRIV_REC_NEXT(13795, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13795(choice, ...) METALANG99_PRIV_REC_NEXT(13796, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13796(choice, ...) METALANG99_PRIV_REC_NEXT(13797, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13797(choice, ...) METALANG99_PRIV_REC_NEXT(13798, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13798(choice, ...) METALANG99_PRIV_REC_NEXT(13799, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13799(choice, ...) METALANG99_PRIV_REC_NEXT(13800, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13800(choice, ...) METALANG99_PRIV_REC_NEXT(13801, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13801(choice, ...) METALANG99_PRIV_REC_NEXT(13802, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13802(choice, ...) METALANG99_PRIV_REC_NEXT(13803, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13803(choice, ...) METALANG99_PRIV_REC_NEXT(13804, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13804(choice, ...) METALANG99_PRIV_REC_NEXT(13805, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13805(choice, ...) METALANG99_PRIV_REC_NEXT(13806, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13806(choice, ...) METALANG99_PRIV_REC_NEXT(13807, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13807(choice, ...) METALANG99_PRIV_REC_NEXT(13808, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13808(choice, ...) METALANG99_PRIV_REC_NEXT(13809, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13809(choice, ...) METALANG99_PRIV_REC_NEXT(13810, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13810(choice, ...) METALANG99_PRIV_REC_NEXT(13811, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13811(choice, ...) METALANG99_PRIV_REC_NEXT(13812, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13812(choice, ...) METALANG99_PRIV_REC_NEXT(13813, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13813(choice, ...) METALANG99_PRIV_REC_NEXT(13814, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13814(choice, ...) METALANG99_PRIV_REC_NEXT(13815, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13815(choice, ...) METALANG99_PRIV_REC_NEXT(13816, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13816(choice, ...) METALANG99_PRIV_REC_NEXT(13817, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13817(choice, ...) METALANG99_PRIV_REC_NEXT(13818, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13818(choice, ...) METALANG99_PRIV_REC_NEXT(13819, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13819(choice, ...) METALANG99_PRIV_REC_NEXT(13820, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13820(choice, ...) METALANG99_PRIV_REC_NEXT(13821, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13821(choice, ...) METALANG99_PRIV_REC_NEXT(13822, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13822(choice, ...) METALANG99_PRIV_REC_NEXT(13823, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13823(choice, ...) METALANG99_PRIV_REC_NEXT(13824, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13824(choice, ...) METALANG99_PRIV_REC_NEXT(13825, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13825(choice, ...) METALANG99_PRIV_REC_NEXT(13826, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13826(choice, ...) METALANG99_PRIV_REC_NEXT(13827, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13827(choice, ...) METALANG99_PRIV_REC_NEXT(13828, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13828(choice, ...) METALANG99_PRIV_REC_NEXT(13829, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13829(choice, ...) METALANG99_PRIV_REC_NEXT(13830, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13830(choice, ...) METALANG99_PRIV_REC_NEXT(13831, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13831(choice, ...) METALANG99_PRIV_REC_NEXT(13832, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13832(choice, ...) METALANG99_PRIV_REC_NEXT(13833, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13833(choice, ...) METALANG99_PRIV_REC_NEXT(13834, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13834(choice, ...) METALANG99_PRIV_REC_NEXT(13835, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13835(choice, ...) METALANG99_PRIV_REC_NEXT(13836, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13836(choice, ...) METALANG99_PRIV_REC_NEXT(13837, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13837(choice, ...) METALANG99_PRIV_REC_NEXT(13838, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13838(choice, ...) METALANG99_PRIV_REC_NEXT(13839, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13839(choice, ...) METALANG99_PRIV_REC_NEXT(13840, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13840(choice, ...) METALANG99_PRIV_REC_NEXT(13841, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13841(choice, ...) METALANG99_PRIV_REC_NEXT(13842, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13842(choice, ...) METALANG99_PRIV_REC_NEXT(13843, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13843(choice, ...) METALANG99_PRIV_REC_NEXT(13844, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13844(choice, ...) METALANG99_PRIV_REC_NEXT(13845, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13845(choice, ...) METALANG99_PRIV_REC_NEXT(13846, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13846(choice, ...) METALANG99_PRIV_REC_NEXT(13847, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13847(choice, ...) METALANG99_PRIV_REC_NEXT(13848, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13848(choice, ...) METALANG99_PRIV_REC_NEXT(13849, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13849(choice, ...) METALANG99_PRIV_REC_NEXT(13850, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13850(choice, ...) METALANG99_PRIV_REC_NEXT(13851, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13851(choice, ...) METALANG99_PRIV_REC_NEXT(13852, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13852(choice, ...) METALANG99_PRIV_REC_NEXT(13853, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13853(choice, ...) METALANG99_PRIV_REC_NEXT(13854, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13854(choice, ...) METALANG99_PRIV_REC_NEXT(13855, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13855(choice, ...) METALANG99_PRIV_REC_NEXT(13856, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13856(choice, ...) METALANG99_PRIV_REC_NEXT(13857, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13857(choice, ...) METALANG99_PRIV_REC_NEXT(13858, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13858(choice, ...) METALANG99_PRIV_REC_NEXT(13859, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13859(choice, ...) METALANG99_PRIV_REC_NEXT(13860, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13860(choice, ...) METALANG99_PRIV_REC_NEXT(13861, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13861(choice, ...) METALANG99_PRIV_REC_NEXT(13862, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13862(choice, ...) METALANG99_PRIV_REC_NEXT(13863, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13863(choice, ...) METALANG99_PRIV_REC_NEXT(13864, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13864(choice, ...) METALANG99_PRIV_REC_NEXT(13865, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13865(choice, ...) METALANG99_PRIV_REC_NEXT(13866, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13866(choice, ...) METALANG99_PRIV_REC_NEXT(13867, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13867(choice, ...) METALANG99_PRIV_REC_NEXT(13868, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13868(choice, ...) METALANG99_PRIV_REC_NEXT(13869, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13869(choice, ...) METALANG99_PRIV_REC_NEXT(13870, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13870(choice, ...) METALANG99_PRIV_REC_NEXT(13871, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13871(choice, ...) METALANG99_PRIV_REC_NEXT(13872, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13872(choice, ...) METALANG99_PRIV_REC_NEXT(13873, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13873(choice, ...) METALANG99_PRIV_REC_NEXT(13874, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13874(choice, ...) METALANG99_PRIV_REC_NEXT(13875, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13875(choice, ...) METALANG99_PRIV_REC_NEXT(13876, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13876(choice, ...) METALANG99_PRIV_REC_NEXT(13877, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13877(choice, ...) METALANG99_PRIV_REC_NEXT(13878, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13878(choice, ...) METALANG99_PRIV_REC_NEXT(13879, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13879(choice, ...) METALANG99_PRIV_REC_NEXT(13880, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13880(choice, ...) METALANG99_PRIV_REC_NEXT(13881, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13881(choice, ...) METALANG99_PRIV_REC_NEXT(13882, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13882(choice, ...) METALANG99_PRIV_REC_NEXT(13883, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13883(choice, ...) METALANG99_PRIV_REC_NEXT(13884, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13884(choice, ...) METALANG99_PRIV_REC_NEXT(13885, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13885(choice, ...) METALANG99_PRIV_REC_NEXT(13886, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13886(choice, ...) METALANG99_PRIV_REC_NEXT(13887, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13887(choice, ...) METALANG99_PRIV_REC_NEXT(13888, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13888(choice, ...) METALANG99_PRIV_REC_NEXT(13889, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13889(choice, ...) METALANG99_PRIV_REC_NEXT(13890, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13890(choice, ...) METALANG99_PRIV_REC_NEXT(13891, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13891(choice, ...) METALANG99_PRIV_REC_NEXT(13892, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13892(choice, ...) METALANG99_PRIV_REC_NEXT(13893, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13893(choice, ...) METALANG99_PRIV_REC_NEXT(13894, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13894(choice, ...) METALANG99_PRIV_REC_NEXT(13895, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13895(choice, ...) METALANG99_PRIV_REC_NEXT(13896, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13896(choice, ...) METALANG99_PRIV_REC_NEXT(13897, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13897(choice, ...) METALANG99_PRIV_REC_NEXT(13898, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13898(choice, ...) METALANG99_PRIV_REC_NEXT(13899, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13899(choice, ...) METALANG99_PRIV_REC_NEXT(13900, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13900(choice, ...) METALANG99_PRIV_REC_NEXT(13901, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13901(choice, ...) METALANG99_PRIV_REC_NEXT(13902, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13902(choice, ...) METALANG99_PRIV_REC_NEXT(13903, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13903(choice, ...) METALANG99_PRIV_REC_NEXT(13904, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13904(choice, ...) METALANG99_PRIV_REC_NEXT(13905, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13905(choice, ...) METALANG99_PRIV_REC_NEXT(13906, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13906(choice, ...) METALANG99_PRIV_REC_NEXT(13907, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13907(choice, ...) METALANG99_PRIV_REC_NEXT(13908, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13908(choice, ...) METALANG99_PRIV_REC_NEXT(13909, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13909(choice, ...) METALANG99_PRIV_REC_NEXT(13910, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13910(choice, ...) METALANG99_PRIV_REC_NEXT(13911, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13911(choice, ...) METALANG99_PRIV_REC_NEXT(13912, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13912(choice, ...) METALANG99_PRIV_REC_NEXT(13913, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13913(choice, ...) METALANG99_PRIV_REC_NEXT(13914, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13914(choice, ...) METALANG99_PRIV_REC_NEXT(13915, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13915(choice, ...) METALANG99_PRIV_REC_NEXT(13916, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13916(choice, ...) METALANG99_PRIV_REC_NEXT(13917, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13917(choice, ...) METALANG99_PRIV_REC_NEXT(13918, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13918(choice, ...) METALANG99_PRIV_REC_NEXT(13919, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13919(choice, ...) METALANG99_PRIV_REC_NEXT(13920, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13920(choice, ...) METALANG99_PRIV_REC_NEXT(13921, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13921(choice, ...) METALANG99_PRIV_REC_NEXT(13922, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13922(choice, ...) METALANG99_PRIV_REC_NEXT(13923, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13923(choice, ...) METALANG99_PRIV_REC_NEXT(13924, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13924(choice, ...) METALANG99_PRIV_REC_NEXT(13925, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13925(choice, ...) METALANG99_PRIV_REC_NEXT(13926, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13926(choice, ...) METALANG99_PRIV_REC_NEXT(13927, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13927(choice, ...) METALANG99_PRIV_REC_NEXT(13928, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13928(choice, ...) METALANG99_PRIV_REC_NEXT(13929, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13929(choice, ...) METALANG99_PRIV_REC_NEXT(13930, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13930(choice, ...) METALANG99_PRIV_REC_NEXT(13931, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13931(choice, ...) METALANG99_PRIV_REC_NEXT(13932, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13932(choice, ...) METALANG99_PRIV_REC_NEXT(13933, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13933(choice, ...) METALANG99_PRIV_REC_NEXT(13934, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13934(choice, ...) METALANG99_PRIV_REC_NEXT(13935, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13935(choice, ...) METALANG99_PRIV_REC_NEXT(13936, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13936(choice, ...) METALANG99_PRIV_REC_NEXT(13937, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13937(choice, ...) METALANG99_PRIV_REC_NEXT(13938, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13938(choice, ...) METALANG99_PRIV_REC_NEXT(13939, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13939(choice, ...) METALANG99_PRIV_REC_NEXT(13940, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13940(choice, ...) METALANG99_PRIV_REC_NEXT(13941, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13941(choice, ...) METALANG99_PRIV_REC_NEXT(13942, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13942(choice, ...) METALANG99_PRIV_REC_NEXT(13943, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13943(choice, ...) METALANG99_PRIV_REC_NEXT(13944, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13944(choice, ...) METALANG99_PRIV_REC_NEXT(13945, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13945(choice, ...) METALANG99_PRIV_REC_NEXT(13946, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13946(choice, ...) METALANG99_PRIV_REC_NEXT(13947, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13947(choice, ...) METALANG99_PRIV_REC_NEXT(13948, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13948(choice, ...) METALANG99_PRIV_REC_NEXT(13949, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13949(choice, ...) METALANG99_PRIV_REC_NEXT(13950, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13950(choice, ...) METALANG99_PRIV_REC_NEXT(13951, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13951(choice, ...) METALANG99_PRIV_REC_NEXT(13952, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13952(choice, ...) METALANG99_PRIV_REC_NEXT(13953, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13953(choice, ...) METALANG99_PRIV_REC_NEXT(13954, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13954(choice, ...) METALANG99_PRIV_REC_NEXT(13955, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13955(choice, ...) METALANG99_PRIV_REC_NEXT(13956, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13956(choice, ...) METALANG99_PRIV_REC_NEXT(13957, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13957(choice, ...) METALANG99_PRIV_REC_NEXT(13958, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13958(choice, ...) METALANG99_PRIV_REC_NEXT(13959, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13959(choice, ...) METALANG99_PRIV_REC_NEXT(13960, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13960(choice, ...) METALANG99_PRIV_REC_NEXT(13961, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13961(choice, ...) METALANG99_PRIV_REC_NEXT(13962, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13962(choice, ...) METALANG99_PRIV_REC_NEXT(13963, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13963(choice, ...) METALANG99_PRIV_REC_NEXT(13964, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13964(choice, ...) METALANG99_PRIV_REC_NEXT(13965, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13965(choice, ...) METALANG99_PRIV_REC_NEXT(13966, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13966(choice, ...) METALANG99_PRIV_REC_NEXT(13967, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13967(choice, ...) METALANG99_PRIV_REC_NEXT(13968, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13968(choice, ...) METALANG99_PRIV_REC_NEXT(13969, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13969(choice, ...) METALANG99_PRIV_REC_NEXT(13970, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13970(choice, ...) METALANG99_PRIV_REC_NEXT(13971, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13971(choice, ...) METALANG99_PRIV_REC_NEXT(13972, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13972(choice, ...) METALANG99_PRIV_REC_NEXT(13973, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13973(choice, ...) METALANG99_PRIV_REC_NEXT(13974, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13974(choice, ...) METALANG99_PRIV_REC_NEXT(13975, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13975(choice, ...) METALANG99_PRIV_REC_NEXT(13976, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13976(choice, ...) METALANG99_PRIV_REC_NEXT(13977, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13977(choice, ...) METALANG99_PRIV_REC_NEXT(13978, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13978(choice, ...) METALANG99_PRIV_REC_NEXT(13979, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13979(choice, ...) METALANG99_PRIV_REC_NEXT(13980, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13980(choice, ...) METALANG99_PRIV_REC_NEXT(13981, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13981(choice, ...) METALANG99_PRIV_REC_NEXT(13982, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13982(choice, ...) METALANG99_PRIV_REC_NEXT(13983, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13983(choice, ...) METALANG99_PRIV_REC_NEXT(13984, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13984(choice, ...) METALANG99_PRIV_REC_NEXT(13985, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13985(choice, ...) METALANG99_PRIV_REC_NEXT(13986, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13986(choice, ...) METALANG99_PRIV_REC_NEXT(13987, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13987(choice, ...) METALANG99_PRIV_REC_NEXT(13988, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13988(choice, ...) METALANG99_PRIV_REC_NEXT(13989, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13989(choice, ...) METALANG99_PRIV_REC_NEXT(13990, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13990(choice, ...) METALANG99_PRIV_REC_NEXT(13991, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13991(choice, ...) METALANG99_PRIV_REC_NEXT(13992, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13992(choice, ...) METALANG99_PRIV_REC_NEXT(13993, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13993(choice, ...) METALANG99_PRIV_REC_NEXT(13994, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13994(choice, ...) METALANG99_PRIV_REC_NEXT(13995, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13995(choice, ...) METALANG99_PRIV_REC_NEXT(13996, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13996(choice, ...) METALANG99_PRIV_REC_NEXT(13997, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13997(choice, ...) METALANG99_PRIV_REC_NEXT(13998, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13998(choice, ...) METALANG99_PRIV_REC_NEXT(13999, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_13999(choice, ...) METALANG99_PRIV_REC_NEXT(14000, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14000(choice, ...) METALANG99_PRIV_REC_NEXT(14001, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14001(choice, ...) METALANG99_PRIV_REC_NEXT(14002, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14002(choice, ...) METALANG99_PRIV_REC_NEXT(14003, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14003(choice, ...) METALANG99_PRIV_REC_NEXT(14004, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14004(choice, ...) METALANG99_PRIV_REC_NEXT(14005, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14005(choice, ...) METALANG99_PRIV_REC_NEXT(14006, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14006(choice, ...) METALANG99_PRIV_REC_NEXT(14007, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14007(choice, ...) METALANG99_PRIV_REC_NEXT(14008, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14008(choice, ...) METALANG99_PRIV_REC_NEXT(14009, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14009(choice, ...) METALANG99_PRIV_REC_NEXT(14010, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14010(choice, ...) METALANG99_PRIV_REC_NEXT(14011, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14011(choice, ...) METALANG99_PRIV_REC_NEXT(14012, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14012(choice, ...) METALANG99_PRIV_REC_NEXT(14013, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14013(choice, ...) METALANG99_PRIV_REC_NEXT(14014, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14014(choice, ...) METALANG99_PRIV_REC_NEXT(14015, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14015(choice, ...) METALANG99_PRIV_REC_NEXT(14016, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14016(choice, ...) METALANG99_PRIV_REC_NEXT(14017, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14017(choice, ...) METALANG99_PRIV_REC_NEXT(14018, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14018(choice, ...) METALANG99_PRIV_REC_NEXT(14019, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14019(choice, ...) METALANG99_PRIV_REC_NEXT(14020, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14020(choice, ...) METALANG99_PRIV_REC_NEXT(14021, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14021(choice, ...) METALANG99_PRIV_REC_NEXT(14022, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14022(choice, ...) METALANG99_PRIV_REC_NEXT(14023, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14023(choice, ...) METALANG99_PRIV_REC_NEXT(14024, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14024(choice, ...) METALANG99_PRIV_REC_NEXT(14025, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14025(choice, ...) METALANG99_PRIV_REC_NEXT(14026, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14026(choice, ...) METALANG99_PRIV_REC_NEXT(14027, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14027(choice, ...) METALANG99_PRIV_REC_NEXT(14028, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14028(choice, ...) METALANG99_PRIV_REC_NEXT(14029, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14029(choice, ...) METALANG99_PRIV_REC_NEXT(14030, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14030(choice, ...) METALANG99_PRIV_REC_NEXT(14031, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14031(choice, ...) METALANG99_PRIV_REC_NEXT(14032, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14032(choice, ...) METALANG99_PRIV_REC_NEXT(14033, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14033(choice, ...) METALANG99_PRIV_REC_NEXT(14034, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14034(choice, ...) METALANG99_PRIV_REC_NEXT(14035, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14035(choice, ...) METALANG99_PRIV_REC_NEXT(14036, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14036(choice, ...) METALANG99_PRIV_REC_NEXT(14037, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14037(choice, ...) METALANG99_PRIV_REC_NEXT(14038, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14038(choice, ...) METALANG99_PRIV_REC_NEXT(14039, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14039(choice, ...) METALANG99_PRIV_REC_NEXT(14040, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14040(choice, ...) METALANG99_PRIV_REC_NEXT(14041, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14041(choice, ...) METALANG99_PRIV_REC_NEXT(14042, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14042(choice, ...) METALANG99_PRIV_REC_NEXT(14043, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14043(choice, ...) METALANG99_PRIV_REC_NEXT(14044, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14044(choice, ...) METALANG99_PRIV_REC_NEXT(14045, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14045(choice, ...) METALANG99_PRIV_REC_NEXT(14046, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14046(choice, ...) METALANG99_PRIV_REC_NEXT(14047, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14047(choice, ...) METALANG99_PRIV_REC_NEXT(14048, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14048(choice, ...) METALANG99_PRIV_REC_NEXT(14049, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14049(choice, ...) METALANG99_PRIV_REC_NEXT(14050, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14050(choice, ...) METALANG99_PRIV_REC_NEXT(14051, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14051(choice, ...) METALANG99_PRIV_REC_NEXT(14052, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14052(choice, ...) METALANG99_PRIV_REC_NEXT(14053, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14053(choice, ...) METALANG99_PRIV_REC_NEXT(14054, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14054(choice, ...) METALANG99_PRIV_REC_NEXT(14055, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14055(choice, ...) METALANG99_PRIV_REC_NEXT(14056, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14056(choice, ...) METALANG99_PRIV_REC_NEXT(14057, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14057(choice, ...) METALANG99_PRIV_REC_NEXT(14058, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14058(choice, ...) METALANG99_PRIV_REC_NEXT(14059, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14059(choice, ...) METALANG99_PRIV_REC_NEXT(14060, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14060(choice, ...) METALANG99_PRIV_REC_NEXT(14061, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14061(choice, ...) METALANG99_PRIV_REC_NEXT(14062, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14062(choice, ...) METALANG99_PRIV_REC_NEXT(14063, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14063(choice, ...) METALANG99_PRIV_REC_NEXT(14064, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14064(choice, ...) METALANG99_PRIV_REC_NEXT(14065, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14065(choice, ...) METALANG99_PRIV_REC_NEXT(14066, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14066(choice, ...) METALANG99_PRIV_REC_NEXT(14067, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14067(choice, ...) METALANG99_PRIV_REC_NEXT(14068, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14068(choice, ...) METALANG99_PRIV_REC_NEXT(14069, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14069(choice, ...) METALANG99_PRIV_REC_NEXT(14070, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14070(choice, ...) METALANG99_PRIV_REC_NEXT(14071, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14071(choice, ...) METALANG99_PRIV_REC_NEXT(14072, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14072(choice, ...) METALANG99_PRIV_REC_NEXT(14073, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14073(choice, ...) METALANG99_PRIV_REC_NEXT(14074, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14074(choice, ...) METALANG99_PRIV_REC_NEXT(14075, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14075(choice, ...) METALANG99_PRIV_REC_NEXT(14076, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14076(choice, ...) METALANG99_PRIV_REC_NEXT(14077, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14077(choice, ...) METALANG99_PRIV_REC_NEXT(14078, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14078(choice, ...) METALANG99_PRIV_REC_NEXT(14079, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14079(choice, ...) METALANG99_PRIV_REC_NEXT(14080, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14080(choice, ...) METALANG99_PRIV_REC_NEXT(14081, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14081(choice, ...) METALANG99_PRIV_REC_NEXT(14082, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14082(choice, ...) METALANG99_PRIV_REC_NEXT(14083, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14083(choice, ...) METALANG99_PRIV_REC_NEXT(14084, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14084(choice, ...) METALANG99_PRIV_REC_NEXT(14085, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14085(choice, ...) METALANG99_PRIV_REC_NEXT(14086, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14086(choice, ...) METALANG99_PRIV_REC_NEXT(14087, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14087(choice, ...) METALANG99_PRIV_REC_NEXT(14088, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14088(choice, ...) METALANG99_PRIV_REC_NEXT(14089, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14089(choice, ...) METALANG99_PRIV_REC_NEXT(14090, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14090(choice, ...) METALANG99_PRIV_REC_NEXT(14091, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14091(choice, ...) METALANG99_PRIV_REC_NEXT(14092, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14092(choice, ...) METALANG99_PRIV_REC_NEXT(14093, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14093(choice, ...) METALANG99_PRIV_REC_NEXT(14094, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14094(choice, ...) METALANG99_PRIV_REC_NEXT(14095, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14095(choice, ...) METALANG99_PRIV_REC_NEXT(14096, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14096(choice, ...) METALANG99_PRIV_REC_NEXT(14097, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14097(choice, ...) METALANG99_PRIV_REC_NEXT(14098, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14098(choice, ...) METALANG99_PRIV_REC_NEXT(14099, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14099(choice, ...) METALANG99_PRIV_REC_NEXT(14100, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14100(choice, ...) METALANG99_PRIV_REC_NEXT(14101, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14101(choice, ...) METALANG99_PRIV_REC_NEXT(14102, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14102(choice, ...) METALANG99_PRIV_REC_NEXT(14103, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14103(choice, ...) METALANG99_PRIV_REC_NEXT(14104, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14104(choice, ...) METALANG99_PRIV_REC_NEXT(14105, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14105(choice, ...) METALANG99_PRIV_REC_NEXT(14106, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14106(choice, ...) METALANG99_PRIV_REC_NEXT(14107, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14107(choice, ...) METALANG99_PRIV_REC_NEXT(14108, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14108(choice, ...) METALANG99_PRIV_REC_NEXT(14109, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14109(choice, ...) METALANG99_PRIV_REC_NEXT(14110, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14110(choice, ...) METALANG99_PRIV_REC_NEXT(14111, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14111(choice, ...) METALANG99_PRIV_REC_NEXT(14112, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14112(choice, ...) METALANG99_PRIV_REC_NEXT(14113, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14113(choice, ...) METALANG99_PRIV_REC_NEXT(14114, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14114(choice, ...) METALANG99_PRIV_REC_NEXT(14115, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14115(choice, ...) METALANG99_PRIV_REC_NEXT(14116, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14116(choice, ...) METALANG99_PRIV_REC_NEXT(14117, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14117(choice, ...) METALANG99_PRIV_REC_NEXT(14118, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14118(choice, ...) METALANG99_PRIV_REC_NEXT(14119, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14119(choice, ...) METALANG99_PRIV_REC_NEXT(14120, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14120(choice, ...) METALANG99_PRIV_REC_NEXT(14121, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14121(choice, ...) METALANG99_PRIV_REC_NEXT(14122, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14122(choice, ...) METALANG99_PRIV_REC_NEXT(14123, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14123(choice, ...) METALANG99_PRIV_REC_NEXT(14124, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14124(choice, ...) METALANG99_PRIV_REC_NEXT(14125, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14125(choice, ...) METALANG99_PRIV_REC_NEXT(14126, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14126(choice, ...) METALANG99_PRIV_REC_NEXT(14127, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14127(choice, ...) METALANG99_PRIV_REC_NEXT(14128, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14128(choice, ...) METALANG99_PRIV_REC_NEXT(14129, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14129(choice, ...) METALANG99_PRIV_REC_NEXT(14130, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14130(choice, ...) METALANG99_PRIV_REC_NEXT(14131, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14131(choice, ...) METALANG99_PRIV_REC_NEXT(14132, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14132(choice, ...) METALANG99_PRIV_REC_NEXT(14133, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14133(choice, ...) METALANG99_PRIV_REC_NEXT(14134, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14134(choice, ...) METALANG99_PRIV_REC_NEXT(14135, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14135(choice, ...) METALANG99_PRIV_REC_NEXT(14136, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14136(choice, ...) METALANG99_PRIV_REC_NEXT(14137, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14137(choice, ...) METALANG99_PRIV_REC_NEXT(14138, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14138(choice, ...) METALANG99_PRIV_REC_NEXT(14139, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14139(choice, ...) METALANG99_PRIV_REC_NEXT(14140, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14140(choice, ...) METALANG99_PRIV_REC_NEXT(14141, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14141(choice, ...) METALANG99_PRIV_REC_NEXT(14142, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14142(choice, ...) METALANG99_PRIV_REC_NEXT(14143, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14143(choice, ...) METALANG99_PRIV_REC_NEXT(14144, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14144(choice, ...) METALANG99_PRIV_REC_NEXT(14145, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14145(choice, ...) METALANG99_PRIV_REC_NEXT(14146, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14146(choice, ...) METALANG99_PRIV_REC_NEXT(14147, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14147(choice, ...) METALANG99_PRIV_REC_NEXT(14148, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14148(choice, ...) METALANG99_PRIV_REC_NEXT(14149, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14149(choice, ...) METALANG99_PRIV_REC_NEXT(14150, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14150(choice, ...) METALANG99_PRIV_REC_NEXT(14151, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14151(choice, ...) METALANG99_PRIV_REC_NEXT(14152, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14152(choice, ...) METALANG99_PRIV_REC_NEXT(14153, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14153(choice, ...) METALANG99_PRIV_REC_NEXT(14154, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14154(choice, ...) METALANG99_PRIV_REC_NEXT(14155, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14155(choice, ...) METALANG99_PRIV_REC_NEXT(14156, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14156(choice, ...) METALANG99_PRIV_REC_NEXT(14157, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14157(choice, ...) METALANG99_PRIV_REC_NEXT(14158, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14158(choice, ...) METALANG99_PRIV_REC_NEXT(14159, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14159(choice, ...) METALANG99_PRIV_REC_NEXT(14160, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14160(choice, ...) METALANG99_PRIV_REC_NEXT(14161, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14161(choice, ...) METALANG99_PRIV_REC_NEXT(14162, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14162(choice, ...) METALANG99_PRIV_REC_NEXT(14163, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14163(choice, ...) METALANG99_PRIV_REC_NEXT(14164, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14164(choice, ...) METALANG99_PRIV_REC_NEXT(14165, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14165(choice, ...) METALANG99_PRIV_REC_NEXT(14166, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14166(choice, ...) METALANG99_PRIV_REC_NEXT(14167, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14167(choice, ...) METALANG99_PRIV_REC_NEXT(14168, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14168(choice, ...) METALANG99_PRIV_REC_NEXT(14169, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14169(choice, ...) METALANG99_PRIV_REC_NEXT(14170, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14170(choice, ...) METALANG99_PRIV_REC_NEXT(14171, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14171(choice, ...) METALANG99_PRIV_REC_NEXT(14172, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14172(choice, ...) METALANG99_PRIV_REC_NEXT(14173, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14173(choice, ...) METALANG99_PRIV_REC_NEXT(14174, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14174(choice, ...) METALANG99_PRIV_REC_NEXT(14175, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14175(choice, ...) METALANG99_PRIV_REC_NEXT(14176, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14176(choice, ...) METALANG99_PRIV_REC_NEXT(14177, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14177(choice, ...) METALANG99_PRIV_REC_NEXT(14178, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14178(choice, ...) METALANG99_PRIV_REC_NEXT(14179, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14179(choice, ...) METALANG99_PRIV_REC_NEXT(14180, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14180(choice, ...) METALANG99_PRIV_REC_NEXT(14181, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14181(choice, ...) METALANG99_PRIV_REC_NEXT(14182, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14182(choice, ...) METALANG99_PRIV_REC_NEXT(14183, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14183(choice, ...) METALANG99_PRIV_REC_NEXT(14184, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14184(choice, ...) METALANG99_PRIV_REC_NEXT(14185, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14185(choice, ...) METALANG99_PRIV_REC_NEXT(14186, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14186(choice, ...) METALANG99_PRIV_REC_NEXT(14187, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14187(choice, ...) METALANG99_PRIV_REC_NEXT(14188, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14188(choice, ...) METALANG99_PRIV_REC_NEXT(14189, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14189(choice, ...) METALANG99_PRIV_REC_NEXT(14190, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14190(choice, ...) METALANG99_PRIV_REC_NEXT(14191, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14191(choice, ...) METALANG99_PRIV_REC_NEXT(14192, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14192(choice, ...) METALANG99_PRIV_REC_NEXT(14193, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14193(choice, ...) METALANG99_PRIV_REC_NEXT(14194, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14194(choice, ...) METALANG99_PRIV_REC_NEXT(14195, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14195(choice, ...) METALANG99_PRIV_REC_NEXT(14196, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14196(choice, ...) METALANG99_PRIV_REC_NEXT(14197, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14197(choice, ...) METALANG99_PRIV_REC_NEXT(14198, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14198(choice, ...) METALANG99_PRIV_REC_NEXT(14199, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14199(choice, ...) METALANG99_PRIV_REC_NEXT(14200, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14200(choice, ...) METALANG99_PRIV_REC_NEXT(14201, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14201(choice, ...) METALANG99_PRIV_REC_NEXT(14202, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14202(choice, ...) METALANG99_PRIV_REC_NEXT(14203, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14203(choice, ...) METALANG99_PRIV_REC_NEXT(14204, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14204(choice, ...) METALANG99_PRIV_REC_NEXT(14205, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14205(choice, ...) METALANG99_PRIV_REC_NEXT(14206, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14206(choice, ...) METALANG99_PRIV_REC_NEXT(14207, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14207(choice, ...) METALANG99_PRIV_REC_NEXT(14208, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14208(choice, ...) METALANG99_PRIV_REC_NEXT(14209, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14209(choice, ...) METALANG99_PRIV_REC_NEXT(14210, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14210(choice, ...) METALANG99_PRIV_REC_NEXT(14211, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14211(choice, ...) METALANG99_PRIV_REC_NEXT(14212, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14212(choice, ...) METALANG99_PRIV_REC_NEXT(14213, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14213(choice, ...) METALANG99_PRIV_REC_NEXT(14214, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14214(choice, ...) METALANG99_PRIV_REC_NEXT(14215, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14215(choice, ...) METALANG99_PRIV_REC_NEXT(14216, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14216(choice, ...) METALANG99_PRIV_REC_NEXT(14217, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14217(choice, ...) METALANG99_PRIV_REC_NEXT(14218, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14218(choice, ...) METALANG99_PRIV_REC_NEXT(14219, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14219(choice, ...) METALANG99_PRIV_REC_NEXT(14220, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14220(choice, ...) METALANG99_PRIV_REC_NEXT(14221, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14221(choice, ...) METALANG99_PRIV_REC_NEXT(14222, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14222(choice, ...) METALANG99_PRIV_REC_NEXT(14223, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14223(choice, ...) METALANG99_PRIV_REC_NEXT(14224, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14224(choice, ...) METALANG99_PRIV_REC_NEXT(14225, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14225(choice, ...) METALANG99_PRIV_REC_NEXT(14226, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14226(choice, ...) METALANG99_PRIV_REC_NEXT(14227, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14227(choice, ...) METALANG99_PRIV_REC_NEXT(14228, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14228(choice, ...) METALANG99_PRIV_REC_NEXT(14229, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14229(choice, ...) METALANG99_PRIV_REC_NEXT(14230, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14230(choice, ...) METALANG99_PRIV_REC_NEXT(14231, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14231(choice, ...) METALANG99_PRIV_REC_NEXT(14232, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14232(choice, ...) METALANG99_PRIV_REC_NEXT(14233, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14233(choice, ...) METALANG99_PRIV_REC_NEXT(14234, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14234(choice, ...) METALANG99_PRIV_REC_NEXT(14235, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14235(choice, ...) METALANG99_PRIV_REC_NEXT(14236, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14236(choice, ...) METALANG99_PRIV_REC_NEXT(14237, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14237(choice, ...) METALANG99_PRIV_REC_NEXT(14238, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14238(choice, ...) METALANG99_PRIV_REC_NEXT(14239, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14239(choice, ...) METALANG99_PRIV_REC_NEXT(14240, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14240(choice, ...) METALANG99_PRIV_REC_NEXT(14241, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14241(choice, ...) METALANG99_PRIV_REC_NEXT(14242, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14242(choice, ...) METALANG99_PRIV_REC_NEXT(14243, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14243(choice, ...) METALANG99_PRIV_REC_NEXT(14244, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14244(choice, ...) METALANG99_PRIV_REC_NEXT(14245, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14245(choice, ...) METALANG99_PRIV_REC_NEXT(14246, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14246(choice, ...) METALANG99_PRIV_REC_NEXT(14247, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14247(choice, ...) METALANG99_PRIV_REC_NEXT(14248, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14248(choice, ...) METALANG99_PRIV_REC_NEXT(14249, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14249(choice, ...) METALANG99_PRIV_REC_NEXT(14250, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14250(choice, ...) METALANG99_PRIV_REC_NEXT(14251, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14251(choice, ...) METALANG99_PRIV_REC_NEXT(14252, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14252(choice, ...) METALANG99_PRIV_REC_NEXT(14253, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14253(choice, ...) METALANG99_PRIV_REC_NEXT(14254, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14254(choice, ...) METALANG99_PRIV_REC_NEXT(14255, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14255(choice, ...) METALANG99_PRIV_REC_NEXT(14256, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14256(choice, ...) METALANG99_PRIV_REC_NEXT(14257, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14257(choice, ...) METALANG99_PRIV_REC_NEXT(14258, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14258(choice, ...) METALANG99_PRIV_REC_NEXT(14259, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14259(choice, ...) METALANG99_PRIV_REC_NEXT(14260, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14260(choice, ...) METALANG99_PRIV_REC_NEXT(14261, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14261(choice, ...) METALANG99_PRIV_REC_NEXT(14262, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14262(choice, ...) METALANG99_PRIV_REC_NEXT(14263, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14263(choice, ...) METALANG99_PRIV_REC_NEXT(14264, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14264(choice, ...) METALANG99_PRIV_REC_NEXT(14265, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14265(choice, ...) METALANG99_PRIV_REC_NEXT(14266, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14266(choice, ...) METALANG99_PRIV_REC_NEXT(14267, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14267(choice, ...) METALANG99_PRIV_REC_NEXT(14268, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14268(choice, ...) METALANG99_PRIV_REC_NEXT(14269, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14269(choice, ...) METALANG99_PRIV_REC_NEXT(14270, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14270(choice, ...) METALANG99_PRIV_REC_NEXT(14271, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14271(choice, ...) METALANG99_PRIV_REC_NEXT(14272, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14272(choice, ...) METALANG99_PRIV_REC_NEXT(14273, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14273(choice, ...) METALANG99_PRIV_REC_NEXT(14274, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14274(choice, ...) METALANG99_PRIV_REC_NEXT(14275, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14275(choice, ...) METALANG99_PRIV_REC_NEXT(14276, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14276(choice, ...) METALANG99_PRIV_REC_NEXT(14277, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14277(choice, ...) METALANG99_PRIV_REC_NEXT(14278, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14278(choice, ...) METALANG99_PRIV_REC_NEXT(14279, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14279(choice, ...) METALANG99_PRIV_REC_NEXT(14280, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14280(choice, ...) METALANG99_PRIV_REC_NEXT(14281, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14281(choice, ...) METALANG99_PRIV_REC_NEXT(14282, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14282(choice, ...) METALANG99_PRIV_REC_NEXT(14283, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14283(choice, ...) METALANG99_PRIV_REC_NEXT(14284, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14284(choice, ...) METALANG99_PRIV_REC_NEXT(14285, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14285(choice, ...) METALANG99_PRIV_REC_NEXT(14286, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14286(choice, ...) METALANG99_PRIV_REC_NEXT(14287, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14287(choice, ...) METALANG99_PRIV_REC_NEXT(14288, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14288(choice, ...) METALANG99_PRIV_REC_NEXT(14289, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14289(choice, ...) METALANG99_PRIV_REC_NEXT(14290, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14290(choice, ...) METALANG99_PRIV_REC_NEXT(14291, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14291(choice, ...) METALANG99_PRIV_REC_NEXT(14292, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14292(choice, ...) METALANG99_PRIV_REC_NEXT(14293, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14293(choice, ...) METALANG99_PRIV_REC_NEXT(14294, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14294(choice, ...) METALANG99_PRIV_REC_NEXT(14295, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14295(choice, ...) METALANG99_PRIV_REC_NEXT(14296, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14296(choice, ...) METALANG99_PRIV_REC_NEXT(14297, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14297(choice, ...) METALANG99_PRIV_REC_NEXT(14298, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14298(choice, ...) METALANG99_PRIV_REC_NEXT(14299, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14299(choice, ...) METALANG99_PRIV_REC_NEXT(14300, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14300(choice, ...) METALANG99_PRIV_REC_NEXT(14301, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14301(choice, ...) METALANG99_PRIV_REC_NEXT(14302, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14302(choice, ...) METALANG99_PRIV_REC_NEXT(14303, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14303(choice, ...) METALANG99_PRIV_REC_NEXT(14304, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14304(choice, ...) METALANG99_PRIV_REC_NEXT(14305, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14305(choice, ...) METALANG99_PRIV_REC_NEXT(14306, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14306(choice, ...) METALANG99_PRIV_REC_NEXT(14307, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14307(choice, ...) METALANG99_PRIV_REC_NEXT(14308, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14308(choice, ...) METALANG99_PRIV_REC_NEXT(14309, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14309(choice, ...) METALANG99_PRIV_REC_NEXT(14310, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14310(choice, ...) METALANG99_PRIV_REC_NEXT(14311, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14311(choice, ...) METALANG99_PRIV_REC_NEXT(14312, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14312(choice, ...) METALANG99_PRIV_REC_NEXT(14313, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14313(choice, ...) METALANG99_PRIV_REC_NEXT(14314, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14314(choice, ...) METALANG99_PRIV_REC_NEXT(14315, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14315(choice, ...) METALANG99_PRIV_REC_NEXT(14316, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14316(choice, ...) METALANG99_PRIV_REC_NEXT(14317, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14317(choice, ...) METALANG99_PRIV_REC_NEXT(14318, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14318(choice, ...) METALANG99_PRIV_REC_NEXT(14319, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14319(choice, ...) METALANG99_PRIV_REC_NEXT(14320, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14320(choice, ...) METALANG99_PRIV_REC_NEXT(14321, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14321(choice, ...) METALANG99_PRIV_REC_NEXT(14322, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14322(choice, ...) METALANG99_PRIV_REC_NEXT(14323, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14323(choice, ...) METALANG99_PRIV_REC_NEXT(14324, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14324(choice, ...) METALANG99_PRIV_REC_NEXT(14325, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14325(choice, ...) METALANG99_PRIV_REC_NEXT(14326, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14326(choice, ...) METALANG99_PRIV_REC_NEXT(14327, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14327(choice, ...) METALANG99_PRIV_REC_NEXT(14328, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14328(choice, ...) METALANG99_PRIV_REC_NEXT(14329, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14329(choice, ...) METALANG99_PRIV_REC_NEXT(14330, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14330(choice, ...) METALANG99_PRIV_REC_NEXT(14331, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14331(choice, ...) METALANG99_PRIV_REC_NEXT(14332, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14332(choice, ...) METALANG99_PRIV_REC_NEXT(14333, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14333(choice, ...) METALANG99_PRIV_REC_NEXT(14334, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14334(choice, ...) METALANG99_PRIV_REC_NEXT(14335, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14335(choice, ...) METALANG99_PRIV_REC_NEXT(14336, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14336(choice, ...) METALANG99_PRIV_REC_NEXT(14337, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14337(choice, ...) METALANG99_PRIV_REC_NEXT(14338, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14338(choice, ...) METALANG99_PRIV_REC_NEXT(14339, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14339(choice, ...) METALANG99_PRIV_REC_NEXT(14340, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14340(choice, ...) METALANG99_PRIV_REC_NEXT(14341, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14341(choice, ...) METALANG99_PRIV_REC_NEXT(14342, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14342(choice, ...) METALANG99_PRIV_REC_NEXT(14343, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14343(choice, ...) METALANG99_PRIV_REC_NEXT(14344, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14344(choice, ...) METALANG99_PRIV_REC_NEXT(14345, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14345(choice, ...) METALANG99_PRIV_REC_NEXT(14346, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14346(choice, ...) METALANG99_PRIV_REC_NEXT(14347, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14347(choice, ...) METALANG99_PRIV_REC_NEXT(14348, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14348(choice, ...) METALANG99_PRIV_REC_NEXT(14349, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14349(choice, ...) METALANG99_PRIV_REC_NEXT(14350, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14350(choice, ...) METALANG99_PRIV_REC_NEXT(14351, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14351(choice, ...) METALANG99_PRIV_REC_NEXT(14352, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14352(choice, ...) METALANG99_PRIV_REC_NEXT(14353, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14353(choice, ...) METALANG99_PRIV_REC_NEXT(14354, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14354(choice, ...) METALANG99_PRIV_REC_NEXT(14355, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14355(choice, ...) METALANG99_PRIV_REC_NEXT(14356, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14356(choice, ...) METALANG99_PRIV_REC_NEXT(14357, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14357(choice, ...) METALANG99_PRIV_REC_NEXT(14358, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14358(choice, ...) METALANG99_PRIV_REC_NEXT(14359, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14359(choice, ...) METALANG99_PRIV_REC_NEXT(14360, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14360(choice, ...) METALANG99_PRIV_REC_NEXT(14361, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14361(choice, ...) METALANG99_PRIV_REC_NEXT(14362, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14362(choice, ...) METALANG99_PRIV_REC_NEXT(14363, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14363(choice, ...) METALANG99_PRIV_REC_NEXT(14364, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14364(choice, ...) METALANG99_PRIV_REC_NEXT(14365, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14365(choice, ...) METALANG99_PRIV_REC_NEXT(14366, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14366(choice, ...) METALANG99_PRIV_REC_NEXT(14367, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14367(choice, ...) METALANG99_PRIV_REC_NEXT(14368, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14368(choice, ...) METALANG99_PRIV_REC_NEXT(14369, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14369(choice, ...) METALANG99_PRIV_REC_NEXT(14370, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14370(choice, ...) METALANG99_PRIV_REC_NEXT(14371, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14371(choice, ...) METALANG99_PRIV_REC_NEXT(14372, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14372(choice, ...) METALANG99_PRIV_REC_NEXT(14373, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14373(choice, ...) METALANG99_PRIV_REC_NEXT(14374, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14374(choice, ...) METALANG99_PRIV_REC_NEXT(14375, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14375(choice, ...) METALANG99_PRIV_REC_NEXT(14376, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14376(choice, ...) METALANG99_PRIV_REC_NEXT(14377, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14377(choice, ...) METALANG99_PRIV_REC_NEXT(14378, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14378(choice, ...) METALANG99_PRIV_REC_NEXT(14379, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14379(choice, ...) METALANG99_PRIV_REC_NEXT(14380, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14380(choice, ...) METALANG99_PRIV_REC_NEXT(14381, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14381(choice, ...) METALANG99_PRIV_REC_NEXT(14382, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14382(choice, ...) METALANG99_PRIV_REC_NEXT(14383, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14383(choice, ...) METALANG99_PRIV_REC_NEXT(14384, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14384(choice, ...) METALANG99_PRIV_REC_NEXT(14385, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14385(choice, ...) METALANG99_PRIV_REC_NEXT(14386, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14386(choice, ...) METALANG99_PRIV_REC_NEXT(14387, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14387(choice, ...) METALANG99_PRIV_REC_NEXT(14388, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14388(choice, ...) METALANG99_PRIV_REC_NEXT(14389, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14389(choice, ...) METALANG99_PRIV_REC_NEXT(14390, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14390(choice, ...) METALANG99_PRIV_REC_NEXT(14391, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14391(choice, ...) METALANG99_PRIV_REC_NEXT(14392, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14392(choice, ...) METALANG99_PRIV_REC_NEXT(14393, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14393(choice, ...) METALANG99_PRIV_REC_NEXT(14394, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14394(choice, ...) METALANG99_PRIV_REC_NEXT(14395, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14395(choice, ...) METALANG99_PRIV_REC_NEXT(14396, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14396(choice, ...) METALANG99_PRIV_REC_NEXT(14397, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14397(choice, ...) METALANG99_PRIV_REC_NEXT(14398, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14398(choice, ...) METALANG99_PRIV_REC_NEXT(14399, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14399(choice, ...) METALANG99_PRIV_REC_NEXT(14400, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14400(choice, ...) METALANG99_PRIV_REC_NEXT(14401, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14401(choice, ...) METALANG99_PRIV_REC_NEXT(14402, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14402(choice, ...) METALANG99_PRIV_REC_NEXT(14403, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14403(choice, ...) METALANG99_PRIV_REC_NEXT(14404, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14404(choice, ...) METALANG99_PRIV_REC_NEXT(14405, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14405(choice, ...) METALANG99_PRIV_REC_NEXT(14406, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14406(choice, ...) METALANG99_PRIV_REC_NEXT(14407, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14407(choice, ...) METALANG99_PRIV_REC_NEXT(14408, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14408(choice, ...) METALANG99_PRIV_REC_NEXT(14409, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14409(choice, ...) METALANG99_PRIV_REC_NEXT(14410, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14410(choice, ...) METALANG99_PRIV_REC_NEXT(14411, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14411(choice, ...) METALANG99_PRIV_REC_NEXT(14412, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14412(choice, ...) METALANG99_PRIV_REC_NEXT(14413, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14413(choice, ...) METALANG99_PRIV_REC_NEXT(14414, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14414(choice, ...) METALANG99_PRIV_REC_NEXT(14415, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14415(choice, ...) METALANG99_PRIV_REC_NEXT(14416, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14416(choice, ...) METALANG99_PRIV_REC_NEXT(14417, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14417(choice, ...) METALANG99_PRIV_REC_NEXT(14418, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14418(choice, ...) METALANG99_PRIV_REC_NEXT(14419, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14419(choice, ...) METALANG99_PRIV_REC_NEXT(14420, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14420(choice, ...) METALANG99_PRIV_REC_NEXT(14421, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14421(choice, ...) METALANG99_PRIV_REC_NEXT(14422, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14422(choice, ...) METALANG99_PRIV_REC_NEXT(14423, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14423(choice, ...) METALANG99_PRIV_REC_NEXT(14424, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14424(choice, ...) METALANG99_PRIV_REC_NEXT(14425, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14425(choice, ...) METALANG99_PRIV_REC_NEXT(14426, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14426(choice, ...) METALANG99_PRIV_REC_NEXT(14427, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14427(choice, ...) METALANG99_PRIV_REC_NEXT(14428, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14428(choice, ...) METALANG99_PRIV_REC_NEXT(14429, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14429(choice, ...) METALANG99_PRIV_REC_NEXT(14430, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14430(choice, ...) METALANG99_PRIV_REC_NEXT(14431, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14431(choice, ...) METALANG99_PRIV_REC_NEXT(14432, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14432(choice, ...) METALANG99_PRIV_REC_NEXT(14433, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14433(choice, ...) METALANG99_PRIV_REC_NEXT(14434, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14434(choice, ...) METALANG99_PRIV_REC_NEXT(14435, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14435(choice, ...) METALANG99_PRIV_REC_NEXT(14436, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14436(choice, ...) METALANG99_PRIV_REC_NEXT(14437, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14437(choice, ...) METALANG99_PRIV_REC_NEXT(14438, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14438(choice, ...) METALANG99_PRIV_REC_NEXT(14439, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14439(choice, ...) METALANG99_PRIV_REC_NEXT(14440, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14440(choice, ...) METALANG99_PRIV_REC_NEXT(14441, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14441(choice, ...) METALANG99_PRIV_REC_NEXT(14442, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14442(choice, ...) METALANG99_PRIV_REC_NEXT(14443, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14443(choice, ...) METALANG99_PRIV_REC_NEXT(14444, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14444(choice, ...) METALANG99_PRIV_REC_NEXT(14445, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14445(choice, ...) METALANG99_PRIV_REC_NEXT(14446, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14446(choice, ...) METALANG99_PRIV_REC_NEXT(14447, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14447(choice, ...) METALANG99_PRIV_REC_NEXT(14448, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14448(choice, ...) METALANG99_PRIV_REC_NEXT(14449, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14449(choice, ...) METALANG99_PRIV_REC_NEXT(14450, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14450(choice, ...) METALANG99_PRIV_REC_NEXT(14451, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14451(choice, ...) METALANG99_PRIV_REC_NEXT(14452, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14452(choice, ...) METALANG99_PRIV_REC_NEXT(14453, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14453(choice, ...) METALANG99_PRIV_REC_NEXT(14454, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14454(choice, ...) METALANG99_PRIV_REC_NEXT(14455, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14455(choice, ...) METALANG99_PRIV_REC_NEXT(14456, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14456(choice, ...) METALANG99_PRIV_REC_NEXT(14457, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14457(choice, ...) METALANG99_PRIV_REC_NEXT(14458, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14458(choice, ...) METALANG99_PRIV_REC_NEXT(14459, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14459(choice, ...) METALANG99_PRIV_REC_NEXT(14460, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14460(choice, ...) METALANG99_PRIV_REC_NEXT(14461, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14461(choice, ...) METALANG99_PRIV_REC_NEXT(14462, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14462(choice, ...) METALANG99_PRIV_REC_NEXT(14463, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14463(choice, ...) METALANG99_PRIV_REC_NEXT(14464, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14464(choice, ...) METALANG99_PRIV_REC_NEXT(14465, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14465(choice, ...) METALANG99_PRIV_REC_NEXT(14466, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14466(choice, ...) METALANG99_PRIV_REC_NEXT(14467, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14467(choice, ...) METALANG99_PRIV_REC_NEXT(14468, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14468(choice, ...) METALANG99_PRIV_REC_NEXT(14469, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14469(choice, ...) METALANG99_PRIV_REC_NEXT(14470, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14470(choice, ...) METALANG99_PRIV_REC_NEXT(14471, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14471(choice, ...) METALANG99_PRIV_REC_NEXT(14472, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14472(choice, ...) METALANG99_PRIV_REC_NEXT(14473, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14473(choice, ...) METALANG99_PRIV_REC_NEXT(14474, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14474(choice, ...) METALANG99_PRIV_REC_NEXT(14475, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14475(choice, ...) METALANG99_PRIV_REC_NEXT(14476, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14476(choice, ...) METALANG99_PRIV_REC_NEXT(14477, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14477(choice, ...) METALANG99_PRIV_REC_NEXT(14478, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14478(choice, ...) METALANG99_PRIV_REC_NEXT(14479, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14479(choice, ...) METALANG99_PRIV_REC_NEXT(14480, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14480(choice, ...) METALANG99_PRIV_REC_NEXT(14481, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14481(choice, ...) METALANG99_PRIV_REC_NEXT(14482, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14482(choice, ...) METALANG99_PRIV_REC_NEXT(14483, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14483(choice, ...) METALANG99_PRIV_REC_NEXT(14484, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14484(choice, ...) METALANG99_PRIV_REC_NEXT(14485, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14485(choice, ...) METALANG99_PRIV_REC_NEXT(14486, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14486(choice, ...) METALANG99_PRIV_REC_NEXT(14487, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14487(choice, ...) METALANG99_PRIV_REC_NEXT(14488, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14488(choice, ...) METALANG99_PRIV_REC_NEXT(14489, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14489(choice, ...) METALANG99_PRIV_REC_NEXT(14490, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14490(choice, ...) METALANG99_PRIV_REC_NEXT(14491, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14491(choice, ...) METALANG99_PRIV_REC_NEXT(14492, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14492(choice, ...) METALANG99_PRIV_REC_NEXT(14493, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14493(choice, ...) METALANG99_PRIV_REC_NEXT(14494, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14494(choice, ...) METALANG99_PRIV_REC_NEXT(14495, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14495(choice, ...) METALANG99_PRIV_REC_NEXT(14496, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14496(choice, ...) METALANG99_PRIV_REC_NEXT(14497, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14497(choice, ...) METALANG99_PRIV_REC_NEXT(14498, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14498(choice, ...) METALANG99_PRIV_REC_NEXT(14499, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14499(choice, ...) METALANG99_PRIV_REC_NEXT(14500, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14500(choice, ...) METALANG99_PRIV_REC_NEXT(14501, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14501(choice, ...) METALANG99_PRIV_REC_NEXT(14502, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14502(choice, ...) METALANG99_PRIV_REC_NEXT(14503, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14503(choice, ...) METALANG99_PRIV_REC_NEXT(14504, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14504(choice, ...) METALANG99_PRIV_REC_NEXT(14505, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14505(choice, ...) METALANG99_PRIV_REC_NEXT(14506, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14506(choice, ...) METALANG99_PRIV_REC_NEXT(14507, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14507(choice, ...) METALANG99_PRIV_REC_NEXT(14508, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14508(choice, ...) METALANG99_PRIV_REC_NEXT(14509, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14509(choice, ...) METALANG99_PRIV_REC_NEXT(14510, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14510(choice, ...) METALANG99_PRIV_REC_NEXT(14511, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14511(choice, ...) METALANG99_PRIV_REC_NEXT(14512, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14512(choice, ...) METALANG99_PRIV_REC_NEXT(14513, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14513(choice, ...) METALANG99_PRIV_REC_NEXT(14514, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14514(choice, ...) METALANG99_PRIV_REC_NEXT(14515, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14515(choice, ...) METALANG99_PRIV_REC_NEXT(14516, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14516(choice, ...) METALANG99_PRIV_REC_NEXT(14517, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14517(choice, ...) METALANG99_PRIV_REC_NEXT(14518, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14518(choice, ...) METALANG99_PRIV_REC_NEXT(14519, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14519(choice, ...) METALANG99_PRIV_REC_NEXT(14520, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14520(choice, ...) METALANG99_PRIV_REC_NEXT(14521, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14521(choice, ...) METALANG99_PRIV_REC_NEXT(14522, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14522(choice, ...) METALANG99_PRIV_REC_NEXT(14523, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14523(choice, ...) METALANG99_PRIV_REC_NEXT(14524, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14524(choice, ...) METALANG99_PRIV_REC_NEXT(14525, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14525(choice, ...) METALANG99_PRIV_REC_NEXT(14526, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14526(choice, ...) METALANG99_PRIV_REC_NEXT(14527, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14527(choice, ...) METALANG99_PRIV_REC_NEXT(14528, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14528(choice, ...) METALANG99_PRIV_REC_NEXT(14529, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14529(choice, ...) METALANG99_PRIV_REC_NEXT(14530, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14530(choice, ...) METALANG99_PRIV_REC_NEXT(14531, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14531(choice, ...) METALANG99_PRIV_REC_NEXT(14532, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14532(choice, ...) METALANG99_PRIV_REC_NEXT(14533, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14533(choice, ...) METALANG99_PRIV_REC_NEXT(14534, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14534(choice, ...) METALANG99_PRIV_REC_NEXT(14535, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14535(choice, ...) METALANG99_PRIV_REC_NEXT(14536, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14536(choice, ...) METALANG99_PRIV_REC_NEXT(14537, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14537(choice, ...) METALANG99_PRIV_REC_NEXT(14538, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14538(choice, ...) METALANG99_PRIV_REC_NEXT(14539, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14539(choice, ...) METALANG99_PRIV_REC_NEXT(14540, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14540(choice, ...) METALANG99_PRIV_REC_NEXT(14541, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14541(choice, ...) METALANG99_PRIV_REC_NEXT(14542, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14542(choice, ...) METALANG99_PRIV_REC_NEXT(14543, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14543(choice, ...) METALANG99_PRIV_REC_NEXT(14544, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14544(choice, ...) METALANG99_PRIV_REC_NEXT(14545, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14545(choice, ...) METALANG99_PRIV_REC_NEXT(14546, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14546(choice, ...) METALANG99_PRIV_REC_NEXT(14547, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14547(choice, ...) METALANG99_PRIV_REC_NEXT(14548, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14548(choice, ...) METALANG99_PRIV_REC_NEXT(14549, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14549(choice, ...) METALANG99_PRIV_REC_NEXT(14550, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14550(choice, ...) METALANG99_PRIV_REC_NEXT(14551, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14551(choice, ...) METALANG99_PRIV_REC_NEXT(14552, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14552(choice, ...) METALANG99_PRIV_REC_NEXT(14553, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14553(choice, ...) METALANG99_PRIV_REC_NEXT(14554, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14554(choice, ...) METALANG99_PRIV_REC_NEXT(14555, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14555(choice, ...) METALANG99_PRIV_REC_NEXT(14556, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14556(choice, ...) METALANG99_PRIV_REC_NEXT(14557, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14557(choice, ...) METALANG99_PRIV_REC_NEXT(14558, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14558(choice, ...) METALANG99_PRIV_REC_NEXT(14559, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14559(choice, ...) METALANG99_PRIV_REC_NEXT(14560, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14560(choice, ...) METALANG99_PRIV_REC_NEXT(14561, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14561(choice, ...) METALANG99_PRIV_REC_NEXT(14562, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14562(choice, ...) METALANG99_PRIV_REC_NEXT(14563, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14563(choice, ...) METALANG99_PRIV_REC_NEXT(14564, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14564(choice, ...) METALANG99_PRIV_REC_NEXT(14565, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14565(choice, ...) METALANG99_PRIV_REC_NEXT(14566, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14566(choice, ...) METALANG99_PRIV_REC_NEXT(14567, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14567(choice, ...) METALANG99_PRIV_REC_NEXT(14568, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14568(choice, ...) METALANG99_PRIV_REC_NEXT(14569, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14569(choice, ...) METALANG99_PRIV_REC_NEXT(14570, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14570(choice, ...) METALANG99_PRIV_REC_NEXT(14571, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14571(choice, ...) METALANG99_PRIV_REC_NEXT(14572, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14572(choice, ...) METALANG99_PRIV_REC_NEXT(14573, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14573(choice, ...) METALANG99_PRIV_REC_NEXT(14574, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14574(choice, ...) METALANG99_PRIV_REC_NEXT(14575, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14575(choice, ...) METALANG99_PRIV_REC_NEXT(14576, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14576(choice, ...) METALANG99_PRIV_REC_NEXT(14577, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14577(choice, ...) METALANG99_PRIV_REC_NEXT(14578, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14578(choice, ...) METALANG99_PRIV_REC_NEXT(14579, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14579(choice, ...) METALANG99_PRIV_REC_NEXT(14580, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14580(choice, ...) METALANG99_PRIV_REC_NEXT(14581, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14581(choice, ...) METALANG99_PRIV_REC_NEXT(14582, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14582(choice, ...) METALANG99_PRIV_REC_NEXT(14583, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14583(choice, ...) METALANG99_PRIV_REC_NEXT(14584, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14584(choice, ...) METALANG99_PRIV_REC_NEXT(14585, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14585(choice, ...) METALANG99_PRIV_REC_NEXT(14586, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14586(choice, ...) METALANG99_PRIV_REC_NEXT(14587, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14587(choice, ...) METALANG99_PRIV_REC_NEXT(14588, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14588(choice, ...) METALANG99_PRIV_REC_NEXT(14589, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14589(choice, ...) METALANG99_PRIV_REC_NEXT(14590, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14590(choice, ...) METALANG99_PRIV_REC_NEXT(14591, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14591(choice, ...) METALANG99_PRIV_REC_NEXT(14592, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14592(choice, ...) METALANG99_PRIV_REC_NEXT(14593, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14593(choice, ...) METALANG99_PRIV_REC_NEXT(14594, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14594(choice, ...) METALANG99_PRIV_REC_NEXT(14595, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14595(choice, ...) METALANG99_PRIV_REC_NEXT(14596, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14596(choice, ...) METALANG99_PRIV_REC_NEXT(14597, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14597(choice, ...) METALANG99_PRIV_REC_NEXT(14598, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14598(choice, ...) METALANG99_PRIV_REC_NEXT(14599, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14599(choice, ...) METALANG99_PRIV_REC_NEXT(14600, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14600(choice, ...) METALANG99_PRIV_REC_NEXT(14601, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14601(choice, ...) METALANG99_PRIV_REC_NEXT(14602, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14602(choice, ...) METALANG99_PRIV_REC_NEXT(14603, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14603(choice, ...) METALANG99_PRIV_REC_NEXT(14604, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14604(choice, ...) METALANG99_PRIV_REC_NEXT(14605, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14605(choice, ...) METALANG99_PRIV_REC_NEXT(14606, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14606(choice, ...) METALANG99_PRIV_REC_NEXT(14607, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14607(choice, ...) METALANG99_PRIV_REC_NEXT(14608, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14608(choice, ...) METALANG99_PRIV_REC_NEXT(14609, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14609(choice, ...) METALANG99_PRIV_REC_NEXT(14610, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14610(choice, ...) METALANG99_PRIV_REC_NEXT(14611, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14611(choice, ...) METALANG99_PRIV_REC_NEXT(14612, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14612(choice, ...) METALANG99_PRIV_REC_NEXT(14613, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14613(choice, ...) METALANG99_PRIV_REC_NEXT(14614, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14614(choice, ...) METALANG99_PRIV_REC_NEXT(14615, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14615(choice, ...) METALANG99_PRIV_REC_NEXT(14616, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14616(choice, ...) METALANG99_PRIV_REC_NEXT(14617, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14617(choice, ...) METALANG99_PRIV_REC_NEXT(14618, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14618(choice, ...) METALANG99_PRIV_REC_NEXT(14619, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14619(choice, ...) METALANG99_PRIV_REC_NEXT(14620, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14620(choice, ...) METALANG99_PRIV_REC_NEXT(14621, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14621(choice, ...) METALANG99_PRIV_REC_NEXT(14622, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14622(choice, ...) METALANG99_PRIV_REC_NEXT(14623, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14623(choice, ...) METALANG99_PRIV_REC_NEXT(14624, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14624(choice, ...) METALANG99_PRIV_REC_NEXT(14625, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14625(choice, ...) METALANG99_PRIV_REC_NEXT(14626, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14626(choice, ...) METALANG99_PRIV_REC_NEXT(14627, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14627(choice, ...) METALANG99_PRIV_REC_NEXT(14628, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14628(choice, ...) METALANG99_PRIV_REC_NEXT(14629, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14629(choice, ...) METALANG99_PRIV_REC_NEXT(14630, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14630(choice, ...) METALANG99_PRIV_REC_NEXT(14631, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14631(choice, ...) METALANG99_PRIV_REC_NEXT(14632, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14632(choice, ...) METALANG99_PRIV_REC_NEXT(14633, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14633(choice, ...) METALANG99_PRIV_REC_NEXT(14634, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14634(choice, ...) METALANG99_PRIV_REC_NEXT(14635, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14635(choice, ...) METALANG99_PRIV_REC_NEXT(14636, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14636(choice, ...) METALANG99_PRIV_REC_NEXT(14637, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14637(choice, ...) METALANG99_PRIV_REC_NEXT(14638, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14638(choice, ...) METALANG99_PRIV_REC_NEXT(14639, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14639(choice, ...) METALANG99_PRIV_REC_NEXT(14640, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14640(choice, ...) METALANG99_PRIV_REC_NEXT(14641, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14641(choice, ...) METALANG99_PRIV_REC_NEXT(14642, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14642(choice, ...) METALANG99_PRIV_REC_NEXT(14643, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14643(choice, ...) METALANG99_PRIV_REC_NEXT(14644, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14644(choice, ...) METALANG99_PRIV_REC_NEXT(14645, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14645(choice, ...) METALANG99_PRIV_REC_NEXT(14646, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14646(choice, ...) METALANG99_PRIV_REC_NEXT(14647, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14647(choice, ...) METALANG99_PRIV_REC_NEXT(14648, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14648(choice, ...) METALANG99_PRIV_REC_NEXT(14649, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14649(choice, ...) METALANG99_PRIV_REC_NEXT(14650, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14650(choice, ...) METALANG99_PRIV_REC_NEXT(14651, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14651(choice, ...) METALANG99_PRIV_REC_NEXT(14652, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14652(choice, ...) METALANG99_PRIV_REC_NEXT(14653, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14653(choice, ...) METALANG99_PRIV_REC_NEXT(14654, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14654(choice, ...) METALANG99_PRIV_REC_NEXT(14655, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14655(choice, ...) METALANG99_PRIV_REC_NEXT(14656, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14656(choice, ...) METALANG99_PRIV_REC_NEXT(14657, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14657(choice, ...) METALANG99_PRIV_REC_NEXT(14658, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14658(choice, ...) METALANG99_PRIV_REC_NEXT(14659, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14659(choice, ...) METALANG99_PRIV_REC_NEXT(14660, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14660(choice, ...) METALANG99_PRIV_REC_NEXT(14661, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14661(choice, ...) METALANG99_PRIV_REC_NEXT(14662, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14662(choice, ...) METALANG99_PRIV_REC_NEXT(14663, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14663(choice, ...) METALANG99_PRIV_REC_NEXT(14664, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14664(choice, ...) METALANG99_PRIV_REC_NEXT(14665, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14665(choice, ...) METALANG99_PRIV_REC_NEXT(14666, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14666(choice, ...) METALANG99_PRIV_REC_NEXT(14667, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14667(choice, ...) METALANG99_PRIV_REC_NEXT(14668, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14668(choice, ...) METALANG99_PRIV_REC_NEXT(14669, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14669(choice, ...) METALANG99_PRIV_REC_NEXT(14670, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14670(choice, ...) METALANG99_PRIV_REC_NEXT(14671, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14671(choice, ...) METALANG99_PRIV_REC_NEXT(14672, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14672(choice, ...) METALANG99_PRIV_REC_NEXT(14673, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14673(choice, ...) METALANG99_PRIV_REC_NEXT(14674, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14674(choice, ...) METALANG99_PRIV_REC_NEXT(14675, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14675(choice, ...) METALANG99_PRIV_REC_NEXT(14676, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14676(choice, ...) METALANG99_PRIV_REC_NEXT(14677, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14677(choice, ...) METALANG99_PRIV_REC_NEXT(14678, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14678(choice, ...) METALANG99_PRIV_REC_NEXT(14679, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14679(choice, ...) METALANG99_PRIV_REC_NEXT(14680, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14680(choice, ...) METALANG99_PRIV_REC_NEXT(14681, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14681(choice, ...) METALANG99_PRIV_REC_NEXT(14682, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14682(choice, ...) METALANG99_PRIV_REC_NEXT(14683, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14683(choice, ...) METALANG99_PRIV_REC_NEXT(14684, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14684(choice, ...) METALANG99_PRIV_REC_NEXT(14685, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14685(choice, ...) METALANG99_PRIV_REC_NEXT(14686, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14686(choice, ...) METALANG99_PRIV_REC_NEXT(14687, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14687(choice, ...) METALANG99_PRIV_REC_NEXT(14688, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14688(choice, ...) METALANG99_PRIV_REC_NEXT(14689, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14689(choice, ...) METALANG99_PRIV_REC_NEXT(14690, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14690(choice, ...) METALANG99_PRIV_REC_NEXT(14691, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14691(choice, ...) METALANG99_PRIV_REC_NEXT(14692, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14692(choice, ...) METALANG99_PRIV_REC_NEXT(14693, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14693(choice, ...) METALANG99_PRIV_REC_NEXT(14694, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14694(choice, ...) METALANG99_PRIV_REC_NEXT(14695, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14695(choice, ...) METALANG99_PRIV_REC_NEXT(14696, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14696(choice, ...) METALANG99_PRIV_REC_NEXT(14697, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14697(choice, ...) METALANG99_PRIV_REC_NEXT(14698, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14698(choice, ...) METALANG99_PRIV_REC_NEXT(14699, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14699(choice, ...) METALANG99_PRIV_REC_NEXT(14700, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14700(choice, ...) METALANG99_PRIV_REC_NEXT(14701, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14701(choice, ...) METALANG99_PRIV_REC_NEXT(14702, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14702(choice, ...) METALANG99_PRIV_REC_NEXT(14703, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14703(choice, ...) METALANG99_PRIV_REC_NEXT(14704, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14704(choice, ...) METALANG99_PRIV_REC_NEXT(14705, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14705(choice, ...) METALANG99_PRIV_REC_NEXT(14706, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14706(choice, ...) METALANG99_PRIV_REC_NEXT(14707, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14707(choice, ...) METALANG99_PRIV_REC_NEXT(14708, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14708(choice, ...) METALANG99_PRIV_REC_NEXT(14709, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14709(choice, ...) METALANG99_PRIV_REC_NEXT(14710, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14710(choice, ...) METALANG99_PRIV_REC_NEXT(14711, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14711(choice, ...) METALANG99_PRIV_REC_NEXT(14712, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14712(choice, ...) METALANG99_PRIV_REC_NEXT(14713, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14713(choice, ...) METALANG99_PRIV_REC_NEXT(14714, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14714(choice, ...) METALANG99_PRIV_REC_NEXT(14715, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14715(choice, ...) METALANG99_PRIV_REC_NEXT(14716, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14716(choice, ...) METALANG99_PRIV_REC_NEXT(14717, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14717(choice, ...) METALANG99_PRIV_REC_NEXT(14718, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14718(choice, ...) METALANG99_PRIV_REC_NEXT(14719, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14719(choice, ...) METALANG99_PRIV_REC_NEXT(14720, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14720(choice, ...) METALANG99_PRIV_REC_NEXT(14721, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14721(choice, ...) METALANG99_PRIV_REC_NEXT(14722, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14722(choice, ...) METALANG99_PRIV_REC_NEXT(14723, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14723(choice, ...) METALANG99_PRIV_REC_NEXT(14724, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14724(choice, ...) METALANG99_PRIV_REC_NEXT(14725, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14725(choice, ...) METALANG99_PRIV_REC_NEXT(14726, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14726(choice, ...) METALANG99_PRIV_REC_NEXT(14727, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14727(choice, ...) METALANG99_PRIV_REC_NEXT(14728, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14728(choice, ...) METALANG99_PRIV_REC_NEXT(14729, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14729(choice, ...) METALANG99_PRIV_REC_NEXT(14730, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14730(choice, ...) METALANG99_PRIV_REC_NEXT(14731, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14731(choice, ...) METALANG99_PRIV_REC_NEXT(14732, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14732(choice, ...) METALANG99_PRIV_REC_NEXT(14733, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14733(choice, ...) METALANG99_PRIV_REC_NEXT(14734, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14734(choice, ...) METALANG99_PRIV_REC_NEXT(14735, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14735(choice, ...) METALANG99_PRIV_REC_NEXT(14736, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14736(choice, ...) METALANG99_PRIV_REC_NEXT(14737, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14737(choice, ...) METALANG99_PRIV_REC_NEXT(14738, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14738(choice, ...) METALANG99_PRIV_REC_NEXT(14739, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14739(choice, ...) METALANG99_PRIV_REC_NEXT(14740, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14740(choice, ...) METALANG99_PRIV_REC_NEXT(14741, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14741(choice, ...) METALANG99_PRIV_REC_NEXT(14742, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14742(choice, ...) METALANG99_PRIV_REC_NEXT(14743, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14743(choice, ...) METALANG99_PRIV_REC_NEXT(14744, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14744(choice, ...) METALANG99_PRIV_REC_NEXT(14745, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14745(choice, ...) METALANG99_PRIV_REC_NEXT(14746, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14746(choice, ...) METALANG99_PRIV_REC_NEXT(14747, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14747(choice, ...) METALANG99_PRIV_REC_NEXT(14748, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14748(choice, ...) METALANG99_PRIV_REC_NEXT(14749, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14749(choice, ...) METALANG99_PRIV_REC_NEXT(14750, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14750(choice, ...) METALANG99_PRIV_REC_NEXT(14751, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14751(choice, ...) METALANG99_PRIV_REC_NEXT(14752, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14752(choice, ...) METALANG99_PRIV_REC_NEXT(14753, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14753(choice, ...) METALANG99_PRIV_REC_NEXT(14754, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14754(choice, ...) METALANG99_PRIV_REC_NEXT(14755, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14755(choice, ...) METALANG99_PRIV_REC_NEXT(14756, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14756(choice, ...) METALANG99_PRIV_REC_NEXT(14757, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14757(choice, ...) METALANG99_PRIV_REC_NEXT(14758, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14758(choice, ...) METALANG99_PRIV_REC_NEXT(14759, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14759(choice, ...) METALANG99_PRIV_REC_NEXT(14760, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14760(choice, ...) METALANG99_PRIV_REC_NEXT(14761, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14761(choice, ...) METALANG99_PRIV_REC_NEXT(14762, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14762(choice, ...) METALANG99_PRIV_REC_NEXT(14763, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14763(choice, ...) METALANG99_PRIV_REC_NEXT(14764, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14764(choice, ...) METALANG99_PRIV_REC_NEXT(14765, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14765(choice, ...) METALANG99_PRIV_REC_NEXT(14766, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14766(choice, ...) METALANG99_PRIV_REC_NEXT(14767, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14767(choice, ...) METALANG99_PRIV_REC_NEXT(14768, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14768(choice, ...) METALANG99_PRIV_REC_NEXT(14769, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14769(choice, ...) METALANG99_PRIV_REC_NEXT(14770, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14770(choice, ...) METALANG99_PRIV_REC_NEXT(14771, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14771(choice, ...) METALANG99_PRIV_REC_NEXT(14772, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14772(choice, ...) METALANG99_PRIV_REC_NEXT(14773, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14773(choice, ...) METALANG99_PRIV_REC_NEXT(14774, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14774(choice, ...) METALANG99_PRIV_REC_NEXT(14775, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14775(choice, ...) METALANG99_PRIV_REC_NEXT(14776, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14776(choice, ...) METALANG99_PRIV_REC_NEXT(14777, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14777(choice, ...) METALANG99_PRIV_REC_NEXT(14778, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14778(choice, ...) METALANG99_PRIV_REC_NEXT(14779, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14779(choice, ...) METALANG99_PRIV_REC_NEXT(14780, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14780(choice, ...) METALANG99_PRIV_REC_NEXT(14781, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14781(choice, ...) METALANG99_PRIV_REC_NEXT(14782, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14782(choice, ...) METALANG99_PRIV_REC_NEXT(14783, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14783(choice, ...) METALANG99_PRIV_REC_NEXT(14784, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14784(choice, ...) METALANG99_PRIV_REC_NEXT(14785, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14785(choice, ...) METALANG99_PRIV_REC_NEXT(14786, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14786(choice, ...) METALANG99_PRIV_REC_NEXT(14787, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14787(choice, ...) METALANG99_PRIV_REC_NEXT(14788, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14788(choice, ...) METALANG99_PRIV_REC_NEXT(14789, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14789(choice, ...) METALANG99_PRIV_REC_NEXT(14790, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14790(choice, ...) METALANG99_PRIV_REC_NEXT(14791, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14791(choice, ...) METALANG99_PRIV_REC_NEXT(14792, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14792(choice, ...) METALANG99_PRIV_REC_NEXT(14793, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14793(choice, ...) METALANG99_PRIV_REC_NEXT(14794, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14794(choice, ...) METALANG99_PRIV_REC_NEXT(14795, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14795(choice, ...) METALANG99_PRIV_REC_NEXT(14796, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14796(choice, ...) METALANG99_PRIV_REC_NEXT(14797, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14797(choice, ...) METALANG99_PRIV_REC_NEXT(14798, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14798(choice, ...) METALANG99_PRIV_REC_NEXT(14799, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14799(choice, ...) METALANG99_PRIV_REC_NEXT(14800, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14800(choice, ...) METALANG99_PRIV_REC_NEXT(14801, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14801(choice, ...) METALANG99_PRIV_REC_NEXT(14802, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14802(choice, ...) METALANG99_PRIV_REC_NEXT(14803, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14803(choice, ...) METALANG99_PRIV_REC_NEXT(14804, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14804(choice, ...) METALANG99_PRIV_REC_NEXT(14805, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14805(choice, ...) METALANG99_PRIV_REC_NEXT(14806, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14806(choice, ...) METALANG99_PRIV_REC_NEXT(14807, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14807(choice, ...) METALANG99_PRIV_REC_NEXT(14808, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14808(choice, ...) METALANG99_PRIV_REC_NEXT(14809, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14809(choice, ...) METALANG99_PRIV_REC_NEXT(14810, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14810(choice, ...) METALANG99_PRIV_REC_NEXT(14811, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14811(choice, ...) METALANG99_PRIV_REC_NEXT(14812, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14812(choice, ...) METALANG99_PRIV_REC_NEXT(14813, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14813(choice, ...) METALANG99_PRIV_REC_NEXT(14814, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14814(choice, ...) METALANG99_PRIV_REC_NEXT(14815, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14815(choice, ...) METALANG99_PRIV_REC_NEXT(14816, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14816(choice, ...) METALANG99_PRIV_REC_NEXT(14817, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14817(choice, ...) METALANG99_PRIV_REC_NEXT(14818, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14818(choice, ...) METALANG99_PRIV_REC_NEXT(14819, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14819(choice, ...) METALANG99_PRIV_REC_NEXT(14820, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14820(choice, ...) METALANG99_PRIV_REC_NEXT(14821, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14821(choice, ...) METALANG99_PRIV_REC_NEXT(14822, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14822(choice, ...) METALANG99_PRIV_REC_NEXT(14823, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14823(choice, ...) METALANG99_PRIV_REC_NEXT(14824, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14824(choice, ...) METALANG99_PRIV_REC_NEXT(14825, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14825(choice, ...) METALANG99_PRIV_REC_NEXT(14826, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14826(choice, ...) METALANG99_PRIV_REC_NEXT(14827, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14827(choice, ...) METALANG99_PRIV_REC_NEXT(14828, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14828(choice, ...) METALANG99_PRIV_REC_NEXT(14829, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14829(choice, ...) METALANG99_PRIV_REC_NEXT(14830, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14830(choice, ...) METALANG99_PRIV_REC_NEXT(14831, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14831(choice, ...) METALANG99_PRIV_REC_NEXT(14832, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14832(choice, ...) METALANG99_PRIV_REC_NEXT(14833, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14833(choice, ...) METALANG99_PRIV_REC_NEXT(14834, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14834(choice, ...) METALANG99_PRIV_REC_NEXT(14835, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14835(choice, ...) METALANG99_PRIV_REC_NEXT(14836, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14836(choice, ...) METALANG99_PRIV_REC_NEXT(14837, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14837(choice, ...) METALANG99_PRIV_REC_NEXT(14838, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14838(choice, ...) METALANG99_PRIV_REC_NEXT(14839, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14839(choice, ...) METALANG99_PRIV_REC_NEXT(14840, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14840(choice, ...) METALANG99_PRIV_REC_NEXT(14841, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14841(choice, ...) METALANG99_PRIV_REC_NEXT(14842, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14842(choice, ...) METALANG99_PRIV_REC_NEXT(14843, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14843(choice, ...) METALANG99_PRIV_REC_NEXT(14844, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14844(choice, ...) METALANG99_PRIV_REC_NEXT(14845, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14845(choice, ...) METALANG99_PRIV_REC_NEXT(14846, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14846(choice, ...) METALANG99_PRIV_REC_NEXT(14847, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14847(choice, ...) METALANG99_PRIV_REC_NEXT(14848, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14848(choice, ...) METALANG99_PRIV_REC_NEXT(14849, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14849(choice, ...) METALANG99_PRIV_REC_NEXT(14850, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14850(choice, ...) METALANG99_PRIV_REC_NEXT(14851, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14851(choice, ...) METALANG99_PRIV_REC_NEXT(14852, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14852(choice, ...) METALANG99_PRIV_REC_NEXT(14853, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14853(choice, ...) METALANG99_PRIV_REC_NEXT(14854, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14854(choice, ...) METALANG99_PRIV_REC_NEXT(14855, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14855(choice, ...) METALANG99_PRIV_REC_NEXT(14856, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14856(choice, ...) METALANG99_PRIV_REC_NEXT(14857, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14857(choice, ...) METALANG99_PRIV_REC_NEXT(14858, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14858(choice, ...) METALANG99_PRIV_REC_NEXT(14859, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14859(choice, ...) METALANG99_PRIV_REC_NEXT(14860, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14860(choice, ...) METALANG99_PRIV_REC_NEXT(14861, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14861(choice, ...) METALANG99_PRIV_REC_NEXT(14862, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14862(choice, ...) METALANG99_PRIV_REC_NEXT(14863, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14863(choice, ...) METALANG99_PRIV_REC_NEXT(14864, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14864(choice, ...) METALANG99_PRIV_REC_NEXT(14865, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14865(choice, ...) METALANG99_PRIV_REC_NEXT(14866, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14866(choice, ...) METALANG99_PRIV_REC_NEXT(14867, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14867(choice, ...) METALANG99_PRIV_REC_NEXT(14868, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14868(choice, ...) METALANG99_PRIV_REC_NEXT(14869, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14869(choice, ...) METALANG99_PRIV_REC_NEXT(14870, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14870(choice, ...) METALANG99_PRIV_REC_NEXT(14871, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14871(choice, ...) METALANG99_PRIV_REC_NEXT(14872, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14872(choice, ...) METALANG99_PRIV_REC_NEXT(14873, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14873(choice, ...) METALANG99_PRIV_REC_NEXT(14874, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14874(choice, ...) METALANG99_PRIV_REC_NEXT(14875, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14875(choice, ...) METALANG99_PRIV_REC_NEXT(14876, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14876(choice, ...) METALANG99_PRIV_REC_NEXT(14877, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14877(choice, ...) METALANG99_PRIV_REC_NEXT(14878, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14878(choice, ...) METALANG99_PRIV_REC_NEXT(14879, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14879(choice, ...) METALANG99_PRIV_REC_NEXT(14880, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14880(choice, ...) METALANG99_PRIV_REC_NEXT(14881, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14881(choice, ...) METALANG99_PRIV_REC_NEXT(14882, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14882(choice, ...) METALANG99_PRIV_REC_NEXT(14883, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14883(choice, ...) METALANG99_PRIV_REC_NEXT(14884, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14884(choice, ...) METALANG99_PRIV_REC_NEXT(14885, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14885(choice, ...) METALANG99_PRIV_REC_NEXT(14886, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14886(choice, ...) METALANG99_PRIV_REC_NEXT(14887, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14887(choice, ...) METALANG99_PRIV_REC_NEXT(14888, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14888(choice, ...) METALANG99_PRIV_REC_NEXT(14889, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14889(choice, ...) METALANG99_PRIV_REC_NEXT(14890, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14890(choice, ...) METALANG99_PRIV_REC_NEXT(14891, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14891(choice, ...) METALANG99_PRIV_REC_NEXT(14892, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14892(choice, ...) METALANG99_PRIV_REC_NEXT(14893, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14893(choice, ...) METALANG99_PRIV_REC_NEXT(14894, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14894(choice, ...) METALANG99_PRIV_REC_NEXT(14895, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14895(choice, ...) METALANG99_PRIV_REC_NEXT(14896, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14896(choice, ...) METALANG99_PRIV_REC_NEXT(14897, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14897(choice, ...) METALANG99_PRIV_REC_NEXT(14898, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14898(choice, ...) METALANG99_PRIV_REC_NEXT(14899, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14899(choice, ...) METALANG99_PRIV_REC_NEXT(14900, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14900(choice, ...) METALANG99_PRIV_REC_NEXT(14901, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14901(choice, ...) METALANG99_PRIV_REC_NEXT(14902, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14902(choice, ...) METALANG99_PRIV_REC_NEXT(14903, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14903(choice, ...) METALANG99_PRIV_REC_NEXT(14904, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14904(choice, ...) METALANG99_PRIV_REC_NEXT(14905, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14905(choice, ...) METALANG99_PRIV_REC_NEXT(14906, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14906(choice, ...) METALANG99_PRIV_REC_NEXT(14907, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14907(choice, ...) METALANG99_PRIV_REC_NEXT(14908, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14908(choice, ...) METALANG99_PRIV_REC_NEXT(14909, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14909(choice, ...) METALANG99_PRIV_REC_NEXT(14910, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14910(choice, ...) METALANG99_PRIV_REC_NEXT(14911, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14911(choice, ...) METALANG99_PRIV_REC_NEXT(14912, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14912(choice, ...) METALANG99_PRIV_REC_NEXT(14913, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14913(choice, ...) METALANG99_PRIV_REC_NEXT(14914, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14914(choice, ...) METALANG99_PRIV_REC_NEXT(14915, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14915(choice, ...) METALANG99_PRIV_REC_NEXT(14916, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14916(choice, ...) METALANG99_PRIV_REC_NEXT(14917, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14917(choice, ...) METALANG99_PRIV_REC_NEXT(14918, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14918(choice, ...) METALANG99_PRIV_REC_NEXT(14919, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14919(choice, ...) METALANG99_PRIV_REC_NEXT(14920, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14920(choice, ...) METALANG99_PRIV_REC_NEXT(14921, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14921(choice, ...) METALANG99_PRIV_REC_NEXT(14922, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14922(choice, ...) METALANG99_PRIV_REC_NEXT(14923, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14923(choice, ...) METALANG99_PRIV_REC_NEXT(14924, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14924(choice, ...) METALANG99_PRIV_REC_NEXT(14925, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14925(choice, ...) METALANG99_PRIV_REC_NEXT(14926, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14926(choice, ...) METALANG99_PRIV_REC_NEXT(14927, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14927(choice, ...) METALANG99_PRIV_REC_NEXT(14928, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14928(choice, ...) METALANG99_PRIV_REC_NEXT(14929, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14929(choice, ...) METALANG99_PRIV_REC_NEXT(14930, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14930(choice, ...) METALANG99_PRIV_REC_NEXT(14931, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14931(choice, ...) METALANG99_PRIV_REC_NEXT(14932, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14932(choice, ...) METALANG99_PRIV_REC_NEXT(14933, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14933(choice, ...) METALANG99_PRIV_REC_NEXT(14934, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14934(choice, ...) METALANG99_PRIV_REC_NEXT(14935, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14935(choice, ...) METALANG99_PRIV_REC_NEXT(14936, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14936(choice, ...) METALANG99_PRIV_REC_NEXT(14937, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14937(choice, ...) METALANG99_PRIV_REC_NEXT(14938, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14938(choice, ...) METALANG99_PRIV_REC_NEXT(14939, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14939(choice, ...) METALANG99_PRIV_REC_NEXT(14940, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14940(choice, ...) METALANG99_PRIV_REC_NEXT(14941, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14941(choice, ...) METALANG99_PRIV_REC_NEXT(14942, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14942(choice, ...) METALANG99_PRIV_REC_NEXT(14943, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14943(choice, ...) METALANG99_PRIV_REC_NEXT(14944, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14944(choice, ...) METALANG99_PRIV_REC_NEXT(14945, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14945(choice, ...) METALANG99_PRIV_REC_NEXT(14946, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14946(choice, ...) METALANG99_PRIV_REC_NEXT(14947, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14947(choice, ...) METALANG99_PRIV_REC_NEXT(14948, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14948(choice, ...) METALANG99_PRIV_REC_NEXT(14949, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14949(choice, ...) METALANG99_PRIV_REC_NEXT(14950, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14950(choice, ...) METALANG99_PRIV_REC_NEXT(14951, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14951(choice, ...) METALANG99_PRIV_REC_NEXT(14952, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14952(choice, ...) METALANG99_PRIV_REC_NEXT(14953, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14953(choice, ...) METALANG99_PRIV_REC_NEXT(14954, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14954(choice, ...) METALANG99_PRIV_REC_NEXT(14955, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14955(choice, ...) METALANG99_PRIV_REC_NEXT(14956, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14956(choice, ...) METALANG99_PRIV_REC_NEXT(14957, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14957(choice, ...) METALANG99_PRIV_REC_NEXT(14958, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14958(choice, ...) METALANG99_PRIV_REC_NEXT(14959, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14959(choice, ...) METALANG99_PRIV_REC_NEXT(14960, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14960(choice, ...) METALANG99_PRIV_REC_NEXT(14961, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14961(choice, ...) METALANG99_PRIV_REC_NEXT(14962, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14962(choice, ...) METALANG99_PRIV_REC_NEXT(14963, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14963(choice, ...) METALANG99_PRIV_REC_NEXT(14964, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14964(choice, ...) METALANG99_PRIV_REC_NEXT(14965, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14965(choice, ...) METALANG99_PRIV_REC_NEXT(14966, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14966(choice, ...) METALANG99_PRIV_REC_NEXT(14967, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14967(choice, ...) METALANG99_PRIV_REC_NEXT(14968, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14968(choice, ...) METALANG99_PRIV_REC_NEXT(14969, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14969(choice, ...) METALANG99_PRIV_REC_NEXT(14970, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14970(choice, ...) METALANG99_PRIV_REC_NEXT(14971, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14971(choice, ...) METALANG99_PRIV_REC_NEXT(14972, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14972(choice, ...) METALANG99_PRIV_REC_NEXT(14973, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14973(choice, ...) METALANG99_PRIV_REC_NEXT(14974, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14974(choice, ...) METALANG99_PRIV_REC_NEXT(14975, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14975(choice, ...) METALANG99_PRIV_REC_NEXT(14976, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14976(choice, ...) METALANG99_PRIV_REC_NEXT(14977, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14977(choice, ...) METALANG99_PRIV_REC_NEXT(14978, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14978(choice, ...) METALANG99_PRIV_REC_NEXT(14979, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14979(choice, ...) METALANG99_PRIV_REC_NEXT(14980, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14980(choice, ...) METALANG99_PRIV_REC_NEXT(14981, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14981(choice, ...) METALANG99_PRIV_REC_NEXT(14982, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14982(choice, ...) METALANG99_PRIV_REC_NEXT(14983, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14983(choice, ...) METALANG99_PRIV_REC_NEXT(14984, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14984(choice, ...) METALANG99_PRIV_REC_NEXT(14985, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14985(choice, ...) METALANG99_PRIV_REC_NEXT(14986, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14986(choice, ...) METALANG99_PRIV_REC_NEXT(14987, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14987(choice, ...) METALANG99_PRIV_REC_NEXT(14988, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14988(choice, ...) METALANG99_PRIV_REC_NEXT(14989, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14989(choice, ...) METALANG99_PRIV_REC_NEXT(14990, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14990(choice, ...) METALANG99_PRIV_REC_NEXT(14991, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14991(choice, ...) METALANG99_PRIV_REC_NEXT(14992, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14992(choice, ...) METALANG99_PRIV_REC_NEXT(14993, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14993(choice, ...) METALANG99_PRIV_REC_NEXT(14994, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14994(choice, ...) METALANG99_PRIV_REC_NEXT(14995, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14995(choice, ...) METALANG99_PRIV_REC_NEXT(14996, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14996(choice, ...) METALANG99_PRIV_REC_NEXT(14997, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14997(choice, ...) METALANG99_PRIV_REC_NEXT(14998, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14998(choice, ...) METALANG99_PRIV_REC_NEXT(14999, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_14999(choice, ...) METALANG99_PRIV_REC_NEXT(15000, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15000(choice, ...) METALANG99_PRIV_REC_NEXT(15001, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15001(choice, ...) METALANG99_PRIV_REC_NEXT(15002, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15002(choice, ...) METALANG99_PRIV_REC_NEXT(15003, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15003(choice, ...) METALANG99_PRIV_REC_NEXT(15004, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15004(choice, ...) METALANG99_PRIV_REC_NEXT(15005, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15005(choice, ...) METALANG99_PRIV_REC_NEXT(15006, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15006(choice, ...) METALANG99_PRIV_REC_NEXT(15007, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15007(choice, ...) METALANG99_PRIV_REC_NEXT(15008, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15008(choice, ...) METALANG99_PRIV_REC_NEXT(15009, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15009(choice, ...) METALANG99_PRIV_REC_NEXT(15010, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15010(choice, ...) METALANG99_PRIV_REC_NEXT(15011, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15011(choice, ...) METALANG99_PRIV_REC_NEXT(15012, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15012(choice, ...) METALANG99_PRIV_REC_NEXT(15013, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15013(choice, ...) METALANG99_PRIV_REC_NEXT(15014, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15014(choice, ...) METALANG99_PRIV_REC_NEXT(15015, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15015(choice, ...) METALANG99_PRIV_REC_NEXT(15016, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15016(choice, ...) METALANG99_PRIV_REC_NEXT(15017, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15017(choice, ...) METALANG99_PRIV_REC_NEXT(15018, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15018(choice, ...) METALANG99_PRIV_REC_NEXT(15019, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15019(choice, ...) METALANG99_PRIV_REC_NEXT(15020, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15020(choice, ...) METALANG99_PRIV_REC_NEXT(15021, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15021(choice, ...) METALANG99_PRIV_REC_NEXT(15022, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15022(choice, ...) METALANG99_PRIV_REC_NEXT(15023, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15023(choice, ...) METALANG99_PRIV_REC_NEXT(15024, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15024(choice, ...) METALANG99_PRIV_REC_NEXT(15025, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15025(choice, ...) METALANG99_PRIV_REC_NEXT(15026, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15026(choice, ...) METALANG99_PRIV_REC_NEXT(15027, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15027(choice, ...) METALANG99_PRIV_REC_NEXT(15028, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15028(choice, ...) METALANG99_PRIV_REC_NEXT(15029, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15029(choice, ...) METALANG99_PRIV_REC_NEXT(15030, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15030(choice, ...) METALANG99_PRIV_REC_NEXT(15031, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15031(choice, ...) METALANG99_PRIV_REC_NEXT(15032, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15032(choice, ...) METALANG99_PRIV_REC_NEXT(15033, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15033(choice, ...) METALANG99_PRIV_REC_NEXT(15034, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15034(choice, ...) METALANG99_PRIV_REC_NEXT(15035, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15035(choice, ...) METALANG99_PRIV_REC_NEXT(15036, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15036(choice, ...) METALANG99_PRIV_REC_NEXT(15037, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15037(choice, ...) METALANG99_PRIV_REC_NEXT(15038, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15038(choice, ...) METALANG99_PRIV_REC_NEXT(15039, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15039(choice, ...) METALANG99_PRIV_REC_NEXT(15040, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15040(choice, ...) METALANG99_PRIV_REC_NEXT(15041, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15041(choice, ...) METALANG99_PRIV_REC_NEXT(15042, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15042(choice, ...) METALANG99_PRIV_REC_NEXT(15043, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15043(choice, ...) METALANG99_PRIV_REC_NEXT(15044, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15044(choice, ...) METALANG99_PRIV_REC_NEXT(15045, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15045(choice, ...) METALANG99_PRIV_REC_NEXT(15046, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15046(choice, ...) METALANG99_PRIV_REC_NEXT(15047, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15047(choice, ...) METALANG99_PRIV_REC_NEXT(15048, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15048(choice, ...) METALANG99_PRIV_REC_NEXT(15049, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15049(choice, ...) METALANG99_PRIV_REC_NEXT(15050, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15050(choice, ...) METALANG99_PRIV_REC_NEXT(15051, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15051(choice, ...) METALANG99_PRIV_REC_NEXT(15052, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15052(choice, ...) METALANG99_PRIV_REC_NEXT(15053, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15053(choice, ...) METALANG99_PRIV_REC_NEXT(15054, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15054(choice, ...) METALANG99_PRIV_REC_NEXT(15055, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15055(choice, ...) METALANG99_PRIV_REC_NEXT(15056, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15056(choice, ...) METALANG99_PRIV_REC_NEXT(15057, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15057(choice, ...) METALANG99_PRIV_REC_NEXT(15058, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15058(choice, ...) METALANG99_PRIV_REC_NEXT(15059, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15059(choice, ...) METALANG99_PRIV_REC_NEXT(15060, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15060(choice, ...) METALANG99_PRIV_REC_NEXT(15061, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15061(choice, ...) METALANG99_PRIV_REC_NEXT(15062, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15062(choice, ...) METALANG99_PRIV_REC_NEXT(15063, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15063(choice, ...) METALANG99_PRIV_REC_NEXT(15064, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15064(choice, ...) METALANG99_PRIV_REC_NEXT(15065, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15065(choice, ...) METALANG99_PRIV_REC_NEXT(15066, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15066(choice, ...) METALANG99_PRIV_REC_NEXT(15067, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15067(choice, ...) METALANG99_PRIV_REC_NEXT(15068, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15068(choice, ...) METALANG99_PRIV_REC_NEXT(15069, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15069(choice, ...) METALANG99_PRIV_REC_NEXT(15070, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15070(choice, ...) METALANG99_PRIV_REC_NEXT(15071, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15071(choice, ...) METALANG99_PRIV_REC_NEXT(15072, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15072(choice, ...) METALANG99_PRIV_REC_NEXT(15073, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15073(choice, ...) METALANG99_PRIV_REC_NEXT(15074, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15074(choice, ...) METALANG99_PRIV_REC_NEXT(15075, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15075(choice, ...) METALANG99_PRIV_REC_NEXT(15076, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15076(choice, ...) METALANG99_PRIV_REC_NEXT(15077, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15077(choice, ...) METALANG99_PRIV_REC_NEXT(15078, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15078(choice, ...) METALANG99_PRIV_REC_NEXT(15079, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15079(choice, ...) METALANG99_PRIV_REC_NEXT(15080, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15080(choice, ...) METALANG99_PRIV_REC_NEXT(15081, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15081(choice, ...) METALANG99_PRIV_REC_NEXT(15082, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15082(choice, ...) METALANG99_PRIV_REC_NEXT(15083, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15083(choice, ...) METALANG99_PRIV_REC_NEXT(15084, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15084(choice, ...) METALANG99_PRIV_REC_NEXT(15085, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15085(choice, ...) METALANG99_PRIV_REC_NEXT(15086, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15086(choice, ...) METALANG99_PRIV_REC_NEXT(15087, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15087(choice, ...) METALANG99_PRIV_REC_NEXT(15088, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15088(choice, ...) METALANG99_PRIV_REC_NEXT(15089, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15089(choice, ...) METALANG99_PRIV_REC_NEXT(15090, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15090(choice, ...) METALANG99_PRIV_REC_NEXT(15091, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15091(choice, ...) METALANG99_PRIV_REC_NEXT(15092, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15092(choice, ...) METALANG99_PRIV_REC_NEXT(15093, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15093(choice, ...) METALANG99_PRIV_REC_NEXT(15094, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15094(choice, ...) METALANG99_PRIV_REC_NEXT(15095, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15095(choice, ...) METALANG99_PRIV_REC_NEXT(15096, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15096(choice, ...) METALANG99_PRIV_REC_NEXT(15097, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15097(choice, ...) METALANG99_PRIV_REC_NEXT(15098, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15098(choice, ...) METALANG99_PRIV_REC_NEXT(15099, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15099(choice, ...) METALANG99_PRIV_REC_NEXT(15100, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15100(choice, ...) METALANG99_PRIV_REC_NEXT(15101, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15101(choice, ...) METALANG99_PRIV_REC_NEXT(15102, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15102(choice, ...) METALANG99_PRIV_REC_NEXT(15103, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15103(choice, ...) METALANG99_PRIV_REC_NEXT(15104, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15104(choice, ...) METALANG99_PRIV_REC_NEXT(15105, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15105(choice, ...) METALANG99_PRIV_REC_NEXT(15106, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15106(choice, ...) METALANG99_PRIV_REC_NEXT(15107, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15107(choice, ...) METALANG99_PRIV_REC_NEXT(15108, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15108(choice, ...) METALANG99_PRIV_REC_NEXT(15109, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15109(choice, ...) METALANG99_PRIV_REC_NEXT(15110, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15110(choice, ...) METALANG99_PRIV_REC_NEXT(15111, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15111(choice, ...) METALANG99_PRIV_REC_NEXT(15112, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15112(choice, ...) METALANG99_PRIV_REC_NEXT(15113, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15113(choice, ...) METALANG99_PRIV_REC_NEXT(15114, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15114(choice, ...) METALANG99_PRIV_REC_NEXT(15115, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15115(choice, ...) METALANG99_PRIV_REC_NEXT(15116, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15116(choice, ...) METALANG99_PRIV_REC_NEXT(15117, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15117(choice, ...) METALANG99_PRIV_REC_NEXT(15118, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15118(choice, ...) METALANG99_PRIV_REC_NEXT(15119, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15119(choice, ...) METALANG99_PRIV_REC_NEXT(15120, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15120(choice, ...) METALANG99_PRIV_REC_NEXT(15121, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15121(choice, ...) METALANG99_PRIV_REC_NEXT(15122, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15122(choice, ...) METALANG99_PRIV_REC_NEXT(15123, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15123(choice, ...) METALANG99_PRIV_REC_NEXT(15124, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15124(choice, ...) METALANG99_PRIV_REC_NEXT(15125, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15125(choice, ...) METALANG99_PRIV_REC_NEXT(15126, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15126(choice, ...) METALANG99_PRIV_REC_NEXT(15127, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15127(choice, ...) METALANG99_PRIV_REC_NEXT(15128, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15128(choice, ...) METALANG99_PRIV_REC_NEXT(15129, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15129(choice, ...) METALANG99_PRIV_REC_NEXT(15130, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15130(choice, ...) METALANG99_PRIV_REC_NEXT(15131, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15131(choice, ...) METALANG99_PRIV_REC_NEXT(15132, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15132(choice, ...) METALANG99_PRIV_REC_NEXT(15133, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15133(choice, ...) METALANG99_PRIV_REC_NEXT(15134, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15134(choice, ...) METALANG99_PRIV_REC_NEXT(15135, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15135(choice, ...) METALANG99_PRIV_REC_NEXT(15136, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15136(choice, ...) METALANG99_PRIV_REC_NEXT(15137, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15137(choice, ...) METALANG99_PRIV_REC_NEXT(15138, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15138(choice, ...) METALANG99_PRIV_REC_NEXT(15139, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15139(choice, ...) METALANG99_PRIV_REC_NEXT(15140, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15140(choice, ...) METALANG99_PRIV_REC_NEXT(15141, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15141(choice, ...) METALANG99_PRIV_REC_NEXT(15142, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15142(choice, ...) METALANG99_PRIV_REC_NEXT(15143, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15143(choice, ...) METALANG99_PRIV_REC_NEXT(15144, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15144(choice, ...) METALANG99_PRIV_REC_NEXT(15145, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15145(choice, ...) METALANG99_PRIV_REC_NEXT(15146, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15146(choice, ...) METALANG99_PRIV_REC_NEXT(15147, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15147(choice, ...) METALANG99_PRIV_REC_NEXT(15148, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15148(choice, ...) METALANG99_PRIV_REC_NEXT(15149, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15149(choice, ...) METALANG99_PRIV_REC_NEXT(15150, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15150(choice, ...) METALANG99_PRIV_REC_NEXT(15151, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15151(choice, ...) METALANG99_PRIV_REC_NEXT(15152, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15152(choice, ...) METALANG99_PRIV_REC_NEXT(15153, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15153(choice, ...) METALANG99_PRIV_REC_NEXT(15154, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15154(choice, ...) METALANG99_PRIV_REC_NEXT(15155, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15155(choice, ...) METALANG99_PRIV_REC_NEXT(15156, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15156(choice, ...) METALANG99_PRIV_REC_NEXT(15157, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15157(choice, ...) METALANG99_PRIV_REC_NEXT(15158, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15158(choice, ...) METALANG99_PRIV_REC_NEXT(15159, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15159(choice, ...) METALANG99_PRIV_REC_NEXT(15160, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15160(choice, ...) METALANG99_PRIV_REC_NEXT(15161, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15161(choice, ...) METALANG99_PRIV_REC_NEXT(15162, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15162(choice, ...) METALANG99_PRIV_REC_NEXT(15163, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15163(choice, ...) METALANG99_PRIV_REC_NEXT(15164, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15164(choice, ...) METALANG99_PRIV_REC_NEXT(15165, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15165(choice, ...) METALANG99_PRIV_REC_NEXT(15166, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15166(choice, ...) METALANG99_PRIV_REC_NEXT(15167, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15167(choice, ...) METALANG99_PRIV_REC_NEXT(15168, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15168(choice, ...) METALANG99_PRIV_REC_NEXT(15169, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15169(choice, ...) METALANG99_PRIV_REC_NEXT(15170, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15170(choice, ...) METALANG99_PRIV_REC_NEXT(15171, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15171(choice, ...) METALANG99_PRIV_REC_NEXT(15172, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15172(choice, ...) METALANG99_PRIV_REC_NEXT(15173, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15173(choice, ...) METALANG99_PRIV_REC_NEXT(15174, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15174(choice, ...) METALANG99_PRIV_REC_NEXT(15175, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15175(choice, ...) METALANG99_PRIV_REC_NEXT(15176, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15176(choice, ...) METALANG99_PRIV_REC_NEXT(15177, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15177(choice, ...) METALANG99_PRIV_REC_NEXT(15178, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15178(choice, ...) METALANG99_PRIV_REC_NEXT(15179, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15179(choice, ...) METALANG99_PRIV_REC_NEXT(15180, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15180(choice, ...) METALANG99_PRIV_REC_NEXT(15181, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15181(choice, ...) METALANG99_PRIV_REC_NEXT(15182, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15182(choice, ...) METALANG99_PRIV_REC_NEXT(15183, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15183(choice, ...) METALANG99_PRIV_REC_NEXT(15184, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15184(choice, ...) METALANG99_PRIV_REC_NEXT(15185, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15185(choice, ...) METALANG99_PRIV_REC_NEXT(15186, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15186(choice, ...) METALANG99_PRIV_REC_NEXT(15187, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15187(choice, ...) METALANG99_PRIV_REC_NEXT(15188, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15188(choice, ...) METALANG99_PRIV_REC_NEXT(15189, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15189(choice, ...) METALANG99_PRIV_REC_NEXT(15190, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15190(choice, ...) METALANG99_PRIV_REC_NEXT(15191, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15191(choice, ...) METALANG99_PRIV_REC_NEXT(15192, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15192(choice, ...) METALANG99_PRIV_REC_NEXT(15193, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15193(choice, ...) METALANG99_PRIV_REC_NEXT(15194, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15194(choice, ...) METALANG99_PRIV_REC_NEXT(15195, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15195(choice, ...) METALANG99_PRIV_REC_NEXT(15196, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15196(choice, ...) METALANG99_PRIV_REC_NEXT(15197, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15197(choice, ...) METALANG99_PRIV_REC_NEXT(15198, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15198(choice, ...) METALANG99_PRIV_REC_NEXT(15199, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15199(choice, ...) METALANG99_PRIV_REC_NEXT(15200, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15200(choice, ...) METALANG99_PRIV_REC_NEXT(15201, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15201(choice, ...) METALANG99_PRIV_REC_NEXT(15202, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15202(choice, ...) METALANG99_PRIV_REC_NEXT(15203, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15203(choice, ...) METALANG99_PRIV_REC_NEXT(15204, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15204(choice, ...) METALANG99_PRIV_REC_NEXT(15205, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15205(choice, ...) METALANG99_PRIV_REC_NEXT(15206, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15206(choice, ...) METALANG99_PRIV_REC_NEXT(15207, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15207(choice, ...) METALANG99_PRIV_REC_NEXT(15208, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15208(choice, ...) METALANG99_PRIV_REC_NEXT(15209, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15209(choice, ...) METALANG99_PRIV_REC_NEXT(15210, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15210(choice, ...) METALANG99_PRIV_REC_NEXT(15211, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15211(choice, ...) METALANG99_PRIV_REC_NEXT(15212, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15212(choice, ...) METALANG99_PRIV_REC_NEXT(15213, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15213(choice, ...) METALANG99_PRIV_REC_NEXT(15214, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15214(choice, ...) METALANG99_PRIV_REC_NEXT(15215, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15215(choice, ...) METALANG99_PRIV_REC_NEXT(15216, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15216(choice, ...) METALANG99_PRIV_REC_NEXT(15217, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15217(choice, ...) METALANG99_PRIV_REC_NEXT(15218, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15218(choice, ...) METALANG99_PRIV_REC_NEXT(15219, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15219(choice, ...) METALANG99_PRIV_REC_NEXT(15220, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15220(choice, ...) METALANG99_PRIV_REC_NEXT(15221, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15221(choice, ...) METALANG99_PRIV_REC_NEXT(15222, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15222(choice, ...) METALANG99_PRIV_REC_NEXT(15223, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15223(choice, ...) METALANG99_PRIV_REC_NEXT(15224, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15224(choice, ...) METALANG99_PRIV_REC_NEXT(15225, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15225(choice, ...) METALANG99_PRIV_REC_NEXT(15226, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15226(choice, ...) METALANG99_PRIV_REC_NEXT(15227, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15227(choice, ...) METALANG99_PRIV_REC_NEXT(15228, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15228(choice, ...) METALANG99_PRIV_REC_NEXT(15229, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15229(choice, ...) METALANG99_PRIV_REC_NEXT(15230, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15230(choice, ...) METALANG99_PRIV_REC_NEXT(15231, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15231(choice, ...) METALANG99_PRIV_REC_NEXT(15232, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15232(choice, ...) METALANG99_PRIV_REC_NEXT(15233, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15233(choice, ...) METALANG99_PRIV_REC_NEXT(15234, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15234(choice, ...) METALANG99_PRIV_REC_NEXT(15235, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15235(choice, ...) METALANG99_PRIV_REC_NEXT(15236, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15236(choice, ...) METALANG99_PRIV_REC_NEXT(15237, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15237(choice, ...) METALANG99_PRIV_REC_NEXT(15238, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15238(choice, ...) METALANG99_PRIV_REC_NEXT(15239, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15239(choice, ...) METALANG99_PRIV_REC_NEXT(15240, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15240(choice, ...) METALANG99_PRIV_REC_NEXT(15241, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15241(choice, ...) METALANG99_PRIV_REC_NEXT(15242, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15242(choice, ...) METALANG99_PRIV_REC_NEXT(15243, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15243(choice, ...) METALANG99_PRIV_REC_NEXT(15244, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15244(choice, ...) METALANG99_PRIV_REC_NEXT(15245, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15245(choice, ...) METALANG99_PRIV_REC_NEXT(15246, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15246(choice, ...) METALANG99_PRIV_REC_NEXT(15247, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15247(choice, ...) METALANG99_PRIV_REC_NEXT(15248, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15248(choice, ...) METALANG99_PRIV_REC_NEXT(15249, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15249(choice, ...) METALANG99_PRIV_REC_NEXT(15250, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15250(choice, ...) METALANG99_PRIV_REC_NEXT(15251, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15251(choice, ...) METALANG99_PRIV_REC_NEXT(15252, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15252(choice, ...) METALANG99_PRIV_REC_NEXT(15253, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15253(choice, ...) METALANG99_PRIV_REC_NEXT(15254, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15254(choice, ...) METALANG99_PRIV_REC_NEXT(15255, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15255(choice, ...) METALANG99_PRIV_REC_NEXT(15256, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15256(choice, ...) METALANG99_PRIV_REC_NEXT(15257, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15257(choice, ...) METALANG99_PRIV_REC_NEXT(15258, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15258(choice, ...) METALANG99_PRIV_REC_NEXT(15259, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15259(choice, ...) METALANG99_PRIV_REC_NEXT(15260, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15260(choice, ...) METALANG99_PRIV_REC_NEXT(15261, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15261(choice, ...) METALANG99_PRIV_REC_NEXT(15262, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15262(choice, ...) METALANG99_PRIV_REC_NEXT(15263, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15263(choice, ...) METALANG99_PRIV_REC_NEXT(15264, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15264(choice, ...) METALANG99_PRIV_REC_NEXT(15265, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15265(choice, ...) METALANG99_PRIV_REC_NEXT(15266, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15266(choice, ...) METALANG99_PRIV_REC_NEXT(15267, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15267(choice, ...) METALANG99_PRIV_REC_NEXT(15268, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15268(choice, ...) METALANG99_PRIV_REC_NEXT(15269, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15269(choice, ...) METALANG99_PRIV_REC_NEXT(15270, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15270(choice, ...) METALANG99_PRIV_REC_NEXT(15271, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15271(choice, ...) METALANG99_PRIV_REC_NEXT(15272, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15272(choice, ...) METALANG99_PRIV_REC_NEXT(15273, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15273(choice, ...) METALANG99_PRIV_REC_NEXT(15274, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15274(choice, ...) METALANG99_PRIV_REC_NEXT(15275, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15275(choice, ...) METALANG99_PRIV_REC_NEXT(15276, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15276(choice, ...) METALANG99_PRIV_REC_NEXT(15277, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15277(choice, ...) METALANG99_PRIV_REC_NEXT(15278, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15278(choice, ...) METALANG99_PRIV_REC_NEXT(15279, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15279(choice, ...) METALANG99_PRIV_REC_NEXT(15280, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15280(choice, ...) METALANG99_PRIV_REC_NEXT(15281, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15281(choice, ...) METALANG99_PRIV_REC_NEXT(15282, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15282(choice, ...) METALANG99_PRIV_REC_NEXT(15283, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15283(choice, ...) METALANG99_PRIV_REC_NEXT(15284, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15284(choice, ...) METALANG99_PRIV_REC_NEXT(15285, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15285(choice, ...) METALANG99_PRIV_REC_NEXT(15286, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15286(choice, ...) METALANG99_PRIV_REC_NEXT(15287, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15287(choice, ...) METALANG99_PRIV_REC_NEXT(15288, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15288(choice, ...) METALANG99_PRIV_REC_NEXT(15289, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15289(choice, ...) METALANG99_PRIV_REC_NEXT(15290, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15290(choice, ...) METALANG99_PRIV_REC_NEXT(15291, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15291(choice, ...) METALANG99_PRIV_REC_NEXT(15292, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15292(choice, ...) METALANG99_PRIV_REC_NEXT(15293, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15293(choice, ...) METALANG99_PRIV_REC_NEXT(15294, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15294(choice, ...) METALANG99_PRIV_REC_NEXT(15295, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15295(choice, ...) METALANG99_PRIV_REC_NEXT(15296, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15296(choice, ...) METALANG99_PRIV_REC_NEXT(15297, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15297(choice, ...) METALANG99_PRIV_REC_NEXT(15298, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15298(choice, ...) METALANG99_PRIV_REC_NEXT(15299, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15299(choice, ...) METALANG99_PRIV_REC_NEXT(15300, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15300(choice, ...) METALANG99_PRIV_REC_NEXT(15301, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15301(choice, ...) METALANG99_PRIV_REC_NEXT(15302, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15302(choice, ...) METALANG99_PRIV_REC_NEXT(15303, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15303(choice, ...) METALANG99_PRIV_REC_NEXT(15304, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15304(choice, ...) METALANG99_PRIV_REC_NEXT(15305, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15305(choice, ...) METALANG99_PRIV_REC_NEXT(15306, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15306(choice, ...) METALANG99_PRIV_REC_NEXT(15307, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15307(choice, ...) METALANG99_PRIV_REC_NEXT(15308, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15308(choice, ...) METALANG99_PRIV_REC_NEXT(15309, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15309(choice, ...) METALANG99_PRIV_REC_NEXT(15310, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15310(choice, ...) METALANG99_PRIV_REC_NEXT(15311, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15311(choice, ...) METALANG99_PRIV_REC_NEXT(15312, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15312(choice, ...) METALANG99_PRIV_REC_NEXT(15313, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15313(choice, ...) METALANG99_PRIV_REC_NEXT(15314, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15314(choice, ...) METALANG99_PRIV_REC_NEXT(15315, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15315(choice, ...) METALANG99_PRIV_REC_NEXT(15316, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15316(choice, ...) METALANG99_PRIV_REC_NEXT(15317, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15317(choice, ...) METALANG99_PRIV_REC_NEXT(15318, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15318(choice, ...) METALANG99_PRIV_REC_NEXT(15319, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15319(choice, ...) METALANG99_PRIV_REC_NEXT(15320, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15320(choice, ...) METALANG99_PRIV_REC_NEXT(15321, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15321(choice, ...) METALANG99_PRIV_REC_NEXT(15322, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15322(choice, ...) METALANG99_PRIV_REC_NEXT(15323, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15323(choice, ...) METALANG99_PRIV_REC_NEXT(15324, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15324(choice, ...) METALANG99_PRIV_REC_NEXT(15325, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15325(choice, ...) METALANG99_PRIV_REC_NEXT(15326, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15326(choice, ...) METALANG99_PRIV_REC_NEXT(15327, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15327(choice, ...) METALANG99_PRIV_REC_NEXT(15328, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15328(choice, ...) METALANG99_PRIV_REC_NEXT(15329, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15329(choice, ...) METALANG99_PRIV_REC_NEXT(15330, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15330(choice, ...) METALANG99_PRIV_REC_NEXT(15331, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15331(choice, ...) METALANG99_PRIV_REC_NEXT(15332, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15332(choice, ...) METALANG99_PRIV_REC_NEXT(15333, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15333(choice, ...) METALANG99_PRIV_REC_NEXT(15334, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15334(choice, ...) METALANG99_PRIV_REC_NEXT(15335, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15335(choice, ...) METALANG99_PRIV_REC_NEXT(15336, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15336(choice, ...) METALANG99_PRIV_REC_NEXT(15337, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15337(choice, ...) METALANG99_PRIV_REC_NEXT(15338, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15338(choice, ...) METALANG99_PRIV_REC_NEXT(15339, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15339(choice, ...) METALANG99_PRIV_REC_NEXT(15340, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15340(choice, ...) METALANG99_PRIV_REC_NEXT(15341, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15341(choice, ...) METALANG99_PRIV_REC_NEXT(15342, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15342(choice, ...) METALANG99_PRIV_REC_NEXT(15343, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15343(choice, ...) METALANG99_PRIV_REC_NEXT(15344, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15344(choice, ...) METALANG99_PRIV_REC_NEXT(15345, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15345(choice, ...) METALANG99_PRIV_REC_NEXT(15346, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15346(choice, ...) METALANG99_PRIV_REC_NEXT(15347, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15347(choice, ...) METALANG99_PRIV_REC_NEXT(15348, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15348(choice, ...) METALANG99_PRIV_REC_NEXT(15349, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15349(choice, ...) METALANG99_PRIV_REC_NEXT(15350, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15350(choice, ...) METALANG99_PRIV_REC_NEXT(15351, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15351(choice, ...) METALANG99_PRIV_REC_NEXT(15352, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15352(choice, ...) METALANG99_PRIV_REC_NEXT(15353, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15353(choice, ...) METALANG99_PRIV_REC_NEXT(15354, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15354(choice, ...) METALANG99_PRIV_REC_NEXT(15355, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15355(choice, ...) METALANG99_PRIV_REC_NEXT(15356, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15356(choice, ...) METALANG99_PRIV_REC_NEXT(15357, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15357(choice, ...) METALANG99_PRIV_REC_NEXT(15358, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15358(choice, ...) METALANG99_PRIV_REC_NEXT(15359, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15359(choice, ...) METALANG99_PRIV_REC_NEXT(15360, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15360(choice, ...) METALANG99_PRIV_REC_NEXT(15361, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15361(choice, ...) METALANG99_PRIV_REC_NEXT(15362, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15362(choice, ...) METALANG99_PRIV_REC_NEXT(15363, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15363(choice, ...) METALANG99_PRIV_REC_NEXT(15364, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15364(choice, ...) METALANG99_PRIV_REC_NEXT(15365, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15365(choice, ...) METALANG99_PRIV_REC_NEXT(15366, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15366(choice, ...) METALANG99_PRIV_REC_NEXT(15367, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15367(choice, ...) METALANG99_PRIV_REC_NEXT(15368, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15368(choice, ...) METALANG99_PRIV_REC_NEXT(15369, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15369(choice, ...) METALANG99_PRIV_REC_NEXT(15370, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15370(choice, ...) METALANG99_PRIV_REC_NEXT(15371, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15371(choice, ...) METALANG99_PRIV_REC_NEXT(15372, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15372(choice, ...) METALANG99_PRIV_REC_NEXT(15373, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15373(choice, ...) METALANG99_PRIV_REC_NEXT(15374, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15374(choice, ...) METALANG99_PRIV_REC_NEXT(15375, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15375(choice, ...) METALANG99_PRIV_REC_NEXT(15376, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15376(choice, ...) METALANG99_PRIV_REC_NEXT(15377, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15377(choice, ...) METALANG99_PRIV_REC_NEXT(15378, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15378(choice, ...) METALANG99_PRIV_REC_NEXT(15379, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15379(choice, ...) METALANG99_PRIV_REC_NEXT(15380, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15380(choice, ...) METALANG99_PRIV_REC_NEXT(15381, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15381(choice, ...) METALANG99_PRIV_REC_NEXT(15382, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15382(choice, ...) METALANG99_PRIV_REC_NEXT(15383, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15383(choice, ...) METALANG99_PRIV_REC_NEXT(15384, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15384(choice, ...) METALANG99_PRIV_REC_NEXT(15385, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15385(choice, ...) METALANG99_PRIV_REC_NEXT(15386, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15386(choice, ...) METALANG99_PRIV_REC_NEXT(15387, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15387(choice, ...) METALANG99_PRIV_REC_NEXT(15388, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15388(choice, ...) METALANG99_PRIV_REC_NEXT(15389, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15389(choice, ...) METALANG99_PRIV_REC_NEXT(15390, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15390(choice, ...) METALANG99_PRIV_REC_NEXT(15391, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15391(choice, ...) METALANG99_PRIV_REC_NEXT(15392, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15392(choice, ...) METALANG99_PRIV_REC_NEXT(15393, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15393(choice, ...) METALANG99_PRIV_REC_NEXT(15394, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15394(choice, ...) METALANG99_PRIV_REC_NEXT(15395, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15395(choice, ...) METALANG99_PRIV_REC_NEXT(15396, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15396(choice, ...) METALANG99_PRIV_REC_NEXT(15397, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15397(choice, ...) METALANG99_PRIV_REC_NEXT(15398, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15398(choice, ...) METALANG99_PRIV_REC_NEXT(15399, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15399(choice, ...) METALANG99_PRIV_REC_NEXT(15400, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15400(choice, ...) METALANG99_PRIV_REC_NEXT(15401, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15401(choice, ...) METALANG99_PRIV_REC_NEXT(15402, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15402(choice, ...) METALANG99_PRIV_REC_NEXT(15403, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15403(choice, ...) METALANG99_PRIV_REC_NEXT(15404, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15404(choice, ...) METALANG99_PRIV_REC_NEXT(15405, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15405(choice, ...) METALANG99_PRIV_REC_NEXT(15406, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15406(choice, ...) METALANG99_PRIV_REC_NEXT(15407, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15407(choice, ...) METALANG99_PRIV_REC_NEXT(15408, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15408(choice, ...) METALANG99_PRIV_REC_NEXT(15409, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15409(choice, ...) METALANG99_PRIV_REC_NEXT(15410, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15410(choice, ...) METALANG99_PRIV_REC_NEXT(15411, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15411(choice, ...) METALANG99_PRIV_REC_NEXT(15412, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15412(choice, ...) METALANG99_PRIV_REC_NEXT(15413, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15413(choice, ...) METALANG99_PRIV_REC_NEXT(15414, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15414(choice, ...) METALANG99_PRIV_REC_NEXT(15415, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15415(choice, ...) METALANG99_PRIV_REC_NEXT(15416, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15416(choice, ...) METALANG99_PRIV_REC_NEXT(15417, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15417(choice, ...) METALANG99_PRIV_REC_NEXT(15418, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15418(choice, ...) METALANG99_PRIV_REC_NEXT(15419, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15419(choice, ...) METALANG99_PRIV_REC_NEXT(15420, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15420(choice, ...) METALANG99_PRIV_REC_NEXT(15421, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15421(choice, ...) METALANG99_PRIV_REC_NEXT(15422, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15422(choice, ...) METALANG99_PRIV_REC_NEXT(15423, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15423(choice, ...) METALANG99_PRIV_REC_NEXT(15424, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15424(choice, ...) METALANG99_PRIV_REC_NEXT(15425, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15425(choice, ...) METALANG99_PRIV_REC_NEXT(15426, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15426(choice, ...) METALANG99_PRIV_REC_NEXT(15427, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15427(choice, ...) METALANG99_PRIV_REC_NEXT(15428, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15428(choice, ...) METALANG99_PRIV_REC_NEXT(15429, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15429(choice, ...) METALANG99_PRIV_REC_NEXT(15430, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15430(choice, ...) METALANG99_PRIV_REC_NEXT(15431, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15431(choice, ...) METALANG99_PRIV_REC_NEXT(15432, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15432(choice, ...) METALANG99_PRIV_REC_NEXT(15433, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15433(choice, ...) METALANG99_PRIV_REC_NEXT(15434, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15434(choice, ...) METALANG99_PRIV_REC_NEXT(15435, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15435(choice, ...) METALANG99_PRIV_REC_NEXT(15436, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15436(choice, ...) METALANG99_PRIV_REC_NEXT(15437, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15437(choice, ...) METALANG99_PRIV_REC_NEXT(15438, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15438(choice, ...) METALANG99_PRIV_REC_NEXT(15439, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15439(choice, ...) METALANG99_PRIV_REC_NEXT(15440, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15440(choice, ...) METALANG99_PRIV_REC_NEXT(15441, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15441(choice, ...) METALANG99_PRIV_REC_NEXT(15442, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15442(choice, ...) METALANG99_PRIV_REC_NEXT(15443, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15443(choice, ...) METALANG99_PRIV_REC_NEXT(15444, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15444(choice, ...) METALANG99_PRIV_REC_NEXT(15445, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15445(choice, ...) METALANG99_PRIV_REC_NEXT(15446, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15446(choice, ...) METALANG99_PRIV_REC_NEXT(15447, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15447(choice, ...) METALANG99_PRIV_REC_NEXT(15448, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15448(choice, ...) METALANG99_PRIV_REC_NEXT(15449, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15449(choice, ...) METALANG99_PRIV_REC_NEXT(15450, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15450(choice, ...) METALANG99_PRIV_REC_NEXT(15451, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15451(choice, ...) METALANG99_PRIV_REC_NEXT(15452, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15452(choice, ...) METALANG99_PRIV_REC_NEXT(15453, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15453(choice, ...) METALANG99_PRIV_REC_NEXT(15454, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15454(choice, ...) METALANG99_PRIV_REC_NEXT(15455, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15455(choice, ...) METALANG99_PRIV_REC_NEXT(15456, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15456(choice, ...) METALANG99_PRIV_REC_NEXT(15457, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15457(choice, ...) METALANG99_PRIV_REC_NEXT(15458, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15458(choice, ...) METALANG99_PRIV_REC_NEXT(15459, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15459(choice, ...) METALANG99_PRIV_REC_NEXT(15460, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15460(choice, ...) METALANG99_PRIV_REC_NEXT(15461, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15461(choice, ...) METALANG99_PRIV_REC_NEXT(15462, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15462(choice, ...) METALANG99_PRIV_REC_NEXT(15463, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15463(choice, ...) METALANG99_PRIV_REC_NEXT(15464, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15464(choice, ...) METALANG99_PRIV_REC_NEXT(15465, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15465(choice, ...) METALANG99_PRIV_REC_NEXT(15466, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15466(choice, ...) METALANG99_PRIV_REC_NEXT(15467, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15467(choice, ...) METALANG99_PRIV_REC_NEXT(15468, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15468(choice, ...) METALANG99_PRIV_REC_NEXT(15469, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15469(choice, ...) METALANG99_PRIV_REC_NEXT(15470, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15470(choice, ...) METALANG99_PRIV_REC_NEXT(15471, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15471(choice, ...) METALANG99_PRIV_REC_NEXT(15472, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15472(choice, ...) METALANG99_PRIV_REC_NEXT(15473, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15473(choice, ...) METALANG99_PRIV_REC_NEXT(15474, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15474(choice, ...) METALANG99_PRIV_REC_NEXT(15475, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15475(choice, ...) METALANG99_PRIV_REC_NEXT(15476, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15476(choice, ...) METALANG99_PRIV_REC_NEXT(15477, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15477(choice, ...) METALANG99_PRIV_REC_NEXT(15478, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15478(choice, ...) METALANG99_PRIV_REC_NEXT(15479, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15479(choice, ...) METALANG99_PRIV_REC_NEXT(15480, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15480(choice, ...) METALANG99_PRIV_REC_NEXT(15481, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15481(choice, ...) METALANG99_PRIV_REC_NEXT(15482, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15482(choice, ...) METALANG99_PRIV_REC_NEXT(15483, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15483(choice, ...) METALANG99_PRIV_REC_NEXT(15484, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15484(choice, ...) METALANG99_PRIV_REC_NEXT(15485, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15485(choice, ...) METALANG99_PRIV_REC_NEXT(15486, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15486(choice, ...) METALANG99_PRIV_REC_NEXT(15487, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15487(choice, ...) METALANG99_PRIV_REC_NEXT(15488, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15488(choice, ...) METALANG99_PRIV_REC_NEXT(15489, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15489(choice, ...) METALANG99_PRIV_REC_NEXT(15490, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15490(choice, ...) METALANG99_PRIV_REC_NEXT(15491, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15491(choice, ...) METALANG99_PRIV_REC_NEXT(15492, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15492(choice, ...) METALANG99_PRIV_REC_NEXT(15493, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15493(choice, ...) METALANG99_PRIV_REC_NEXT(15494, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15494(choice, ...) METALANG99_PRIV_REC_NEXT(15495, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15495(choice, ...) METALANG99_PRIV_REC_NEXT(15496, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15496(choice, ...) METALANG99_PRIV_REC_NEXT(15497, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15497(choice, ...) METALANG99_PRIV_REC_NEXT(15498, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15498(choice, ...) METALANG99_PRIV_REC_NEXT(15499, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15499(choice, ...) METALANG99_PRIV_REC_NEXT(15500, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15500(choice, ...) METALANG99_PRIV_REC_NEXT(15501, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15501(choice, ...) METALANG99_PRIV_REC_NEXT(15502, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15502(choice, ...) METALANG99_PRIV_REC_NEXT(15503, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15503(choice, ...) METALANG99_PRIV_REC_NEXT(15504, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15504(choice, ...) METALANG99_PRIV_REC_NEXT(15505, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15505(choice, ...) METALANG99_PRIV_REC_NEXT(15506, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15506(choice, ...) METALANG99_PRIV_REC_NEXT(15507, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15507(choice, ...) METALANG99_PRIV_REC_NEXT(15508, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15508(choice, ...) METALANG99_PRIV_REC_NEXT(15509, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15509(choice, ...) METALANG99_PRIV_REC_NEXT(15510, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15510(choice, ...) METALANG99_PRIV_REC_NEXT(15511, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15511(choice, ...) METALANG99_PRIV_REC_NEXT(15512, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15512(choice, ...) METALANG99_PRIV_REC_NEXT(15513, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15513(choice, ...) METALANG99_PRIV_REC_NEXT(15514, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15514(choice, ...) METALANG99_PRIV_REC_NEXT(15515, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15515(choice, ...) METALANG99_PRIV_REC_NEXT(15516, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15516(choice, ...) METALANG99_PRIV_REC_NEXT(15517, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15517(choice, ...) METALANG99_PRIV_REC_NEXT(15518, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15518(choice, ...) METALANG99_PRIV_REC_NEXT(15519, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15519(choice, ...) METALANG99_PRIV_REC_NEXT(15520, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15520(choice, ...) METALANG99_PRIV_REC_NEXT(15521, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15521(choice, ...) METALANG99_PRIV_REC_NEXT(15522, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15522(choice, ...) METALANG99_PRIV_REC_NEXT(15523, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15523(choice, ...) METALANG99_PRIV_REC_NEXT(15524, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15524(choice, ...) METALANG99_PRIV_REC_NEXT(15525, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15525(choice, ...) METALANG99_PRIV_REC_NEXT(15526, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15526(choice, ...) METALANG99_PRIV_REC_NEXT(15527, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15527(choice, ...) METALANG99_PRIV_REC_NEXT(15528, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15528(choice, ...) METALANG99_PRIV_REC_NEXT(15529, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15529(choice, ...) METALANG99_PRIV_REC_NEXT(15530, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15530(choice, ...) METALANG99_PRIV_REC_NEXT(15531, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15531(choice, ...) METALANG99_PRIV_REC_NEXT(15532, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15532(choice, ...) METALANG99_PRIV_REC_NEXT(15533, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15533(choice, ...) METALANG99_PRIV_REC_NEXT(15534, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15534(choice, ...) METALANG99_PRIV_REC_NEXT(15535, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15535(choice, ...) METALANG99_PRIV_REC_NEXT(15536, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15536(choice, ...) METALANG99_PRIV_REC_NEXT(15537, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15537(choice, ...) METALANG99_PRIV_REC_NEXT(15538, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15538(choice, ...) METALANG99_PRIV_REC_NEXT(15539, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15539(choice, ...) METALANG99_PRIV_REC_NEXT(15540, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15540(choice, ...) METALANG99_PRIV_REC_NEXT(15541, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15541(choice, ...) METALANG99_PRIV_REC_NEXT(15542, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15542(choice, ...) METALANG99_PRIV_REC_NEXT(15543, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15543(choice, ...) METALANG99_PRIV_REC_NEXT(15544, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15544(choice, ...) METALANG99_PRIV_REC_NEXT(15545, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15545(choice, ...) METALANG99_PRIV_REC_NEXT(15546, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15546(choice, ...) METALANG99_PRIV_REC_NEXT(15547, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15547(choice, ...) METALANG99_PRIV_REC_NEXT(15548, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15548(choice, ...) METALANG99_PRIV_REC_NEXT(15549, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15549(choice, ...) METALANG99_PRIV_REC_NEXT(15550, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15550(choice, ...) METALANG99_PRIV_REC_NEXT(15551, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15551(choice, ...) METALANG99_PRIV_REC_NEXT(15552, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15552(choice, ...) METALANG99_PRIV_REC_NEXT(15553, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15553(choice, ...) METALANG99_PRIV_REC_NEXT(15554, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15554(choice, ...) METALANG99_PRIV_REC_NEXT(15555, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15555(choice, ...) METALANG99_PRIV_REC_NEXT(15556, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15556(choice, ...) METALANG99_PRIV_REC_NEXT(15557, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15557(choice, ...) METALANG99_PRIV_REC_NEXT(15558, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15558(choice, ...) METALANG99_PRIV_REC_NEXT(15559, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15559(choice, ...) METALANG99_PRIV_REC_NEXT(15560, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15560(choice, ...) METALANG99_PRIV_REC_NEXT(15561, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15561(choice, ...) METALANG99_PRIV_REC_NEXT(15562, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15562(choice, ...) METALANG99_PRIV_REC_NEXT(15563, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15563(choice, ...) METALANG99_PRIV_REC_NEXT(15564, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15564(choice, ...) METALANG99_PRIV_REC_NEXT(15565, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15565(choice, ...) METALANG99_PRIV_REC_NEXT(15566, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15566(choice, ...) METALANG99_PRIV_REC_NEXT(15567, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15567(choice, ...) METALANG99_PRIV_REC_NEXT(15568, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15568(choice, ...) METALANG99_PRIV_REC_NEXT(15569, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15569(choice, ...) METALANG99_PRIV_REC_NEXT(15570, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15570(choice, ...) METALANG99_PRIV_REC_NEXT(15571, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15571(choice, ...) METALANG99_PRIV_REC_NEXT(15572, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15572(choice, ...) METALANG99_PRIV_REC_NEXT(15573, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15573(choice, ...) METALANG99_PRIV_REC_NEXT(15574, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15574(choice, ...) METALANG99_PRIV_REC_NEXT(15575, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15575(choice, ...) METALANG99_PRIV_REC_NEXT(15576, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15576(choice, ...) METALANG99_PRIV_REC_NEXT(15577, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15577(choice, ...) METALANG99_PRIV_REC_NEXT(15578, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15578(choice, ...) METALANG99_PRIV_REC_NEXT(15579, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15579(choice, ...) METALANG99_PRIV_REC_NEXT(15580, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15580(choice, ...) METALANG99_PRIV_REC_NEXT(15581, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15581(choice, ...) METALANG99_PRIV_REC_NEXT(15582, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15582(choice, ...) METALANG99_PRIV_REC_NEXT(15583, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15583(choice, ...) METALANG99_PRIV_REC_NEXT(15584, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15584(choice, ...) METALANG99_PRIV_REC_NEXT(15585, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15585(choice, ...) METALANG99_PRIV_REC_NEXT(15586, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15586(choice, ...) METALANG99_PRIV_REC_NEXT(15587, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15587(choice, ...) METALANG99_PRIV_REC_NEXT(15588, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15588(choice, ...) METALANG99_PRIV_REC_NEXT(15589, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15589(choice, ...) METALANG99_PRIV_REC_NEXT(15590, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15590(choice, ...) METALANG99_PRIV_REC_NEXT(15591, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15591(choice, ...) METALANG99_PRIV_REC_NEXT(15592, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15592(choice, ...) METALANG99_PRIV_REC_NEXT(15593, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15593(choice, ...) METALANG99_PRIV_REC_NEXT(15594, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15594(choice, ...) METALANG99_PRIV_REC_NEXT(15595, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15595(choice, ...) METALANG99_PRIV_REC_NEXT(15596, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15596(choice, ...) METALANG99_PRIV_REC_NEXT(15597, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15597(choice, ...) METALANG99_PRIV_REC_NEXT(15598, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15598(choice, ...) METALANG99_PRIV_REC_NEXT(15599, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15599(choice, ...) METALANG99_PRIV_REC_NEXT(15600, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15600(choice, ...) METALANG99_PRIV_REC_NEXT(15601, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15601(choice, ...) METALANG99_PRIV_REC_NEXT(15602, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15602(choice, ...) METALANG99_PRIV_REC_NEXT(15603, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15603(choice, ...) METALANG99_PRIV_REC_NEXT(15604, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15604(choice, ...) METALANG99_PRIV_REC_NEXT(15605, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15605(choice, ...) METALANG99_PRIV_REC_NEXT(15606, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15606(choice, ...) METALANG99_PRIV_REC_NEXT(15607, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15607(choice, ...) METALANG99_PRIV_REC_NEXT(15608, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15608(choice, ...) METALANG99_PRIV_REC_NEXT(15609, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15609(choice, ...) METALANG99_PRIV_REC_NEXT(15610, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15610(choice, ...) METALANG99_PRIV_REC_NEXT(15611, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15611(choice, ...) METALANG99_PRIV_REC_NEXT(15612, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15612(choice, ...) METALANG99_PRIV_REC_NEXT(15613, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15613(choice, ...) METALANG99_PRIV_REC_NEXT(15614, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15614(choice, ...) METALANG99_PRIV_REC_NEXT(15615, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15615(choice, ...) METALANG99_PRIV_REC_NEXT(15616, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15616(choice, ...) METALANG99_PRIV_REC_NEXT(15617, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15617(choice, ...) METALANG99_PRIV_REC_NEXT(15618, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15618(choice, ...) METALANG99_PRIV_REC_NEXT(15619, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15619(choice, ...) METALANG99_PRIV_REC_NEXT(15620, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15620(choice, ...) METALANG99_PRIV_REC_NEXT(15621, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15621(choice, ...) METALANG99_PRIV_REC_NEXT(15622, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15622(choice, ...) METALANG99_PRIV_REC_NEXT(15623, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15623(choice, ...) METALANG99_PRIV_REC_NEXT(15624, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15624(choice, ...) METALANG99_PRIV_REC_NEXT(15625, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15625(choice, ...) METALANG99_PRIV_REC_NEXT(15626, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15626(choice, ...) METALANG99_PRIV_REC_NEXT(15627, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15627(choice, ...) METALANG99_PRIV_REC_NEXT(15628, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15628(choice, ...) METALANG99_PRIV_REC_NEXT(15629, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15629(choice, ...) METALANG99_PRIV_REC_NEXT(15630, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15630(choice, ...) METALANG99_PRIV_REC_NEXT(15631, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15631(choice, ...) METALANG99_PRIV_REC_NEXT(15632, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15632(choice, ...) METALANG99_PRIV_REC_NEXT(15633, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15633(choice, ...) METALANG99_PRIV_REC_NEXT(15634, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15634(choice, ...) METALANG99_PRIV_REC_NEXT(15635, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15635(choice, ...) METALANG99_PRIV_REC_NEXT(15636, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15636(choice, ...) METALANG99_PRIV_REC_NEXT(15637, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15637(choice, ...) METALANG99_PRIV_REC_NEXT(15638, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15638(choice, ...) METALANG99_PRIV_REC_NEXT(15639, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15639(choice, ...) METALANG99_PRIV_REC_NEXT(15640, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15640(choice, ...) METALANG99_PRIV_REC_NEXT(15641, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15641(choice, ...) METALANG99_PRIV_REC_NEXT(15642, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15642(choice, ...) METALANG99_PRIV_REC_NEXT(15643, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15643(choice, ...) METALANG99_PRIV_REC_NEXT(15644, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15644(choice, ...) METALANG99_PRIV_REC_NEXT(15645, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15645(choice, ...) METALANG99_PRIV_REC_NEXT(15646, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15646(choice, ...) METALANG99_PRIV_REC_NEXT(15647, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15647(choice, ...) METALANG99_PRIV_REC_NEXT(15648, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15648(choice, ...) METALANG99_PRIV_REC_NEXT(15649, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15649(choice, ...) METALANG99_PRIV_REC_NEXT(15650, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15650(choice, ...) METALANG99_PRIV_REC_NEXT(15651, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15651(choice, ...) METALANG99_PRIV_REC_NEXT(15652, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15652(choice, ...) METALANG99_PRIV_REC_NEXT(15653, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15653(choice, ...) METALANG99_PRIV_REC_NEXT(15654, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15654(choice, ...) METALANG99_PRIV_REC_NEXT(15655, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15655(choice, ...) METALANG99_PRIV_REC_NEXT(15656, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15656(choice, ...) METALANG99_PRIV_REC_NEXT(15657, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15657(choice, ...) METALANG99_PRIV_REC_NEXT(15658, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15658(choice, ...) METALANG99_PRIV_REC_NEXT(15659, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15659(choice, ...) METALANG99_PRIV_REC_NEXT(15660, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15660(choice, ...) METALANG99_PRIV_REC_NEXT(15661, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15661(choice, ...) METALANG99_PRIV_REC_NEXT(15662, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15662(choice, ...) METALANG99_PRIV_REC_NEXT(15663, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15663(choice, ...) METALANG99_PRIV_REC_NEXT(15664, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15664(choice, ...) METALANG99_PRIV_REC_NEXT(15665, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15665(choice, ...) METALANG99_PRIV_REC_NEXT(15666, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15666(choice, ...) METALANG99_PRIV_REC_NEXT(15667, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15667(choice, ...) METALANG99_PRIV_REC_NEXT(15668, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15668(choice, ...) METALANG99_PRIV_REC_NEXT(15669, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15669(choice, ...) METALANG99_PRIV_REC_NEXT(15670, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15670(choice, ...) METALANG99_PRIV_REC_NEXT(15671, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15671(choice, ...) METALANG99_PRIV_REC_NEXT(15672, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15672(choice, ...) METALANG99_PRIV_REC_NEXT(15673, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15673(choice, ...) METALANG99_PRIV_REC_NEXT(15674, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15674(choice, ...) METALANG99_PRIV_REC_NEXT(15675, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15675(choice, ...) METALANG99_PRIV_REC_NEXT(15676, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15676(choice, ...) METALANG99_PRIV_REC_NEXT(15677, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15677(choice, ...) METALANG99_PRIV_REC_NEXT(15678, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15678(choice, ...) METALANG99_PRIV_REC_NEXT(15679, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15679(choice, ...) METALANG99_PRIV_REC_NEXT(15680, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15680(choice, ...) METALANG99_PRIV_REC_NEXT(15681, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15681(choice, ...) METALANG99_PRIV_REC_NEXT(15682, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15682(choice, ...) METALANG99_PRIV_REC_NEXT(15683, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15683(choice, ...) METALANG99_PRIV_REC_NEXT(15684, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15684(choice, ...) METALANG99_PRIV_REC_NEXT(15685, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15685(choice, ...) METALANG99_PRIV_REC_NEXT(15686, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15686(choice, ...) METALANG99_PRIV_REC_NEXT(15687, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15687(choice, ...) METALANG99_PRIV_REC_NEXT(15688, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15688(choice, ...) METALANG99_PRIV_REC_NEXT(15689, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15689(choice, ...) METALANG99_PRIV_REC_NEXT(15690, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15690(choice, ...) METALANG99_PRIV_REC_NEXT(15691, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15691(choice, ...) METALANG99_PRIV_REC_NEXT(15692, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15692(choice, ...) METALANG99_PRIV_REC_NEXT(15693, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15693(choice, ...) METALANG99_PRIV_REC_NEXT(15694, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15694(choice, ...) METALANG99_PRIV_REC_NEXT(15695, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15695(choice, ...) METALANG99_PRIV_REC_NEXT(15696, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15696(choice, ...) METALANG99_PRIV_REC_NEXT(15697, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15697(choice, ...) METALANG99_PRIV_REC_NEXT(15698, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15698(choice, ...) METALANG99_PRIV_REC_NEXT(15699, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15699(choice, ...) METALANG99_PRIV_REC_NEXT(15700, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15700(choice, ...) METALANG99_PRIV_REC_NEXT(15701, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15701(choice, ...) METALANG99_PRIV_REC_NEXT(15702, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15702(choice, ...) METALANG99_PRIV_REC_NEXT(15703, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15703(choice, ...) METALANG99_PRIV_REC_NEXT(15704, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15704(choice, ...) METALANG99_PRIV_REC_NEXT(15705, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15705(choice, ...) METALANG99_PRIV_REC_NEXT(15706, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15706(choice, ...) METALANG99_PRIV_REC_NEXT(15707, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15707(choice, ...) METALANG99_PRIV_REC_NEXT(15708, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15708(choice, ...) METALANG99_PRIV_REC_NEXT(15709, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15709(choice, ...) METALANG99_PRIV_REC_NEXT(15710, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15710(choice, ...) METALANG99_PRIV_REC_NEXT(15711, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15711(choice, ...) METALANG99_PRIV_REC_NEXT(15712, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15712(choice, ...) METALANG99_PRIV_REC_NEXT(15713, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15713(choice, ...) METALANG99_PRIV_REC_NEXT(15714, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15714(choice, ...) METALANG99_PRIV_REC_NEXT(15715, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15715(choice, ...) METALANG99_PRIV_REC_NEXT(15716, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15716(choice, ...) METALANG99_PRIV_REC_NEXT(15717, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15717(choice, ...) METALANG99_PRIV_REC_NEXT(15718, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15718(choice, ...) METALANG99_PRIV_REC_NEXT(15719, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15719(choice, ...) METALANG99_PRIV_REC_NEXT(15720, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15720(choice, ...) METALANG99_PRIV_REC_NEXT(15721, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15721(choice, ...) METALANG99_PRIV_REC_NEXT(15722, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15722(choice, ...) METALANG99_PRIV_REC_NEXT(15723, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15723(choice, ...) METALANG99_PRIV_REC_NEXT(15724, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15724(choice, ...) METALANG99_PRIV_REC_NEXT(15725, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15725(choice, ...) METALANG99_PRIV_REC_NEXT(15726, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15726(choice, ...) METALANG99_PRIV_REC_NEXT(15727, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15727(choice, ...) METALANG99_PRIV_REC_NEXT(15728, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15728(choice, ...) METALANG99_PRIV_REC_NEXT(15729, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15729(choice, ...) METALANG99_PRIV_REC_NEXT(15730, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15730(choice, ...) METALANG99_PRIV_REC_NEXT(15731, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15731(choice, ...) METALANG99_PRIV_REC_NEXT(15732, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15732(choice, ...) METALANG99_PRIV_REC_NEXT(15733, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15733(choice, ...) METALANG99_PRIV_REC_NEXT(15734, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15734(choice, ...) METALANG99_PRIV_REC_NEXT(15735, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15735(choice, ...) METALANG99_PRIV_REC_NEXT(15736, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15736(choice, ...) METALANG99_PRIV_REC_NEXT(15737, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15737(choice, ...) METALANG99_PRIV_REC_NEXT(15738, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15738(choice, ...) METALANG99_PRIV_REC_NEXT(15739, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15739(choice, ...) METALANG99_PRIV_REC_NEXT(15740, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15740(choice, ...) METALANG99_PRIV_REC_NEXT(15741, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15741(choice, ...) METALANG99_PRIV_REC_NEXT(15742, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15742(choice, ...) METALANG99_PRIV_REC_NEXT(15743, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15743(choice, ...) METALANG99_PRIV_REC_NEXT(15744, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15744(choice, ...) METALANG99_PRIV_REC_NEXT(15745, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15745(choice, ...) METALANG99_PRIV_REC_NEXT(15746, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15746(choice, ...) METALANG99_PRIV_REC_NEXT(15747, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15747(choice, ...) METALANG99_PRIV_REC_NEXT(15748, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15748(choice, ...) METALANG99_PRIV_REC_NEXT(15749, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15749(choice, ...) METALANG99_PRIV_REC_NEXT(15750, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15750(choice, ...) METALANG99_PRIV_REC_NEXT(15751, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15751(choice, ...) METALANG99_PRIV_REC_NEXT(15752, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15752(choice, ...) METALANG99_PRIV_REC_NEXT(15753, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15753(choice, ...) METALANG99_PRIV_REC_NEXT(15754, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15754(choice, ...) METALANG99_PRIV_REC_NEXT(15755, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15755(choice, ...) METALANG99_PRIV_REC_NEXT(15756, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15756(choice, ...) METALANG99_PRIV_REC_NEXT(15757, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15757(choice, ...) METALANG99_PRIV_REC_NEXT(15758, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15758(choice, ...) METALANG99_PRIV_REC_NEXT(15759, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15759(choice, ...) METALANG99_PRIV_REC_NEXT(15760, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15760(choice, ...) METALANG99_PRIV_REC_NEXT(15761, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15761(choice, ...) METALANG99_PRIV_REC_NEXT(15762, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15762(choice, ...) METALANG99_PRIV_REC_NEXT(15763, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15763(choice, ...) METALANG99_PRIV_REC_NEXT(15764, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15764(choice, ...) METALANG99_PRIV_REC_NEXT(15765, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15765(choice, ...) METALANG99_PRIV_REC_NEXT(15766, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15766(choice, ...) METALANG99_PRIV_REC_NEXT(15767, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15767(choice, ...) METALANG99_PRIV_REC_NEXT(15768, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15768(choice, ...) METALANG99_PRIV_REC_NEXT(15769, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15769(choice, ...) METALANG99_PRIV_REC_NEXT(15770, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15770(choice, ...) METALANG99_PRIV_REC_NEXT(15771, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15771(choice, ...) METALANG99_PRIV_REC_NEXT(15772, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15772(choice, ...) METALANG99_PRIV_REC_NEXT(15773, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15773(choice, ...) METALANG99_PRIV_REC_NEXT(15774, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15774(choice, ...) METALANG99_PRIV_REC_NEXT(15775, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15775(choice, ...) METALANG99_PRIV_REC_NEXT(15776, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15776(choice, ...) METALANG99_PRIV_REC_NEXT(15777, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15777(choice, ...) METALANG99_PRIV_REC_NEXT(15778, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15778(choice, ...) METALANG99_PRIV_REC_NEXT(15779, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15779(choice, ...) METALANG99_PRIV_REC_NEXT(15780, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15780(choice, ...) METALANG99_PRIV_REC_NEXT(15781, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15781(choice, ...) METALANG99_PRIV_REC_NEXT(15782, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15782(choice, ...) METALANG99_PRIV_REC_NEXT(15783, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15783(choice, ...) METALANG99_PRIV_REC_NEXT(15784, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15784(choice, ...) METALANG99_PRIV_REC_NEXT(15785, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15785(choice, ...) METALANG99_PRIV_REC_NEXT(15786, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15786(choice, ...) METALANG99_PRIV_REC_NEXT(15787, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15787(choice, ...) METALANG99_PRIV_REC_NEXT(15788, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15788(choice, ...) METALANG99_PRIV_REC_NEXT(15789, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15789(choice, ...) METALANG99_PRIV_REC_NEXT(15790, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15790(choice, ...) METALANG99_PRIV_REC_NEXT(15791, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15791(choice, ...) METALANG99_PRIV_REC_NEXT(15792, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15792(choice, ...) METALANG99_PRIV_REC_NEXT(15793, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15793(choice, ...) METALANG99_PRIV_REC_NEXT(15794, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15794(choice, ...) METALANG99_PRIV_REC_NEXT(15795, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15795(choice, ...) METALANG99_PRIV_REC_NEXT(15796, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15796(choice, ...) METALANG99_PRIV_REC_NEXT(15797, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15797(choice, ...) METALANG99_PRIV_REC_NEXT(15798, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15798(choice, ...) METALANG99_PRIV_REC_NEXT(15799, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15799(choice, ...) METALANG99_PRIV_REC_NEXT(15800, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15800(choice, ...) METALANG99_PRIV_REC_NEXT(15801, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15801(choice, ...) METALANG99_PRIV_REC_NEXT(15802, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15802(choice, ...) METALANG99_PRIV_REC_NEXT(15803, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15803(choice, ...) METALANG99_PRIV_REC_NEXT(15804, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15804(choice, ...) METALANG99_PRIV_REC_NEXT(15805, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15805(choice, ...) METALANG99_PRIV_REC_NEXT(15806, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15806(choice, ...) METALANG99_PRIV_REC_NEXT(15807, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15807(choice, ...) METALANG99_PRIV_REC_NEXT(15808, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15808(choice, ...) METALANG99_PRIV_REC_NEXT(15809, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15809(choice, ...) METALANG99_PRIV_REC_NEXT(15810, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15810(choice, ...) METALANG99_PRIV_REC_NEXT(15811, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15811(choice, ...) METALANG99_PRIV_REC_NEXT(15812, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15812(choice, ...) METALANG99_PRIV_REC_NEXT(15813, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15813(choice, ...) METALANG99_PRIV_REC_NEXT(15814, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15814(choice, ...) METALANG99_PRIV_REC_NEXT(15815, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15815(choice, ...) METALANG99_PRIV_REC_NEXT(15816, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15816(choice, ...) METALANG99_PRIV_REC_NEXT(15817, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15817(choice, ...) METALANG99_PRIV_REC_NEXT(15818, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15818(choice, ...) METALANG99_PRIV_REC_NEXT(15819, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15819(choice, ...) METALANG99_PRIV_REC_NEXT(15820, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15820(choice, ...) METALANG99_PRIV_REC_NEXT(15821, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15821(choice, ...) METALANG99_PRIV_REC_NEXT(15822, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15822(choice, ...) METALANG99_PRIV_REC_NEXT(15823, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15823(choice, ...) METALANG99_PRIV_REC_NEXT(15824, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15824(choice, ...) METALANG99_PRIV_REC_NEXT(15825, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15825(choice, ...) METALANG99_PRIV_REC_NEXT(15826, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15826(choice, ...) METALANG99_PRIV_REC_NEXT(15827, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15827(choice, ...) METALANG99_PRIV_REC_NEXT(15828, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15828(choice, ...) METALANG99_PRIV_REC_NEXT(15829, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15829(choice, ...) METALANG99_PRIV_REC_NEXT(15830, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15830(choice, ...) METALANG99_PRIV_REC_NEXT(15831, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15831(choice, ...) METALANG99_PRIV_REC_NEXT(15832, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15832(choice, ...) METALANG99_PRIV_REC_NEXT(15833, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15833(choice, ...) METALANG99_PRIV_REC_NEXT(15834, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15834(choice, ...) METALANG99_PRIV_REC_NEXT(15835, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15835(choice, ...) METALANG99_PRIV_REC_NEXT(15836, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15836(choice, ...) METALANG99_PRIV_REC_NEXT(15837, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15837(choice, ...) METALANG99_PRIV_REC_NEXT(15838, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15838(choice, ...) METALANG99_PRIV_REC_NEXT(15839, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15839(choice, ...) METALANG99_PRIV_REC_NEXT(15840, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15840(choice, ...) METALANG99_PRIV_REC_NEXT(15841, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15841(choice, ...) METALANG99_PRIV_REC_NEXT(15842, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15842(choice, ...) METALANG99_PRIV_REC_NEXT(15843, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15843(choice, ...) METALANG99_PRIV_REC_NEXT(15844, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15844(choice, ...) METALANG99_PRIV_REC_NEXT(15845, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15845(choice, ...) METALANG99_PRIV_REC_NEXT(15846, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15846(choice, ...) METALANG99_PRIV_REC_NEXT(15847, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15847(choice, ...) METALANG99_PRIV_REC_NEXT(15848, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15848(choice, ...) METALANG99_PRIV_REC_NEXT(15849, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15849(choice, ...) METALANG99_PRIV_REC_NEXT(15850, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15850(choice, ...) METALANG99_PRIV_REC_NEXT(15851, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15851(choice, ...) METALANG99_PRIV_REC_NEXT(15852, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15852(choice, ...) METALANG99_PRIV_REC_NEXT(15853, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15853(choice, ...) METALANG99_PRIV_REC_NEXT(15854, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15854(choice, ...) METALANG99_PRIV_REC_NEXT(15855, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15855(choice, ...) METALANG99_PRIV_REC_NEXT(15856, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15856(choice, ...) METALANG99_PRIV_REC_NEXT(15857, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15857(choice, ...) METALANG99_PRIV_REC_NEXT(15858, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15858(choice, ...) METALANG99_PRIV_REC_NEXT(15859, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15859(choice, ...) METALANG99_PRIV_REC_NEXT(15860, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15860(choice, ...) METALANG99_PRIV_REC_NEXT(15861, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15861(choice, ...) METALANG99_PRIV_REC_NEXT(15862, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15862(choice, ...) METALANG99_PRIV_REC_NEXT(15863, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15863(choice, ...) METALANG99_PRIV_REC_NEXT(15864, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15864(choice, ...) METALANG99_PRIV_REC_NEXT(15865, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15865(choice, ...) METALANG99_PRIV_REC_NEXT(15866, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15866(choice, ...) METALANG99_PRIV_REC_NEXT(15867, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15867(choice, ...) METALANG99_PRIV_REC_NEXT(15868, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15868(choice, ...) METALANG99_PRIV_REC_NEXT(15869, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15869(choice, ...) METALANG99_PRIV_REC_NEXT(15870, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15870(choice, ...) METALANG99_PRIV_REC_NEXT(15871, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15871(choice, ...) METALANG99_PRIV_REC_NEXT(15872, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15872(choice, ...) METALANG99_PRIV_REC_NEXT(15873, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15873(choice, ...) METALANG99_PRIV_REC_NEXT(15874, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15874(choice, ...) METALANG99_PRIV_REC_NEXT(15875, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15875(choice, ...) METALANG99_PRIV_REC_NEXT(15876, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15876(choice, ...) METALANG99_PRIV_REC_NEXT(15877, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15877(choice, ...) METALANG99_PRIV_REC_NEXT(15878, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15878(choice, ...) METALANG99_PRIV_REC_NEXT(15879, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15879(choice, ...) METALANG99_PRIV_REC_NEXT(15880, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15880(choice, ...) METALANG99_PRIV_REC_NEXT(15881, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15881(choice, ...) METALANG99_PRIV_REC_NEXT(15882, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15882(choice, ...) METALANG99_PRIV_REC_NEXT(15883, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15883(choice, ...) METALANG99_PRIV_REC_NEXT(15884, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15884(choice, ...) METALANG99_PRIV_REC_NEXT(15885, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15885(choice, ...) METALANG99_PRIV_REC_NEXT(15886, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15886(choice, ...) METALANG99_PRIV_REC_NEXT(15887, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15887(choice, ...) METALANG99_PRIV_REC_NEXT(15888, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15888(choice, ...) METALANG99_PRIV_REC_NEXT(15889, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15889(choice, ...) METALANG99_PRIV_REC_NEXT(15890, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15890(choice, ...) METALANG99_PRIV_REC_NEXT(15891, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15891(choice, ...) METALANG99_PRIV_REC_NEXT(15892, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15892(choice, ...) METALANG99_PRIV_REC_NEXT(15893, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15893(choice, ...) METALANG99_PRIV_REC_NEXT(15894, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15894(choice, ...) METALANG99_PRIV_REC_NEXT(15895, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15895(choice, ...) METALANG99_PRIV_REC_NEXT(15896, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15896(choice, ...) METALANG99_PRIV_REC_NEXT(15897, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15897(choice, ...) METALANG99_PRIV_REC_NEXT(15898, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15898(choice, ...) METALANG99_PRIV_REC_NEXT(15899, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15899(choice, ...) METALANG99_PRIV_REC_NEXT(15900, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15900(choice, ...) METALANG99_PRIV_REC_NEXT(15901, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15901(choice, ...) METALANG99_PRIV_REC_NEXT(15902, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15902(choice, ...) METALANG99_PRIV_REC_NEXT(15903, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15903(choice, ...) METALANG99_PRIV_REC_NEXT(15904, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15904(choice, ...) METALANG99_PRIV_REC_NEXT(15905, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15905(choice, ...) METALANG99_PRIV_REC_NEXT(15906, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15906(choice, ...) METALANG99_PRIV_REC_NEXT(15907, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15907(choice, ...) METALANG99_PRIV_REC_NEXT(15908, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15908(choice, ...) METALANG99_PRIV_REC_NEXT(15909, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15909(choice, ...) METALANG99_PRIV_REC_NEXT(15910, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15910(choice, ...) METALANG99_PRIV_REC_NEXT(15911, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15911(choice, ...) METALANG99_PRIV_REC_NEXT(15912, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15912(choice, ...) METALANG99_PRIV_REC_NEXT(15913, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15913(choice, ...) METALANG99_PRIV_REC_NEXT(15914, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15914(choice, ...) METALANG99_PRIV_REC_NEXT(15915, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15915(choice, ...) METALANG99_PRIV_REC_NEXT(15916, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15916(choice, ...) METALANG99_PRIV_REC_NEXT(15917, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15917(choice, ...) METALANG99_PRIV_REC_NEXT(15918, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15918(choice, ...) METALANG99_PRIV_REC_NEXT(15919, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15919(choice, ...) METALANG99_PRIV_REC_NEXT(15920, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15920(choice, ...) METALANG99_PRIV_REC_NEXT(15921, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15921(choice, ...) METALANG99_PRIV_REC_NEXT(15922, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15922(choice, ...) METALANG99_PRIV_REC_NEXT(15923, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15923(choice, ...) METALANG99_PRIV_REC_NEXT(15924, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15924(choice, ...) METALANG99_PRIV_REC_NEXT(15925, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15925(choice, ...) METALANG99_PRIV_REC_NEXT(15926, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15926(choice, ...) METALANG99_PRIV_REC_NEXT(15927, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15927(choice, ...) METALANG99_PRIV_REC_NEXT(15928, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15928(choice, ...) METALANG99_PRIV_REC_NEXT(15929, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15929(choice, ...) METALANG99_PRIV_REC_NEXT(15930, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15930(choice, ...) METALANG99_PRIV_REC_NEXT(15931, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15931(choice, ...) METALANG99_PRIV_REC_NEXT(15932, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15932(choice, ...) METALANG99_PRIV_REC_NEXT(15933, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15933(choice, ...) METALANG99_PRIV_REC_NEXT(15934, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15934(choice, ...) METALANG99_PRIV_REC_NEXT(15935, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15935(choice, ...) METALANG99_PRIV_REC_NEXT(15936, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15936(choice, ...) METALANG99_PRIV_REC_NEXT(15937, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15937(choice, ...) METALANG99_PRIV_REC_NEXT(15938, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15938(choice, ...) METALANG99_PRIV_REC_NEXT(15939, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15939(choice, ...) METALANG99_PRIV_REC_NEXT(15940, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15940(choice, ...) METALANG99_PRIV_REC_NEXT(15941, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15941(choice, ...) METALANG99_PRIV_REC_NEXT(15942, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15942(choice, ...) METALANG99_PRIV_REC_NEXT(15943, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15943(choice, ...) METALANG99_PRIV_REC_NEXT(15944, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15944(choice, ...) METALANG99_PRIV_REC_NEXT(15945, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15945(choice, ...) METALANG99_PRIV_REC_NEXT(15946, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15946(choice, ...) METALANG99_PRIV_REC_NEXT(15947, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15947(choice, ...) METALANG99_PRIV_REC_NEXT(15948, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15948(choice, ...) METALANG99_PRIV_REC_NEXT(15949, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15949(choice, ...) METALANG99_PRIV_REC_NEXT(15950, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15950(choice, ...) METALANG99_PRIV_REC_NEXT(15951, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15951(choice, ...) METALANG99_PRIV_REC_NEXT(15952, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15952(choice, ...) METALANG99_PRIV_REC_NEXT(15953, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15953(choice, ...) METALANG99_PRIV_REC_NEXT(15954, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15954(choice, ...) METALANG99_PRIV_REC_NEXT(15955, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15955(choice, ...) METALANG99_PRIV_REC_NEXT(15956, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15956(choice, ...) METALANG99_PRIV_REC_NEXT(15957, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15957(choice, ...) METALANG99_PRIV_REC_NEXT(15958, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15958(choice, ...) METALANG99_PRIV_REC_NEXT(15959, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15959(choice, ...) METALANG99_PRIV_REC_NEXT(15960, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15960(choice, ...) METALANG99_PRIV_REC_NEXT(15961, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15961(choice, ...) METALANG99_PRIV_REC_NEXT(15962, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15962(choice, ...) METALANG99_PRIV_REC_NEXT(15963, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15963(choice, ...) METALANG99_PRIV_REC_NEXT(15964, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15964(choice, ...) METALANG99_PRIV_REC_NEXT(15965, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15965(choice, ...) METALANG99_PRIV_REC_NEXT(15966, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15966(choice, ...) METALANG99_PRIV_REC_NEXT(15967, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15967(choice, ...) METALANG99_PRIV_REC_NEXT(15968, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15968(choice, ...) METALANG99_PRIV_REC_NEXT(15969, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15969(choice, ...) METALANG99_PRIV_REC_NEXT(15970, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15970(choice, ...) METALANG99_PRIV_REC_NEXT(15971, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15971(choice, ...) METALANG99_PRIV_REC_NEXT(15972, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15972(choice, ...) METALANG99_PRIV_REC_NEXT(15973, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15973(choice, ...) METALANG99_PRIV_REC_NEXT(15974, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15974(choice, ...) METALANG99_PRIV_REC_NEXT(15975, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15975(choice, ...) METALANG99_PRIV_REC_NEXT(15976, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15976(choice, ...) METALANG99_PRIV_REC_NEXT(15977, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15977(choice, ...) METALANG99_PRIV_REC_NEXT(15978, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15978(choice, ...) METALANG99_PRIV_REC_NEXT(15979, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15979(choice, ...) METALANG99_PRIV_REC_NEXT(15980, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15980(choice, ...) METALANG99_PRIV_REC_NEXT(15981, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15981(choice, ...) METALANG99_PRIV_REC_NEXT(15982, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15982(choice, ...) METALANG99_PRIV_REC_NEXT(15983, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15983(choice, ...) METALANG99_PRIV_REC_NEXT(15984, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15984(choice, ...) METALANG99_PRIV_REC_NEXT(15985, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15985(choice, ...) METALANG99_PRIV_REC_NEXT(15986, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15986(choice, ...) METALANG99_PRIV_REC_NEXT(15987, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15987(choice, ...) METALANG99_PRIV_REC_NEXT(15988, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15988(choice, ...) METALANG99_PRIV_REC_NEXT(15989, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15989(choice, ...) METALANG99_PRIV_REC_NEXT(15990, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15990(choice, ...) METALANG99_PRIV_REC_NEXT(15991, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15991(choice, ...) METALANG99_PRIV_REC_NEXT(15992, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15992(choice, ...) METALANG99_PRIV_REC_NEXT(15993, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15993(choice, ...) METALANG99_PRIV_REC_NEXT(15994, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15994(choice, ...) METALANG99_PRIV_REC_NEXT(15995, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15995(choice, ...) METALANG99_PRIV_REC_NEXT(15996, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15996(choice, ...) METALANG99_PRIV_REC_NEXT(15997, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15997(choice, ...) METALANG99_PRIV_REC_NEXT(15998, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15998(choice, ...) METALANG99_PRIV_REC_NEXT(15999, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_15999(choice, ...) METALANG99_PRIV_REC_NEXT(16000, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16000(choice, ...) METALANG99_PRIV_REC_NEXT(16001, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16001(choice, ...) METALANG99_PRIV_REC_NEXT(16002, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16002(choice, ...) METALANG99_PRIV_REC_NEXT(16003, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16003(choice, ...) METALANG99_PRIV_REC_NEXT(16004, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16004(choice, ...) METALANG99_PRIV_REC_NEXT(16005, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16005(choice, ...) METALANG99_PRIV_REC_NEXT(16006, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16006(choice, ...) METALANG99_PRIV_REC_NEXT(16007, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16007(choice, ...) METALANG99_PRIV_REC_NEXT(16008, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16008(choice, ...) METALANG99_PRIV_REC_NEXT(16009, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16009(choice, ...) METALANG99_PRIV_REC_NEXT(16010, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16010(choice, ...) METALANG99_PRIV_REC_NEXT(16011, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16011(choice, ...) METALANG99_PRIV_REC_NEXT(16012, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16012(choice, ...) METALANG99_PRIV_REC_NEXT(16013, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16013(choice, ...) METALANG99_PRIV_REC_NEXT(16014, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16014(choice, ...) METALANG99_PRIV_REC_NEXT(16015, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16015(choice, ...) METALANG99_PRIV_REC_NEXT(16016, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16016(choice, ...) METALANG99_PRIV_REC_NEXT(16017, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16017(choice, ...) METALANG99_PRIV_REC_NEXT(16018, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16018(choice, ...) METALANG99_PRIV_REC_NEXT(16019, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16019(choice, ...) METALANG99_PRIV_REC_NEXT(16020, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16020(choice, ...) METALANG99_PRIV_REC_NEXT(16021, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16021(choice, ...) METALANG99_PRIV_REC_NEXT(16022, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16022(choice, ...) METALANG99_PRIV_REC_NEXT(16023, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16023(choice, ...) METALANG99_PRIV_REC_NEXT(16024, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16024(choice, ...) METALANG99_PRIV_REC_NEXT(16025, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16025(choice, ...) METALANG99_PRIV_REC_NEXT(16026, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16026(choice, ...) METALANG99_PRIV_REC_NEXT(16027, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16027(choice, ...) METALANG99_PRIV_REC_NEXT(16028, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16028(choice, ...) METALANG99_PRIV_REC_NEXT(16029, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16029(choice, ...) METALANG99_PRIV_REC_NEXT(16030, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16030(choice, ...) METALANG99_PRIV_REC_NEXT(16031, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16031(choice, ...) METALANG99_PRIV_REC_NEXT(16032, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16032(choice, ...) METALANG99_PRIV_REC_NEXT(16033, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16033(choice, ...) METALANG99_PRIV_REC_NEXT(16034, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16034(choice, ...) METALANG99_PRIV_REC_NEXT(16035, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16035(choice, ...) METALANG99_PRIV_REC_NEXT(16036, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16036(choice, ...) METALANG99_PRIV_REC_NEXT(16037, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16037(choice, ...) METALANG99_PRIV_REC_NEXT(16038, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16038(choice, ...) METALANG99_PRIV_REC_NEXT(16039, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16039(choice, ...) METALANG99_PRIV_REC_NEXT(16040, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16040(choice, ...) METALANG99_PRIV_REC_NEXT(16041, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16041(choice, ...) METALANG99_PRIV_REC_NEXT(16042, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16042(choice, ...) METALANG99_PRIV_REC_NEXT(16043, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16043(choice, ...) METALANG99_PRIV_REC_NEXT(16044, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16044(choice, ...) METALANG99_PRIV_REC_NEXT(16045, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16045(choice, ...) METALANG99_PRIV_REC_NEXT(16046, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16046(choice, ...) METALANG99_PRIV_REC_NEXT(16047, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16047(choice, ...) METALANG99_PRIV_REC_NEXT(16048, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16048(choice, ...) METALANG99_PRIV_REC_NEXT(16049, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16049(choice, ...) METALANG99_PRIV_REC_NEXT(16050, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16050(choice, ...) METALANG99_PRIV_REC_NEXT(16051, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16051(choice, ...) METALANG99_PRIV_REC_NEXT(16052, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16052(choice, ...) METALANG99_PRIV_REC_NEXT(16053, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16053(choice, ...) METALANG99_PRIV_REC_NEXT(16054, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16054(choice, ...) METALANG99_PRIV_REC_NEXT(16055, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16055(choice, ...) METALANG99_PRIV_REC_NEXT(16056, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16056(choice, ...) METALANG99_PRIV_REC_NEXT(16057, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16057(choice, ...) METALANG99_PRIV_REC_NEXT(16058, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16058(choice, ...) METALANG99_PRIV_REC_NEXT(16059, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16059(choice, ...) METALANG99_PRIV_REC_NEXT(16060, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16060(choice, ...) METALANG99_PRIV_REC_NEXT(16061, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16061(choice, ...) METALANG99_PRIV_REC_NEXT(16062, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16062(choice, ...) METALANG99_PRIV_REC_NEXT(16063, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16063(choice, ...) METALANG99_PRIV_REC_NEXT(16064, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16064(choice, ...) METALANG99_PRIV_REC_NEXT(16065, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16065(choice, ...) METALANG99_PRIV_REC_NEXT(16066, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16066(choice, ...) METALANG99_PRIV_REC_NEXT(16067, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16067(choice, ...) METALANG99_PRIV_REC_NEXT(16068, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16068(choice, ...) METALANG99_PRIV_REC_NEXT(16069, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16069(choice, ...) METALANG99_PRIV_REC_NEXT(16070, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16070(choice, ...) METALANG99_PRIV_REC_NEXT(16071, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16071(choice, ...) METALANG99_PRIV_REC_NEXT(16072, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16072(choice, ...) METALANG99_PRIV_REC_NEXT(16073, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16073(choice, ...) METALANG99_PRIV_REC_NEXT(16074, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16074(choice, ...) METALANG99_PRIV_REC_NEXT(16075, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16075(choice, ...) METALANG99_PRIV_REC_NEXT(16076, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16076(choice, ...) METALANG99_PRIV_REC_NEXT(16077, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16077(choice, ...) METALANG99_PRIV_REC_NEXT(16078, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16078(choice, ...) METALANG99_PRIV_REC_NEXT(16079, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16079(choice, ...) METALANG99_PRIV_REC_NEXT(16080, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16080(choice, ...) METALANG99_PRIV_REC_NEXT(16081, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16081(choice, ...) METALANG99_PRIV_REC_NEXT(16082, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16082(choice, ...) METALANG99_PRIV_REC_NEXT(16083, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16083(choice, ...) METALANG99_PRIV_REC_NEXT(16084, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16084(choice, ...) METALANG99_PRIV_REC_NEXT(16085, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16085(choice, ...) METALANG99_PRIV_REC_NEXT(16086, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16086(choice, ...) METALANG99_PRIV_REC_NEXT(16087, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16087(choice, ...) METALANG99_PRIV_REC_NEXT(16088, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16088(choice, ...) METALANG99_PRIV_REC_NEXT(16089, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16089(choice, ...) METALANG99_PRIV_REC_NEXT(16090, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16090(choice, ...) METALANG99_PRIV_REC_NEXT(16091, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16091(choice, ...) METALANG99_PRIV_REC_NEXT(16092, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16092(choice, ...) METALANG99_PRIV_REC_NEXT(16093, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16093(choice, ...) METALANG99_PRIV_REC_NEXT(16094, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16094(choice, ...) METALANG99_PRIV_REC_NEXT(16095, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16095(choice, ...) METALANG99_PRIV_REC_NEXT(16096, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16096(choice, ...) METALANG99_PRIV_REC_NEXT(16097, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16097(choice, ...) METALANG99_PRIV_REC_NEXT(16098, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16098(choice, ...) METALANG99_PRIV_REC_NEXT(16099, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16099(choice, ...) METALANG99_PRIV_REC_NEXT(16100, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16100(choice, ...) METALANG99_PRIV_REC_NEXT(16101, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16101(choice, ...) METALANG99_PRIV_REC_NEXT(16102, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16102(choice, ...) METALANG99_PRIV_REC_NEXT(16103, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16103(choice, ...) METALANG99_PRIV_REC_NEXT(16104, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16104(choice, ...) METALANG99_PRIV_REC_NEXT(16105, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16105(choice, ...) METALANG99_PRIV_REC_NEXT(16106, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16106(choice, ...) METALANG99_PRIV_REC_NEXT(16107, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16107(choice, ...) METALANG99_PRIV_REC_NEXT(16108, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16108(choice, ...) METALANG99_PRIV_REC_NEXT(16109, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16109(choice, ...) METALANG99_PRIV_REC_NEXT(16110, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16110(choice, ...) METALANG99_PRIV_REC_NEXT(16111, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16111(choice, ...) METALANG99_PRIV_REC_NEXT(16112, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16112(choice, ...) METALANG99_PRIV_REC_NEXT(16113, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16113(choice, ...) METALANG99_PRIV_REC_NEXT(16114, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16114(choice, ...) METALANG99_PRIV_REC_NEXT(16115, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16115(choice, ...) METALANG99_PRIV_REC_NEXT(16116, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16116(choice, ...) METALANG99_PRIV_REC_NEXT(16117, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16117(choice, ...) METALANG99_PRIV_REC_NEXT(16118, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16118(choice, ...) METALANG99_PRIV_REC_NEXT(16119, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16119(choice, ...) METALANG99_PRIV_REC_NEXT(16120, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16120(choice, ...) METALANG99_PRIV_REC_NEXT(16121, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16121(choice, ...) METALANG99_PRIV_REC_NEXT(16122, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16122(choice, ...) METALANG99_PRIV_REC_NEXT(16123, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16123(choice, ...) METALANG99_PRIV_REC_NEXT(16124, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16124(choice, ...) METALANG99_PRIV_REC_NEXT(16125, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16125(choice, ...) METALANG99_PRIV_REC_NEXT(16126, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16126(choice, ...) METALANG99_PRIV_REC_NEXT(16127, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16127(choice, ...) METALANG99_PRIV_REC_NEXT(16128, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16128(choice, ...) METALANG99_PRIV_REC_NEXT(16129, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16129(choice, ...) METALANG99_PRIV_REC_NEXT(16130, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16130(choice, ...) METALANG99_PRIV_REC_NEXT(16131, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16131(choice, ...) METALANG99_PRIV_REC_NEXT(16132, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16132(choice, ...) METALANG99_PRIV_REC_NEXT(16133, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16133(choice, ...) METALANG99_PRIV_REC_NEXT(16134, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16134(choice, ...) METALANG99_PRIV_REC_NEXT(16135, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16135(choice, ...) METALANG99_PRIV_REC_NEXT(16136, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16136(choice, ...) METALANG99_PRIV_REC_NEXT(16137, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16137(choice, ...) METALANG99_PRIV_REC_NEXT(16138, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16138(choice, ...) METALANG99_PRIV_REC_NEXT(16139, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16139(choice, ...) METALANG99_PRIV_REC_NEXT(16140, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16140(choice, ...) METALANG99_PRIV_REC_NEXT(16141, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16141(choice, ...) METALANG99_PRIV_REC_NEXT(16142, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16142(choice, ...) METALANG99_PRIV_REC_NEXT(16143, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16143(choice, ...) METALANG99_PRIV_REC_NEXT(16144, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16144(choice, ...) METALANG99_PRIV_REC_NEXT(16145, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16145(choice, ...) METALANG99_PRIV_REC_NEXT(16146, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16146(choice, ...) METALANG99_PRIV_REC_NEXT(16147, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16147(choice, ...) METALANG99_PRIV_REC_NEXT(16148, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16148(choice, ...) METALANG99_PRIV_REC_NEXT(16149, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16149(choice, ...) METALANG99_PRIV_REC_NEXT(16150, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16150(choice, ...) METALANG99_PRIV_REC_NEXT(16151, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16151(choice, ...) METALANG99_PRIV_REC_NEXT(16152, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16152(choice, ...) METALANG99_PRIV_REC_NEXT(16153, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16153(choice, ...) METALANG99_PRIV_REC_NEXT(16154, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16154(choice, ...) METALANG99_PRIV_REC_NEXT(16155, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16155(choice, ...) METALANG99_PRIV_REC_NEXT(16156, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16156(choice, ...) METALANG99_PRIV_REC_NEXT(16157, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16157(choice, ...) METALANG99_PRIV_REC_NEXT(16158, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16158(choice, ...) METALANG99_PRIV_REC_NEXT(16159, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16159(choice, ...) METALANG99_PRIV_REC_NEXT(16160, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16160(choice, ...) METALANG99_PRIV_REC_NEXT(16161, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16161(choice, ...) METALANG99_PRIV_REC_NEXT(16162, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16162(choice, ...) METALANG99_PRIV_REC_NEXT(16163, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16163(choice, ...) METALANG99_PRIV_REC_NEXT(16164, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16164(choice, ...) METALANG99_PRIV_REC_NEXT(16165, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16165(choice, ...) METALANG99_PRIV_REC_NEXT(16166, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16166(choice, ...) METALANG99_PRIV_REC_NEXT(16167, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16167(choice, ...) METALANG99_PRIV_REC_NEXT(16168, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16168(choice, ...) METALANG99_PRIV_REC_NEXT(16169, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16169(choice, ...) METALANG99_PRIV_REC_NEXT(16170, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16170(choice, ...) METALANG99_PRIV_REC_NEXT(16171, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16171(choice, ...) METALANG99_PRIV_REC_NEXT(16172, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16172(choice, ...) METALANG99_PRIV_REC_NEXT(16173, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16173(choice, ...) METALANG99_PRIV_REC_NEXT(16174, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16174(choice, ...) METALANG99_PRIV_REC_NEXT(16175, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16175(choice, ...) METALANG99_PRIV_REC_NEXT(16176, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16176(choice, ...) METALANG99_PRIV_REC_NEXT(16177, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16177(choice, ...) METALANG99_PRIV_REC_NEXT(16178, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16178(choice, ...) METALANG99_PRIV_REC_NEXT(16179, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16179(choice, ...) METALANG99_PRIV_REC_NEXT(16180, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16180(choice, ...) METALANG99_PRIV_REC_NEXT(16181, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16181(choice, ...) METALANG99_PRIV_REC_NEXT(16182, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16182(choice, ...) METALANG99_PRIV_REC_NEXT(16183, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16183(choice, ...) METALANG99_PRIV_REC_NEXT(16184, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16184(choice, ...) METALANG99_PRIV_REC_NEXT(16185, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16185(choice, ...) METALANG99_PRIV_REC_NEXT(16186, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16186(choice, ...) METALANG99_PRIV_REC_NEXT(16187, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16187(choice, ...) METALANG99_PRIV_REC_NEXT(16188, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16188(choice, ...) METALANG99_PRIV_REC_NEXT(16189, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16189(choice, ...) METALANG99_PRIV_REC_NEXT(16190, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16190(choice, ...) METALANG99_PRIV_REC_NEXT(16191, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16191(choice, ...) METALANG99_PRIV_REC_NEXT(16192, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16192(choice, ...) METALANG99_PRIV_REC_NEXT(16193, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16193(choice, ...) METALANG99_PRIV_REC_NEXT(16194, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16194(choice, ...) METALANG99_PRIV_REC_NEXT(16195, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16195(choice, ...) METALANG99_PRIV_REC_NEXT(16196, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16196(choice, ...) METALANG99_PRIV_REC_NEXT(16197, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16197(choice, ...) METALANG99_PRIV_REC_NEXT(16198, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16198(choice, ...) METALANG99_PRIV_REC_NEXT(16199, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16199(choice, ...) METALANG99_PRIV_REC_NEXT(16200, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16200(choice, ...) METALANG99_PRIV_REC_NEXT(16201, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16201(choice, ...) METALANG99_PRIV_REC_NEXT(16202, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16202(choice, ...) METALANG99_PRIV_REC_NEXT(16203, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16203(choice, ...) METALANG99_PRIV_REC_NEXT(16204, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16204(choice, ...) METALANG99_PRIV_REC_NEXT(16205, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16205(choice, ...) METALANG99_PRIV_REC_NEXT(16206, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16206(choice, ...) METALANG99_PRIV_REC_NEXT(16207, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16207(choice, ...) METALANG99_PRIV_REC_NEXT(16208, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16208(choice, ...) METALANG99_PRIV_REC_NEXT(16209, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16209(choice, ...) METALANG99_PRIV_REC_NEXT(16210, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16210(choice, ...) METALANG99_PRIV_REC_NEXT(16211, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16211(choice, ...) METALANG99_PRIV_REC_NEXT(16212, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16212(choice, ...) METALANG99_PRIV_REC_NEXT(16213, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16213(choice, ...) METALANG99_PRIV_REC_NEXT(16214, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16214(choice, ...) METALANG99_PRIV_REC_NEXT(16215, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16215(choice, ...) METALANG99_PRIV_REC_NEXT(16216, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16216(choice, ...) METALANG99_PRIV_REC_NEXT(16217, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16217(choice, ...) METALANG99_PRIV_REC_NEXT(16218, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16218(choice, ...) METALANG99_PRIV_REC_NEXT(16219, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16219(choice, ...) METALANG99_PRIV_REC_NEXT(16220, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16220(choice, ...) METALANG99_PRIV_REC_NEXT(16221, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16221(choice, ...) METALANG99_PRIV_REC_NEXT(16222, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16222(choice, ...) METALANG99_PRIV_REC_NEXT(16223, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16223(choice, ...) METALANG99_PRIV_REC_NEXT(16224, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16224(choice, ...) METALANG99_PRIV_REC_NEXT(16225, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16225(choice, ...) METALANG99_PRIV_REC_NEXT(16226, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16226(choice, ...) METALANG99_PRIV_REC_NEXT(16227, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16227(choice, ...) METALANG99_PRIV_REC_NEXT(16228, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16228(choice, ...) METALANG99_PRIV_REC_NEXT(16229, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16229(choice, ...) METALANG99_PRIV_REC_NEXT(16230, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16230(choice, ...) METALANG99_PRIV_REC_NEXT(16231, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16231(choice, ...) METALANG99_PRIV_REC_NEXT(16232, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16232(choice, ...) METALANG99_PRIV_REC_NEXT(16233, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16233(choice, ...) METALANG99_PRIV_REC_NEXT(16234, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16234(choice, ...) METALANG99_PRIV_REC_NEXT(16235, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16235(choice, ...) METALANG99_PRIV_REC_NEXT(16236, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16236(choice, ...) METALANG99_PRIV_REC_NEXT(16237, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16237(choice, ...) METALANG99_PRIV_REC_NEXT(16238, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16238(choice, ...) METALANG99_PRIV_REC_NEXT(16239, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16239(choice, ...) METALANG99_PRIV_REC_NEXT(16240, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16240(choice, ...) METALANG99_PRIV_REC_NEXT(16241, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16241(choice, ...) METALANG99_PRIV_REC_NEXT(16242, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16242(choice, ...) METALANG99_PRIV_REC_NEXT(16243, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16243(choice, ...) METALANG99_PRIV_REC_NEXT(16244, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16244(choice, ...) METALANG99_PRIV_REC_NEXT(16245, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16245(choice, ...) METALANG99_PRIV_REC_NEXT(16246, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16246(choice, ...) METALANG99_PRIV_REC_NEXT(16247, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16247(choice, ...) METALANG99_PRIV_REC_NEXT(16248, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16248(choice, ...) METALANG99_PRIV_REC_NEXT(16249, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16249(choice, ...) METALANG99_PRIV_REC_NEXT(16250, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16250(choice, ...) METALANG99_PRIV_REC_NEXT(16251, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16251(choice, ...) METALANG99_PRIV_REC_NEXT(16252, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16252(choice, ...) METALANG99_PRIV_REC_NEXT(16253, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16253(choice, ...) METALANG99_PRIV_REC_NEXT(16254, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16254(choice, ...) METALANG99_PRIV_REC_NEXT(16255, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16255(choice, ...) METALANG99_PRIV_REC_NEXT(16256, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16256(choice, ...) METALANG99_PRIV_REC_NEXT(16257, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16257(choice, ...) METALANG99_PRIV_REC_NEXT(16258, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16258(choice, ...) METALANG99_PRIV_REC_NEXT(16259, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16259(choice, ...) METALANG99_PRIV_REC_NEXT(16260, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16260(choice, ...) METALANG99_PRIV_REC_NEXT(16261, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16261(choice, ...) METALANG99_PRIV_REC_NEXT(16262, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16262(choice, ...) METALANG99_PRIV_REC_NEXT(16263, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16263(choice, ...) METALANG99_PRIV_REC_NEXT(16264, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16264(choice, ...) METALANG99_PRIV_REC_NEXT(16265, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16265(choice, ...) METALANG99_PRIV_REC_NEXT(16266, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16266(choice, ...) METALANG99_PRIV_REC_NEXT(16267, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16267(choice, ...) METALANG99_PRIV_REC_NEXT(16268, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16268(choice, ...) METALANG99_PRIV_REC_NEXT(16269, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16269(choice, ...) METALANG99_PRIV_REC_NEXT(16270, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16270(choice, ...) METALANG99_PRIV_REC_NEXT(16271, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16271(choice, ...) METALANG99_PRIV_REC_NEXT(16272, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16272(choice, ...) METALANG99_PRIV_REC_NEXT(16273, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16273(choice, ...) METALANG99_PRIV_REC_NEXT(16274, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16274(choice, ...) METALANG99_PRIV_REC_NEXT(16275, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16275(choice, ...) METALANG99_PRIV_REC_NEXT(16276, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16276(choice, ...) METALANG99_PRIV_REC_NEXT(16277, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16277(choice, ...) METALANG99_PRIV_REC_NEXT(16278, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16278(choice, ...) METALANG99_PRIV_REC_NEXT(16279, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16279(choice, ...) METALANG99_PRIV_REC_NEXT(16280, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16280(choice, ...) METALANG99_PRIV_REC_NEXT(16281, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16281(choice, ...) METALANG99_PRIV_REC_NEXT(16282, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16282(choice, ...) METALANG99_PRIV_REC_NEXT(16283, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16283(choice, ...) METALANG99_PRIV_REC_NEXT(16284, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16284(choice, ...) METALANG99_PRIV_REC_NEXT(16285, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16285(choice, ...) METALANG99_PRIV_REC_NEXT(16286, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16286(choice, ...) METALANG99_PRIV_REC_NEXT(16287, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16287(choice, ...) METALANG99_PRIV_REC_NEXT(16288, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16288(choice, ...) METALANG99_PRIV_REC_NEXT(16289, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16289(choice, ...) METALANG99_PRIV_REC_NEXT(16290, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16290(choice, ...) METALANG99_PRIV_REC_NEXT(16291, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16291(choice, ...) METALANG99_PRIV_REC_NEXT(16292, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16292(choice, ...) METALANG99_PRIV_REC_NEXT(16293, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16293(choice, ...) METALANG99_PRIV_REC_NEXT(16294, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16294(choice, ...) METALANG99_PRIV_REC_NEXT(16295, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16295(choice, ...) METALANG99_PRIV_REC_NEXT(16296, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16296(choice, ...) METALANG99_PRIV_REC_NEXT(16297, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16297(choice, ...) METALANG99_PRIV_REC_NEXT(16298, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16298(choice, ...) METALANG99_PRIV_REC_NEXT(16299, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16299(choice, ...) METALANG99_PRIV_REC_NEXT(16300, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16300(choice, ...) METALANG99_PRIV_REC_NEXT(16301, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16301(choice, ...) METALANG99_PRIV_REC_NEXT(16302, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16302(choice, ...) METALANG99_PRIV_REC_NEXT(16303, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16303(choice, ...) METALANG99_PRIV_REC_NEXT(16304, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16304(choice, ...) METALANG99_PRIV_REC_NEXT(16305, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16305(choice, ...) METALANG99_PRIV_REC_NEXT(16306, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16306(choice, ...) METALANG99_PRIV_REC_NEXT(16307, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16307(choice, ...) METALANG99_PRIV_REC_NEXT(16308, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16308(choice, ...) METALANG99_PRIV_REC_NEXT(16309, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16309(choice, ...) METALANG99_PRIV_REC_NEXT(16310, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16310(choice, ...) METALANG99_PRIV_REC_NEXT(16311, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16311(choice, ...) METALANG99_PRIV_REC_NEXT(16312, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16312(choice, ...) METALANG99_PRIV_REC_NEXT(16313, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16313(choice, ...) METALANG99_PRIV_REC_NEXT(16314, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16314(choice, ...) METALANG99_PRIV_REC_NEXT(16315, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16315(choice, ...) METALANG99_PRIV_REC_NEXT(16316, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16316(choice, ...) METALANG99_PRIV_REC_NEXT(16317, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16317(choice, ...) METALANG99_PRIV_REC_NEXT(16318, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16318(choice, ...) METALANG99_PRIV_REC_NEXT(16319, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16319(choice, ...) METALANG99_PRIV_REC_NEXT(16320, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16320(choice, ...) METALANG99_PRIV_REC_NEXT(16321, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16321(choice, ...) METALANG99_PRIV_REC_NEXT(16322, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16322(choice, ...) METALANG99_PRIV_REC_NEXT(16323, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16323(choice, ...) METALANG99_PRIV_REC_NEXT(16324, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16324(choice, ...) METALANG99_PRIV_REC_NEXT(16325, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16325(choice, ...) METALANG99_PRIV_REC_NEXT(16326, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16326(choice, ...) METALANG99_PRIV_REC_NEXT(16327, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16327(choice, ...) METALANG99_PRIV_REC_NEXT(16328, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16328(choice, ...) METALANG99_PRIV_REC_NEXT(16329, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16329(choice, ...) METALANG99_PRIV_REC_NEXT(16330, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16330(choice, ...) METALANG99_PRIV_REC_NEXT(16331, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16331(choice, ...) METALANG99_PRIV_REC_NEXT(16332, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16332(choice, ...) METALANG99_PRIV_REC_NEXT(16333, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16333(choice, ...) METALANG99_PRIV_REC_NEXT(16334, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16334(choice, ...) METALANG99_PRIV_REC_NEXT(16335, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16335(choice, ...) METALANG99_PRIV_REC_NEXT(16336, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16336(choice, ...) METALANG99_PRIV_REC_NEXT(16337, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16337(choice, ...) METALANG99_PRIV_REC_NEXT(16338, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16338(choice, ...) METALANG99_PRIV_REC_NEXT(16339, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16339(choice, ...) METALANG99_PRIV_REC_NEXT(16340, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16340(choice, ...) METALANG99_PRIV_REC_NEXT(16341, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16341(choice, ...) METALANG99_PRIV_REC_NEXT(16342, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16342(choice, ...) METALANG99_PRIV_REC_NEXT(16343, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16343(choice, ...) METALANG99_PRIV_REC_NEXT(16344, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16344(choice, ...) METALANG99_PRIV_REC_NEXT(16345, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16345(choice, ...) METALANG99_PRIV_REC_NEXT(16346, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16346(choice, ...) METALANG99_PRIV_REC_NEXT(16347, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16347(choice, ...) METALANG99_PRIV_REC_NEXT(16348, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16348(choice, ...) METALANG99_PRIV_REC_NEXT(16349, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16349(choice, ...) METALANG99_PRIV_REC_NEXT(16350, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16350(choice, ...) METALANG99_PRIV_REC_NEXT(16351, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16351(choice, ...) METALANG99_PRIV_REC_NEXT(16352, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16352(choice, ...) METALANG99_PRIV_REC_NEXT(16353, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16353(choice, ...) METALANG99_PRIV_REC_NEXT(16354, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16354(choice, ...) METALANG99_PRIV_REC_NEXT(16355, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16355(choice, ...) METALANG99_PRIV_REC_NEXT(16356, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16356(choice, ...) METALANG99_PRIV_REC_NEXT(16357, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16357(choice, ...) METALANG99_PRIV_REC_NEXT(16358, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16358(choice, ...) METALANG99_PRIV_REC_NEXT(16359, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16359(choice, ...) METALANG99_PRIV_REC_NEXT(16360, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16360(choice, ...) METALANG99_PRIV_REC_NEXT(16361, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16361(choice, ...) METALANG99_PRIV_REC_NEXT(16362, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16362(choice, ...) METALANG99_PRIV_REC_NEXT(16363, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16363(choice, ...) METALANG99_PRIV_REC_NEXT(16364, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16364(choice, ...) METALANG99_PRIV_REC_NEXT(16365, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16365(choice, ...) METALANG99_PRIV_REC_NEXT(16366, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16366(choice, ...) METALANG99_PRIV_REC_NEXT(16367, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16367(choice, ...) METALANG99_PRIV_REC_NEXT(16368, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16368(choice, ...) METALANG99_PRIV_REC_NEXT(16369, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16369(choice, ...) METALANG99_PRIV_REC_NEXT(16370, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16370(choice, ...) METALANG99_PRIV_REC_NEXT(16371, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16371(choice, ...) METALANG99_PRIV_REC_NEXT(16372, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16372(choice, ...) METALANG99_PRIV_REC_NEXT(16373, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16373(choice, ...) METALANG99_PRIV_REC_NEXT(16374, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16374(choice, ...) METALANG99_PRIV_REC_NEXT(16375, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16375(choice, ...) METALANG99_PRIV_REC_NEXT(16376, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16376(choice, ...) METALANG99_PRIV_REC_NEXT(16377, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16377(choice, ...) METALANG99_PRIV_REC_NEXT(16378, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16378(choice, ...) METALANG99_PRIV_REC_NEXT(16379, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16379(choice, ...) METALANG99_PRIV_REC_NEXT(16380, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16380(choice, ...) METALANG99_PRIV_REC_NEXT(16381, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16381(choice, ...) METALANG99_PRIV_REC_NEXT(16382, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16382(choice, ...) METALANG99_PRIV_REC_NEXT(16383, choice)(__VA_ARGS__)
#define METALANG99_PRIV_REC_16383(choice, ...) METALANG99_PRIV_REC_NEXT(16384, choice)(__VA_ARGS__)

#endif // METALANG99_EVAL_REC_UNROLL_H
